#!/usr/bin/env python

"""
Class to wrap the Automatic Antenna Integration

$Id $ v1.0 - smartin 09/2015
$Id $ v2.0 - smartin 11/2015 First commit
$Id $ v2.1 - debugging of paths
$Id $ v3.0 - merging baseline and pointing using the same observation
$Id $ v3.1 - adding 3 repetitions to the tiltmeter runs.
$Id $ v4.0 - smartin 11/2016 updated version for AOS-64. New communication path among other changes
$Id $ v4.1 - smartin 11/2016 fixed some bugs regarding the checking of .casa/init.py file
$Id $ v4.2 - smartin 12/2016 changed communication directory to AIV for operators to be able to access...
$Id $ v4.3 - smartin 01/2016 Automatic update of AN0,AW0, new path/pythonpath check allowing only default if not in -t
"""

import os
import sys
import subprocess
import time
import getpass
import signal
from optparse import OptionParser, SUPPRESS_HELP

__author__ = 'smartin'


# Logger to direct output to both terminal and log file
class AntennaIntegratorLogger(object):
    def __init__(self, filedescriptor=''):
        self.terminal = sys.stdout
        self.log = open(time.strftime("%Y%m%d_%H:%M_"+filedescriptor+".log"), "a")

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)


class AntennaIntegrator:

    def __init__(self, arrayname='', antennalist='', step='', simulation=False, datareduction=False, testing=False):
        # Defines the system standard output and error to the defined Logger
        sys.stdout = AntennaIntegratorLogger('AntennaIntegration')
        sys.stderr = AntennaIntegratorLogger('AntennaIntegration')
        # Initializes input parameter
        self.arrayName = arrayname
        self.antennaListToIntegrate = antennalist
        # Initializing the steps dictionary
        self.aisteps = {
            '1': "Initial Check",
            '2': "Pad Tilt Terms",
            '3': "Initial IA offset",
            '4': "Check Delays",
            '5': "Baseline",
            '6': "Pointing Model"
        }
        self.startingstep = step
        self.joinbaselinepointingobservation = True
        # Initializes server variables
        self.justsimulate = simulation  # Variable for debugging with no connection to server
        self.justreduce = datareduction  # Variable to indicate that only data reduction will be needed
        self.testing = testing # Variable for activating testing mode
        # if in reduction mode, it will also activate simulation mode
        if self.justreduce:
            self.justsimulate = True
        # file to transfer information with aos-gns
        # old path self.commandsfilepath = '/mnt/nas-red/storage1/data2/data'
        # old path self.commandsfilepathatAOS q= '/mnt/nas-red.osf.alma.cl/storage1/data2/data'
        self.defaultAIVpath = '/groups/science/AIV'
        self.commandsfilepath = '/groups/science/data/AAI'
        self.commandsfilepathatAOS = '/mnt/nas-05.osf.alma.cl/groups/science/data/AAI'
        self.commandsfile = '%s/AntennaIntegrationCommunication.txt' % self.commandsfilepath
        # and the output from AOS executions
        self.aos_execution_output_file = '%s/AntennaIntegratorAOS.output' % self.commandsfilepath
        self.last_aos_execution_output = ''
        # If communication file exists we delete it to start from scratch
        if os.path.isfile(self.commandsfile):
            subprocess.call(['rm', '%s' % self.commandsfile])
        self.comandsfile_modifiedtime = 0
        self.exec_uid = ''
        self.asdm_uid = ''
        # Check basic paths and directories
        self.analysis_scripts_path = ''
        self.analysis_R104working = ''
        self.check_paths()
        self.check_directory()

    def cancel_execution(self, signum, frame):
        signal.signal(signal.SIGINT, original_sigint)
        try:
            if raw_input(" \n Quit the AUTOMATED ANTENNA INTEGRATION [y/n]").lower().startswith('y'):
                self.connection_file_to_aosgns(disconnect=True)
                sys.exit(1)
        except KeyboardInterrupt:
            print "\n OK, OK, QUITTING AAI"
            self.connection_file_to_aosgns(disconnect=True)
            sys.exit(1)

    def check_paths(self):
        is_path_modified = False
        original_path = os.environ['PATH']
        print "\033[31mChecking environmental variables and CASA init.py file\033[39m"

        #
        # ANALYSIS SCRIPTS
        #
        # Check the environmental PATH includes the analysis_scripts directory
        # Needed to call other required scripts from that directory
        ospath = str(os.environ['PATH'])
        if ospath.__contains__('analysis_scripts'):
            print 'PATH contains analysis_scripts'
            self.analysis_scripts_path = [s for s in ospath.split(os.pathsep) if 'analysis_scripts' in s][0]
        else:
            print 'System PATH should contain %s/analysis_scripts' % self.defaultAIVpath
            self.analysis_scripts_path = '%s/analysis_scripts' % self.defaultAIVpath
            print 'Adding default path %s to environmental variable PATH' % self.analysis_scripts_path
            os.environ['PATH'] += os.pathsep + self.analysis_scripts_path
            is_path_modified = True
        # We will allow the use of alternative repositories only with the testing option
        if self.analysis_scripts_path != '%s/analysis_scripts' % self.defaultAIVpath:
            if self.testing:
                #  We make sure that files in there are executable.
                print "Making sure files in %s have executable rights" % self.analysis_scripts_path
                subprocess.call('chmod +x %s/*' % self.analysis_scripts_path, shell=True)
            else:
                print "ERROR: analysis_scripts path %s " \
                      "is different from the default path %s/analysis_scripts." \
                      % (self.analysis_scripts_path, self.defaultAIVpath)
                print "USE THE OPTION -t AND RUN AT YOUR OWN RISK"
                sys.exit()
        # Check for existence of the directory
        if not os.path.isdir(self.analysis_scripts_path):
            while not os.path.isdir(self.analysis_scripts_path):
                print 'ERROR: %s does not exist.' % self.analysis_scripts_path
                self.analysis_scripts_path = raw_input('Please enter the full path of the analysis_scripts: ')
            print 'Adding %s to environmental variable PATH' % self.analysis_scripts_path
            os.environ['PATH'] += os.pathsep + self.analysis_scripts_path
            is_path_modified = True

        #
        #  R10.4_WORKING
        #
        # Check the environmental PATH includes the R10.4_WORKING directory
        # Needed to call other required scripts from that directory
        if ospath.__contains__('scripts/R10.4_WORKING'):
            print 'PATH contains scripts/R10.4_WORKING'
            self.analysis_R104working = [s for s in ospath.split(os.pathsep) if 'R10.4_WORKING' in s][0]
        else:
            print 'System PATH should contain %s/scripts/R10.4_WORKING' % self.defaultAIVpath
            self.analysis_R104working = '%s/scripts/R10.4_WORKING' % self.defaultAIVpath
            print 'Adding default path %s to environmental variable PATH' % self.analysis_R104working
            os.environ['PATH'] += os.pathsep + self.analysis_R104working
            is_path_modified = True
        # We will allow the use of alternative repositories only with the testing option
        if self.analysis_R104working != '%s/scripts/R10.4_WORKING' % self.defaultAIVpath:
            if self.testing:
                #  We make sure that files in there are executable.
                print "Making sure files in %s have executable rights" % self.analysis_R104working
                subprocess.call('chmod +x %s/*' % self.analysis_R104working, shell=True)
            else:
                print "ERROR: scripts/R10.4_WORKING path %s " \
                      "is different from the default path %s/scripts/R10.4_WORKING." \
                      % (self.analysis_R104working, self.defaultAIVpath)
                print "USE THE OPTION -t AND RUN AT YOUR OWN RISK"
                sys.exit()
        # Check for existence of the directory
        if not os.path.isdir(self.analysis_R104working):
            while not os.path.isdir(self.analysis_R104working):
                print 'ERROR: %s does not exist. Exitting' % self.analysis_R104working
                self.analysis_R104working = raw_input(
                    'Please enter the full path of the science/scripts/R10.4_WORKING: ')
            print 'Adding %s to environmental variable PATH' % self.analysis_R104working
            os.environ['PATH'] += os.pathsep + self.analysis_R104working
            is_path_modified = True

        # Similar with /opt/telcalsa/current/bin, now /opt/telcalsa_default/current
        # Needed to reduce the allskypoiting run
        telcalsapath = '/opt/telcalsa_default/current'
        ospath = str(os.environ['PATH'])
        if ospath.__contains__(telcalsapath):
            print 'PATH contains %s ' % telcalsapath
        else:
            print 'System PATH should contain %s' % telcalsapath
            print 'Adding %s to environmental variable PATH' % telcalsapath
            os.environ['PATH'] += os.pathsep + telcalsapath
            is_path_modified = True

        # Similar with /opt/tpoint/bin
        # Needed to reduce the allskypoiting run
        tpoint_path = '/opt/tpoint/bin'
        ospath = str(os.environ['PATH'])
        if ospath.__contains__(tpoint_path):
            print 'PATH contains %s' % tpoint_path
        else:
            print 'System PATH should contain %s' % tpoint_path
            print 'Adding %s to environmental variable PATH' % tpoint_path
            os.environ['PATH'] += os.pathsep + tpoint_path
            is_path_modified = True

        if is_path_modified:
            print 'Original PATH'
            print original_path
            print 'Modified PATH'
            print os.environ['PATH']

        is_path_modified = False
        original_path = os.environ.get('PYTHONPATH')
        if original_path is None:
            print 'PYTHONPATH is not set'
            os.environ['PYTHONPATH'] = self.analysis_scripts_path + os.pathsep + self.analysis_R104working
            is_path_modified = True
        else:
            ospath = str(original_path)
            if ospath.__contains__(self.analysis_scripts_path):
                print 'PYTHONPATH contains %s' % self.analysis_scripts_path
            else:
                print 'PYTHONPATH should contain %s' % self.analysis_scripts_path
                print 'Adding %s to environmental variable PYTHONPATH' % self.analysis_scripts_path
                os.environ['PYTHONPATH'] = self.analysis_scripts_path + os.pathsep + os.environ['PYTHONPATH']
                is_path_modified = True
            if ospath.__contains__(self.analysis_R104working):
                print 'PYTHONPATH contains %s' % self.analysis_R104working
            else:
                print 'PYTHONPATH should contain %s' % self.analysis_R104working
                print 'Adding %s to environmental variable PYTHONPATH' % self.analysis_R104working
                os.environ['PYTHONPATH'] = self.analysis_R104working + os.pathsep + os.environ['PYTHONPATH']
                is_path_modified = True
            if ospath.__contains__('TMCDBAccessClient'):
                print 'PYTHONPATH contains TMCDBAccessClient'
            else:
                print 'PYTHONPATH should contain TMCDBAccessClient'
                print 'Adding /opt/TMCDBAccessClient/bin to environmental variable PYTHONPATH'
                os.environ['PYTHONPATH'] += os.pathsep + '/opt/TMCDBAccessClient/bin'
                is_path_modified = True

        if is_path_modified:
            print 'Original PYTHONPATH'
            print original_path
            print 'Modified PYTHONPATH'
            print os.environ['PYTHONPATH']

        # Creating environmental variable TMCDB_CONFIGURATION_NAME
        # Needed for the AllSkyPointing pipeline reduction
        print "Checking TMCDB_CONFIGURATION_NAME"
        if (os.environ.get('TMCDB_CONFIGURATION_NAME') is None) or \
                (os.environ['TMCDB_CONFIGURATION_NAME'] != 'CURRENT.AOS'):
            os.environ['TMCDB_CONFIGURATION_NAME'] = 'CURRENT.AOS'
            print "TMCDB_CONFIGURATION_NAME initialized to CURRENT.AOS"
        else:
            print "TMCDB_CONFIGURATION_NAME is CURRENT.AOS"

        # Check if ~/.casa/init.py exist. If not, create it and if it does,
        # adds the line to load CASA_TELCAL if missing
        # Needed to reduce the baseline run
        print "Checking if .casa/init.py exists and contains execfile(os.getenv(\'CASA_TELCAL\'))"
        casainitfilepath = os.path.expanduser('~/.casa/init.py')
        print casainitfilepath
        if os.path.isfile(casainitfilepath):
            print 'init.py exists'
            casainitfile = open(casainitfilepath, "r+")
            if not casainitfile.read().__contains__('execfile(os.getenv(\'CASA_TELCAL\'))'):
                casainitfile.write('execfile(os.getenv(\'CASA_TELCAL\'))')
                print 'execfile(os.getenv(\'CASA_TELCAL\')) added to init.py'
            else:
                print 'init.py contains execfile(os.getenv(\'CASA_TELCAL\'))'
        else:
            if not os.path.isdir(os.path.expanduser('~/.casa')):
                print "~/.casa directory does not exist. Creating..."
                os.mkdir(os.path.expanduser('~/.casa'))
            casainitfile = open(casainitfilepath, "w")
            casainitfile.write('execfile(os.getenv(\'CASA_TELCAL\'))')
            print 'init.py created and initialized'
        casainitfile.close()

        # TODO All communication should be done in the AAI directory instead of the /tmp !!!!
        # Check whether needed directories for other scripts exist and create them if needed
        neededdirectories= ['/tmp/tiltmeterfiles',
                            '/tmp/trend_analysis',
                            '/tmp/trend_analysis/position_models',
                            '/tmp/trend_analysis/pointing_models',
                            '/tmp/trend_analysis/delay_models']
        print 'Checking needed directories'
        for neededdir in neededdirectories:
            if os.path.isdir(neededdir):
                print '%s exists' % neededdir
            else:
                print '%s does not exist. Creating.' %neededdir
                os.mkdir(neededdir)
                subprocess.call(['chmod', '777', neededdir])

    # Check current directory and ask whether to create a subfolder
    @staticmethod
    def check_directory():
        print "\033[31mChecking working directory\033[39m"
        # Default name of the Antenna Integration Folder
        subfolder_name = time.strftime("%Y%m%d_AntennaIntegration")
        # If we are not in this subfolder
        if os.getcwd().__contains__(subfolder_name):
            print 'Running on default subfolder %s ' % subfolder_name
        else:
            # Check if that subfolder exists. If so, we change directory
            if os.path.isdir(subfolder_name):
                os.chdir(subfolder_name)
                print "Antenna Integration directory found.\nResults will be stored in %s" % os.getcwd()
            # Else we suggest the creation of such directory
            else:
                print 'This script will create output files and datafiles in current directory %s' % os.getcwd()
                answer = raw_input("Do you want it to be stored in subfolder %s {[y]/n}:)" % subfolder_name)
                if answer != 'n' and answer != 'no':
                    os.mkdir(subfolder_name)
                    os.chdir(subfolder_name)

    # Prints an error message and exits the program
    @staticmethod
    def exitwitherrormessage(exitmessage):
        print exitmessage
        sys.exit(0)

    # User Input requires:
    # - ArrayName
    # - List of Antennas to Integrate [empty for single step on ALL]
    # - Starting step (or single step for all antennas) to begin integration
    def input_menu(self):
        print "AntennaIntegrator.py executed by \033[31m %s\033[49m\033[39m" % getpass.getuser()
        self.antennaListToIntegrate = self.antennaListToIntegrate.replace(' ', '')
        if self.antennaListToIntegrate == '':
            # In case it is provided blank separated, we replace by commas
            self.antennaListToIntegrate = raw_input("LIST OF ANTENNAS TO INTEGRATE "
                                                    "(comma separated with no spaces): ").replace(' ', ',')
        elif self.antennaListToIntegrate.lower() == 'all':
            self.antennaListToIntegrate = ''
        print 'Antennas to integrate: \033[31m %s \033[39m' % self.antennaListToIntegrate
        for keys, values in sorted(self.aisteps.iteritems()):
            print ' ' + keys + ': ' + values
        try:
            if self.startingstep == '0':
                self.startingstep = int(raw_input("CHOOSE STARTING STEP: "))
            else:
                self.startingstep = int(self.startingstep)
        except ValueError:
                self.exitwitherrormessage('ERROR: Input is not a number.')
        print self.startingstep
        if self.startingstep <= 0 or self.startingstep > len(self.aisteps):
            self.exitwitherrormessage('ERROR: Not a valid step number')
        if self.arrayName == '':
            if self.startingstep < 4:
                self.array_required(False)
            else:
                self.array_required(True)

    def array_required(self, correlatorneeded=True):
        if not self.justreduce:
            print "******************************************************************************"
            if correlatorneeded:
                print "Create a MANUAL ARRAY including the antennas to be integrated plus other"
                print "antennas up to ~10 for interferometric observations."
                print "Array should include DA41 and/or DA45 and/or DA49."
                print "CORRELATOR NEEDED"
            else:
                print "Create a MANUAL ARRAY inlcuding ONLY THE ANTENNAS TO BE INTEGRATED."
                print "CORRELATOR IS NOT NEEDED at this stage"
            print "******************************************************************************"
            self.arrayName = raw_input("ARRAY NAME: ")
            print 'Using Array name \033[31m %s \033[39m' % self.arrayName
            print "Making sure Band 3 is up"
            self.radio_setup()

    # Controls the connection (connect/disconnect) to AOS machine for executing scripts
    def connection_file_to_aosgns(self, disconnect=False):
        # It will not communicate with server if in simulation or in reduction mode
        if not self.justsimulate and not self.justreduce:
            # Disconnect communication with AOS-GNA
            if disconnect:
                # Writes the -exit- command so aos-gns disconnects
                commandfile = file(self.commandsfile, 'a')
                print 'DISCONNECTING. No more commands will be sent.'
                commandfile.write('\n-exit-')
                commandfile.close()
                # Remove the link to the file that is red from aos-gns
                subprocess.call(['cp', self.commandsfile, os.getcwd()])
            # Create file to communicate commands to aog-gns
            else:
                commandfile = file(self.commandsfile, 'w')
                commandfile.write('Commands to be executed:\n')
                commandfile.close()
                subprocess.call(['chmod', '666', self.commandsfile])

    def is_commands_file_modified(self):
        last_modified = os.stat(self.commandsfile).st_mtime
        if last_modified > self.comandsfile_modifiedtime:
            self.comandsfile_modifiedtime = last_modified
            return True
        else:
            return False

    # Sends a command to server and returns the output. Does nothing in simulation mode
    def run_command_in_aosgns(self, commandtoexecute):
        if self.justreduce:
            userinputuid = raw_input("Enter the ExecUID of the dataset to be reduced: ")
            self.exec_uid = userinputuid
            self.asdm_uid = self.exec_uid.replace('uid:', 'uid/').replace('/', '_')
        else:
            print "Executing in aos-gns.aiv.alma.cl: \033[34m %s \033[39m" % commandtoexecute
            if not self.justsimulate:
                commandfile = file(self.commandsfile, 'a')
                commandfile.write('\n%s-toexecute-' % commandtoexecute)
                commandfile.close()
                # This is the name of the file where we will store the output log from the AOS execution
                self.last_aos_execution_output = time.strftime("%Y%m%d_%H:%M_"'AOSoutput.log')
                # And we wait until the end of the execution
                while True:
                    if self.is_commands_file_modified():
                        commandfile = file(self.commandsfile, 'r')
                        commandfilealllines = commandfile.readlines()
                        commandfileline = commandfilealllines[len(commandfilealllines)-1]
                        # If executed, we check whether it comes with a uid file
                        if commandfileline.__contains__('-executed-'):
                            if commandfileline.__contains__('uid'):
                                newuid = commandfileline.split('-executed-')[1].replace('\n', '')
                                print "Execution returned the following UID: %s" % newuid
                                if newuid.__contains__('uid'):
                                    self.exec_uid = newuid
                                    self.asdm_uid = self.exec_uid.replace('uid:', 'uid/').replace('/', '_')
                            else:
                                self.exec_uid = 'none'
                                self.asdm_uid = 'none'
                            # We copy the output from the execution to a local file
                            if os.path.isfile(self.aos_execution_output_file):
                                subprocess.call(['mv',
                                                self.aos_execution_output_file,
                                                self.last_aos_execution_output])
                            else:
                                print 'WARNING: The execution in AOS returned no output log file'
                            break
                    time.sleep(1)

    # Send a command to the local machine. Does nothing in simulation mode
    def run_command_in_localhost(self, commandline):
        print "Executing in red-osf.osf.alma.cl: \033[33m %s \033[39m" % commandline
        if self.justreduce or not (self.justsimulate or self.asdm_uid.__contains__('simulation')):
            exitstatus = subprocess.call(commandline, shell=True)
            # exitstatus = subprocess.call(commandline.split(' '))
            return exitstatus
        else:
            return 0

    # Exports the ASDM
    def export_asdm(self, metadata_only=False):
        # Basic checks of the uid retrieved
        if self.asdm_uid == 'none':
            print "Execution did not return a valid UID"
            newuid = raw_input("ENTER to continue or input the right uid:")
            if newuid.lower().__contains__('uid'):
                self.exec_uid = newuid
                self.asdm_uid = self.exec_uid.replace('uid:', 'uid/').replace('/', '_')
            else:
                return False
        elif self.asdm_uid.__contains__('simulation'):
            print "Simulated observation. Simulated asdmExport"
            return True
        elif self.justsimulate and not self.justreduce:
            print "Simulated antenna integration. Simulated asdmExport"
            self.asdm_uid = 'uid__simulation_uid_000'
            return True
        else:
            # Check if ASDM file exists and if not, asdmExport
            if not os.path.exists(self.asdm_uid):
                if metadata_only:
                    cmd_asdmexport = 'asdmExport -m %s' % self.exec_uid
                else:
                    cmd_asdmexport = 'asdmExport %s' % self.exec_uid
                print '%s file does not exist. Will proceed with\n %s' % (self.asdm_uid, cmd_asdmexport)
                self.run_command_in_localhost(cmd_asdmexport)
                while os.path.isfile('%s/ASDM.xml.missing' % self.asdm_uid):
                    subprocess.call(['rm', '-rf', '%s' % self.asdm_uid])
                    print 'Data is not yet ready. Waiting for 60 seconds'
                    time.sleep(60)
                    self.run_command_in_localhost(cmd_asdmexport)
            else:
                print '%s file does exist. Will not be exported' % self.asdm_uid
            return True

    # Reload the pointing model
    def radio_setup(self, reloadpoint=False):
        if reloadpoint:
            command_to_execute = 'radioSetup.py -b 3 -r -p --arrayName=%s' % self.arrayName
        else:
            command_to_execute = 'radioSetup.py -b 3 --arrayName=%s' % self.arrayName
        self.run_command_in_aosgns(command_to_execute)

    # METHODS TO IMPLEMENT THE INDIVIDUAL STEPS OF THE INTEGRATION
    def step_initial_checks(self):
        # ONLY WORKING FOR ONE ANTENNA. SO WE MAKE A LOOP
        # AND EXECUTE THE SCRIPT FOR EACH ANTENNA
        for ants in self.antennaListToIntegrate.split(','):
            # Run the necessary scripts on aos-gns and the output is sent to a
            command_to_execute = '/groups/science/scripts/R10.4_WORKING/util/DRXcheck.py -a %s;' \
                                 '/groups/science/analysis_scripts/compareTilt.py -a %s;' \
                                 '/users/ahales/scripts//getMetroMode.py -a %s' % (ants, ants, ants)
            self.run_command_in_aosgns(command_to_execute)
            # We move the output file (already stored locally red-osf to the right file where
            # ant_move_check.py expects to find it
            command_to_execute = 'cp %s /tmp/drx_metro_results_%s_last.txt' % (self.last_aos_execution_output, ants)
            self.run_command_in_localhost(command_to_execute)
            # and save a copy locally
            command_to_execute = 'cp /tmp/drx_metro_results_%s_last.txt ./' % ants
            self.run_command_in_localhost(command_to_execute)
            # Executing the ant_move_check in AntennaIntegrator mode
            command_to_execute = 'ant_move_check.py -a %s --AI' % ants
            self.run_command_in_localhost(command_to_execute)
            # We remove the file to avoid confusion
            command_to_execute = 'rm /tmp/drx_metro_results_%s_last.txt' % ants
            self.run_command_in_localhost(command_to_execute)
            # Pause between each antenna. Not displayed if it is the last antenna in the list
            if ants != self.antennaListToIntegrate.split(',')[len(self.antennaListToIntegrate.split(','))-1]:
                raw_input("Press ENTER to proceed to next antenna")

    def step_pad_tilt_terms(self):
        # Divide the antenna list in the different groups
        ant_pm = []
        ant_dv = []
        ant_da = []
        for ants in self.antennaListToIntegrate.split(','):
            if ants.__contains__('PM'):
                ant_pm.append(ants)
            elif ants.__contains__('DV'):
                ant_dv.append(ants)
            elif ants.__contains__('DA'):
                ant_da.append(ants)
        # Executing the Tiltmeter runs 3 times
        # todo, make sure 3 are needed!!!
        for tiltruns in range(1):
            if len(ant_pm) > 0:
                command_to_execute = 'pmTiltMeter.py -a %s' % ','.join(ant_pm)
                self.run_command_in_aosgns(command_to_execute)
            if len(ant_dv) > 0:
                command_to_execute = 'dvTiltMeterCalPinch.py -a %s' % ','.join(ant_dv)
                self.run_command_in_aosgns(command_to_execute)
            if len(ant_da) > 0:
                # TODO DSOANT-90 dealing broken tiltmeters
                command_to_execute = 'aemTiltmeterCal.py -a %s' % ','.join(ant_da)
                self.run_command_in_aosgns(command_to_execute)
        # We create directory from aos-gns which will be accessible by red-osf
        command_to_execute = "mkdir " \
                             "%s/tmpTiltmeterFiles" % self.commandsfilepathatAOS
        self.run_command_in_aosgns(command_to_execute)
        # We copy the needed tiltmeter files into that common directory
        command_to_execute = "cp /groups/science/data/RADIO/POINT/Tiltmeter/*_tiltmeterValues_backup.txt " \
                             "%s/tmpTiltmeterFiles" % self.commandsfilepathatAOS
        self.run_command_in_aosgns(command_to_execute)
        # In red-osf we copy the required tiltmeter files into the directory where
        # EditAnAwFromTilts.py expects them
        command_to_execute = 'cp ' \
                             '%s/tmpTiltmeterFiles/*.txt ' \
                             '/tmp/tiltmeterfiles/' % self.commandsfilepath
        self.run_command_in_localhost(command_to_execute)
        command_to_execute = 'chmod 776 /tmp/tiltmeterfiles/*'
        self.run_command_in_localhost(command_to_execute)
        # Computing AN and AW
        for ants in self.antennaListToIntegrate.split(','):
            if not ants.__contains__('CM'):
                # We execute the EditAnAwFromTilts in not standalone mode
                # (It does assume the files are already in the right place. Done before this loop)
                command_to_execute = 'EditAnAwFromTilts.py -a %s --AI' % ants
                self.run_command_in_localhost(command_to_execute)
                # done it is automated now.
                # raw_input("Press ENTER after you updated AN0 and AW0 in TMCDB for %s: " % ants)
                if ants.__contains__('DV'):
                    print "\033[31mWARNING: After you press enter you will need to go to the aos-gns terminal " \
                          "an provide the requested password\033[39m"
                    print "\033[31mIMPORTANT: You will have to enter it TWICE and then press ENTER again\033[39m"
                    raw_input("\033[31m***Press ENTER to make sure you read this message***\033[39m")
                    command_to_execute = 'setACUMetrology -a %s' % ants
                    self.run_command_in_aosgns(command_to_execute)
            else:
                print 'No TiltMeters in Ant %s ... skipping' % ants
        # For all antennas, now we run steDailySetup to reset the metrology values.
        command_to_execute = 'steDailySetup -a %s -s' % self.antennaListToIntegrate.replace(',', ' ')
        self.run_command_in_aosgns(command_to_execute)
        # As the setACUMetrology sends the DV antennas to shutdown, we have to set them to autonomous
        if len(ant_dv) > 0:
            command_to_execute = 'mountAxisToAutonomous.py -a %s -s' % ' '.join(ant_dv)
            self.run_command_in_aosgns(command_to_execute)

    def step_ia_determination(self, initial=True):
        # Execution of observation in aos-gns
        if initial:
            command_to_execute = 'ObsCalAzSearch.py --array=%s' % self.arrayName
        else:  # DEPRECATED
            point_band = raw_input("Receiver Band to perform pointing: ")
            print point_band
            point_source = raw_input("Bright QSO to point: ")
            print point_source
            command_to_execute = 'InterferometricPointing.py -b %s -a %s -o %s'\
                                 % (point_band, self.arrayName, point_source)
        self.run_command_in_aosgns(command_to_execute)
        # No need to import the data, as it is done within the script.
        # But we do it for consistency and for checks within export_asdm
        if self.export_asdm():
            command_to_execute = 'EditIApointoffset.py -a %s -f %s' % (self.antennaListToIntegrate, self.asdm_uid)
            exitstatus = self.run_command_in_localhost(command_to_execute)
            if exitstatus == 0:
                self.radio_setup(reloadpoint=True)

    def step_check_delays(self):
        # Executing PAD Delay observation
        command_to_execute = 'ObsCalPadDelay.py --array=%s' % self.arrayName
        self.run_command_in_aosgns(command_to_execute)
        # Exporting Delay Model from TMCDB
        if self.export_asdm():
            infile = '%s-PadDelayModel-ORIG.xml' % self.asdm_uid
            command_to_execute = 'ExportDelayModel.py -c CURRENT.AOS -f %s' % infile
            self.run_command_in_localhost(command_to_execute)
            print "PAD Delay model from configuration 'CURRENT.AOS' saved to %s " % infile
            # Calculating the modified Delay Model based on the data
            outfile = infile.replace('ORIG', 'NEW')
            command_to_execute = 'EditDelays.py -f %s -i %s -P' % (self.asdm_uid, infile)
            self.run_command_in_localhost(command_to_execute)
            # Importing results into TMCDB
            answer = raw_input('Would you like to proceed with the Import steps. '
                               'Be careful, if answer is "y", you will proceed with '
                               'importing these model to the TMCDB. {y|[n]} ?').lower()
            if answer == 'y' or answer == 'yes':
                print "UPDATING PAD Delays IN TMCDB"
                # TODO this was to avoid the error because the NEW.xml does not exist
                outfile = 'DelayModel-current.xml'
                command_to_execute = 'ImportDelayModel.py -f %s -c CURRENT.AOS -m \'Automatic update of PAD delays ' \
                                     'based on dataset %s\'' % (outfile, self.asdm_uid)
                self.run_command_in_localhost(command_to_execute)
                print "New PAD delay loaded in the TMCDB"
            else:
                print 'You have chosen to do the Edit manually.'
                print 'To do so, run the following command.'
                print '\033[95m ImportDelayModel.py -f %s -c CURRENT.AOS -m ' \
                      '"Updating PAD delays [YOUR COMMENT]" \033[0m' \
                      % outfile

    # def step_ia_ie_ca_determination(self):
        # /groups/science/data/SISTENV/configSISTEnv
        # /groups/science/data/SISTENV/scripts/station3/radioPointing/DetectorPointing.py
    #    command_to_execute = 'InterferometricPointing.py -b 3 -f 5 -r 6 -a %s'\
    #                             % self.arrayName
    #    self.run_command_in_aosgns(command_to_execute)
    #    if self.export_asdm(True):
    #        return True

    def step_baseline(self):
        # Executing Baseline run, or Pointing run,
        # if the observation will be used for both Baseline and Pointing determination
        if self.joinbaselinepointingobservation:
            command_to_execute = 'ObsCalAllSkyPointing.py --array=%s' % self.arrayName
        else:
            command_to_execute = 'ObsCalBaselines.py --array=%s' % self.arrayName
        self.run_command_in_aosgns(command_to_execute)
        # Exporting Delay Model from TMCDB
        if self.export_asdm():
            # ------------------------------
            # Pipeline assisted calibration
            # ------------------------------
            # Provide instructions for data reduction
            # print 'NOW we will try to automatically run the baseline track.'
            # print 'This is not yet tested in the frame of the AntennaIntegrator'
            # print 'If it happens to fail:'
            print 'Run the following commands in CASA for reducing the baseline observations.'
            print 'NOTE(2016-07): If it does not work, follow the manual reduction instructions in the web'
            print '\033[95mcasapy-telcal \033[0m'
            # print '\033[95mimport sys \033[0m'
            # print '\033[95msys.path.append(\'%s\') \033[0m' % self.analysis_scripts_path
            print '\033[95mimport BaselineReducer as BR \033[0m'
            print '\033[95mblr = BR.BaselineDataReducer() \033[0m'
            print '\033[95mblr.runOnce(asdm =\'%s\')  \033[0m' % self.asdm_uid
            print 'IMPORTANT NOTE (valid as of 20-12-2015):'
            print 'The baselineReducer detects phase jumps in the first run. Then in a second iteration it suggests'
            print 'flagging them. If scans are to be flagged for the reference antenna it will result in an error'
            print 'As a temporary workaround, use ignoredScans=\'X,Y,Z\' option in blr.runOnce()'
            print 'where X,Y,Z are the scans which were to be flagged in the reference antenna'
            raw_input('When done PRESS ENTER to continue:')

            # Invoking calibration in casapy-telcal
            # casacommandfile = open('AntennaIntegratorBaselinecasacommand.py', 'w')
            # casacommandfile.write('import sys\n')
            # casacommandfile.write('sys.path.append(\'%s\')\n' % self.analysis_scripts_path_to_add)
            # casacommandfile.write('import BaselineReducer as BR\n')
            # casacommandfile.write('blr = BR.BaselineDataReducer()\n')
            # casacommandfile.write('blr.runOnce(asdm =\'%s\')\n' % self.asdm_uid)
            # casacommandfile.close()
            # os.system('casapy-telcal -c AntennaIntegratorBaselinecasacommand.py')
            print self.asdm_uid
            antposresultfile = [s for s in os.listdir('./') if '%s' % self.asdm_uid in s and 'antpos' in s]
            if antposresultfile == []:
                antposresultfile = ''
            else:
                # We get the first file in case there are multiple
                antposresultfile = antposresultfile[0]
            answer = raw_input('If any antenna was modified. Will now run\n'
                               'UpdatePosition.sh %s\n'
                               'Proceed: y/[n]' % antposresultfile)
            if answer.lower() == 'y':
                if antposresultfile == '':
                    antposresultfile = raw_input("Please introduce the _antpos_result directory name to be used: ")
                command_to_execute = 'UpdatePosition.sh %s' % antposresultfile
                self.run_command_in_localhost(command_to_execute)

    def step_pointing_model(self, firstiteration=False):
        if firstiteration:
            # We assume we are within the beam and we can do interferometric pointing
            # Otherwise we should use the Total Power methods.
            command_to_execute = 'InterferometricPointing.py -b 3 -f 3 -e 30 -r 1 -a %s' % self.arrayName
            self.run_command_in_aosgns(command_to_execute)
        # Observing Pointing Run, only if it is not done already in the baseline step
        elif not self.joinbaselinepointingobservation:
            # Executing Pointing run
            command_to_execute = 'ObsCalAllSkyPointing.py --array=%s' % self.arrayName
            self.run_command_in_aosgns(command_to_execute)
        # Exporting Delay Model from TMCDB
        if self.export_asdm():
            # ------------------------------
            # Pipeline assisted calibration
            # ------------------------------
            if self.antennaListToIntegrate == '':
                command_to_execute = 'ReducePointingData.py -f %s' % self.asdm_uid
            else:
                command_to_execute = 'ReducePointingData.py -f %s -a %s' % (self.asdm_uid, self.antennaListToIntegrate)
            self.run_command_in_localhost(command_to_execute)
            self.radio_setup(reloadpoint=True)

    # Runs the individual integration steps to be run according to the user selection
    def run_integration_steps(self):
        # First we update the common science directory. Updated information will be used by other scripts.
        print "UPDATING COMMON SCIENCE DIRECTORY..."
        p = subprocess.Popen('/opt/bin/update-sciencedir-redosf', stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        print "...UPDATED\n"
        # If the list of antennas for integrations is empty (meaning ALL) or running in reduction mode,
        # then only one step is executed at a time
        # And it will therefore not share the same observation between baseline and pointing
        if self.antennaListToIntegrate == '' or self.justreduce:
            stepstorun = range(int(self.startingstep), int(self.startingstep)+1)
            self.joinbaselinepointingobservation = False
        # Otherwise it runs all steps starting with the one selected by the user
        else:
            stepstorun = range(int(self.startingstep), len(self.aisteps)+1)
        # Goes through all steps and execute them
        print "READY to proceed to integration steps"
        for step in stepstorun:
            startorrepeat = True
            repetitionnumber = 0
            print ""
            print "\033[41m\033[30m**********************************\033[49m\033[39m"
            print "\033[41m\033[30m*** RUNNING INTEGRATION STEP %s ***\033[49m\033[39m" % step
            print "\033[41m\033[30m**********************************\033[49m\033[39m"
            while startorrepeat:
                startorrepeat = False
                print "\033[31m%s\033[39m" % ai.aisteps[str(step)].upper()
                if repetitionnumber > 0:
                    print "\033[31mRepetition %s\033[39m" % repetitionnumber
                if step == 1:
                    self.step_initial_checks()
                elif step == 2:
                    self.step_pad_tilt_terms()
                elif step == 3:
                    self.step_ia_determination(True)
                elif step == 4:
                    # At this point we need the correlator in the array as well as other reference antennas
                    if repetitionnumber == 0:
                        # If this is the only step, the array will already be created in the input_menu()
                        if self.startingstep != step:
                            self.array_required(True)
                        print "\033[1;41m\033[30mWARNING!!!WARNING!!!WARNING!!!WARNING!!!WARNING!!!\033[49m\033[0;39m"
                        print "Before starting interferometric integration we will associate weather"
                        print "stations to each PAD. For this purpose "
                        print "MAKE SURE NO OBSERVATION IS RUNNING IN ANY ARRAY!!!"
                        print "Answer no if you want to avoid this step"
                        answer = raw_input("Proceed? (y/[n])")
                        if answer == 'y' or answer == 'Y' or answer == 'yes' or answer == 'YES' or answer == 'Yes':
                            command_to_execute = 'antennaToWeatherStation.py'
                            self.run_command_in_aosgns(command_to_execute)
                    self.step_check_delays()
                elif step == 5:
                    self.step_baseline()
                elif step == 6:
                    # If the starting step is the pointing, there cannot be a join baseline observation
                    if self.startingstep == step:
                        self.joinbaselinepointingobservation = False
                    self.step_pointing_model()
                # Pause between each step. Not displayed if it is the last step in the list
                if step != stepstorun[len(stepstorun)-1]:
                    answer = raw_input('ENTER to Continue to step %s, (r) to repeat step %s, '
                                       '(mc/mr) to add message to log and continue/repeat (q) to quit:'
                                       % (step+1, step)).lower()
                    if answer.lower().__contains__('q'):
                        return
                    if answer.lower().__contains__('r'):
                        startorrepeat = True
                        repetitionnumber += 1
                    if answer.lower().__contains__('m'):
                        print '\033[32m %s \033[39m' % raw_input('\nUSER MESSAGE:')
                        if answer.lower().__contains__("r"):
                            raw_input("ENTER to continue to repeat this step")
                        elif answer.lower().__contains__("c"):
                            raw_input("ENTER to continue to next step")
                else:
                    print 'FINISHED THE %s REQUESTED INTEGRATION STEPS' % len(stepstorun)


if __name__ == '__main__':
    parser = OptionParser()
    parser.add_option("-a", "--arrayName", dest="inputarrayname", default='', help="Array Name")
    parser.add_option("-l", "--antennaList", dest="inputantennalist", default='',
                      help="List of Antennas to Integrate (comma separated with no spaces)")
    parser.add_option("-s", "--step", dest="step", default='0',
                      help="Starting step.")
    parser.add_option("-R", "--datareduction", action='store_true', dest="datareduction", default=False,
                      help="Execute only to reduce already taken datasets. UIDs will be requested.")
    parser.add_option("-S", "--simulation", action='store_true', dest="simulation", default=False,
                      help="Execute in simulation. No commands will be executed.")
    parser.add_option("-t", "--testing", action='store_true', dest="testing", default=False, help=SUPPRESS_HELP)
    (options, args) = parser.parse_args()
    ai = AntennaIntegrator(options.inputarrayname, options.inputantennalist, options.step, options.simulation,
                           options.datareduction, options.testing)
    # To handle the Ctrl+C signal
    original_sigint = signal.getsignal(signal.SIGINT)
    signal.signal(signal.SIGINT, ai.cancel_execution)
    # Starting integration
    ai.connection_file_to_aosgns()
    ai.input_menu()
    ai.run_integration_steps()
    ai.connection_file_to_aosgns(disconnect=True)
