#!/usr/bin/env python

 
"""
Script Edit IA pointing offset on the TMCDB, based on the latest Tiltmenter measurement.

Usage: 

EditAnAwFromTilts.py -a DVXX,DAXX -f ASDM_FILENAME
EditAnAwFromTilts.py -a DVXX,DAXX -f uid://  (will asdmExport metadata)

Script reads the horizontal pointing offset from the ASDM_FILENAME and computes IA pointing parameter and 
creates a new XML file for each of the given antennas

Alternative usage: (for single antenna only):

EditAnAwFromTilts.py -a DVXX -i IAnew   : Directly inputs the IA parameter into TMCDB
EditAnAwFromTilts.py -a DVXX -H Horizontal_offset -E Elevation   : Calculates IA based on the H_offset and elevation
                                                                   and input result into TMCDB

Script then gives the option to do the ImportPointingModel.py command (user has to type 'y' or 'yes).

Otherwise it prints the ImportPointingModel.py command on screen (and the new IA value).

2015-08-14  S. Martin
#"""

import os
import sys
import getpass
from optparse import OptionParser
from pylab import *
from asdm import *
from AntennaIntegrator import AntennaIntegratorLogger
import subprocess

username = getpass.getuser()

sys.path.append('/users/%s/AIV/science/scripts/R10.4_WORKING/' % username)

sys.path.append('/users/%s/AIV/science/analysis_scripts/' % username)


#################################
# Defining some useful functions
#################################
def grep_file(file_to_grep, search_term):
    line_found = ''
    for read_line in open(file_to_grep, 'r'):
        if re.search(search_term, read_line):
            line_found = read_line
            if read_line is None:
                print 'no matches found'
    return line_found


def system_except(cmd):
    ret = os.system(cmd)
    if ret != 0:
        raise Exception("Command '%s' returned non-zero exit status %d" % (cmd, ret))


def continue_after_warning(warning_message):
    print warning_message
    input_answer = raw_input('Continue? {[y]|n} ?').lower()
    if input_answer == 'n' or input_answer == 'no':
        sys.exit()


# Runs the appropriate tc_pointing command to calculate the pointing solutions in the dataset
def cal_pointing_table_regenerate(asdm_file):
    print '\033[31mAttempting tc_pointing in casapy-telcal\033[39m'
    if os.listdir('%s/ASDMBinary' % asdm_file) == []:
        print "ASDMBinary data not found. Running asdmExport again"
        os.system('asdmExport %s' % asdm_file.replace('uid_', 'uid:').replace('_', '/'))

    casacommand = 'tc_pointing(' \
                  'asdm="%s",' \
                  'dataorigin="totalpower",' \
                  'antenna="%s",' \
                  'fitwidth=False,' \
                  'polaraverage=True,' \
                  'planetfit=False,' \
                  'calresult="%s",' \
                  'showplot=True,' \
                  'verbose=False)' % (asdm_file, options.antenna, asdm_file)
    casacommandfile = open('EditIAcalpointingregenerate.py', 'w')
    casacommandfile.write(casacommand)
    casacommandfile.close()
    os.system('casapy-telcal -c EditIAcalpointingregenerate.py')
    if os.path.isfile('%s/CalPointing.xml' % asdm_file):
        print "CalPointing.xml succesfully generated"
    else:
        print "\033[91mCalPointing.xml could not be generated\033[90m"
        print "\033[91mOBSERVATION SHOULD BE REPEATED. OTHERWISE IA CANNOT BE DETERMINED FOR THE ANTENNAS ABOVE\033[90m"
        print "\033[91mData may be ok, but are likely flagged. If problem persists it should be reported\033[90m"
        sys.exit()


# Checks if CalPointing.xml exists and if not, run tc_pointing in casapy-telcal to generate the table.
def cal_pointing_table_check(asdm_file):
    # Check the CalPointing.xml file exist
    if os.path.isfile('%s/CalPointing.xml' % asdm_file):
        print "CalPointing.xml found"
    else:
        # If no ASDMBinary, asdmExport again
        print '\033[31mWARNING: CalPointing.xml not found\033[39m'
        cal_pointing_table_regenerate(asdm_file)


# Generate Plots for evaluation of the results. ASDM binary data must be present (not checked)
def cal_pointing_generate_plots(asdm_file):
    # AutoGenerate plots
    # Invoking calibration in casapy-telcal
    casacommandfile = open('AntennaIntegratorGenerateIAplots.py', 'w')
    casacommandfile.write('asdmplot(\n')
    casacommandfile.write('   asdm="%s",\n' % asdm_file)
    casacommandfile.write('   xaxis="azoffset",\n')
    casacommandfile.write('   yaxis="totalpower",\n')
    casacommandfile.write('   antenna="%s",\n' % options.antenna)
    casacommandfile.write('   baseband = "BB_1",\n')
    casacommandfile.write('   polarization="XX",\n')
    casacommandfile.write('   verbose=False)\n')
    casacommandfile.write('import pylab\n')
    casacommandfile.write('fig = pylab.gcf()\n')
    casacommandfile.write('fig.savefig("%s.plot.png")\n' % asdm_file)
    #casacommandfile.write('! eog %s.plot.png\n')
    casacommandfile.close()
    casacommand = 'casapy-telcal -c AntennaIntegratorGenerateIAplots.py'
    subprocess.call(casacommand.split(' '))
    subprocess.Popen(['eog', '%s.plot.png' % asdm_file])


# Method to guide the user through the checking of the results provided by TelCal
def cal_pointing_double_check():
    print 'Results from TelCal and tc_pointing (offline) may not be reliable.'
    print 'Check the AZ offset suggested above by TelCal with the signal in the figure.'
    answer = raw_input('Can the source be seen and is the fit consistent? {y|[n]} ?').lower()
    if answer == 'y' or answer == 'yes':
        print "You can go ahead with the integration"
        return True
    else:
        answer = raw_input('Can you see the source in the plot? {y|[n]} ?').lower()
        if answer == 'y' or answer == 'yes':
            print 'We do need a workaround'
            print 'Estimate the AZ offset from the figure and then run'
            print 'EditIApointoffset.py -a XXXX -H [measured Az offset in figure] -E [elevation of observation]'
            print 'For a better precision you can use the cursor in CASA running'
            print 'casapy-telcal'
            print 'execfile(AntennaIntegratorGenerateIAplots.py)'
        else:
            print 'Something is seriously wrong. We need to do things by hand'
            print 'From an almaop@aos-gns console run'
            print 'radioSetup -y ArrayXXX -r -p -b 3'
            print 'pointingSpiralMap.py -a DXXX -A ArrayXXX'
            print 'See full instructions in:'
            print ' https://wikis.alma.cl/bin/view/DSOANT/AntennaIntegration#Determine_the_IA_pointing_offset'
        raw_input('Press ENTER when done')
        raw_input('Did you find a solution for IA and update the IA with EditIApointoffset.py?{y|[n]} ?').lower()
        if answer == 'y' or answer == 'yes':
            sys.exit(0)  # This will be interpreted as IA update found and modified
        else:
            return False


# Read the CalPointing Table from the ASDM and returns
# ia_from_asdm: Dictionary defined as {'ant': ['relativeOffAZ', 'elevation', 'IAoffset', 'IAold', 'IAnew']}
# antennalist_nosolutions: Antennas for which no solutions were found
def extract_iaoffset_from_asdm(asdm_file, antennalist):
    # Check if ASDM file exists and if not, asdmExport  -> Not needed if asdmExport is done before
    if not os.path.exists(asdm_file) or 'uid:' in asdm_file:
        asdmfile2 = asdm_file.replace('uid_', 'uid:')
        asdmfile2 = asdmfile2.replace('_', '/')
        cmd_asdmexport = 'asdmExport -m %s' % asdmfile2
        continue_after_warning("%s file does not exist. Will proceed with\n"
                               "%s" % (asdm_file, cmd_asdmexport))
        os.system(cmd_asdmexport)

    # check/generate CalPointing.xml
    cal_pointing_table_check(asdm_file)

    # Open the ASDM
    asdm = ASDM()
    try:
        po = ASDMParseOptions()
        po.asALMA()
        asdm.setFromFile(asdm_file, po)
    except:
        #  TODO
        #   e = sys.exc_info()
        #  print e
        #  print 'File %s exist but could not be open ' % asdm_file
        print 'Some execption was found when loading asdm %s' % asdm_file

    # Extracting calPointingTable
    arcsec = 206264.8062470
    cal_pointing_table = asdm.calPointingTable().get()
    # We will store the antennas with solutions found
    antennalist_withsolutions = []
    # We create a list copy of the antennalist
    # We will remove antennas as they are found in the CalPointingTable
    # The remaining antennas are those with no solutions available
    antennalist_nosolutions = antennalist[:]
    # Dictionary with antenna as key
    ia_from_asdm = {'ant': ['relativeOffAZ', 'elevation', 'IAoffset', 'IAold', 'IAnew']}
    for row in cal_pointing_table:
        if row.antennaName() in antennalist:
            antennalist_nosolutions.remove(row.antennaName())
            antennalist_withsolutions.append(row.antennaName())
            coll_offset_relative_az = row.collOffsetRelative()[0][0].get()*arcsec
            direction_el = row.direction()[1].get()
            asdm_ia_offset = coll_offset_relative_az/math.cos(direction_el)
            ia_from_asdm[row.antennaName()] = [round(coll_offset_relative_az, 2),
                                               round(math.degrees(direction_el), 2),
                                               round(asdm_ia_offset, 2), 0, 0]
    if len(antennalist_nosolutions) > 0:
        print "\033[91mWARNING: No solutions in the CalPointing Table for the Antennas: %s\033[90m" \
              % ', '.join(antennalist_nosolutions)
        print "\033[91mOBSERVATION SHOULD BE REPEATED. OTHERWISE IA CANNOT BE DETERMINED FOR THE ANTENNAS ABOVE\033[90m"
        print "\033[91mData may be ok, but are likely flagged. If problem persists it should be reported\033[90m"
    return ia_from_asdm, antennalist_withsolutions, antennalist_nosolutions

###############################
# Redirecting stdout and stderr
###############################
sys.stdout = AntennaIntegratorLogger('EditIApointoffset')
sys.stderr = AntennaIntegratorLogger('EditIApointoffset')

#################################
# Options
#################################

parser = OptionParser()

parser.add_option("-a", "--antenna", dest="antenna", help="Antenna to get parameters. Ex.: DV01", default="")
parser.add_option("-f", "--file", dest="asdmfile", help="ASDM file to be used", default="noasdm")
parser.add_option("-i", "--ia_new", dest="ia_term", help="New IA term to update TMCDB", default="")
parser.add_option("-H", "--h_offset", dest="h_offset", help="Measured H offset", default="")
parser.add_option("-E", "--elevation", dest="elevation", help="Elevation of the Measurement", default="")

(options, args) = parser.parse_args()


if options.antenna == "":
    print "ERROR:Antenna name needs to be specified"
    sys.exit()
antlist = options.antenna.split(',')
antlist_withsolutions = []  # Initialization of the list of antennas WITH solutions found in the data
antlist_nosolution = []  # Initialization of the list of antennas WITHOUT solutions found in the data
asdmfile = options.asdmfile


###############################
# Checking right parameters are input calculating the new IA term
###############################

# From ASDM information
if options.asdmfile != 'noasdm':
    # If ASDM is provided it will overrride any other parameter entered as input. Send Warning.
    if options.ia_term != "" or options.h_offset != "" or options.elevation != "":
        continue_after_warning("WARNING:ASDM FILE HAS BEEN SPECIFIED\n"
                               "Any input value of IA, H offset or Elevation will not be taken into account")
    # if in format uid://XXXX/XXXXXXX/XXX change it to uid___XXXX_XXXXXXX_XXX
    asdmfile = options.asdmfile
    if 'uid:' in asdmfile:
        asdmfile = asdmfile.replace('uid:', 'uid_')
        asdmfile = asdmfile.replace('/', '_')
    # Extracting relevant information from ASDM
    IAresults, antlist_withsolutions, antlist_nosolution = extract_iaoffset_from_asdm(asdmfile, antlist)

# From IA direct input for one single antenna
elif options.ia_term != "":
    if options.h_offset != "" or options.elevation != "":
        continue_after_warning("WARNING: IA PARAMETER HAS BEEN SPECIFIED\n"
                               "Any input value of H offset or Elevation will not be taken into account ")
    if len(antlist) == 1:
        IAresults = {antlist[0]: [0, 0, 0, 0, options.ia_term]}
        antlist_withsolutions = antlist
    else:
        print "Direct IA insertion only available for one antenna"
        sys.exit()

# From H offset and Elevation input parameters for one single antenna
else:
    if options.h_offset == "" or options.elevation == "":
        print "Not enough parameters provided. Needed any of these"
        print "- ASDM FILE"
        print "- IA"
        print "- H offset and Elevation"
        sys.exit()
    if len(antlist) == 1:
        h_offset = float(options.h_offset)
        elevation = float(options.elevation)
        ia_offset = round(h_offset/math.cos(math.radians(elevation)), 2)
        IAresults = {antlist[0]: [h_offset, elevation, ia_offset, 0, 0]}
        antlist_withsolutions = antlist
    else:
        print "Using H offset and elevation to calculate IA only available for one antenna"
        sys.exit()

# We only continue if there are solutions found for any antenna. If not, exit.
if len(antlist_withsolutions) > 0:
    print "#########################################"
    print "Processing  Antennas %s" % antlist_withsolutions
    print "#########################################"
else:
    sys.exit()

print '### Exporting TMCDB PointingModel ###'
# Dictionary with the output files for each antenna
outfile = dict()
for ant in antlist_withsolutions:
    #################################
    # Input/Output File
    #################################
    tmcdbModel = os.popen('ExportPointingModel.py -a %s -c CURRENT.AOS -f %s-PointingModel-%s-ORIG.xml' %
                          (ant, asdmfile, ant)).read()
    infile = tmcdbModel.split()[-1]
    print "Pointing models for antenna %s from configuration 'CURRENT.AOS' saved to %s " % (ant, infile)
    outfile[ant] = infile.replace('ORIG', 'NEW')

    ################################
    # Edit IA
    ################################
    term = 'Coeff name="IA"'
    print "Old xml file is %s " % infile
    old_line = grep_file(infile, term)
    ia_old_str = old_line.split("\"")[3]
    ia_old = float(ia_old_str)
    IAresults[ant][3] = ia_old
    # calculate ia_new with the calculated offset.
    # If already initialized (element 4 in array=ia_new)means direct user input
    if IAresults[ant][4] == 0:
        # we add the calculated offset to the old value
        ia_new = ia_old+IAresults[ant][2]
        IAresults[ant][4] = ia_new
    else:
        ia_new = IAresults[ant][4]
    new_line = old_line.replace(ia_old_str, str(ia_new))
    print "new xml file is %s " % outfile[ant]

    old_f = open(infile, 'r')
    new_f = open(outfile[ant], 'w')
    for line in old_f:
        if term in line:
            line = new_line
        new_f.write(line)
    new_f.close()
    old_f.close()
    os.system('chmod ugo+rwx -R %s' % outfile[ant])


# If the IA parameter is coming from an observations we will generate plots to show to the AOD
# TODO if we ever remove this, we could just load the meta data, not needed the binary
if options.asdmfile != 'noasdm':
    cal_pointing_generate_plots(asdmfile)

print '-------------------------------------------------------------------'
print 'NOTE: The following changes will be applied to TMCDB if accepted'
print '{0:8} - {1:8} - {2:8} - {3:8} ===  {4:8} -> {5:8}'\
    .format('ANTENNA', 'OffsetAz', '    Elev', '  IA off', '  IA old', '  IA new')
for ant in antlist_withsolutions:
    print '{0:8} - {1:8} - {2:8} - {3:8} ===  {4:8} -> {5:8}'\
        .format(ant, IAresults[ant][0], IAresults[ant][1], IAresults[ant][2], IAresults[ant][3], IAresults[ant][4])
print '-------------------------------------------------------------------'

# TODO Until we are fully sure that the fitting works properly
if options.asdmfile != 'noasdm':
    goahead = cal_pointing_double_check()
    if not goahead:
        sys.exit(1)  # This will be interpreted as IA update not found or not modified

answer = raw_input('Would you like to proceed with the Import steps. Be careful, if answer is "y", '
                   'you will proceed with importing these model to the TMCDB. {y|[n]} ?').lower()

if answer == 'y' or answer == 'yes':
    print "UPDATING IA VALUES IN TMCDB"
    for ant in antlist_withsolutions:
        if asdmfile != 'noasdm':
            cmd_import = 'ImportPointingModel.py -f %s -c CURRENT.AOS -m "Automatic update of %s IA ' \
                         'based dataset %s " ' % (outfile[ant], ant, asdmfile)
        else:
            cmd_import = 'ImportPointingModel.py -f %s -c CURRENT.AOS -m "Automatic update of %s IA ' \
                         'based on user input "' % (outfile[ant], ant)
        print cmd_import
        system_except(cmd_import)
        print "New IA loaded in the TMCDB for %s" % ant
        print 'NOTE: It is necessary to run "radioSetup.py -b 3 -p --arrayName=ArrayXXX" so that ' \
              'the updated pointing model in TMCDB is reloaded.'
        print 'WARNING: If running the AntennaIntegrator, this will be done automatically!!!'
    sys.exit(0)  # This will be interpreted as IA update found and modified.
    # This exit was outside the for loop, otherwise
    # it is not looping over the antennas with solutions
    # TODO maybe set exit(0) if the lenght of antlist_withsolutions >0???
else:
    print 'You have chosen to do the Edit manually.'
    print "Edit IA values in the TMCDB Manually "
    print "or run"
    for ant in antlist_withsolutions:
        print '\033[95mImportPointingModel.py -f %s  -c CURRENT.AOS -m "Updating %s IA [YOUR COMMENT] \033[0m" '\
              % (outfile[ant], ant)
    print 'NOTE: After that command it will be necessary to run ' \
          '\033[95mradioSetup.py -b 3 -p --arrayName=ArrayXXX\033[0m so that ' \
          'the updated pointing model in TMCDB is reloaded.'
    sys.exit(1)  # This will be interpreted as IA update not found or not modified


