#!/usr/bin/env python
#ExtractFocusRun.py
#written by RLucas (2009-04-17)
#revised following the re-introduction of the (absolute) Focus Pos. in 8.0.3 (rkneissl, 2011-06-20)
#
#Example:
# ExtractFocusRun -f uid___X54_X35fa_X1 uid___X54_X3a30_X1 [...] -o 
# focus_xyz_20090417.txt
#
#$Id: ExtractFocusRun.py,v 1.6 2012/09/10 17:26:28 rkneissl Exp $

from string import *
import os,sys,time
import casac
from numpy import *
from taskinit import *
import math as m

from asdm import *
from optparse import OptionParser
from pylab import *
import scipy.optimize

def makeTimeStamp() :
    return time.strftime('%Y-%m-%dT%H:%M:%S')

def files_callback(option, opt_str, value, parser):
    """ callback for a  variable number of arguments for option --files
    """
    assert value is None
    value = []
    def floatable(str):
        try:
            float(str)
            return True
        except ValueError:
            return False

    for arg in parser.rargs:
        # stop on --foo like options
        if arg[:2] == "--" and len(arg) > 2:
            break
        # stop on -a, but not on -3 or -3.0
        if arg[:1] == "-" and len(arg) > 1 and not floatable(arg):
            break
        value.append(arg)

    del parser.rargs[:len(value)]
    setattr(parser.values, option.dest, value)

def getBand(fr):
    if fr < 115e9:
        band='ALMA_RB_03'
    elif fr < 270e9:
        band='ALMA_RB_06'
    elif fr < 360e9:
        band='ALMA_RB_07'
    else:
        band='ALMA_RB_09'
    return band

class ExtractFocus:

    def __init__(self):
        self.temperature={}
        self.elevation={}
        self.azimuth={}
        self.focus={}
        self.focuscul={}
        self.pos={}
        self.poscul={}
        self.model1={}
        self.model2={}
        self.model3={}
        self.error={}
        self.antenna=['PM01','PM02','PM03','PM04','DV01','DV02','DV03','DV04','DV05','DV06','DV07','DV08','DV09','DV10','DV11','DV12','DV13','DV14','DV15','DV16','DV17','DV18','DV19','DV20','DV21','DV22','DV23','DV24','DV25','DA41','DA42','DA43','DA44','DA45','DA46','DA47','DA48','DA49','DA50','DA51','DA52','DA53','DA54','DA55','DA56','DA57','DA58','DA59','DA60','DA61','DA62','DA63','DA64','DA65','CM01','CM02','CM03','CM04','CM05','CM06','CM07','CM08','CM09','CM10','CM11','CM12']
        self.axes=['X','Y','Z']
        self.bands=['ALMA_RB_03', 'ALMA_RB_06' , 'ALMA_RB_07', 'ALMA_RB_09']
        self.lines = []
        self.startTime = ''
        self.endTime = ''
        self.files = []
        for axis in self.axes:
            for band in self.bands:
                for antenna in self.antenna:
                    self.temperature[(axis,band,antenna)] = []
                    self.elevation[(axis,band,antenna)] = []
                    self.azimuth[(axis,band,antenna)] = []
                    self.focus[(axis,band,antenna)] = []
                    self.focuscul[(axis,band,antenna)] = []
                    self.pos[(axis,band,antenna)] = []
                    self.poscul[(axis,band,antenna)] = []
                    self.error[(axis,band,antenna)] = []
                    self.model1[(axis,band,antenna)] = []
                    self.model2[(axis,band,antenna)] = []
                    self.model3[(axis,band,antenna)] = []
        self.A={}
        self.B={}
                
    def readFile(self,uid):
        """
        read one data set and extract focus data calibration and focus model 
        """
        self.files.append(uid)
        a = ASDM()
        a.setFromFile(uid,True)
        mm=1e3
        calFocus = a.calFocusTable().get()
        antTable = a.antennaTable().get()
        antnum={}
        for row in antTable:
            id = row.antennaId()
            ids = str(id).split('_')
            antn = int(ids[1])
            antenna = row.name()
            antnum[antenna] = antn
        nants = len(antnum)
        models = zeros((nants,14))
        focusModel = a.focusModelTable().get()
        for row in focusModel:
            modelId = row.focusModelId()
            if modelId == 0:
                id = row.antennaId()
                ids = str(id).split('_')
                antn = int(ids[1])
                models[antn,:] = row.coeffVal()
        models = models*mm
        #print antn,models
        bandnum={'ALMA_RB_03':3, 'ALMA_RB_06':6, 'ALMA_RB_07':7, 'ALMA_RB_09':9}
        axisnum={'X':1,'Y':2,'Z':3}
        self.lines.append('#see http://wikis.alma.cl/bin/view/AIV/ExtractFocusRun for a detailed explanation of the output')
        self.lines.append('#Focus-offset,offset-error,Focus-position: [mm], Azimuth,Elevation: [deg], Temperature: [K]')
        self.lines.append('#--Date/time-----------------|--Temp-------|Ant-|Axis|--Band------|Foc-offs-|offs-err-|--Az----|--El--|Foc-pos|')
        for row in calFocus:
            fr = row.frequencyRange()[0].get()
            antenna = row.antennaName()
            temperature = row.ambientTemperature()
            temp = float(str(temperature).split('K')[0])
            band = getBand(fr)
            for i in range(3):
                if not row.wereFixed()[i]:
                    axis = self.axes[i] 
                    focus = row.offset()[0][i].get()*mm
                    error = row.offsetError()[0][i].get()*mm
                    pos = row.position()[0][i].get()*mm
            az, el = row.pointingDirection()
            azd = degrees(az.get())
            eld = degrees(el.get())
            kind = (axis, band, antenna)
            if axis == 'Z':
                res =  models[antnum[antenna],12]*(m.sin(m.radians(eld))-m.sin(m.radians(50.))) + models[antnum[antenna],13]*(temp-273.15) + models[antnum[antenna],11]
                resm1 = models[antnum[antenna],12]
                resm2 = models[antnum[antenna],13]
                resm3 = models[antnum[antenna],11]
            if axis == 'Y':
                res =  models[antnum[antenna],6]*(m.cos(m.radians(eld))-m.cos(m.radians(50.))) + models[antnum[antenna],7]
                resm1 = models[antnum[antenna],6]
                resm2 = 0.
                resm3 = models[antnum[antenna],7]
            if axis == 'X':
                res = models[antnum[antenna],3]
                resm1 = 0.
                resm2 = 0.
                resm3 = models[antnum[antenna],3]                
            self.elevation[kind].append(degrees(el.get()))
            self.azimuth[kind].append(degrees(az.get()))
            self.temperature[kind].append(temp)
            focus2=focus
            if len(self.focuscul[kind]) > 0:
                focus2=focus2+self.focuscul[kind][-1]
            #pos2 = res+focus2-focus
            pos2 = res
            self.focuscul[kind].append(focus2)
            self.focus[kind].append(focus)
            self.pos[kind].append(pos2+focus)
            self.error[kind].append(error)
            self.model1[kind].append(resm1)
            self.model2[kind].append(resm2)
            self.model3[kind].append(resm3)
            if self.startTime=='':
                self.startTime=row.startValidTime().toFITS()
            self.endTime = row.startValidTime().toFITS()
            self.lines.append('%s %s %s %s %s %s  %8.4f %8.4f %8.2f %8.2f %8.4f %8.4f' % \
                              (row.startValidTime().toFITS(), temperature, antenna, axis, '  ', band, focus,
                               error, azd, eld, pos, pos2))
        return

    def writeData(self,outfile):
        out_file = open(outfile,'w')
        for line in self.lines:
            out_file.write(line+'\n')
        out_file.close()
                    
    def plot(self, interactive):

        os.system('rm *%s*.pdf'%(self.files[0]))
        
        self.bandsym={'ALMA_RB_03':'g','ALMA_RB_06':'b','ALMA_RB_07':'r','ALMA_RB_09':'m'}
        self.axiscol={'X':'x','Y':'o','Z':'s'}
        ydep={'X': cos, 'Y': cos, 'Z': sin} 

        fig = 1
        for antenna in self.antenna:
            print "%s" %antenna
            figure(fig,figsize=(12,8));clf();
            subplot(3, 1, 1)
            subplots_adjust(hspace=0.37)
            blen=0
            for band in self.bands:
                blen += len(self.focus[('Z',band,antenna)])
            #print blen
            if blen > 0:
                for axis in self.axes:
                    for band in self.bands:
                        if len(self.elevation[(axis,band,antenna)]) > 0:
                            code = self.bandsym[band] + self.axiscol[axis]
                            errorbar(self.elevation[(axis,band,antenna)],self.focus[(axis,band,antenna)],yerr=self.error[(axis,band,antenna)],fmt=code, label = axis)
                            
                figtext(0.02, 0.98, '%s %s'%(self.files.__str__(),antenna), fontsize=12, ha='left',va='top')
                figtext(0.02, 0.96, self.startTime+' - '+self.endTime, fontsize=12, ha='left',va='top')
                title('Relative Focus offset')
                xlabel('elevation [degrees]')
                ylabel(' focus [mm]')
                legend(loc=2)
                
                subplot(3, 1, 2)
                subplots_adjust(hspace=0.37)
                axis = 'Z'
                for band in self.bands:
                    lenel=len(self.elevation[(axis,band,antenna)])
                    if lenel > 0:
                        code = self.bandsym[band] + self.axiscol[axis]
                        coeff=[]
                        for i in range(lenel):
                            coeff.append(self.pos[(axis,band,antenna)][i]-self.model1[(axis,band,antenna)][0]*(sin(self.elevation[(axis,band,antenna)][i]*pi/180.)-sin(m.radians(50))))
                        #print coeff,self.temperature[(axis,band,antenna)],self.pos[(axis,band,antenna)]
                        errorbar(self.temperature[(axis,band,antenna)],coeff,yerr=self.error[(axis,band,antenna)],fmt=code)
                title('Absolute Z focus wrt Temperature')
                ylabel('focus [mm]')
                xlabel('temperature[Kelvin]')
                
                subplot(3, 1, 3)
                subplots_adjust(hspace=0.37)
                temp0=273.15
                for axis in self.axes:
                    for band in self.bands:
                        if len(self.elevation[(axis,band,antenna)]) > 0:
                            code = self.bandsym[band] + self.axiscol[axis]
                            errorbar(self.elevation[(axis,band,antenna)],self.pos[(axis,band,antenna)],yerr=self.error[(axis,band,antenna)],fmt=code)
                            elsup = arange(0.,90.,1)
                            for i in range(0,len(self.model1[(axis,band,antenna)])-1):
                                #print self.model1[(axis,band,antenna)][i],self.model2[(axis,band,antenna)][i],self.model3[(axis,band,antenna)][i]
                                foc=self.model3[(axis,band,antenna)][i]+self.model1[(axis,band,antenna)][i]*(ydep[axis](elsup*pi/180.)-ydep[axis](50.*pi/180.))+self.model2[(axis,band,antenna)][i]*(temp0-273.15)
                                plot(elsup,foc,self.bandsym[band]+'-')

                xlabel('elevation[degrees]')
                title('Absolute focus')
                ylabel('focus[mm]')

                if interactive:
                    show()
                savefig('%s_%s.pdf'%(self.files[0],antenna), format='pdf')
                fig +=1
                
        print "Converting into a single PDF"
        os.system('convert -adjoin *.pdf focus_results_%s.pdf'%(self.files[0]))
        os.system('rm uid*.pdf')
        print 'Results and plots are visible in focus_results_%s.pdf'%(self.files[0])
                
    def fit(self, axis, band=None):
        ytext={'X': 0.90, 'Y': 0.86, 'Z': 0.82} 
        ydep={'X': cos, 'Y': cos, 'Z': sin} 
        col={'X': 'red', 'Y': 'green', 'Z': 'blue'} 
        xx = []
        yy = []
        for b in self.bands:
            if (band == None) or (b == band):
                xx.extend(self.elevation[(axis,b)])
                yy.extend(self.focuscul[(axis,b)])
        ax = array(xx)*pi/180.
        ay = array(yy)
        fitfunc= lambda p, x: p[0] + p[1]*ydep[axis](x)
        errfunc = lambda p, x, y: fitfunc(p, x)-y
        p0=[0.,1.]
        p1, success = scipy.optimize.leastsq(errfunc, copy.copy(p0), args=(ax, ay))
        self.B[axis] = p1[1]
        self.A[axis] = p1[0]
        el = arange(0.,90.,1)
        fit = self.A[axis] + self.B[axis]*ydep[axis](el*pi/180.)
        plot(ax, ay, 'r.')
        plot(el,fit,self.axiscol[axis]+'-')
        ytext={'X': 0.90, 'Y': 0.86, 'Z': 0.82} 
        figtext(0.02,
                ytext[axis],
                '%s in red:   %6.3f + %6.3f %s(el)' % (axis, self.A[axis], self.B[axis],
                                                       str(ydep[axis])), 
                fontsize=12, ha='left',va='top')
        

# Execution starts here
if __name__ == '__main__':

    # Initialize parser
    parser = OptionParser()

    parser.add_option("-f", "--files", dest="files",
                  help="Input files to be analysed (separated by white space)",
                  action="callback", callback=files_callback)

    parser.add_option("-o", "--outfile", dest="outfile",
                      help="OUtput file name",
                      action='store')

    parser.add_option("-i","--interactive",dest="interactive",metavar="interactive", action = 'store_true',
                     default=False, help="Use this option if you want to produce plots as you go. Default False") 

    (options, args) = parser.parse_args()

    files = options.files
    if options.outfile is None:
        ts = makeTimeStamp()
        outfile = 'focus_results_%s.txt'%ts
    else:
        outfile = options.outfile
    
        
    ef = ExtractFocus()

    print "Reading in data ..."
    for uid in files:
        ef.readFile(uid)

    print "Writing output to %s" %outfile
    ef.writeData(outfile)

    print "Generating plots for ..."
    ef.plot(options.interactive)

        


