#!/usr/bin/env python
#
#NAME: 
#  FrontEndTuningAnalysis.py
#
#PURPOSE:
#  To analyzethe data taken during tests executed with FrontEndTuning.py.  The 
#  analysis consists of:
#    1) Plotting the Receiver temperature vs. frequency for each antenna
#    2) Plotting the locking time vs. frequency 
#
#INPUTS:
#  FILENAME: path to file
#    The path to the file FrontEndTuningTest_[Date].txt produced by 
#    the FrontEndTuning.py script.
#
#EXAMPLES:
#  FrontEndTuningAnalysis.py -f '/data/RADIO/SYSTEM/DV03_DV05/20100421/FrontEndTuningTest_20100421.txt'
#
#Written by Lance Simms on April 23, 2010
#
#Modified:
#
#$Id: FrontEndTuningAnalysis.py,v 1.1 2010/04/23 20:37:06 lsimms Exp $
###################################################################
###################################################################
from numpy import *
from optparse import OptionParser
import pdb
import matplotlib.pylab as mplot

def ReadFETData(FileName):

  #Open the file and get the size
  fd=open(FileName, 'r').readlines()
  fd_size = size(fd)

  #Remove any lines with the header
  j = 0 #Number of lines removed
  for i, line in zip(arange(fd_size), fd):
    if line.find('SkyFrq') != -1 : 
      fd.pop(i-j)
      j+=1

  #Get new size with header lines removed
  fd_size = size(fd)

  #Create arrays to hold the values
  Antenna  = zeros(fd_size, chararray)
  Band     = zeros(fd_size, int)
  Freqs    = zeros(fd_size, float)
  LockTime = zeros(fd_size, float)
  IFPowH0  = zeros(fd_size, float)
  IFPowH1  = zeros(fd_size, float)
  THot0    = zeros(fd_size, float)
  THot1    = zeros(fd_size, float)
  THot2    = zeros(fd_size, float)
  IFPowC0  = zeros(fd_size, float)
  IFPowC1  = zeros(fd_size, float)
  TAmb0    = zeros(fd_size, float)
  TAmb1    = zeros(fd_size, float)
  TAmb11   = zeros(fd_size, float)
  TAmb22   = zeros(fd_size, float)
  LSFreq   = zeros(fd_size, float)
  WVAFreq  = zeros(fd_size, float)
  CRSTune  = zeros(fd_size, float)
  PmCrnt   = zeros(fd_size, float)
  IFPwr    = zeros(fd_size, float)
  LckVolt  = zeros(fd_size, float)
  Times    = zeros(fd_size, float)
  Trxs     = zeros(fd_size, float)

  for i, fd_line in zip(arange(fd_size), fd):
    fd_line = fd_line.split()
    Antenna[i] = fd_line[0]
    Band[i]    = fd_line[1]
    Freqs[i]   = float(fd_line[2])
    LockTime[i]= float(fd_line[3])
    IFPowH0[i] = 10**((float(fd_line[4]))/10)/1000
    IFPowH1[i] = 10**((float(fd_line[5]))/10)/1000
    THot0[i]   = float(fd_line[6])
    THot1[i]   = float(fd_line[7])
    THot2[i]   = float(fd_line[8])
    IFPowC0[i] = 10**((float(fd_line[9]))/10)/1000
    IFPowC1[i] = 10**((float(fd_line[10]))/10)/1000
    TAmb0[i]   = float(fd_line[11])
    TAmb1[i]   = float(fd_line[12])
    TAmb11[i]  = float(fd_line[13])
    TAmb22[i]  = float(fd_line[14])
    WVAFreq[i] = float(fd_line[15])
    CRSTune[i] = float(fd_line[16])
    PmCrnt[i]  = float(fd_line[17])
    IFPwr[i]   = float(fd_line[18])
    LckVolt[i] = float(fd_line[19])
    Times[i]   = float(fd_line[20])
    Trxs[i]    = float(fd_line[21])

  return     Antenna, Band, Freqs, IFPowH0, IFPowH1, THot0, THot1, THot2,\
             IFPowC0, IFPowC1, TAmb0, TAmb1, TAmb11, TAmb22,\
             WVAFreq, CRSTune, PmCrnt, IFPwr, LckVolt, Times, Trxs

###################################
#Main execution starts here
###################################
# Execution starts here
if __name__ == '__main__':

  # Initialize parser
  parser = OptionParser()

  parser.add_option("-f", "--filename", dest="filename",
                help="Input file to be analysed",
                action="store")
  (options, args) = parser.parse_args()
  
  #Get the filename
  filename = options.filename
  
  #Get the date of the filename
  dateloc  = filename.find('Test_')
  date     = filename[dateloc+5:dateloc+13]
  #Read all the values from the file
  Antenna, Band, Freqs, IFPowH0, IFPowH1, THot0, THot1, THot2,\
           IFPowC0, IFPowC1, TAmb0, TAmb1, TAmb11, TAmb22,\
           WVAFreq, CRSTune, PmCrnt, IFPwr, LckVolt, Times, Trxs = \
           ReadFETData(filename)

  #Find out the unique antenna values
  uniAnt = unique(Antenna)

  #Plot numbers need initialization
  PlotNum = 0

  for Ant in uniAnt:
    #Get indices for this antenna
    ThisAntVals = (where(Antenna==Ant))[0]
    ThisTrxs    = Trxs[ThisAntVals]
    ThisFreqs   = Freqs[ThisAntVals]
  
    #Plot
    mplot.figure(PlotNum)
    mplot.plot(ThisFreqs, ThisTrxs, 'bo')
    mplot.title('Receiver Temperature for Front End of ' + Ant)
    mplot.xlabel('Frequency [GHz]')
    mplot.ylabel('Temperature [K]')
    
    #Save the file
    mplot.savefig('ReceiverTemperatures_'+Ant+'_'+date+'.png')

    #increase the plot number
    PlotNum+=1

  #Inform the user that matplotlib will hang since python is not configured to
  #run in interactive mode on these machines
  print '\nIn order to get back to the shell, you must close the plots (unless someone decides to set interactive mode in the python environment).\n'
   
  #Show the plots
  mplot.show()
