#!/usr/bin/env python
"""

Goal: span the whole receiver band and observe a target at each frequency setting to perform a wideband survey of
Tsys and sideband ratio
      

$Id: SideBandRatio-survey.py,v 1.1 2012/09/14 00:27:41 jbulger Exp $

Example: SideBandRatio-survey.py -s 1924-292 -b 3

Modifications:
2012-09-13: spectralSpec for ACA & BL correlators

"""
testName = "spectralSurvey.py"

# Options section
from optparse import OptionParser
import sys
import math
import pylab as pl

parser = OptionParser()

parser.add_option("-a", "--arrayName",
                  dest="arrayName",
                  metavar="ArrayName",
                  help="The name of the array to be used.")

parser.add_option("-b", "--band",
                  dest="band",
                  metavar="Band",type=int,
                  help="The band of observation. Can be 3,6,7,9 only")

parser.add_option("-d", "--subscanDuration",
                  dest='subscanDuration',
                  default = 60,
                  type = float,
                  help = 'Time spent, on each source.')

parser.add_option("-c", "--channelAverageDuration",
                  dest="channelAverageDuration",
                  metavar="ChannelAverageDuration",
                  default=4.0, type = float,
                  help="Duration of channel average integration in seconds")

parser.add_option("-e", "--minEl",
                  dest="minEl",
                  metavar="MinElevation",
                  default = 10.0, type = float,
                  help="Minimum elevation required on source (degrees)")

parser.add_option("-i", "--integrationDuration",
                  dest="integrationDuration",
                  metavar="IntegrationDuration",
                  default = 4.0, type = float,
                  help="Duration of spectral integration in seconds")

parser.add_option("-n", "--mode",
                  dest="mode",
                  metavar="Correlator Mode",
                  default = "TDM",
                  help="The correlator mode: TDM or FDM. TDM=default")

parser.add_option("-s", "--source",
                  dest="source",
                  metavar="Source",
                  help="The source name. Can be a comma separated list.")

parser.add_option("-u","--dumpt", dest="dumpt",
                  metavar="dumpt",
                  default = 4.032, type = float,
                  help="correlator dump time. Default=4.032s. " + \
                  "Make it a multiple of 48ms")

parser.add_option("-P", "--polarization",
                  dest="polarization",
                  metavar="Polarization",
                  default = "2",
                  help="The polarization")

parser.add_option("-T", "--tsys",
                  dest = "tsys",
                  metavar = "Tsys", 
                  action="store_true",
                  default= True,
                  help = "If true (=default), atmCalTarget (a.k.a. tsys measurement) done before each tuning")

parser.add_option("-D", "--DelayCal",
                  dest = "DelayCal",
                  metavar = "DelayCal", 
                  action="store_false",
                  default= True,
                  help = "If true (default), DelayCal observation is not done at each tuning. Use False only for Tsys measurements. Currently it is set not to do the DelayCal")

parser.add_option("-F", "--freqListIndice",
                  dest="freqListIndice",
                  metavar="FreqListIndice",
                  help="Indice of the frequency list item(s) to be tuned. " + \
                  "Counted from 1, separated by commas. " + \
                  "If not specified, tuned to all frequencies in the list.")

parser.add_option("-N", "--subscanRepeats",
                  dest="subscanRepeats",
                  metavar="SubscanRepeats",
                  default = 1, type = int,
                  help="The number of subscans to repeat on each source")

parser.add_option("--180", "--180",
                  dest="do180degPhaseSwitch",
                  metavar="Do180degPhaseSwitch",
                  action="store_true", default=False,
                  help = "If set, the 180 degree phase switching is turned on.  If not set, it turns it off.")

#parser.add_option("-O", "--LOOffset",
#                  dest="LOOffset",
#                  metavar="LOOffset",
#                  action="store_true", default=False,
#                  help="If set, LOOffsetting will be enabled with the factors set in the tmcdb for the various basebands.")

parser.add_option("--image","--imageSideband", dest="imageSideband",
                  metavar="ImageSideband", default=False,
                  action="store_true",
                  help="If this option is set along with -O for LOOffsetting, the signal sideband is rejected and only the image sideband is returned")


(options, args) = parser.parse_args()

ssD = options.subscanDuration
if ssD > 100:
    print " We are currently limiting the subscanDuration to less than 100s. ( subscan file size limit = 2Gb)."
    print " For longer scans, just use the -d 100 -N 12 option to repeat 12 subscans per scans "
    sys.exit()

ssR = options.subscanRepeats
scanDuration = ssD*(ssR-0.5)

import generalUtil   as gU
import sourceCatalog as sC
import time
from CCL.FieldSource import *
from CCL.FLOOG import FLOOG
from CCL.LO2 import LO2
from CCL.FrontEnd import FrontEnd

# Setup Section
from CCL.Global import *
from Observation.DelayCalTarget import DelayCalTarget
from Observation.AtmCalTarget   import AtmCalTarget
from Observation.AmplitudeCalTarget   import AmplitudeCalTarget
from PyDataModelEnumeration import PyCalibrationDevice,PyScanIntent,PySubscanIntent
from Observation.SBRatioCalTarget import SBRatioCalTarget
from Observation.DelayCalTarget import DelayCalTarget
import Control
from time import sleep

if options.arrayName is None : 
    arrayName = gU.getArrayName()
    if arrayName is None:
        sys.exit(-1)
else: 
    arrayName = options.arrayName

setArrayName(arrayName)
array = getArray()
antennaList = array.antennas()
ant0=antennaList[0]
numberAntennas = len(antennaList)

# Import CSVLogger and set up logger tool:
from CSVLogger import CSVLogger
#Imports for Error Handling
import ACSErrTypeCommonImpl

logger = CSVLogger("spectralSurvey.py", arrayName)

#define spectra list
if options.band == 3:
    subbands = [[85.959e9, 87.834e9, 98.004e9, 99.879e9],
                [89.709e9, 91.584e9, 101.754e9, 103.629e9],
                [93.459e9, 95.334e9, 105.504e9, 107.379e9],
                [97.205e9, 99.080e9, 109.250e9, 111.125e9],
                [100.955e9, 102.830e9, 113.000e9, 114.875e9]]
elif options.band == 6:
    subbands = [[212.525e9, 214.400e9, 227.600e9, 229.475e9],
                [216.275e9, 218.150e9, 231.350e9, 233.225e9],
                [220.025e9, 221.900e9, 235.100e9, 236.975e9],
                [223.775e9, 225.650e9, 238.850e9, 240.725e9],
                [242.600e9, 244.475e9, 258.475e9, 260.350e9],
                [246.350e9, 248.225e9, 262.225e9, 264.100e9],
                [250.100e9, 251.975e9, 265.975e9, 267.850e9],
                [253.850e9, 255.725e9, 269.725e9, 271.600e9],
                [256.125e9, 258.000e9, 272.000e9, 273.875e9]]
elif options.band == 7:
    subbands =[[276.125e9, 278.000e9, 288.044e9, 289.919e9],
               [279.875e9, 281.750e9, 291.794e9, 293.669e9],
               [283.625e9, 285.500e9, 295.544e9, 297.419e9],
               [299.294e9, 301.169e9, 311.213e9, 313.088e9],
               [303.044e9, 304.919e9, 314.963e9, 316.838e9],
               [306.794e9, 308.669e9, 318.713e9, 320.588e9],
               [322.463e9, 324.338e9, 334.382e9, 336.257e9],
               [326.213e9, 328.088e9, 338.132e9, 340.007e9],
               [329.963e9, 331.838e9, 341.882e9, 343.757e9],
               [345.632e9, 347.507e9, 357.551e9, 359.426e9],
               [349.382e9, 351.257e9, 361.301e9, 363.176e9],
               [353.132e9, 355.007e9, 365.051e9, 366.926e9],
               [358.022e9, 359.897e9, 370.103e9, 371.978e9],
               [354.434e9, 356.309e9, 366.353e9, 368.228e9],
               [284.825e9, 286.700e9, 296.744e9, 298.619e9],
               [307.994e9, 309.869e9, 319.913e9, 321.788e9],
               [331.163e9, 333.038e9, 343.082e9, 344.957e9]]
elif options.band == 9:
    subbands = [[603.022e9, 604.897e9, 606.772e9, 608.647e9],
                [610.522e9, 612.397e9, 614.272e9, 616.147e9],
                [618.022e9, 619.897e9, 621.772e9, 623.647e9],
                [625.522e9, 627.397e9, 629.272e9, 631.147e9],
                [633.022e9, 634.897e9, 636.772e9, 638.647e9],
                [640.522e9, 642.397e9, 644.272e9, 646.147e9],
                [648.022e9, 649.897e9, 651.772e9, 653.647e9],
                [655.522e9, 657.397e9, 659.272e9, 661.147e9],
                [663.022e9, 664.897e9, 666.772e9, 668.647e9],
                [670.522e9, 672.397e9, 674.272e9, 676.147e9],
                [678.022e9, 679.897e9, 681.772e9, 683.647e9],
                [685.522e9, 687.397e9, 689.272e9, 691.147e9],
                [693.022e9, 694.897e9, 696.772e9, 698.647e9],
                [700.522e9, 702.397e9, 704.272e9, 706.147e9],
                [708.022e9, 709.897e9, 711.772e9, 713.647e9],
                [713.353e9, 715.228e9, 717.103e9, 718.978e9]]
else:
    print "No band %s tunings yet. Go back home" % options.band
    sys.exit(-1)

if options.freqListIndice == None:
    subband = subbands
else:
    subband = []
    for idx in [int(item) for item in options.freqListIndice.split(',')]:
        if (idx < 1) or (idx > len(subbands)):
            print "Freq list index %d is out of range (1-%d). Ignored." % \
                (idx, len(subbands))
        else:
            subband.append(subbands[idx-1])
    if len(subband) == 0:
        print "The indices given with -F option seem invalid.  Go back home"
        sys.exit()

n_tunings = pl.shape(subband)[0]

# define other hard wired variables:
corrmode = options.mode
width = 1
sourceName = options.source
fe = FrontEnd(ant0) 
floog = FLOOG(ant0)
lo2 = LO2(ant0, 0)

# Overall try / except to make sure we call endExecution
try:
    array.beginExecution()
    mode = array.getSingleFieldInterferometryObservingMode()
        
    # Initial CSVLogger statement to note command-line options and ExecBlock UID:
    localAsdm = array.getExecBlockUID()
    logger.logOptionsASDM(options, localAsdm)
    
    from CCL.SpectralSpec import *
    # Loop over tunings
    for j in range(n_tunings):
        
        obsFreq = gU.getFreq(pl.mean(subband[j][:]))
        if obsFreq == None :
            print "Stopping because of a bad frequency.  Think about it and get back to me."
            sys.exit()
        bandNum = gU.getBand(obsFreq, False, True)
        basebandFreqs = subband[j][:]
        basebandNames = ['BB_1', 'BB_2', 'BB_3', 'BB_4']
        print "Baseband frequencies are ", basebandFreqs, " Hz.  "
                        
        numBaseBand = len(basebandFreqs)
        frequencies = gU.makeList(basebandFreqs)
    
        # Get the Spectral Spec
        #spectralSpec = CCL.SpectralSpec.SpectralSpec().CreateBLCorrelatorSpec(\
        spectralSpec = gU.createCorrelatorSpec\                       
                       (corrMode = corrmode,
                       dump = options.dumpt,
                       channelAverage = options.channelAverageDuration,
                       integration = options.integrationDuration,
                       baseband = basebandNames,
                       bwd = width,
                       frequency = obsFreq,
                       basebandFrequency = frequencies,
                       pol = options.polarization)

        
        #spectralSpecTDM= CCL.SpectralSpec.SpectralSpec().CreateBLCorrelatorSpec(\
        spectralSpecTDM = gU.createCorrelatorSpec\
                       (corrMode = 'TDM',
                        dump = options.dumpt,
                        channelAverage = options.channelAverageDuration,
                        integration = options.integrationDuration,
                        baseband = basebandNames,
                        bwd = width,
                        frequency = obsFreq,
                        basebandFrequency = frequencies,
                        pol = options.polarization)
                   
        spectralSpec.FrequencySetup.receiverBand = unicode('ALMA_RB_%2.2d'%bandNum)
        spectralSpecTDM.FrequencySetup.receiverBand = unicode('ALMA_RB_%2.2d'%bandNum)
        
        if options.do180degPhaseSwitch :
            print "enabling 180 Walsh switching"
            spectralSpec.set180degSwitching(True)

# always use 2LO offsetting in this test:
        spectralSpec.setLOOffsetting("TWO_LOS")
        spectralSpecTDM.setLOOffsetting("TWO_LOS")
                    
        # Define our source List
        if sourceName is None :
            sourceList = sC.getContinuumList(fluxLimit=options.fluxLimit, doPlanet=False)
        else :
            try:
                if sourceName.lower() in sC.sourceContinuum.keys() :
                    sourceList = {sourceName:sC.sourceContinuum[sourceName.lower()]}
                elif sourceName.lower() in sC.sourceSpectralLine.keys() :
                    sourceList = {sourceName:sC.sourceSpectralLine[sourceName.lower()]}
                elif sourceName.lower() in sC.sourcePlanets.keys() :
                    sourceList = {sourceName:sC.sourcePlanets[sourceName.lower()]}
            except Exception, ex:
                print ex
                logger.logError("Source %s is not found in the planet catalog.  Giving up." % sourceName)
                raise

        for source in sourceList:
            src = sC.getEquatorialSource(source, band=bandNum)
            if source in sC.sourceLimited.keys() :
                doObs = mode.isObservable(src, 1800.+scanDuration)
            else :
                mode.setElevationLimit(math.radians(options.minEl))
                doObs = mode.isObservable(src, 300)
            if doObs:
                try:
                    sbRatioCal = SBRatioCalTarget(src, spectralSpec)
                    sbRatioCal.setSubscanDuration(options.subscanDuration)
                    sbRatioCal.setIntegrationTime(sbRatioCal.getSubscanDuration()*
                                                    (options.subscanRepeats-0.5))
                    sbRatioCal.setOnlineProcessing(True)
                    sbRatioCal.execute(mode)

                                        
                    if (options.tsys):
                        atm = AtmCalTarget(src,spectralSpecTDM,doHotLoad=True)
                        atm.setOnlineProcessing(True)
                        atm.setDataOrigin('FULL_RESOLUTION_AUTO')
                        atm.setDoZero(False)
                        atm.setSubscanDuration(5.)
                        atm.setIntegrationTime(1.)
                        atm.execute(mode)
                        
#                    if (options.DelayCal):
#                        delayCal = DelayCalTarget(src, spectralSpec)
#                        delayCal.setSubscanDuration(ssD)
#                        scanDuration = delayCal.getSubscanDuration()*(ssR-0.5)
#                        delayCal.setIntegrationTime(scanDuration)
#                        logger.logInfo("Observing %s" % src.sourceName)
#                        delayCal.execute(mode)
                        
                    lo1_freq = fe.getLO1Frequency()
                    floog_freq = floog.GET_FREQ()[0]
                    lo2_fts = lo2.GET_FREQ()[0]
                    lo2_freq = lo2.GET_LO2_FREQUENCY()[0]
                    print "LO1, FLOOG, LO2, LO2-FTS (Hz) =  ", lo1_freq, floog_freq, lo2_freq, lo2_fts
                    sleep(3.0)
                        
                except Exception, ex:
                    print ex
                    logger.logError("Error observing %s" % src.sourceName)
                    raise
            else:
                logger.logInfo("Skipping %s" % src.sourceName)
                    
            # Reset the elevation limit
            mode.resetLimits()

    logger.logInfo("Observing done...")
    status = (Control.SUCCESS, testName)

except KeyboardInterrupt:
    logger.logError("^C Pressed!")
    logger.logInfo("Trying to close down...")
    status = (Control.SUCCESS, testName + ' (interrupted)')
    
except Exception, ex:
    logger.logCritical("Some sort of exception was thrown... go look at logs.")
    helperException = ACSErrTypeCommonImpl.UnknownExImpl(exception=ex)
    helperException.Print()
    status = (Control.FAIL, "Exception thrown in "+ testName)
    
finally:
    try:
        mode.resetLimits()
    except:
        logger.logError("no observing mode ...")

array.endExecution(*status)
logger.logInfo("End of Execution, uid is %s" % localAsdm)
logger.logInfo("Trying to export the asdm...")

gU.autoExportAsdm("RADIO", "SYSTEM", antennaList, localAsdm)
logger.logInfo("End of Execution, asdm exported ")

# Final CSVLogger statement to note successful completion:
logger.logComplete()

