#!/usr/bin/env python
"""
OTF test script
   usage: analyzeOTFPSC.py -f position.dat

   2010-08-17 rsimon:  add option doplots 
   2010-07-16 tsawada: compatibility with the new PSC data format
   2010-03-02 tsawada: remove invalid data
   2010-01-15 tsawada: workaround to skip extra messages seen on 2010-01-14
   2009-08-12 tsawada

Analyze PSC data and assess the performance of the antenna.

Pointing fluctuation (RMS) during tracking is calculated.
usage:
  analyzeOTFPSC.py -f position.dat  (see help for more options)

$Id: analyzeOTFPSC.py,v 1.3 2010/08/17 18:11:43 rsimon Exp $
"""

import os, sys
import pylab as pl
from optparse import OptionParser

###--------------------------------
### definitions of functions
###--------------------------------

def string_pass(passed):
    """
    If True return "PASSED ", else return "FAILED!".
    """
    if passed:
        return "PASSED "
    else:
        return "FAILED!"

def hms_to_sec(hmsstring):
    """
    Return seconds (float) from a "HH:MM:SS.s" or "DD:MM:SS.s" string.
    
    The argument hmsstring is "HH:MM:SS.s", "MM:SS.s", or "SS.s".
    Negative sign (if exists) must be the first character of hmsstring.
    """
    result = 0.
    sign = 1.

    hms = hmsstring.strip().split(":")
    if hms[0][0] == "-":
        sign = -1.
        hms[0] = "0"+hms[0][1:]
    nlen = len(hms)
    for i in range(nlen):
        result = result + float(hms[i]) * 60.**(nlen-1-i)

    return sign*result

def normalize_timestamp(timearray):
    """
    Return normalized array of timestamps [seconds].

    If the timestamps cross 0/24 hours,
    [12 to 24] hours are shifted to [-12 to 0] hours.
    """
    if (max(timearray)-min(timearray)) > 43200.:
        return timearray - 86400.*(timearray > 43200.)
    else:
        return timearray

def azel_to_dec(azim, elev, lat):
    """
    Return declination from given az, el, and latitude [all in degrees].
    """
    dtor = pl.pi/180.
    #ha  = ( pl.arctan2( -pl.sin(azim*dtor)*pl.cos(elev*dtor), \
    #       -pl.cos(azim*dtor)*pl.cos(elev*dtor)*pl.sin(lat*dtor) + \
    #        pl.sin(elev*dtor)*pl.cos(lat*dtor) ) \
    #      ) / dtor
    dec = ( pl.arcsin( pl.cos(azim*dtor)* \
            pl.cos(elev*dtor)*pl.cos(lat*dtor) + \
            pl.sin(elev*dtor)*pl.sin(lat*dtor) ) \
          ) / dtor

    return dec

def read_psc(filename, return_array = False):
    """
    Read a PositionStreamClient data and return a 10-key dictionary.

    Each entry is an array (if return_array is True) or a list (otherwise).
    Keys are:
    timeStamp [seconds since UT0:00],
    cmdAz, cmdEl, mteAz, mteEl, oteAz, oteEl, encAz, encEl [degrees],
    onSource  [bool]
    """
    keys = ['timeStamp', 'cmdAz', 'cmdEl', 'mteAz', 'mteEl',
        'oteAz', 'oteEl', 'encAz', 'encEl', 'onSource']

    data = [[] for i in xrange(10)]
    for line in open(filename, "r"):
        fields = line.replace("invalid", "-99999.0").strip().split()
        if fields[0].find("T") == 10:
            ### remove date: for new PSC R7.1.1
            fields[0] = fields[0][11:]
        if ("0" <= fields[0][0] <= "2") and (fields[0][2] == ":"):
            for i, val in enumerate(fields):
                if i == 0:
                    data[i].append(hms_to_sec(val))
                elif i < 9:
                    data[i].append(float(val))
                elif i == 9:
                    data[i].append(val == "On")

    pscdata = {}
    for i, key in enumerate(keys):
        if return_array:
            pscdata[key] = pl.array(data[i])
        else:
            pscdata[key] = data[i]

    return pscdata

def timerange_psc(pscdata, trange):
    """
    Mask PSC data with the time range.

    The input and output are dictionary of arrays.
    """
    datetime = normalize_timestamp(pscdata['timeStamp'])
    pscdata['timeStamp'] = datetime

    in_trange = (datetime != pl.NaN)
    if not trange == None:
        trange = trange.split(':')
        if len(trange) == 2:
            in_trange = (float(trange[0]) <= datetime) & \
                        (datetime <= float(trange[1]))
                       
    if len(datetime[in_trange]) == 0:
        print "WARNING: Time range is invalid and ignored."
        in_trange = (datetime != pl.NaN)

    for key in pscdata.keys():
        pscdata[key] = pscdata[key][in_trange]

    return pscdata

def remove_invalid(pscdata):
    """
    Remove columns which contain invalid (-99999.) data.
    """
    keys = ['cmdAz', 'cmdEl', 'mteAz', 'mteEl', 'oteAz', 'oteEl']
    isvalid = (pscdata[keys[0]] > -100000.)
    for key in keys:
        isvalid = isvalid & (pscdata[key] != -99999.)
    for key in pscdata.keys():
        pscdata[key] = pscdata[key][isvalid]

    return pscdata

def judge_onscan(pscdata, masktime):
    """
    Return an array of on-scan flag.
    
    Given an dictionary of PSC data, the part in which the antenna was
    commanded to follow non-sidereal track is regarded as 'on-scan'.
    """
    dtor = pl.pi / 180.

    sidereal_vel = 360. / hms_to_sec("23:56:04") * pl.cos( \
        azel_to_dec(pscdata['oteAz'], pscdata['oteEl'], -23.0) * dtor)
    onscan = ( \
        abs(pl.sqrt(pscdata['cmdVelAz']**2 + pscdata['cmdVelEl']**2) - \
        sidereal_vel[:-1]) > 2./3600.)
    off_to_on = onscan & pl.append(False, pl.diff(onscan))
    for tzero in pscdata['timeStamp'][off_to_on]:
        onscan[((tzero <= pscdata['timeStamp']) &
                  (pscdata['timeStamp'] <= tzero+masktime))[:-1]] = False
    on_to_off = onscan & pl.append(pl.diff(onscan), False)
    onscan[on_to_off] = False

    return onscan

def process_psc(pscfilename, rmsthreshold, masktime, yrange, trange, doplots):
    """
    Analyze PSC data and assess the performance of the antenna.

    Pointing fluctuation (RMS) during tracking is calculated.
    """
    dtor = pl.pi/180.

    #--------------------------------
    # the data file is read and stored in a dictionary
    # (see docstring of the read_psc function); and then
    # trimmed with the given time range
    #--------------------------------
    data = remove_invalid(timerange_psc(read_psc(pscfilename, True), trange))

    #--------------------------------
    # add some more entries into the dictionary
    #--------------------------------
    cos_el    = pl.cos(data['cmdEl']*dtor)
    time_step = pl.diff(data['timeStamp'])

    data['errAz'] = (data['oteAz']-data['cmdAz'])*cos_el
    data['errEl'] =  data['oteEl']-data['cmdEl']
    data['cmdVelAz'] = pl.diff(data['cmdAz'])/time_step*cos_el[:-1]
    data['cmdVelEl'] = pl.diff(data['cmdEl'])/time_step

    #--------------------------------
    # distinguish 'on-scan' and 'off-scan' parts.
    # the part in which the antenna was commanded to follow
    # non-sidereal track is regarded as 'on-scan' (see docstring)
    #--------------------------------
    onscan = judge_onscan(data, masktime)
    offscan = (onscan == False)

    mean_az = pl.mean(data['oteAz'][onscan])
    mean_el = pl.mean(data['oteEl'][onscan])

    #--------------------------------
    # check the tracking error
    #--------------------------------
    # nominal RMS (x, y, total) in degrees
    err_std = [pl.sqrt(pl.mean(data['errAz'][onscan]**2)), \
               pl.sqrt(pl.mean(data['errEl'][onscan]**2)), \
               pl.sqrt(pl.mean(data['errAz'][onscan]**2 + \
                               data['errEl'][onscan]**2))]
    print "%s: RMS = %.3f\" (%.3f\" in Az, %.3f\" in El) at Az=%.1f, El=%.1f" \
           % (string_pass(err_std[2]*3600. < rmsthreshold), \
           err_std[2]*3600., err_std[0]*3600., err_std[1]*3600., \
           mean_az, mean_el)
    # RMS after removing constant offsets
    err_mean = [pl.mean(data['errAz'][onscan]), \
                pl.mean(data['errEl'][onscan])]
    err_std = [pl.sqrt(err_std[0]**2 - err_mean[0]**2), \
               pl.sqrt(err_std[1]**2 - err_mean[1]**2), \
               pl.sqrt(pl.mean(
                   (data['errAz'][onscan]-err_mean[0])**2 + \
                   (data['errEl'][onscan]-err_mean[1])**2))]
    print "After removing constant offsets (%.3f\" in Az, %.3f\" in El)" % \
          (err_mean[0]*3600., err_mean[1]*3600.)
    print "%s: RMS = %.3f\" (%.3f\" in Az, %.3f\" in El) at Az=%.1f, El=%.1f" \
          % (string_pass(err_std[2]*3600. < rmsthreshold), \
           err_std[2]*3600., err_std[0]*3600., err_std[1]*3600., \
           mean_az, mean_el)

    #--------------------------------
    # check the turn-around performance
    # skipped now (2009-06) -- to be implemented
    #--------------------------------

    #--------------------------------
    # plot the results unless doplots=n
    #--------------------------------
    if doplots == "n":
        sys.exit()
    pl.clf()

    # timestamp vs tracking error
    pl.subplot(2, 1, 1)
    pl.plot(data['timeStamp'][onscan], data['errAz'][onscan]*3600.,
            "r.", ms=1.0, label="dAz*cos(El)")
    pl.plot(data['timeStamp'][onscan], data['errEl'][onscan]*3600.,
            "b.", ms=1.0, label="dEl")
    pl.xlabel("Time [seconds since UT0:00]")
    pl.ylabel("Real-Cmd [arcsec]")
    pl.title("%s: (Az,El)=(%.1f,%.1f) [deg] (Mask = %.1f [s])" % \
             (pscfilename, mean_az, mean_el, masktime))
    if (yrange > 0):
        pl.ylim(-yrange, yrange)
    # keywords labelsep and pad will be deprecated in matplotlib 0.98:
    # shall be replaced with labelspacing and borderpad, respectively
    pl.legend(labelsep = 0, \
              pad = 0.1, \
              numpoints = 1, \
              prop = pl.matplotlib.font_manager.FontProperties(size = 10))

    # az vs el
    pl.subplot(2, 1, 2)
    pl.plot(data['cmdAz'], data['cmdEl'],
            "g.", ms=0.5, label="Commanded")
    pl.plot(data['oteAz'][onscan], data['oteEl'][onscan],
            "r.", ms=1.0, label="On-src (Scanning)")
    pl.plot(data['oteAz'][offscan], data['oteEl'][offscan],
            "b.", ms=1.0, label="Off-src (Sidereal)")
    pl.xlabel("Az [deg]")
    pl.ylabel("El [deg]")
    # keywords labelsep and pad will be deprecated in matplotlib 0.98:
    # shall be replaced with labelspacing and borderpad, respectively
    pl.legend(labelsep = 0, \
              pad = 0.1, \
              numpoints = 1, \
              prop = pl.matplotlib.font_manager.FontProperties(size = 10))

    pl.show()

###--------------------------------
### end definitions
###--------------------------------

PARSER = OptionParser()
PARSER.set_defaults(fileName     = "position.dat",
                    rmsThreshold = "2.0",
                    maskTime     = "0.8",
                    yRange       = "0",
                    doPlots      = "y")
PARSER.add_option("-f", "--file", dest="fileName", metavar="FileName",
                  help="The name of the PSC data file")
PARSER.add_option("-r", "--rms", dest="rmsThreshold", metavar="RmsThreshold",
                  help="The Pass/Fail criterion of tracking accuracy [\"]")
PARSER.add_option("-m", "--mask", dest="maskTime", metavar="MaskTime",
                  help="The duration of mask at the beginning of scans [s]")
PARSER.add_option("-y", "--yrange", dest="yRange", metavar="YRange",
                  help="Limit of y-axis of error plot [\"]")
PARSER.add_option("-t", "--trange", dest="tRange", metavar="TRange",
                  help="Range of timestamps to be analyzed "+
                  "[s; two numbers separated by a colon]")
PARSER.add_option("-d", "--doplots", dest="doPlots", metavar="DoPlots",
                  help="Display the plots [y] or [n]")
(OPTIONS, ARGS) = PARSER.parse_args()

if not os.path.exists(OPTIONS.fileName):
    print "File \"%s\" does not exist." % OPTIONS.fileName
    sys.exit()

process_psc(OPTIONS.fileName, 
            float(OPTIONS.rmsThreshold),
            float(OPTIONS.maskTime),
            float(OPTIONS.yRange),
            OPTIONS.tRange,
            OPTIONS.doPlots)
