import os
import re
import pickle
import glob
import matplotlib.ticker 
import time
import shutil
import pylab as pl
import datetime
import casadef
import numpy as np
from analysisUtils import primaryBeamArcsec, numberOfChannelsInCube, getFitsBeam 
from analysisUtils import imviewField, roundFiguresToString, transition
from analysisUtils import angularScaleBaseline, imagePrimaryBeam, restFrequency
from analysisUtils import surmiseCycleFromImage, invertChannelRanges, flattenMask
from analysisUtils import imheadlist, makeList
from analysisUtils import version as auversion
from taskinit import *
from imhead_cli import imhead_cli as imhead
from imstat_cli import imstat_cli as imstat
from imsubimage_cli import imsubimage_cli as imsubimage
from immoments_cli import immoments_cli as immoments
from makemask_cli import makemask_cli as makemask

# global variable
startTime = 0

ANALYZECONTSUB_REVISION_STRING = "$Id: analyzecontsub.py,v 1.27 2016/03/07 17:34:10 thunter Exp $" 
rsyncPermissions = '--no-p'

def findSpectralAxis(img):
    """
    CASA images typically have the fourth axis as spectral, but exportfits 
    changes this to the third axis.  This function finds where it is.
    Returns: a number from 0 to 3 (most typically 2 or 3)
    """
    myia = createCasaTool(iatool)
    myia.open(img)
    mycs = myia.coordsys()
    try:
        iax = mycs.findaxisbyname("spectral")
    except:
        print "ERROR: can't find spectral axis.  Assuming it is 3."
        iax = 3
    mycs.done()
    myia.close()
    return iax

def createCasaTool(mytool):
    """
    A wrapper to handle the changing ways in which casa tools are invoked.
    Todd Hunter
    """
    if (type(casac.Quantity) != type):  # casa 4.x
        myt = mytool()
    else:  # casa 3.x
        myt = mytool.create()
    return(myt)

def version():
    """
    Returns the complete CVS version string for the current file.
    """
    return(ANALYZECONTSUB_REVISION_STRING)
 
def getCvsVersion(versionString):
    """
    Extracts the meaningful fields from a CVS version string.
    Example: 'analyzecontsub.py,v 1.24 2013/01/18 19:26:56'
    """
    tokens = versionString.split()        
    myVersion = tokens[1] + ' ' + tokens[2] + ' ' + tokens[3] + ' ' + tokens[4]
    return(myVersion)

def myCopyfile(a, b):
    """
    Checks whether a specified file exists in a specified location and
    if not, it copies it there, preserving the timestamps.
    a: source file
    b: destination
    """
    if (os.path.dirname(a) == b):
        return
    copyIndent = ' '*10         
#    print "%srsync from %s to %s" % (copyIndent, a, b)
    #shutil.copyfile(a,b)
    os.system('rsync -au %s %s %s' % (rsyncPermissions,a,b))

def addplot(webpagedir, myplot, width=800, br=True, debug=False):
    """
    Copies a single plot file into the specified web area, and returns 
    the text that places plot on the current html page.
    """
    myplot0 = os.path.basename(myplot)
    if (br):
        brk = '<br>'
    else:
        brk = ''
    mystring = """<a href="%s"><img width="%d" src="%s"></img></a>%s\n\n""" % (myplot0, width, myplot0, brk)
    destination = webpagedir+myplot0
    myCopyfile(myplot, destination)
    return(mystring)

def addplotrow(webpagedir, title, plotForColumn1='', plotForColumn2='',
               width=380, pdf1='', pdf2='', textForColumn1='', textForColumn2='', 
               justifyTitle='left',colspanTitle=1,bold=False, debug=False,
               justifyColumn1='left',valignColumn1='middle'):
    """
    Copies up to 2 plot files into the web area, and adds a row to the current html table
    If pdf1 is not blank, then the link is made to the PDF file rather than the full-sized
    version of the png.
    """
    copyIndent = ' '*10         
    plot1="c1_"+plotForColumn1.split('/')[-1]
    plot2="c2_"+plotForColumn2.split('/')[-1]
    if (1==0):    # debugging use
        if (textForColumn1 == ''):
            textForColumn1 = plotForColumn1
        if (textForColumn2 == ''):
            textForColumn2 = plotForColumn2
    if bold:
        textForColumn1 = '<b>' + textForColumn1
        textForColumn2 = '<b>' + textForColumn2
    bgcolor = '#E8E8E8'
    if (title.find('<b>') < 0):
        title = '<b>' + title
    mystring = """<tr><td align="%s" bgcolor="%s" colspan="%d">%s</td>""" % (justifyTitle,bgcolor,colspanTitle,title)
    if (colspanTitle == 1):
        bgcolor = 'white'
    if (colspanTitle < 3):
        if (pdf1 == ''):
            if (plot1 == 'c1_'):
                mystring += """<td bgcolor=%s align=%s valign=%s>%s</td>\n\n""" % (bgcolor,justifyColumn1,valignColumn1,textForColumn1)
            else:
                mystring += """<td align=%s valign=%s>%s<br><a href="%s"><img width="%d" src="%s"></img></a></td>\n""" % (justifyColumn1,valignColumn1,textForColumn1, plot1, width, plot1)
        else:
            mystring = """<td align=%s valign=%s>%s<br><a href="%s"><img width="%d" src="%s"></img></a></td>\n""" % (justifyColumn1, valignColumn1, textForColumn1, os.path.basename(pdf1), width, plot1)
        if (pdf2 == ''):
            if (plot2 == 'c2_'):
                mystring += """<td bgcolor=%s>%s</td></tr>\n\n""" % (bgcolor,textForColumn2)
            else:
                mystring += """<td>%s<br><a href="%s"><img width="%d" src="%s"></img></a></td></tr>\n\n""" % (textForColumn2, plot2, width, plot2)
        else:
            mystring += """<td>%s<br><a href="%s"><img width="%d" src="%s"></img></a></td></tr>\n\n""" % (textForColumn2, os.path.basename(pdf2), width, plot2)
    else:
        mystring += """</tr>\n\n"""
    # Make sure that the plots and PDFs get copied to the webpage, automatically.
    if os.path.exists(plotForColumn1):
        destination = webpagedir+plot1
        myCopyfile(plotForColumn1, destination)
    elif (plotForColumn1 != ''):
        print "ERROR: plotForColumn1 = %s does not exist!" % (plotForColumn1)

    if os.path.exists(plotForColumn2):
        destination = webpagedir+plot2
        myCopyfile(plotForColumn2, destination)
    elif (plotForColumn2 != ''):
        print "ERROR: plotForColumn2 = %s does not exist!" % (plotForColumn2)
        
    if os.path.exists(pdf1):
        destination = webpagedir+os.path.basename(pdf1)
        myCopyfile(pdf1, destination)
    elif (pdf1 != ''):
        print "ERROR: pdf1 = %s does not exist!" % (pdf1)
    if os.path.exists(pdf2):
        destination = webpagedir+os.path.basename(pdf2)
        myCopyfile(pdf2, destination)
    elif (pdf2 != ''):
        print "ERROR: pdf2 = %s does not exist!" % (pdf2)
    return(mystring)

def removeIfPresent(img):
    """
    Removes the specified directory, if it is present.
    """
    if os.path.exists(img):
        try:
            shutil.rmtree(img) 
        except:
            print "Could not remove file, perhaps because it is in use."
            cmd = "/usr/sbin/lsof +D %s" % (img)
            print "Running " + cmd
            os.system(cmd)
            print "You should quit or kill the offending process (if any)."
            sys.exit(0)

def showElapsedTime(mystring=''):
    # the trailing comma prevents the newline so that ...10...20 will be on same line
    line = "%s at UT %s %s" % (computeElapsedTime(), time.strftime('%H:%M:%S',time.gmtime()), mystring) 
    if (mystring.lower().find('done') == 0):
        print line
    else:
        print line,
    return line

def buildMomentImage(img, moment, resultsDir, includepix=-1, suffix='', chans=''):
    """
    Takes a CASA cube, and creates the specified moment image, writing it to
    the desired directory.
    moment: integer moment (e.g. 0..8=peak)
    """
    img_mom = os.path.basename(img) + suffix + '.mom%d' % (moment)
    if (resultsDir != ''):
        img_mom = resultsDir + '/' + img_mom
    removeIfPresent(img_mom) # prevent table cache issue
    if (includepix > -1):
        includepix = [includepix,1e9]
    print "Running immoments('%s',moments=[%d],outfile='%s',includepix=%s,chans='%s')" % (img,moment,img_mom,str(includePix),chans)
    showElapsedTime("immoments")
    immoments(img, moments=[moment], outfile=img_mom, includepix=includepix, chans=chans)
    showElapsedTime("imhead")
    img_mom_head = au.imheadlist(img_mom) 
    return(img_mom, img_mom_head)

def buildImagePng(img, radius, resultsDir, minIntensity=None, maxIntensity=None,
                  maxIntensityMinFraction=None, contourImage=None, levels=[], 
                  contourImage2=None, levels2=[], verbose=False, 
                  colorwedge=True, contourColor='white', trim=True):
    """
    Takes a CASA image and creates a color plot with the specified image size
    and intensity range.  See help(au.imviewField) for the meaning of the other
    parameters.
    """
    levels = makeList(levels)
    if (img == ''): return img
    plotfile = os.path.basename(img) + '.png'
    if (resultsDir != ''):
        plotfile = resultsDir + '/' + plotfile
    if verbose:
        print "buildImagePng: minIntensity = ", minIntensity
        print "buildImagePng: maxIntensity = ", maxIntensity
    img_png = imviewField(img, radius, minIntensity, maxIntensity=maxIntensity, 
                          maxIntensityMinFraction=maxIntensityMinFraction,
                          colorwedge=colorwedge, plotfile=plotfile, 
                          contourImage=contourImage, levels=levels,
                          contourImage2=contourImage2, levels2=levels2, 
                          verbose=verbose, contourColor=contourColor, trim=trim)
    os.system('chmod g+w ' + img_png)
    return img_png

def initializeWebpage(directoriesAndVersions, user, host, date, casaVersion, 
                      py0, py1, py2, py3, py4, cwd, commandLine, webpagedir):
    """
    Returns a string containing the header portion of a new .html file.
    """
    webpage = """
    <html>
      <head>
         <title>analyzecontsub</title>
         <style type="text/css">
            html, td {font-size: 90%%;}
            .headline {font-size:200%%; font-weight: bold;}              
            table {border-collapse:collapse;}
            table, th, td {border: 1px solid black;}           
            td {padding:0.25em}
            td {vertical-align:top;}
            td[colspan] {background-color: #F3F3F3;}
            .innertable td, .innertable tr, .innertable {border: none; padding: 0; padding-right: 1em;}            
            b.passquality {color: #227722;}
            b.nopassquality {color: #772222;}
         </style>
      </head>
      <body>
      <div class="headline"><a href="%s">analyzecontsub</a></div><br>
      <H3>Location of data and version information</H3>
      %s
      This report was generated by %s on %s (%.1fGB) at <b>%s</b> from CASA %s (python %d.%d.%d %s %d).<br>
      PWD = %s<br>
      Command line:<br>%s<br><br>
    """ % (fileToWebAddress(os.path.dirname(webpagedir.rstrip('/'))),directoriesAndVersions, user, host, getMemorySize()/1e9, date, casaVersion, py0,py1,py2,py3,py4,
           cwd, commandLine)
    return webpage

def beamString(head, roundMajorAxis=3, roundPA=4):
    """
    Reads a CASA image header dictionary and produces a string of the format 
    "X.XX x Y.YY at ZZ.ZZ deg".  If no beam is found in the header, then return 'no beam'.
    roundMajorAxis: number of significant figures to show for bmaj and bmin
    roundPA: number of significant figures to show for position angle
    """
    if ('beammajor') in head:
        bmaj = head['beammajor']['value']
        bmin = head['beamminor']['value']
        bpa = head['beampa']['value']
    elif ('perplanebeams' in head):
        nchan = head['perplanebeams']['nChannels']
        bmaj = head['perplanebeams']["*"+str(int(np.floor(nchan/2.0)))]['major']['value']
        bmin = head['perplanebeams']["*"+str(int(np.floor(nchan/2.0)))]['minor']['value']
        bpa = head['perplanebeams']["*"+str(int(np.floor(nchan/2.0)))]['positionangle']['value']
    else:
        print "Did not find 'beammajor' or 'perplanebeams' in header.  Is this a dirty image?"
        return 'no beam'
    string = '%s" x %s" at %s deg' % (roundFiguresToString(bmaj,roundMajorAxis), 
                                      roundFiguresToString(bmin,roundMajorAxis), 
                                      roundFiguresToString(bpa,roundPA))
    return string

def string_stats2(stats, stats2, keyword, digits):
    """
    Reads a keyword out of two dictionaries and passes them to build_stat_string2
    to create a string.
    """
    if (stats == ''):
        return('')
    return(build_stat_string2(stats[keyword], stats2[keyword], digits))

def string_stats(stats, keyword, digits):
    """
    Reads a keyword out of a dictionary and pass it to build_stat_string to create a string
    """
    if (stats == ''):
        return('')
    return(build_stat_string(stats[keyword], digits))

def build_stat_string(jy, digits):
    return(roundFiguresToString(jy,digits) + ' Jy/b * km/s')

def build_stat_string2(jy, jy2, digits):
    if (len(jy2) > 0):
        return(roundFiguresToString(jy,digits) + 5*'&emsp;' + roundFiguresToString(jy2,digits) + ' Jy/b * km/s')
    else:
        return(roundFiguresToString(jy,digits) + ' Jy/b * km/s')

def aggregateBandwidthString(bw, chanWidth):
    return('aggregate bandwidth = %g MHz (%d chan)'%(bw*1000, abs(bw*1e9/chanWidth)))

def getHeader(img, psf='', omitBeam=False):
    """
    Reads the header of an image.  If the beam is not present, then look for it in
    the specified psf (or try to find a psf) and copy it into the original header.
    """
    if (not os.path.exists(img)):
        print "Could not find image: ", img
        return
    showElapsedTime("imhead")
    head = imheadlist(img, omitBeam)
    # Add some information to the header dictionary
    head['imagefile'] = os.path.basename(img)
    head['npix'] = head['shape'][0] * head['shape'][1]
    if beamString(head) == 'no beam':
        if (psf == ''):
            imageStrings = [m.start() for m in re.finditer(r".image", img)]
            if (len(imageStrings) < 1):
                imageStrings = [m.start() for m in re.finditer(r".residual", img)]
                if (len(imageStrings) < 1):
                    imageStrings = [m.start() for m in re.finditer(r".pbcor", img)]
                    if (len(imageStrings) < 1):
                        imageStrings = [m.start() for m in re.finditer(r".pbc", img)]
                        if (len(imageStrings) < 1):
                            imageStrings = [m.start() for m in re.finditer(r".mask", img)]
            if (len(imageStrings) > 0):
                finalImage = imageStrings[-1]
                psf = img[:finalImage] + '.psf'
                print "Looking for psf = ", psf
        if os.path.exists(psf):
            print "psf imhead", # the comma prevents the newline so that ...10...20 will be on same line
            headpsf = imheadlist(psf) 
            if ('beammajor' in headpsf):
                head['beammajor'] = headpsf['beammajor']
                head['beamminor'] = headpsf['beamminor']
                head['bpa'] = headpsf['beampa']
            elif ('perplanebeams' in headpsf):
                head['perplanebeams'] = headpsf['perplanebeams']
            else:
                print "But still no beam found in psf."
        else:
            print "Did not find psf image"
    return head

def imageSizeString(header):
    cdelt1arcsec = np.degrees(abs(header['cdelt1']))*3600 
    cdelt2arcsec = np.degrees(abs(header['cdelt2']))*3600
    string = '%dx%d=%d pixels of %g" (%.2f"x%.2f")' % (header['shape'][0], header['shape'][1],
                                                   header['shape'][0]*header['shape'][1],
                                                   cdelt1arcsec,
                                                   header['shape'][0]*cdelt1arcsec,
                                                   header['shape'][1]*cdelt2arcsec)
    return string

def spectralSizeString(header, chanWidth, specaxis):
    nchan = header['shape'][specaxis]
    string = '%d channels * %s kHz = %s MHz' % (nchan,roundFiguresToString(abs(chanWidth*1e-3),4),
                                                roundFiguresToString(nchan*abs(chanWidth*1e-6),4))
    return string

def fileToWebAddress(webfile):
    if (webfile == ''):
        return webfile
    return('https://safe.nrao.edu/'+webfile.replace('/home/safe.nrao.edu/content',''))

def getMemorySize():
    """
    Returns memory size in bytes
    """
    return(os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES'))

def analyzecontsubs(images=[], imsubImages=[], uvcontsubImages=[], 
                    webfile='analyzecontsub/index.html',
                    imsubDir='', uvcontsubDir='', pblimit=0.5, 
                    webarea=True, imagedir='', verbose=False,
                    minmaxScale='imsub', imageRadius=0):
    """
    Accepts a list of imsub images and a list of corresponding uvcontsub
    images and generates a webpage showing statistics of the images, moment 0 and 8 images and
    spectra.  
    Inputs:
    Specify either 'images' or ('imsubImages' and 'uvcontsubImages')
    images: a list of strings, to which _imagesub.tclean.image or _uvcontsub.tclean.image will be appended
    imsubImages: a list of strings for imagesub images
    uvcontsubImages: a list of strings for the uvcontsub images
    webfile: full or relative path of html file to produce, e.g. '2013.1.00099.S/index.html'
    webarea: True: treat webfile as a relative directory, and precede it with
                   /home/safe.nrao.edu/content/alma/PipelineTestResults/
    imagedir: directory to write new images (if any), default: current working directory
    pblimit: compute statistics over the central primary beam with radius equal to this 
             Gaussian beam sensitivity level (0.0..1.0)
    minmaxScale: 'imsub' or 'uvcontsub'; sets the min/max colorbar of residual image
    imageRadius: radius of image (in arcsec) to display as png (<=0 means whole image)

    Example execution:
    analyzecontsub.analyzecontsubs(images=['calibrated_NGC_1068.ms_spw0',calibrated_NGC_1068.ms_spw1',
                                           'calibrated_NGC_1068.ms_spw1',calibrated_NGC_1068.ms_spw3']
                                   webfile='contsub.html',
                                   imsubDir='/lustre/naasc/users/cbrogan/Pipeline_Imaging_Tests/2013.1.00014.S/',
                                   uvcontsubDir='/lustre/naasc/users/cbrogan/Pipeline_Imaging_Tests/2013.1.00014.S/'
    """
    global startTime
    startTime = time.time()
    spacing = 54
    commandLine = "%s.analyzecontsubs(images=%s,<br>" % (analyzecontsub.__name__, images) + \
                   + spacing*'&nbsp;' + "imsubImages=%s,<br>" % (imsubImages) + \
                   + spacing*'&nbsp;' + "uvcontsubImages=%s,<br>" % (uvcontsubImages)  + \
                   + spacing*'&nbsp;' + "imsubDir='%s',<br>" % (imsubDir)  + \
                   + spacing*'&nbsp;' + "uvcontsubDir='%s',<br>" % (uvcontsubDir)  + \
                   + spacing*'&nbsp;' + "webfile='%s',<br>" % (webfile) + \
                   + spacing*'&nbsp;' + "webarea=%s,<br>" % (webarea) + \
                   + spacing*'&nbsp;' + "imagedir='%s',<br>" % (imagedir) + \
                   + spacing*'&nbsp;' + "pblimit=%g,<br>" % (pblimit) + \
                   + spacing*'&nbsp;' + "imageRadius=%g,<br>" % (imageRadius) + \
                   + spacing*'&nbsp;' + "minmaxScale='%s')<br>" % (minmaxScale)

    if (len(images) > 0):
        imsubImages = [i+'_imagesub.tclean.image' for i in images]
        uvcontsubImages = [i+'_uvcontsub.tclean.image' for i in images]
    if (type(imsubImages) == str):
        imsubImages = imsubImages.split(',')
    if (type(uvcontsubImages) == str):
        uvcontsubImages = uvcontsubImages.split(',')
    if (len(imsubImages) != len(uvcontsubImages)):
        print "Mismatch in number of imsub (%d) vs. uvcontsub (%d) images." % (len(imsubImages), len(uvcontsubImages))
        return
    if (webfile == ''):
        webfile = 'index.html'
    continuumResults = ''
    if webarea:
        if (webfile[0] != '/'):
            webfile = '/home/safe.nrao.edu/content/alma/PipelineTestResults/' + webfile
        else:
            print "webfile begins with a /, so I will assume this is a full path."
        continuumResults = lookForFile(os.path.dirname(webfile.replace('analyzecontsub','analyzemsimage'))+'/continuum.html')
        continuumResults = fileToWebAddress(continuumResults)

    webpagedir = os.path.dirname(webfile) + '/'
    if (webpagedir == '/'): 
        webpagedir = ''
    elif (not os.path.exists(webpagedir)):
        print "----- Making web subdirectories:  %s -----" % (webpagedir)
        os.makedirs(webpagedir)
        os.system('chmod g+w ' + webpagedir)

    directoriesAndVersions = 'ac version = %s<br>\n' % (version())
    directoriesAndVersions += 'au version = %s<br>\n' % (auversion())
    py0,py1,py2,py3,py4 = sys.version_info

    # Prepend image list with imsubDir
    if (imsubDir != ''):
        imsubImages = [imsubDir.rstrip('/') + '/' + p for p in imsubImages]
    if (uvcontsubDir != ''):
        uvcontsubImages = [uvcontsubDir.rstrip('/') + '/' + p for p in uvcontsubImages]

    for i in range(len(imsubImages)):
        if (not os.path.exists(imsubImages[i])):
            print "Could not find imsubImage = ", imsubImages[i]
            return
        if (not os.path.exists(uvcontsubImages[i])):
            print "Could not find uvcontsubImage = ", uvcontsubImages[i]
            return

    webpage = initializeWebpage(directoriesAndVersions, os.getenv('USER'), os.getenv('HOST'), 
                                str(datetime.datetime.now()), 
                                casadef.casa_version+' r'+casadef.subversion_revision, 
                                py0,py1,py2,py3,py4,os.getcwd(), commandLine, webpagedir)
    totalTime = 0
    # copy the results file
    textfile = lookForFile(imsubDir+'/calibrated_*_results.txt')
    acsTxt = webpagedir + os.path.basename(textfile.replace('.txt','.analyzecontsub.txt'))
    if (textfile != ''):
        myCopyfile(textfile, webpagedir+os.path.basename(textfile))
        os.system('chmod g+w ' + webpagedir+os.path.basename(textfile))
        webpage += '<a href="%s">Link to benchmark script results file</a><BR><BR>' % (os.path.basename(textfile))
        webpage += '<a href="%s">Link to extra results file</a><BR><BR>' % (os.path.basename(acsTxt))
        if (len(continuumResults) > 0):
            webpage += '<a href="%s">Link to continuum results page</a><BR><BR>' % (continuumResults)
    f = open(webfile,'w')
    f.write(webpage)
    f.close()
    os.system('chmod g+w ' + webfile)
    txt = open(acsTxt,'w')
    txt.close()
    os.system('chmod g+w ' + webpagedir+os.path.basename(acsTxt))
    for i in range(len(imsubImages)):
        # keep webarea=False to avoid double insertion of that directory
        result = analyzecontsub(imsubImages[i], uvcontsubImages[i], acsTxt, pblimit,
                                appendhtml=True, webfile=webfile, webarea=False,
                                imagedir=imagedir, verbose=verbose, 
                                minmaxScale=minmaxScale, imageRadius=imageRadius)
        if result==None: return
        totalTime += result
    finishWebpage(webfile)
    print "Total elapsed time = %.1f sec = %.1f min" % (totalTime, totalTime/60.)

def escapeSlash(s):
    """
    Protect all appearances of forward slash with a backslash.
    """
#    return(s.replace('/','\/'))
    return('"'+s+'"')

def runImstat(Image, PB, Mask=False, channels='', region='', level=0.5, 
              innerAnnulusLevel=None, location=-1, verbose=True, pbMin=None,
              MaskFlattened=False, insideMask=False):
    """
    Runs imstat on an image, applying a mask based on the primary beam response
    and/or another mask image, over all channels or specific channel selection.
    PB: False, or the name of a primary beam image
    Mask: False, or the name of a mask image;  if specified, then use only pixels 
          outside of this mask (i.e. with value=0 meaning use the uncleaned area)
    insideMask: if True, then use values inside the mask (1); False => outside (0)
    Note: PB and Mask images should be of the same dimensionality (incl. number of channels)
    channels: passed to the chans parameter of imstat
    region: passed to the region parameter of imstat
    level: if PB is not False, then use pixels above this response level
    innerAnnulusLevel: if PB is not False, then use this level to define the
                       inner radius of an annulus from which to select pixels
    location: integer used for debugging where in the method this call originated
    pbMin: minimum value in the PB image
    """
    print "Running imstat %d on " % (location), os.path.basename(Image)
    # level only matters if PB exists
    if PB:
        if pbMin == None:
            # Sometimes the manual image is small and does not extend very far into the PB
            showElapsedTime("imstat on PB")
            pbMin = imstat(PB)['min']
    if (PB and Mask):
        if (numberOfChannelsInCube(PB) == 1):
            if (numberOfChannelsInCube(Mask) > 1):
                Mask = MaskFlattened
        if insideMask:
            mask = '%s>%g && %s>0.5' % (escapeSlash(PB), level, escapeSlash(Mask))
        else:
            mask = '%s>%g && %s<0.5' % (escapeSlash(PB), level, escapeSlash(Mask))
        if verbose: print "Using PB mask and clean mask: imstat('%s', mask='%s', stretch=True)" % (Image,mask)
        showElapsedTime("imstat")
        Stats = imstat(Image, mask=mask, stretch=True) 
    elif PB:
        if (innerAnnulusLevel == None):
            mask = '%s>%g' % (escapeSlash(PB),level)
            if verbose: 
                print "Using PB mask: imstat('%s', mask='%s', chans='%s')" % (Image, mask, channels)
            showElapsedTime("imstat")
            Stats = imstat(Image, mask=mask, chans=channels)
        else:
            if (pbMin > innerAnnulusLevel):
                print "Image is not large enough to reach the specified PB level."
                Stats = ''
            else:
                mask = '%s>%g && %s<%g'%(escapeSlash(PB),level,escapeSlash(PB),innerAnnulusLevel)
                if verbose: print "Using PB annular mask: imstat('%s', mask='%s', chans='%s', stretch=True)" % (Image, mask, channels)
                showElapsedTime("imstat")
                Stats = imstat(Image, mask=mask, chans=channels, stretch=True)
    elif Mask:
        if insideMask:
            mask = '%s>0.5'%(escapeSlash(Mask))
        else:
            mask = '%s<0.5'%(escapeSlash(Mask))
        if verbose: print "Using clean mask: imstat('%s', mask='%s', chans='%s', region='%s')" % (Image, mask, channels, region)
        showElapsedTime("imstat")
        Stats = imstat(Image, mask=mask, region=region, chans=channels)
    else:
        if verbose: print "Using region,chans: imstat('%s', region='%s', chans='%s')" % (Image, region, channels)
        showElapsedTime("imstat")
        Stats = imstat(Image, region=region, chans=channels) #  regionName and chans and/or stokes cannot be specified simultaneously              
    showElapsedTime("done imstat")
    return Stats

def imagePixelsInMask(maskImage, validValue=1):
    """
    Counts the number of pixels with a non-zero value in a mask image.
    """
    pixels = 0
    if (maskImage != False):
        if (os.path.exists(maskImage)):
            myia = createCasaTool(iatool)
            myia.open(maskImage)
            data = myia.getregion()
            pixels = len(np.where(data == validValue)[0])
            myia.close()
    return pixels

def readContDat(contDat):
    """
    Reads a summary text file created by findContinuum.py
    """
    f = open(contDat,'r')
    selection, png, bw = f.readlines()[0].strip('\n').split()
    print "Read %s: " % (contDat), selection, png, float(bw)
    return(selection, png, float(bw))

def rreplace(s, old, new='', occurrence=1):
    """
    Replaces the final n appearances of <old> in <s> with <new>.
    Giving simply the first 2 arguments is equivalent to rstrip 
    but only removing the final appearance of <old>.
    e.g. rreplace('a.image.image','.image','.model') -> a.image.model
    """
    li = s.rsplit(old, occurrence)
    return new.join(li)

def computeElapsedTime():
    """
    Takes the difference between the current time and the global variable start time
    and returns a string in minutes and seconds elapsed.
    """
    seconds = time.time()-startTime
    return('%02dm%02ds ' % (int(seconds/60),int(seconds-60*int(seconds/60))))

def lookForFile(png):
    """
    Looks for a file matching a wildcard and returns the newest one
    """
    a = glob.glob(png)
    if (len(a) == 0):
        return ''
    else:
        a = max(glob.iglob(png), key=os.path.getctime)
        return a

def analyzecontsub(imsubImage, uvcontsubImage, acsTxt, pblimit=0.5, cleanMask='',
                   appendhtml=False, webfile='', webarea=False, 
                   imagedir='',verbose=False, minmaxScale='imsub', 
                   balanceIntensityLimitsLineFreeMoment0=True,
                   imageRadius=0):
    """
    Compares two images (or cubes) and generates an html page with
    moment images and statistics. 
    Inputs:
    imsubImage: the image of an spw generated by imsub
    uvcontsubImage: the corresponding image of the same spw generated by uvcontsub
    webfile: full path to write the webfile (and accompanying images)
    webarea: True: treat webfile as a relative directory, and precede it with
                   /home/safe.nrao.edu/content/alma/PipelineTestResults/
             Note: this should be set False when called from analyzecontsubs
    imagedir: directory to write new images (if any), default: current working directory
    appendhtml: if True, then append rather than overwrite to the webfile
    pblimit: compute statistics over the central primary beam with radius equal to this 
             Gaussian beam sensitivity level (0.0..1.0)
    imageMask: if specified, avoid using pixels from the clean mask in statistics
    minmaxScale: 'imsub' or 'uvcontsub'; sets the min/max colorbar of residual image
    imageRadius: radius of image (in arcsec) to display as png (<=0 means whole image)
    Returns:
    * execution time in seconds
    """
    global startTime
    if (startTime == 0):
        startTime = time.time()
    singleStartTime = time.time()
        
    webpagedir = os.path.dirname(webfile) + '/'
    if webarea:
        webpagedir = '/home/safe.nrao.edu/content/alma/PipelineTestResults/' + webpagedir
    if (webpagedir == '/'): 
        webpagedir = ''
    elif (not os.path.exists(webpagedir)):
        print "----- Making web subdirectories:  %s -----" % (webpagedir)
        os.makedirs(webpagedir)

    if (not os.path.exists(imagedir) and imagedir != ''):
        print "----- Making image subdirectories:  %s -----" % (imagedir)
        os.makedirs(imagedir)
    if (imagedir != ''):
        if (imagedir[-1] != '/'): imagedir += '/'

    # Check that both images are of the same dimensionality
    if verbose: print "Reading nchan from ", imsubImage
    # Getting the header information from the original images
    imsubHead = getHeader(imsubImage)
    if verbose: print "Running numberOfChannelsInCube('%s', returnChannelWidth=True)" % (imsubImage)
    imsubChannels, imsubChanWidth = numberOfChannelsInCube(imsubImage, returnChannelWidth=True, header=imsubHead)
    uvcontsubHead = getHeader(uvcontsubImage)
    uvcontsubChannels, uvcontsubChanWidth = numberOfChannelsInCube(uvcontsubImage, returnChannelWidth=True, header=uvcontsubHead)

    if (not appendhtml):
        if webfile == '':
            htmlfile = "%s_compared_to_%s.html" % (os.path.basename(pipelineImage), os.path.basename(manualImage))
            webfile = webpagedir + htmlfile
        directoriesAndVersions = ''
        directoriesAndVersions += 'ami version = %s<br>\n' % (version())
        directoriesAndVersions += 'imsub image = %s<br>\n' % (imsubImage)
        directoriesAndVersions += 'uvcontsub image = %s<br>\n' % (uvcontsubImage)
        amiVersion = getCvsVersion(version())
        py0,py1,py2,py3,py4 = sys.version_info
        commandLine = "analyzecontsub.analyzecontsub(imsubImage='%s', uvcontsubImage='%s', webfile='%s')" % (imsubImage, uvcontsubImage, webfile)

        webpage = initializeWebpage(directoriesAndVersions, os.getenv('USER'), os.getenv('HOST'), 
                                    str(datetime.datetime.now()), 
                                    casadef.casa_version+' r'+casadef.subversion_revision, 
                                    py0,py1,py2,py3,py4,os.getcwd(), commandLine, webpagedir)
    else:
        webpage = ''

    imsubSpecAxis = findSpectralAxis(imsubImage)
    uvcontsubSpecAxis = findSpectralAxis(uvcontsubImage)
    bmaj,bmin,bpa,cdelt1,cdelt2,naxis1,naxis2,freqGHz = getFitsBeam(imsubImage)
    primaryBeam = primaryBeamArcsec(frequency=freqGHz, showEquation=False)
    if verbose: print "primary beam = %.1f arcsec for %.3f GHz" % (primaryBeam, freqGHz)

    dirtyImage = lookForFile(uvcontsubImage[:uvcontsubImage.find('_uvcontsub.')]+'_dirty.residual')
    dirtyHead = imheadlist(dirtyImage, omitBeam=True)
    dirtyChannels, dirtyChanWidth = numberOfChannelsInCube(dirtyImage, returnChannelWidth=True, header=dirtyHead)

    spwString = os.path.basename(imsubImage)[os.path.basename(imsubImage).find('spw'):].split('_')[0]
    spwInfo = "channels: %s<br>width: %.0f kHz" % (dirtyChannels, dirtyChanWidth*1e-3)
    spwID = int(spwString.lstrip('spw'))
    sourceName = os.path.basename(imsubImage).split('.ms_spw')[0].split('alibrated_')[1]
    vis = imsubImage.split('_spw%d'%spwID)[0]

    # Separate images into individual tables
    dirtyFindContinuumPng = lookForFile(uvcontsubImage[:uvcontsubImage.find('_uvcontsub.')]+'_dirty.residual.meanSpectrum*.png')
    webpage += '<table><tr><td><font size=12>%s</font></td><td><b>initial findContinuum result:</b> used to define<br> continuum channels and line-free channels for uvcontsub</td><td><b>Continuum clean image</b><br> with contours showing primary beam response: [0.5]</tr>' % (spwString)
    imsubPB = imsubImage.replace('.image','.pb')
    uvcontsubPB = uvcontsubImage.replace('.image','.pb')
    continuumImage = imsubImage.rstrip('_imagesub.tclean.image').replace('.ms','.cont') + '.image'
    continuumPng = buildImagePng(continuumImage, imageRadius, webpagedir, 
                                 contourImage=imsubPB, levels=0.5, 
                                 contourColor='yellow', verbose=True)
    if (continuumPng == ''):
        continuumText = ''
    else:
        continuumText = "<a href="+os.path.basename(continuumPng)+"><img width=500 src="+os.path.basename(continuumPng)+"></img></a>"

    webpage += '<tr><td valign=middle>%s</td> <td><a href="%s"><img width=500 src="%s"></img></a></td><td>%s</td></tr></table>' % (spwInfo, os.path.basename(dirtyFindContinuumPng),os.path.basename(dirtyFindContinuumPng), continuumText)
    myCopyfile(dirtyFindContinuumPng, webpagedir+os.path.basename(dirtyFindContinuumPng))

    webpage += '<table border=1>'
    webpage += '<tr> <th>Common Information</th></tr>\n'
    beamMajor = float(beamString(imsubHead).split('"')[0])
    beamMinor = float(beamString(imsubHead).split('"')[1].split()[-1])
    geometricMeanBeam = (beamMajor * beamMinor)**0.5
    if ('date-obs' in imsubHead):
        cycle = surmiseCycleFromImage(imsubImage, imsubHead)
        observationDate = 'observed on %s (ALMA Cycle %d)' % (imsubHead['date-obs'], cycle)
    webpage += addplotrow(webpagedir, 'Primary beam = %s" for frequency = %s GHz;</b> approximate max baseline = %sm; %s' % (roundFiguresToString(primaryBeam,3), roundFiguresToString(freqGHz,5), roundFiguresToString(angularScaleBaseline(freqGHz,beamMinor),2), observationDate),colspanTitle=3) 

    print "Running au.transition('%s',%d,'%s')" % (vis,spwID,sourceName)
    transitionString = transition(vis, spwID, sourceName)[0]
    webpage += addplotrow(webpagedir, title='Transition & Rest Frequency',
                          textForColumn1=transitionString,
                          textForColumn2=roundFiguresToString(1e-9*restFrequency(vis, spwID, sourceName)[0], 8)+' GHz')

    webpage += '<tr> <th>Item</th> <th>imsub results</th> <th>uvcontsub results</th> </tr>\n'
    flushWebpage(webfile,webpage)
    webpage = ''

    webpage += addplotrow(webpagedir, title='Filename',
                          textForColumn1=os.path.basename(imsubImage),
                          textForColumn2=os.path.basename(uvcontsubImage))

    webpage += addplotrow(webpagedir, title='Image size',
                          textForColumn1=imageSizeString(imsubHead),
                          textForColumn2=imageSizeString(uvcontsubHead))

    webpage += addplotrow(webpagedir, title='Spectral axis',
                          textForColumn1=spectralSizeString(imsubHead, imsubChanWidth, imsubSpecAxis),
                          textForColumn2=spectralSizeString(uvcontsubHead, uvcontsubChanWidth, uvcontsubSpecAxis))

    webpage += addplotrow(webpagedir, 'Restoring beam', 
                          textForColumn1=beamString(imsubHead),
                          textForColumn2=beamString(uvcontsubHead))

    imsubMom0 = imsubImage+'.MOM0'
    imsubMom8 = imsubImage+'.PEAKMOM'
    imsubMom0LineFree = imsubImage+'.MOM0.LINEFREE'
    imsubMom8LineFree = imsubImage+'.PEAKMOM.LINEFREE'
    uvcontsubMom0 = uvcontsubImage+'.MOM0'
    uvcontsubMom8 = uvcontsubImage+'.PEAKMOM'
    uvcontsubMom0LineFree = uvcontsubImage+'.MOM0.LINEFREE'
    uvcontsubMom8LineFree = uvcontsubImage+'.PEAKMOM.LINEFREE'


    if (numberOfChannelsInCube(imsubPB) > 1):
        newimsubPB = imagedir + rreplace(os.path.basename(imsubPB),'.pb','.1chan.pb')
        if (os.path.exists(newimsubPB)):
            shutil.rmtree(newimsubPB)
        imsubimage(imsubPB,chans='0',outfile=newimsubPB)
        imsubPB = newimsubPB
    if (numberOfChannelsInCube(uvcontsubPB) > 1):
        newuvcontsubPB = imagedir + rreplace(os.path.basename(uvcontsubPB),'.pb','.1chan.pb')
        if (os.path.exists(newuvcontsubPB)):
            shutil.rmtree(newuvcontsubPB)
        imsubimage(uvcontsubPB,chans='0',outfile=newuvcontsubPB)
        uvcontsubPB = newuvcontsubPB

    # Moment 0 statistics
    imsubFlattenedMask = lookForFile(imsubImage.replace('.image','')+'*.flattenmask')
    uvcontsubFlattenedMask = lookForFile(uvcontsubImage.replace('.image','')+'*.flattenmask')
    pixelsInFlattenedMask = imagePixelsInMask(uvcontsubFlattenedMask)
    # The benchmark script only constructs the uvcontsub*flattenmask
    if imsubFlattenedMask == '':
        imsubFlattenedMask = uvcontsubFlattenedMask

    # region pb>0.5
    imsubMom0LFStats = runImstat(imsubMom0LineFree, imsubPB, location=0, level=pblimit)
    uvcontsubMom0LFStats = runImstat(uvcontsubMom0LineFree, uvcontsubPB, location=1, level=pblimit)
    imsubMom0Stats = runImstat(imsubMom0, imsubPB, location=2, level=pblimit)
    uvcontsubMom0Stats = runImstat(uvcontsubMom0, uvcontsubPB, location=3, level=pblimit)

    # region "between" pb>0.5 and the mask
    print "runImstat('%s','%s','%s',location=10,level=%f)" % (imsubMom0LineFree, imsubPB, imsubFlattenedMask, 
                                                              pblimit)
    imsubMom0LFStatsBetween = runImstat(imsubMom0LineFree, imsubPB, imsubFlattenedMask, 
                                        location=10, level=pblimit)
    uvcontsubMom0LFStatsBetween = runImstat(uvcontsubMom0LineFree, uvcontsubPB, uvcontsubFlattenedMask, 
                                            location=11, level=pblimit)
    imsubMom0StatsBetween = runImstat(imsubMom0, imsubPB, imsubFlattenedMask, 
                                      location=12, level=pblimit)
    uvcontsubMom0StatsBetween = runImstat(uvcontsubMom0, uvcontsubPB, uvcontsubFlattenedMask, 
                                          location=13, level=pblimit)

    imsubMom8LFStatsBetween = runImstat(imsubMom8LineFree, imsubPB, imsubFlattenedMask, 
                                        location=10, level=pblimit)
    uvcontsubMom8LFStatsBetween = runImstat(uvcontsubMom8LineFree, uvcontsubPB, uvcontsubFlattenedMask, 
                                            location=11, level=pblimit)
    imsubMom8StatsBetween = runImstat(imsubMom8, imsubPB, imsubFlattenedMask, 
                                      location=12, level=pblimit)
    uvcontsubMom8StatsBetween = runImstat(uvcontsubMom8, uvcontsubPB, uvcontsubFlattenedMask, 
                                          location=13, level=pblimit)

    # region within the mask
    imsubMom0LFStatsInMask = runImstat(imsubMom0LineFree, False, imsubFlattenedMask, 
                                       location=20, level=pblimit, insideMask=True)
    uvcontsubMom0LFStatsInMask = runImstat(uvcontsubMom0LineFree, False, uvcontsubFlattenedMask, 
                                           location=21, level=pblimit, insideMask=True)
    imsubMom0StatsInMask = runImstat(imsubMom0, False, imsubFlattenedMask, 
                                     location=22, level=pblimit, insideMask=True)
    uvcontsubMom0StatsInMask = runImstat(uvcontsubMom0, False, uvcontsubFlattenedMask, 
                                         location=23, level=pblimit, insideMask=True)

    # region outside the mask
    imsubMom0LFStatsOutMask = runImstat(imsubMom0LineFree, False, imsubFlattenedMask, 
                                       location=20, level=pblimit, insideMask=False)
    uvcontsubMom0LFStatsOutMask = runImstat(uvcontsubMom0LineFree, False, uvcontsubFlattenedMask, 
                                           location=21, level=pblimit, insideMask=False)
    imsubMom0StatsOutMask = runImstat(imsubMom0, False, imsubFlattenedMask, 
                                     location=22, level=pblimit, insideMask=False)
    uvcontsubMom0StatsOutMask = runImstat(uvcontsubMom0, False, uvcontsubFlattenedMask, 
                                         location=23, level=pblimit, insideMask=False)


    # Moment 8 statistics
    # region "between" pb>0.5 and the mask
    imsubMom8LFStats = runImstat(imsubMom8LineFree, imsubPB, imsubFlattenedMask, 
                                 location=30, level=pblimit)
    uvcontsubMom8LFStats = runImstat(uvcontsubMom8LineFree, uvcontsubPB, uvcontsubFlattenedMask, 
                                     location=31, level=pblimit)
    imsubMom8Stats = runImstat(imsubMom8, imsubPB, imsubFlattenedMask, 
                               location=32, level=pblimit)
    uvcontsubMom8Stats = runImstat(uvcontsubMom8, uvcontsubPB, uvcontsubFlattenedMask, 
                                   location=33, level=pblimit)

    # region pb>0.5
    imsubMom8LFStats = runImstat(imsubMom8LineFree, imsubPB, location=40, level=pblimit)
    uvcontsubMom8LFStats = runImstat(uvcontsubMom8LineFree, uvcontsubPB, location=41, level=pblimit)
    imsubMom8Stats = runImstat(imsubMom8, imsubPB, location=42, level=pblimit)
    uvcontsubMom8Stats = runImstat(uvcontsubMom8, uvcontsubPB, location=43, level=pblimit)

    # region within the mask
    imsubMom8LFStatsInMask = runImstat(imsubMom8LineFree, False, imsubFlattenedMask,
                                       location=50, level=pblimit, insideMask=True)
    uvcontsubMom8LFStatsInMask = runImstat(uvcontsubMom8LineFree, False, uvcontsubFlattenedMask, 
                                           location=51, level=pblimit, insideMask=True)
    imsubMom8StatsInMask = runImstat(imsubMom8, False, imsubFlattenedMask,
                                     location=52, level=pblimit, insideMask=True)
    uvcontsubMom8StatsInMask = runImstat(uvcontsubMom8, False, uvcontsubFlattenedMask, 
                                         location=53, level=pblimit, insideMask=True)

    # region outside the mask
    imsubMom8LFStatsOutMask = runImstat(imsubMom8LineFree, False, imsubFlattenedMask,
                                       location=50, level=pblimit, insideMask=False)
    uvcontsubMom8LFStatsOutMask = runImstat(uvcontsubMom8LineFree, False, uvcontsubFlattenedMask, 
                                           location=51, level=pblimit, insideMask=False)
    imsubMom8StatsOutMask = runImstat(imsubMom8, False, imsubFlattenedMask,
                                     location=52, level=pblimit, insideMask=False)
    uvcontsubMom8StatsOutMask = runImstat(uvcontsubMom8, False, uvcontsubFlattenedMask, 
                                         location=53, level=pblimit, insideMask=False)

    if (minmaxScale == 'imsub'):
        minIntensity0 = imsubMom0Stats['min'][0]
        minIntensity8 = imsubMom8Stats['min'][0]
        maxIntensity0 = imsubMom0Stats['max'][0]
        maxIntensity8 = imsubMom8Stats['max'][0]
        minIntensityLF0 = imsubMom0LFStats['min'][0]
        minIntensityLF8 = imsubMom8LFStats['min'][0]
        maxIntensityLF0 = imsubMom0LFStats['max'][0]
        maxIntensityLF8 = imsubMom8LFStats['max'][0]
    else:
        minIntensity0 = uvcontsubMom0Stats['min'][0]
        minIntensity8 = uvcontsubMom8Stats['min'][0]
        maxIntensity0 = uvcontsubMom0Stats['max'][0]
        maxIntensity8 = uvcontsubMom8Stats['max'][0]
        minIntensityLF0 = uvcontsubMom0LFStats['min'][0]
        minIntensityLF8 = uvcontsubMom8LFStats['min'][0]
        maxIntensityLF0 = uvcontsubMom0LFStats['max'][0]
        maxIntensityLF8 = uvcontsubMom8LFStats['max'][0]
    if balanceIntensityLimitsLineFreeMoment0:
        a = np.max(np.abs([minIntensityLF0,maxIntensityLF0]))
        minIntensityLF0 = -a
        maxIntensityLF0 = +a
    imsubMom0Png = buildImagePng(imsubMom0, imageRadius, webpagedir, 
                                 minIntensity0, maxIntensity0, 
                                 contourImage=imsubPB, levels=0.5)
    imsubMom8Png = buildImagePng(imsubMom8, imageRadius, webpagedir, 
                                 minIntensity8, maxIntensity8,
                                 contourImage=imsubPB, levels=0.5)
    imsubMom0LineFreePng = buildImagePng(imsubMom0LineFree, imageRadius, webpagedir, 
                                         minIntensityLF0, maxIntensityLF0,
                                         contourImage=imsubPB, levels=0.5)
    imsubMom8LineFreePng = buildImagePng(imsubMom8LineFree, imageRadius, webpagedir, 
                                         minIntensityLF8, maxIntensityLF8,
                                         contourImage=imsubPB, levels=0.5)

    uvcontsubMom0Png = buildImagePng(uvcontsubMom0, imageRadius, webpagedir, 
                                     minIntensity0, maxIntensity0,
                                     contourImage=uvcontsubPB, levels=0.5)
    uvcontsubMom8Png = buildImagePng(uvcontsubMom8, imageRadius, webpagedir, 
                                     minIntensity8, maxIntensity8,
                                     contourImage=uvcontsubPB, levels=0.5)
    uvcontsubMom0LineFreePng = buildImagePng(uvcontsubMom0LineFree, imageRadius, webpagedir, 
                                             minIntensityLF0, maxIntensityLF0,
                                             contourImage=uvcontsubPB, levels=0.5)
    uvcontsubMom8LineFreePng = buildImagePng(uvcontsubMom8LineFree, imageRadius, webpagedir, 
                                             minIntensityLF8, maxIntensityLF8,
                                             contourImage=uvcontsubPB, levels=0.5)

    uvcontsubFindContinuumPng = lookForFile(uvcontsubImage+'.meanSpectrum*png')
    imsubFlattenedMaskPng = buildImagePng(imsubFlattenedMask, imageRadius, webpagedir, 0, 1, 
                                          colorwedge=False, contourImage=imsubPB, levels=0.5)
    uvcontsubFlattenedMaskPng = buildImagePng(uvcontsubFlattenedMask, imageRadius, webpagedir, 0, 1, 
                                              colorwedge=False, contourImage=imsubPB, levels=0.5)

    webpage += addplotrow(webpagedir, title='<b>new findContinuum result:</b><br>computed over flattened mask<br>and used for moment creation',
                          textForColumn1='same as uvcontsub',justifyColumn1='center',
                          plotForColumn2=uvcontsubFindContinuumPng)
    webpage += addplotrow(webpagedir, title='<b>Flattened mask<br></b>with contour showing pblevel=0.5',
                          textForColumn1='same as uvcontsub',justifyColumn1='center',
                          plotForColumn2=uvcontsubFlattenedMaskPng)
    webpage += addplotrow(webpagedir, title='<b>Peak<br></b>with contour showing pblevel=0.5',
                          plotForColumn1=imsubMom8Png, plotForColumn2=uvcontsubMom8Png)
    webpage += addplotrow(webpagedir, title='<b>Moment 0<br></b>with contour showing pblevel=0.5',
                          plotForColumn1=imsubMom0Png, plotForColumn2=uvcontsubMom0Png)
    if pixelsInFlattenedMask < 1:
        textForColumn = 'within pb>0.5'+3*'&emsp;'+'(no mask)'
    else:
        textForColumn = 'within pb>0.5'+3*'&emsp;'+'within mask'

    webpage += addplotrow(webpagedir, title='Region', justifyTitle='right',
                          textForColumn1=textForColumn, textForColumn2=textForColumn, bold=True)
    webpage += addplotrow(webpagedir, title='Moment 0 peak', justifyTitle='right',
                          textForColumn1=string_stats2(imsubMom0Stats,imsubMom0StatsInMask,'max',3),
                          textForColumn2=string_stats2(uvcontsubMom0Stats,uvcontsubMom0StatsInMask,'max',3))
    txt = open(acsTxt,'a')
    txtDict = {}
    txt.write('Moment Properties for spw=%s\n' % spwString.lstrip('spw'))
    txtDict['spw'] = int(spwString.lstrip('spw'))

    myvalue = imsubMom0Stats['max'][0]
    txt.write('Peak in imsub moment 0 (>0.5 pb) = %f\n' % myvalue)
    txtDict['peak_in_imsub_moment0_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom0StatsInMask['max'][0]
        txt.write('Peak in imsub moment 0 (within mask) = %f\n' % myvalue)
    else:
        txt.write('Peak in imsub moment 0 (within mask) = 0\n')
        myvalue = 0
    txtDict['peak_in_imsub_moment0_within_mask'] = myvalue

    myvalue = uvcontsubMom0Stats['max'][0]
    txt.write('Peak in uvcontsub moment 0 (>0.5 pb) = %f\n' % myvalue)
    txtDict['peak_in_uvcontsub_moment0_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        mypeak = uvcontsubMom0StatsInMask['max'][0]
        txt.write('Peak in uvcontsub moment 0 (within mask) = %f\n'%mypeak)
    else:
        mypeak = 0
        txt.write('Peak in uvcontsub moment 0 (within mask) = 0\n')
    txtDict['peak_in_uvcontsub_moment0_within_mask'] = mypeak

    webpage += addplotrow(webpagedir, title='Moment 0 flux', justifyTitle='right',
                          textForColumn1=string_stats2(imsubMom0Stats,imsubMom0StatsInMask,'flux',3),
                          textForColumn2=string_stats2(uvcontsubMom0Stats,uvcontsubMom0StatsInMask,'flux',3))

    myvalue = imsubMom0Stats['flux'][0]
    txt.write('Flux in imsub moment 0 (>0.5 pb) = %f\n' % myvalue)
    txtDict['flux_in_imsub_moment0_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom0StatsInMask['flux'][0]
        txt.write('Flux in imsub moment 0 (within mask) = %f\n'%myvalue)
    else:
        myvalue = 0
        txt.write('Flux in imsub moment 0 (within mask) = 0\n')
    txtDict['flux_in_imsub_moment0_within_mask'] = myvalue

    myvalue = uvcontsubMom0Stats['flux'][0]
    txt.write('Flux in uvcontsub moment 0 (>0.5 pb) = %f\n' % myvalue)
    txtDict['flux_in_uvcontsub_moment0_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = uvcontsubMom0StatsInMask['flux'][0]
        txt.write('Flux in uvcontsub moment 0 (within mask) = %f\n'%myvalue)
    else:
        myvalue = 0
        txt.write('Flux in uvcontsub moment 0 (within mask) = 0\n')
    txtDict['flux_in_uvcontsub_moment0_within_mask'] = myvalue

    if (len(imsubMom0StatsInMask['flux']) > 0):
        webpage += addplotrow(webpagedir, title='Moment 0 flux/rms', justifyTitle='right',
                          textForColumn1=roundFiguresToString(imsubMom0Stats['flux']/imsubMom0StatsBetween['rms'],3)+'&emsp;'*5+\
                                         roundFiguresToString(imsubMom0StatsInMask['flux']/imsubMom0StatsBetween['rms'],3),
                          textForColumn2=roundFiguresToString(uvcontsubMom0Stats['flux']/uvcontsubMom0StatsBetween['rms'],3)+'&emsp;'*5+\
                                         roundFiguresToString(uvcontsubMom0StatsInMask['flux']/uvcontsubMom0StatsBetween['rms'],3)
                          )
    else:
        webpage += addplotrow(webpagedir, title='Moment 0 flux/rms1', justifyTitle='right',
                          textForColumn1=roundFiguresToString(imsubMom0Stats['flux']/imsubMom0StatsBetween['rms'],3),
                          textForColumn2=roundFiguresToString(uvcontsubMom0Stats['flux']/uvcontsubMom0StatsBetween['rms'],3)
                          )

    myvalue = imsubMom0Stats['flux'][0]/imsubMom0StatsBetween['rms'][0]
    txt.write('Flux/rms in imsub moment 0 (>0.5 pb) = %f\n'%(myvalue))
    txtDict['flux_over_rms_in_imsub_moment0_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom0StatsInMask['flux'][0]/imsubMom0StatsBetween['rms'][0]
        txt.write('Flux/rms in imsub moment 0 (within mask) = %f\n'%(myvalue))
    else:
        myvalue = 0
        txt.write('Flux/rms in imsub moment 0 (within mask) = 0\n')
    txtDict['flux_over_rms_in_imsub_moment0_within_mask'] = myvalue

    myvalue = uvcontsubMom0Stats['flux'][0]/uvcontsubMom0StatsBetween['rms'][0]
    txt.write('Flux/rms in uvcontsub moment 0 (>0.5 pb) = %f\n'%(myvalue))
    txtDict['flux_over_rms_in_uvcontsub_moment0_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = uvcontsubMom0StatsInMask['flux'][0]/uvcontsubMom0StatsBetween['rms'][0]
        txt.write('Flux/rms in uvcontsub moment 0 (within mask) = %f\n'%(myvalue))
    else:
        myvalue = 0
        txt.write('Flux/rms in uvcontsub moment 0 (within mask) = 0\n')
    txtDict['flux_over_rms_in_uvcontsub_moment0_within_mask'] = myvalue

    webpage += addplotrow(webpagedir, title='<b>Moment 0 rms1</b><br>(within pb>0.5 and outside mask)', justifyTitle='right',
                          textForColumn1=string_stats(imsubMom0StatsBetween,'rms',3),
                          textForColumn2=string_stats(uvcontsubMom0StatsBetween,'rms',3))

    myvalue = imsubMom0StatsBetween['rms'][0]
    txt.write('rms in imsub moment 0 (>0.5 pb and outside mask) = %f\n'%myvalue)
    txtDict['rms_in_imsub_moment0_above_0.5pb_outside_mask'] = myvalue

    myvalue = uvcontsubMom0StatsBetween['rms'][0]
    txt.write('rms in uvcontsub moment 0 (>0.5 pb and outside mask) = %f\n'%myvalue)
    txtDict['rms_in_uvcontsub_moment0_above_0.5pb_outside_mask'] = myvalue

    webpage += addplotrow(webpagedir, title='<b>Moment 0 rms2</b><br>(outside the mask)', justifyTitle='right',
                          textForColumn1=string_stats(imsubMom0StatsOutMask,'rms',3),
                          textForColumn2=string_stats(uvcontsubMom0StatsOutMask,'rms',3))

    myvalue = imsubMom0StatsOutMask['rms'][0]
    txt.write('rms in imsub moment 0 (outside mask) = %f\n'%myvalue)
    txtDict['rms_in_imsub_moment0_outside_mask'] = myvalue
    
    myvalue = uvcontsubMom0StatsOutMask['rms'][0]
    txt.write('rms in uvcontsub moment 0 (outside mask) = %f\n'%myvalue)
    txtDict['rms_in_uvcontsub_moment0_outside_mask'] = myvalue

    webpage += addplotrow(webpagedir, title='<b>Peak (line-free)<br></b>with contour showing pblevel=0.5',
                          plotForColumn1=imsubMom8LineFreePng, plotForColumn2=uvcontsubMom8LineFreePng)

    webpage += addplotrow(webpagedir, title='<b>Moment 0 (line-free)<br></b>with contour showing pblevel=0.5',
                          plotForColumn1=imsubMom0LineFreePng, plotForColumn2=uvcontsubMom0LineFreePng)

    webpage += addplotrow(webpagedir, title='Region', justifyTitle='right',
                          textForColumn1=textForColumn,textForColumn2=textForColumn, bold=True)
    webpage += addplotrow(webpagedir, title='Moment 0 line-free peak', justifyTitle='right',
                          textForColumn1=string_stats2(imsubMom0LFStats,imsubMom0LFStatsInMask,'max',3),
                          textForColumn2=string_stats2(uvcontsubMom0LFStats,uvcontsubMom0LFStatsInMask,'max',3))

    myvalue = imsubMom0LFStats['max'][0]
    txt.write('Peak in imsub line-free moment 0 (>0.5 pb) = %f\n'%myvalue)
    txtDict['peak_in_imsub_linefree_moment0_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom0LFStatsInMask['max'][0]
        txt.write('Peak in imsub line-free moment 0 (within mask) = %f\n'%myvalue)
    else:
        myvalue = 0
        txt.write('Peak in imsub line-free moment 0 (within mask) = 0\n')
    txtDict['peak_in_imsub_linefree_moment0_within_mask'] = myvalue

    myvalue = uvcontsubMom0LFStats['max'][0]
    txt.write('Peak in uvcontsub line-free moment 0 (>0.5 pb) = %f\n'%myvalue)
    txtDict['peak_in_uvcontsub_linefree_moment0_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = uvcontsubMom0LFStatsInMask['max'][0]
        txt.write('Peak in uvcontsub line-free moment 0 (within mask) = %f\n'%myvalue)
    else:
        myvalue = 0
        txt.write('Peak in uvcontsub line-free moment 0 (within mask) = 0\n')
    txtDict['peak_in_uvcontsub_linefree_moment0_within_mask'] = myvalue

    webpage += addplotrow(webpagedir, title='Moment 0 line-free flux', justifyTitle='right',
                          textForColumn1=string_stats2(imsubMom0LFStats,imsubMom0LFStatsInMask,'flux',3),
                          textForColumn2=string_stats2(uvcontsubMom0LFStats,uvcontsubMom0LFStatsInMask,'flux',3))

    myvalue = imsubMom0LFStats['flux'][0]
    txt.write('Flux in imsub line-free moment 0 (>0.5 pb) = %f\n'%myvalue)
    txtDict['flux_in_imsub_linefree_moment0_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom0LFStatsInMask['flux'][0]
        txt.write('Flux in imsub line-free moment 0 (within mask) = %f\n'%myvalue)
    else:
        myvalue = 0
        txt.write('Flux in imsub line-free moment 0 (within mask) = 0\n')
    txtDict['flux_in_imsub_linefree_moment0_within_mask'] = myvalue

    myvalue = uvcontsubMom0LFStats['flux'][0]
    txt.write('Flux in uvcontsub line-free moment 0 (>0.5 pb) = %f\n'%myvalue)
    txtDict['flux_in_uvcontsub_linefree_moment0_above_0.5pb'] = myvalue
    
    if pixelsInFlattenedMask > 0:
        value = uvcontsubMom0LFStatsInMask['flux']
        txt.write('Flux in uvcontsub line-free moment 0 (within mask) = %f\n'%value)
    else:
        value = 0
        txt.write('Flux in uvcontsub line-free moment 0 (within mask) = 0\n')
    txtDict['flux_in_uvcontsub_linefree_moment0_within_mask'] = myvalue

    if (len(imsubMom0LFStatsInMask['flux']) > 0):
        webpage += addplotrow(webpagedir, title='Mom. 0 line-free flux/rms', justifyTitle='right',
                          textForColumn1=roundFiguresToString(imsubMom0LFStats['flux']/imsubMom0LFStatsBetween['rms'],3)+'&emsp;'*5+\
                                         roundFiguresToString(imsubMom0LFStatsInMask['flux']/imsubMom0LFStatsBetween['rms'],3),
                          textForColumn2=roundFiguresToString(uvcontsubMom0LFStats['flux']/uvcontsubMom0LFStatsBetween['rms'],3)+'&emsp;'*5+\
                                         roundFiguresToString(uvcontsubMom0LFStatsInMask['flux']/uvcontsubMom0LFStatsBetween['rms'],3))
    else:
        webpage += addplotrow(webpagedir, title='Mom. 0 line-free flux/rms', justifyTitle='right',
                              textForColumn1=roundFiguresToString(imsubMom0LFStats['flux']/imsubMom0LFStatsBetween['rms'],3),
                              textForColumn2=roundFiguresToString(uvcontsubMom0LFStats['flux']/uvcontsubMom0LFStatsBetween['rms'],3))

    myvalue = imsubMom0LFStats['flux'][0]/imsubMom0LFStatsBetween['rms'][0]
    txt.write('Flux/rms in imsub line-free moment 0 (>0.5 pb) = %f\n'%(myvalue))
    txtDict['flux_over_rms_in_imsub_linefree_moment0_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom0LFStatsInMask['flux'][0]/imsubMom0LFStatsBetween['rms'][0]
        txt.write('Flux/rms in imsub line-free moment 0 (within mask) = %f\n'%(myvalue))
    else:
        myvalue = 0
        txt.write('Flux/rms in imsub line-free moment 0 (within mask) = 0\n')
    txtDict['flux_over_rms_in_imsub_linefree_moment0_within_mask'] = myvalue

    myvalue = uvcontsubMom0LFStats['flux'][0]/uvcontsubMom0LFStatsBetween['rms'][0]
    txt.write('Flux/rms in uvcontsub line-free moment 0 (>0.5 pb) = %f\n'%(myvalue))
    txtDict['flux_over_rms_in_uvcontsub_linefree_moment0_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = uvcontsubMom0LFStatsInMask['flux'][0]/uvcontsubMom0LFStatsBetween['rms'][0]
        txt.write('Flux/rms in uvcontsub line-free moment 0 (within mask) = %f\n'%(myvalue))
    else:
        myvalue = 0
        txt.write('Flux/rms in uvcontsub line-free moment 0 (within mask) = 0\n')
    txtDict['flux_over_rms_in_uvcontsub_linefree_moment0_within_mask'] = myvalue

    webpage += addplotrow(webpagedir, title='<b>Moment 0 line-free rms1</b><br>(within pb>0.5 and outside mask)', justifyTitle='right',
                          textForColumn1=string_stats(imsubMom0LFStatsBetween,'rms',3),
                          textForColumn2=string_stats(uvcontsubMom0LFStatsBetween,'rms',3))

    myvalue = imsubMom0LFStatsBetween['rms'][0]
    txt.write('rms in imsub line-free moment 0 (>0.5 pb and outside mask) = %f\n'%(myvalue))
    txtDict['rms_in_imsub_linefree_moment0_above_0.5pb_outside_mask'] = myvalue
    
    myvalue = uvcontsubMom0LFStatsBetween['rms'][0]
    txt.write('rms in uvcontsub line-free moment 0 (>0.5 pb and outside mask) = %f\n'%myvalue)
    txtDict['rms_in_uvcontsub_linefree_moment0_above_0.5pb_outside_mask'] = myvalue

    webpage += addplotrow(webpagedir, title='<b>Moment 0 line-free rms2</b><br>(outside the mask)', justifyTitle='right',
                          textForColumn1=string_stats(imsubMom0LFStatsOutMask,'rms',3),
                          textForColumn2=string_stats(uvcontsubMom0LFStatsOutMask,'rms',3))

    myvalue = imsubMom0LFStatsOutMask['rms'][0]
    txt.write('rms in imsub line-free moment 0 (outside mask) = %f\n'%myvalue)
    txtDict['rms_in_imsub_linefree_moment0_outside_mask'] = myvalue
    
    myvalue = uvcontsubMom0LFStatsOutMask['rms'][0]
    txt.write('rms in uvcontsub line-free moment 0 (outside mask) = %f\n\n'%myvalue)
    txtDict['rms_in_uvcontsub_linefree_moment0_outside_mask'] = myvalue

    # Now write out the moment 8 stats

    myvalue = imsubMom8Stats['max'][0]
    txt.write('Peak in imsub moment 8 (>0.5 pb) = %f\n' % myvalue)
    txtDict['peak_in_imsub_moment8_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom8StatsInMask['max'][0]
        txt.write('Peak in imsub moment 8 (within mask) = %f\n' % myvalue)
    else:
        txt.write('Peak in imsub moment 8 (within mask) = 0\n')
        myvalue = 0
    txtDict['peak_in_imsub_moment8_within_mask'] = myvalue

    myvalue = uvcontsubMom8Stats['max'][0]
    txt.write('Peak in uvcontsub moment 8 (>0.5 pb) = %f\n' % myvalue)
    txtDict['peak_in_uvcontsub_moment8_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        mypeak = uvcontsubMom8StatsInMask['max'][0]
        txt.write('Peak in uvcontsub moment 8 (within mask) = %f\n'%mypeak)
    else:
        mypeak = 0
        txt.write('Peak in uvcontsub moment 8 (within mask) = 0\n')
    txtDict['peak_in_uvcontsub_moment8_within_mask'] = mypeak

    myvalue = imsubMom8Stats['flux'][0]
    txt.write('Flux in imsub moment 8 (>0.5 pb) = %f\n' % myvalue)
    txtDict['flux_in_imsub_moment8_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom8StatsInMask['flux'][0]
        txt.write('Flux in imsub moment 8 (within mask) = %f\n'%myvalue)
    else:
        myvalue = 0
        txt.write('Flux in imsub moment 8 (within mask) = 0\n')
    txtDict['flux_in_imsub_moment8_within_mask'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = uvcontsubMom8StatsInMask['flux'][0]
        txt.write('Flux in uvcontsub moment 8 (within mask) = %f\n'%myvalue)
    else:
        myvalue = 0
        txt.write('Flux in uvcontsub moment 8 (within mask) = 0\n')
    txtDict['flux_in_uvcontsub_moment8_within_mask'] = myvalue

    myvalue = imsubMom8Stats['flux'][0]/imsubMom8StatsBetween['rms'][0]
    txt.write('Flux/rms in imsub moment 8 (>0.5 pb) = %f\n'%(myvalue))
    txtDict['flux_over_rms_in_imsub_moment8_above_0.5pb'] = myvalue
    
    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom8StatsInMask['flux'][0]/imsubMom8StatsBetween['rms'][0]
        txt.write('Flux/rms in imsub moment 8 (within mask) = %f\n'%(myvalue))
    else:
        myvalue = 0
        txt.write('Flux/rms in imsub moment 8 (within mask) = 0\n')
    txtDict['flux_over_rms_in_imsub_moment8_within_mask'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = uvcontsubMom8StatsInMask['flux'][0]/uvcontsubMom8StatsBetween['rms'][0]
        txt.write('Flux/rms in uvcontsub moment 8 (within mask) = %f\n'%(myvalue))
    else:
        myvalue = 0
        txt.write('Flux/rms in uvcontsub moment 8 (within mask) = 0\n')
    txtDict['flux_over_rms_in_uvcontsub_moment8_within_mask'] = myvalue

    myvalue = uvcontsubMom8Stats['flux'][0]
    txt.write('Flux in uvcontsub moment 8 (>0.5 pb) = %f\n' % myvalue)
    txtDict['flux_in_uvcontsub_moment8_above_0.5pb'] = myvalue

    myvalue = uvcontsubMom8Stats['flux'][0]/uvcontsubMom8StatsBetween['rms'][0]
    txt.write('Flux/rms in uvcontsub moment 8 (>0.5 pb) = %f\n'%(myvalue))
    txtDict['flux_over_rms_in_uvcontsub_moment8_above_0.5pb'] = myvalue

    myvalue = imsubMom8StatsBetween['rms'][0]
    txt.write('rms in imsub moment 8 (>0.5 pb and outside mask) = %f\n'%myvalue)
    txtDict['rms_in_imsub_moment8_above_0.5pb_outside_mask'] = myvalue

    myvalue = uvcontsubMom8StatsBetween['rms'][0]
    txt.write('rms in uvcontsub moment 8 (>0.5 pb and outside mask) = %f\n'%myvalue)
    txtDict['rms_in_uvcontsub_moment8_above_0.5pb_outside_mask'] = myvalue

    myvalue = imsubMom8StatsOutMask['rms'][0]
    txt.write('rms in imsub moment 8 (outside mask) = %f\n'%myvalue)
    txtDict['rms_in_imsub_moment8_outside_mask'] = myvalue
    
    myvalue = uvcontsubMom8StatsOutMask['rms'][0]
    txt.write('rms in uvcontsub moment 8 (outside mask) = %f\n'%myvalue)
    txtDict['rms_in_uvcontsub_moment8_outside_mask'] = myvalue

    myvalue = imsubMom8LFStats['max'][0]
    txt.write('Peak in imsub line-free moment 8 (>0.5 pb) = %f\n'%myvalue)
    txtDict['peak_in_imsub_linefree_moment8_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom8LFStatsInMask['max'][0]
        txt.write('Peak in imsub line-free moment 8 (within mask) = %f\n'%myvalue)
    else:
        myvalue = 0
        txt.write('Peak in imsub line-free moment 8 (within mask) = 0\n')
    txtDict['peak_in_imsub_linefree_moment8_within_mask'] = myvalue

    myvalue = uvcontsubMom8LFStats['max'][0]
    txt.write('Peak in uvcontsub line-free moment 8 (>0.5 pb) = %f\n'%myvalue)
    txtDict['peak_in_uvcontsub_linefree_moment8_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = uvcontsubMom8LFStatsInMask['max'][0]
        txt.write('Peak in uvcontsub line-free moment 8 (within mask) = %f\n'%myvalue)
    else:
        myvalue = 0
        txt.write('Peak in uvcontsub line-free moment 8 (within mask) = 0\n')
    txtDict['peak_in_uvcontsub_linefree_moment8_within_mask'] = myvalue

    myvalue = imsubMom8LFStats['flux'][0]
    txt.write('Flux in imsub line-free moment 8 (>0.5 pb) = %f\n'%myvalue)
    txtDict['flux_in_imsub_linefree_moment8_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom8LFStatsInMask['flux'][0]
        txt.write('Flux in imsub line-free moment 8 (within mask) = %f\n'%myvalue)
    else:
        myvalue = 0
        txt.write('Flux in imsub line-free moment 8 (within mask) = 0\n')
    txtDict['flux_in_imsub_linefree_moment8_within_mask'] = myvalue

    myvalue = uvcontsubMom8LFStats['flux'][0]
    txt.write('Flux in uvcontsub line-free moment 8 (>0.5 pb) = %f\n'%myvalue)
    txtDict['flux_in_uvcontsub_linefree_moment8_above_0.5pb'] = myvalue
    
    if pixelsInFlattenedMask > 0:
        value = uvcontsubMom8LFStatsInMask['flux']
        txt.write('Flux in uvcontsub line-free moment 8 (within mask) = %f\n'%value)
    else:
        value = 0
        txt.write('Flux in uvcontsub line-free moment 8 (within mask) = 0\n')
    txtDict['flux_in_uvcontsub_linefree_moment8_within_mask'] = myvalue

    myvalue = imsubMom8LFStats['flux'][0]/imsubMom8LFStatsBetween['rms'][0]
    txt.write('Flux/rms in imsub line-free moment 8 (>0.5 pb) = %f\n'%(myvalue))
    txtDict['flux_over_rms_in_imsub_linefree_moment8_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = imsubMom8LFStatsInMask['flux'][0]/imsubMom8LFStatsBetween['rms'][0]
        txt.write('Flux/rms in imsub line-free moment 8 (within mask) = %f\n'%(myvalue))
    else:
        myvalue = 0
        txt.write('Flux/rms in imsub line-free moment 8 (within mask) = 0\n')
    txtDict['flux_over_rms_in_imsub_linefree_moment8_within_mask'] = myvalue

    myvalue = uvcontsubMom8LFStats['flux'][0]/uvcontsubMom8LFStatsBetween['rms'][0]
    txt.write('Flux/rms in uvcontsub line-free moment 8 (>0.5 pb) = %f\n'%(myvalue))
    txtDict['flux_over_rms_in_uvcontsub_linefree_moment8_above_0.5pb'] = myvalue

    if pixelsInFlattenedMask > 0:
        myvalue = uvcontsubMom8LFStatsInMask['flux'][0]/uvcontsubMom8LFStatsBetween['rms'][0]
        txt.write('Flux/rms in uvcontsub line-free moment 8 (within mask) = %f\n'%(myvalue))
    else:
        myvalue = 0
        txt.write('Flux/rms in uvcontsub line-free moment 8 (within mask) = 0\n')
    txtDict['flux_over_rms_in_uvcontsub_linefree_moment8_within_mask'] = myvalue

    myvalue = imsubMom8LFStatsBetween['rms'][0]
    txt.write('rms in imsub line-free moment 8 (>0.5 pb and outside mask) = %f\n'%(myvalue))
    txtDict['rms_in_imsub_linefree_moment8_above_0.5pb_outside_mask'] = myvalue
    
    myvalue = uvcontsubMom8LFStatsBetween['rms'][0]
    txt.write('rms in uvcontsub line-free moment 8 (>0.5 pb and outside mask) = %f\n'%myvalue)
    txtDict['rms_in_uvcontsub_linefree_moment8_above_0.5pb_outside_mask'] = myvalue

    myvalue = imsubMom8LFStatsOutMask['rms'][0]
    txt.write('rms in imsub line-free moment 8 (outside mask) = %f\n'%myvalue)
    txtDict['rms_in_imsub_linefree_moment8_outside_mask'] = myvalue
    
    myvalue = uvcontsubMom8LFStatsOutMask['rms'][0]
    txt.write('rms in uvcontsub line-free moment 8 (outside mask) = %f\n\n'%myvalue)
    txtDict['rms_in_uvcontsub_linefree_moment8_outside_mask'] = myvalue




    txt.close()
    pickleFile = rreplace(acsTxt, '.txt', '.spw%d.p'%txtDict['spw'])
    print "Writing pickle file = ", pickleFile
    pickle.dump({spwString:txtDict}, open(pickleFile,'wb'))
    os.system('chmod g+w ' + pickleFile)

    # Finished with this pair of images
    webpage += '</table>\n'
    webpage += 'Elapsed time = %s<br><br><br><br>' % (showElapsedTime())
    elapsedTime = time.time()-singleStartTime
    if not appendhtml:
        if os.path.exists(webfile): 
            os.remove(webfile)
        finishWebpage(webfile, webpage)
    else:
        f = open(webfile,'a')
        f.write(webpage)
        f.close()
    print "%.1f sec = %.1f min elapsed time" % (elapsedTime, elapsedTime/60.)
    return(elapsedTime)

def flushWebpage(webfile,webpage):
    f = open(webfile,"a")
    f.write(webpage)
    f.close()

def finishWebpage(webfile, data=''):
    """
    Write an optional final batch of data to the specified text file, then
    close the html tags.
    """
    f = open(webfile,"a")
    print >>f, data + """
         </body>
         </html>
      """
    f.close()
    print "Finished webpage = ", webfile
