#!/usr/bin/env python
import os, sys, traceback, email, errno, mimetypes, xml.dom.minidom, StringIO, getopt, re, exceptions, operator, string
from string import *
from tokenize import *
from token import *

from xml.dom.minidom import parse, parseString

class ArrayTime:

    """
    A very limited subset of the ArrayTime class as it's defined in the C++ or Java implementations. We define here 
    only what's required to build a string expressing a date and a time (readable by a human eye) from an integer value
    representing a number of nanoseconds since an origin of time which is 15 October 1582 00:00:00 UTC (Gregorian time).
    """

    numberSigDigitsInASecond = 9
    unitsInASecond           = 1000000000L
    unitsInADayL             = 86400000000000L
    julianDayOfBase          = 2400000.5
 
    def unitToJD(self, nanocount):
        return (1.0 * nanocount / self.unitsInADayL) + self.julianDayOfBase

    def getDateTime(self, nanocount) :
        """
        Return this time expressed by nanocount as an array of integers denoting the following:
        year,
        month (varies from 1 to 12),
        day (varies from 1 to 28, 29, 30, or 31),
        hour (varies from 0 to 23),
        minute (varies from 0 to 59),
        second (varies from 0 to 59), and
        the number of nanoseconds that remain in this fraction of a second.
        """
        fractionOfADay = nanocount % self.unitsInADayL

        if fractionOfADay < 0:
            fractionOfADay = self.unitsInADayL - fractionOfADay
        
        nsec = fractionOfADay / self.unitsInASecond
        n6   = int(fractionOfADay - (nsec * 1000000000L))
        nmin = nsec / 60
        n5   = int(nsec - nmin * 60)
        nhr  = nmin / 60
        n4   = int(nmin - nhr * 60L)
        n3   = int(nhr)

        jd = self.unitToJD(nanocount)

    # For this algorithm see Meeus, chapter 7, p. 63.
        x = jd + 0.5 # Make the 12h UT adjustment.
        Z = int(x)
        F = x - Z
        A = Z
        alpha = 0
        
        if Z >= 2299161 :
            alpha = int((Z - 1867216.25) / 36524.25)
            A = Z + 1 + alpha - (int)(alpha / 4)

        B = A + 1524
        C = int((B - 122.1) / 365.25)
        D = int(365.25 * C)
        E = int((B - D) / 30.6001)
        day = B - D - int(30.6001 * E) + F
        
        if E < 14 :
            month = E - 1
        else:
            month = E - 13

        if month > 2 :
            year = C - 4716
        else:
            year = C - 4715

        n2 = int(day)
        n1 = month
        n0 = year
        
        return [n0, n1, n2, n3, n4, n5, n6]


    def toFITS(self,nanocount) :
        """ 
        Given nanocount an integer expected to represent an integer number of nanoseconds since the time origin
        return a string which represents a number of nanoseconds since the origin of time.
        """
        unit = self.getDateTime(nanocount)
        yy = unit[0]
        mm = unit[1]
        dd = unit[2]
        hh = unit[3]
        min = unit[4]
        sec = unit[5]

        s = str(yy); s = s +  '-'
        if mm < 10 :
            s = s +  '0'
        s = s +  str(mm); s = s +  '-'

        if (dd < 10) :
            s = s +  '0'
        s = s +  str(dd)
        s = s +  'T'

        if hh < 10 :
            s = s +  '0'
        s = s +  str(hh)
        s = s +  ':'

        if min < 10 :
            s = s +  '0'
        s = s +  str(min)
        s = s +  ':'

        if sec < 10 :
            s = s +  '0'
        s = s +  str(sec)

        # apply fractions of a second
        frac = str(unit[6])
        s = s +  '.'
        # The statement below is sensitive to the number of significant
        # digits in a fraction.  If units are nanoseconds,
        # then we need we will have 9 significant digits in a fraction 
        # string.
        tmp = "0000000000000000"
        s = s +  tmp[0:self.numberSigDigitsInASecond - len(frac)]
        s = s +  frac
        return s

class SDMDataObjectStreamReader:

    def position(self):
        return self.f.tell()

    def nextLine(self):
        line = self.f.readline()
        if len(line) > 0 and line[len(line)-1] == "\n":
            line = line[:-1]
        self.currentLine = line
        return line

    def HeaderField2Pair(self, hf):
        name = None
        value = None
        colonIndex = string.find(hf, ":")
        if colonIndex == -1:
            return (name, value)

        if colonIndex > 0:
            name = hf[:colonIndex]

        if colonIndex < len(hf) - 1:
            value = hf[colonIndex + 1:]

        return (name, value)

    def requireHeaderField(self, hf):
        name, value = self.HeaderField2Pair(self.nextLine())
        #print name, value
        if string.upper(name) != hf:
            print 'Missing "', hf, '" field.'
            sys.exit(1)
        return name, value

    def unquote(self, s):
        if len(s) < 2:
            return s
        if s[0] == '"' and s[len(s)-1] == '"':
            return s[1:-1]
        if s[0] == "'" and s[len(s)-1] == "'":
            return s[1:-1]
        return s

    def requireBoundaryInCT(self, ctValue):
        cvValueItems = [item.lstrip().rstrip() for item in ctValue.split(";")] 
        cvValueItemsNameValue = [item.partition("=") for item in cvValueItems]
        boundaryValues = [item[2] for item in cvValueItemsNameValue if string.upper(item[0]) == 'BOUNDARY' and item[2] != ''] 
        if boundaryValues == []:
            return None
        else:
            return self.unquote(boundaryValues[0])

    def skipUntilEmptyLine(self, maxSkips):
        numSkip = 0
        line = self.nextLine()
        while line != '' and numSkip <= maxSkips :
            line = self.nextLine()
            numSkip+=1
        if numSkip > maxSkips:
            print "could not find an expected empty line"
            sys.exit(1)

    def requireMIMEHeader(self):
        # MIME-Version
        line = self.nextLine()
        name, value = self.HeaderField2Pair(line)
        #print name, value
        #if line != "MIME-Version: 1.0":
        if not line.endswith("IME-Version: 1.0"):
            print 'Missing "MIME-Version: 1.0"'
            Sys.exit(1)

        # Content-Type
        name, value = self.requireHeaderField("CONTENT-TYPE")
        # boundary in value
        boundary_1 = self.requireBoundaryInCT(value)
        self.boundary_1 = boundary_1
        #print boundary_1

        # Content-Description
        #print self.requireHeaderField("CONTENT-DESCRIPTION")

        # Content-Location
        #print self.requireHeaderField("CONTENT-LOCATION")

        # Look for an empty line during 10 lines maximum.
        self.skipUntilEmptyLine(10)

        return boundary_1

    def accumulateUntilBoundary(self, boundary, maxLines):
        numLines = 0
        line = self.nextLine()
        result = ""
        while  numLines <= maxLines and string.find(line, "--"+boundary) != 0:
            result += line
            numLines += 1
            line = self.nextLine()

        if numLines > maxLines:
            print "Could not find '--" + boundary + "' in the next " + maxLines + "."
            sys.exit(1)

        return result

    def requireBoundary(self, boundary, maxLines):
        numLines = 0
        line = self.nextLine()
        while numLines <= maxLines and line != "--"+boundary:
            numLines += 1
            line = self.nextLine()

        if numLines > maxLines:
            print "Could not read the expected boundary '"+boundary+"'"
            sys.exit(1)

    def lookForBinaryPartSize(self, dom, partName):
        result = None
        elements = dom.getElementsByTagName(partName)
        if elements != []:
            size = elements.item(0).getAttribute("size")
            if size == "":
                print "Missing 'size' attribute in element '", element.nodeName, "."
                sys.exit(1)
            result = int(size)
        return result

    def requireCrossDataType(self, dom):
        result = None
        elements = dom.getElementsByTagName("crossData")
        if elements == []:
            print "Missing  'crossData' element in '" , dom.toprettyxml() , "'"
            sys.exit(1)

        result = elements.item(0).getAttribute("type")
        if result == "":
            print "Missing 'type' attribute in element '", element.nodeName, "."
            sys.exit(1)

        return result

    def requireSDMDataHeaderMIMEPart(self):
        #
        # Require the presence of boundary
        self.requireBoundary(self.boundary_1, 0)

        #
        # Ignore header fields
        #
        #requireHeaderField(f, "CONTENT-TYPE")
        #requireHeaderField(f, "CONTENT-TRANSFER-ENCODING")
        #requireHeaderField(f, "CONTENT-LOCATION")

        self.skipUntilEmptyLine(10)
        sdmDataHeader = self.accumulateUntilBoundary(self.boundary_1, 100)
        dom = parseString(sdmDataHeader)
        self.binaryPartSize = {}

        for partName in ["actualDurations", "actualTimes", "autoData", "crossData", "zeroLags", "flags"]:
            size = self.lookForBinaryPartSize(dom, partName)
            if size != None:
                self.binaryPartSize[partName] = size

        #print dom.toprettyxml()
        #print self.binaryPartSize
        self.sdmDataHeaderDOM = dom

    def requireSDMDataSubsetMIMEPart(self):
        name, value = self.requireHeaderField("CONTENT-TYPE")
        self.boundary_2  = self.requireBoundaryInCT(value)
        
        self.requireHeaderField("CONTENT-DESCRIPTION")
        self.requireBoundary(self.boundary_2, 10)
        self.skipUntilEmptyLine(10)

        sdmDataSubsetHeader = self.accumulateUntilBoundary(self.boundary_2, 100)
        sdmDataSubsetHeaderDOM = parseString(sdmDataSubsetHeader)

        done = False
        crossDataDesc = {}
        autoDataDesc = {}
        autoDataDesc["type"] = "FLOAT32_TYPE"
        flagsDesc = {}
        flagsDesc["type"] = "INT16_TYPE"
        flagsDesc["present"] = False
        while not done :
            self.requireHeaderField("CONTENT-TYPE")
            name, value = self.requireHeaderField("CONTENT-LOCATION")
            result = re.match("([0-9]+/)+(actualDurations|actualTimes|autoData|crossData|zeroLags|flags)\.bin", value.lstrip().rstrip())
            if result == None:
                print "Could not identify the part name in '" , value , "'"
                sys.exit(1)

            binaryPartName = result.group(2)

            if not self.binaryPartSize.has_key(binaryPartName):
                print "The size of '" + binaryPartName + "' has not been announced in the  data header"
                sys.exit(1)

            if binaryPartName == "crossData":
                crossDataType = self.requireCrossDataType(sdmDataSubsetHeaderDOM)
                crossDataDesc["type"] = str(crossDataType)
                            
            self.skipUntilEmptyLine(10)
            if binaryPartName == "actualDurations" or binaryPartName == "actualTimes":
                numberOfBytesPerValue = 8 
            elif binaryPartName == "autoData" :
                numberOfBytesPerValue = 4
            elif binaryPartName == "crossData" :
                if crossDataType == "INT16_TYPE":
                    numberOfBytesPerValue = 2
                elif crossDataType == "INT32_TYPE" or crossDataType == "FLOAT32_TYPE" :
                    numberOfBytesPerValue = 4
            elif binaryPartName == "flags":
                numberOfBytesPerValue = 4
            elif binaryPartName == "zeroLags":
                numberOfBytesPerValue = 4

            numberOfBytesToRead = numberOfBytesPerValue * self.binaryPartSize[binaryPartName]
            if binaryPartName == "crossData":
                crossDataDesc["base"] = self.position()
            elif binaryPartName == "autoData":
                autoDataDesc["base"] = self.position()
            elif binaryPartName == "flags":
                flagsDesc["base"] = self.position()
                flagsDesc["present"] = True

            bytes = self.f.read(numberOfBytesToRead) 
            if len(bytes) < numberOfBytesToRead:
                print "EOF reached while reading a binary attachment ('" + binaryPartName + "')"
                sys.exit(1)

            #print "I have read " , self.binaryPartSize[binaryPartName] , " values for " , binaryPartName

            line = self.nextLine() # Absorb the nl right after the last byte of the binary attachment
            line = self.nextLine() # This should boundary_2

            if string.find(line, "--"+self.boundary_2) != 0:
                print "Unexpected '", line, "' after the binary part '", binaryPartName, "'" 
                sys.exit(1)

            done = line == "--"+self.boundary_2+"--"

            integrationStartTime = int((sdmDataSubsetHeaderDOM.getElementsByTagName("time"))[0].childNodes[0].nodeValue)
            integrationInterval  = int((sdmDataSubsetHeaderDOM.getElementsByTagName("interval"))[0].childNodes[0].nodeValue)
        return (integrationStartTime, integrationInterval, crossDataDesc, autoDataDesc, flagsDesc)

    def open(self, f):
        self.f = f
        self.boundary_1 = self.requireMIMEHeader()
        self.requireSDMDataHeaderMIMEPart()

    def getSDMDataHeader(self):
        return self.sdmDataHeaderDOM

    def hasData(self):
        return self.currentLine != "--"+self.boundary_1+"--"

    def getData(self):
        midpoint, interval, cdd, add, fdd = self.requireSDMDataSubsetMIMEPart()
        line = self.nextLine()
        return (midpoint, interval, cdd, add, fdd)

    def processBDF(self, f):
        print "in processBDF"
        boundary_1 = self.requireMIMEHeader()

        self.requireSDMDataHeaderMIMEPart(boundary_1)


MSMainRowNum = 0

def traverseALMARadiometricData(numAntenna, dataDesc, flagsDesc, basebands):
    global MSMainRowNum
    position = dataDesc["base"]
    numberOfBytesPerValue = 0
    numChan = 0
    numPol = 0

    if dataDesc["type"] == "FLOAT32_TYPE" :
        numberOfBytesPerValue = 4
    else:
        print "Unknown data type '%s' for radiometric data" % dataDesc["type"]

    for time in range(int(dataDesc["numTime"])):
        for antenna in range(numAntenna):
            for baseband in basebands:
                for spectralWindow in baseband:
                    numChan = spectralWindow["numSpectralPoint"]
                    numPol = len(spectralWindow["sdPolProducts"].split())
                    osio = StringIO.StringIO()
                    MSMainRowNum = MSMainRowNum + 1
                    osio.write(MSMainRowNum)
                    osio.write("|%s" %  dataDesc["bdfName"])
                    osio.write("|float")
                    osio.write("|%s" % dataDesc["type"])
                    osio.write("|auto")
                    osio.write("|%d" % numChan )
                    osio.write("|%d" % numPol )
                    osio.write("|%d" % position)
                    s = osio.getvalue()
                    osio.close()
                    print s
                    
                    position = position + numberOfBytesPerValue * numPol * numChan   # Because everything is real.
                    

def traverseALMABaseband(dataDesc, basebands, base):
    global MSMainRowNum
    position = base
    numberOfBytesPerValue = 0
    numChan = 0
    numPol = 0
    
    if dataDesc["crossOrAuto"] == "cross" :

        numberOfBytesPerValue = 4
        if dataDesc["type"] == "INT16_TYPE":
            numberOfBytesPerValue = 2
            
        for baseband in basebands:
            for spectralWindow in baseband:
                numChan = spectralWindow["numSpectralPoint"]
                numPol = len(spectralWindow["crossPolProducts"].split())
                osio = StringIO.StringIO()
                MSMainRowNum = MSMainRowNum + 1
                osio.write(MSMainRowNum)
                osio.write("|%s" %  dataDesc["bdfName"])
                osio.write("|complex")
                osio.write("|%s" % dataDesc["type"])
                osio.write("|cross:%s" % spectralWindow["scaleFactor"])
                osio.write("|%d" % numChan)
                osio.write("|%d" % numPol) 
                osio.write("|%d" % position)
                position = position + 2 * numberOfBytesPerValue * numChan * numPol
                s = osio.getvalue()
                osio.close()
                print s

    elif dataDesc["crossOrAuto"] == "auto" :
        numberOfBytesPerValue = 4

        for baseband in basebands:
            for spectralWindow in baseband:
                numChan = spectralWindow["numSpectralPoint"]
                numPol = len(spectralWindow["sdPolProducts"].split())
                mixedAutoDataType = dataDesc["correlationMode"] == "CROSS_AND_AUTO" and spectralWindow["sdPolProducts"] == "XX XY YY"
                osio = StringIO.StringIO()
                MSMainRowNum = MSMainRowNum + 1
                osio.write(MSMainRowNum)
                osio.write("|%s" %  dataDesc["bdfName"])
                if mixedAutoDataType:
                    osio.write("|mixed")
                else:
                    osio.write("|float")

                osio.write("|%s" % dataDesc["type"])
                osio.write("|auto")
                osio.write("|%d" % numChan )
                osio.write("|%d" % numPol )
                osio.write("|%d" % position)
                s = osio.getvalue()
                osio.close()
                print s

                if mixedAutoDataType :
                    position = position + numberOfBytesPerValue * 4 * numChan   # Because = 1 (real) + 2 (complex) + 1 (real)
                else:
                    position = position + numberOfBytesPerValue * numPol * numChan   # Because everything is real.
                
    else:
        print "Invalid value of crossOrAuto '%s'" % dataDesc["crossOrAuto"]
        exit

    return position

def traverseALMABAL(numAntenna, crossDataDesc, basebands):
    base = crossDataDesc["base"]
    for antenna0 in range (1, numAntenna+1):
        for antenna1 in range (antenna0+1, numAntenna+1):
            base = traverseALMABaseband(crossDataDesc, basebands, base)

def traverseALMAANT(numAntenna, autoDataDesc, basebands):
    base = autoDataDesc["base"]
    for antenna0 in range(numAntenna):
        base = traverseALMABaseband(autoDataDesc, basebands, base)
        

def traverseALMACrossAndAutoData(numAntenna, crossDataDesc, autoDataDesc, flagsDesc, basebands):
    # The BAL first
    traverseALMABAL(numAntenna, crossDataDesc, basebands)

    # Then the ANT
    traverseALMAANT(numAntenna, autoDataDesc, basebands)

def main() :
    #
    # The path to the ASDM is expected to be argv[1]
    #
    try:
        options, arg = getopt.getopt(sys.argv[1:], ":h", ["help"])
    except getopt.GetoptError, err:
            usage();
            sys.exit(2)

    #
    # Build the ordered list of the BDF files from the content of the ASDM Main table
    #
    asdmMainFile = arg[0] + "/Main.xml"
    
    asdmMainDOM = xml.dom.minidom.parse(asdmMainFile)
    bdfNames = [ string.replace(string.replace(entityRef.getAttribute("entityId"), "/", "_", ), ":", "_" ) for entityRef in asdmMainDOM.getElementsByTagName("EntityRef") ]

    crossDataAxesRE = re.compile("BAL( +BAB)?( +APC)?( +SPP)?( +POL)?") 
    autoDataAxesRE = re.compile("ANT( +BAB)?( +APC)?( +SPP)?( +POL)?")    
    flagsAxesRE = re.compile("BAL ANT BAB( +POL)?")
    radiometerFlagsAxesRE = re.compile("TIM ANT")

    #
    # now process sequentially the bdf files
    #
    for bdfName in bdfNames :
        #print "> Processing %s" %  bdfName
        try :
            f  = open(arg[0]+"/ASDMBinary/"+bdfName)
        except IOError, e:
            print e;
            continue

        sdosr = SDMDataObjectStreamReader()
        sdosr.open(f)
        globalHeaderDOM = sdosr.getSDMDataHeader()
        #
        # What's the processor type ?
        #
        processorType = (globalHeaderDOM.getElementsByTagName("processorType"))[0].childNodes[0].nodeValue

        #
        # What's the correlation mode ?
        #
        correlationMode = (globalHeaderDOM.getElementsByTagName("correlationMode"))[0].childNodes[0].nodeValue

        #
        # What's the spectral resolution type ?
        #
        spectralResolution = (globalHeaderDOM.getElementsByTagName("spectralResolution"))[0].childNodes[0].nodeValue

        flagsDesc = {}
        flagsDesc["correlationMode"] = correlationMode
        elems = globalHeaderDOM.getElementsByTagName("flags")
        if len(elems) > 0:
            flagsDesc["size"] = int(elems[0].getAttribute("size"))
            flagsDesc["axes"] = str(elems[0].getAttribute("axes"))

        #
        # The process depends on the processor type.
        #
        if processorType == "CORRELATOR":

            if "axes" in flagsDesc and flagsAxesRE.match(flagsDesc["axes"]) == None:
                print "These axes '%s' are not recognize for flags." % flagsDesc["axes"]
                continue

            crossDataDesc = {}
            crossDataDesc["bdfName"] = bdfName
            crossDataDesc["crossOrAuto"] = "cross"
            crossDataDesc["correlationMode"] = correlationMode

            elems = globalHeaderDOM.getElementsByTagName("crossData")
            if len(elems) > 0:
                crossDataDesc["size"] = int(elems[0].getAttribute("size"))
                crossDataDesc["axes"] = str(elems[0].getAttribute("axes"))
                
            if crossDataAxesRE.match(crossDataDesc["axes"]) == None:
                print "These axes '%s' are not recognized for cross data." % crossDataDesc["axes"] 
                continue

            autoDataDesc = {}
            autoDataDesc["bdfName"] = bdfName
            autoDataDesc["crossOrAuto"] = "auto"
            autoDataDesc["correlationMode"] = correlationMode

            elems = globalHeaderDOM.getElementsByTagName("autoData")
            if len(elems) > 0:
                autoDataDesc["size"] = int(elems[0].getAttribute("size"))
                autoDataDesc["axes"] = str(elems[0].getAttribute("axes"))
                normalized = str(elems[0].getAttribute("normalized"))
                if normalized == None:
                    autoDataDesc["normalized"] = False
                else:
                    autoDataDesc["normalized"] = normalized == "true"

            if autoDataAxesRE.match(autoDataDesc["axes"]) == None:
                print "These axes '%s' are not recognized for auto data." % autoDataDesc["axes"] 
                continue

            numAntenna = int(globalHeaderDOM.getElementsByTagName("numAntenna")[0].childNodes[0].nodeValue)
            
            #
            # listify the baseband elements
            #
            basebands          = []
            for bb in globalHeaderDOM.getElementsByTagName("baseband"):
                spws           = []
                for spw in globalHeaderDOM.getElementsByTagName("spectralWindow"):
                    attributes = {"sdPolProducts": str(spw.getAttribute("sdPolProducts")),
                                  "crossPolProducts": str(spw.getAttribute("crossPolProducts")),
                                  "scaleFactor": float(spw.getAttribute("scaleFactor")),
                                  "numSpectralPoint": int(spw.getAttribute("numSpectralPoint")),
                                  "numBin" : int(spw.getAttribute("numBin"))}
                    spws.append(attributes)
                basebands.append(spws)

            #
            # Then let's traverse the BDF.
            #
            while sdosr.hasData() :
                midpoint, interval, cdd, add, fdd = sdosr.getData()
                for k, v in cdd.iteritems(): 
                    crossDataDesc[k]=v
                    
                for k, v in add.iteritems():
                    autoDataDesc[k]=v
                
                flagsDesc = {}
                for k, v in fdd.iteritems():
                    flagsDesc[k]=v
                
                if flagsDesc["present"] :
                    print "There are flags in '%s' at '%s' (processor = '%s', correlationMode = '%s', spectralResolution = '%s') " % (bdfName,
                                                                                                                                          ArrayTime().toFITS(midpoint - interval/2),
                                                                                                                                          processorType,
                                                                                                                                          correlationMode,
                                                                                                                                          spectralResolution)


            #  traverseALMACrossAndAutoData(numAntenna, crossDataDesc, autoDataDesc, flagsDesc, basebands)

        elif processorType == "RADIOMETER":

            if "axes" in flagsDesc and radiometerFlagsAxesRE.match(flagsDesc["axes"]) == None:
                print "These axes '%s' are not recognize for flags." % flagsDesc["axes"]
                continue

            radiometricDataDesc = {}
            radiometricDataDesc["bdfName"] = bdfName
            radiometricDataDesc["crossOrAuto"] = "auto"
            radiometricDataDesc["correlationMode"] = correlationMode
            radiometricDataDesc["numTime"] = (globalHeaderDOM.getElementsByTagName("numTime"))[0].childNodes[0].nodeValue

            elems = globalHeaderDOM.getElementsByTagName("autoData")
            if len(elems) > 0:
                radiometricDataDesc["size"] = int(elems[0].getAttribute("size"))
                radiometricDataDesc["axes"] = str(elems[0].getAttribute("axes"))
                normalized = str(elems[0].getAttribute("normalized"))
                if normalized == None:
                    radiometricDataDesc["normalized"] = False
                else:
                    radiometricDataDesc["normalized"] = normalized == "true"        
            

            basebands          = []
            for bb in globalHeaderDOM.getElementsByTagName("baseband"):
                spws           = []
                for spw in globalHeaderDOM.getElementsByTagName("spectralWindow"):
                    attributes = {"sdPolProducts": str(spw.getAttribute("sdPolProducts")),
                                  "numSpectralPoint": int(spw.getAttribute("numSpectralPoint")),
                                  "numBin" : int(spw.getAttribute("numBin"))}
                    spws.append(attributes)
                basebands.append(spws)

            midpoint, interval, cdd, add, fdd = sdosr.getData()
            for k, v in add.iteritems():
                radiometricDataDesc[k]=v

            for k, v in fdd.iteritems():
                flagsDesc[k] = v

            if flagsDesc["present"] :
                print "There are flags in '%s' at '%s' (processor = '%s', correlationMode = '%s', spectralResolution = '%s') " % (bdfName,
                                                                                                                                      ArrayTime().toFITS(midpoint - interval/2),
                                                                                                                                      processorType,
                                                                                                                                      correlationMode,
                                                                                                                                      spectralResolution)
                           
            #traverseALMARadiometricData(numAntenna, radiometricDataDesc, flagsDesc, basebands) 

#
# the main entry point
#
if __name__ == "__main__":
    main()
