# Bojan Nikolic <b.nikolic@mrao.cam.ac.uk>
"""
Casa utilities
"""
import os
from extprog import wvrgcal


def comparePh(msin,
              **kwargs):
    """
    Compare the corrected and uncorrected phasees
    """
    if "antenna" not in kwargs:
        kwargs["antenna"]="*&*"
    plotxy(msin,
           yaxis="phase", 
           datacolumn="data",
           **kwargs)
    plotxy(msin,
           yaxis="phase", 
           datacolumn="corrected",
           overplot=True,
           **kwargs)

def comparePhBl(msin,
                amax=3,
                **kwargs):
    """
    Compare on each baseline and save to a file
    """
    pref,junk=os.path.splitext(os.path.basename(msin))
    for i in range(amax):
        for j in range(i+1,amax):
            comparePh(msin,
                      antenna="%i&%i" % (i,j),
                      figfile=pref+("-%i-%i.png" %(i,j)),
                      **kwargs)

def wvrSummary(msin):
    """
    Create a summary plot of WVR teting data
    """
    pref,junk=os.path.splitext(os.path.basename(msin))
    gcal=pref+".gcal"
    clearstat()
    wvrgcal(msin, 
            gcal, 
            toffset=-1.0,
            segsource=True)
    applycal(vis=msin, 
             gaintable=gcal, 
             spwmap=[0,0,0,0])
    comparePhBl(msin,
                spw="1")

def compareFields(msin,
                  fl,
                  **kwargs):
    for i,f in enumerate(fl):
        plotxy(msin,
               field=f,
               overplot=((i and True) or False),
               **kwargs)

# Parameters that are used to select visibilities
_selectVis=["spw", 
            "field",
            "time"]
# Parameters to the calibration solving routines
_solveCal=["type", "t", "combine", "minblperant", "solnorm", "minsnr"]

def parsSel(kwargs,
            pars):
    """
    Return only parameters in the given list
    
    """
    res={}
    for k in kwargs.keys():
        if k in pars:
            res[k]=kwargs[k]
    return res
    
def bandPass(msin,
             tabout,
             **kwargs):
    """
    Compute the bandpass solution working around shortcomings the
    "task" interface. 

    Example call:
    >>> bandPass("/home/bnikolic/data/almagen/20100312/uid___X02_X4a3be_X1.ms", 
                 "test3", 
                 type='B',
                 t="inf",
                 combine="scan", minblperant=2,
                 solnorm=False, 
                 minsnr=-2,
                 spw="0:10~240",
                 field="3c273")

    """
    cb.open(msin)
    cb.selectvis(**parsSel(kwargs,
                           _selectVis))
    
    cb.setsolve(table=tabout,
                **parsSel(kwargs,
                          _solveCal))
    cb.solve()
    cb.close()

def bandPassTimeSeg(msin,
                    tabout,
                    ival=10,
                    **kwargs):
    """
    Compute bandpass for a segment of time (i.e., sub-scan duration)
    of a MS.

    :param ival: Use the first ival seconds of the scan to compute the
    bandpass phase
    """
    tb.open(msin)
    tbeg=tb.getcol("TIME")[0]
    tend=tbeg+ival
    tb.close()
    tsel=""
    tu= qa.quantity()
    tu["unit"]="s"
    tu["value"]=tbeg
    tsel+=qa.time(tu,form=["ymd"])
    tsel+="~"
    tu["value"]=tend
    tsel+=qa.time(tu,form=["ymd"])
    bandPass(msin,
             tabout,
             time=tsel,
             **kwargs)
             
             
    
