# Script to image and assess the properties of long baseline calibrators.
# Runs in CASA 4.5.0
# Expects to find the *.split.cal measurement set and the .fluxscale file.
# Unless you want to limit spws (i.e. exclude very narrow ones for
# speed) nothing should need to be changed.

# If the analysis fails (usually only on check source) it's an
# indication that the source is non-point like. The image and png
# should be created regardless.
#
# C. Brogan (Nov 2015)
# T. Hunter (Jan 2016)

###################################
# DATA PROPERTIES
###################################
import numpy as np
import analysisUtils as au
import glob 
from taskinit import *
from imfit_cli import imfit_cli as imfit
try:
    from tclean_cli import tclean_cli as tclean
except:
    print "checksource.py: Cannot import tclean"

def writeOut(f,line):
    print line
    f.write(line+'\n')

def version(short=False):
    """
    Returns the CVS revision number as a string.
    """
    myversion = "$Id: checksource.py,v 1.13 2016/12/15 16:03:04 thunter Exp $"
    if (short):
        myversion = myversion.split()[2]
    return myversion

def checksource(overwrite=True, verbose=False, subdir='', splitcal_vis=''):
    """
    Images the phasecal and check source in a dataset and reports statistics.
    Expects to find the *.split.cal measurement set and the corresponding
    .fluxscale file for it.
    Inputs:
    overwrite: if True, overwrite any existing image from a previous execution
    splitcal_vis: defaults to *.cal, but can be specified as list of strings, 
                  or a comma-delimited string
    Outputs:
    png image plots of each calibrator, and an ASCII file for each dataset
    The name of the ASCII file, and a list of pngs are returned.
    """
    # Read the dataset(s) and get properties
    if (splitcal_vis == ''):
        vislist = glob.glob('*.cal')
    else:
        if (type(splitcal_vis) == str):
            vislist = splitcal_vis.split(',')
        else:
            vislist = splitcal_vis
    print "Checking datasets: ", vislist
    mymsmd = au.createCasaTool(msmdtool)
    if (len(subdir) > 0):
        if (os.path.exists(subdir)):
            if (subdir[-1] != '/'): 
                subdir += '/'
    pnglist = []
    textfiles = []
    for vis in vislist:
        mymsmd.open(vis)
        freq=mymsmd.meanfreq(0,unit='GHz')
        # Check Source
        check=mymsmd.fieldsforintent('OBSERVE_CHECK_SOURCE*',True)[0]
        checkid=mymsmd.fieldsforintent('OBSERVE_CHECK_SOURCE*',False)[0]
        checkpos=mymsmd.phasecenter(checkid)
        # Phase calibrator
        phase=mymsmd.fieldsforintent('CALIBRATE_PHASE*',True)[0]
        phaseid=mymsmd.fieldsforintent('CALIBRATE_PHASE*',False)[0]
        phasepos=mymsmd.phasecenter(phaseid)
        if ('OBSERVE_TARGET#ON_SOURCE' in mymsmd.intents()):
            nScienceFields= len(mymsmd.fieldsforintent('OBSERVE_TARGET*',False))
            science = mymsmd.fieldsforintent('OBSERVE_TARGET*',True)[0]
            scienceid = mymsmd.fieldsforintent('OBSERVE_TARGET*',False)[0]
        else:
            nScienceFields = 0
        mymsmd.done()

        floatcell = au.pickCellSize(vis, maxBaselinePercentile=99, 
                                    verbose=verbose)
        cell = au.pickCellSize(vis, maxBaselinePercentile=99, cellstring=True, 
                               verbose=verbose)
#        imsize = int(au.nextValidImsize(int(5.0/floatcell))) # valid when we only had checksources for synthBeam < 0.25
        imsize = int(au.nextValidImsize(int(np.max([5.0,5.0*au.estimateSynthesizedBeam(vis)])/floatcell))) 
        print "imsize = ", imsize
        region='circle[[%dpix , %dpix], 15pix ]' % (imsize/2,imsize/2)

        if False:
            # original method (for bands 3-6 only)
            cell = str(np.round(0.015*(100/freq),3))+'arcsec'
            if freq < 116.0:
                imsize = [320,320]
                region='circle[[160pix , 160pix] ,15pix ]'
            else:
                imsize = [680,680]
                region='circle[[340pix , 340pix] ,15pix ]'

        ###################################
        # IMAGE 
        ###################################
        weighting = 'briggs'
        robust = 0.5
        niter = 50
        threshold = '0.0mJy'
        spw=''
        separation = au.angularSeparationOfTwoFields(vis,checkid,phaseid)
        if (nScienceFields > 0):
            separation_pcal_science = au.angularSeparationOfTwoFields(vis,scienceid,phaseid)
            separation_check_science = au.angularSeparationOfTwoFields(vis,scienceid,checkid)

        fieldtype = ['checksource','phasecal']
        field = [check,phase]
        for i,cal in enumerate(field):
            if (not os.path.exists(cal+'_'+vis+'.image') or overwrite):
                os.system('rm -rf '+cal+'_'+vis+'.*')
                if verbose:
                    print "Running tclean('%s', field='%s', cell=%s, imsize=%s, ...)" % (vis, cal, str(cell), str(imsize))
                tclean(vis=vis,
                       imagename=cal+'_'+vis,
                       field=cal,spw=spw,
                       specmode='mfs',
                       deconvolver='hogbom',
                       imsize = imsize, 
                       cell= cell, 
                       weighting = weighting, 
                       robust = robust,
                       niter = niter, 
                       threshold = threshold, 
                       interactive = False,
                       mask = region,
                       gridder = 'standard')
            png = subdir+fieldtype[i]+'_'+cal+'_'+vis+'.image.png'
            pnglist.append(png)
            au.imviewField(cal+'_'+vis+'.image',radius=30*floatcell,
                           contourImage=cal+'_'+vis+'.mask',levels=[1],
                           plotfile=png)


        ###################################
        # ANALYZE
        ###################################
        ###########
        # PHASE
        ###########
        imagename=phase+'_'+vis
        if verbose:
            print "Running imfit('%s', region='%s')" % (imagename+'.image', region)
        # Fit the phase source to get position and flux
        imagefit=imfit(imagename=imagename+'.image',
                       region=region)      
        fitresults=au.imfitparse(imagefit)

        # Compare the Positions
        phasepos_obs=au.direction2radec(phasepos)
        if fitresults != None:
            phasepos_fit=','.join(fitresults.split()[:2])
            phasepos_diff=au.angularSeparationOfStrings(phasepos_obs,phasepos_fit,verbose=False)*3600.

        # Compare the Flux densities
        selffluxfile=glob.glob('*.fluxscale')[0]
        selfflux=au.fluxscaleParseLog(selffluxfile,field=phase)[0][0]

        peakIntensity = au.imagePeak(imagename+'.image')
        if fitresults != None:
            phaseflux_fit=float(fitresults.split()[2])
            phaseCoherence = 100*peakIntensity/phaseflux_fit
            phaseflux_diff=100*(selfflux-phaseflux_fit)/selfflux

        # Print the final results and save to file
        textfile = subdir+'calimage_results_'+vis+'.txt'
        textfiles.append(textfile)
        f = open(textfile,'w')
        f.write('\n*************************************************************\n\n')
        line = 'CHECK_SOURCE IMAGE ANALYSIS REPORT (version %s)\n' % version(short=True)
        writeOut(f,line)
        info = au.getFitsBeam(imagename+'.image')
        synthBeam = (info[0]*info[1])**0.5
        if fitresults == None:
            line = "Phasecal %s: imfit failed" % (phase)
        else:
            line= "Phasecal %s: Position difference = %s arcsec = %s synth.beam, Flux %% difference = %s"%(phase,au.roundFiguresToString(phasepos_diff,3), au.roundFiguresToString(phasepos_diff/synthBeam,3), au.roundFiguresToString(phaseflux_diff,3))

            writeOut(f,line)
            line = "    coherence = peakIntensity/fittedFluxDensity = %s%%" % (au.roundFiguresToString(phaseCoherence,3))
        writeOut(f,line)
        f.close()

        ###########
        # CHECK
        ###########
        imagename=check+'_'+vis
        # Fit the check source to get position and flux
        if verbose:
            print "Running imfit('%s', region='%s')" % (imagename+'.image', region)
        imagefit=imfit(imagename=imagename+'.image',
                       region=region)      
        fitresults=au.imfitparse(imagefit, deconvolved=True)
        info = au.getFitsBeam(imagename+'.image')
        synthMajor, synthMinor = info[0:2]
        synthBeam = (info[0]*info[1])**0.5

        # Compare the Positions
        checkpos_obs=au.direction2radec(checkpos)
        if fitresults != None:
            checkpos_fit=','.join(fitresults.split()[:2])
            checkpos_diff=au.angularSeparationOfStrings(checkpos_obs,checkpos_fit,
                                                        verbose=False)*3600.

        # Compare the Flux densities
        selffluxfile=glob.glob('*.fluxscale')[0]
        results = au.fluxscaleParseLog(selffluxfile,field=check)
        peakIntensity = au.imagePeak(imagename+'.image')
        if (results != None and fitresults != None):
            selfflux=results[0][0] 
            checkflux_fit=float(fitresults.split()[2])

            checkflux_diff=100*(selfflux-checkflux_fit)/selfflux
            checkCoherence = 100*peakIntensity/checkflux_fit
        if fitresults != None:
            if verbose: 
                print "Checksource fitresults: ", fitresults
            deconvolvedMajor = float(fitresults.split()[5])
            deconvolvedMinor = float(fitresults.split()[7])

        # Print the final results and save to file
        f=open(textfile,'a')
        if fitresults == None:
            line = "Checksource %s: imfit failed" % (phase)
        else:
            if (results != None):
                line= "\nChecksource %s: Position difference = %s arcsec = %s synth.beam, Flux %% difference = %s"%(check ,au.roundFiguresToString(checkpos_diff,3),au.roundFiguresToString(checkpos_diff/synthBeam,3),au.roundFiguresToString(checkflux_diff,3))
                writeOut(f,line)
                line = "    coherence = peakIntensity/fittedFluxDensity = %s%%" % (au.roundFiguresToString(checkCoherence,3))
            else:
                line= "\nChecksource %s: Position difference = %s arcsec = %s synth.beam" % (check ,au.roundFiguresToString(checkpos_diff,3),au.roundFiguresToString(checkpos_diff/synthBeam,3))
            writeOut(f,line)
            line = "    beam size = %s x %s arcsec" % (au.roundFiguresToString(synthMajor,3), au.roundFiguresToString(synthMinor,3))
            writeOut(f,line)
            line = "    apparent deconvolved size = %s x %s arcsec = %s synth.beam area" % (au.roundFiguresToString(deconvolvedMajor,2), au.roundFiguresToString(deconvolvedMinor,2), au.roundFiguresToString(deconvolvedMajor*deconvolvedMinor/(synthBeam**2),2))
        writeOut(f,line)
        line = "    angular separation of phasecal to checksource = %s degree" % (au.roundFiguresToString(separation,3))
        writeOut(f,line)
        if (nScienceFields > 0):
            if (nScienceFields > 1):
                modifier = 'first'
            else:
                modifier = 'only'
            line = "    angular separation of phasecal to %s science field (%d) = %s degree" % (modifier,scienceid,au.roundFiguresToString(separation_pcal_science,3))
            writeOut(f,line)
            line = "    angular separation of checksource to %s science field (%d) = %s degree" % (modifier,scienceid,au.roundFiguresToString(separation_check_science,3))
            writeOut(f,line)
        f.close()
    # end 'for' loop over vislist
    return textfiles, pnglist


