#! /usr/bin/env python

import taskinit

def readCaiMap(fname):
    """
    Read CAI map from CCC container log file.
    Both a full file and just the lines for the CAI map should work.
    Output is a list of length 64 with either antenna name strings of Nonje as elements.
    """
    f = open(fname)
    caiMap = []
    numCai = 0
    for i in range(64):
        caiMap.append(None)
    for line in f:
        if len(line) < 8: continue
        firstTwo = line[0:2]
        if firstTwo not in ["CM", "DA", "DV", "PM"]: continue
        ant = line[0:4]
        colon = line[4:7]
        if colon != " : ": continue
        caiStr = line[7:-1].strip()
        cai = int(caiStr)
        if cai < 0 or cai > 63: continue
        caiMap[cai] = ant
        numCai += 1
        if numCai >= 64: break
    f.close()
    return caiMap

def getMSAntennas(vis):
    mymsmd=taskinit.msmdtool()
    mymsmd.open(vis)
    ants = mymsmd.antennanames()
    mymsmd.close()
    return ants

def lowCaiAntennaSelection(vis, caiMapFile, sortByName=True):
    """
    Returns a string list of baselines composed of antennas that are all < CAI-32
    e.g. 'DV01&DV02;DV05&DV08'  
    """
    if caiMapFile is None:
        caiMapFile="acsStartContainer_cppContainer_2016-11-24_20.20.30.358"
    caiMap = readCaiMap(caiMapFile)
    antsInMS = getMSAntennas(vis)
    print "CAI map: 0-31: ", caiMap[:32]
    print "CAI map: 32-63: ", caiMap[32:]
    print "antennas in ms: ", antsInMS
    antSel = ""
    for i in range(32):
        ant0 = caiMap[i]
        if ant0 is None: continue
        if ant0 not in antsInMS: continue
        for j in range(1+i,32):
            ant1 = caiMap[j]
            if ant1 is None: continue
            if ant1 not in antsInMS: continue
            if len(antSel) != 0:
                antSel = antSel + ";"
            if sortByName:
                ant0_temp, ant1_temp = sorted([ant0,ant1])
            antSel = antSel + ant0_temp + "&" + ant1_temp
    
    print antSel
    return antSel

def crossCaiAntennaSelection(vis, caiMapFile, sortByName=True):
    """
    Returns a string of baselines that cross the 31-32 CAI barrier
    e.g. 'DV01&DV02;DV05&DV08'  
    """
    if caiMapFile is None:
        caiMapFile="acsStartContainer_cppContainer_2016-11-24_20.20.30.358"
    caiMap = readCaiMap(caiMapFile)
    antsInMS = getMSAntennas(vis)
    print "CAI map: 0-31: ", caiMap[:32]
    print "CAI map: 32-63: ", caiMap[32:]
    print "antennas in ms: ", antsInMS
    antSel = ""
    for i in range(32):
        ant0 = caiMap[i]
        if ant0 is None: continue
        if ant0 not in antsInMS: continue
        for j in range(32,64):
            ant1 = caiMap[j]
            if ant1 is None: continue
            if ant1 not in antsInMS: continue
            if len(antSel) != 0:
                antSel = antSel + ";"
            if sortByName:
                ant0_temp, ant1_temp = sorted([ant0,ant1])
            antSel = antSel + ant0_temp + "&" + ant1_temp
    
    print antSel
    return antSel
    
    
    
