"""
delayJumpDetector.py: detect delay jumps by utilizing bandpass solution.

usage:
import delayJumpDetector as djd
djd.detectDelayJumps('uid___XXX_XXX_XXX.ms')
"""
import os 
import sys
import pylab as pl
import casadef
from casa import *
from taskinit import *
import analysisUtils as aU

def detectOutliers(values, threshold=10., edgetrim=5):
    """
    Detect outlying samples from an array of values.

    threshold: if a value is deviated from the mean by more than
               threshold*sigma, it is identified as an outlier,
               where mean and sigma are not "raw" ones but
               "robust" ones (see below)
    edgetrim: minimum and maximum 1/edgetrim samples are ignored
              to determine robust (trimmed) mean and sigma
    """
    n_samples = len(values)
    subsample = pl.sort(values)[n_samples/edgetrim:-n_samples/edgetrim]
    std = subsample.std()
    mean = subsample.mean()
    return (pl.fabs(values-mean) > threshold*std)

def readCoeffs(bpcaltable, spw='1'):
    """
    Read polynomial coeffs and timestamps of bandpass table.
    """
    # obtain CAL_DESC_IDs corresponding to spws.
    # should be [0, 1, ...] ...
    spws = [int(item) for item in spw.split(',')]
    tb.open('%s/CAL_DESC' % bpcaltable)
    spwin = tb.getcol('SPECTRAL_WINDOW_ID').squeeze()
    tb.close()
    caldescs = [pl.arange(len(spwin))[spwin==s][0] for s in spws]

    coeffs = {}
    tb.open(bpcaltable)
    nants = len(pl.unique(tb.getcol('ANTENNA1')))
    for i in range(nants):
        coeffs[i] = {}
        for j in range(len(spws)):
            q = tb.query('ANTENNA1==%d && CAL_DESC_ID==%d' % (i, caldescs[j]))
            coeffs[i][spws[j]] = {'timestamp': q.getcol('TIME').squeeze(),
                'acoeff': q.getcol('POLY_COEFF_AMP').squeeze(),
                'pcoeff': q.getcol('POLY_COEFF_PHASE').squeeze()}
    tb.close()
    return coeffs


def detectDelayJumps(msname='', spw='', scan='', solint='int', refant=0,
                     intent='CALIBRATE_PHASE#ON_SOURCE'):
    """
    Detect delay jumps using bandpass(bandtype=BPOLY).

    Outliers in the amplitude and phase slopes of the bandpass solution
    are identified as delay jumps.
    spw: spws to be analyzed (can be a comma-separated list)
    scan: scans to be analyzed (can be a comma-separated list)
    solint: time interval for solving bandpass
    refant: reference antenna ID (integer) for bandpass solution

    If spw and/or scan are left blank and CASA version is >= 4.1.0,
    they were set according to the given intent parameter.
    """

    if not os.path.exists(msname):
        print 'ERROR: msname "%s" is not found.' % msname
        return

    antnames = aU.getAntennaNames(msname)
    nants = len(antnames)

    if spw == '':
        if casadef.casa_version >= '4.1.0':
            #mymsmd = createCasaTool(msmdtool)
            msmd.open(msname)
            phase_spw = msmd.spwsforintent(intent)
            wvrspws = msmd.wvrspws()
            spw = []
            for s in phase_spw:
                if len(msmd.chanfreqs(s)) > 4 and s not in wvrspws:
                    spw.append(s)
            if len(spw) < 1:
                print 'Found no non-wvr, multi-channel spws for intent=%s' \
                    % intent
                return
            print 'Found %d non-wvr, multi-channel spws for intent=%s: %s' \
                % (len(spw), intent, str(spw))
            spw = ','.join([str(i) for i in spw])
            msmd.close()
        else:
            print 'You must specify 1 or more spws via the spw parameter.'
            return

    if scan == '' and casadef.casa_version >= '4.1.0':
        msmd.open(msname)
        scans = msmd.scansforintent(intent)
        if len(scans) > 1:
            scan = ','.join([str(i) for i in scans])
        msmd.close()

    bpcaltable = '%s.bpcal' % msname
    os.system('rm -rf %s' % bpcaltable)
    bandpass(vis = msname,
        caltable = bpcaltable,
        spw = spw,
        scan = scan,
        solint = solint,
        refant = '%s' % refant,
        bandtype = 'BPOLY',
        degamp = 1,
        degphase = 1,
        visnorm = True,
        maskedge = 10)

    coeffs = readCoeffs(bpcaltable, spw)
    spws = [int(item) for item in spw.split(',')]
    timestamp = coeffs[0][spws[0]]['timestamp']
    outliers = []
    # hereafter 2-pol spws are assumed
    for i in range(nants):
        for j in range(len(spws)):
            outX = (detectOutliers(coeffs[i][spws[j]]['acoeff'][1, :]) | \
                    detectOutliers(coeffs[i][spws[j]]['pcoeff'][1, :]))
            outY = (detectOutliers(coeffs[i][spws[j]]['acoeff'][3, :]) | \
                    detectOutliers(coeffs[i][spws[j]]['pcoeff'][3, :]))
            outliers.append(outX)
            outliers.append(outY)
    outliers = pl.array(outliers)
    olexist = outliers.sum(0, dtype=bool)
    antspwpol = pl.array([('%sspw%dpol%d' % (a, s, p)) \
        for a in antnames for s in spws for p in [0, 1]])
    if olexist.sum(dtype=bool):
        ts = timestamp[olexist]
        ol = outliers[:, olexist]
        for i in range(len(ts)):
            #print mjd2iso8601(ts[i]), ', '.join(antspwpol[ol[:, i]])
            print qa.time('%fs' % ts[i], form='fits')[0], \
                ', '.join(antspwpol[ol[:, i]])
    else:
        print 'no delay jumps were detected.'
