#!/usr/bin/env python
"""
Show azimuth and/or elevation of astronomical object vs time.

Usage (see help for details):
  elplot.py [-d 2011-01-19] [-x LST] [-y AZ,EL] [-s '3c273,1924-292,Saturn']
    [-S '5:35/-5:27/Ori,17:46/-29.0/GC'] [-Q] [-G] [-c False] [-t Title]

2016-08-02 tsawada: more cosmetics
2016-07-31 tsawada: cosmetic changes on AZ and PA plots
                    adjusted label positions in AZ and PA plots
                    added the capability to plot parallactic angle
                    added '-t' option
                    removed warning for CLT (thanks to Chilean Government)
2015-08-18 tsawada: added a warning for using CLT
2015-08-02 tsawada: added '-G' option
2015-03-08 tsawada: vertical line to show the current time become available for
                    all x axis options.  also added '-c' option to disable it
2015-03-08 tsawada: added timezone to x-axis label
2013-11-11 tsawada: cleaned up the code to calculate planet RA/Dec
2013-09-24 tsawada: added '-l' option
2012-12-13 tsawada: added the capability to specify the names of "-S" sources
                    and made some cosmetic changes
2012-09-05 tsawada: removed dummy raw_input at the end of the script
2012-04-09 tsawada: added pluto, ceres, pallas, juno, and vesta data
2012-02-27 tsawada: added CLT and CLST support
2011-10-06 tsawada: added '-Q' option
2011-01-25 tsawada: disabled masking EL<0 part of AZ plot because
                    it does not work well with the currently-installed-in-STE
                    version of matplotlib
2011-01-19 tsawada: added '-S' option
2010-12-31 tsawada: minor cosmetic changes
2010-12-30 tsawada: added '-y' option
2010-??-?? tsawada

$Id: elplot.py,v 1.27 2016/08/03 14:43:56 tsawada Exp $
"""

import sys
import getpass
username = getpass.getuser()
sys.path.append('/users/%s/AIV/science/scripts/R10.4_WORKING/'%username)
import time
import datetime
import pylab as pl
import sourceCatalog as sC
from optparse import OptionParser

### define constants ###

# longitude (E) and latitude (N) of the sites in degrees
locations = {
    'Chajnantor': (-67.75, -23.0),
    'Nobeyama': (138.48, 35.93),
}
defaultlocation = 'Chajnantor'

# timezones (other than UTC and LST)
timezones = {
    'CLT' : -4,
    'CLST': -3,
}

### define methods ###

def convert_colonsep_to_float(value):
    """
    Convert a colon-delimited string (like dd:mm:ss.s) into a float number.
    """
    value = value.strip()
    if value[0] == '-':
        sign = -1.
        value = value[1:]
    else:
        sign = 1.

    values = [float(item) for item in value.split(':')]
    result = 0.
    for (i, num) in enumerate(values):
        result += num * 60.**(-i)

    return sign*result


def normalize_angle(value):
    """
    Normalize angle (radians) into 0 - 2pi.
    """
    return value % (2.*pl.pi)


def convert_date_to_t2000(thedate):
    """
    Return time passed since 2000-01-01 12:00 UT (in unit of 36525 days).
    """
    return ((thedate-datetime.date(2000, 1, 1)).days-0.5)/36525.


def get_planet_orbits():
    """
    Return orbital elements (and diameters) of planets.

    The listed parameters are:
    - Semimajor axis [AU]
    - Eccentricity
    - Inclination [degrees]
    - Longitude of the ascending node at 2000-01-01T12:00 [degrees]
    - Argument of periapsis at 2000-01-01T12:00 [degrees]
    - Mean anomaly at 2000-01-01T12:00 [degrees]
    - Variation of longitude of the ascending node [degrees/36525days]
    - Variation of argument of periapsis [degrees/36525days]
    - Variation of mean anomaly [degrees/36525days]
    - Diameter [km]
    """
    return {
        "mercury": (0.3871, 0.2056, 7.00, 48.33, 29.13, 174.79,
            1.19, 0.37, 149472.52, 4878.),
        "venus": (0.7233, 0.0068, 3.39, 76.68, 54.88, 50.42,
            0.90, 0.50, 58517.81, 12104.),
        "earth": (1.0000, 0.0167, 0.00, 0.00, 102.94, 357.53,
            0.00, 1.72, 35999.05, 12756.),
        "mars": (1.5237, 0.0934, 1.85, 49.56, 286.50, 19.37,
            0.77, 1.07, 19139.86, 6787.),
        "jupiter": (5.2026, 0.0485, 1.30, 100.46, 273.87, 20.02,
            1.02, 0.59, 3034.69, 142800),
        "saturn": (9.5549, 0.0555, 2.49, 113.67, 339.39, 317.02,
            0.88, 1.09, 1221.55, 120660.),
        "uranus": (19.218, 0.0463, 0.77, 74.01, 99.00, 141.05,
            0.52, 0.97, 428.38, 51118),
        "neptune": (30.110, 0.0090, 1.77, 131.78, 276.34, 256.22,
            1.10, 0.32, 218.46, 49528),
        "pluto": (39.540, 0.2490, 17.15, 110.32, 113.82, 14.33,
            1.35, 0.04, 145.29, 2390),
        "ceres": (2.767, 0.078, 10.6, 80.54, 72.50, 7.41,
            -1.48, 0.00, 7817.98, 941),
        "pallas": (2.771, 0.231, 34.8, 173.21, 310.00, 353.17,
            -1.02, 0.00, 7796.46, 544),
        "juno": (2.669, 0.257, 13.0, 170.23, 248.00, 240.74,
            -3.06, 0.000, 8249.91, 233),
        "vesta": (2.362, 0.090, 7.1, 104.00, 149.70, 340.82,
            0.00, 0.000, 9919.81, 529)
    }


def get_planet_xyz(orbits, thedate):
    """
    Return ecliptic (X, Y, Z) of planets for a give date.
    """
    t2000 = convert_date_to_t2000(thedate)

    planetxyz = {}
    for theplanet in orbits.keys():
        ### a = semimajor axis; e = eccentricity; i = inclunation;
        ### Omega0 = long. of asc. node (Omega) at 2000.0;
        ### omega0 = arg. of periapsis (omega) at 2000.0;
        ### l0 = mean anomaly (l) at 2000.0;
        ### Omegadot = dOmega/dt; omegadot = domega/dt; ldot = dl/dt;
        ### d = diameter
        (a, e, i, Omega0, omega0, l0, Omegadot, omegadot, ldot, d) = \
            orbits[theplanet]
        (i, Omega, omega, l) = pl.radians((i, \
            Omega0+Omegadot*t2000, omega0+omegadot*t2000, l0+ldot*t2000))
        l = normalize_angle(l)

        ### solve Kepler's equation
        u = l
        du = 1e9
        while pl.fabs(du)>1e-7:
            du = (l-u+e*pl.sin(u))/(1.-e*pl.cos(u))
            u += du

        ### get (x, y) coordinates on the orbital plane
        (xx, yy) = (a*(pl.cos(u)-e), a*pl.sqrt(1.-e**2)*pl.sin(u))

        ### convert to ecliptic lon/lat (lm, bt) and then to ecliptic (x, y, z)
        r = pl.sqrt(xx**2+yy**2)
        f = pl.arctan2(yy, xx)
        a1 = pl.cos(omega+f)
        b1 = pl.cos(i)*pl.sin(omega+f)
        c1 = pl.sin(i)*pl.sin(omega+f)
        bt = pl.arctan2(c1, pl.sqrt(a1**2+b1**2))
        lm = pl.arctan2(b1, a1) + Omega
        planetxyz[theplanet] = pl.array([r*pl.cos(bt)*pl.cos(lm), \
            r*pl.cos(bt)*pl.sin(lm), r*pl.sin(bt)])

    return planetxyz


def get_planet_radec_from_xyz(planetxyz):
    """
    Return (RA, Dec) of planets.
    """
    i_ecl = pl.radians(23.44)
    planetradec = {}
    for theplanet in planetxyz.keys():
        if theplanet == "earth":
            continue
        (xx, yy, zz) = planetxyz[theplanet] - planetxyz["earth"]
        x1 = xx
        y1 = -zz*pl.sin(i_ecl) + yy*pl.cos(i_ecl)
        z1 = zz*pl.cos(i_ecl) + yy*pl.sin(i_ecl)
        planetradec[theplanet] = (normalize_angle(pl.arctan2(y1, x1)),
            pl.arctan2(z1, pl.sqrt(x1**2+y1**2)))

    return planetradec


def get_sun_radec(thedate):
    """
    Return (RA, Dec) of the Sun.
    """
    i_ecl = pl.radians(23.44)
    t2000 = convert_date_to_t2000(thedate)
    gsun = 357.53 + 36000.77*t2000
    lsun = pl.radians(gsun - 77.06 + 1.91 * pl.sin(pl.radians(gsun)))
    tmpval = pl.sqrt((pl.cos(lsun)*pl.cos(lsun) + \
        pl.cos(i_ecl)*pl.cos(i_ecl)*pl.sin(lsun)*pl.sin(lsun)))
    dec = pl.arctan2((pl.sin(i_ecl)*pl.sin(lsun)), tmpval)
    ra = normalize_angle( pl.arctan(pl.cos(i_ecl)*pl.tan(lsun)) +
        (pl.cos(lsun) < 0.)*pl.pi )

    return (ra, dec)


def get_planet_radec(thedate):
    """
    Return planets' (RA, Dec).
    """
    planetradec = get_planet_radec_from_xyz(\
        get_planet_xyz(get_planet_orbits(), thedate))
    planetradec['sun'] = get_sun_radec(thedate)

    return planetradec


def get_utc_to_lst_offset(thedate, sitelon):
    """
    Return the offset between UTC and LST for give date and site.
    """
    return sitelon + pl.radians(15.* (
        6.646065556 + 2400.051262/36525.*( \
        (thedate-datetime.date(1900, 1, 1)).days+0.5)))


def get_sC_source_radec(sourcedict):
    """
    Extract RA and Dec values from a sourceCatalog entry.
    """
    return pl.radians([15.*convert_colonsep_to_float(sourcedict['ra']),
        convert_colonsep_to_float(sourcedict['dec'].replace('.', ':', 2))])


def get_utc_offset_string(tz):
    """
    Generate a string to express timezone (UTC+HH:MM).
    """
    if tz < 0:
        sign = '-'
        tz = -tz
    else:
        sign = '+'
    hr = int(tz)
    mn = int(round(60.*(tz-hr)))
    return 'UTC%s%d:%02d' % (sign, hr, mn)

################################
# MAIN
################################

tzkeys = ['UTC', 'LST'] + timezones.keys()

### options ###

parser = OptionParser()
parser.add_option('-d', '--date', dest='date', metavar='Date',
    help='Date (YYYY-MM-DD).  default: today')
parser.add_option('-s', '--source', dest='sources', metavar='Source(s)',
    help='Source(s) to be plotted.  Solar system objects ' + \
        '(Sun, planets, Pluto, and the "big four" asteroids) and/or the ' + \
        'ones defined in sourceCatalog.  ' + \
        'Multiple items can be concatenated by commas.  ' + \
        'default: all planets and the sun (if -S/-Q are not given) ' + \
        'or none (otherwise)')
parser.add_option('-S', '--radecsource',
    dest='radecsources', metavar='radecSource(s)',
    help='Source(s) to be plotted, as a set of RA, Dec, and name ' + \
        '(optional) separated by slashes (like HH:MM/dd:mm or ' + \
        'HH:MM/dd:mm/name).  ' + \
        'Multiple items can be concatenated by commas.  ' + \
        'default: none')
parser.add_option('-Q', '--brightquasars',
    dest='brightquasars', metavar='brightQuasars',
    action='store_true', default=False,
    help='Plot "standard" set of bright quasars.  default: False')
parser.add_option('-G', '--gridsources',
    dest='gridsources', metavar='gridSources',
    action='store_true', default=False,
    help='Plot the ALMA Grid sources.  default: False')
parser.add_option('-x', '--xaxis', dest='xaxis', metavar='Xaxis',
    default='UTC', help='X axis (%s).  default: UTC' % \
    ', '.join(tzkeys[:-1]+['or '+tzkeys[-1]]))
parser.add_option('-y', '--yaxis', dest='yaxis', metavar='Yaxis',
    default='EL', help='Y axis (AZ, EL, PA, or ' + \
    'their comma-concatenation).  default: EL')
parser.add_option('-l', '--location', dest='location', metavar='Location',
    default=defaultlocation, help='Observatory name (%s).  default: %s' % \
    (', '.join(locations.keys()), defaultlocation))
parser.add_option('-c', '--static', dest='static', metavar='static',
    action='store_true', default=False,
    help='Static mode, i.e., do not draw vertical line for current time.  ' + \
    'default: False')
parser.add_option('-t', '--title', dest='title', metavar='Title',
    default='', help='Plot title.  default: "" (->date)')
(options, args) = parser.parse_args()

if options.date == None:
    thedate = datetime.date.today()
else:
    dlist = [int(item) for item in options.date.split('-')]
    thedate = datetime.date(dlist[0], dlist[1], dlist[2])

if not options.xaxis in tzkeys:
    print '"-x" option should be %s.  Force set to UTC.' % \
        ', '.join(tzkeys[:-1] + ['or '+tzkeys[-1]])
    options.xaxis = 'UTC'
yaxes = sorted(options.yaxis.split(','))
validYaxes = ['AZ', 'EL', 'PA']
if set(yaxes+validYaxes) != set(validYaxes):
    print '"-y" option should be AZ, EL, PA, or ' + \
        'their comma-concatenation.  Force set to EL.'
    options.xaxis = 'EL'
if not options.location in locations.keys():
    print 'location "%s" is not found in the list.  Force set to %s.' % \
        (options.location, defaultlocation)
    options.location = defaultlocation

if options.sources == None and options.radecsources == None and \
    options.brightquasars == False and options.gridsources == False:
    options.sources = 'Sun,Mercury,Venus,Mars,Jupiter,Saturn,Uranus,Neptune'

if options.sources != None:
    sources = [item.strip() for item in options.sources.split(',')]
if options.radecsources != None:
    radecsources = [item.strip() for item in options.radecsources.split(',')]

if options.xaxis in ['UTC', 'LST']:
    xaxislabel = options.xaxis
else:
    xaxislabel = '%s (%s)' % \
        (options.xaxis, get_utc_offset_string(timezones[options.xaxis]))

#if options.xaxis == 'CLT':
#        print '*** WARNING: CLT (UTC-4) has been decommissioned! ***'


### make source catalog ###

planetsource = get_planet_radec(thedate)
contsource = sC.sourceContinuum
linesource = sC.sourceSpectralLine

sourceradec = {}

if options.sources != None:
    for source in sources:
        lsource = source.lower()
        if lsource in contsource.keys():
            sourceradec[source] = get_sC_source_radec(contsource[lsource])
        elif lsource in linesource.keys():
            sourceradec[source] = get_sC_source_radec(linesource[lsource])
        elif lsource in planetsource.keys():
            sourceradec[source] = planetsource[lsource]
        else:
            print 'source "%s" is not found in the catalog.' % source

if options.radecsources != None:
    for radecsource in radecsources:
        (ra, dec, name) = (radecsource+'//').split('/')[:3]
        if name == '':
            name = radecsource
        sourceradec[name] = \
            (pl.radians(15.*convert_colonsep_to_float(ra)), \
             pl.radians(    convert_colonsep_to_float(dec)))

if options.brightquasars:
    for source in ['0006-063', '3c84', '0423-013', '0522-364', '0538-440',
                   '0730-116', '0854+201', '1058+015', '3c273', '3c279',
                   '1337-129', '1751+096', '1924-292', '3c454.3']:
        lsource = source.lower()
        if lsource in contsource.keys():
            sourceradec[source] = get_sC_source_radec(contsource[lsource])
        elif lsource in linesource.keys():
            sourceradec[source] = get_sC_source_radec(linesource[lsource])
        elif lsource in planetsource.keys():
            sourceradec[source] = planetsource[lsource]
        else:
            print 'source "%s" is not found in the catalog.' % source

if options.gridsources:
    for source in ['2355-534', '0237+288', '0238+166', '3c84', '0334-401',
                   '0423-013', '0510+180', '0519-454', '0522-364',
                   '0538-440', '0637-752', '0750+125', '0854+201',
                   '1037-295', '1058+015', '1107-448', '1146+399', '3c273',
                   '3c279', '1337-129', '1427-421', '1517-243', '1550+054',
                   '3c345', '1613-586', '1733-130', '1751+096', '1924-292',
                   '2025+337', '2056-472', '2148+069', '3c454.3', '2232+117',
                   '2258-279']:
        lsource = source.lower()
        if lsource in contsource.keys():
            sourceradec[source] = get_sC_source_radec(contsource[lsource])
        elif lsource in linesource.keys():
            sourceradec[source] = get_sC_source_radec(linesource[lsource])
        elif lsource in planetsource.keys():
            sourceradec[source] = planetsource[lsource]
        else:
            print 'source "%s" is not found in the catalog.' % source

### plot ###
def plot() :
    (sitelon, sitelat) = pl.radians(locations[options.location])

    xtime = pl.radians(pl.arange(721)/2.)
    if options.xaxis == 'LST':
        lst = xtime
    elif options.xaxis == 'UTC':
        lst = xtime + get_utc_to_lst_offset(thedate, sitelon)
    else:
        lst = xtime + get_utc_to_lst_offset(thedate, sitelon) - \
              pl.radians(timezones[options.xaxis]*15)

    nplots = len(yaxes)
    plots = dict((k, i) for (i, k) in enumerate(yaxes))

    
    subpls = []
    for i in range(nplots):
        subpls.append(plf.add_subplot(nplots, 1, i+1))

    for source in sourceradec.keys():
        (ra, dec) = sourceradec[source]
        azelpa = {
            'AZ': normalize_angle(
            pl.arctan2(-pl.cos(dec)*pl.sin(lst-ra) , \
                       (pl.cos(sitelat)*pl.sin(dec)-pl.sin(sitelat) * \
                        pl.cos(dec)*pl.cos(lst-ra)))),
            'EL': pl.arcsin(pl.cos(pl.pi/2.-sitelat)*pl.sin(dec) + \
                            pl.sin(pl.pi/2.-sitelat)*pl.cos(lst-ra)*pl.cos(dec)),
            'PA': pl.arctan2(pl.cos(sitelat)*pl.sin(lst-ra),
                    pl.sin(sitelat)*pl.cos(dec)-pl.cos(sitelat)*pl.sin(dec)*pl.cos(lst-ra))
            }
        try:
            imaxel = pl.where(azelpa['EL']==azelpa['EL'].max())[0][0]
        except:
            imaxel = 0
        try:
            irise = pl.where(pl.diff(pl.sign(azelpa['EL']-azelpa['EL'].max()/1.5))!=0)[0][0] + 1
        except:
            irise = 0
        for key in plots.keys():
            xvalue = xtime.copy()
            yvalue = azelpa[key].copy()
            if key == 'EL':
                plotsym = '-'
                labelx = pl.degrees(xtime[imaxel])/15.
                labely = pl.degrees(yvalue[imaxel])
            else:
                labelx = pl.degrees(xtime[irise])/15.
                labely = pl.degrees(yvalue[irise])
                yvalue[azelpa['EL']<0.] = pl.nan
                yjump = pl.where((pl.fabs(pl.diff(yvalue))>pl.pi/2))[0]
                for ijump in sorted(yjump, reverse=True):
                    xvalue = pl.insert(xvalue, [ijump+1, ijump+1, ijump+1],
                        [xvalue[ijump+1], pl.nan, xvalue[ijump]])
                    if yvalue[ijump] < yvalue[ijump+1]:
                        yvalue = pl.insert(yvalue, [ijump+1, ijump+1, ijump+1],
                            [yvalue[ijump+1]-2.*pl.pi, pl.nan, yvalue[ijump]+2.*pl.pi])
                    else:
                        yvalue = pl.insert(yvalue, [ijump+1, ijump+1, ijump+1],
                            [yvalue[ijump+1]+2.*pl.pi, pl.nan, yvalue[ijump]-2.*pl.pi])
            subpls[plots[key]].plot(pl.degrees(xvalue)/15., pl.degrees(yvalue), '-')
            subpls[plots[key]].annotate(source,
                xy=(labelx, labely), xycoords='data', size='small')

    for key in plots.keys():
        subpls[plots[key]].set_xlim(0, 24)
        subpls[plots[key]].set_ylim({'AZ': 0, 'EL': 0, 'PA': -180}[key],
            {'AZ': 360, 'EL': 90, 'PA': 180}[key])
        subpls[plots[key]].grid(True, alpha=0.5)
        if plots[key] == 0:
            if options.title == '':
                subpls[plots[key]].set_title(thedate.isoformat())
            else:
                subpls[plots[key]].set_title(options.title)
        if plots[key]+1 >= nplots:
            subpls[plots[key]].set_xlabel('%s [h]' % xaxislabel)
        subpls[plots[key]].set_ylabel('%s [deg]' % \
            {'AZ': 'Azimuth', 'EL': 'Elevation', 'PA': 'Parallactic Angle'}[key])
        subpls[plots[key]].set_xticks(pl.arange(25))
        subpls[plots[key]].set_yticks(
            pl.arange({'AZ': 0, 'EL': 0, 'PA': -6}[key],
                      {'AZ': 13, 'EL': 10, 'PA': 7}[key]) * \
                {'AZ': 30, 'EL': 10, 'PA': 30}[key])

    if not options.static:
        now = datetime.datetime.utcnow()
        hrs = now.hour + now.minute/60. + now.second/3600.
        if options.xaxis == 'UTC':
            pass
        elif options.xaxis == 'LST':
            hrs += pl.degrees(get_utc_to_lst_offset(now.date(), sitelon))/15.
        else:
            hrs += timezones[options.xaxis]
        hrs %= 24.
        if (options.xaxis == 'LST') or (abs((thedate-now.date()).days) <= 1):
            for key in plots.keys():
                subpls[plots[key]].plot([hrs, hrs], [-360, 360], 'r--')

pl.ion()
pl.show()
plf = pl.figure()
while True :
    pl.clf()
    plot()
    pl.pause(300)
