"""
A script to perform Gaussian fit of total-power map(s).

Launch CASA and type the following:
  imfiles     = 'uid___Xxx_Xxxxx_X1.image' # Image file(s) to be processed
                                           #   no defaults
                                           #   use a list for multiple items
  showfigures = 2                          # 1 to show images; 2 to save PNG
                                           #   default 0 (don't show nor save)
  fitlogfile  = 'gaussfitlog.txt'          # Name of log file
                                           #   default 'gaussfitmap_log.txt'
  execfile('gaussfitotfmap.py')

In order to process multiple images, 'glob' package is helpful to obtain
a list of image files:
  import glob
  imfiles = glob.glob('*.image')
  imfiles.sort()                          # optional: to sort the log
  execfile('gaussfitotfmap.py')

The script calls the imfit task.  The box (mask) is set to
(peakX, peakY) +/- (2*HWHM, 2*HWHM), where HWHM is a roughly-estimated
half-width-at-half-maximum value.

2010-01-18 tsawada: modified to work in CASA 3.0.0 (imfit now returns a
                    dictionary)
2009-10-14 tsawada: fixed a bug that filenames in the log file were incorrect
                    (enbugged yesterday)
                    use bicubic interpolation instead of bilinear
2009-10-13 tsawada: regrid the image if the pixels are not square-shaped
                    (imfit returns wrong values for non-square-pixel images)
2009-09-07 tsawada: rewrite Stokes of the image to 'I' (imfit of the newest
                    CASA test build does not handle 'XX' or 'YY')
                    beamshape_error was not reset in the except branch: fixed
2009-08-12 tsawada: optimize the box by detecting edges of valid data
                    shrink the box from 12*HWHM to 4*HWHM
2009-08-10 tsawada: fixed a bug in image_gauss2d
                    (axis ratio was wrong; PA did not work)
                    change savefigure to showfigures to allow 3 branches
2009-08-07 tsawada: first version
"""
import os
import pylab as pl

###--------------------------------
### define method(s)
###--------------------------------

def imregrid_bicubic(imagename, template, output):
    """
    Perform bicubic interpolation/regridding of an image.

    Based on the official imregrid() method; a slight modification has
    been made (method = 'cubic').
    """

    casalog.origin('imregrid')
    # First check to see if the output file exists.  If it
    # does then we abort.  CASA doesn't allow files to be
    # over-written, just a policy.
    if ( len( output ) < 1 ):
        output = 'imregrid_results.im'
        casalog.post( "The output paramter is empty, consequently the" \
                      +" regridded image will be\nsaved on disk in file, " \
                      + outfile, 'WARN' )
    if ( len( output ) > 0 and os.path.exists( output ) ):
        raise Exception, 'Output file, '+output+\
              ' exists. imregrid can not proceed, please\n'+\
              'remove it or change the output file name.'

    try:

        if os.access(imagename, os.F_OK) is False:
            raise TypeError, '***Image %s is inaccessible***' % imagename
        if os.access(template, os.F_OK) is False:
            raise TypeError, '***Template image %s is inaccessible***' % \
                template

        ia.open(template)
        csys=ia.coordsys()
        shap=ia.shape()
        ia.done()
        ia.open(imagename)
        ib=ia.regrid(outfile=output, shape=shap, csys=csys.torecord(),
                  method='cubic', overwrite=True)
        ia.done()
        ib.done()
        tb.clearlocks()
        return True

    except Exception, instance:
        print '*** Error ***', instance
        try:
            ia.close()
            tb.clearlocks()
        except:
            pass
        return False

def regrid_image(image):
    """
    Regrid the image onto a square-pixel coordinate.

    Pixel numbers and increments are hard-coded to 512x512 and 1"x1".
    imregrid_bicubic() is called instead of imregrid().
    """
    templatefile = 'tmp_regrid_template.image'
    regridfile = 'tmp_regrid.image'

    ia.open(image)
    csys = ia.coordsys()
    shap = ia.summary()['header']['shape']
    ia.close()

    tmp_referencepixel = csys.referencepixel()['numeric']
    tmp_referencepixel[0] = 256
    tmp_referencepixel[1] = 256
    csys.setreferencepixel(tmp_referencepixel.tolist())

    tmp_increment = csys.increment()['numeric']
    tmp_increment[0] = radians(-1./3600.)
    tmp_increment[1] = radians( 1./3600.)
    csys.setincrement(tmp_increment.tolist())

    shap[0] = 512
    shap[1] = 512

    if os.path.isdir(templatefile):
        os.system('rm -r %s' % templatefile)
    ia.newimagefromshape(outfile = templatefile, \
        shape = shap.tolist(), csys = csys.torecord())

    if os.path.isdir(regridfile):
        os.system('rm -r %s' % regridfile)
    # default('imregrid')
    imregrid_bicubic(imagename = image, template = templatefile, output = regridfile)

    if os.path.isdir(templatefile):
        os.system('rm -r %s' % templatefile)

    return regridfile

def calc_box(image, peakpos, cdelt):
    """
    Return a box for imfit.

    image : 2-dimensional array
    peakpos: 2-item list of the position (pixels) of peak
    cdelt: 2-item list of cdelt value

    This method first gets an approximate value of
    half-width-at-half-maximum along -x, +x, -y, and +y directions
    and adopt the maximum one.  Then the mask is set to be
    (Xpeak, Ypeak) +/- 2*HWHM.  The edges of valid image (i.e.,
    image > 0) are used instead if the mask is beyond them.
    """
    imsize = [len(image), len(image[0])]
    peakval = image.max()
    hwhm_tmp = 0

    # get an approximate value of half-width-at-half-maximum
    # and the edge of the region having non-blank value

    # search for -x direction
    index = peakpos[0]
    while index >= 0:
        if image[index][peakpos[1]] > peakval/2.:
            hwhm_tmp = peakpos[0] - index
        if image[index][peakpos[1]] <= 0.:
            break
        index -= 1
    edge_x_min = index+1
    hwhm = hwhm_tmp * abs(cdelt[0])

    # search for +x direction
    index = peakpos[0]
    while index < imsize[0]:
        if image[index][peakpos[1]] > peakval/2.:
            hwhm_tmp = index - peakpos[0]
        if image[index][peakpos[1]] <= 0.:
            break
        index += 1
    edge_x_max = index-1
    hwhm = max(hwhm, hwhm_tmp * abs(cdelt[0]))

    ### search for -y direction
    index = peakpos[1]
    while index >= 0:
        if image[peakpos[0]][index] > peakval/2.:
            hwhm_tmp = peakpos[1] - index
        if image[peakpos[0]][index] <= 0.:
            break
        index -= 1
    edge_y_min = index+1
    hwhm = max(hwhm, hwhm_tmp * abs(cdelt[1]))

    ### search for +y direction
    index = peakpos[1]
    while index < imsize[1]:
        if image[peakpos[0]][index] > peakval/2.:
            hwhm_tmp = index - peakpos[1]
        if image[peakpos[0]][index] <= 0.:
            break
        index += 1
    edge_y_max = index-1
    hwhm = max(hwhm, hwhm_tmp * abs(cdelt[1]))

    # box is set to 4 x HWHM centered at peak position
    return '%d,%d,%d,%d' % (
        max(peakpos[0] - 2*int(round(hwhm/abs(cdelt[0]))), edge_x_min), \
        max(peakpos[1] - 2*int(round(hwhm/abs(cdelt[1]))), edge_y_min), \
        min(peakpos[0] + 2*int(round(hwhm/abs(cdelt[0]))), edge_x_max), \
        min(peakpos[1] + 2*int(round(hwhm/abs(cdelt[1]))), edge_y_max)  \
    )


def image_gauss2d(xarr, yarr, xpeak, ypeak, bmaj, bmin, pa):
    """
    Returns an elliptical Gaussian as 2-D list of float numbers.

    xarr, yarr   : 2-D list of x and y coordinates.
    xpeak, ypeak : Peak position in x and y.
    bmaj, bmin   : Major and minor axes of the Gaussian.
                   # The unit of these 6 parameters are arbitrary
                   # but should be common.
    pa           : Position angle of the Gaussian in degrees.
    """
    sin_pa = pl.sin(pa * pl.pi/180.)
    cos_pa = pl.cos(pa * pl.pi/180.)
    two_sigma_x_sq = 2. * bmin*bmin / (8. * pl.log(2.))
    two_sigma_y_sq = 2. * bmaj*bmaj / (8. * pl.log(2.))

    result = xarr.copy()

    for i in range(len(xarr)):
        for j in range(len(xarr[0])):
            x = xarr[i][j] - xpeak
            y = yarr[i][j] - ypeak
            rot_x =   x * cos_pa - y * sin_pa
            rot_y =   x * sin_pa + y * cos_pa
            result[i][j] = pl.exp( -(rot_x*rot_x/two_sigma_x_sq + \
                rot_y*rot_y/two_sigma_y_sq) )

    return result

###--------------------------------
### end defs
###--------------------------------

localsymbols = locals().keys()

if not 'imfiles' in localsymbols:
    print "ERROR: variable imfiles is not defined."
else:
    ### imfiles must be a list
    if isinstance(imfiles, list):
        imagefiles = imfiles
    else:
        imagefiles = [imfiles]

    if  'showfigures' in localsymbols:
        gaussplotlevel = showfigures
    else:
        print "INFO: showfigures is not defined --> 0"
        gaussplotlevel = 0

    ### open a file to log the results
    if  'fitlogfile' in localsymbols:
        fitlogfilename = fitlogfile
    else:
        fitlogfilename = 'gaussfitmap_log.txt'
        print "INFO: fitlogfile is not defined. Use \"%s\"." % \
            fitlogfilename

    logfileo = open(fitlogfilename, 'w')
    logfileo.write("#filename\tpeakX[\"]\tpeakY[\"]\t"+ \
        "Bmaj[\"]\tBmaj_err[\"]\tBmin[\"]\tBmin_err[\"]\t"+ \
        "PA[deg]\tPA_err[deg]\n")

    default('imfit')

    #--------------------------------
    # main loop: execute imfit for each entry of imfiles
    #--------------------------------
    for imagename_org in imagefiles:
        if not os.path.isdir(imagename_org):
            print "ERROR: an entry \"%s\" " % imagename_org + \
                "in imfiles is not found and skipped"
            continue

        # regrid the image if the pixels are not square-shaped
        ia.open(imagename_org)
        incr = ia.summary()['header']['incr'].tolist()
        ia.close()
        if 0.99 < fabs(incr[0]/incr[1]) < 1.01:
            pixel_is_square = True
            imagename = imagename_org
        else:
            pixel_is_square = False
            imagename = regrid_image(imagename_org)

        ia.open(imagename)
        #
        image = ia.getchunk(dropdeg=True)
        summary = ia.summary()
        stat = ia.statistics()
        # rewrite Stokes to 'I' --workaround for imfit to work
        tmpcoordsys = ia.coordsys()
        tmpcoordsys.setstokes('I')
        ia.setcoordsys(tmpcoordsys.torecord())
        #
        ia.close()

        # imsize = summary['header']['shape'][:2]
        peakpos = stat['maxpos'][:2]
        # peakval = stat['max'][0]
        cdelt = summary['header']['incr'][:2] * 180./pl.pi

        box = calc_box(image, peakpos, cdelt)
        # fit it
        try:
            # for CASA 2.4.0:
            # fitresult = imfit().getcomponent(0)
            # for CASA 3.0.0:
            fitresult = imfit()['results']['component0']

            # peak position (x, y) in arcsec
            peakpos = [ \
                qa.convert(fitresult['shape']['direction']['m0'], \
                    'arcsec')['value'], \
                qa.convert(fitresult['shape']['direction']['m1'], \
                    'arcsec')['value'] \
            ]
            # major and minor axes in arcsec, and position angle in deg
            beamshape = [ \
                qa.convert(fitresult['shape']['majoraxis'], \
                    'arcsec')['value'], \
                qa.convert(fitresult['shape']['minoraxis'], \
                    'arcsec')['value'], \
                qa.convert(fitresult['shape']['positionangle'], \
                    'deg')['value'] \
            ]
            # and their errors ...
            beamshape_error = [ \
                qa.convert(fitresult['shape']['majoraxiserror'], \
                    'arcsec')['value'], \
                qa.convert(fitresult['shape']['minoraxiserror'], \
                    'arcsec')['value'], \
                qa.convert(fitresult['shape']['positionangleerror'], \
                    'deg')['value'] \
            ]

            # destroy regridded temporal image
            if pixel_is_square == False:
                os.system('rm -r %s' % imagename)
                imagename = imagename_org

            # show (and save) images
            if gaussplotlevel > 0:
                if pixel_is_square == False:
                    # reload the original (i.e., non-square-pixel) data
                    # for drawing a map ...
                    ia.open(imagename)
                    image = ia.getchunk(dropdeg=True)
                    summary = ia.summary()
                    stat = ia.statistics()
                    ia.close()
                    cdelt = summary['header']['incr'][:2] * 180./pl.pi

                blcf = stat['blcf'].split(',')
                trcf = stat['trcf'].split(',')
                xaxismax = qa.quantity(blcf[0], 'deg')['value']
                yaxismin = qa.quantity(blcf[1], 'deg')['value']
                xaxismin = qa.quantity(trcf[0], 'deg')['value']
                yaxismax = qa.quantity(trcf[1], 'deg')['value']
                if xaxismax < xaxismin:
                    xaxismin -= 360.

                (xcoord, ycoord) = pl.mgrid[ \
                    xaxismax:(xaxismin+cdelt[0]/2.):cdelt[0], \
                    yaxismin:(yaxismax+cdelt[1]/2.):cdelt[1] ]
                gaussim = image_gauss2d(xcoord, ycoord, \
                    peakpos[0]/3600., peakpos[1]/3600., \
                    beamshape[0]/3600., beamshape[1]/3600., \
                    beamshape[2])
                xcoord = pl.mgrid[ \
                    xaxismax:(xaxismin+cdelt[0]/2.):cdelt[0] ]
                ycoord = pl.mgrid[ \
                    yaxismin:(yaxismax+cdelt[1]/2.):cdelt[1] ]

                pl.close(0)
                pl.figure(0)
                pl.imshow(pl.transpose(image), origin='lower', \
                    extent=[xaxismax, xaxismin, yaxismin, yaxismax], \
                    aspect='equal', interpolation='nearest')
                pl.hot()
                pl.xlabel('X [degrees]')
                pl.ylabel('Y [degrees]')
                pl.title(imagename)
                pl.colorbar()
                pl.hold(True)
                contourlevs = pl.array([0.9, 0.5, 0.1, 0.01])
                # pl.contour(-ycoord, -xcoord, pl.transpose(gaussim), \
                pl.contour(xcoord, ycoord, pl.transpose(gaussim), \
                    contourlevs, aspect='equal', colors='white')
                pl.imshow(pl.transpose(image), origin='lower', \
                    extent=[xaxismax, xaxismin, yaxismin, yaxismax], \
                    aspect='equal', interpolation='nearest')
                if gaussplotlevel > 1:
                    pl.savefig(imagename+'.png', format='png')

        except:  # if fit (or plot) failed, set all results to zero
            peakpos = [0, 0]
            beamshape = [0, 0, 0]
            beamshape_error = [0, 0, 0]
            imagename = imagename_org

        logfileo.write("%s\t" % imagename + \
            "%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.1f\t%.1f\n" % \
            (peakpos[0], peakpos[1], \
             beamshape[0], beamshape_error[0], \
             beamshape[1], beamshape_error[1], \
             beamshape[2], beamshape_error[2]) )

    logfileo.close()

