#!/opt/local/bin/python
# This script convolves a Gaussian with the default GJinc function
# implemented in casapy.  It simulates the expected FWHM of a single-dish
# ALMA image.
#   GJinc = Gaussian times Jinc, truncated at the radius of the first null
# - Todd Hunter
# June 20, 2013
#
import numpy as np
import pylab as pb
from scipy.special import j1
from scipy.signal import convolve

pixelsize = 10 # arcsec
fwhm = 53 # arcsec, i.e. the result of 1.18*lambda/D  (D=12m)
jwidth = 1.55*pixelsize
# casa default gwidth=-1, which means 2.52*sqrt(log(2)) * pixel
gwidth = 2.52*(np.log(2)**0.5)*pixelsize

def gjinc(x):
    result = jinc(x) * np.exp(-(x/gwidth)**2)
    return result

def jinc(x):
    result = j1(np.pi*x/jwidth) / (np.pi*x/jwidth)
    for i in range(len(x)):
        if (abs(x[i]) < 1e-8):
            result[i] = 0.5
    return result*2

def gauss(x, fwhm):
    sigma = fwhm/2.3548201
    result = np.exp(-(x**2/(2*sigma**2)))
    return result

def trunc(result):
    # casa default truncate=-1, which means truncate at radius of first null
    for r in range(len(result)/2,len(result)):
        if (result[r]<0):
            truncateBeyond = r
            break
    for r in range(len(result)/2,0,-1):
        if (result[r]<0):
            truncateBefore = r
            break
    mask = np.zeros(len(result))
    mask[truncateBefore:truncateBeyond] = 1
    result *= mask
    return result

def findFWHM(x,y):
    halfmax = np.max(y)*0.5
    for r in range(len(result)/2,len(result)):
        if (result[r]<halfmax):
            r0 = r
            break
    for r in range(len(result)/2,0,-1):
        if (result[r]<halfmax):
            r1 = r
            break
    return(np.abs(x[r1]-x[r0]))
    
myxaxis = np.arange(-2*fwhm,2*fwhm+0.1,0.2)
mygaussian = gauss(myxaxis,fwhm)
myjinc = jinc(myxaxis)
mygjinc = trunc(gjinc(myxaxis))
result = convolve(myjinc, mygaussian, mode='same')
result /= np.max(result)
myfwhm = findFWHM(myxaxis,result)
pb.clf()
desc = pb.subplot(111)
pb.plot(myxaxis, mygaussian, 'k', myxaxis, myjinc, 'r', myxaxis, mygjinc,'b',myxaxis,result,'g')
lx = 0.05
pb.text(0.55,0.94,r'Gaussian$\otimes$GJinc (FWHM=%.1f")'%(myfwhm), color='g',transform=desc.transAxes)
pb.text(lx,0.94,'Gaussian (FWHM=%.1f")'%(fwhm), color='k',transform=desc.transAxes)
pb.text(lx,0.88,'Jinc (jwidth=%.1f")'%jwidth, color='r',transform=desc.transAxes)
pb.text(lx,0.82,'GJinc (gwidth=%.1f")'%gwidth, color='b',transform=desc.transAxes)
pb.xlabel('Arcsec')
desc.xaxis.grid(True,which='major')
desc.yaxis.grid(True,which='major')
pb.xlim([-1.5*fwhm,1.5*fwhm])
pb.ylim([-0.15,1.02])
pb.savefig('gjinc.png')
pb.show()
