"""
Script Name : imageBeamMap.py
Goal: Create Beammaps for Total Power Observations. Beammaps for XX and YY polarizations are
      independently created. 

Usage: Within CASA:
       import imageBeamMap as ibm # (or other name)
       ibm.makeMap(msfile,  aid, spwnum, scanum)
       example: ibm.makeMap('uid___X54_X2119_X1.ms',0, 0, 7)
       reload(ibm) # if code changes

Keywords:
       msfile   : Measurement set
       aid      : antenna id to analyse when there is more than 1 antenna
       spwnum   : ID of spectral window to be imaged
       scanum   : is now defined in the data itself.
Comments: 
       Order 1 baseline fitting supported.
       Beam maps for both polarizations are produced.
               
"""

import os
import glob
from math import *
import ephem

#from scipy import *
#from numpy import *
from pylab import *
from taskinit import *

from imfit import *
from imhead import *
import analysisUtils as aU
from deconv_planet import npToMap,mapToNP, fitObs

def makeMap(msfile, aid, spwnum, scanum=-1):

   if scanum == -1:
      print "You have no specified the scan number to be analyzed, please specify it."
      return
   
   #get info about observations.
   tb.open(msfile+'/ANTENNA')
   pos=tb.getcol('POSITION')[:,aid]
   ant=tb.getcol('NAME')[aid]
   pad=tb.getcol('STATION')[aid]
   tb.close()
   
   # get name of source.
   tb.open(msfile+ '/SOURCE')
   src=tb.getcol('NAME')[0]
   tb.close()
     
   #get frequency and define image size and pixel size
   tb.open(msfile+'/SPECTRAL_WINDOW')
   freq=tb.getcol('CHAN_FREQ')
   freq=freq[0,0]
   tb.close()
   if freq < 116e9:
      band='3'
      cellx='10arcsec'
      nx=60
   elif freq < 275e9:
      band='6'
      cellx='6arcsec'
      nx=50
   elif freq < 373e9:
      band='7'
      cellx='4arcsec'
      nx=50
   elif freq < 720e9:
      band = '9'
      cellx='2arcsec'
      nx=50

   if src == 'Moon':
      cellx='60arcsec'
      nx=250

   # get scan table, state table and time table.
   tb.open(msfile)
   scan=tb.getcol('SCAN_NUMBER')
   t0=tb.getcol('TIME')
   tb.close()

   # get focus values
   focus=aU.getMeasFocus(msfile, aid, scanum)
   if size(shape(focus)) > 1:
      focus= focus[:,aid]
      
   #define site position according to position of antennas
   osfSite=me.position('ITRF',qa.quantity(pos[0],'m'),qa.quantity(pos[1],'m'),qa.quantity(pos[2],'m'))
   me.doframe(osfSite)
   print ant+ ' located at '+ pad + ' observing '+src +' at Band'+ band +' Scan  '+ str(scanum)
   imagefile=msfile.split('/')[-1][:-3]+'_'+ant+'_'+pad+'_'+src+'_band'+band+'_spw' \
              +str(spwnum)+'_scan'+str(scanum)
   print '%s' % imagefile

   #find time of first beam map scan
   f=find(scan==int(scanum))
   scanbeg=t0[f[0]]
   print scanbeg
  
   #read pointing  offset data and get row of scan of interest
   tb.open(msfile+'/POINTING')
   t=tb.getcol('TIME')
   dir=tb.getcol('DIRECTION')
   off=tb.getcol('POINTING_OFFSET')
   tar=tb.getcol('TARGET')
  # tb.putcol("DIRECTION", off)
   nrow=tb.nrows()
   tb.close()
   rownum=0
   mintimediff=1000000
   for k in range(nrow):
      if(abs(scanbeg - t[k]) < mintimediff):
         rownum=k
         mintimediff=abs(scanbeg-t[k])
   print 'rownumber', rownum, mintimediff, 'among nrows=',nrow
   t0str=str(t[rownum])+'s'
   xoff=off[0,0,rownum]/cos(dir[1,0,rownum]);yoff=off[1,0,rownum]
   az0=tar[0,0,rownum]*180/pi;el0=tar[1,0,rownum]*180/pi
   
   #define phase center
   me.doframe(me.epoch('utc', t0str))
   # phcen=me.measure(me.direction(src), 'AZELGEO')
   phcen_2=me.direction('AZELGEO', qa.quantity(dir[0,0,rownum]-xoff,'rad'),qa.quantity(dir[1,0,rownum]-yoff,'rad'))
     
   phcen=phcen_2
   
   
   # Processing the MS
   im.open(msfile)
   asel="ANTENNA1==%1.0g && ANTENNA2== %1.0g" % (aid, aid)
   im.selectvis(spw=spwnum, field='0',scan=str(scanum), taql=asel)      #Select the data
   im.setvp(dovp=True)                                    # Use default PB    
   im.weight('natural')
   
   #Imaging polarization XX
   #im.defineimage(nx=nx, cellx=cellx, stokes='XX', phasecenter=phcen, movingsource=src) 
   #im.setoptions(ftmachine='sd',location=osfSite)#,gridfunction='BOX')
   #im.makeimage(type='singledish', image=imagefile+'_XX_DIR.image')  
           
   #  using pointing offsets 
   im.defineimage(nx=nx, cellx=cellx, stokes='XX', phasecenter='AZELGEO 0d0m 0d0m',movingsource='', spw=[spwnum])
   im.setoptions(ftmachine='sd',location=osfSite)#, gridfunction='BOX')
   im.setsdoptions(pointingcolumntouse='POINTING_OFFSET') #Adding KUMAR Patch here
   im.makeimage(type='singledish', image=imagefile+'_XX_OFF.image')    

   #Imaging polarization YY
   #im.defineimage(nx=nx, cellx=cellx, stokes='YY', phasecenter=phcen, movingsource=src) 
   #im.setoptions(ftmachine='sd',location=osfSite)#,gridfunction='BOX')
   #im.makeimage(type='singledish', image=imagefile+'_YY_DIR.image')     

   # using  pointing offset column
   im.defineimage(nx=nx, cellx=cellx, stokes='YY', phasecenter='AZELGEO 0d0m 0d0m',movingsource='', spw=[spwnum])
   im.setoptions(ftmachine='sd',location=osfSite)#, gridfunction='BOX')
   im.setsdoptions(pointingcolumntouse='POINTING_OFFSET')  #Adding KUMAR Patch here
   im.makeimage(type='singledish', image=imagefile+'_YY_OFF.image')  

   #subtract background form image
   subtract_background(imagefile+'_XX_OFF.image')
   subtract_background(imagefile+'_YY_OFF.image') 

   #make plots
   makePlot(1,imagefile, 'XX', az0, el0, t0str, focus)
   makePlot(2,imagefile, 'YY', az0, el0, t0str, focus)

   return 
#fitting deconv gaussian in tod.
#import reduc_oof as r
#from deconv_planet import npToMap,mapToNP, fitTS
#from pylab import *
#import pyfits
#
#msfile='uid___X02_X4a3c3_X1.ms'
#aid=0
#fitsfile=r.main(msfile, aid)
#print fitsfile

#planet=pyfits.open(fitsfile)[0].header[7]
#radius=pyfits.open(fitsfile)[0].header[8]
#map_c_bestfit, r, ch=fitTS(fitsfile, radius,256, radians(500.0/3600.0),fit=["amp", 'x0','y0','sigma','diff','rho'],ic=[1, 0,  0,  radians(60.0/3600)**2, 0, 0])
#param=convert_params(r)


# plotting and   gaussian fitting function.
def  makePlot(f,imagefile, pol, az0, el0, t0str, focus):
   
   close(f)
   close(f+2)
   close(f+4)
   close(f+6)
   close(f+8)
   close(f+10)
   close(f+14)
   # myim=ia.fitpolynomial(order=2, axis=0, mask='indexin( 0, [15,30]) || indexin( 1, [15,30])')
   # ia.unlock()

   fname0='%s_%s_OFF' % (imagefile,  pol)
   src=imagefile.split('_')[8]
   im0=ia.open(fname0+'.image')
   #ia.setbrightnessunit('K')

   z0=ia.getchunk(dropdeg=True)
   s=ia.summary()
   st0=ia.statistics()
   ia.close()

   # imhead(imagename=fname+'.image',mode='put', hdkey='bunit',hdvalue='Jy/pixel')
   # imhead(imagename=fname+'.image',mode='put', hdkey='crval3',hdvalue='XX')
  
   blcf=st0['blcf']
   azmax=float(blcf.split(',')[0][:-3])
   elmin=float(blcf.split(',')[1][:-3])
   trcf=st0['trcf']
   azmin=float(trcf.split(',')[0][:-3])-360
   elmax=float(trcf.split(',')[1][:-3]) 
   dim=s['header']['shape']
   delta=s['header']['incr']
   mapsize=abs(degrees(dim[0]*delta[0])*3600) # in arcsec
   
   blc=st0['blc']
   trc=st0['trc']
   pix=st0['maxpos'][0:2]
   mm=st0['max'][0]
   
   # fit 2D elliptical gaussian
   wd=15
   arcsec2deg=3600.
   #bx=str(max(pix[0]-wd,0)) +','+str(max(pix[1]-wd,0))+',' + str(min(pix[0]+wd,dim[0]-1))+','+ str(min(pix[1]+wd,dim[1]-1))
   bx=str(blc[0]) +','+str(blc[1])+',' + str(trc[0])+','+ str(trc[1])
   print pix, bx, mapsize
   
   # try:
   if 1:
      fitpar=imfit(fname0+'.image',box=bx, stokes=pol,\
                   residual=fname0+'.residual',\
                   model=fname0+'.gaussfit')
      fit_comp=fitpar['results']['component0']
       
      # i1 = ia.convertflux('1.0989e+05Jy', major='0.020834deg', minor=' 0.020168deg', topeak=T);
      # ia.brightnessunit()

      #open the residual image and  get residual stats in 10dB contour
      ia.open(fname0+'.residual')
      z2=ia.getchunk(dropdeg=True)  
      mtext='%s.image > %s' %(fname0, mm/10)
      st2=ia.statistics(mask=mtext) 
      percent_rms=st2['rms'][0]*100/mm
      ia.close()

       
      # deconvolove source to get original beam
      if src in ['Mercury','Mars','Venus', 'Jupiter','Saturn', 'Neptune', 'Uranus']:
         print 'Deconvolution of planet size from observed beam'
         tt=qa.time(t0str, form='fits')
         dt=tt.replace('-','/').replace('T',' ')
         exec 'j=ephem.%s()' %src
         j.compute(dt)
         diameter=j.size
         planet_radius=diameter/2. # in arcsec
         ad="%s arcsec" %diameter
         print 'Assumed diameter of %s is %s' %(src, ad)
         f2s=sqrt(8*log(2))

         map_c=npToMap(z0, mapsize)
         map_c_bestfit, r, ch=fitObs(map_c, planet_radius, mapsize, ic=[1.0, 0, 0, 50.0, 0, 0])
         z0_bestfit=mapToNP(map_c_bestfit)
         param_de=convert_params(r)
         print param_de[1]*f2s, param_de[2]*f2s 
         print sqrt(r[3])*f2s
         print fit_comp['shape']['majoraxis']['value'],\
               fit_comp['shape']['minoraxis']['value']
      else:
         f2s=sqrt(8*log(2))
         r=[0,0,0,0,0,0]
         param_de=convert_params(r)
         diameter=0
        
      #save the fit params
      amplitude= fit_comp['flux']['value'][0]
      amplitude_err= fit_comp['flux']['error'][0]
      center_x= fit_comp['shape']['direction']['m0']['value']*180/pi  #in deg
      center_x_err = fit_comp['shape']['direction']['error']['latitude']['value']/arcsec2deg
      center_y = fit_comp['shape']['direction']['m1']['value']*180/pi  #in deg
      center_y_err = fit_comp['shape']['direction']['error']['longitude']['value']/arcsec2deg
      fwhma = fit_comp['shape']['majoraxis']['value']/arcsec2deg    #in deg
      fwhma_err  = fit_comp['shape']['majoraxiserror']['value']/arcsec2deg
      fwhmb  = fit_comp['shape']['minoraxis']['value']/arcsec2deg   #in deg
      fwhmb_err  = fit_comp['shape']['minoraxiserror']['value']/arcsec2deg
      fwhma_de  = param_de[1]*f2s/arcsec2deg
      fwhmb_de  = param_de[2]*f2s/arcsec2deg
      pa   = fit_comp['shape']['positionangle']['value']
      pa_err    = fit_comp['shape']['positionangleerror']['value']
      inpPars  = [amplitude,center_x,center_y,fwhma,fwhmb,pa+90]  #0,180  vert, 45 topleft bottomright
      
      Xin,Yin  = mgrid[azmax:azmin:degrees(delta[0]),elmin:elmax:degrees(delta[1])]
      fit2 = image_gauss2d(Xin, Yin, center_x, center_y, fwhma, fwhmb, pa)
         
      # plot beam in linear space
      figure(f)
      imshow(transpose(z0), origin='lower',extent=[azmax, azmin, elmin, elmax],aspect='equal',interpolation='nearest')
      hot()
      [x,y]=drawCircle(0.7*azmax,0.7*elmin,diameter/3600, 1000)
      plot(x,y,'b')
      xlabel('AZ (deg)'); ylabel('EL (deg)')
      tit1=imagefile +' ' +pol
      title(tit1)
      colorbar()
      hold(True)
      V=array([0.9, 0.5, 0.1, 0.01])#* amplitude
      contour(-Yin ,-Xin,transpose(fit2),V, aspect='equal',colors='white')
      imshow(transpose(z0), origin='lower',extent=[azmax, azmin, elmin, elmax],aspect='equal',interpolation='nearest')
     
      
      t0='FWHMa = %3.4g (%1.2g)' % (fwhma*arcsec2deg , fwhma_err*arcsec2deg)
      t1='FWHMb = %3.4g (%1.2g)' % (fwhmb*arcsec2deg , fwhmb_err*arcsec2deg)
      t0d='FWHMa_de = %3.4g (%1.2g)' % (fwhma_de*arcsec2deg , fwhma_err*arcsec2deg)
      t1d='FWHMb_de = %3.4g (%1.2g)' % (fwhmb_de*arcsec2deg , fwhmb_err*arcsec2deg)
      t2='Pa      = %3.4g (%1.2g)'    % (pa , pa_err)
      t3='%s diameter = %3.4g arcsec' %(src,diameter)
      figtext(0.43, 0.85, t0, color='w')
      figtext(0.43, 0.8225, t1, color='w')
      figtext(0.43, 0.80, t0d, color='w')
      figtext(0.43, 0.775, t1d, color='w')
      figtext(0.43, 0.75, t2, color='w')
      figtext(0.28,0.24, t3, color='b')
      figtext(0.89, 0.4, 'uncalibrated Jy/pixel', rotation='vertical')
      savefig(fname0+'.png', format='png')
      
      # plot beam in log space.
      figure(f+2)
      zlog=10*log10(abs(z0+1e-20))
      mml=ceil(zlog.max())
      imshow(transpose(zlog), origin='lower',extent=[azmax, azmin, elmin, elmax],aspect='equal',vmin=0, vmax=mml)
      hot()
      xlabel('AZ (deg)'); ylabel('EL (deg)')
      tit1=imagefile + ' '+pol
      title(tit1)
      figtext(0.89, 0.4, 'dB', rotation='vertical')
      colorbar()
      savefig(fname0+'_log.png', format='png')

      # plot difference beam
      if pol == 'YY':
         fname1='%s_XX_OFF' % (imagefile)
         im1=ia.open(fname1+'.image')
         z1=ia.getchunk(dropdeg=True)
         # open the model image
         ia.open(fname0+'.gaussfit')
         fit3x=ia.getchunk(dropdeg=True)
         ia.open(fname1+'.gaussfit')
         fit3y=ia.getchunk(dropdeg=True)
         ia.close()
         figure(f+10)
         diff=transpose(z1/z1.max())-transpose(z0/z0.max())
         imshow(diff, origin='lower',extent=[azmax, azmin, elmin, elmax],aspect='equal',interpolation='nearest')
         hot()
         xlabel('AZ (deg)'); ylabel('EL (deg)')
         tit1=imagefile +'  XX - YY '
         title(tit1)
         colorbar()
         hold(True)
         diff_fit=(transpose(fit3y/fit3y.max())- transpose(fit3x/fit3x.max()))
         V=array([0.6*diff_fit.max(), 0.6*diff_fit.min()])#* amplitude
         contour(-Yin ,-Xin,diff_fit,V, aspect='equal',colors='green')
         imshow(transpose(z0/z0.max())-transpose(z1/z1.max()), origin='lower',extent=[azmax, azmin, elmin, elmax],aspect='equal',interpolation='nearest')
         hold(True)
         savefig(fname0+'_diffpoint.png', format='png')
      
      # plot residual
      figure(f+4)
      res=z2*100/mm
      imshow(transpose(res), origin='lower',extent=[azmax, azmin, elmin, elmax],aspect='equal',vmin=-10, vmax=10)
      hold(True)
      contour(-Yin ,-Xin,transpose(fit2),array([0.1]), aspect='equal',colors='white')
      imshow(transpose(res), origin='lower',extent=[azmax, azmin, elmin, elmax],aspect='equal',vmin=-10, vmax=10)
      hot()
      xlabel('AZ (deg)'); ylabel('EL (deg)')
      tit1=imagefile + ' '+pol+ ' residual'
      t0='RMS Residual in 10dB contour = %1.3g%%  ' % percent_rms
      figtext(0.33, 0.85, t0, color='w')
      figtext(0.89, 0.4, '% of boresight amplitude', rotation='vertical')
      title(tit1)
      colorbar()
      savefig(fname0+'_residual.png', format='png')
     
      #plot radial profile
      figure(f+6)
      [bin, rp]=radial(z0,dim[0:2],pix )
      [binf, rpf]=radial(transpose(fit2), dim[0:2], pix)
      pixsize=degrees(abs(delta[0]))*3600.
      semilogy(bin*pixsize, abs(rp)/mm,'r-o')
      semilogy(binf*pixsize, abs(rpf),'b')
      xlim([0, 0.8*max(bin*pixsize)])
      ylim([1e-5, 1.1])
      xlabel('radius from peak (arcsec)')
      ylabel('abs (amplitude)')
      tit=imagefile + ' '+pol+ ' radial profile'
      title(tit)
      savefig(fname0+'_radial.png', format='png')

      #plot horizontal and vertical cut 
      figure(f+8)
      subplot(2,1,1)
      plot(Xin[:, pix[0]],zlog[:,pix[0]])
      ylabel(' horizontal cut')
      ylim([-20, mml]); #xlim([-2, 2])
      tit=imagefile + ' '+pol+ ' cuts'
      title(tit)
      subplot(2,1,2)
      plot(Yin[pix[0],:],zlog[pix[0],:])
      xlabel('AZ/EL (deg)'); ylabel('vertical cut')
      ylim([-20, mml]);# xlim([-2, 2])
      savefig(fname0+'_cuts.png', format='png')
      
      
      # test viewer scripting
      # vc = viewer(connection=True)
      # vc.cwd('./')
      # vc.load(fname+'.image',displaytype='raster')
      # vc.restore(restore_path,new_window=False)
      # vc.load(fname+'.gaussfit',displaytype='contour')
      # vc.output(fname+'.png',scale=2.0)
      
      
   # except:  # if fit (or plot) failed, set all results to zero
   if 0:
      amplitude_err=  NaN
      amplitude    =  NaN
      center_x     =  NaN
      center_x_err =  NaN 
      center_y     =  NaN
      center_y_err =  NaN
      fwhma        =  NaN
      fwhma_err    =  NaN
      fwhmb        =  NaN
      fwhmb_err    =  NaN
      fwhma_de     =  NaN
      fwhmb_de     =  NaN
      ad           =  NaN
      pa           =  NaN
      pa_err       =  NaN
      percent_rms  =  NaN
               
   # output the parameters to file.
   outfile=open(fname0+'.param','w')
   outfile.write('Output  parameters for %s_XX_OFF.image\n' % imagefile)
   outfile.write('Format:time (mjd s), az0, el0, src diam,  Amp, Amp_err, cen_x, cen_x_err,\
   cen_y,cen_y_err, fwhma, fwhma_err ,fwhmb, fwhmb_err, fwhma_de ,fwhmb_de, pa,\
   pa_err, percent_rms, focusx, focusy, focusz\n')
   
   outfile.write('%s %3.4g %3.4g %2.4g %3.5g %3.5g %3.5g %3.5g %3.5g %3.5g %3.5g %3.5g %3.5g %3.5g %3.5g %3.5g %3.5g %3.5g %1.2g %3.5g %3.5g %3.5g\n'
                 % (t0str[0:-1], az0, el0, diameter, amplitude, amplitude_err, center_x,
                    center_x_err, center_y, center_y_err, fwhma, fwhma_err ,fwhmb,
                    fwhmb_err,fwhma_de,fwhmb_de,
                    pa, pa_err, percent_rms, focus[0], focus[1], focus[2]))
   outfile.close()

   return


def subtract_background(filename):
   # subtract background.  For each row in the image,take the first Nbase and last Nbase nonzero index
   # and fit a polynomial of order deg from from those two set if indices.
   ia.open(filename)
   x=1 # number of pixels to exclude at edge.
   nbase=3 # number of points to use at each edge to do baseline subtraction.
   deg=1 # degree of baseline sibrtraction (1= linear)
   image = ia.getchunk(dropdeg=True)
   for i in range(image.shape[0]):
      if (image[:,i].any() != 0):
         nzIndex=image[:,i].nonzero()[0]
         nonZero=image[nzIndex,i]
         xdata=concatenate((nzIndex[x:nbase+x], nzIndex[-x-nbase:-x]),0)# concatenating index
         ydata=concatenate((nonZero[x:nbase+x], nonZero[-x-nbase:-x]),0)
         polycoeffs=polyfit(xdata, ydata, deg)
         yfit=polyval(polycoeffs, nzIndex)
         image[nzIndex,i] = image[nzIndex,i] - yfit
   ia.putchunk(image)
   ia.close()
   return

def radial(im,dim, pix):

   try:
      # calculate average radial profile
      y, x = indices((dim[0],dim[1])) # first determine radii of all pixels
      r = sqrt((x-pix[0])**2+(y-pix[1])**2)
      ind = argsort(r.flat) # get sorted indices
      
      # needed to arange image values too
      sr = r.flat[ind] # sorted radii
      sim = im.flat[ind] # image values sorted by radii
      
      rind=arange(0, size(ind), 10)
      nr = rind[1:] - rind[:-1] # number in radius bin
      sim = sim*1. 
      csim = cumsum(sim) # cumulative sum to figure out sums for each radii bin 
      tbin = csim[rind[1:]] - csim[rind[:-1]] # sum for image values in radius bins 
      radialprofile = tbin/nr 
      
      # semilogy(sr,abs(sim),'.')
      # semilogy(sr[rind[0:-1]], abs(radialprofile),'ro')
   except:

      sr=zeros(size(im))*NaN
      rind=arange(0, size(sr),10)
      radialprofile=sr
      
   return [sr[rind[0:-1]], radialprofile]
   
   
def one2two(ind, dim1, dim2):
   ind1=ind/dim1
   ind2=mod(ind, dim2)
   return [ind1, ind2]

def pol2cart(th,r): 
    x = r*cos(th) 
    y = r*sin(th) 
    return (x,y) 
    
def drawCircle(cx,cy,radius,np): 
    theta = linspace(0,2*pi,np) 
    rho = ones((1,np))*radius 
    x,y = pol2cart(theta,rho) 
    x = x + cx 
    y = y + cy 
    return [x[0,:],y[0,:]]
   

def image_gauss2d(xarr, yarr, xpeak, ypeak, bmaj, bmin, pa):
    """
    Returns an elliptical Gaussian as 2-D list of float numbers.
   
    xarr, yarr   : 2-D list of x and y coordinates.
    xpeak, ypeak : Peak position in x and y.
    bmaj, bmin   : Major and minor axes of the Gaussian.
                   # The unit of these 6 parameters are arbitrary
                   # but should be common.
    pa           : Position angle of the Gaussian in degrees.
    """
    sin_pa = sin(pa * pi/180.)
    cos_pa = cos(pa * pi/180.)
    two_sigma_x_sq = 2. * bmin*bmin / (8. * log(2.))
    two_sigma_y_sq = 2. * bmaj*bmaj / (8. * log(2.))

    result = xarr.copy()

    for i in range(len(xarr)):
        for j in range(len(xarr[0])):
            x = xarr[i][j] - xpeak
            y = yarr[i][j] - ypeak
            rot_x =   x * cos_pa - y * sin_pa
            rot_y =   x * sin_pa + y * cos_pa
            result[i][j] = exp( -(rot_x*rot_x/two_sigma_x_sq + \
                rot_y*rot_y/two_sigma_y_sq) )

    return result

def convert_params(r):
   """
   to convert the parameters from Bojan's 2D gaussian to the parameters in image_gauss2d
   Bojan's gaussian
   r[0]=Amp, r[1]=center_x, r[2]=center_y
   r[3]=sigma, r[4]=rho, r[5]=diff
   return amp * exp ( -1.0 * ( 1.0/(2*sigma)   *  (pow(dx,2) + pow(dy,2) +
   rho *  ( dx*dy)  +
   diff   *  ( pow(dx,2) - pow(dy,2 )))));

   pa=0.5*atan2(rho, 2*diff)
   2 * bmin^2 = 2 * sigma * cos(2 pa)/ (cos(2 pa)- diff) 
   2 * bmax^2 = 2 * sigma * cos(2 pa) / (cos 2 pa)- diff)
   """
   A=r[0]
   sigma=r[3]
   rho=r[4]
   diff=r[5]

   pa=0.5*atan2(rho, 2*diff)
   bmin = sqrt(sigma * cos(2*pa) / (cos(2*pa)- diff)) 
   bmax = sqrt(sigma * cos(2*pa) / (cos(2*pa)+ diff))
   result=[degrees(pa),bmin,bmax]

   return result
   
def pipeline(ant):
   """
   This will run the beam map pipeline on your data.     
   """
   #read beam data so far
   data=read_beamdatasofar()
   filelist=data[ant]['uid']
   i=0
   while i < size(filelist):
      asdmfile=filelist[i]
      #f=raw_input('%s: Enter (analyse), 4(to skip 4 ahead), c(break) anything else to skip ' % asdmfile)
      f=''
      if f == '':
         location=locate(asdmfile)[0]
         msfile=location+'.ms'
         print 'msfile = %s' % msfile
         # if asdm has not been converted to MS, do it.
         #if not os.path.isdir(msfile):
         aU.batchasdm2MS(asdmfile, ocorr='ao')
         # if STATE  table exists fid which scans are beam maps
         scanums=aU.findBeamMap(msfile)
         #scanum=raw_input('Which scan number? (1,3,5,7):  ')
         tb.open(msfile+'/ANTENNA')
         antennas=tb.getcol('NAME')
         id=find(antennas==ant)[0]
         for scanum in scanums:
            # filelist=glob.glob('*.ms')
            # filelist=sort(filelist)
            focus=aU.getMeasFocus(msfile, id, scanum)
            makeMap(msfile, id, 0, scanum)
         i=i+1
      elif f == 'c': break
      elif type(f) == type(str()) and f.isdigit(): i=i+int(f)
      else: i=i+1
   return

def locate(file):
   #first look in current dir
   datadir='~/alma_analysis/beam_map_analysis/data_ms'
   location=os.popen('find %s -name %s.ms' % (datadir,file)).read().strip()
   if location != '':
      #print file,  location
      location=location.strip(file+'.ms') 
   else:
      datadir="/data/RADIO"
      location=os.popen('find  %s/AMPCAL %s/BEAMMAP -name %s' %(datadir, datadir, file)).read().strip()
      if location != '':
         #print file,  location
         directory=location.strip(file)
         return [location, directory]
      else:
         print 'could not find this file in local dir or in /data/RADIO/'
         return
   

def read_beamdatasofar():
   host='dbarkats@wwwpub.sco.alma.cl'
   f='public_html/beam_map_data_sofar.txt'
   cmd='scp  %s:%s ./%s' % (host, f, 'beam_map_data_sofar.tmp')
   os.system(cmd)
   
   b=open('beam_map_data_sofar.tmp')
   obs=b.readlines() 
   data={}
   for i in range(size(obs)):
      line=obs[i]
      # print  i, line
      if line.startswith('#-----'):
         antenna=line.split('#-----')[1][0:4]
         data[antenna]={}
         attList = ['uid','band','src', 'focusflag', 'focusx','focusy','focusz', 'scanum']
         for i in attList: data[antenna][i] = []
      elif line[0] != '\n' and line[0] != '#':   
         nelem=len(line.split(','))
         data[antenna]['uid'].append(line.split(',')[1].lstrip())
         data[antenna]['band'].append(line.split(',')[2].lstrip())
         data[antenna]['src'].append(line.split(',')[3].lstrip())        
         if nelem > 5:
            
            data[antenna]['focusflag'].append(line.split(',')[4].lstrip())
            focus_tmp= line.split(',')[5].lstrip()
            data[antenna]['scanum'].append(line.split(',')[6].lstrip())
            #p1=focus_tmp.find('(')
            #p2=focus_tmp.find(')')
            #data[antenna]['focusx'].append(float(focus_tmp[p1+1:p2].split(' ')[0]))
            #data[antenna]['focusy'].append(float(focus_tmp[p1+1:p2].split(' ')[1]))
            #data[antenna]['focusz'].append(float(focus_tmp[p1+1:p2].split(' ')[2]))
         
   return data

def getCalFocus(msfile):
   tb.open(msfile+'/ASDM_CALFOCUS')
   ant=tb.getcol('antennaName')
   off=tb.getcol('offset')
   wf=tb.getcol('wereFixed')
   tb.close()
   
   if off.size < 1:
      print "No TelCal Focus in this ASDM, you need to read the measured values instead."
      return False
   else:
      print "Focus values from TelCal fit results."
      off=off[0]
      
   #focus=zeros([3,size(ant)])
   um=1e6
   n_measx=size(find(wf[0] == False))
   n_measy=size(find(wf[1] == False))
   n_measz=size(find(wf[2] == False))
   n_meas=max([n_measx, n_measy, n_measz])
   focx=empty([n_meas])
   focy=empty([n_meas])
   focz=empty([n_meas])
   
   for i in range(n_measx):
      focx[i]= off[0][~wf[0]][i]*um
   for i in range(n_measy):
      focy[i]= off[1][~wf[1]][i]*um
   for i in range(n_measz):
      focz[i]= off[2][~wf[2]][i]*um
   for i in range(n_meas):
      print "%s (X  Y  Z) %s (%.0f %.0f %.0f) " % (ant[i], msfile, focx[i], focy[i], focz[i])

   if (n_measx != n_measy) or (n_measx != n_measz) or (n_measz != n_measy):
      return False
   else:
      return True
