"""
A script to produce and show map(s) from total-power OTF dataset(s).

Launch CASA and type the following:
  sdfiles    = 'uid___Xxx_Xxxxx_X1'      # ASDM or MS dataset(s) to be procesed
                                         #   no defaults
                                         #   use a list for multiple items
  oblongpix  = True                      # Use oblong pixels along scans
                                         #   defaulted to False (square pix)
  pols       = 'XX'                      # Polarization(s) to be processed
                                         #   defaulted to 'XX'
                                         #   use a list for multiple items
  showparams = True                      # Print parameters of sdtpimaging
                                         #   defaulted to False
  showviewer = True                      # Open the image with viewer
                                         #   defaulted to False
  execfile('imageotfmap.py')

In order to process multiple datasets, 'glob' package is helpful to
obtain a list of ASDM files:
  import glob
  sdfiles = glob.glob('*_X1')
  execfile('imageotfmap.py')

The script calls the sdtpimaging task.  CASA revision newer than 8237
is needed to process datasets taken at sampling intervals faster than
48 ms.

2010-03-12 tsawada: importasdm included
2010-01-18 tsawada: workaround for MS in R7.0 (add CORRECTED_DATA if not exist)
                    now works for not only planets, but also extrasolar objects
2009-09-04 tsawada: adjust imsize
2009-08-17 tsawada: remove 'scandir' option (now auto-detected)
2009-08-14 tsawada: add 'scandir' option
2009-08-13 tsawada: add 'oblongpix' option
2009-08-05 tsawada: add 'pols', 'showparams' and 'showviewer' options
                    improved the way to check if the variables are defined
2009-08-04 tsawada: first version
"""
import os

### define method(s)

def add_corrected_data_column(sdfile):
    """
    Add "CORRECTED_DATA" column if not exist.
    """
    tb.open(sdfile)
    colnames = tb.colnames()
    tb.close()
    if not ('CORRECTED_DATA' in colnames):
        im.open(sdfile, usescratch=True)
        im.close()

def get_antenna_name(sdfile):
    """
    Return the antenna name stored in given MS.
    """
    tb.open(sdfile + '/ANTENNA')
    antenna = tb.getcol('NAME')[0]
    tb.close()
    return antenna

def get_scan_dir(sdfile):
    """
    Return the scan direction from given MS.
    """
    tb.open(sdfile + '/POINTING')
    pointoff = tb.getcol('POINTING_OFFSET')
    tb.close()
    offx = pointoff[0][0][:-2]
    offy = pointoff[1][0][:-2]
    velx = abs(offx[0] - offx[1])
    vely = abs(offy[0] - offy[1])
    if velx > (10.*vely):
        return 'X'
    elif vely > (10.*velx):
        return 'Y'
    else:
        return 'Z'

def get_frequency(sdfile):
    """
    Return the frequency stored in given MS.
    """
    tb.open(sdfile + '/SPECTRAL_WINDOW')
    frequency = tb.getcol('CHAN_FREQ')[0,0]
    tb.close()
    return frequency

def get_source_name(sdfile):
    """
    Return the source name stored in given MS.
    """
    tb.open(sdfile + '/SOURCE')
    srcname = tb.getcol('NAME')[0]
    tb.close()
    return srcname

def get_band_number(frequency):
    """
    Return the band number from given frequency [Hz].
    """
    freq = frequency / 1e9
    if    31.3 <= freq <=  45.:
        return 1
    # elif  68.  <= freq <=  90.:
    #     return 2
    elif  84.  <= freq <= 116.:
        return 3
    elif 125.  <= freq <= 163.:
        return 4
    elif 163.  <= freq <= 211.:
        return 5
    elif 211.  <= freq <= 275.:
        return 6
    elif 275.  <= freq <= 373.:
        return 7
    elif 385.  <= freq <= 500.:
        return 8
    elif 602.  <= freq <= 720.:
        return 9
    elif 787.  <= freq <= 950.:
        return 10
    else:
        return 0

### end defs

lightspeed = 2.99792458e8
rad_to_arcsec = 180. / 3.14159 * 3600.

localsymbols = locals().keys()

if not 'sdfiles' in localsymbols:
    print "ERROR: variable sdfiles is not defined."
else:
    ### define sdfilelist as a list
    if isinstance(sdfiles, list):
        sdfilelist = sdfiles
    else:
        sdfilelist = [sdfiles]

    #--------------------------------
    # check if parameters are set
    #--------------------------------
    if 'pols' in localsymbols:
        ### polarizations must be a list
        if isinstance(pols, list):
            polarizations = pols
        else:
            polarizations = [pols]
    else:
        print "INFO: variable pols is not defined  --> ['XX']"
        polarizations = ['XX']

    if 'oblongpix' in localsymbols:
        useoblongpix = oblongpix
    else:
        print "INFO: variable useoblong is not defined  --> False"
        useoblongpix = False

    if 'showparams' in localsymbols:
        doparams = showparams
    else:
        print "INFO: variable showparams is not defined  --> False"
        doparams = False

    if 'showviewer' in localsymbols:
        doviewer = showviewer
    else:
        print "INFO: variable showviewer is not defined  --> False"
        doviewer = False

    #--------------------------------
    # call asap_init() if ASAP has not yet loaded in the session
    #--------------------------------
    if not 'sdtpimaging' in localsymbols:
        asap_init()

    #--------------------------------
    # set paramters common for every execution
    #--------------------------------
    default('sdtpimaging')

    calmode = 'baseline'
    masklist = [80]

    createimage = True
    phasecenter = 'AZELGEO 0d0m0 0d0m0'
    pointingcolumn = 'pointing_offset'

    #--------------------------------
    # main loop: execute sdtpimaging for every entry of sdfiles
    #--------------------------------
    for sdfilename in sdfilelist:
        if not os.path.isdir(sdfilename):
            print "ERROR: an entry \"%s\" " % sdfilename + \
                "in sdfiles is not found and skipped"
            continue

        if sdfilename.split('.')[-1] == 'ms':
            sdfile = sdfilename
        else:
            sdfile = sdfilename+'.ms'

        if os.path.isdir(sdfile):
            print "MeasurementSet %s is found" % sdfile
        else:
            print "MeasurementSet %s is not found. Calling asdm2MS" % sdfile
            importasdm(asdm=sdfilename, vis=sdfile, ocorr_mode='ao')
            print "asdm2MS done"

        add_corrected_data_column(sdfile)
        antenna = get_antenna_name(sdfile)
        # scandir = get_scan_dir(sdfile)
        frequency = get_frequency(sdfile)
        srcname = get_source_name(sdfile)
        bandnumber = get_band_number(frequency)

        if useoblongpix:
            scandirection = get_scan_dir(sdfile)
        else:
            scandirection = 'Z'

        # set cellsize to be FWHM/4 (row spacing in TPOTF.py)
        cellsize = '%.3farcsec' % (lightspeed / frequency / 12. * \
            1.22 / 4. * rad_to_arcsec)
        cell = [cellsize, cellsize]
        imsize = [32, 32]

        if scandirection is 'X':
            cell[0] = '1arcsec'
            imsize[0] = int(imsize[0] * \
                qa.quantity(cellsize, 'arcsec')['value'] / \
                qa.quantity(cell[0], 'arcsec')['value'])
        elif scandirection is 'Y':
            cell[1] = '1arcsec'
            imsize[1] = int(imsize[1] * \
                qa.quantity(cellsize, 'arcsec')['value'] / \
                qa.quantity(cell[1], 'arcsec')['value'])

        #--------------------------------
        # loop for polarizations
        #--------------------------------
        for stokes in polarizations:
            imagename = "%s_%s_B%d_%s_%s.image" % \
                (sdfile.split('.')[0], antenna, bandnumber, \
                 srcname[:3], stokes)

            if doparams:
                inp(sdtpimaging)

            sdtpimaging()

            #--------------------------------
            # open with the viewer
            #--------------------------------
            if doviewer:
                default('viewer')

                infile = imagename
                displaytype = 'raster'

                viewer()
