import matplotlib.dates
from matplotlib.ticker import MultipleLocator 
import numpy as np
from taskinit import tbtool
from analysisUtils import createCasaTool, mjdSecondsListToDateTime, snrToPhaseRmsDegrees
from analysisUtils import getAntennaNamesFromCaltable, resizeFontsOnly, mjdsecToDate
import pylab as pb

def plotcal(caltable, caltable2='', subplot=511, plotfile='', 
            antenna='DA43,DV05,DV06,DV10,DV13', labelSnr=False, errorbarFactor=1.0,
            iteration='antenna', yaxis='phase', plotrange=[0,0,-99,99], obsid='0,1,2',
            field='', spw=''):
    """
    This function was written for the Advanced Calibration chapter of the
    synthesis imaging book 2016.  It can produce a plot-cal like plot, and
    allows overlaying of phase solutions from two different temporal gain 
    tables, showing the errorbars computed from the SNR column.
    Inputs:
    caltable: caltable to be plotted in black points
    caltable2: caltabel to be plotted in red points
    subplot: rows,columns,1
    plotfile: name of plotfile to produce
    antenna: comma-delimited string list of antenna names or IDs
    labelSnr: if True, then write the SNR value for each point (at -80 deg phase)
    errorbarFactor:  error bars will be +- this value times Eq. 7-51, 7-52 of 
                     the Synthesis Imaging book (volume I)
    iteration: only 'antenna' is supported (i.e. 1 panel per antenna)
    yaxis: only 'phase' is supported
    plotrange: 4 values, only the latter 2 are used (to set y-axis limits)
    obsid: a comma-delimited string list of obs IDs to select
    field: currently not supported
    spw: currently not supported
    -Todd Hunter
    """
    if (iteration!='antenna'):
        print "Only iteration='antenna' is supported"
        return
    if (yaxis != 'phase'):
        print "Only yaxis='phase' is supported"
        return
    if (field != ''):
        print "Only field='' is supported"
        return
    if (spw != ''):
        print "Only spw='' is supported"
        return
    obsid = [int(i) for i in obsid.split(',')]
    mytb = createCasaTool(tbtool)
    antennaNames = getAntennaNamesFromCaltable(caltable)
    mytb.open(caltable)
    gains = mytb.getcol('CPARAM')
    phase = np.degrees(np.arctan2(np.imag(gains),np.real(gains)))
    snr = mytb.getcol('SNR')
    mjdsec = mytb.getcol('TIME')
    fields = mytb.getcol('FIELD_ID')
    flags = mytb.getcol('FLAG')
    spws = mytb.getcol('SPECTRAL_WINDOW_ID')
    obsids = mytb.getcol('OBSERVATION_ID')
    antenna1 = mytb.getcol('ANTENNA1')
    uniqueObsids = np.unique(obsids)
    if (len(obsid) == 0):
        obsid = uniqueObsids
    uniqueAntennas = np.unique(antenna1)
    mytb.close()
    if (caltable2 != ''):
        antennaNames2 = getAntennaNamesFromCaltable(caltable2)
        mytb.open(caltable2)
        gains = mytb.getcol('CPARAM')
        phase2 = np.degrees(np.arctan2(np.imag(gains),np.real(gains)))
        snr2 = mytb.getcol('SNR')
        mjdsec2 = mytb.getcol('TIME')
        fields2 = mytb.getcol('FIELD_ID')
        flags2 = mytb.getcol('FLAG')
        spws2 = mytb.getcol('SPECTRAL_WINDOW_ID')
        obsids2 = mytb.getcol('OBSERVATION_ID')
        antenna1_2 = mytb.getcol('ANTENNA1')
        uniqueObsids2 = np.unique(obsids2)
        uniqueAntennas2 = np.unique(antenna1_2)
        if (len(uniqueObsids) != len(uniqueObsids2)):
            print "The two caltables must have the same number of obs ids"
            return
        mytb.close()

    if (antenna == ''):
        antenna = uniqueAntennas
    else:
        if (antenna.split(',')[0].isdigit()):
            antenna = [int(i) for i in antenna.split(',')]
        else:
            antenna = [list(antennaNames).index(n) for n in antenna.split(',')]
    pb.clf()
    pb.ion()
    rows = subplot/100
    cols = (subplot-rows*100)/10
    color2 = 'r'
    for row in range(rows):
        desc = pb.subplot(rows, cols, row+1)
        if (iteration=='antenna'):
            ant = antenna[row]
        idx = np.where((antenna1==ant) * np.in1d(obsids,obsid) * (flags[0][0]==0))
        list_of_date_times = mjdSecondsListToDateTime(mjdsec[idx])
        timeplot = pb.date2num(list_of_date_times)
        pb.errorbar(timeplot, phase[0][0][idx], yerr=errorbarFactor*snrToPhaseRmsDegrees(snr[0][0][idx]), 
                    ms=3, ls='None', color='k', fmt='o')
        if (labelSnr):
            for i in range(len(idx[0])):
                pb.text(timeplot[i], -80, '%.1f'%(snr[0][0][idx][i]),size=8)
        desc.xaxis_date()
        desc.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%H:%M'))
        if (row == rows-1):
            pb.xlabel('UT (HH:MM on %s)' % (mjdsecToDate(mjdsec[idx][0])))
        else:
            desc.set_xticklabels([])
        pb.ylabel('Phase (deg)', size=10)
        majorLocator = MultipleLocator(50)
        desc.yaxis.set_major_locator(majorLocator)
        resizeFontsOnly(desc,10)
        pb.title('Antenna %s' % (antennaNames[ant]), size=10)
        if (caltable2 != ''):
            pb.hold(True)
            idx = np.where((antenna1_2==ant) * np.in1d(obsids2,obsid) * (flags2[0][0]==0))
            list_of_date_times = mjdSecondsListToDateTime(mjdsec2[idx])
            timeplot = pb.date2num(list_of_date_times)
            pb.errorbar(timeplot, phase2[0][0][idx], yerr=errorbarFactor*snrToPhaseRmsDegrees(snr2[0][0][idx]), 
                        ms=3, ls='None', color=color2, fmt='o')
            if (row == rows-1):
                pb.text(0.34,0.70,caltable,transform=desc.transAxes,color='k', ha='center')
                pb.text(0.34,0.45,caltable2,transform=desc.transAxes,color=color2, ha='center')
        if (plotrange[2:] != [0,0]):
            pb.ylim(plotrange[2:])
        x0,x1 = pb.xlim()
        x0 -= (x1-x0)*0.03
        x1 += (x1-x0)*0.03
        pb.xlim([x0,x1])
    pb.subplots_adjust(hspace=0.30)
    pb.draw()
    if plotfile != '':
        if plotfile == True:
            plotfile = caltable + '.png'
        pb.savefig(plotfile)
