"""
A script to plot interferometric phase along with TMC monitor data.

Usage:
  - put this script (and MS dataset) in a directory
  - launch casapy
  - execfile('phase_vs_llc.py') in casapy prompt
  - plot_llc_phase(msfile='uid...', ant=['DV01', 'PM03'])
  - plot_monitor_data(ant='DV01', date='2010-04-24')

2010-05-03 tsawada/ahales

$Id: phase_vs_llc.py,v 1.3 2010/05/13 22:02:16 tsawada Exp $
"""
import os
import datetime
import tmUtils as tU

def plot_llc_phase(msfile='', ant=[], trange=[], vffile=['', ''], \
    cntrfile=['', '']):
    """
    Plot time series of interferometric phase and LLC counts/voltage.

    "execfile" this file in CASA prompt and type, e.g.,
    plot_llc_phase(msfile='uid___A002_X7f55d_X1.ms',
        ant=['DV01', 'PM03'],
        trange=['2010-04-20T02:00:00', '2010-04-20T03:00:00'],
        vffile=['DV01VF.txt', 'PM03VF.txt'],
        cntrfile=['DV01CNTR.txt', 'PM03CNTR.txt'])
    All parameters other than msfile are optional.
    If any of vffile and cntrfile name is blank, the data are automatically
    retrieved, read, and removed.
    """

    ### antenna IDs for which the phase is plotted
    antennanames = get_antenna_names(msfile)
    print 'The following antennas are found in MS: %s.' % \
        (', '.join(antennanames))

    if len(ant) == 2:
        want0 = pl.where(antennanames == ant[0])[0]
        want1 = pl.where(antennanames == ant[1])[0]
        if len(want0)==0 or len(want1)==0:
            raise Exception, \
                'Given antenna(s) were not found in ANTENNA table.'
        antids = [want0[0], want1[0]]
    else:
        antids = [0, 1]
        ant = [antennanames[0], antennanames[1]]
        print "Variable 'ant' is not given or is invalid. " + \
            "%s and %s are used." % (ant[0], ant[1])
    
    # select full-band data (hopefully, not quite sure)
    #   assuming DATA_DESC_ID is equivalent to Spectral Window ID and
    #   SPWs with NUM_CHAN==1 are averaged data from the correlator
    spwid = get_spw_ids_with_numchan(msfile, 1)
    if len(spwid) == 0:
        raise Exception, "could not find NUM_CHAN==1 spw"
    
    #
    # read correlator data
    #
    data = []
    for spw in spwid:
        data.append( get_correlator_data(msfile, antids[0], antids[1], spw) )
    
    for i in range(len(data)):
        data[i]['phase'] = \
            pl.arctan2(data[i]['value'].imag, data[i]['value'].real)
    
    #
    # plot phase
    #
    plf = pl.figure()

    if len(trange) == 2:
        timerange = trange
    else:
        timerange = [data[0]['datetime'][0].isoformat(), \
            data[0]['datetime'][-1].isoformat()]
    
    subpl = plf.add_subplot(3, 1, 1)
    for i in range(len(data)):
        tU.show_time_series_in_subplot(subpl, \
            data[i]['datetime'], data[i]['phase'], \
            timerange[0], timerange[1], title=msfile, \
            label=('SPW %d' % spwid[i]), ylabel='Phase [rad]')
    subpl.legend(prop = pl.matplotlib.font_manager.FontProperties(size=10))
    plf.show()
    
    #
    # read/plot TMC data
    #
    subpl = plf.add_subplot(3, 1, 2)
    for i in range(len(ant)):
        vffilei = vffile[i]
        if vffilei == '':
            tU.show_monitor_data_in_subplot(subpl, ant[i], 'LLC', \
                'VF_MON_VF_MON_0', timerange[0], timerange[1], title='', \
                yrange=[0, 5], ylabel='Fast stretcher voltage [V]', \
                label=ant[i])
        else:
            if is_mongui_data(vffilei):
                mondata = read_mongui_data(vffilei)
            else:
                mondata = tU.read_tmc_data_file(vffilei, removefile=False)
            tU.show_time_series_in_subplot(subpl, \
                mondata['datetime'], mondata['value'], \
                timerange[0], timerange[1], title='', yrange=[0, 5], \
                ylabel='Fast stretcher voltage [V]', label=ant[i])
    subpl.legend(prop = pl.matplotlib.font_manager.FontProperties(size=10))
    plf.show()

    subpl = plf.add_subplot(3, 1, 3)
    for i in range(len(ant)):
        cntrfilei = cntrfile[i]
        if cntrfilei == '':
            tU.show_monitor_data_in_subplot(subpl, ant[i], 'LLC', \
                'CNTR_CNTR_0', timerange[0], timerange[1], title='', \
                ylabel='Fringe counts', \
                label=ant[i])
        else:
            if is_mongui_data(cntrfilei):
                mondata = read_mongui_data(cntrfilei)
            else:
                mondata = tU.read_tmc_data_file(cntrfilei, removefile=False)
            tU.show_time_series_in_subplot(subpl, \
                mondata['datetime'], mondata['value'], \
                timerange[0], timerange[1], ylabel='Fringe Count', title='', \
                label=ant[i])
    subpl.legend(prop = pl.matplotlib.font_manager.FontProperties(size=10))
    plf.show()


def plot_monitor_data(ant='', date='', vffile='', cntrfile=''):
    """
    Plot time series of LLC counts/voltage for given antenna and date.

    "execfile" this file in CASA prompt and type, e.g.,
    plot_monitor_data('DV01', '2010-04-29')

    vffile and cntrfile are optional: TMCDB data and MonitorGUI output are
    both accepted.  If omitted, TMCDB data are automatically retrieved, read,
    and removed.
    """

    plf = pl.figure()

    # Fast stretcher voltage
    subpl = plf.add_subplot(2, 1, 1)
    if vffile == '':
        tU.show_monitor_data_in_subplot(subpl, ant, 'LLC', \
            'VF_MON_VF_MON_0', date, ylabel='Fast Stretcher Voltage [V]', \
            yrange=[0, 5], title=ant)
    else:
        if is_mongui_data(vffile):
            mondata = read_mongui_data(vffile)
        else:
            mondata = tU.read_tmc_data_file(vffile, removefile=False)
        tU.show_time_series_in_subplot(subpl, \
            mondata['datetime'], mondata['value'], \
            date, ylabel='Fast Stretcher Voltage [V]', \
            yrange=[0, 5], title=ant)

    # Fringe count
    subpl = plf.add_subplot(2, 1, 2)
    if cntrfile == '':
        tU.show_monitor_data_in_subplot(subpl, ant, 'LLC', \
            'CNTR_CNTR_0', date, ylabel='Fringe Count', title='')
    else:
        if is_mongui_data(cntrfile):
            mondata = read_mongui_data(cntrfile)
        else:
            mondata = tU.read_tmc_data_file(cntrfile, removefile=False)
        tU.show_time_series_in_subplot(subpl, \
            mondata['datetime'], mondata['value'], \
            date, ylabel='Fringe Count', title='')

    plf.show()


############################################################
# aux methods

def get_spw_ids_with_numchan(msfile, numchan):
    """
    Return an array of Spectral Window IDs whose NUM_CHAN are numchan.
    """
    tb.open(msfile+'/SPECTRAL_WINDOW')
    nchan = tb.getcol('NUM_CHAN')
    tb.close()

    return pl.where(nchan==numchan)[0]


def get_datetime_from_casa_timestamp(timestamp):
    """
    Return a datetime.datetime object from CASA timestamp.
    
    timestamp should be a float value (4.7e9 or something like that).
    """
    return tU.get_datetime_from_isodatetime( \
        qa.time(('%fs' % timestamp), prec=9, form='fits') )


def get_correlator_data(msfile, ant1, ant2, spw):
    """
    Read correlator data for given pair of antennas and spw id.

    Returned values are
    {'datetime': [datetime], 'value': [value]}.
    """
    tb.open(msfile)

    antlist = [ant1, ant2]
    antlist.sort()
    q = tb.query('ANTENNA1==%d && ANTENNA2==%d && DATA_DESC_ID==%d' % \
        (antlist[0], antlist[1], spw))
    data = q.getcol('DATA').squeeze()
    timec = q.getcol('TIME_CENTROID').squeeze()

    tb.close()

    datetimelist = [get_datetime_from_casa_timestamp(timestamp) for \
        timestamp in timec]

    return {'datetime': datetimelist, 'value': data}


def get_antenna_names(msfile):
    """
    Read antenna table in given msfile.

    Returned value is an array of antenna names.
    """
    tb.open(msfile+'/ANTENNA')
    antnames = tb.getcol('NAME')
    tb.close()

    return antnames


def read_mongui_data(filename):
    """
    Read given MonitorGUI data file.

    Return is a dictionary of lists of datetime.datetime and float values:
    {'datetime': [datetime], 'value': [value]}
    """
    if not os.path.exists(filename):
        raise Exception, ("File %s is not found." % filename)

    datetimelist = []
    valuelist = []
    for line in open(filename, 'r'):
        if ('T' in line) or ('A' in line) or (len(line.strip()) == 0):
            continue

        (datetimestr, valuestr) = line.split(',')

        datetimelist.append( \
            tU.get_datetime_from_isodatetime('T'.join(datetimestr.split())) )
        valuelist.append(float(valuestr))

    return {'datetime': datetimelist, 'value': valuelist}


def is_mongui_data(filename):
    """
    Return a Boolean value if the given data file came from MonitorGUI.

    Assumes MonitorGUI files contain 'TIME' in their 1st line.
    """
    if not os.path.exists(filename):
        raise Exception, ("File %s is not found." % filename)

    f = open(filename, 'r')
    line = f.readline()
    f.close()

    return ('TIME' in line)
