#!/usr/bin/env python
"""
Produce a time-series plot of TMC monitor data as a PNG image.

Usage (see help for details):
plot_tmc.py -a DV01 -d LLC -m VF_MON_0 -s 2010-04-28 -e 2010-04-29 \
  -y 0,5

This retrieves TMCDB datasets via HTTP from tmc-services.aiv.alma.cl and
remove them automatically once the plot is made.
Requires 'tmUtils.py' to work.

tsawada 2010-05-27: added options to remove discontinuity (stepwise change) and
                    outliers
tsawada 2010-04-30

$Id: plot_tmc.py,v 1.5 2015/02/18 14:56:10 tsawada Exp $
"""

import sys
import optparse
import tmUtils as tU

PARSER = optparse.OptionParser()
PARSER.set_defaults( \
    antenna='DV01', \
    device='LLC', \
    monitorpoint='CNTR_0', \
    index=0, \
    removediscontinuity=False, \
    removeoutlier=False, \
    showplot=False)

PARSER.add_option('-a', '--antenna', dest='antenna', metavar='Antenna', \
    help='Antenna name (e.g., DV01)')
PARSER.add_option('-d', '--device', dest='device', metavar='Device', \
    help='Device name (e.g., LLC)')
PARSER.add_option('-m', '--monitor', dest='monitorpoint', \
    metavar='Monitorpoint',
    help='Monitor point name (e.g., CNTR_0)')
PARSER.add_option('-i', '--index', dest='index', \
    metavar='Index', type='int',
    help='0-based index of the monitor data (if there are multiple columns)')
PARSER.add_option('-s', '--startdate', dest='startdate', metavar='Startdate', \
    help='Start date (e.g., 2010-04-28) or datetime (2010-04-28T00:00:00)')
PARSER.add_option('-e', '--enddate', dest='enddate', metavar='Enddate', \
    help='End date (e.g., 2010-04-29) or datetime (2010-04-29T00:00:00)' + \
    ': optional')
PARSER.add_option('-y', '--yrange', dest='yrange', metavar='Yrange', \
    help='Range of Y-axis (comma-separated values; e.g., -5,5): optional')
PARSER.add_option('-D', '--removediscontinuity', \
    dest='removediscontinuity', metavar='RemoveDiscontinuity', \
    action='store_true', \
    help='Remove discontinuities (stepwise changes) in time series')
PARSER.add_option('-o', '--removeoutlier', \
    dest='removeoutlier', metavar='RemoveOutlier', \
    action='store_true', \
    help='Remove outlying data points in time series')
PARSER.add_option('-p', '--showplot', dest='showplot', metavar='Showplot', \
    action='store_true', \
    help='Keep the plot window opened until Return is hit')

(OPTIONS, ARGS) = PARSER.parse_args()

if OPTIONS.startdate == None:
    print '-s/--startdate option is required.'
    sys.exit()

if OPTIONS.yrange == None:
    yrange = None
else:
    yrangelist = OPTIONS.yrange.split(',')
    if len(yrangelist) == 2:
        yrange = [float(yrangelist[0]), float(yrangelist[1])]
    else:
        print '-y/--yrange option is in invalid format.  Ignored.'
        yrange = None
    
try:
    pngfiles = tU.plot_monitor_data_to_png(OPTIONS.antenna, OPTIONS.device, \
        OPTIONS.monitorpoint, OPTIONS.startdate, OPTIONS.enddate, yrange, \
        OPTIONS.showplot, OPTIONS.removediscontinuity, OPTIONS.removeoutlier, \
        OPTIONS.index)
except Exception, ex:
    print 'Got an exception: ', ex
    raise
else:
    print 'The plot was saved as %s' % ', '.join(pngfiles)
