#!/usr/bin/env python
"""
Calculate shadowing for antennas on AOS pads.

Usage: shadowingACApads.py [-a PM03@J505,DV01@N601] [-A DV02@N606,DV04@T701]
Maps and a shadowing diagram are shown in separate windows.
If the script is executed on the AOS STE without "-a" option, it gets the
current configuration of antennas on pads.
Tip: use "shadowingACApads.py | sort -n -k 2" to sort the baseline lengthes.

2016-12-20 tsawada: workaround to ignore "TF0X" pads
2014-11-16 tsawada: added "-p" option
2014-08-24 tsawada: adapted to new format of configuredAntennas
2013-02-27 tsawada: added "-d" option
2013-02-04 tsawada: much faster calculation of shadowing diagram
2012-04-26 tsawada: added "-x" option
2012-02-27 tsawada: added "-M" option
2011-08-13 tsawada: compatible with CM antennas (7-m diameter is assumed,
                    although the back structures are somewhat larger)
2011-07-31 tsawada: antenna height is now taken into account
                    (DA and PM 7.5 m, DV 7 m)
2011-06-05 tsawada: removed the workaround for AIV-4902
2011-05-31 tsawada: add DV13
2011-05-30 tsawada: add DV05, remove PM02
2011-05-24 tsawada: workaround for AIV-4902 (configuredAntennas not working)
2011-01-25 tsawada: replaced deprecated plotting options of matplotlib
2010-12-31 tsawada: pad positions are now read from PadData.xml
2010-12-30 tsawada: added A109 pad position
2010-08-24 tsawada: shadowing diagram now has legends
2010-07-17 tsawada: modified to show baseline lengths
2010-06-04 tsawada: added A106 pad position
2010-05-26 tsawada: added "-A" option
2010-05-21 tsawada

$Id: shadowingACApads.py,v 1.19 2017/02/08 18:36:54 tsawada Exp $
"""

import os
import sys
import re
import pylab as pl
from optparse import OptionParser

ant_diam = {"DA": 12., "DV": 12., "PM": 12., "CM": 7.}
ant_height = {"DA": 7.5, "DV": 7.0, "PM": 7.5, "CM": 7.5}

### define method(s)

def get_configured_antennas():
    """
    Returns a dictionary of {"AntennaName": "PadName", ...}.
    """
    antennas = {}

    try:
        versioninfo = os.popen('showRunningVersion').read()
    except:
        print "Could not execute 'showRunningVersion'."
        raise

    if (versioninfo.find('aos') < 0) and (versioninfo.find('ape1') < 0):
        print "WARNING: this script should be executed on AOS/APE1 STE, " + \
            "or give antennas by hand (e.g, -a PM03@J505,DV01@N601)."
        return antennas

    isantinfo = False
    for line in versioninfo.split('\n'):
        if line[:2] == '=>':
            if line.find('Configured Antennas') >= 0:
                isantinfo = True
            else:
                isantinfo = False
            continue
        if isantinfo:
            for l in line.strip().split('|'):
                words = l.strip().split()
                if len(words) > 2:
                    if words[1]=='in':
                        antennas[words[0]] = words[3]
                    elif words[1]=='->':
                        antennas[words[0]] = words[2]

    return antennas


def geoToAlma(lon, lat, geo):
    """
    Convert the geocentric coordinates into the local (horizontal) ones.
    """
    alma = [0, 0, 0]
    alma[0] = - geo[0]*pl.sin(lon) + \
                geo[1]*pl.cos(lon)
    alma[1] = - geo[0]*pl.cos(lon)*pl.sin(lat) - \
                geo[1]*pl.sin(lon)*pl.sin(lat) + \
                geo[2]*pl.cos(lat)
    alma[2] =   geo[0]*pl.cos(lon)*pl.cos(lat) + \
                geo[1]*pl.sin(lon)*pl.cos(lat) + \
                geo[2]*pl.sin(lat)
    return alma


def parse_pad_data(padfile):
    """
    Parse PadData.xml file and returns a dictionary of the coordinates.

    The positions are converted from the geocentric coordinate into the
    horizontal positions in meters.  The conversion may not be very accurate.
    The returned dictionary is {"padname": pl.array([x, y, z]), ...}.
    """
    sitelon = pl.radians(-67.75)
    sitelat = pl.radians(-23.02)

    pads = {}
    for line in open(padfile):
        (name, x, y, z) = ('', 0., 0., 0.)
        if 'name' in line:
            for field in re.sub('[<>/"]', '', line).split():
                entry = field.split('=')
                if entry[0] == 'name':
                    if re.match('^[A-Z][0-9][0-9][0-9]$', entry[1]):
                        name = entry[1]
                    elif re.match('^TF[0-9]+$', entry[1]):
                        name = 'TF%02d' % int(entry[1][2:])
                elif entry[0] == 'x':
                    x = float(entry[1])
                elif entry[0] == 'y':
                    y = float(entry[1])
                elif entry[0] == 'z':
                    z = float(entry[1])
        if name != '':
            pads[name] = pl.array(geoToAlma(sitelon, sitelat, [x, y, z]))

    # assume J501 position is (-13.25, 156.53, 30.0) in the local coordinate
    refpos = pads['J501'] - pl.array([-13.25, 156.53, 30.0])
    for key in pads.keys():
        pads[key] = pads[key] - refpos

    return pads


def get_antenna_name_on_pad(antennas, pad):
    """
    Return the antenna name on the specified pad.

    antennas: a dictionary of antennas {"AntennaName": "PadName", ...}
    pad: pad name
    """
    ant = '----'
    for key in antennas.keys():
        if antennas[key] == pad:
            ant = key
    return ant


def project_pad_position(padpos, az, el, ant):
    """
    Project pad pos (x, y, z) in meters seen from (az, el) in radians.
    """
    sinaz = pl.sin(az)
    cosaz = pl.cos(az)
    sinel = pl.sin(el)
    cosel = pl.cos(el)
    padposZ = padpos[2] + ant_height[ant[:2]]

    return (cosaz*padpos[0] - sinaz*padpos[1], \
        sinel*sinaz*padpos[0] + sinel*cosaz*padpos[1] - \
        cosel*padposZ, \
        cosel*sinaz*padpos[0] + cosel*cosaz*padpos[1] + \
        sinel*padposZ \
    )


def calc_xy_dist(pad0, pad1):
    """
    Calc projected distance on x-y plane.
    """
    return pl.sqrt((pad0[0]-pad1[0])**2 + (pad0[1]-pad1[1])**2)


def calc_z_dist(pad0, pad1):
    """
    Calc projected distance on x-y plane.
    """
    return (pad0[2]-pad1[2])


def draw_pad_map_in_subplot(subpl, pads, antennas, xlimit=None, ylimit=None,
    showemptypads=True, showbaselinelength=False):
    """
    Draw a map of pads and antennas on them.

    subpl: a pylab.subplot instance
    pads: a dictionary of antennas {"Name": (X, Y, Z), ...}
    antennas: a dictionary of antennas {"AntennaName": "PadName", ...}
    xlimit, ylimit: lists (or tuples, arrays) for the x and y axis limits.
                    if not given, automatically adjusted.
    showemptypads: set False not to draw pads and their names
    showbaselinelength: set True to display baseline length
    """
    if showemptypads:
        for pad in pads.keys():
            padpos = pads[pad]
            if pad[:1] in ['J', 'N']:
                radius = ant_diam['CM']/2.
            else:
                radius = ant_diam['PM']/2.
            circ = pl.Circle(padpos[:2], radius)
            subpl.add_artist(circ)
            circ.set_alpha(0.5)
            circ.set_facecolor([1.0, 1.0, 1.0])
            tt = subpl.text(padpos[0]-4., padpos[1]-3., pad)
            pl.setp(tt, size='small', alpha=0.5)

    (xmin, xmax, ymin, ymax) = (9e9, -9e9, 9e9, -9e9)
    for ant in antennas.keys():
        radius = ant_diam[ant[:2]]/2.
        #if not antennas[ant] in pads.keys():
        #    continue
        padpos = pads[antennas[ant]]
        circ = pl.Circle(padpos[:2], radius=radius)
        subpl.add_artist(circ)
        circ.set_alpha(1.0)
        circ.set_facecolor([0.8, 0.8, 0.8])
        subpl.text(padpos[0]-4, padpos[1], ant)
        if padpos[0] < xmin: xmin = padpos[0]
        if padpos[0] > xmax: xmax = padpos[0]
        if padpos[1] < ymin: ymin = padpos[1]
        if padpos[1] > ymax: ymax = padpos[1]

    subpl.set_xlabel('X [m]')
    subpl.set_ylabel('Y [m]')
    plotwidth = max(xmax-xmin, ymax-ymin) * 6./10. # extra 1/10 is the margin
    (xcenter, ycenter) = ((xmin+xmax)/2., (ymin+ymax)/2.)
    if xlimit == None:
        subpl.set_xlim(xcenter-plotwidth, xcenter+plotwidth)
    else:
        subpl.set_xlim(xlimit[0], xlimit[1])
    if ylimit == None:
        subpl.set_ylim(ycenter-plotwidth, ycenter+plotwidth)
    else:
        subpl.set_ylim(ylimit[0], ylimit[1])

    ants = antennas.keys()
    ants.sort()
    if showbaselinelength:
        for i in range(len(ants)-1):
            #if not antennas[ants[i]] in pads.keys():
            #    continue
            for j in range(i+1, len(ants)):
                #if not antennas[ants[j]] in pads.keys():
                #    continue
                print '  %s-%s: %.1f m' % (ants[i], ants[j], 
                    calc_xy_dist(pads[antennas[ants[i]]], 
                        pads[antennas[ants[j]]]))

    return None


def draw_shadowing_diagram_in_subplot(subpl, pads, antennas, plotant, dec):
    """
    Draw a shadowing diagram (number of shadowed pairs for given Az, El).

    subpl: a pylab.subplot instance
    pads: a dictionary of antennas {"Name": (X, Y, Z), ...}
    antennas: a dictionary of antennas {"AntennaName": "PadName", ...}
    plotant: 
    """
    contour_colors = ['blue', 'green', 'red', 'cyan', 'magenta', 'black', \
        '#996600', '#9966ff', '#009999', '#663399']
    az = pl.arange(0, 361, 1)
    el = pl.arange(0, 181, 1)/2.

    (az_deg, el_deg) = pl.meshgrid(az, el)
    (az_rad, el_rad) = (pl.radians(az_deg), pl.radians(el_deg))
    shadowcount = {}
    projected_pads = {}
    for ant in antennas.keys():
        pad = antennas[ant]
        projected_pads[pad] = project_pad_position(pads[pad],
            az_rad, el_rad, ant)

    for pad0 in projected_pads:
        ant0 = get_antenna_name_on_pad(antennas, pad0)
        ant0_diam = ant_diam[ant0[:2]]
        shadowcount[ant0] = pl.zeros((181, 361))
        for pad1 in projected_pads:
            ant1 = get_antenna_name_on_pad(antennas, pad1)
            ant1_diam = ant_diam[ant1[:2]]
            if ant0==ant1:
                continue
            shadowcount[ant0] += 1. * (
                (calc_xy_dist(projected_pads[pad0], projected_pads[pad1])
                    < (ant0_diam+ant1_diam)/2.) &
                (calc_z_dist(projected_pads[pad0], projected_pads[pad1])
                    < 0.) )

    subpl.set_xticks(pl.arange(12)*30.)
    subpl.set_yticks(pl.arange(10)*10.)
    subpl.grid()
    subpl.set_xlabel('Azimuth [deg]')
    subpl.set_ylabel('Elevation [deg]')
    for (i, ant) in enumerate(sorted(antennas.keys())):
        if len(plotant)==0:
            doplot = True
        else:
            doplot = False
            for p in plotant:
                if re.match(p, ant) != None:
                    doplot = True
                    break
        if doplot:
            subpl.contour(az_deg, el_deg, shadowcount[ant], [0.5], \
                colors=contour_colors[i % 10])
            subpl.plot([360], [90], color=contour_colors[i % 10], label=ant)
    subpl.legend(labelspacing=0, borderpad=0.1, \
        prop=pl.matplotlib.font_manager.FontProperties(size=9))

    if -90. <= dec <= 90.:
        decli = pl.radians(dec)
        sitelon = pl.radians(-67.75)
        sitelat = pl.radians(-23.02)
        ha = pl.radians(pl.arange(360.))
        objaz = ( pl.arctan2(-pl.cos(decli)*pl.sin(ha) , \
            (pl.cos(sitelat)*pl.sin(decli)-pl.sin(sitelat) * \
            pl.cos(decli)*pl.cos(ha))) ) % (2.*pl.pi)
        objel = pl.arcsin(pl.cos(pl.pi/2.-sitelat)*pl.sin(decli) + \
            pl.sin(pl.pi/2.-sitelat)*pl.cos(ha)*pl.cos(decli))
        subpl.plot(pl.degrees(objaz), pl.degrees(objel), '.')

    subpl.set_xlim(0, 360)
    subpl.set_ylim(0, 90)

    return None


### end defs

################################################################

PARSER = OptionParser()
PARSER.add_option("-a", "--antenna", dest="antennas", metavar="Antennas", \
    help="List of AntennaName@PadName separated by commas.")
PARSER.add_option("-A", "--append", dest="appendedantennas", \
    metavar="AntennasToBeAppended", \
    help="List of AntennaName@PadName separated by commas, " + \
        "which is appended to the current configuration.")
PARSER.add_option("-x", "--exclude", dest="excludeant", \
    metavar="Excluded_Antennas_or_Pads", \
    help="Comma-separated list of AntennaName or PadName to be excluded " + \
        "from the current configuration.")
PARSER.add_option("-p", "--plot", dest="plotant", \
    metavar="AntennasToBePlotted", \
    help="Comma-separated list of AntennaName (or part of it -- more " + \
        "precisely, regular expression which matches antenna names) " + \
        "to be plotted in the shadowing diagram.")
PARSER.add_option("-M", "--maponly", dest="maponly", metavar="Maponly", \
    action='store_true', default=False,
    help="Show pad/antenna map only; not shadowing diagram.")
PARSER.add_option("-d", "--declination", dest="dec", metavar="Dec", \
    type="float", default=-99.,
    help="Declination of an object (degrees) whose trajectory is " + \
        "plotted on the shadowing diagram.")

(OPTIONS, ARGS) = PARSER.parse_args()

if OPTIONS.antennas == None:
    config_antennas = get_configured_antennas()
else:
    config_antennas = {}
    for antpad in OPTIONS.antennas.split(','):
        (ant, pad) = antpad.strip().split('@')[:2]
        config_antennas[ant] = pad

if OPTIONS.appendedantennas != None:
    for antpad in OPTIONS.appendedantennas.split(','):
        (ant, pad) = antpad.strip().split('@')[:2]
        config_antennas[ant] = pad

if (len(config_antennas) > 0) and (OPTIONS.excludeant != None):
    ex_antennas = OPTIONS.excludeant.split(',')
    for ant in config_antennas.keys():
        if (ant in ex_antennas) or (config_antennas[ant] in ex_antennas):
            del(config_antennas[ant])

if len(config_antennas) < 1:
    print "No valid antennas were given."
    sys.exit()

if OPTIONS.plotant == None:
    plotant = []
else:
    plotant = OPTIONS.plotant.split(',')

alma_pads = parse_pad_data('/groups/science/PadData/PadData.xml')

for ant in config_antennas.keys():
    if not config_antennas[ant] in alma_pads.keys():
        sys.stderr.write("Position of pad %s is not given.\n" % \
            config_antennas[ant])

# map: ACA pads only
plf1 = pl.figure(1)
subpl1 = plf1.add_subplot(1, 1, 1, aspect='equal')
draw_pad_map_in_subplot(subpl1, alma_pads, config_antennas,
    xlimit=(-65, 35), ylimit=(100, 200), showbaselinelength=True)
plf1.show()

# map: the whole array
plf2 = pl.figure(2)
subpl2 = plf2.add_subplot(1, 1, 1, aspect='equal')
draw_pad_map_in_subplot(subpl2, alma_pads, config_antennas, showemptypads=False)
plf2.show()

# shadowing diagram
if not OPTIONS.maponly:
    plf3 = pl.figure(3)
    subpl3 = plf3.add_subplot(1, 1, 1)
    draw_shadowing_diagram_in_subplot(subpl3, alma_pads, config_antennas, \
        plotant, OPTIONS.dec)
    plf3.show()

sys.stderr.write('Hit RETURN to quit: ')
dummy = raw_input()
