#!/usr/bin/env python
"""
Calculate u-v coverange with antennas on AOS pads for given (HA, Dec).

Usage: uvcoverageACApads.py [-a PM03@J505,DV01@N601] [-H -2,2] [-d -40]
 -a: comma-separated list of ANTENNA@PAD.  If the script is executed on the
     AOS STE without "-a" option, it gets the current configuration.
 -A: comma-separated list of ANTENNA@PAD, which is appended to the current
     configuration (or the one given by "-a" option).
 -x: comma-separated list of ANTENNA or PAD, which is exluded from the
     current configuration defined with the above options.
 -H: hour angle range (hours, comma-separated).  If omitted, assumed to be 0.
 -d: declination of the object (degrees).

2014-09-06 tsawada: adapted to new format of configuredAntennas
2011-06-05 tsawada: removed the workaround for AIV-4902
2011-05-31 tsawada: add DV13
2011-05-30 tsawada: add DV05, remove PM02
2011-05-24 tsawada: workaround for AIV-4902 (configuredAntennas not working)
2010-12-31 tsawada: pad positions are now read from PadData.xml
2010-12-30 tsawada: added A109 pad position
2010-09-12 tsawada: added "-x" option
2010-06-04 tsawada: added A106 pad position
2010-05-26 tsawada: added "-A" option
2010-05-22 tsawada

$Id: uvcoverageACApads.py,v 1.9 2014/09/06 09:00:06 tsawada Exp $
"""

import os
import sys
import re
import pylab as pl
from optparse import OptionParser

### define method(s)

# def tmp_configured_antennas():
#     """
#     Workaround for AIV-4902
#     """
#     return [
#         'DV01 in pad A137',
#         'DV02 in pad A015',
#         'DV04 in pad J505',
#         'DV05 in pad A067',
#         'DV06 in pad T704',
#         'DV07 in pad A004',
#         'DV08 in pad A072',
#         'DV09 in pad A008',
#         'DV10 in pad A009',
#         'DV11 in pad A016',
#         'DV12 in pad A011',
#         'DV13 in pad A075',
#         'PM01 in pad T702',
#         'PM03 in pad J504',
#     ]


def get_configured_antennas():
    """
    Returns a dictionary of {"AntennaName": "PadName", ...}.
    """
    antennas = {}

    versioninfo = os.popen('showRunningVersion').read()

    if versioninfo.find('aos') < 0:
        print "WARNING: this script should be executed on AOS-STE, " + \
            "or give antennas by hand (e.g, -a PM03@J505,DV01@N601)."
        return antennas

    isantinfo = False
    for line in versioninfo.split('\n'):
        if line[:2] == '=>':
            if line.find('Configured Antennas') >= 0:
                isantinfo = True
            else:
                isantinfo = False
            continue
        if isantinfo:
            for l in line.strip().split('|'):
                words = l.strip().split()
                if len(words) > 2:
                    if words[1]=='in':
                        antennas[words[0]] = words[3]
                    elif words[1]=='->':
                        antennas[words[0]] = words[2]

    return antennas


def get_aca_pad_positions():
    """
    Returns relative ACA pad positions (X, Y, Z) in meters.

    This method is obsolete, as the positions are now read from PadData.xml.
    """
    return {
        "J501": (-13.250, 156.530, 30.0),
        "J502": (-21.250, 152.640, 30.0),
        "J503": (-14.180, 146.970, 30.0),
        "J504": ( -1.200, 151.670, 30.0),
        "J505": (-30.520, 160.830, 30.0),
        "J506": (-28.360, 146.810, 30.0),
        "J507": ( -2.170, 161.280, 30.0),
        "J508": (-14.390, 166.080, 30.0),
        "J509": (-23.240, 166.170, 30.0),
        "J510": (-23.840, 138.310, 30.0),
        "J511": (-15.990, 134.050, 30.0),
        "J512": ( -6.060, 143.360, 30.0),
        "N601": ( -2.170, 169.350, 30.0),
        "N602": (-14.390, 174.150, 30.0),
        "N603": (-23.240, 174.250, 30.0),
        "N604": (-23.840, 130.220, 30.0),
        "N605": (-13.900, 126.250, 30.0),
        "N606": ( -4.330, 135.380, 30.0),
        "T701": (-52.230, 179.550, 30.0),
        "T702": (-47.230, 120.050, 30.0),
        "T703": ( 19.270, 126.050, 29.7),
        "T704": ( 19.770, 181.550, 30.0),
        "A106": ( 87.770, 787.050, 29.4),
        "A109": (-525.230, 411.050, 26.1)
    }


def geoToAlma(lon, lat, geo):
    alma = [0, 0, 0]
    alma[0] = - geo[0]*pl.sin(lon) + \
                geo[1]*pl.cos(lon)
    alma[1] = - geo[0]*pl.cos(lon)*pl.sin(lat) - \
                geo[1]*pl.sin(lon)*pl.sin(lat) + \
                geo[2]*pl.cos(lat)
    alma[2] =   geo[0]*pl.cos(lon)*pl.cos(lat) + \
                geo[1]*pl.sin(lon)*pl.cos(lat) + \
                geo[2]*pl.sin(lat)
    return alma


def parse_pad_data(padfile):
    """
    Parse PadData.xml file and returns a dictionary of the coordinates.

    The positions are converted into the local coordinates in meters from the
    geocentric ones.  The conversion may not be very accurate.
    """
    sitelon = pl.radians(-67.75)
    sitelat = pl.radians(-23.02)

    pads = {}
    for line in open(padfile):
        (name, x, y, z) = ('', 0., 0., 0.)
        if 'name' in line:
            for field in re.sub('[<>/"]', '', line).split():
                entry = field.split('=')
                if entry[0] == 'name':
                    if re.match('^[A-Z][0-9][0-9][0-9]$', entry[1]):
                        name = entry[1]
                elif entry[0] == 'x':
                    x = float(entry[1])
                elif entry[0] == 'y':
                    y = float(entry[1])
                elif entry[0] == 'z':
                    z = float(entry[1])
        if name != '':
            pads[name] = pl.array(geoToAlma(sitelon, sitelat, [x, y, z]))

    # assume J501 position is (-13.25, 156.53, 30.0) in the local coordinate
    refpos = pads['J501'] - pl.array([-13.25, 156.53, 30.0])
    for key in pads.keys():
        pads[key] = pads[key] - refpos

    return pads


def convert_pad_positions_to_xyz(pads, sitelat):
    """
    Convert horizontal (x,y,z) coordinate into geocentric (X,Y,Z).

    Input and output are both dictionaries: {"PadName": (x, y, z), ...}
    """
    result = {}
    for key in pads.keys():
        pos = pads[key]
        result[key] = ( \
            -pos[1]*pl.sin(sitelat) + pos[2]*pl.cos(sitelat), \
            pos[0], \
            pos[1]*pl.cos(sitelat) + pos[2]*pl.sin(sitelat) \
        )

    return result

def convert_xyz_to_uvw(xyz, ha, dec):
    """
    Convert geocentric (X,Y,Z) coordinate into (u,v,w) for given (HA, Dec).

    Input and output are both tuples of 3 values.
    """
    return ( \
        pl.sin(ha)*xyz[0] + pl.cos(ha)*xyz[1], \
        -pl.sin(dec)*pl.cos(ha)*xyz[0] + pl.sin(dec)*pl.sin(ha)*xyz[1] + \
            pl.cos(dec)*xyz[2], \
        pl.cos(dec)*pl.cos(ha)*xyz[0] - pl.cos(dec)*pl.sin(ha)*xyz[1] + \
            pl.sin(dec)*xyz[2] \
    )

### end defs

################################################################

PARSER = OptionParser()
PARSER.add_option("-a", "--antenna", dest="antennas", metavar="Antennas", \
    help="Comma-separated list of AntennaName@PadName.")
PARSER.add_option("-A", "--append", dest="appendedantennas", \
    metavar="AntennasToBeAppended", \
    help="Comma-separated list of AntennaName@PadName, " + \
        "which is appended to the current configuration.")
PARSER.add_option("-x", "--exclude", dest="excludeant", \
    metavar="Excluded_Antennas_or_Pads", \
    help="Comma-separated list of AntennaName or PadName to be excluded " + \
    "from the current configuration.")
PARSER.add_option("-d", "--declination", dest="decl", metavar="Declination", \
    help="Declination of the object (in degrees)")
PARSER.add_option("-H", "--hourangle", dest="hourangle", metavar="HourAngle", \
    help="Hour angle range (comma-separated float values in hours)")

(OPTIONS, ARGS) = PARSER.parse_args()

# evaluate Hour Angle
if OPTIONS.hourangle == None:
    print "Hour angle is assumed to 0."
    halist = [0., 0.]
else:
    halist = OPTIONS.hourangle.split(',')
    if len(halist) > 1:
        halist  = [pl.radians(15.*float(item)) for item in halist[:2]]
    else:
        halist = [pl.radians(15.*float(halist[0])) for i in range(2)]

# evaluate Declination
if OPTIONS.decl == None:
    print "Declination is assumed to 0."
    decl = 0.
else:
    decl = pl.radians(float(OPTIONS.decl))

# evaluate Antennas
if OPTIONS.antennas == None:
    config_antennas = get_configured_antennas()
else:
    config_antennas = {}
    for antpad in OPTIONS.antennas.split(','):
        (ant, pad) = antpad.strip().split('@')[:2]
        config_antennas[ant] = pad

if OPTIONS.appendedantennas != None:
    for antpad in OPTIONS.appendedantennas.split(','):
        (ant, pad) = antpad.strip().split('@')[:2]
        config_antennas[ant] = pad

if (len(config_antennas) > 0) and (OPTIONS.excludeant != None):
    ex_antennas = OPTIONS.excludeant.split(',')
    for ant in config_antennas.keys():
        if (ant in ex_antennas) or (config_antennas[ant] in ex_antennas):
            del(config_antennas[ant])

if len(config_antennas) < 2:
    print "Not sufficient number of antennas was given."
    sys.exit()

# aca_pads = convert_pad_positions_to_xyz(get_aca_pad_positions(), \
#     pl.radians(-23.))
aca_pads = convert_pad_positions_to_xyz( \
    parse_pad_data('/groups/science/PadData/PadData.xml'), \
    pl.radians(-23.))

for ant in config_antennas.keys():
    if not config_antennas[ant] in aca_pads.keys():
        sys.stderr.write("Position of pad %s is not given.\n" % \
            config_antennas[ant])

# main loop
u = []
v = []
for k in range(101):
    hourangle = halist[0] + k/100.*(halist[1]-halist[0])
    uvw = {}
    for ant in config_antennas.keys():
        pos = aca_pads[config_antennas[ant]]
        uvw[ant] = convert_xyz_to_uvw(pos, hourangle, decl)

    for ant0 in uvw.keys():
        for ant1 in uvw.keys():
            if ant0 == ant1:
                continue
            u.append(uvw[ant0][0]-uvw[ant1][0])
            v.append(uvw[ant0][1]-uvw[ant1][1])

uvmax = pl.append(u, v).max()
axlim = 10.*pl.ceil(uvmax*1.2/10.)

plf = pl.figure()
subpl = plf.add_subplot(1, 1, 1, aspect='equal')
subpl.plot(u, v, '.')
subpl.set_xlabel('U [m]')
subpl.set_ylabel('V [m]')
subpl.set_xlim(-axlim, axlim)
subpl.set_ylim(-axlim, axlim)
plf.show()

dummy = raw_input('Hit RETURN to quit: ')
