# Script to address SCIREQ-114.
# This script operates on a (list of) calibrated MS that has had applycal run. For 
# real PL operations, that behavior will have to be changed to preapplying on the
# fly, within existing hif tasks. There are probably some MSIter tools that can 
# make it faster. 
#
# Must be run inside CASA.
# Parent calling script will need to import analysisUtils as aU (for a few handy functions)
import os
import shutil
import glob
import numpy as np
import time as timeUtilities
import analysisUtils as aU
import pylab as pl
from taskinit import *  # needed for things like msmdtool
import casadef
if casadef.casa_version >= '4.6.0':
    from visstat2_cli import visstat2_cli as visstat2
else:
    print "visstat2 is not available in this version of CASA, so visibilityOutliers may not work."
    from visstat_cli import visstat_cli as visstat2

BAD_ANTENNA_REASON = 'corrected_amp_outlier_too_many_bad_baselines'
BAD_BASELINE_REASON = 'corrected_amp_outlier_too_many_bad_integrations'
BAD_INTEGRATION_REASON = 'corrected_amp_outlier_quack_single_integration'

def version(short=True):
    """
    Returns the CVS revision number (as a string).
    """
    myversion = "$Id: visibilityOutliers.py,v 1.74 2017/03/06 18:47:16 thunter Exp $"
    if (short):
        myversion = myversion.split()[2]
    return myversion

def writeModificationLine(flagTemplate):
    """
    Appends a comment line to the flag template file indicating that it has 
    been modified by this task, along with version information.
    """
    f = open(flagTemplate, 'a')
    f.write('# Modified on %s by visibilityOutliers.makeFlagTemplates() version %s\n' % (aU.mjdToUT(aU.getMJD()),version()))
    f.close()

def initializeFlagTemplate(flagTemplate):
    """
    Write a new flagtemplate file, putting the standard comment lines at the 
    top, along with the version information of this task.
    """
    f = open(flagTemplate, 'w')
    f.write('# User flagging commands file for the calibration pipeline\n')
    f.write('# Created on %s by visibilityOutliers.makeFlagTemplates() version %s\n' % (aU.mjdToUT(aU.getMJD()),version()))
    f.write('#\n')
    f.write('# Examples\n')
    f.write('# Note: Do not put spaces inside the reason string !\n')
    f.write('#\n')
    f.write("# mode='manual' correlation='YY' antenna='DV01;DV08;DA43;DA48&DV23' spw='21:1920~2880' autocorr=False reason='bad_channels'\n")
    f.write("# mode='manual' spw='25:0~3;122~127' reason='stage8_2'\n")
    f.write("# mode='manual' antenna='DV07' timerange='2013/01/31/08:09:55.248~2013/01/31/08:10:01.296' reason='quack'\n")
    f.write('#\n')
    f.close()

def addBadAntennaToFlagTemplate(flagTemplate, antennaName, spw, intent, 
                                pol, poln, timestamp='', reason=BAD_ANTENNA_REASON,
                                integrationTime=1.0):
    """
    Add a line to the specified flagtemplate file.
    antennaName: string
    spw: int
    intent: string
    pol: integer
    poln: tuple that maps from 0->'XX' and 1->'YY', i.e. ['XX','YY']
    timestamp: (optional) single value in MJD seconds as a string
    reason: string (with no spaces)
    """
    if not os.path.exists(flagTemplate):
        initializeFlagTemplate(flagTemplate)
    f = open(flagTemplate, 'a')
    if (timestamp != ''):
        timeWindow = 0.5*integrationTime
        timeModifier = "timerange='%s~%s'" % (aU.mjdsecToUT(float(timestamp)-timeWindow,
                                                                prec=8,measuresToolFormat=True),
                                              aU.mjdsecToUT(float(timestamp)+timeWindow,
                                                                prec=8,measuresToolFormat=True))
    else:
        timeModifier = ''
    if pol == '':
        f.write("mode='manual' antenna='%s' spw='%d' intent='%s' %s reason='%s'\n" % (antennaName,spw,intent,timeModifier,reason))
    else:
        f.write("mode='manual' antenna='%s' spw='%d' intent='%s' correlation='%s' %s reason='%s'\n" % (antennaName,spw,intent,poln[pol],timeModifier,reason))
    f.close()
    return

def addBadEntry(badEntries, newitem):
    """
    Takes an existing list of potential entries for the flagtemplate file and add a new item,
    consolidating polarizations into a single entry where both are set to be flagged.
    Entries are dictionaries.
    newitem: a dictionary of format: {'name': antenna_selection, 'spw': spw, 'intent': intent, 'pol': corr}
                       with 'timestamp' and 'integrationTime' being optional keys
             antenna_selection is either a single antenna or single baseline
    Returns: a new, modified list of all the entries
    """
    newlist = []
    placed = False
    for entry in badEntries:
        if (newitem['name']==entry['name'] and newitem['spw']==entry['spw'] and newitem['intent']==entry['intent']):
            if 'timestamp' in newitem.keys() and 'timestamp' in entry.keys():
                if (newitem['timestamp'] == entry['timestamp']):
                    placed = True
                    # modify the existing entry
                    print "Consolidating time-stamped polarization flags for %s spw %d %s %s" % (entry['name'], entry['spw'], entry['intent'], entry['timestamp'])
                    newlist.append({'name': newitem['name'], 'spw': newitem['spw'], 'intent': newitem['intent'], 'pol': '', 'timestamp': newitem['timestamp'], 'integrationTime': newitem['integrationTime']})
                else: # keep the existing entry
                    newlist.append(entry)
            else:
                placed = True
                # modify the existing entry
                print "Consolidating polarization flags for %s spw %d %s" % (entry['name'], entry['spw'], entry['intent'])
                newlist.append({'name': newitem['name'], 'spw': newitem['spw'], 'intent': newitem['intent'], 'pol': ''})
        else: # keep the existing entry
            newlist.append(entry)
    if not placed:
        if 'timestamp' in newitem.keys():
            newlist.append({'name': newitem['name'], 'spw': newitem['spw'], 'intent': newitem['intent'], 'pol': newitem['pol'], 'timestamp':newitem['timestamp'], 'integrationTime':newitem['integrationTime']})
        else:
            newlist.append({'name': newitem['name'], 'spw': newitem['spw'], 'intent': newitem['intent'], 'pol': newitem['pol']})
    return newlist
        
def addBadBaselineToFlagTemplate(flagTemplate, baselineName, spw, intent, pol, poln):
    """
    Add a line to the specified flagtemplate file.
    spw: integer
    intent: string
    pol: integer
    poln: tuple that maps from 0->'XX' and 1->'YY', i.e. ['XX','YY']
    """
    if not os.path.exists(flagTemplate):
        initializeFlagTemplate(flagTemplate)
    f = open(flagTemplate, 'a')
    if pol == '':
        f.write("mode='manual' antenna='%s' spw='%d' intent='%s' reason='%s'\n" % (baselineName,spw,intent,BAD_BASELINE_REASON))
    else:
        f.write("mode='manual' antenna='%s' spw='%d' intent='%s' correlation='%s' reason='%s'\n" % (baselineName,spw,intent,poln[pol],BAD_BASELINE_REASON))
    f.close()
    return

def baselineName(this_bad_bl, antennaNames):
    """
    Builds the baseline antenna name string from baseline antenna ID string
    this_bad_bl: string like '6&22'
    antennaNames: an array resulting from a call to msmd.antennanames()
    Returns: a string, e.g. 'DA47&DV25'
    """
    return "%s&%s" % (antennaNames[int(this_bad_bl.split('&')[0])],
                      antennaNames[int(this_bad_bl.split('&')[1])])

def makeFileRoot(vis, field, intent, spw, poln, metric, flag_nsigma, doChannelAveragingInPython=False, outdir=None):
    """
    Generates the root name of a png to be created for the plot of the specified
    data selection and metric information.
    vis: will generally include a full absolute path
    Note: It intentionally leaves out the dot between poln and metric so that either script 
    (flagoutliers2pol.py and flagoutliers.py) will only find pngs created by itself
    """
    fileroot = vis+".field"+str(field)+"."+intent+".spw"+str(spw)+"."+poln+metric+"."+str(flag_nsigma)+"sig.flagged"
    if outdir != '' and outdir != None:
        if not os.path.exists(outdir+'/visibilityOutliers'):
            print "Running os.makedirs(%s)" % (outdir+'/visibilityOutliers')
            os.makedirs(outdir+'/visibilityOutliers')
        fileroot = outdir+'/visibilityOutliers/'+os.path.basename(fileroot)
    else:
        # fileroot = fileroot.replace('/working/','/products/')
        fileroot = fileroot.replace('/products/','/working/')  # restore data does not have a working dir
        fileroot = fileroot.replace('/working/','/working/visibilityOutliers/')
        if (fileroot.find('/working/') < 0):
            # This is the casapipescript case, where there is not working directory
            fileroot = 'visibilityOutliers/' + os.path.basename(fileroot)
        if not os.path.exists(os.path.dirname(fileroot)):
            print "Running os.makedirs(%s)" % (os.path.dirname(fileroot))
            os.makedirs(os.path.dirname(fileroot))
    if doChannelAveragingInPython:
        fileroot += '_python'
    return fileroot

def identifyOutliers(vis, mymsmd, metrics=['ratio','vectordiff','scalardiff'], 
                     intents=['*FLUX*','*BANDPASS*','*PHASE*'],
                     spws=None, uvmax=10000, flag_nsigma={'ratio':3.0, 'vectordiff':10., 'scalardiff':3.0}, 
                     tooManyBaselinesFraction=0.18, tooManyIntegrationsFraction=0.09,
                     catchNegativeOutliers={'ratio':3.7,'vectordiff':False,'scalardiff':3.7}, labelAmplitudeLevel=0.0, 
                     doChannelAveragingInPython=False, makePlots=True, verbose=False,
                     flagTemplateMetric='scalardiff', outdir=None, poln=['XX','YY'],
                     relaxationSigma=8.0, relaxationFactor=2.0, checkForAntennaBasedBadIntegrations=0.06,
                     positiveSigmaAntennaBased=5.8, scienceTargetFactor=3.0):
    """
    Input parameters:
    vis: measurement set
    mymsmd: an msmd instance for this measurement set
    metrics: a list of metric names. valid metrics are:
      "ratio":      pl.absolute(corrected / model)
      "vectordiff"  pl.absolute(corrected - model)
      "scalardiff"  pl.absolute(corrected) - pl.absolute(model)
    intents: list of intents (or comma-delimited string) to analyze
    spws: list of spws as integer IDs (default is all science spws)
    uvmax: maximum uvdistance to consider (in meters)
    flag_nsigma: dictionary indicating which sigma multiplier to use for each metric
    tooManyBaselinesFraction: if more than this fraction of baselines is bad per antenna, then flag that antenna
    tooManyIntegrationsFraction: if more than this fraction of integrations is bad on baseline, then flag that baseline
    catchNegativeOutliers: dictionary specifing the flag_nsigma value to use in the negative direction
    labelAmplitudeLevel: the relative amplitude level to write as threshold in the x-axis label
    doChannelAveragingInPython: if True, use pl.average; otherwise, use taql query to compute mean in c++
    relaxationSigma: if there are any negative outliers beyond this level, then relax the tooManyFractions
    relaxationFactor: the factor by which to raise the tooManyFractions values
    checkForAntennaBasedBadIntegrations: if >0, invokes an additional heuristic (meant for Cycle 5)
          value is the maximum fraction of individual timestamps to flag. Recommended value = 0.06
    positiveSigmaAntennaBased: consider positive values above this level as outliers, but only
        if checkForAntennaBasedBadIntegrations > 0
    scienceTargetFactor: if *TARGET* is included in intents, then use this value to scale flag_nsigma 
        and catchNegativeOutliers for this intent
    flagTemplateMetric: this is the metric whose results are used to build flag template files

    Returns: 6 items (an integer, 3 floats and two dictionaries):
    * entries (number of outliers)
    * timeSpentAveraging (seconds)
    * timeSpentPlotting (seconds)
    * timeSpentGettingIntegrationTime (seconds)
    * modelMedian dictionary: keyed by intent and spw
    * badAntennaEntries dictionary: keyed by 'scalardiff' and 'ratio'
    """
    if flagTemplateMetric not in metrics:
        print "flagTemplateMetric must be present in metrics"
        return
    timeSpentAveraging = 0
    timeSpentPlotting = 0
    if type(intents) == str:
        intents = intents.split(',')
    if spws==None:  # default to all science spws
        spws = aU.getScienceSpws(vis).split(",")
    badBaselineEntries = {'ratio':[], 'scalardiff':[]}
    badAntennaEntries = {'ratio':[], 'scalardiff':[]}
    antennaBasedBadIntegrationEntries = {'ratio':[], 'scalardiff':[]}
    plotsWithBadData = []
    cc = aU.overlayColors # ['b','g','c','y','r']
    myms = aU.createCasaTool(mstool)
    mytb = aU.createCasaTool(tbtool)
    modelMedian = {}
    for intent in intents:
        intentNoStar = intent.replace('*','')
        field = mymsmd.fieldsforintent(intent)[0]
        fieldname = mymsmd.namesforfields(field)[0]
        antennaNames = mymsmd.antennanames()
        print "intent = ", intents[0]
        scan = mymsmd.scansforintent(intents[0])
        scanString = ','.join([str(i) for i in scan])
        scanStringList = '['+scanString+']'
        # Check if BANDPASS and FLUX share the same scan, and adjust plot label accordingly
        intentLabel = intentNoStar
        sharedIntent = False
        intentFactor = 1.0
        if (intents[0].find('BANDPASS') >= 0):
            if 'CALIBRATE_FLUX#ON_SOURCE' in mymsmd.intents():
                otherscans = mymsmd.scansforintent('CALIBRATE_FLUX#ON_SOURCE')
                if any(i in otherscans for i in scan):
                    intentLabel += '=FLUX'
                    sharedIntent = True
            elif 'CALIBRATE_AMPLITUDE#ON_SOURCE' in mymsmd.intents():
                otherscans = mymsmd.scansforintent('CALIBRATE_FLUX#ON_SOURCE')
                if any(i in otherscans for i in scan):
                    intentLabel += '=FLUX'
                    sharedIntent = True
        elif (intents[0].find('FLUX') >= 0):
            if 'CALIBRATE_BANDPASS#ON_SOURCE' in mymsmd.intents():
                otherscans = mymsmd.scansforintent('CALIBRATE_BANDPASS#ON_SOURCE')
                if any(i in otherscans for i in scan):
                    intentLabel += '=BANDPASS'
                    sharedIntent = True
        elif (intents[0].find('AMPLITUDE') >= 0):
            # Support older data which had AMPLITUDE instead of FLUX
            if 'CALIBRATE_BANDPASS#ON_SOURCE' in mymsmd.intents():
                otherscans = mymsmd.scansforintent('CALIBRATE_BANDPASS#ON_SOURCE')
                if any(i in otherscans for i in scan):
                    intentLabel += '=BANDPASS'
                    sharedIntent = True
        elif (intents[0].find('TARGET') >= 0):
            intentFactor = scienceTargetFactor
            
        intentLabel += '='+fieldname
        if (len(scan) == 0):
            print "No scans with intent: ", intents[0]
            continue
        else:
            print "Found scans=%s with intent " % (scanStringList), intents[0]
        modelMedian[intent] = {}
        if checkForAntennaBasedBadIntegrations > 0:
            antennaBasedBadIntegrationEntries = {'ratio':[], 'scalardiff':[]}
        if len(scan) > 1:
            multiScan = 2
        else:
            multiScan = 1
        for spw in spws:
            startTime = timeUtilities.time()
            myscan = scan[0]
            # The following is necessary for bandwidth switching observations.
            for myscan in scan:
                if (myscan in mymsmd.scansforspw(spw)):
                    break
                print "Ignoring scan %d because it does not have spw %d" % (myscan, spw)
            integrationTime = mymsmd.exposuretime(scan=myscan, spwid=spw)['value'] 
            timeSpentGettingIntegrationTime = timeUtilities.time()-startTime 
            dd = mymsmd.datadescids(spw=spw)[0]
            spwfreq = mymsmd.meanfreq(spw)*1e-9
            myms.open(vis)
            myms.msselect({'field':str(field),'intent':str(intent),'spw':str(spw), 'scan':scanString})
            startTime = timeUtilities.time()
            if doChannelAveragingInPython:
                data = myms.getdata(['corrected_data','model_data','uvdist','antenna1','antenna2','flag'])
                myms.done()
                corrs, chans, rows = data['corrected_data'].shape
                # assume axes are corr,chan,time
                corrected=pl.zeros([corrs,rows],dtype='complex')
                model=pl.zeros([corrs,rows],dtype='complex')
                modelMedian[intent][spw] = np.median(np.absolute(model[np.where(model != 0j)]))
                flag=pl.zeros([corrs,rows],dtype='bool')
                # compute average over the unflagged channels
                for i in range(rows):
                    for corr in range(corrs): # should be 2 for 2 corrs
                        z = pl.where(data['flag'][corr,:,i]==False)
                        if len(z)>0:
                            corrected[corr,i]=pl.average(data['corrected_data'][corr,z,i])
                            model[corr,i]=pl.average(data['model_data'][corr,z,i])
                        else:
                            flag[corr,i]=True # entire spectrum is flagged
            else:
                data = myms.getdata(['uvdist','antenna1','antenna2','flag','time'])
                myms.done()
#                query = "[select means(MODEL_DATA[FLAG],1) from "+vis+" where DATA_DESC_ID=="+str(dd)+" && FIELD_ID=="+str(field)+"]"
                query = "[select means(MODEL_DATA[FLAG],1) from "+vis+" where DATA_DESC_ID=="+str(dd)+" && FIELD_ID=="+str(field)+ " && SCAN_NUMBER in "+scanStringList + "]"
                if verbose:
                    print "Running mytb.calc(%s)" % (query)
                model =     mytb.calc(query)['0']
#                modelMedian[intent][spw] = np.median(model)
                modelMedian[intent][spw] = np.median(np.absolute(model[np.where(model != 0j)]))
#                query = "[select means(CORRECTED_DATA[FLAG],1) from "+vis+" where DATA_DESC_ID=="+str(dd)+" && FIELD_ID=="+str(field)+"]"
                query = "[select means(CORRECTED_DATA[FLAG],1) from "+vis+" where DATA_DESC_ID=="+str(dd)+" && FIELD_ID=="+str(field)+ " && SCAN_NUMBER in "+scanStringList + "]"
                if verbose:
                    print "Running mytb.calc(%s)" % (query)
                result = mytb.calc(query)
                mytb.done()
                corrected = result['0']
                corrs, rows = model.shape
                flag = np.zeros([corrs,rows],dtype='bool')
                for corr in range(corrs):
                    # Maybe this can be done in one step, but I could not figure it out
                    idx = np.where(corrected[corr] == 0j)[0]
                    if (len(idx) > 0):
                        # check that it is actually flagged! because it might not be...
                        nchan = len(data['flag'][corr,:,0])
                        for i in idx:
                            if (np.sum(data['flag'][corr,:,i]) == nchan):
                                flag[corr][i] = True
                            else:
                                if verbose: print "Not ignoring value of 0j because not all %d channels are flagged. Time = %s" % (nchan,aU.mjdsecToUT(data['time'][i]))
# old method below:  
# which will mistakenly ignore cases (i.e. flag them internally) where the corrected data spectrum is all zero but not flagged
#                   flag[corr][idx] = True
#                   if verbose:
#                       print "pol%d zero-values: %d/%d" % (corr, len(np.where(flag[0] == True)[0]), len(flag[0]))
            if verbose:
                print "model: min=%f, median=%f, mean=%f, max=%f" % (np.min(model), np.median(model), np.mean(model), np.max(model))
            timeSpentAveraging += timeUtilities.time()-startTime
            ants = pl.unique(data['antenna1'])
            nant = len(ants)
            nbaselines = nant*(nant-1)/2
            for imetric in range(len(metrics)):
                metric=metrics[imetric]
                print "------------------------"
                print intent,spw,metric
                if metric=="ratio":
                    cmetric=pl.absolute(corrected/model)
                elif metric=="vectordiff":
                    cmetric=pl.absolute(corrected-model)
                elif metric=="scalardiff":
                    cmetric=pl.absolute(corrected)-pl.absolute(model)
                else:
                    print "ERROR: metric "+metric+" not known."

                # collect baselines
                bad_ant1={}
                bad_ant2={}
                bad_bl={}   # all bad baselines
                bad_bls={}  # unique bad baselines
                if makePlots:
                    bad_bls_intfraction={}  # fraction of integrations that were bad (for examining/setting new thresholds)
                legendDescriptor = {}
                nflagged = {}
                for corr in range(corrs):  # loop over 2 polarizations
                    ########### FIRST PLOT  ######################################################
                    unflagged = pl.where(flag[corr]==False)  # unflagged contains the indices of the unflagged points
                    if makePlots: 
                        startTime = timeUtilities.time()
                        pl.ion()
                        pl.clf()
                        desc = pl.subplot(111)
                        # plot all unflagged data as black points
                        pl.plot(data['uvdist'][unflagged], cmetric[corr][unflagged],'k,')
                        timeSpentPlotting += timeUtilities.time()-startTime
                    if (metric=="ratio"):
                        z0 = pl.where(data['uvdist']<uvmax)[0]
                        if makePlots: 
                            # Draw blue vertical line at uvmax considered
                            startTime = timeUtilities.time()
                            pl.plot([uvmax,uvmax],pl.ylim(),'b')
                            timeSpentPlotting += timeUtilities.time()-startTime
                    else:
                        z0=pl.where(data['uvdist']>0)[0]
#                        z0=pl.arange(rows)
                    # z0 now contains an index of points that match the uvmax criterion
                    ptsWithinUvmax = pl.where( (flag[corr][z0]==False) * (pl.isnan(cmetric[corr][z0])==False) )[0]
                    goodmetric = cmetric[corr][z0][ptsWithinUvmax]
                    med = pl.median(goodmetric)
                    rms = aU.MAD(goodmetric)
                    lowerThreshold = -relaxationSigma*rms*intentFactor + med
                    upperThreshold = positiveSigmaAntennaBased*rms*intentFactor + med
                    if makePlots: 
                        if (len(goodmetric) > 0):
                            if verbose:
                                if checkForAntennaBasedBadIntegrations:
                                    print "pol%d: len(z0)=%d, len(goodmetric)=%d, MAD=%f, rms=%f, min=%f, max=%f, threshold: <%f, >%f" % (corr, len(z0), len(goodmetric),rms,np.std(goodmetric),np.min(goodmetric),np.max(goodmetric),lowerThreshold,upperThreshold)
                                else:
                                    print "pol%d: len(z0)=%d, len(goodmetric)=%d, MAD=%f, rms=%f, min=%f, max=%f, threshold: <%f" % (corr, len(z0), len(goodmetric),rms,np.std(goodmetric),np.min(goodmetric),np.max(goodmetric),lowerThreshold)
                            ylims = (np.min(goodmetric), np.max(goodmetric))
                            pl.ylim(ylims)
                        else:
                            ylims = pl.ylim()

                    # The following equation does not take into account that some baselines could have 
                    # different numbers of timestamps after applying the uvmax criterion.  Should do a
                    # more involved calculation.
#                    timestampsToConsider = len(goodmetric) / nbaselines
                    # Here is the more involved calculation to find number of integrations per
                    # baseline that satify the uvmax criterion:
                    timestampsToConsider = {}
                    for ant1 in range(nant):
                        for ant2 in range(nant):
                            timestampsToConsider['%i&%i'%(ant1,ant2)] = len(pl.where((data['antenna1'][z0]==ant1) * (data['antenna2'][z0]==ant2))[0])
                    if makePlots:
                        # Draw green horizontal lines at positive sigmas
                        startTime = timeUtilities.time()
                        x0 = pl.xlim()[0]
                        if metric=="ratio":
                            x1 = uvmax
                        else:
                            x1 = pl.xlim()[1]
                        for nsig in list(pl.arange(1,flag_nsigma[metric]*intentFactor))+[flag_nsigma[metric]*intentFactor]:  # pl.maximum(5,flag_nsigma[metric])):
                            pl.plot([x0,x1], med+nsig*rms*pl.array([1,1]), 'g')

                        if (len(ptsWithinUvmax) > 0):
                            # make sure that highest green line is shown
                            y1 = np.max([ylims[1], med+(flag_nsigma[metric]*intentFactor+1)*rms])
                            if (catchNegativeOutliers[metric] != False):
                                y0 = pl.minimum(ylims[0], med-(catchNegativeOutliers[metric]*intentFactor+1)*rms)
                            else:
                                y0 = ylims[0]
#                            print "computed ylimits of %s over " % str([y0,y1]), [x0,x1]
                            pl.ylim(y0,y1)
                        else:
                            print "No unflagged data found"
                        timeSpentPlotting += timeUtilities.time()-startTime
                    bad_ant1[corr] = []
                    bad_ant2[corr] = []
                    bad_bl[corr] = []
                    goodmetric = cmetric[corr,z0]

                    if checkForAntennaBasedBadIntegrations > 0:
                        # Also detect positive outliers
                        highSigma = pl.where( (flag[corr,z0]==False) * (pl.isnan(goodmetric)==False) * ((goodmetric<lowerThreshold) | (goodmetric>upperThreshold)) )[0]
                    else:
                        highSigma = pl.where( (flag[corr,z0]==False) * (pl.isnan(goodmetric)==False) * (goodmetric<lowerThreshold) )[0]
                    if len(highSigma) > 0:
                        if checkForAntennaBasedBadIntegrations>0 and metric in ['scalardiff','ratio']:
                            # If all high points are concentrated to only one (or less than 10%) of the timestamp, then do not modify
                            # the thresholdScaleFactor because it is likely just one or a few bad integrations.  Otherwise, if it is
                            # more than that amount, then it is likely to be general decorrelation which we are trying to avoid flagging.
                            timestamps = data['time'][z0]
                            uniqueTimestamps = np.unique(timestamps)
                            uniqueHighTimestamps = np.unique(timestamps[highSigma])
                            # Proposed future improvement: detect time-adjacent groups of bad integrations
                            # i.e. consider adjacent bad integrations as one group;   (Cycle 6?)
                            # Q: how to find adjacent ones? 
                            # A: they will be separated by the integrationTime
                            # highTimestampGroups = 1 + len(np.where(np.diff(uniqueHighTimestamps) > integrationTime*1.5)[0])
                            # The first 'if' expression would be true in cases of heavy decorrelation, so the second expression is also 
                            # needed, but is made less restrictive in order to allow more quack-worthy integrations to be caught.
#                            if (highTimestampGroups <= np.max([1,len(uniqueTimestamps)*checkForAntennaBasedBadIntegrations]) and
#                                len(uniqueHighTimestamps) <= np.max([1,len(uniqueTimestamps)*checkForAntennaBasedBadIntegrations*2])):
                            # For now (Cycle 5), for simplicity, simply double the threshold for multi-scan calibrators
                            if (len(uniqueHighTimestamps) <= np.max([1,len(uniqueTimestamps)*checkForAntennaBasedBadIntegrations*multiScan])):
                                print "All %s points beyond [-%.1f,+%.1f] sigma arise from only 1 (or <= %d%%) of the timestamps (%d/%d)." % (poln[corr], relaxationSigma, positiveSigmaAntennaBased*intentFactor, int(round(checkForAntennaBasedBadIntegrations*100*multiScan)), len(uniqueHighTimestamps), len(uniqueTimestamps))
#                                print "All %s points beyond [-%.1f,+%.1f] sigma arise from only 1 (or <= %d%%) contiguous groups (%d/%d) and only 1 (or <=%d%%) of timestamps (%d/%d)." % (poln[corr], relaxationSigma, positiveSigmaAntennaBased, int(round(checkForAntennaBasedBadIntegrations*100)), highTimestampGroups, len(uniqueTimestamps), int(round(checkForAntennaBasedBadIntegrations*100*2)), len(uniqueHighTimestamps), len(uniqueTimestamps))
                                for highTimestamp in uniqueHighTimestamps:
                                    # For each of the few bad timestamps, determine if one antenna dominates, or if a majority are affected, and set flags accordingly
                                    idx = pl.where(timestamps[highSigma] == highTimestamp)[0]
                                    timestampAntlist = list(data['antenna1'][z0][highSigma][idx]) + list(data['antenna2'][z0][highSigma][idx])
                                    # Check if a single antenna is responsible
                                    timestampBadAnt = max(set(timestampAntlist), key=timestampAntlist.count)  # find the mode
                                    baselinesPerAntenna = len(antennaNames)-1
                                    if (timestampAntlist.count(timestampBadAnt) >= 0.5*len(highSigma[idx])):
                                        # only add bad timestamps to the flagtemplate (and the plot legend) if they do
                                        # not contain completely bad antennas
                                        print "    Antenna %2s=%s is responsible for half or more of them (%d/%d) at %s. Will flag it." % (timestampBadAnt,antennaNames[timestampBadAnt],timestampAntlist.count(timestampBadAnt),len(highSigma[idx]), aU.mjdsecToUT(highTimestamp))
                                        # bad_ants is not yet defined here, so we may have to remove this entry later
                                        antennaBasedBadIntegrationEntries[metric] = addBadEntry(antennaBasedBadIntegrationEntries[metric],
                                                               {'pol':corr, 'timestamp':highTimestamp, 'name':antennaNames[timestampBadAnt], 
                                                                'spw':spw, 'intent':intent, 'integrationTime': integrationTime})
                                    else:
                                        # Check if most antennas are responsible
                                        affectedAntennas = 0
                                        highestCountForAntenna = 0
                                        for myant in range(len(antennaNames)):
                                            if (timestampAntlist.count(myant) > highestCountForAntenna):
                                                highestCountForAntenna = timestampAntlist.count(myant)
                                        for myant in range(len(antennaNames)):
                                            if (timestampAntlist.count(myant) >= np.max([1,highestCountForAntenna-1])):
                                                affectedAntennas += 1
                                        if (affectedAntennas > len(antennaNames)/3.0):
                                            print "    A significant (>1/3) fraction (%d/%d) of the antennas are affected at %s, will flag them all." % (affectedAntennas, len(antennaNames), aU.mjdsecToUT(highTimestamp))
                                            antennaBasedBadIntegrationEntries[metric] = addBadEntry(antennaBasedBadIntegrationEntries[metric],
                                                                                                    {'pol':corr, 'timestamp':highTimestamp,'name':'', 'spw':spw, 'intent':intent, 'integrationTime':integrationTime})
                                thresholdScaleFactor = 1.0 
                                # better for regression page: 31,47,326,332,342,348,424,428,439,440,444,567,736,756,1385; 
                                # worse for page: 60,62,1292;  different but debatably neutral: 943,949,950,962,978,1287,1288,1289,1290,1294
                            else:
                                if verbose:
                                    print "Not all points arise from only 1 (or < %d%%) of the timestamps (%d/%d)" % (int(round(checkForAntennaBasedBadIntegrations*100*multiScan)), len(uniqueHighTimestamps), len(uniqueTimestamps))
                                thresholdScaleFactor = relaxationFactor
                                if verbose:
                                    print "Setting thresholdScaleFactor to %.1f for %s" % (thresholdScaleFactor, poln[corr])
                        else:
                            thresholdScaleFactor = relaxationFactor
                    else:
                        thresholdScaleFactor = 1.0
                    
                    z=pl.where( (flag[corr,z0]==False) * (pl.isnan(goodmetric)==False) * (goodmetric>(flag_nsigma[metric]*intentFactor*rms+med)) )[0]
#                    print "metric=%s, catchNegativeOutliers=%s" % (metric, catchNegativeOutliers[metric])
                    if (catchNegativeOutliers[metric] != False and metric in ['scalardiff','ratio']):
                        # Draw magenta horizontal lines at negative sigmas
                        z2 = pl.where( (flag[corr,z0]==False) * (pl.isnan(goodmetric)==False) * (goodmetric<(-catchNegativeOutliers[metric]*intentFactor*rms+med)) )[0]
                        z = pl.unique(list(z) + list(z2))  # probably do not need a unique here, but it does not hurt
                        if makePlots:
                            startTime = timeUtilities.time()
                            for nsig in list(pl.arange(1.0,catchNegativeOutliers[metric]*intentFactor))+[catchNegativeOutliers[metric]*intentFactor]: 
                                pl.plot([x0,x1],med-nsig*rms*pl.array([1,1]),'m')
                            timeSpentPlotting += timeUtilities.time()-startTime
                            if (thresholdScaleFactor > 1.0):
                                pl.plot([x0,x1], med-relaxationSigma*rms*pl.array([1,1]), 'm--')
                            if checkForAntennaBasedBadIntegrations > 0:
                                pl.plot([x0,x1], med+positiveSigmaAntennaBased*intentFactor*rms*pl.array([1,1]), 'g--')

                    # z now contains an index of points that are outliers
                        
                    for i in z:  # for each outlier point, build the name of the baseline it arose from
                        bad_ant1[corr].append(data['antenna1'][z0[i]])
                        bad_ant2[corr].append(data['antenna2'][z0[i]])
                        bad_bl[corr].append("%i&%i"%(data['antenna1'][z0[i]],data['antenna2'][z0[i]]))                

                    # New improvement = TRH 2016/10/30
                    # Don't just flag the whole baseline for one bad integration, but use a threshold.
                    # Unfortunately, our numpy version is too old to do this most efficiently:
#                    baselinesWithBadData, counts = np.unique(bad_bl[corr], return_counts=True)
                    # So instead must convert to a list and use .count(). - TRH.
                    baselinesWithBadData = np.unique(bad_bl[corr])
                    bad_bl_list = list(bad_bl[corr])
                    bad_bls[corr] = []
                    bad_bls_intfraction[corr] = []
                    for baseline in baselinesWithBadData:
                        badTimestamps = bad_bl_list.count(baseline)
                        if (badTimestamps > tooManyIntegrationsFraction*timestampsToConsider[baseline]*thresholdScaleFactor and
                            badTimestamps > 1): # the > 1 is to ensure we are not in small number statistics
                            bad_bls[corr].append(baseline)
                            bad_bls_intfraction[corr].append(100*float(badTimestamps)/timestampsToConsider[baseline])
#                            print "Marking %s because it has %d/%d bad timestamps" % (baselineName(baseline, antennaNames),badTimestamps,timestampsToConsider[baseline])
#                        else:
#                            print "Ignoring %d/%d bad timestamps on %s" % (badTimestamps,timestampsToConsider[baseline],baselineName(baseline, antennaNames))

                    # The following 9 lines have been moved here from originally further down in order to define bad_ants for usage earlier. - 2017-02-23
                    nflagged[corr] = pl.zeros(nant,dtype='int')
                    for this_bad_bl in bad_bls[corr]:
                        a,b = this_bad_bl.split("&")
                        nflagged[corr][int(a)] += 1
                        nflagged[corr][int(b)] += 1
                    zbad = pl.where(nflagged[corr]>=(tooManyBaselinesFraction*thresholdScaleFactor*(nant-1)))[0]
                    bad_ants = []
                    if (len(zbad)>0):
                        bad_ants=ants[zbad]
                    #######################################

                    if makePlots:
                        startTime = timeUtilities.time()
                        badPoints = 0
                        colorIndex = 0  
                        # mark each bad baseline in colored points
                        for i,this_bad_bl in enumerate(bad_bls[corr]):
                            this_bad_bl_name = baselineName(this_bad_bl, antennaNames)
                            zbl = bad_bl[corr].index(this_bad_bl)
                            z = pl.where( (data['antenna1']==bad_ant1[corr][zbl]) * (data['antenna2']==bad_ant2[corr][zbl]) )[0]
                            badPoints += len(z)
                            pl.plot(data['uvdist'][z],cmetric[corr,z],'o',label=this_bad_bl_name+' (%.0f%%)'%bad_bls_intfraction[corr][i],
                                    color=cc[colorIndex])
                            colorIndex += 1
                            # cycle the colors, if necessary
                            if colorIndex>=len(cc): colorIndex=0
                    
                        if checkForAntennaBasedBadIntegrations > 0:
                            if metric in antennaBasedBadIntegrationEntries.keys():
                                # mark each bad antenna-based integration in colored points
                                for this_bad_int in antennaBasedBadIntegrationEntries[metric]:
                                    if ((this_bad_int['pol']==corr or this_bad_int['pol']=='') and 
                                         this_bad_int['spw']==spw and this_bad_int['intent']==intent):
                                        if len(this_bad_int['name']) > 0:
                                            antennaID = antennaNames.index(this_bad_int['name'])
                                            z = pl.where(((data['antenna1']==antennaID) | (data['antenna2']==antennaID)) * (data['time']==this_bad_int['timestamp']))[0]
                                        else:  # most antennas show as outliers
                                            antennaID = -1
                                            z = pl.where(data['time']==this_bad_int['timestamp'])[0]
                                        if antennaID not in bad_ants:
                                            badPoints += len(z)
                                            pl.plot(data['uvdist'][z],cmetric[corr,z],'o',color=cc[colorIndex],
                                                    label=this_bad_int['name']+' '+aU.mjdsecToUT(this_bad_int['timestamp']))
                                            colorIndex += 1
                                            # cycle the colors, if necessary
                                            if colorIndex>=len(cc): colorIndex=0

                        fileroot = makeFileRoot(vis, field, intentNoStar, spw, poln[corr], metric, flag_nsigma[metric], doChannelAveragingInPython, outdir)
                        if (badPoints > 0):
                            legendDescriptor[corr] = pl.legend(loc="best",prop={"size":8},numpoints=1)
                        else:
                            legendDescriptor[corr] = None
                        pl.title(os.path.basename(vis) + ' - spw %s - %s - %s - %.0fsec' % (str(spw), poln[corr], intentLabel, integrationTime),size=12) 
                        if labelAmplitudeLevel>0.0 and metric=='ratio':
                            pl.xlabel('uvdist (m), %.1fGHz, blue line: %.0fm = amp>%.2f' % (spwfreq, uvmax, labelAmplitudeLevel)) 
                        else:
                            pl.xlabel('uvdist (m), %.1fGHz'%(spwfreq))  
                        pl.ylabel('Amplitude ' + metric + ' (scaled MAD: %.3f, npts: %d)'%(rms,len(ptsWithinUvmax)))
                        pl.savefig(fileroot+".png")

                    if makePlots:
                        ########### SECOND PLOT  ######################################################
                        pl.clf()
                        desc = pl.subplot(111)
                        # plot all unflagged data as black points
                        pl.plot(data['uvdist'][unflagged],cmetric[corr][unflagged],'k,')
                        timeSpentPlotting += timeUtilities.time()-startTime
# These 9 lines have been moved earlier in order to define bad_ants earlier.  - 2017-02-23
#                    nflagged[corr] = pl.zeros(nant,dtype='int')
#                    for this_bad_bl in bad_bls[corr]:
#                        a,b = this_bad_bl.split("&")
#                        nflagged[corr][int(a)] += 1
#                        nflagged[corr][int(b)] += 1
#                    zbad = pl.where(nflagged[corr]>=(tooManyBaselinesFraction*thresholdScaleFactor*(nant-1)))[0]
#                    bad_ants = []
#                    if (len(zbad)>0):
#                        bad_ants=ants[zbad]
                    if (len(zbad)>0):
                        if makePlots:
                            startTime = timeUtilities.time()
                            for bad_ant in bad_ants:
                                z = pl.where( (data['antenna1']==bad_ant) | (data['antenna2']==bad_ant) )[0]
                                pl.plot(data['uvdist'][z],cmetric[corr,z],'rx',lw=2)
                            timeSpentPlotting += timeUtilities.time()-startTime
                        # List the bad antennas by name (and how many baselines were bad on each one)
                        for i,antennaID in enumerate(pl.unique(bad_ants)):
                            if makePlots:
                                if (i==0 and metric==flagTemplateMetric):
                                    plotsWithBadData.append(fileroot+"_toomany.png")
                                label = 'badAnt=%s (%d/%d)'%(antennaNames[antennaID], nflagged[corr][antennaID],nant-1)
                                if i >= 25:
                                    axcenter = 0.75
                                else:
                                    axcenter = 0.5
                                pl.text(axcenter, 0.95-i*0.04, label, ha='center', transform=desc.transAxes, size=12)
                            if metric in ['scalardiff','ratio']:
                                badAntennaEntries[metric] = addBadEntry(badAntennaEntries[metric],
                                   {'name': antennaNames[antennaID], 'spw': spw, 'intent': intent, 'pol': corr})
                    badPoints = 0
                    colorIndex = 0  
                    for i,this_bad_bl in enumerate(bad_bls[corr]):
                        ant1 = int(this_bad_bl.split('&')[0])
                        ant2 = int(this_bad_bl.split('&')[1])
                        if (ant1 not in bad_ants and ant2 not in bad_ants and 
                            bad_bls_intfraction[corr][i] >= 100*np.min([1.0,relaxationFactor*tooManyIntegrationsFraction])):
                            # only add bad baselines to the flagtemplate (and the plot legend) if they do 
                            # not contain bad antennas, and exceed the relaxationFactor times the threshold 
                            # for tooManyIntegrations.  This means that not every plot showing outlier data in the
                            # first plot will have a corresponding entry in the flag template file.  It is only those 
                            # that show outlier data in the second plot.
                            this_bad_bl_name = baselineName(this_bad_bl, antennaNames)
                            if makePlots:
                                startTime = timeUtilities.time()
                                z = pl.where( (data['antenna1']==ant1) * (data['antenna2']==ant2) )[0]
                                badPoints += len(z)
                                pl.plot(data['uvdist'][z],cmetric[corr,z],'o',lw=2,label=this_bad_bl_name+' (%.0f%%)'%bad_bls_intfraction[corr][i],
                                        color=cc[colorIndex])
                                colorIndex += 1
                                # cycle the colors, if necessary
                                if colorIndex>=len(cc): colorIndex=0
                                timeSpentPlotting += timeUtilities.time()-startTime
                            if metric in ['scalardiff','ratio']:
                                badBaselineEntries[metric] = addBadEntry(badBaselineEntries[metric],
                                   {'name': this_bad_bl_name, 'spw': spw, 'intent': intent, 'pol': corr})
                    if makePlots:
                        startTime = timeUtilities.time()
                        if checkForAntennaBasedBadIntegrations > 0:
                            if metric in antennaBasedBadIntegrationEntries.keys():
                                # mark each bad antenna-based integration in colored points
                                for this_bad_int in antennaBasedBadIntegrationEntries[metric]:
                                    if ((this_bad_int['pol']==corr or this_bad_int['pol']=='') and
                                        this_bad_int['spw']==spw and this_bad_int['intent']==intent):
                                        if len(this_bad_int['name']) > 0:
                                            antennaID = antennaNames.index(this_bad_int['name'])
                                            z = pl.where(((data['antenna1']==antennaID) | (data['antenna2']==antennaID)) * (data['time']==this_bad_int['timestamp']))[0]
                                        else: # most antennas show as outliers
                                            antennaID = -1
                                            z = pl.where(data['time']==this_bad_int['timestamp'])[0]
                                        if (antennaID not in bad_ants):
                                            if verbose:
                                                print "%d not in bad_ants: " % (antennaID), bad_ants
                                            badPoints += len(z)
                                            pl.plot(data['uvdist'][z],cmetric[corr,z],'o',color=cc[colorIndex],
                                                    label=this_bad_int['name']+' '+aU.mjdsecToUT(this_bad_int['timestamp']))
                                            colorIndex += 1
                                            # cycle the colors, if necessary
                                            if colorIndex>=len(cc): colorIndex=0
                        if badPoints == 0:
                            if legendDescriptor[corr] != None:
                                legendDescriptor[corr].set_visible(False)
                        else:
                            if metric==flagTemplateMetric:
                                if fileroot+"_toomany.png" not in plotsWithBadData: plotsWithBadData.append(fileroot+"_toomany.png")
                            pl.legend(loc="best",prop={"size":8},numpoints=1)
                        if metric=="ratio":
                            # Draw blue vertical line at uvmax considered
                            pl.plot([uvmax,uvmax],pl.ylim(),'b')
                        # Draw green horizontal lines at positive sigmas
                        for nsig in list(pl.arange(1,flag_nsigma[metric]*intentFactor))+[flag_nsigma[metric]*intentFactor]:  # pl.maximum(5,flag_nsigma[metric])):
                            pl.plot([x0,x1], med+nsig*rms*pl.array([1,1]), 'g')
                        if (catchNegativeOutliers[metric] != False and metric in ['scalardiff','ratio']):
                            # Draw magenta horizontal lines at negative sigmas
                            for nsig in list(pl.arange(1.0,catchNegativeOutliers[metric]*intentFactor))+[catchNegativeOutliers[metric]*intentFactor]: 
                                pl.plot([x0,x1],med-nsig*rms*pl.array([1,1]),'m')
                            if (thresholdScaleFactor > 1.0):
                                pl.plot([x0,x1], med-relaxationSigma*rms*pl.array([1,1]), 'm--')
                            if checkForAntennaBasedBadIntegrations > 0:
                                pl.plot([x0,x1], med+positiveSigmaAntennaBased*intentFactor*rms*pl.array([1,1]), 'g--')
                        pl.title(os.path.basename(vis) + ' - spw %s - %s - %s - toomany' % (str(spw), poln[corr], intentLabel), size=12)  # TRH
                        if labelAmplitudeLevel > 0.0 and metric=='ratio':
                            pl.xlabel('uvdist (m), %.1fGHz, blue line: %.0fm = amp>%.2f' % (spwfreq, uvmax,labelAmplitudeLevel))  # TRH
                        else:
                            pl.xlabel('uvdist (m), %.1fGHz'%(spwfreq))  # TRH
                        pl.ylim(ylims[0],ylims[1])
                        pl.ylabel('Amplitude ' + metric) # TRH
                        pl.savefig(fileroot+"_toomany.png")
                        timeSpentPlotting += timeUtilities.time()-startTime
                # end 'for' loop over corr
            # end for loop over metrics
        # end 'for' loop over spws
    # end 'for' loop over intents

    # write out a new flagTemplate file, using only one of the heuristiscs
    flagTemplateMetric = 'scalardiff'
    flagTemplate = vis.replace('.ms','_flagtemplate.txt')
    flagTargetsTemplate = vis.replace('.ms','_flagtargetstemplate.txt')
    if (outdir != None):
        flagTemplate = outdir + '/' + os.path.basename(flagTemplate)
    targetEntries = 0
    for entry in badAntennaEntries[flagTemplateMetric]:
        if 'TARGET' in entry['intent']:
            myFlagTemplate = flagTargetsTemplate
            targetEntries += 1
        else:
            myFlagTemplate = flagTemplate
        addBadAntennaToFlagTemplate(myFlagTemplate, entry['name'], entry['spw'], entry['intent'], entry['pol'], poln)
            
    for entry in badBaselineEntries[flagTemplateMetric]:
        if 'TARGET' in entry['intent']:
            myFlagTemplate = flagTargetsTemplate
        else:
            myFlagTemplate = flagTemplate
        addBadBaselineToFlagTemplate(myFlagTemplate, entry['name'], entry['spw'], entry['intent'], entry['pol'], poln)

    if checkForAntennaBasedBadIntegrations > 0:  
        for entry in antennaBasedBadIntegrationEntries[flagTemplateMetric]:
            # check if whole antenna has been flagged already
            skip = False
            for aentry in badAntennaEntries[flagTemplateMetric]:
                if ((entry['name']==aentry['name']) and (entry['spw']==aentry['spw']) and (entry['intent']==aentry['intent']) 
                      and (entry['pol']==aentry['pol'] or aentry['pol'] == '')):
                    if aentry['pol'] == '':
                        flagpol = "''"
                    else:
                        flagpol = poln[int(aentry['pol'])]
                    if entry['pol'] == '':
                        newflagpol = "''"
                    else:
                        newflagpol = poln[int(entry['pol'])]

                    print "Not writing timestamp flag for antenna %s, spw %s, intent %s, pol %s, since whole antenna is already flagged for pol %s." % (entry['name'], entry['spw'], entry['intent'], newflagpol, flagpol)
                    skip = True
                    break
            if skip: continue
            if 'TARGET' in entry['intent']:
                myFlagTemplate = flagTargetsTemplate
                targetEntries += 1
            else:
                myFlagTemplate = flagTemplate
            addBadAntennaToFlagTemplate(myFlagTemplate, entry['name'], entry['spw'], entry['intent'], entry['pol'], 
                                        poln, timestamp=entry['timestamp'], reason=BAD_INTEGRATION_REASON, 
                                        integrationTime=entry['integrationTime'])

    if len(intents) > 1:
        # propagate flags to science intent if present in both flux and phase
        # this won't work in the normal case where we call this function with a single intent!
        # Need to do it at the higher level by looking at the flag template file.
        for entry in badBaselineEntries[flagTemplateMetric]:
            if (entry['intent'].find('FLUX') >= 0):
                for other in badBaselineEntries[flagTemplateMetric]:
                    print "Checking for common flag in PHASE calibrator"
                    if (entry['name']==other['name'] and entry['spw']==other['spw'] and 
                        entry['pol']==other['pol'] and other['intent'].find('PHASE') >= 0):
                        print "Propagating flag for %s spw %s pol %s to science target" % (entry['name'], str(entry['spw']), str(entry['pol']))
                        addBadBaselineToFlagTemplate(flagTemplate, entry['name'], entry['spw'], '*OBSERVE_TARGET*', 
                                                     entry['pol'], poln)
            
    entries = len(badAntennaEntries[flagTemplateMetric]) + len(badBaselineEntries[flagTemplateMetric]) 
    if checkForAntennaBasedBadIntegrations > 0:
        entries += len(antennaBasedBadIntegrationEntries[flagTemplateMetric])
    if (entries > 0):
        print "Wrote ", flagTemplate
    if (targetEntries > 0):
        print "Wrote ", flagTargetsTemplate
    return entries, timeSpentAveraging, timeSpentPlotting, timeSpentGettingIntegrationTime, modelMedian, badAntennaEntries, plotsWithBadData, sharedIntent

def getProjectIDandMOUSFromDirectoryName(mousdir):
    """
    Extract project and MOUS from directory name in either of the following 
    possible formats:
    * 2016.1.00165.S_uid___A001_X88e_X1d5
    * 2015.1.01469.S.MOUS_uid___A001_X5a4_X27a.SBNAME.PMN_J195_a_07_TE-analysis
    """
    projectID = mousdir.split('/')[0].split('_')[0]
    result = mousdir.split('/')[0].split('_products')[0].split('.S_')
    if len(result)>1:
        mous = result[1]
    else:
        result = mousdir.split('/')[0].split('_products')[0].split('.S.MOUS_')
        if len(result)>1:
            mous = result[1].split('.')[0]
        else:
            mous = 'uid___A001_Xabc_Xabc'
    return projectID, mous

def getPolarizations(vis, spw):
    """
    Returns a list of polarizations in the data, e.g.: ['XX','YY']
    """
    corrtype, polarizations, npols = getCorrType(vis, spw)
    return polarizations

def corrTypeToString(ptype):
    """
    Converts a correlation type integer (e.g. 9) to a string (e.g. 'XX')
    """
    ptypes = ['Undefined','I','Q','U','V','RR','RL','LR','LL','XX','XY','YX','YY']
    mystring = ptypes[ptype]
    return(mystring)

def getCorrType(vis, spw):
    """
    Open the DATA_DESCRIPTION_ID table.  Find the polarization_id of the first
    spw in the list of spwsToPlot, then read the CORR_TYPE from POLARIZATION
    table.
    spw: integer list of spw IDs, or comma-delimited string
    """
    if type(spw) == str:
        spw = [int(i) for i in spw.split(',')]
    elif  type(spw) == int or type(spw) == np.int32:
        spw = [spw]
    mytb = aU.createCasaTool(tbtool)
    mytb.open(vis+'/DATA_DESCRIPTION')
    spws = mytb.getcol('SPECTRAL_WINDOW_ID')
    polarization_id = mytb.getcol('POLARIZATION_ID')
    mytb.close()
    pol_id = 0
    mytb.open(vis+'/POLARIZATION')
    for myspw in spw:
        row = list(spws).index(myspw)
        if (row >= 0):
            pol_id = polarization_id[row]
            corr_type = mytb.getcell('CORR_TYPE',pol_id)
    mytb.close()
    corr_type_string = []
    if (len(corr_type) == 4):
        print "This is a 4-polarization dataset."
        if (corr_type[0] in [5,6,7,8]):
            corr_type = [5,8]
        elif (corr_type[0] in [9,10,11,12]):
            corr_type = [9,12]
        else:
            print "Unsupported polarization types = ", corr_type
            return(corr_type, corr_type_string)
    # This overrides the len(gain_table) because it can have length=2 even when only 1 pol present
    nPolarizations = len(corr_type)
    for ct in corr_type:
        corr_type_string.append(corrTypeToString(ct))
    return(corr_type, corr_type_string, nPolarizations)

def insertBookmarks(pdf, newpdf, pages=[], titles=[], levels=None, pdftk='pdftk', cleanup=True):
    """
    Inserts bookmarks into a PDF.
    pages: list of integers
    titles: list of strings
    levels: list of levels (default is all 1)
    cleanup: if True, remove the input pdf after making the new one, and the intermediate .txt files
    """
    data = '%s_metadata.txt' % pdf
    newdata = '%s_newmetadata.txt' % pdf
    os.system(pdftk+' %s dump_data > %s' % (pdf,data))
    f = open(data, 'r')
    lines = f.readlines()
    f.close()
    o = open(newdata, 'w')
    for j,line in enumerate(lines):
        o.write(line)
        if (line.find('NumberOfPages') >= 0):
            break
    for i in range(len(pages)):
        if (levels == None): 
            level = 1
        else:
            level = levels[i]
        o.write('BookmarkBegin\nBookmarkTitle: %s\nBookmarkLevel: %d\nBookmarkPageNumber: %d\n' % (titles[i], level, pages[i]))
    for line in lines[j+1:]:
        o.write(line)
    o.close()
    cmd = pdftk + ' %s update_info_utf8 %s output %s' % (pdf,newdata,newpdf)
    print "Running ", cmd
    os.system(cmd)
    if cleanup:
        os.remove(data)
        os.remove(newdata)
        os.remove(pdf)

def buildFlagTemplateFilename(vis):
    return vis.replace('.ms','_flagtemplate.txt')    

def buildFlagTargetsTemplateFilename(vis):
    return vis.replace('.ms','_flagtargetstemplate.txt')    

def propagateFlags(flagTemplate, fromIntents=['PHASE','FLUX'], toIntent='OBSERVE_TARGET', verbose=False):
    """
    Propagate the too_many_baselines flags to science intent if present in all specified calibration intents
    for same combination of baseline/spw/pol.
    fromIntents: a list of intents, e.g. ['PHASE','FLUX']
    toIntent: a single intent, e.g. 'OBSERVE_TARGET'
    """
    print "Checking for any baseline flags to propagate..."
    shutil.copyfile(flagTemplate,flagTemplate+'_pre-propagation')
    oldTemplate = open(flagTemplate,'r')
    oldTemplateLines = oldTemplate.readlines()
    oldTemplate.close()
    newTemplate = open(flagTemplate,'w')
    baselineLines = []
    for line in oldTemplateLines:
        if line.find(BAD_BASELINE_REASON) > 0 and line.find('&') > 0:
            baselineLines.append(line)
        newTemplate.write(line)
    # Now, write another line for each baseline line, if necessary
    additionalLines = 0
    for i,line in enumerate(baselineLines):
        if verbose: print "Checking baselineLine: ", line
        intent = line.split("intent='")[1].split("'")[0]
        if (len(fromIntents) == 1 and intent.find(fromIntents[0]) > 0):
            newTemplate.write(line.replace('PHASE',toIntent))
        else:
            antenna = line.split("antenna='")[1].split("'")[0]
            spw = line.split("spw='")[1].split("'")[0]
            if (line.find('correlation') < 0):
                corr = '(all)'
            else:
                corr = line.split("correlation='")[1].split("'")[0]
            for line2 in baselineLines[i:]:
                if verbose: print "    Checking against baselineLine2: ", line2
                antenna2 = line2.split("antenna='")[1].split("'")[0]
                spw2 = line2.split("spw='")[1].split("'")[0]
                intent2 = line2.split("intent='")[1].split("'")[0]
                if (line2.find('correlation') < 0):
                    corr2 = '(all)'
                else:
                    corr2 = line2.split("correlation='")[1].split("'")[0]
                if (antenna==antenna2 and spw==spw2 and corr==corr2):
                    if (len(fromIntents) == 2):
                        newReason = 'propagated_from_%s' % ('+'.join(fromIntents)).replace('*','')
                        if (intent.find(fromIntents[0])>0 and intent2.find(fromIntents[1])>0):
                            print "Propagating baseline flag %s spw %s pol %s to science target" % (antenna,spw,corr)
                            newTemplate.write(line.replace(fromIntents[0],toIntent).replace(BAD_BASELINE_REASON,newReason))
                            additionalLines += 1
                        elif (intent2.find(fromIntents[0])>0 and intent.find(fromIntents[1])>0):
                            print "Propagating baseline flag %s spw %s pol %s to science target" % (antenna,spw,corr)
                            newTemplate.write(line2.replace(fromIntents[0],toIntent).replace(BAD_BASELINE_REASON,newReason))
                            additionalLines += 1
                    else:
                        print "%d fromIntents is not supported" % (len(fromIntents))
                elif verbose:
                    print "    no match"
    os.remove(flagTemplate+'_pre-propagation')
    newTemplate.close()
    return additionalLines

def makeFlagTemplates(mousdir, metrics=['ratio','scalardiff'], 
                      intents=['BANDPASS','PHASE','FLUX'], 
                      amplitudeLevelForFirstNull=0.75, tooManyBaselinesFraction=0.18, 
                      tooManyIntegrationsFraction=0.09,
                      flag_nsigma={'ratio':3.0, 'scalardiff':3.0}, 
                      catchNegativeOutliers={'ratio':3.7, 'scalardiff': 3.7}, 
                      runSetJyIfNecessary=False, rebuildPlots=True, relabelPlots=True, 
                      buildPDF=True, showvis=[], makePlots=True,
                      doChannelAveragingInPython=False, verbose=False, flagTemplateMetric='scalardiff', 
                      outdir=None, pageOffset=0, spws='', relaxationSigma=8.0, 
                      relaxationFactor=2.0, checkForAntennaBasedBadIntegrations=0.06,
                      positiveSigmaAntennaBased=5.8, scienceTargetFactor=3.0,
                      convert='convert', pdftk='pdftk', overwrite=False):
    """
    Compares the corrected data to the model for the specified intents in all the measurement 
    sets of a calibrated MOUS in order to identify outliers in channel-averaged amplitude 
    (per intent, per spw, per pol).  Generates flagtemplate.txt files with baselined-based 
    and antenna-based flags with an option to generate summary plots of amplitude 
    differences and ratio vs. uv-distance.  Appends to existing flagtemplate.txt if overwrite
    is not specified, but also writes a backup file.  Ignores any "_target.ms" that may be
    present alongside the .ms
    Input parameters:    
    mousdir: the directory name of the MOUS
    intents: list of intents on which to operate (list, or comma-delimited string, wildcards will be added)
          valid values: 'BANDPASS', 'PHASE', 'FLUX', 'TARGET'
    amplitudeLevelForFirstNull: for ephemeris targets, only examine data down to this theoretical amplitude 
         level with respect to the peak.
    tooManyIntegrationsFraction: if more than this fraction of integrations is bad on baseline, then flag that baseline
    tooManyBaselinesFraction: if more than this fraction of baselines is bad on an antenna, then flag that antenna
    flag_nsigma: dictionary indicating which sigma multiplier to use for each metric; can also be a 
             single value which will be used to fill the dictionary of metrics
    catchNegativeOutliers: dictionary indicating the flag_nsigma value to use in the negative direction;
             can also be a single value which will be used to fill the dictionary of metrics
    runSetJyIfNecessary: only needed when operating on a restored pipeline calibration 
                         because it does not try to restore the model column
    rebuildPlots: this is only needed if you are testing bugfixes in the plot generation
    relabelPlots: this is only needed if you are testing bugfixes in the plot labeling
    labelAmplitudeLevel: the relative amplitude level to write as threshold in the x-axis label
    doChannelAveragingInPython: if True, use pl.average; otherwise, use taql query to compute mean in c++
    showvis: if specified, if the name of a measurement set does not appear in this list, it will be skipped
    makePlots: if True, then plots will be written into <mousdir>/working/visibilityOutliers/
    buildPDF: if True, then PDF will be written to the current working directory
    outdir: where to place the flagtemplate file(s) and summary pdf; 
            default: the current location of vis (e.g. the pipeline 'working' directory)
    pageOffset: value to add to the page numbers
    spws: if blank, then examine all science spws
    relaxationSigma: if there are any negative outliers beyond this level, then relax the values
              of tooManyIntegrationsFraction and tooManyBaselinesFraction (to avoid flagging
              swaths of data suffering general decorrelation)
    relaxationFactor: the factor by which to raise the tooManyFractions values
    checkForAntennaBasedBadIntegrations: if >0, invokes an additional heuristic (meant for Cycle 5)
          value is the maximum fraction of individual timestamps to flag. Recommended value = 0.06
    positiveSigmaAntennaBased: consider positive values above this level as outliers, but only
        if checkForAntennaBasedBadIntegrations > 0
    scienceTargetFactor: if *TARGET* is included in intents, then use this value to scale flag_nsigma 
        and catchNegativeOutliers for this intent
    flagTemplateMetric: this is the metric whose results are used to build flag template files
    convert: path to the ImageMagick program called 'convert' (if not found by 'which convert')
    pdftk: path to the ImageMagick program called 'pdftk' (if not found by 'which pdftk')
    overwrite: if True, then overwrite the current flagtemplate files, otherwise append

    Returns a list:
    if buildPDF: 7 values: [timeSpentAveraging, timeSpentVisstat2, timeSpentPlotting,
                            timeSpentMontaging, timeSpentGettingIntegrationTime, totalTime, flagTemplateFileList]
    else: 13 values: [montagePngs, bookmarkTitles, bookmarkPages, bookmarkLevels, pageCtr,
                      timeSpentAveraging, timeSpentVisstat2, timeSpentPlotting,
                      timeSpentMontaging, timeSpentGettingIntegrationTime, 
                      plotsWithBadData, totalTime, 
                      flagTemplateFileList]
      in both cases the first five items and final item are lists
    """
    if type(flag_nsigma) != dict:
        flag_nsigma_value = flag_nsigma
        flag_nsigma = {}
        for m in metrics:
            flag_nsigma[m] = flag_nsigma_value
    if type(catchNegativeOutliers) != dict:
        catchNegativeOutliersValue = catchNegativeOutliers
        catchNegativeOutliers = {}
        for m in metrics:
            catchNegativeOutliers[m] = catchNegativeOutliersValue
    if (type(mousdir) != str):
        print "Requested MOUS directory is not a string, it is: ", type(mousdir)
        return
    if (mousdir == './'):
        mousdir = os.getcwd()
    if (not os.path.exists(mousdir)):
        print "Could not find directory"
        return
    beginTime = timeUtilities.time()
    mymsmd = aU.createCasaTool(msmdtool)
    projectID = None
    mous = None
    if (mousdir[:2] == '20' and mousdir.find('uid') >= 0):
        projectID, mous = getProjectIDandMOUSFromDirectoryName(mousdir)
    else:
        tokens = mousdir.split('/')
#       typical tree: 2015.1.00870.S_2017_01_04T21_28_27.081/SOUS_uid___A001_X2d6_Xc1/GOUS_uid___A001_X2d6_Xc2/MOUS_uid___A001_X2d6_Xc3
        for token in tokens:
            if (token[:2] == '20'):
                if (token.find('uid') >= 0):
                    projectID, mous = getProjectIDandMOUSFromDirectoryName(token)
                else:
                    projectID = token.split('_')[0]
            elif (token[:8] == 'MOUS_uid'):
                # This will replace the timestamp e.g. "2017_01_04T21_28_27.081"
                mous = token.split('MOUS_')[1]
        if (projectID == None):
            print "Could not parse project ID from directory path."
            projectID = '0000.1.00000.S'
        if (mous == None):
            print "Could not parse MOUS name from directory path"
            mous = 'uid___A001_Xabc_Xabc'
    print "project ID: %s,  MOUS: %s" % (projectID, mous)
    if (type(intents) == str):
        intents = ['*%s*' % i.upper() for i in intents.split(',')]
        print "Using intents: ", intents
    # check if this is "real" working directory, as it will then contain a flux.csv file
    # (as opposed to a "fake" working directory created by a previous run of this script to hold visibilityOutliers)
    if (projectID == '0000.1.00000.S'):
        runSetjy = False
        # look for manually reducted dataset
        vislist = glob.glob(mousdir+'/uid*ms.split')
        if (len(vislist) == 0):
            # look for pipeline restored dataset
            vislist = glob.glob(mousdir+'/uid*.ms')
    else:
        if (os.path.exists(mousdir+'/working/flux.csv')):
            vislist = glob.glob(mousdir+'/working/uid*ms')
            runSetjy = False
        else:
            vislist = glob.glob(mousdir+'/products/uid*ms')
            runSetjy = runSetJyIfNecessary
        if (len(vislist)==0 and mousdir.find('working') > 0):
            # Check if someone specified the working directory rather than the MOUS, and proceed accordingly
            mousdir = mousdir.replace('/working','')
            if (os.path.exists(mousdir+'/working/flux.csv')):
                vislist = glob.glob(mousdir+'/working/uid*ms')
                runSetjy = False
        if (len(vislist) == 0):
            print "No measurement sets found. Is the MOUS directory correct?"
            return
    # skip the pipeline-created target.ms
    realvis = []
    plotsWithBadData = []
    for vis in vislist:
        if vis.find('_target.ms') < 0:
            realvis.append(vis)
    vislist = realvis
    if runSetjy:
        casacommands = glob.glob(mousdir+'/products/*casa_commands.log')
        os.chdir(mousdir+'/products')
        os.system('grep -A 4 "setjy(f" %s | grep -v "\-\-" > runsetjy.py' % os.path.basename(casacommands[0]))
        aU.fixPythonScript('runsetjy.py')
        print "Running runsetjy_fixed.py"
        execfile('runsetjy_fixed.py')
        os.chdir(cwd)
    pageCtr = 1
    bookmarkTitles = []
    bookmarkPages = []
    bookmarkLevels = []
    montagePngs = []
    timeSpentAveraging = 0
    timeSpentVisstat2 = 0
    timeSpentPlotting = 0
    timeSpentMontaging = 0
    timeSpentGettingIntegrationTime = 0
    if type(metrics) == str:
        metrics = metrics.split(',')
    flagStatementsDict = {}
    print "vislist = ", vislist
    mytb = aU.createCasaTool(tbtool)
    for ivis, vis in enumerate(vislist):
        visbase = os.path.basename(vis)
        if len(showvis) > 0:
            if (visbase not in showvis):
                if visbase.rstrip('.ms') not in showvis:
                    print "Skipping ", visbase
                    continue
        if rebuildPlots and (outdir==None or outdir=='' or outdir == os.path.dirname(vis)):
            flagTemplate = buildFlagTemplateFilename(vis)
            flagTemplateBackup = flagTemplate.replace('.txt','.bak')
            if os.path.exists(flagTemplate):
                if not os.path.exists(flagTemplateBackup):
                    if overwrite:
                        print "Renaming existing _flagtemplate.txt file and creating new one."
                        os.rename(flagTemplate, flagTemplateBackup)
                    else:
                        print "Copying existing _flagtemplate.txt file to .bak and appending to existing one."
                        shutil.copyfile(flagTemplate, flagTemplateBackup)
                        writeModificationLine(flagTemplate)
                else:
                    if overwrite:
                        print "Removing existing _flagtemplate.txt file and creating new one."
                        os.remove(flagTemplate)
                    else:
                        writeModificationLine(flagTemplate)
            else:
                print "No prior _flagtemplate.txt file exists, will write a new one."
        if ('*TARGET*' in intents or 'TARGET' in intents) and rebuildPlots and (outdir==None or outdir=='' or outdir == os.path.dirname(vis)):
            flagTargetsTemplate = buildFlagTargetsTemplateFilename(vis)
            flagTargetsTemplateBackup = flagTargetsTemplate.replace('.txt','.bak')
            if os.path.exists(flagTargetsTemplate):
                if not os.path.exists(flagTargetsTemplateBackup):
                    if overwrite:
                        print "Renaming existing _flagtargetstemplate.txt file and creating new one."
                        os.rename(flagTargetsTemplate, flagTargetsTemplateBackup)
                    else:
                        print "Copying existing _flagtargetstemplate.txt file to .bak and appending to existing one."
                        shutil.copyfile(flagTargetsTemplate, flagTargetsTemplateBackup)
                        writeModificationLine(flagTargetsTemplate)
                else:
                    if overwrite:
                        print "Removing existing _flagtargetstemplate.txt file and creating new one."
                        os.remove(flagTargetsTemplate)
                    else:
                        writeModificationLine(flagTargetsTemplate)
            else:
                print "No prior _flagtargetstemplate.txt file exists, will write a new one."
            
        bookmarkTitles.append(projectID+', '+mous.split('_')[-1]+', '+vis.split('_')[-1])
        bookmarkPages.append(pageCtr+pageOffset)
        bookmarkLevels.append(1)
        mytb.open(vis)
        colnames = mytb.colnames()
        mytb.close()
        if 'MODEL_DATA' not in colnames:
            print "This task cannot be run on datasets that lack a MODEL_DATA column. To remedy this:"
            print "If this is a manually reduced dataset, then re-calibrate it using setjy(usescratch=True)."
            print "If this is a result of casapipescript.py, then run hif_setjy(pipelinemode='automatic')."
            continue
        mymsmd.open(vis)
        nants = mymsmd.nantennas()
        scienceSpws = aU.getScienceSpws(vis, mymsmd=mymsmd)
        scienceSpwsList = [int(i) for i in scienceSpws.split(',')]
        if spws == '':
            spws = scienceSpws
            spwslist = [int(j) for j in spws.split(',')]
        elif type(spws) == str:
            spwslist = [int(j) for j in spws.split(',')]
        elif type(spws) != list:
            spwslist = [spws]
            spws = str(spws)
        elif type(spws) == list or type(spws) == np.ndarray:
            spwslist = spws
            spws = ','.join([str(i) for i in spwslist])
        for myspw in spwslist:
            if myspw not in scienceSpwsList:
                print "spw %d not in science spws: " %(myspw), scienceSpws
                return
        print "  Running visstat2('%s', axis='uvrange', spw='%s')" % (vis,str(spws))
        startTime = timeUtilities.time()
        data = visstat2(vis, axis='uvrange', spw=spws) 
        visstatUvmax = aU.visstat2uvrange(data)[1]
        timeSpentVisstat2 += timeUtilities.time()-startTime
#        polarizations = ['XX', 'YY']  # lazy method
        polarizations = getPolarizations(vis, spws[0])
        flagStatementsDict[vis] = 0
        for intent in intents:
            if ('*' not in intent):
                intent = '*' + intent + '*'
            intentNoStar = intent.replace('*','')
            field = mymsmd.fieldsforintent(intent)[0]
            if rebuildPlots:
                uvmax = None
                if (intent.find('FLUX') >= 0):
                    ephemerisFields = aU.getEphemerisFields(vis)
                    if (len(ephemerisFields) > 0):
                        # finds uvdistance where amplitude has dropped to specified value
                        result = aU.findNullForObservedPlanet(vis, spwslist, [amplitudeLevelForFirstNull,0.20], intent=intent, verbose=False)
                        if result != None:
                            uvmax,uvmaxForYLimits = result
                            labelAmplitudeLevel = amplitudeLevelForFirstNull
                if (uvmax == None):
                    labelAmplitudeLevel = 0.0
                    uvmax = visstatUvmax
                    uvmaxForYLimits = uvmax
                    uvmaxType = 'no'
                else:
                    uvmaxType = 'intentionally limited'
                cmd = "  Running identifyOutliers with %s uvmax (%.1fm)" % (uvmaxType, uvmax)
                casalog.post(cmd)
                print cmd
                # The following command processes all spws and pols independently
                result = identifyOutliers(vis, mymsmd, metrics, [intent], spwslist, uvmax, flag_nsigma,
                                          tooManyBaselinesFraction, tooManyIntegrationsFraction,
                                          catchNegativeOutliers, labelAmplitudeLevel,
                                          doChannelAveragingInPython, makePlots, verbose, flagTemplateMetric, 
                                          outdir, polarizations, relaxationSigma, relaxationFactor,
                                          checkForAntennaBasedBadIntegrations, positiveSigmaAntennaBased) 
                flagStatementsDict[vis] += result[0]
                timeSpentAveraging += result[1]
                timeSpentPlotting += result[2]
                timeSpentGettingIntegrationTime += result[3]
                modelMedian = result[4][intent]
                badAntennaEntries = result[5]
                sharedIntent = result[7]
                if not sharedIntent or intent.find('FLUX') < 0:
                    plotsWithBadData += result[6]
                else:
                    print "Flux scan == bandpass scan; will not send duplicate plot to the PDF."
            else:
                flagStatementsDict[vis] = -1 
                badAntennaEntries = {'ratio':[], 'scalardiff':[]}

            timeStart = timeUtilities.time()
            if flagStatementsDict[vis] == 0:
                # This is simply to help us manually cull out combinations of dataset/intent 
                # that do not need any flagging from future test runs.
                bookmarkTitles.append(intentNoStar + ' (field %s) no outliers'%str(field))
            else:
                bookmarkTitles.append(intentNoStar + ' (field %s)'%str(field))
            bookmarkPages.append(pageCtr+pageOffset)
            bookmarkLevels.append(2)
            if outdir == '' or outdir==None:
                if os.path.exists('%s/working'%mousdir):
                    plotdir = '%s/working/visibilityOutliers' % (mousdir)
                else:
                    # This is the case of casapipescript
                    plotdir = '%s/visibilityOutliers' % (mousdir)
            else:
                plotdir = '%s/visibilityOutliers' % (outdir)

            print "Set plotdir = ", plotdir
            for spw in spwslist:
                bookmarkTitles.append(str(spw) + ' (page %d)' % (pageCtr+pageOffset))
                bookmarkPages.append(pageCtr+pageOffset)
                bookmarkLevels.append(3)
                for pol in polarizations:
                    pageCtr += 1
                    os.system('rm -f %s/%s.field%s.%s.spw%s.%s*_label*png' % (plotdir,visbase,str(field),intentNoStar,str(spw),pol))
                    ratio = sorted(glob.glob(plotdir+'/%s.field%s.%s.spw%s.%sratio*png'%(visbase,str(field),intentNoStar,str(spw),pol)))
                    # name will look like: uid*.ms.field0.FLUX.spw19.XXratio.5.0sig.flagged.png
                    print "ratio png = ", ratio
                    png = ratio[0].split('ratio.')[0] + '.montage.png'
                    if (rebuildPlots or relabelPlots or not os.path.exists(png)):
                        scalar = sorted(glob.glob(plotdir+'/%s.field%s.%s.spw%s.%sscalar*png'%(visbase, str(field),intentNoStar,str(spw),pol)))
                        print "  Building montage plot for %s spw %s pol %s" % (intentNoStar, str(spw), pol)
                        if 'vectordiff' in metrics:
                            vector = sorted(glob.glob(plotdir+'/%s.field%s.%s.spw%s.%svector*png'%(visbase,str(field),intentNoStar,str(spw),pol)))
                            png5 = vector[0]
                            png6 = vector[1]
                        else:
                            vector = ['','']
                            png5 = None
                            png6 = None
                        ratioLeftLabel = '%s, ratio (+%.1fsigma,-%.1fsigma)'%(projectID,flag_nsigma['ratio'],catchNegativeOutliers['ratio'])
                        if len(badAntennaEntries['ratio']) > 0:
                            ratioRightLabel = '%s, (red X = too_many_baselines)'%(mous)
                        else:
                            ratioRightLabel = '%s (%d antennas)'%(mous,nants)
                        scalardiffLeftLabel = 'scalar diff (%+.1fsigma,-%.1fsigma) model=%.3fJy'%(flag_nsigma['scalardiff'],catchNegativeOutliers['scalardiff'],modelMedian[spw])
                        if len(badAntennaEntries['scalardiff']) > 0:
                            scalardiffRightLabel = '%s, (red X = too_many_baselines)' % (projectID)
                        else:
                            scalardiffRightLabel = '%s, %s (%d antennas)' % (projectID,mous,nants)
                        if 'vectordiff' in flag_nsigma:
                            vectordiffLeftLabel = 'vector diff (+%.1fsigma)' % (flag_nsigma['vectordiff'])
                            vectordiffRightLabel = '(red X = too_many_baselines)'
                        else:
                            vectordiffLeftLabel = ''
                            vectordiffRightLabel = ''
                        if verbose:
                            print "Running au.montagePngs('%s', '%s', '%s', png3='%s', png4='%s', png5='%s', png6='%s')" % (ratio[0], ratio[1], png, scalar[0], scalar[1], vector[0], vector[1])
                        aU.montagePngs(ratio[0], ratio[1], png, png3=scalar[0], png4=scalar[1], png5=png5, png6=png6, pointsize=30,
                                       geometry='600x440+4+4', verbose=False, convert=convert,
                                       labelImages=[ratioLeftLabel, ratioRightLabel, scalardiffLeftLabel, 
                                                    scalardiffRightLabel, vectordiffLeftLabel, vectordiffRightLabel])
                    montagePngs.append(png)
            timeSpentMontaging += timeUtilities.time() - timeStart
        # end 'for' loop over intents
        mymsmd.close()
        if rebuildPlots and (outdir==None or outdir=='' or outdir == os.path.dirname(vis)):
            if os.path.exists(flagTemplate):
                # propagate flags to science intent if present in both flux and phase
                additionalLines = propagateFlags(flagTemplate, fromIntents=['PHASE','FLUX'], toIntent='OBSERVE_TARGET')
                flagStatementsDict[vis] += additionalLines
            elif os.path.exists(flagTemplateBackup):
                shutil.copyfile(flagTemplateBackup,flagTemplate)
        vis.replace('.ms','')
    # end 'for' loop over vislist
    if buildPDF and len(montagePngs)>0:
        if outdir == None or outdir=='':
            pdfdir = os.path.dirname(vislist[0])
        else:
            pdfdir = outdir
        pdf = pdfdir + '/%s_%s_visibilityAllPlots.pdf' % (projectID, mous)
        aU.buildPdfFromPngs(montagePngs,pdf)        
        if os.path.exists(pdftk) or os.system('which '+pdftk)==0:
            # build bookmarks
            newpdf = pdf.replace('.pdf','_bookmarked.pdf')
            insertBookmarks(pdf, newpdf, bookmarkPages, bookmarkTitles, bookmarkLevels, pdftk=pdftk)
            pdf = newpdf
        if len(plotsWithBadData) >= 1:
            pdf2 = pdfdir + '/%s_%s_visibilityOutliers.pdf' % (projectID, mous)
            aU.buildPdfFromPngs(plotsWithBadData, pdf2)
            print "Wrote ", pdf2
        print "Wrote ", pdf
        returnValue = [timeSpentAveraging, timeSpentVisstat2, timeSpentPlotting,
                       timeSpentMontaging, timeSpentGettingIntegrationTime]
    else:
        returnValue = [montagePngs, bookmarkTitles, bookmarkPages, bookmarkLevels, pageCtr,
                       timeSpentAveraging, timeSpentVisstat2, timeSpentPlotting,
                       timeSpentMontaging, timeSpentGettingIntegrationTime, plotsWithBadData]
    totalTime = timeUtilities.time() - beginTime
    returnValue += [totalTime]
    cmd = "timeSpentVisstat2:    %.2f sec = %.1f%% of total time: %.1f sec" % (timeSpentVisstat2, 100*timeSpentVisstat2/totalTime, totalTime)
    casalog.post(cmd)
    print cmd
    cmd = "timeSpentPlotting:    %.1f sec = %.0f%% of total time" % (timeSpentPlotting, 100*timeSpentPlotting/totalTime)
    casalog.post(cmd)
    print cmd
    if rebuildPlots:
        cmd = "timeSpentAveraging:   %.1f sec = %.0f%% of total time" % (timeSpentAveraging, 100*timeSpentAveraging/totalTime)
        print cmd
        casalog.post(cmd)
        cmd = "timeSpentMontaging:   %.1f sec = %.0f%% of total time" % (timeSpentMontaging, 100*timeSpentMontaging/totalTime)
        print cmd
        casalog.post(cmd)
#        cmd = "timeSpentGettingIntegrationTime: %.1f sec = %.0f%% of total time" % (timeSpentGettingIntegrationTime, 100*timeSpentGettingIntegrationTime/totalTime)
#        print cmd
#        casalog.post(cmd)
    cmd = 'flag statements per ms:'
    if (len(flagStatementsDict.keys()) == 0):
        cmd += ' none'
    casalog.post(cmd)
    print cmd
    for vis in flagStatementsDict:
        cmd = '%s: %d' % (os.path.basename(vis), flagStatementsDict[vis])
        casalog.post(cmd)
        print cmd
    newFlagTemplates = []
    for vis in flagStatementsDict:
        if (flagStatementsDict[vis] > 0):
            if outdir == '' or outdir == None:
                flagTemplateFile = vis.replace('.ms','_flagtemplate.txt')
            else:
                flagTemplateFile = outdir+'/'+os.path.basename(vis.replace('.ms','_flagtemplate.txt'))
            cmd = 'wrote '+flagTemplateFile
            newFlagTemplates.append(flagTemplateFile)
            casalog.post(cmd)
            print cmd

    return returnValue + [newFlagTemplates]
    
    
