# PACKAGING SCRIPT FOR QA2 (CYCLE 0/1/2/3/4).
# VERSION 4.2 (Nov 2016)
#
# $Id: QA2_Packaging_module.py,v 1.52 2016/11/10 14:34:14 dpetry Exp $

# MAIN CHANGES: 
# 1.- Deals with datasets where the original ASDMs were removed. 
# 2.- Does not assume anymore that there is one analyzed ASDM per subfolder.
# 3.- New 'append' option, to append data to an already-existing member id.
# 4.- Updated 'datafind' function to deal with the new cal format of CASA 3.4.
# 5.- Added make_tar_ticket function.
# 6.- Added new way of giving QA2 filenames (with wildcards; see line 120).
#-------
# 7.- Most WARNING messages redirected to a log file.
# 8.- Log files are concatenated.
# 9.- Folders of combined products don't need to have an MS.
#-------
# 10.- Follows symbolic links (USE WITH CAUTION! mode='hard' MAY NOT WORK!)
# 11.- Checks existence of origpath.
#-------
# 12.- Checks existence of CORRECTED_DATA column
# 13.- Gzips the calibration tables if desired (i.e., gzip_caltables = True).
#--------
# 14.- Restores the original flags before splitting the a-priori cal'd data.
# 15.- Saves flag state before the final split of the data.
# 16.- New parameter (noms). Specify whether to save (or not) the raw and calibrated ms.
#      Defaults to True (i.e., do not save them).
# 17.- Small bug in README corrected (new line after proj. code).
# 18.- Added new keyword: PIscript (i.e., the path to the "scriptforPI.py" script).
# 19.- Corrected problem with spurious packaged scripts (usually related to "__init.py" and similars). 
# 20.- Added "Tree=True" for search of combined products. Also, the uid is not added to the "scriptforPI" 
# 21.- Several updates: print function, better error reporting, new indentation (2 spaces), removed
#        assumption about several EBs analyzed in the SAME directory, removed question about removal of 
#        existing package folder, wvrgcal log added to the log files, added the "style" keyword,
#        preliminary preparation for pipeline products, a few more sanity checks.
# 22.- Added functionality to support packaging of pipeline-calibrated data (staging for the ARCs
#      and final packaging after imaging)
# 23.- Ellusive BUG related to the copy of scriptForPI.py
# 24.- Added functionality to incorporate ephemeris files in the package if present
# 25.- For pipe1 style: Fixed bug in creation on ticket.zip files (dummy scripts)
#      and included scriptForFluxCalibration.py when present
# 26.- For pipe1 style: Fixed another bug in ticket creation: excluded unwanted weblog.
#      Improved handling of scriptForPI copy. - DP
# 27.- Corrected bug related to mask corruption. - IMV
# 28.- Pipeline README - IMV
# 29.- Implemented Pipeline TP and solved problem with tar hard-linking (os.rename). Made message printing a bit less chatty
# 30.- antennapos.csv.
# 31.- Added new function "postProcess": zipping the *.flux images, and gzipping the .mask images.
# 32.- Revised handling of antennapos.csv
# 33.- ERROR -> WARNING when looking for pipeline optional files.
# 34.- Corrected bug of closed logfile during writing of PIPELINE_README.
# 35.- Introduced styles cycle4-pipe and -nopipe; accommodated tgz instead of tar.gz
# 36.- Added new pipeline products
# 37.- Added FITS file sanitation
# 38.- Accommodated weblog file names ending in *.tgz also in ticket file creation

import tarfile, zipfile, gzip
import os, sys, shutil, errno
import operator, glob
import filecmp
import itertools
from split_cli import split_cli as split  # This is the correct method.
from flagmanager_cli import flagmanager_cli as flagmanager

import analysisUtils as aU

from taskinit import gentools
tb = gentools(['tb'])[0]

join = os.path.join 
base = os.path.basename 
name_split = os.path.split 



####################################################
#         AREA OF CONFIGURATION OPTIONS            #
####################################################

# If append = ''    -> re-generate the whole 'packpath' (i.e., 
#                      remove everything before packaging).
# If append = 'group' -> append a new group_id folder.
# If append = 'member' -> append a new member_id folder
#                         to the group of highest id
# If append = 'product' -> append the data to the (already existing) 
#                          member_id with the highest id (in the group 
#                          of highest id). If the script finds more 
#                          than one member id in the data, new ids will
#                          be created in the same group.

# If mode = 'fake' -> create package folder structure with
#                     empty files (for fast debugging).
# If mode = 'copy' -> copy files.
# If mode = 'hard' -> create hard links (destination and 
#                     origin must be in the same partition).
# If mode = 'move' -> move the files and create symbolic
#                     links at the origin. NOT recommended.
# If mode = 'ticket' -> like fake, but copied the files for the JIRA ticket (and generates it). 
################################



################################
# THESE ARE USED TO FINE-TUNE THE ALGORITHM 
# AND ARE USED *ALWAYS*:
#global required_folders_CAL, required_files_QA2
#global rawdatacopy, caldatacopy, rawdatafolder, caldatafolder, caldatacopynames
#global tablefolder,tableplotsfolder,logsfolder,qafolder,fitsfolder
#global scriptfolder,QA2_subfolders, Priori_Extension
#global scriptPattern, scriptNotUiAppend, logExtensions
##################



def setDefaults(pipeline,pipemode,cycle=1):

    global required_folders_CAL, required_files_QA2
    global rawdatacopy, caldatacopy, rawdatafolder, caldatafolder, caldatacopynames
    global tablefolder,tableplotsfolder,logsfolder,qafolder,fitsfolder
    global scriptfolder,QA2_subfolders, Priori_Extension,QA2_others
    global scriptPattern, scriptNotUiAppend, logExtensions, log2Calib
  
  
    if pipeline==False:
  
###########################################
# CASE OF HUMAN QA:

# List of patterns that should appear in the names of all scripts that should be packaged:
        scriptPattern = ['scriptFor']
    
# Patterns in the names of scripts that shouldn't be modified by appending the UID:
        if cycle in [0,1,2]:
          scriptNotUiAppend = ['scriptForImaging','scriptForFluxEqualization']
        else:
          scriptNotUiAppend = ['scriptForImaging','scriptForImagingPrep']
     

# Extensions of all files that should be considered as log files:
        logExtensions = ['.log','.log.txt']
        log2Calib = ['.wvrgcal'] # File extensions that should be copied in the calibration directory.
    
# Extension of the ms with a-priori calibration. This is needed
# for the algorithm to find the different EBs.
# NOTICE THAT ALL THE DOTS IN THE EXTENSION (i.e., all the '.')
# MUST BE INCLUDED IN THIS STRING:
        Priori_Extension = '.ms.split' # (i.e., the name is uid__XXX_XXXXX_X.split.ms).
    
# List of extensions (or full names) of folders/files that are *required*
# for the packaging:
#  If not found, a WARNING (i.e., not ERROR) WILL BE WRITTEN IN THE LOG FILE.

# NOTICE that ALL calibration tables are detected and packaged, but the existence 
# of those on this list is always CHECKED: 

# 1.- Folders for the calibration-tables folder:
        required_folders_CAL = ['.ms.split.cal.ampli_inf','.ms.split.bandpass',
          '.ms.split.phase_inf','.wvr','.wvr.smooth','.tsys','.tsys.fdm','.antpos',
          '.ap_pre_bandpass','.bandpass','.phase_int','.ampli_inf','.bandpass_smooth20flat_ri',
          '.flux_inf']
    
# 2.- Files and directories *required* for the QA2 folder (add as desired):
#  If not found, a WARNING (i.e., not ERROR) WILL BE WRITTEN IN THE LOG FILE.
        required_files_QA2 = ['allFluxes.txt','qa2']
    
    
##################


###########################
# Choose different datasets to copy (as many as elements in the list):
# [] -> No data (only tables, plots, images).
# '' -> original asdm.
# '.EXTENSION' -> Any MS named like '*.EXTENSION' 

#rawdatacopy = ['','.ms.split'] # -> i.e., both, the asdm and the a-priori calibrated ms.
        rawdatacopy = ['.ms.split'] # -> i.e., the a-priori calibrated ms.
        caldatacopy = ['.ms.split.cal','.contsub'] # -> i.e., fully calibrated ms.
        caldatacopynames = ['calibrated.ms'] # -> These are not extensions, but full names.
    
# Folder names used in the packaging:

        tablefolder='calibration' # where calibration tables will be saved (in ASDM/MS basis).
        tableplotsfolder='calibration'     # where plots of the cal tables will be saved (in ASDM/MS basis).
        rawdatafolder = 'raw'  # only used if the rawdatacopy list is not empty.
        caldatafolder='calibrated' # only used if the caldatacopy list is not empty.
        logsfolder = 'log'
        qafolder = 'qa'  # where qa2-related stuff will be saved.
        fitsfolder = 'product'
        scriptfolder = 'script'
    
    
# Just in case we miss some QA2 "special" files (or folders), here is a list 
# of all the kind of QA2 files that we may find (please, use wildcards):
        QA2_others = ['*.jpg','*qa2.txt','*qa2-summ.txt','*.qa2.pdf','*textfile.txt',
         '*qa2_part*.png','*.tbl','*.html.tgz','*.html']
    
    else:
  
#########################################
# PIPELINE PARAMETERS:

        scriptPattern = ['scriptFor']
        scriptNotUiAppend = ['scriptForImaging','scriptForFluxEqualization','scriptForImagingPrep']
        logExtensions = ['.log','.log.txt']
        log2Calib = ['.wvrgcal']
    
    
        Priori_Extension = '.ms.split'
    
        required_folders_CAL = ['.ms.split.cal.ampli_inf','.ms.split.bandpass',
         '.ms.split.phase_inf','.wvr','.wvr.smooth','.tsys','.tsys.fdm','.antpos',
         '.ap_pre_bandpass','.bandpass','.phase_int','.ampli_inf','.bandpass_smooth20flat_ri',
         '.flux_inf']
    
        required_files_QA2 = ['allFluxes.txt','qa2']
    
        rawdatacopy = ['.ms.split'] # -> i.e., the a-priori calibrated ms.
        caldatacopy = ['.ms.split.cal','.contsub'] # -> i.e., fully calibrated ms.
        caldatacopynames = ['calibrated.ms'] # -> These are not extensions, but full names.
    
        tablefolder='calibration' # where calibration tables will be saved (in ASDM/MS basis).
        tableplotsfolder='calibration'     # where plots of the cal tables will be saved (in ASDM/MS basis).
        rawdatafolder = 'raw'  # only used if the rawdatacopy list is not empty.
        caldatafolder='calibrated' # only used if the caldatacopy list is not empty.
        logsfolder = 'log'
        qafolder = 'qa'  # where qa2-related stuff will be saved.
        fitsfolder = 'product'
        scriptfolder = 'script'
    
        QA2_others = ['*.jpg','*qa2.txt','*qa2-summ.txt','*.qa2.pdf','*textfile.txt',
         '*qa2_part*.png','*.tbl','*.html.tgz','*.html']
    
    
    
    
    
    
####################################################
#      END OF AREA OF CONFIGURATION OPTIONS        #
####################################################






####################################
# BEGINNING OF AUXILIARY FUNCTIONS #
####################################



###############################################################
# SPECIAL COPYING (I.E., GENERATE EMPTY FILES, HARD LINKS, OR #
# SYMBOLIC LINKS IN THE DESTINATION FOLDER)                   #
###############################################################

def special_cp(origin,destination,mode):
    """
    Copy `origin` to `destination`, by default using a hardlink,
    move-and-symlink-back, or fake-file-creation mode.
  
    Parameters
    ----------
    mode : 'hard' | 'fake' | 'move' | 'copy'
      hard: hardlink the files to the destination (will fail on cross-filesystem
            copies)
      fake: `touch` the target filename at the destination, creating an empty
            file
      move: Move the file to the destination, then symbolically link the destination
            back to the origin
      copy: Use `shutil.copy2` to copy the source to the destination.  This
            behavior is also achieved by calling `cpall`
      ticket: like copy
      
    """

    if mode=='ticket':
        mode = 'copy'
  
    if os.path.isfile(origin) or os.path.islink(origin):
        if os.path.exists(destination):
            lprint('WARNING: FILE '+base(destination)+' ALREADY EXISTS!')
        else:
            if mode == 'hard':
                try:
                    os.link(origin,destination)
                except:
                    lprint('HARD LINKING FAILED! ABORTING!')
                    lprint('hard mode only works if the QA2 directory and the package directory are',True,True)
                    lprint('in the same partition. Please, try again with mode=\'copy\'',True,True)
                    raise ValueError("hard-linking failed! Try with  mode=\'copy\'")
            elif mode == 'fake':
                open(destination,'w').close()
            elif mode == 'copy':
                shutil.copy2(origin, destination)
            elif mode == 'move':
                shutil.move(origin,destination)
                os.symlink(destination,origin)
        
    else: # origin is a directory
  
        if os.path.exists(destination):
            lprint('WARNING: FOLDER '+base(destination)+' ACCESSED MORE THAN ONCE!',False,True)
        else:
            os.makedirs(destination)
            for name in os.listdir(origin):
                currfile = join(origin,name)
                if os.path.islink(currfile) or os.path.isfile(currfile):
                    if mode=='hard':
                        try:
                            os.link(currfile,join(destination,name))
                        except:
                            lprint('HARD LINKING FAILED! ABORTING!')
                            lprint('hard mode only works if the QA2 directory and the package directory are',True,True)
                            lprint('in the same partition. Please, try again with mode=\'copy\'',True,True)
                            raise ValueError("hard-linking failed! Try with  mode=\'copy\'")
                    elif mode=='fake':
                        open(join(destination,name),'w').close()
                    elif mode == 'copy':
                        shutil.copy2(currfile, join(destination,name))
                    elif mode == 'move':
                        shutil.move(currfile,join(destination,name))
                        os.symlink(join(destination,name),currfile)
                elif os.path.isdir(currfile):
                    currdest = join(destination,name)
                    special_cp(currfile,currdest,mode)
          
          
########################################################
# COPY FILE TO DIRECTORY (OR DIRECTORY TO DIRECTORY):  #
########################################################

def cpall(origin,destination,mode):

    newpath = join(destination,base(origin))
  
    if mode in ['fake','hard','move']:
        special_cp(origin,newpath,mode)
    
    else:
  
        if os.path.isfile(origin):
            if os.path.exists(newpath):
                lprint('WARNING: FILE '+base(origin)+' IS FOUND MORE THAN ONCE!',False,True)
            else:
                shutil.copy2(origin,destination)
        else:
            if os.path.exists(newpath):
                lprint('WARNING: FOLDER '+base(origin)+' IS ACCESSED MORE THAN ONCE!',False,True)
            else:
                shutil.copytree(origin, newpath)
        
        
##################################################
# FIND ALL FILES (OR FOLDERS) MATCHING 
# A PATTERN, EXTENSION, OR NAME. RETURNS A LIST 
# WITH THEIR ABSOLUTE PATHS:
# if tree = True, the search is done recursively
##################################################

def findall(whole_folder, folder, pattern = '',extension='',name='',
            kind='',exclude_parent='',tree=True):

    if kind == 'file':
        it = 2
    elif kind == 'folder':
        it = 1
    else:
        return []
    
    lext = len(extension)
    lnam = len(name)
    lexcl = len(exclude_parent)
    found = []
  
    for step in whole_folder:
  
        if tree:
            findit = (folder in step[0])
        else:
            findit = (folder == step[0])
         
        if findit:
            paths = filter(lambda x: (pattern in x) and (lext==0 or (lext>0 and x[-lext:]==extension)) and (lnam==0 or (lnam>0 and x==name)) and (lexcl==0 or (lexcl>0 and exclude_parent not in step[0])), step[it])
            abspath = [join(step[0],path) for path in paths]
            found += abspath
      
      
    return found
  
  
  
########################################################
# RETURN A LIST OF ALL FILES (OR FOLDERS) IN A FOLDER  #
# kind IS THE TYPE OF ENTITY TO LOOK FOR. IT CAN BE    #
# EITHER 'file' OR 'folder' (IS 'file' BY DEFAULT)     #
########################################################

def lsall(whole_folder,folder, kind='file'):
 
    ls = []
    names = os.listdir(folder)
  
    for name in names:
        fullname = join(folder,name)
    if kind=='file' and os.path.isfile(fullname):
        ls.append(fullname)
    elif kind=='folder' and os.path.isdir(fullname):
        ls.append(fullname)
    
    return ls
  
  
  
##############################################################
# Return a list of all asdms, ms, or cal tables in a folder. #
# If there is no matching, the list will be empty:           #
# If tree=True, the search is done recursively               #
# if kind=='priori', return the asdms, but based on the      #
# names of the a-priori-calibrated ms (i.e., even if the     #
# original ASDMs were removed, the function would still      #
# 'find' them.                                               #
##############################################################

def datafind(whole_folder,folder,kind='asdm',tree=False):
    found = []
  
    ndots = Priori_Extension.count('.')
    nchar = len(Priori_Extension)
  
    if kind=='asdm':
        match = ['ASDMBinary'] ; notmatch=['']
    elif kind in ['ms','priori']:
        match = ['DATA_DESCRIPTION']; notmatch=['']
    elif kind=='table':
        match = ['ANTENNA','CAL_DESC'] ; notmatch = ['DATA_DESCRIPTION']
    elif kind=='img':
        match=['logtable'] ; notmatch=['']
    
    
    if tree:
        folder_walk = filter(lambda x: folder in name_split(x[0])[0],whole_folder)
    else:
        folder_walk = filter(lambda x: folder == name_split(x[0])[0],whole_folder)
    
    for subfold in folder_walk:
        filtmatch = filter(lambda x: x in subfold[1]+subfold[2],match)
        filtnotmatch = filter(lambda x: x in subfold[1]+subfold[2],notmatch)
        if len(filtmatch)>0 and len(filtnotmatch)==0:
            found.append(subfold[0])
      
    if kind=='priori':
        filtfound = filter(lambda x: base(x)[:3]=='uid' and base(x).count('.')==ndots and x[-nchar:]==Priori_Extension,found)
        return [asdm[:-nchar] for asdm in filtfound]  # Expected names of the ASDMs used in the reduction
    else:
        return found
    
    
#################################################
# Find  *subfolders*  with one or more ms:      #
# (useful to classify member ids automatically) #
#################################################

def find_id(whole_folder, folder, tree=False):
    """
    Search a path for ASDMs or MS files to determine the observation ID
    """
  
    subfs = filter(lambda x: os.path.isdir(join(folder,x)),os.listdir(folder))
    subfolders = [join(folder,subf) for subf in subfs]
  
    combprods = [] 
    members = [] 
    folderms = datafind(whole_folder,folder,kind='ms',tree=tree)
    folderimg = datafind(whole_folder,folder,kind='img',tree=tree)
  
    for i,dirsub in enumerate(subfolders):
  
        mslist = datafind(whole_folder,dirsub,kind='ms',tree=tree)
        asdmlist = datafind(whole_folder,dirsub,kind='priori',tree=tree)
        here_asdm = len(asdmlist)>0
    
# If there is at least one analyzed asdm, this subfolder is made of members of a group:
        if here_asdm:
            members += asdmlist # FULL PATH+NAME OF THE (MAYBE-REMOVED?) ASDM.
# If there is no result of an asdm analysis, and folder is not an ms,
# this subfolder is likely to be that of a combined product:
        elif (dirsub not in folderms) and (dirsub not in folderimg):
            combprods.append(dirsub)
      
    return [members,combprods]
  
  
###################################
# Just return empty string if i=1 #
# and str(i) otherwise:           #
###################################

def idname(i):
    if i==1:
        return ''
    else:
        return str(i)
      
      


##########################################
# Post-processing of package content:    #
##########################################

def postProcess(packpath='./DEST', mode='hard'):


  iterator = os.walk(packpath,followlinks=True)
  whole_package = [it for it in iterator]

# Sanitise all fits images for CAS-9131 CASA 4.7 bug

  if not (mode=='ticket' or mode=='fake'):

      all_fits = findall(whole_package,packpath,extension='.fits',kind='file')

      lprint('Will sanitise all fits images for CAS-9131 CASA 4.7 bug:')

      for fimage in all_fits:
          lprint('   Sanitising '+fimage)
          aU.fixFitsHeader(fimage)

# Gzip and tar.gzip the flux and mask images:

  all_flux = findall(whole_package,packpath,extension='.flux.fits',kind='file')
  all_mask = findall(whole_package,packpath,extension='.mask',kind='folder')

  for flux in all_flux:
    flin = open(flux,'r')
    gzipfile = gzip.open(flux+'.gz','w')
    shutil.copyfileobj(flin, gzipfile)
    gzipfile.close()
    flin.close()
    os.remove(flux)


  for mask in all_mask:
    caltabtar = tarfile.open(mask+'.tgz','w:gz')
    caltabtar.add(mask,arcname=base(mask))
    caltabtar.close()
    shutil.rmtree(mask)





#############################################
# Create/update tar and zip files with the  #
# content of the qa2, README, and scripts:  #
#############################################

def make_tar_ticket(packpath='./DEST', group=1, member=1,
                    qafolder='qa',
                    scriptfolder='script'):

# Name of the ticket tar file (it will be written in the package parent directory):
    auxtar = name_split(packpath)
    if auxtar[1] == '':
        auxtar = name_split(auxtar[0])  # in case path ends with '/'.
    
    tarfilename = join(auxtar[1]+'.ticket.tar')
    zipfilename = join(auxtar[1]+'.ticket.zip')
  
  
# Work only with selected groups/members:
    grdir = 'group_ouss_id'+idname(group)
    memdir = 'member_ouss_id'+idname(member)
  
    if member==1:
        memappend = ''
    else:
        memappend = '.member_id'+idname(member)
    
    datadir = join(packpath,'sg_ouss_id',grdir,memdir)
    currwd = os.getcwd()
  
  
    try:
  
# Get total size of the member-id folder:
        size = os.popen('du -sh ' + datadir)
        sizefile = open(join(datadir,'tar_size.txt'+memappend),'w')
        sizefile.write(size.readline())
        sizefile.close()
    
# Generate tree:
        struc = tree(packpath,6)
        strucfile = open(join(datadir,'tar_content.txt'+memappend),'w')
        for line in struc:
            strucfile.write(line)
        strucfile.close()
    
    
# Change to current member/group folder:
        os.chdir(datadir)
    
# Open tar file:
        if os.path.exists(join(currwd,tarfilename)):
            shutil.move(join(currwd,tarfilename),tarfilename)
            shutil.move(join(currwd,zipfilename),zipfilename)
      
        if member==1:
            mytar = tarfile.open(tarfilename,'w')
            myzip = zipfile.ZipFile(zipfilename,'w')
        else:
            mytar = tarfile.open(tarfilename,'a')
            myzip = zipfile.ZipFile(zipfilename,'a')
      
       
      
        copydirs = [qafolder,scriptfolder]

        avoidnames = ['weblog.tar.gz',
                      'weblog.tgz']
    
        for tocopy in copydirs:
# Add content to zip and tar files:
            for basedir, dirs, files in os.walk(tocopy,followlinks=True):
                for file in files:
                    doavoid=False
                    for avoidname in avoidnames:
                        if (file.find(avoidname) == len(file)-len(avoidname)):
                            doavoid=True
                    if doavoid:
                        continue
                    fn = join(basedir, file)
                    myzip.write(fn,join(basedir,file+memappend))
                    mytar.add(fn,arcname=join(basedir,file+memappend))
          
          
          
# Add README:
        mytar.add('README',arcname='README'+memappend)
        myzip.write('README','README'+memappend)
    
# Add summary:
        mytar.add('tar_content.txt'+memappend)
        mytar.add('tar_size.txt'+memappend)
        myzip.write('tar_content.txt'+memappend)
        myzip.write('tar_size.txt'+memappend)
        mytar.close()
        myzip.close()
        os.remove('tar_content.txt'+memappend)
        os.remove('tar_size.txt'+memappend)
    
    
# Finish:
        os.chdir(currwd)
        shutil.move(join(datadir,tarfilename),join(currwd,tarfilename))
        shutil.move(join(datadir,zipfilename),join(currwd,zipfilename))

 
    except Exception as excpt:
        lprint('TAR FAILED!')
        lprint(excpt)
        os.chdir(currwd)
        raise
    
    
###############################
# Append uid to the filename, #
# if it is not appended:      #
###############################
def append_uid(uid,filepath):
    EBspl = filter(lambda x: len(x)>0,uid.split('_'))[1:]  # Name components of the uid.
    splfile = name_split(filepath)
    if sum([EBsp in splfile[1] for EBsp in EBspl])==0:
        os.rename(filepath,join(splfile[0],uid+'__'+splfile[1]))
    
    
###################
# A PYTHONIC TREE #
###################

def tree(startpath, max_level):
    """
    Alphabetically sorted tree
    (each level is sorted: directories, then files)
    
    See http://stackoverflow.com/questions/9727673/list-directory-tree-structure-using-python

    Robust:
    tree('testpack5/',6) == tree('testpack5',6) == tree('./testpack5',6) == tree('./testpack5/',6)
    """
    lines = []
    startpath = startpath.rstrip('/')
    for root, dirs, files in os.walk(startpath):
        dirs.sort()
        level = root.replace(startpath,'').count(os.sep)
        if level < max_level:
            indent = '|   ' * (level) + '|-- '
            lines.append('{0}{1}/\n'.format(indent, os.path.basename(root)))
            if level+1 < max_level:
                subindent = '|   ' * (level+1) + '|-- '
                for f in sorted(files):
                    lines.append('{0}{1}\n'.format(subindent, f))

    return lines
     
     
#############################################
# PRINT MESSAGE TO SCREEN AND/OR TO LOGFILE #
#############################################

# We do not use the usual CASA logging, since the packaging
# is done after QA2 (so no info on packaging should be saved
# in the CASA logs).

def lprint(message,doscreen=True,dolog=True):
    if doscreen:
        print message
    if dolog:
        logfile.write(str(message)+'\n')
    
###################
# RECURSIVE MKDIR #
###################

def mkdir_p(path):
    """ Make directories recursively.  See http://stackoverflow.com/questions/600268/mkdir-p-functionality-in-python """
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



#########################
# TOTAL-POWER PACKAGING #
#########################

def TotalPowerPackaging(whole_folder, origpath, packpath, readme, mode, PIscript):

#####################################
# PACKAGE DIRECTORIES:
  packdirs = ['calibration','log','product','qa','script']

# DATA TO BE ARCHIVED IN TAR FORMAT. IT GOES INTO "calibration"
  NEEDEDTAR = ['plots','flagversions']

# DATA TO BE ADDED TO THE qa FOLDER. NO TARRING
  NEEDED = ['sdlist','png','listobs']

#####################################





# LIST ALL SUBDIRECTORIES:
#  subfs = filter(lambda x: os.path.isdir(join(whole_folder[0][0],x)),os.listdir(whole_folder[0][0]))
  subfs = [join(whole_folder[0][0],x) for x in os.listdir(whole_folder[0][0]) if os.path.isdir(join(whole_folder[0][0],x)) ]


  EBs = []
  COMB = ''
# CHECK WHICH SUBDIRS CORRESPOND TO EBs:
  for subf in subfs:
    content = os.listdir(subf)
    for cont in content:
      if 'SDCalibration.py' in os.path.basename(cont):
        EBs.append(subf)
        break
# CHECK WHICH SUBDIR SHALL HAVE COMBINED PRODUCTS:
      if 'SDimaging.py' in os.path.basename(cont):
        COMB = subf
        break

# IF NO COMB. FOLDER FOUND, COMBINED PRODUCTS SHOULD BE IN PARENT:
  if len(COMB)==0:
    COMB = whole_folder[0][0]

 # print EBs, COMB

# GENERATE DIRECTORY STRUCTURE:
  packpath2 = join(packpath,'sg_ouss_id','group_ouss_id','member_ouss_id')
  if os.path.exists(packpath2):
    lprint('DESTINATION DIRECTORY EXISTS! ABORTING!')
    lprint(packpath2)

    raise ValueError('Destination directory exists!')
  else:
    mkdir_p(packpath2)
    for d in packdirs:
      mkdir_p(os.path.join(packpath2,d))



  for EB in EBs:
     
     EBpath,EBname = name_split(EB)

     lprint('\n ORIGINAL ASDM: %s'%EBname)
     msname = EBname[EBname.find('uid__'):]
     if msname[-3:] != '.ms':
        msname += '.ms'

# PACKAGE CALIBRATION INFO (IN TAR ARCHIVES):
     for need in NEEDEDTAR:
       files = findall(whole_folder,EB,extension=need,kind='file')
       files += findall(whole_folder,EB,extension=need,kind='folder')
# Check that all needed calibration info is here:
       if len(files)==0:
         lprint('ERROR: NO %s IN %s'%(need,EBname))
         raise ValueError('NO %s FOUND!'%need)
       caltabtar = tarfile.open(os.path.join(packpath2,packdirs[0],'%s.%s.tgz'%(msname,need)),'w:gz')
       for fil in files:
         caltabtar.add(fil,arcname=base(fil))
       caltabtar.close()

# PACKAGE QA FILES:
     for need in NEEDED:
  #     print need,EB
       files = findall(whole_folder,EB,extension=need,kind='file',tree=False)
       files += findall(whole_folder,EB,extension=need,kind='folder',tree=False)
       if len(files)==0:
         lprint('ERROR: NO %s IN %s'%(need,EBname))
         raise ValueError('NO %s FOUND!'%need)
       for fil in files:
         cpall(fil,os.path.join(packpath2,packdirs[3]),mode='copy')

# PACKAGE LOGS:
     logs = findall(whole_folder,EB,kind='file',extension='.log')
#     for log in logs:
#        cpall(log,os.path.join(packpath2,packdirs[1],EBname + os.path.basename(log)),mode='copy')


# Concatenate logs for this SB:
     all_logs = open(join(packpath2,packdirs[1],msname+'_casapy.log'),'w')
     for log in logs:
        if 'ipython' not in base(log):
          logf = open(log)
          all_logs.write('\n\n  --- CONTENT OF '+base(log)+': --- \n\n')
          shutil.copyfileobj(logf,all_logs)
          logf.close()
     all_logs.close()

     all_logs = open(join(packpath2,packdirs[1],msname+'_ipython.log'),'w')
     for log in logs:
        if 'casapy' not in base(log):
          logf = open(log)
          all_logs.write('\n\n  --- CONTENT OF '+base(log)+': --- \n\n')
          shutil.copyfileobj(logf,all_logs)
          logf.close()
     all_logs.close()


# Are there images here? Add the EB uid name, in that case:
     imgs = findall(whole_folder,EB,kind='file',extension='.fits',tree=True)
     imgs += findall(whole_folder,EB,kind='file',extension='.flux.fits.gz',tree=True)
     for img in imgs:
       if EBname not in os.path.basename(img):
         outname = EBname + os.path.basename(img)
       else:
         outname = os.path.basename(img)
       cpall(img,os.path.join(packpath2,packdirs[2]),mode='copy')

  lprint('\n Getting combined products')
  
# Copy all scripts and images:
  scripts = findall(whole_folder,whole_folder[0][0],kind='file',extension='.py',tree=True)
  imgs = findall(whole_folder,COMB,kind='file',extension='.fits',tree=True)
  imgs += findall(whole_folder,COMB,kind='file',extension='.flux.fits.gz',tree=True)
  for pys in scripts:
    Standard = False
    for snam in ['scriptForPI.py','scriptForSDCalibration.py','scriptForSDImaging.py','scriptForSDimaging.py']:
      if base(pys).endswith(snam):
        Standard = True
        break
    if not Standard:
         lprint('WARNING! SCRIPT %s IS NOT STANDARD!'%base(pys))

    cpall(pys,os.path.join(packpath2,packdirs[4],os.path.basename(pys)),mode='copy')

  for img in imgs:
    cpall(img,os.path.join(packpath2,packdirs[2],os.path.basename(img)),mode='copy')
            
# Copy PIscript if it exists:
  if os.path.exists(PIscript):
    cpall(PIscript,os.path.join(packpath2,packdirs[4],os.path.basename(PIscript)),mode='copy')
  else:
    lprint('WARNING! The script for PI is not found!',True,True)

# Prepare README file:
# Generate tree folder:

  postProcess(packpath, mode)   

  make_readme(readmepath=packpath,
       readmefile=readme,
       outpath=os.path.join(packpath2, 'README'),
       levels=6)


  make_tar_ticket(packpath=packpath, group=1, member=1)














##############################
# END OF AUXILIARY FUNCTIONS #
##############################





#################################
# BEGINNING OF MAIN SCRIPT CODE #
#################################



def QA_Packager(origpath='./ORIG', packpath='./DEST',
                readme='./README.header.txt', mode='fake', append='',
                gzip_caltables=True, style=None, PIscript='',
                noms=True, gzip_logfiles=True):

    """
    Please visit:
        http://www.eso.org/projects/alma/arc/tw/bin/view/External/PackagingScript
    for user information.
    Notice that the new 'style' keyword OVERWRITES the contents of 'noms'. The
    later is only kept for backwards compatibility. Examples of style keyword:
        'cycle0-nopipe', 'cycle1-nopipe', 'cycle2-pipe1'
    """
   
    ###########################
    # Some definitions and preliminary checks/steps:

    global required_folders_CAL, required_files_QA2
    global rawdatacopy, caldatacopy, rawdatafolder, caldatafolder, caldatacopynames
    global tablefolder,tableplotsfolder,logsfolder,qafolder,fitsfolder
    global scriptfolder,QA2_subfolders, Priori_Extension,QA2_others
    global scriptPattern, scriptNotUiAppend, logExtensions,log2Calib
    global logfile
   
    pipeline=False
    pipemode=0

    # There's a separate packager for the pipeline now
    use_pipeline_packager = False

    currcycle=1
   
    if style is not None:

        cmode, pmode = style.split('-') 
        currcycle = int(cmode[5:])
        

# FIRST CHECK IF THIS IS CAN BE A TP DATASET:
        if cmode in ['cycle2','cycle3','cycle4']:
           SDCompatible=True

        style_valid = True
        if cmode in ('cycle1', 'cycle2', 'cycle3', 'cycle4'):
            noms=True
            if pmode in ['pipe','pipe1'] and cmode != 'cycle1': #style[:11] == 'cycle2-pipe':
                return QA_Pipeline_Packager(origpath,
                                            packpath,
                                            readme,
                                            PIscript,
                                            mode=mode, 
                                            cycle=currcycle)
        elif cmode == 'cycle0':
            noms=False
        else:
            style_valid = False
       
        if pmode == 'nopipe': #style[6:]=='-nopipe':
            pipeline=False
            pipemode=0
        elif pmode == 'pipe1':
            pipeline=True
            pipemode=1
        else:
            style_valid = False

        if not style_valid:
            raise ValueError("Invalid style specified")
    else:
        print "NOTE: instead of using the `noms` parameter, better switch to the `style='cycleN-(no)pipe'` approach."
     

    if mode=='ticket':
        cp_ticket= True; mode = 'fake';
    else:
        cp_ticket=False

   
   
    setDefaults(pipeline,pipemode,cycle=currcycle)
   
    packLog = 'Packager.log.txt'
    logfile = open(packLog,'w')
   
    from time import gmtime, strftime
    lprint('Packaging script log file. Date: %s\n\n'%strftime("%Y-%m-%d %H:%M:%S", gmtime()),False,True)
   
   
   
    if not os.path.isfile(readme):
        lprint('README FILE DOES NOT EXIST OR IS NOT A FILE! ABORTING!')
        lprint('Please, set the value of the \'readme\' keyword correctly!',True,False)
        logfile.close()
        raise ValueError("Nonexistent README file!")
     
       
    if not os.path.isdir(origpath):
        lprint(origpath + ' does not exist or is not a directory! Aborting!')
        lprint('\'origpath\' should be the directory where you have all your QA2 data',True,False)
        logfile.close()
        raise ValueError("Nonexistent QA path!")
     
     
    if origpath[-1]=='/':
        origpath = origpath[:-1]
    if packpath[-1]=='/':
        packpath = packpath[:-1]
     
    if packpath==origpath:
        lprint('\'packpath\' and \'origpath\' CANNOT be the same directory!')
        lprint('Please, set \'packpath\' to the directory where you want to have your packaged data',True,False)
        lprint('Such directory should NOT be in the QA2 directory',True,False)
        lprint('ABORTING!')
        logfile.close()
        raise ValueError("QA path and package path should NOT coincide!")
     
    readmepath = packpath
    packpath2 = join(packpath,'sg_ouss_id','group_ouss_id')
   
     
# Determine destination group id (and minimum member id, in
# case we append the data to member ids).

    curr_grp = 1
    min_memb_id = 1
    prod_memb_id = 1
   
    if os.path.exists(packpath2):
        if append == '':
            lprint('Destination directory already exists. ABORTING!')
            lprint('Unless you set the \'append\' keyword to either \'group\', \'member\', or \'product\',',True,False)
            lprint('the destination (i.e., package) directory should NOT exist',True,False)
            logfile.close()
            raise ValueError("package path exists, but append is not set!")
        else:
            exist = True ; i=2
            removedata = True
            while exist:
                pathtemp = packpath2
                packpath2 = join(packpath,'sg_ouss_id','group_ouss_id'+str(i))
                if os.path.exists(packpath2):
                    i += 1
                else:
                    exist = False
           
            if append == 'group':
                packpath = packpath2
                curr_grp = i
             
            elif append in ['member','product']:
                packpath = pathtemp
                curr_grp = i-1
                exist = os.path.exists(join(pathtemp,'member_ouss_id')) 
                while exist:
                    min_memb_id += 1
                    packpath2 = join(packpath,'member_ouss_id'+str(min_memb_id))
                    if os.path.exists(packpath2)==False:
                        exist = False
                    elif append == 'product':
                        prod_memb_id = min_memb_id
             
            if append == 'product':
                min_memb_id = prod_memb_id
         
    else:
        packpath = packpath2
        removedata=True
     
     
###########################


# Walk through the whole folder:

    iterator = os.walk(origpath,followlinks=True)
    whole_folder = [it for it in iterator]
   
##########################################################
# FIND MEMBERS IDS:
##########################################################

    members=[]
    combprods=[]
    EBs = []
   




######################
# CHECK IF THIS IS A SINGLE-DISH DATASET:
    isSD = False
    for pysc in [os.path.basename(x) for x in findall(whole_folder,whole_folder[0][0],kind='file',extension='.py',tree=True)]:
      if 'SDCalibration.py' in pysc or 'SDimaging.py' in pysc:
        isSD = True
        break

    if not isSD:
      jyp = findall(whole_folder,whole_folder[0][0],kind='file', name='jyperk.csv',tree=True)
      if len(jyp)>0:
        isSD = True


#######################################
## IF THIS IS A SINGLE-DISH SD:
    if isSD:
      lprint('\n\n\nTOTAL-POWER DATASET FOUND!\nSTYLE, APPEND, AND MODE WILL BE OVERRIDEN\n\n\n')
      TotalPowerPackaging(whole_folder,origpath, readmepath, readme, mode, PIscript)
      return
#######################################



    if noms: 
        if not os.path.isfile(PIscript):
            lprint('THE SCRIPT FOR PI DOES NOT EXIST OR IS NOT A FILE! ABORTING!')
            lprint('Please, set the value of the \'PIscript\' keyword correctly!',True,False)
            logfile.close()
            raise ValueError("Nonexistent scriptForPI!")




#########################################################
# Do we have the results of an asdm analysis in the parent folder? 
# If so, there must be only one member id:
# This is the most simple situation.

    here_asdm = datafind(whole_folder,whole_folder[0][0],kind='priori')
   
    if len(here_asdm) > 0:
        members = [whole_folder[0][0]]
        EBs = [here_asdm]
        combprods = [[]]
     
# TRY TO CLASSIFY DATA AUTOMATICALLY:
    else:
   
##################################################################
# If there is no asdm-related stuff in the parent folder, each subfolder 
# may be either a different member id or a combined-product folder.

# First, we check if the subfolders are likely to be of a unique 
# member id:
        combprods = [[]]
        asdms,combs = find_id(whole_folder, whole_folder[0][0],
                              tree=False)
     
#####################
    # Proposed code by Todd, awaiting comments and approval.
    # Note by Ivan (Feb 2014): it doesn't seem to be necessary, is it?
        if False: # proposed code is not used
            if (len(asdms) > 1):
                my_asdm_names = []
                for ctr in range(len(asdms)):
                    print "%d) %s" % (ctr, asdms[ctr])
                    my_asdm_names.append(asdms[ctr].split('/')[-1])
                setnames = list(set(my_asdm_names))
                if (len(setnames) == 1):
                    selection = raw_input('I found multiple folders associated with the same ASDM name.  Which one contains the calibration results? ')
                    if (int(selection) not in range(len(asdms))):
                        print "Invalid selection!"
                        lprint('PROBLEM WITH MULTIPLE FOLDERS WITH THE SAME ASDM: %s'%setnames[0])
                        logfile.close()
                        raise
                    asdms = [asdms[int(selection)]]
#####################

        if len(asdms)>0:
            # There are subfolders with asdms. Hence,
            # there is only 1 member id (i.e., the parent folder).
            members = [whole_folder[0][0]]
            EBs = [asdms] # FULL PATH+NAME OF THE (MAYBE REMOVED?) ASDMs
       
# LOOK FOR COMBINED PRODUCTS:
            if len(combs)>0:
                         # There is/are subfolder(s) unrelated to 
                         # asdms. Hence, these must be combined-product folder(s). 
                combprods = [combs]
         
            if len(whole_folder[0][2])>0: # There are *files* in the parent
                combprods[0] += [whole_folder[0][0]]  # So these files are combined products!
         
         
############################################################
# If the above didn't work, we check if the subfolders are likely 
# to correspond to different member ids:

        else:
     
            nodata = True
            combprods=[]
            for i,subfolder in enumerate(whole_folder[0][1]):
                asdms,combs = find_id(whole_folder,
                                      join(whole_folder[0][0],subfolder),
                                      tree=False)
         
                if len(asdms)>0:
                    if nodata:
                        nodata = False
             
                    members.append(subfolder)
                    EBs.append(asdms)
                    combprods.append(combs)
                    # It could also happen that the combined product is the 
                    # subfolder itself!
                    fullpathsubf = join(whole_folder[0][0],subfolder)
                    whole_idx = [whole_folder[j][0] for j in range(len(whole_folder))].index(fullpathsubf)
                    if len(whole_folder[whole_idx][2])>0:   # There are files in this subfolder!
                    # It is very likely that these are combined-product files:
                        combprods[-1] += [join(whole_folder[0][0],subfolder)]
                    else: # If there is just one set of asdms for this member,
                          # located one level above the other asdms, we attach it
                          # as a new member id.
                        asdms = datafind(whole_folder,join(whole_folder[0][0],subfolder),kind='priori')
                        if len(asdms)>0:
                            members.append(subfolder)
                            EBs.append(asdms)
                            combprods.append([])
               
               
            if nodata == True:
                lprint('NO A-PRIORI-CALIBRATED MSs WERE FOUND IN NEITHER PARENT, SUB, NOR SUBSUB-FOLDERS!')
                lprint('Are you sure that the keyword \'origpath\' points to your QA2 directory?',True,False)
                logfile.close()
                raise ValueError("Bad QA path!")
##################################################################


##################################################################


    nmembers = len(members)
    nEBs = [len(EB) for EB in EBs]
   
   
####################################
# order the members by name:
    if nmembers>1:
        alphordmem = sorted(enumerate(members), key=operator.itemgetter(1))
    else:
        alphordmem = [(0,members[0])]
     
    suffix = [idname(i) for i in range(min_memb_id,min_memb_id+nmembers)]
#####################################



#####################################
# Set the final folder names:

    prodmems = []
   
    for i,member in enumerate(alphordmem):
        iord = member[0]
        prodmems.append('member_ouss_id'+suffix[iord])
     
##################################################



# BEGIN THE PACKAGING!!!

    for imem,member in enumerate(members):
        lprint('\n Packaging data for member %i'%(imem+1))
     
# Path to save the products for this member id:
        prodfolder = join(packpath,prodmems[imem])
        if os.path.exists(prodfolder) and removedata:
            lprint('ALL DATA FROM THE *PACKAGE* FOLDER FOR %s WILL BE *REMOVED*.'%prodmems[imem])
            if mode != 'move':
                lprint('THE DATA FROM THE ORIGINAL QA2 FOLDER WILL NOT BE TOUCHED.')
            shutil.rmtree(prodfolder)
       
# Keep track of all copied tables, flagtables, and plots.
# This is not necessary now, but could be useful in the future:
        allpngs = {}
        allpltfolders = {}
        tabcopied = {}
        flagcopied = {}
        plotcopied = {}
        fixplanets_set = {}
     
     
        for iEB,EB in enumerate(EBs[imem]):
     
            EBpath,EBname = name_split(EB)
            lprint('\n ORIGINAL ASDM: %s'%EBname)
       
            ###############
            # Initialize the copy-tracking dictionaries:
            allpngs[EBpath] = []
            allpltfolders[EBpath] = []
            tabcopied[EBpath] = []
            flagcopied[EBpath] = []
            plotcopied[EBpath] = []
            ###############



            # Walk through the whole member directory:
            member_walk = filter(lambda x: EBpath in x[0],whole_folder)
       
            # Find all the calibration tables:
            allcaltabs = datafind(whole_folder,EBpath,kind='table',tree=True)
            caltabs = filter(lambda x: EBname in base(x), allcaltabs)
            if len(caltabs) != len(allcaltabs):
                lprint('WARNING: THE FOLLOWING TABLES DO NOT HAVE UID ATTACHED TO THEIR NAMES:',False,True)
                for tab in filter(lambda x: x not in caltabs,allcaltabs):
                    lprint(base(tab),False,True)
                lprint('THESE TABLES WILL *NOT* BE PACKAGED!',False,True)
         
            # Check if all the required CAL folders exist:
            calnames = map(base,caltabs)
            for table in required_folders_CAL:
                lt = len(table)
                seltab = [len(caln)>= lt and table in caln[-lt:] for caln in calnames]
                if sum(seltab)==0:
                    lprint('WARNING: NO TABLE %s FOR %s'%(str(table),str(EBname)),False,True) 
           
            # Find the calibration FILES:
            calfiles = []
            for cext in log2Calib:
                calfiles += findall(whole_folder,EBpath,extension=cext,kind='file')
            if len(calfiles) ==0:
                lprint('WARNING: NO FILES FOUND MATCHING: '+', '.join(log2Calib),False,True)
         
         
            # Find all the ms:
            allms = datafind(whole_folder,EBpath,kind='ms',tree=True)
            ms = filter(lambda x: EBname in base(x) or base(x) in caldatacopynames, allms)
            if len(ms) != len(allms):
                lprint('WARNING: THE FOLLOWING MS DO NOT HAVE UID ATTACHED TO THEIR NAMES:',False,True)
                for tab in filter(lambda x: x not in ms,allms):
                    lprint(base(tab),False,True)
                lprint('THESE MS WILL *NOT* BE PACKAGED!',False,True)
         
            # Find the flagversions folder(s):
            allflagvers = findall(whole_folder,EBpath,extension='.flagversions',kind='folder')
            flagvers = filter(lambda x: EBname in base(x), allflagvers)
            if len(flagvers) != len(allflagvers):
                lprint('WARNING: THE FOLLOWING FLAGVERS DO NOT HAVE UID ATTACHED TO THEIR NAMES:',False,True)
                for tab in filter(lambda x: x not in flagvers,allflagvers):
                    lprint(base(tab),False,True)
                lprint('THESE FLAGVERS WILL *NOT* BE PACKAGED!',False,True)
         
            # Find the fits images:
            fits = findall(whole_folder,EBpath,extension='.fits',kind='file')
            fits += findall(whole_folder,EBpath,extension='.flux.fits.gz',kind='file')
       
            # Find the masks images:
            masks = findall(whole_folder,EBpath,extension='.mask',kind='folder')
            masks += findall(whole_folder,EBpath,extension='.mask.tgz',kind='folder')
       
            # Find the logs:
            logs = []
            for extd in logExtensions:
                logs += findall(whole_folder,EBpath,extension=extd,kind='file')
         
            # Find the scripts:
            scripts = []
            for pat in scriptPattern:
                scripts += findall(whole_folder, EBpath, pattern=pat,
                                   extension='.py', kind='file',
                                   exclude_parent="qa2_scripts")

            # Required QA2 files (for all the EBs in this folder!!):
            qa2s = []
            for reqfile in required_files_QA2:
                findlist = findall(whole_folder,EBpath,extension=reqfile,kind='file')
                if len(findlist)==0:
                    lprint('WARNING: NO FILE '+reqfile,False,True)
                else:
                    qa2s += findlist
           
            
            # Find all the folders full of plots (i.e., png images):
            for step in member_walk:
                nopng = filter(lambda x: x[-4:] != '.png', step[2])
                if len(nopng)==0 and len(step[1])==0 and len(step[2])>0:
                    allpltfolders[EBpath].append(step[0])
                    allpngs[EBpath] += step[2]
           
            # Find any other subfolders with QA2 content:

            QA2_subfs = []
       
            for step in member_walk:
                for qa in QA2_others:
                    QA2files = glob.glob(join(step[0],qa))
                    allpngs[EBpath] += filter(lambda x: x[-4:] == '.png',QA2files)
                    for QAfile in QA2files:
                        # Avoid overlap with other tables
                        if QAfile not in caltabs:
                            QA2_subfs.append(QAfile)
             
             
            plotfolders = filter(lambda x: EBname in base(x), allpltfolders[EBpath])
       
            
            ############################################################
            # COPY FILES AND FOLDERS TO THE PACKAGE!!!
            #  (FINALLY!)

            # Data:

            EBraws = []
       
            if noms:
                rawdatacopy=[]
                caldatacopynames=[]
                caldatacopy=[]
         
            if len(rawdatacopy)>0:
                destfolder = join(prodfolder,rawdatafolder)
                if os.path.exists(destfolder)==False:
                    os.makedirs(destfolder)
           
            for copy in rawdatacopy:
                ms2copy = filter(lambda x: x[-len(copy):] == copy,ms)
                if len(ms2copy)==0:
                    lprint('WARNING: %s NOT FOUND.'%copy,False,True)
                else:
                    ###########
                    # NEW CODE. RUN SPLIT AND COPY ONLY THE NECESSARY DATA:
                    EBraws.append(ms2copy[0])
                    if mode != 'fake':
                        ##############################
                        # Change on 16 March 2013. Restore original flags before split:
                        flagmanager(vis=ms2copy[0], mode= 'save', versionname='BeforeFinalSplit')
                        flagmanager(vis=ms2copy[0], mode= 'restore', versionname='Original')
                        ##############################
                        split(vis=ms2copy[0], outputvis=join(destfolder,base(ms2copy[0])), datacolumn='data')
                        ##############################
                        # Change on 16 March 2013. Put final flags back in the non-packaged data:
                        flagmanager(vis=ms2copy[0], mode= 'restore', versionname='BeforeFinalSplit')
                        ##############################
                    else:
                        cpall(ms2copy[0],destfolder,mode)
                    ###########


            thisrawcopied = False  # To check if calibrated datasets are found and copied for each SB
       
            if len(caldatacopy+caldatacopynames)>0:
       
                destfolder = join(prodfolder,caldatafolder)
                if os.path.exists(destfolder)==False:
                    os.makedirs(destfolder)
           
                for copy in caldatacopy:
                    ms2copy = filter(lambda x: x[-len(copy):] == copy,ms)
           
                    if len(ms2copy)==0:
                        lprint('WARNING: %s NOT FOUND.'%copy,False,True)
                    else:
           
                        thisrawcopied = True
                        ###########
                        # NEW CODE. RUN SPLIT AND COPY ONLY THE NECESSARY DATA:
                        if mode != 'fake':
             
                            tb.open(ms2copy[0])
                            datacol = tb.colnames()
                            tb.close()
                            if 'CORRECTED_DATA' in datacol:
                                split(vis=ms2copy[0], outputvis=join(destfolder,base(ms2copy[0])), datacolumn='corrected')
                            else:
                                lprint('WARNING: %s DOES NOT HAVE CORRECTED COLUMN! WILL PACKAGE THE DATA COLUMN.'%ms2copy[0])
                                split(vis=ms2copy[0], outputvis=join(destfolder,base(ms2copy[0])), datacolumn='data')
                 
                        else:
                            cpall(ms2copy[0],destfolder,mode)
                        ###########


                for copy in caldatacopynames:
                    ms2copy = filter(lambda x: base(x) == copy,ms)
                    if len(ms2copy)==0:
                        lprint('WARNING: %s NOT FOUND.'%copy,False,True)
                    else:
                        thisrawcopied = True
                        ###########
                        # NEW CODE. RUN SPLIT AND COPY ONLY THE NECESSARY DATA:
                        if mode != 'fake':
            
                            tb.open(ms2copy[0])
                            datacol = tb.colnames()
                            tb.close()
                            if 'CORRECTED_DATA' in datacol:
                                split(vis=ms2copy[0], outputvis=join(destfolder,base(ms2copy[0])), datacolumn='corrected')
                            else:
                                lprint('WARNING: %s DOES NOT HAVE CORRECTED COLUMN! WILL PACKAGE THE DATA COLUMN.'%ms2copy[0])
                                split(vis=ms2copy[0], outputvis=join(destfolder,base(ms2copy[0])), datacolumn='data')
                
                        else:
                            cpall(ms2copy[0],destfolder,mode)
                        ###########



            ###########
            # NEW CODE. RUN SPLIT AND COPY ONLY THE NECESSARY DATA:
            if thisrawcopied==False and noms==False:  # No calibrated datasets have been copied.
                for thisraw in EBraws:
                    if mode != 'fake':
                        tb.open(thisraw)
                        datacol = tb.colnames()
                        tb.close()
                        if 'CORRECTED_DATA' in datacol:
                            split(vis=thisraw, outputvis=join(destfolder,base(thisraw)+'.cal'), datacolumn='corrected')
                        else:
                            lprint('WARNING: %s DOES NOT HAVE CORRECTED COLUMN! WILL PACKAGE THE DATA COLUMN.'%thisraw)
                            split(vis=thisraw, outputvis=join(destfolder,base(thisraw)+'.cal'), datacolumn='data')
                    else:
                        cpall(thisraw,destfolder,mode)
                        os.rename(join(destfolder,base(thisraw)),join(destfolder,base(thisraw)+'.cal'))
                        ###########



            # Calibration tables (on an ASDM/MS basis):
            destfolder = join(prodfolder,tablefolder)

            tabfolder = None

            for table in caltabs:
                #     if EBname in base(table):
                tabfolder = join(destfolder,EBname+'.'+tablefolder)
                if os.path.exists(tabfolder)==False:
                    os.makedirs(tabfolder)
                cpall(table,tabfolder,mode)
                tabcopied[EBpath].append(table)
       
            for flagver in flagvers:
                #     if EBname in base(flagver):
                tabfolder = join(destfolder,EBname+'.'+tablefolder)
                if os.path.exists(tabfolder)==False:
                    os.makedirs(tabfolder)
                cpall(flagver,tabfolder,mode)
                flagcopied[EBpath].append(flagver)
       
       
            if gzip_caltables:
                if tabfolder is None:
                    lprint("WARNING: No tables found!")
                else:
                    caltabtar = tarfile.open(tabfolder+'.tgz','w:gz')
                    caltabtar.add(tabfolder,arcname=base(tabfolder))
                    caltabtar.close()
                    shutil.rmtree(tabfolder)
         
         
            # Calibration files:
            for calfile in calfiles:
                cpall(calfile,destfolder,mode)
         
         
            destfolder = join(prodfolder,tableplotsfolder)
       
       
            for plotf in plotfolders:
                if EBname in base(plotf):
                    tabfolder = join(destfolder,EBname+'.'+tablefolder+'.plots')
                    if os.path.exists(tabfolder)==False:
                        os.makedirs(tabfolder)
                    cpall(plotf,tabfolder,mode)
                    plotcopied[EBpath].append(plotf)
           
            # Logs:
            destfolder = join(prodfolder,logsfolder)
       
            if len(logs)>0:
                if os.path.exists(destfolder)==False:
                    os.makedirs(destfolder)
           
                # Make a gzip file of logs for this EB:
                if gzip_logfiles:
                    logtar = tarfile.open(join(destfolder,EBname+'.log.tgz'),'w:gz')
                    for log in logs:
                        if base(log) != packLog and 'ipython' not in base(log):
                            logtar.add(log,arcname=base(log))
                    logtar.close()
                else:
         
                    # Concatenate logs for this SB:
                    all_logs = open(join(destfolder,EBname+'.'+tablefolder+'.log'),'w')
                    for log in logs:
                        if base(log) != packLog and 'ipython' not in base(log):
                            logf = open(log)
                            all_logs.write('\n\n  --- CONTENT OF '+base(log)+': --- \n\n')
                            shutil.copyfileobj(logf,all_logs)
                            logf.close()
                    all_logs.close()
           
            else:
                lprint('WARNING: NO LOGS FOUND!',False,True)
         
            # QA2:
            qamode = {0:mode,1:'copy'}[cp_ticket]
       
            destfolder = join(prodfolder,qafolder)
            tarqa2 = destfolder  # for the ticket tar file.
        
            if os.path.exists(destfolder)==False:
                os.makedirs(destfolder)
         
            for qa2 in qa2s:
                cpall(qa2,destfolder,qamode)
                append_uid(EBname,join(destfolder,base(qa2)))
         
            for qa2subf in QA2_subfs:
                cpall(qa2subf,destfolder,qamode)
                append_uid(EBname,join(destfolder,base(qa2subf)))
         
         
            # FITS images:
            if len(fits)>0:
                destfolder = join(prodfolder,fitsfolder)
                if os.path.exists(destfolder)==False:
                    os.makedirs(destfolder)
            else:
                lprint('WARNING: NO FITS IMAGES FOUND!',False,True)
         
            for image in fits:
                cpall(image,destfolder,mode)
                append_uid(EBname,join(destfolder,base(image)))
         
            # masks:
            if len(masks)>0:
                destfolder = join(prodfolder,fitsfolder)
                if os.path.exists(destfolder)==False:
                    os.makedirs(destfolder)
            else:
                lprint('WARNING: NO MASKS FOUND!',False,True)
         
            for image in masks:
                cpall(image,destfolder,mode)
                append_uid(EBname,join(destfolder,base(image)))
         
            # Scripts:
            destfolder = join(prodfolder,scriptfolder)
            if len(scripts)>0:
                tarscript = destfolder  # for the tar ticket file
                if os.path.exists(destfolder)==False:
                    os.makedirs(destfolder)
            else:
                lprint('WARNING: NO SCRIPTS FOUND!',False,True)
         
            fixplanets = False
            for script in scripts:
                cpall(script,destfolder,qamode)
                toappend = os.path.basename(PIscript) not in script
                for pat in scriptNotUiAppend:
                    toappend = toappend and (pat not in script)
                if toappend:
                    append_uid(EBname,join(destfolder,base(script)))
                
                # Check whether ephemerides are needed
                if 'Calibration' in script:
                    fixplanets += contains_fixplanets(script)
            fixplanets_set[EBname] = fixplanets
           
            # Add the scriptforPI:
            if noms and not os.path.exists(join(destfolder,os.path.basename(PIscript))):
                if not os.path.exists(destfolder):
                    # This can occur if there are no scripts in the QA directory
                    os.makedirs(destfolder)
                cpall(PIscript,destfolder,qamode)
         

        try:
            find_and_copy_ephemerides(whole_folder, EBs[imem], fixplanets_set,
                                      calibration_directory=join(prodfolder,
                                                                 tablefolder),
                                      qamode=qamode)
        except ValueError as ex:
            print "ERROR: ",ex.message
            rmpath = packpath.replace("/sg_ouss_id",'').replace("/group_ouss_id",'')
            print "Removing {rmpath}".format(rmpath=rmpath)
            shutil.rmtree(rmpath)
            return
         
        # Combined products (if any):
        if len(combprods[imem])>0:
            for combfolder in combprods[imem]:
                lprint('\n COMBINED PRODUCT: %s'%combfolder)
                combfits = findall(whole_folder,combfolder,tree=True,extension='.fits',kind='file')
                combmasks = findall(whole_folder,combfolder,tree=True,extension='.mask',kind='folder')
                combfits += findall(whole_folder,combfolder,tree=True,extension='.flux.fits.gz',kind='file')
                combmasks += findall(whole_folder,combfolder,tree=True,extension='.mask.tgz',kind='folder')

                comblogs = []; combscripts = []
                for exten in logExtensions:
                    comblogs += findall(whole_folder,combfolder,tree=True,extension=exten,kind='file')
                for pat in scriptPattern:
                    combscripts += findall(whole_folder,combfolder,pattern=pat,tree=True,extension='.py',kind='file',exclude_parent='qa2_scripts')
           
                destfolder = join(prodfolder,fitsfolder)
                if len(combfits)>0:
                    if os.path.exists(destfolder)==False:
                        os.makedirs(destfolder)
                else:
                    lprint('WARNING: NO FITS FILE FOR COMBINED PRODUCTS!',False,True)
            
                for combfit in combfits:
                    cpall(combfit,destfolder,mode)
           
                destfolder = join(prodfolder,fitsfolder)
                if len(combmasks)>0:
                    if os.path.exists(destfolder)==False:
                        os.makedirs(destfolder)
                else:
                    lprint('WARNING: NO MASK FOR COMBINED PRODUCTS!',False,True)
           
                for combmask in combmasks:
                    cpall(combmask,destfolder,mode)
           
                destfolder = join(prodfolder,logsfolder)
         
                if len(comblogs)>0:
         
                    if os.path.exists(destfolder)==False:
                        os.makedirs(destfolder)
             
# Make a gzip file of logs:
                    if gzip_logfiles:
                        logtar = tarfile.open(join(destfolder,base(combfolder)+'.log.tgz'),'w:gz')
                        for comblog in comblogs:
                            if base(comblog) != packLog and 'ipython' not in base(comblog):
                                logtar.add(comblog,arcname=base(comblog))
                        logtar.close()
             
                    else:
# Concatenate logs:
                        all_logs = open(join(destfolder,base(combfolder)+'.log'),'w')
                        for comblog in comblogs: 
                            if base(comblog) != packLog and 'ipython' not in base(comblog):
                                comblogf = open(comblog)
                                all_logs.write('\n\n  --- CONTENT OF '+base(comblog)+': --- \n\n')
                                shutil.copyfileobj(comblogf,all_logs)
                                comblogf.close()
                        all_logs.close()
             
                else:
                    lprint('WARNING: NO LOGS FOR COMBINED PRODUCTS!',False,True)
           
           
                destfolder = join(prodfolder,scriptfolder)
                if len(comblogs)>0:
                    if os.path.exists(destfolder)==False:
                        os.makedirs(destfolder)
                else:
                    lprint('WARNING: NO SCRIPTS FOR COMBINED PRODUCTS!',False,True)
           
                for combscript in combscripts:
                    cpall(combscript,destfolder,qamode)
           
           
        ######################################
        # COPY THE TABLES AND PLOTS THAT DIDN'T MATCH ANY ASDM NAME:
        # (we make use of the copy-tracking dictionaries)

        #  for path in allpngs.keys():

        #   destfolder = join(prodfolder,qafolder)
        #   allpathpngs = findall(whole_folder,path,extension='.png',kind='file',tree=False)
        #   uncopied = filter(lambda x: base(x) not in allpngs[path],allpathpngs)
        #   for pngfile in uncopied:
        #    cpall(pngfile,destfolder,mode)
        #    append_uid(EBname,join(destfolder,base(pngfile)))

        #########################################

        # README FILE (for this member_id folder):

        # Generate tree folder:

        postProcess(readmepath, mode)   


        make_readme(readmepath=readmepath,
                    readmefile=readme,
                    outpath=join(prodfolder, 'README'),
                    member_to_include=prodmems[imem],
                    levels=6)
     
     
     
        # Generate ticket tar file (if there is more than one member id, it
        # will append):
        make_tar_ticket(packpath=readmepath, group=curr_grp,
                        member=min_memb_id+imem, qafolder=qafolder,
                        scriptfolder=scriptfolder)
      
    lprint('\n FINISHED!',True,False) 
    lprint('  THERE MAY BE SPECIAL MESSAGES IN THE FILE %s.'%packLog,True,False)
    # lprint('  PLEASE, READ THEM',True,False)


# TRY TO CORRECT THE ELLUSIVE scriptForPI BUG:
    destfolder = join(prodfolder,scriptfolder)
    if not os.path.exists(join(destfolder,base(PIscript))):
      if os.path.exists(PIscript):
        os.system('cp %s %s/.'%(PIscript,destfolder))
      else:
        lprint('\n ERROR WITH scriptForPI!\n',True,True)
        raise Exception("ERROR WITH scriptForPI!")


    logfile.close()


def find_and_copy_ephemerides(whole_folder, EBs, fixplanets_set,
                              calibration_directory=None,
                              qamode=None):
    """
    Find all ephemerides in all EBs.
    Will raise an exception if multiple ephemerides exist with the same name
    but different contents

    Parameters
    ----------
    whole_folder : list
        The whole_folder walk list used throughout this package
    EBs : list
        A list of EB paths
    fixplanets_set : dict
        A dictionary that has an entry of True or False for each EB name
        indicating whether that EB contains a calibration script with the word
        'fixplanets' in it
    calibration_directory : str
        The output calibration directory into which all ephemeride files will
        be copied
    """
    # Find .eph files:
    ephemerides = {}
    for EB in EBs:
        EBpath,EBname = name_split(EB)
        ephemerides[EBname] = (findall(whole_folder, EBpath, extension='.eph',
                                       kind='file') +
                               findall(whole_folder, EBpath, extension='.eph',
                                       kind='folder'))
        # Per-EB warnings: disabled
        #if fixplanets_set[EBname] and len(ephemerides[EBname]) == 0:
        #    lprint("WARNING: No ephemeris files found,"
        #           " but fixplanets was used in calibration script "
        #           "in EB {EBname}.".format(EBname=EBname))
        #elif not fixplanets_set[EBname] and len(ephemerides[EBname]) > 0:
        #    lprint("WARNING: Ephemeris files found in EB {EBname}, but "
        #           "fixplanets was not used in calibration "
        #           "script.".format(EBname=EBname))

    all_ephemeris_files = [eph
                           for ephlist in ephemerides.values()
                           for eph in ephlist]

    if all_ephemeris_files and not any(fixplanets_set.values()):
        lprint("WARNING: Ephemeris files were found, but 'fixplanets' was "
               "never used in a calibration script.")
    if not all_ephemeris_files and any(fixplanets_set.values()):
        lprint("WARNING: 'fixplanets' was set in at least one calibration "
               "script, but there were no ephemeris files found.")

    check_if_ephemerides_unique(all_ephemeris_files)

    for eph_file in get_unique_files(all_ephemeris_files):
        cpall(eph_file, calibration_directory, qamode)


def make_readme(readmepath, readmefile, outpath, levels=6,
                member_to_include='member_ouss_id'):
    """
    Generate the README file

    Parameters
    ----------
    readmepath: str
        The root path for determining the structure of the README
    readmefile: str
        Path to an existing (or nonexisting) readme file to copy the header
        from
    outpath: str
        The location to write the output readme file (with directory structure)
        to
    levels: int
        The number of tree levels to include in the tree diagram
    member_to_include: 'member_ouss_id'
        A `member_ouss_id` thing.
    """

    # Ensure the output file exists before creating the tree
    touch(outpath)

    fifo = tree(readmepath, levels)

    # Read the README header:
    with open(readmefile, 'r') as READHEAD:
        HeadLines = READHEAD.readlines()
 
    # Change project code:
    projname = name_split(readmepath.strip("/"))
   
    for i,line in enumerate(HeadLines):
        if len(line)>0 and 'Project code:' in line:
            HeadLines[i] = 'Project code: '+projname[1]+'\n'
         
    # Write README:
    with open(outpath,'w') as README:
     
        for line in HeadLines:
            README.write(line)
       
        README.write('\n\n')
     
        goodmemb = member_to_include
        badmemb = 'member_ouss_i'
     
        isgood = True;
        for i,line in enumerate(fifo):
     
            if (badmemb in line) and (goodmemb not in line):
                isgood = False
            if (goodmemb in line):
                isgood = True
            if isgood:
                README.write(line)

def contains_subpath(dirpath, matchstring):
    """
    Check if a file path contains a sub-path matching 'matchstring'
    
    Does the same thing as 'datafind' but for a single folder
    """
    files = [base(x) for x in glob.glob(join(dirpath,'*'))]
    if matchstring in files:
        return True
    else:
        return False
   
def QA_Pipeline_Packager(staged_root, delivery_root, readme='', PIscript='', mode='copy',cycle=1):
    """
    Packaging script for pipeline-processed *and* imaged data

    *assumes* a directory structure of the form:
    |-- staged_root/
    |   |-- sg_ouss_id/
    |   |   |-- group_ouss_id/
    |   |   |   |-- member_ouss_id/
    |   |   |   |   |-- log/
    |   |   |   |   |-- qa/
    |   |   |   |   |-- product/
    |   |   |   |   |-- script/
    |   |   |   |   |-- calibration/
    |   |   |   |   |-- calibrated/

    Parameters
    ----------
    pipeline_root : str
        The root directory of the processed data
    staged_root : str
        The root directory of the output renamed, restructured, processed data
    readme : str
        Path to the README file.  Assumed to be 'staged/sous/gous/mous/README'
        if not specified
    PIscript : str
        (optional) Path to the scriptForPI.py script to include
    mode : ['fake','hard','copy','move','ticket']
        The copy mode.  'fake' will create empty files, 'hard' will create
        hardlinks, 'move' will move the file and create symbolic links back to
        the original location
    """

    print "Packaging from ", staged_root, " to directory ", delivery_root, " in mode \"", mode, "\" including pipeline files."
    

    global logfile


# Check whether this is a TP dataset:
    subdirs = 'sg_ouss_id/group_ouss_id/member_ouss_id'
    isTP = os.path.exists(os.path.join(staged_root, subdirs, 'calibration/jyperk.csv'))

    
    packLog = 'PipelinePackager.log.txt'
    logfile = open(packLog,'w')

    from time import gmtime, strftime
    lprint('Pipeline Packaging script log file. Date: %s\n\n' % 
           strftime("%Y-%m-%d %H:%M:%S", gmtime()),False,True)

    if isTP:
      lprint('\nThis is a Total Power dataset\n',True,True)

    dirs = ['log', 'qa', 'product', 'script', 'calibration']

    for dd in dirs:
        mkdir_p(join(delivery_root, subdirs, dd))
    
    # The output script names are predictable and verifiable
    # The following files *MUST* be present
    if isTP:
      scripts_required = ['casa_pipescript.py',
                          'PPR*xml',]
    else:
      scripts_required = ['casa_piperestorescript.py',
                        'casa_pipescript.py',
                        'scriptForImaging.py',
                        'PPR*xml',]
    # These should be copied, but no exceptions should be raised if they are
    # missing: these can be present if it was an analysis where some execblocks
    # were processed manually.
    if cycle in [0,1,2]:
      scripts_optional = ['scriptForFluxCalibration.py',
                        '*.scriptForCalibration.py',]
    else:
 #   elif cycle in [3]:
      scripts_optional = ['scriptForImagingPrep.py',
                        '*.scriptForCalibration.py',]


    if PIscript=='' and not isTP: # path to scriptForPI.py not provided, needs to be in package
        scripts_required.append('scriptForPI.py')


    if isTP:

      copy_to_dir = {
        # The output product names: wildcards including all .fits
        # and all masks needed to produce the images
        'product': ['*.fits','*.flux.fits.gz'],
        'calibrated': ['*.fits','*.flux.fits.gz'],
        # qa: wildcards
        'qa': ['*weblog.t*gz'],
        # log: wildcards
        'log': ['*casa_commands.log',
                'casa*log'],
        # calibration: 
        'calibration': ['*calapply.txt',
                        '*flagversions.t*gz',
                        '*caltables.t*gz',
                        'jyperk.csv',
                        '*flagtemplate.txt',
                        ],
        'script': scripts_optional,
      }


    else:

      copy_to_dir = {
        # The output product names: wildcards including all .fits
        # and all masks needed to produce the images
        'product': ['*.fits','*.flux.fits.gz',
                    '*.mask','*.mask.tgz'],
        'calibrated': ['*.fits','*.flux.fits.gz',
                       '*.mask','*.mask.tgz'],
        # qa: wildcards
        'qa': ['*_textfile.txt',
               '*qa2_part*png',
               '*weblog.t*gz'],
        # log: wildcards
        'log': ['*casa_commands.log',
                'casa*log'],
        # calibration: 
        'calibration': ['*calapply.txt',
                        '*flagversions.t*gz',
                        '*caltables.t*gz',
                        'flux.csv',
                        '*flagtemplate.txt'],
        'script': scripts_optional,
      }

      if cycle in [3,4]:
        copy_to_dir['calibration'].append('antennapos.csv')
        copy_to_dir['calibration'].append('*flagtargetstemplate.txt')
        copy_to_dir['calibration'].append('*cont.dat')

    # optional files about which one should not complain if they are absent
    if isTP:
      if cycle in [0,1,2]:
        optional_wildcards = ['*_textfile.txt',
                          '*qa2_part*png',
                          'casa*log',
                          '*.mask.tgz', '*.flux.fits.gz',
                          'scriptForFluxCalibration.py',
                          '*.scriptForCalibration.py']
      else:
  #    elif cycle in [3]:
        optional_wildcards = ['*_textfile.txt',
                          '*qa2_part*png',
                          'casa*log',
                          '*.mask.tgz', '*.flux.fits.gz',
                          'scriptForImagingPrep.py',
                          '*.scriptForCalibration.py']
    else:
      if cycle in [0,1,2]:
        optional_wildcards = ['*.mask', '*.mask.tgz',
                          '*_textfile.txt',
                          '*qa2_part*png',
                          'casa*log',
                          '*.mask.tgz', '*.flux.fits.gz',
                          'scriptForFluxCalibration.py',
                          '*.scriptForCalibration.py']
      else:
     # elif cycle in [3]:
        optional_wildcards = ['*.mask', '*.mask.tgz',
                          '*_textfile.txt',
                          '*qa2_part*png',
                          'casa*log',
                          'antennapos.csv',
                          '*.mask.tgz', '*.flux.fits.gz',
                          'scriptForImagingPrep.py',
                          '*.scriptForCalibration.py',
                          '*flagtargetstemplate.txt',
                          '*cont.dat']

  
    # remapping  calibrated -> product
    # (can be used for other remapping; in the example below calibrated/ files
    # will be copied to product/)
    # duplicate lines here are to aid those unfamiliar with dictionary
    # comprehensions - the next two lines perform identical operations
  #  dirname_maps = {d:d for d in copy_to_dir} # dictionary comprehension
    dirname_maps = dict([(d,d) for d in copy_to_dir]) # dict(list comprehension)
    dirname_maps['calibrated'] = 'product'
    
    # acquire UIDs:
    calibrated_path = join(staged_root, subdirs, 'calibrated')
    uids = calibrated_filenames = [base(fn)
                                   for fn in glob.glob(join(calibrated_path,"uid*ms"))
                                   if contains_subpath(fn, 'DATA_DESCRIPTION')]

    # Copy required scripts
    for fglob in scripts_required:
        staged_scriptdir = join(staged_root, subdirs, 'script')
        files = glob.glob(join(staged_scriptdir, fglob))

        # Deal with special cases
        if fglob == 'scriptForImaging.py' and len(files) == 0:
            # Special case: look for scriptForImaging in two places
            files = glob.glob(join(staged_root, subdirs, 'calibrated', fglob))
            staged_scriptdir = join(staged_root, subdirs, 'calibrated')

        if len(files) == 0:
            raise ValueError("MISSING: {0}/{1}".format(staged_scriptdir, fglob))

        for fn in files:
            special_cp(join(staged_scriptdir, base(fn)),
                       join(delivery_root, subdirs, 'script', base(fn)),
                       mode=mode)

    # Ephemerides cannot be in pipeline-produced data
    ## BEGIN ephemerides section
    ## Check for flux calibration script to determine whether a warning is
    ## needed for missing ephemerides
    #calibration_scripts = glob.glob(join(staged_root, subdirs, 'script',
    #                                     '*Calibration.py'))
    #fixplanets = False
    #for script_filename in calibration_scripts:
    #    with open(script_filename, 'r') as ff:
    #        if 'fixplanets' in ff.read():
    #            fixplanets = True
    ## Copy the ephemerides from *anywhere*
    #target_dirname = 'calibration'
    #fglob = join(staged_root, subdirs, '*', '*.eph')
    #files = glob.glob(fglob)
    #if len(files) > 1:
    #    # after this, files should have length 1
    #    files = get_first_ephemeride(files)
    #if len(files) == 1:
    #    special_cp(files[0],
    #               join(delivery_root, subdirs, target_dirname,
    #                    base(files[0])), mode=mode)
    #    if not fixplanets:
    #        lprint("WARNING: Ephemeris files found, but "
    #               "fixplanets was not set.")
    #elif len(files) == 0:
    #    if fixplanets:
    #        lprint("WARNING: No ephemeris files found,"
    #               " but fixplanets was set.")
    #else:
    #    raise Exception("Bad conditions reached.  Failure in packager.")
    ## END ephemerides section


    # Copy everything else
    for dirname in copy_to_dir:
        target_dirname = dirname_maps[dirname]
        staged_dirname = join(staged_root, subdirs, dirname)
        for fglob in copy_to_dir[dirname]:
            files = glob.glob(join(staged_dirname, fglob))

            if len(files)==0 and fglob not in optional_wildcards:
                print "WARNING: no match for ", fglob, " in directory ", staged_dirname
               # return
            
            for fn in files:
                special_cp(join(staged_dirname, base(fn)),
                           join(delivery_root, subdirs, target_dirname,
                                base(fn)), mode=mode)
                
    # special case: also get casa*.log from "calibrated"
    if (len(glob.glob(os.path.join(staged_root, subdirs, 'calibrated', 'casa*.log')))>0):
        files = glob.glob(os.path.join(staged_root, subdirs, 'calibrated', 'casa*.log'))
        for fn in files:
            special_cp(fn,
                       join(delivery_root, subdirs, 'log', base(fn)),
                       mode=mode)

    # special case: also get scriptForFluxCalibration.py from "calibrated"
    if cycle in [0,1,2]:
      files_calibrated_scriptforfluxcal = glob.glob(join(staged_root, subdirs,
                                                       'calibrated',
                                                       'scriptForFluxCalibration.py'))
    else:
#    elif cycle in [3]:
      files_calibrated_scriptforfluxcal = glob.glob(join(staged_root, subdirs,
                                                       'calibrated',
                                                       'scriptForImagingPrep.py'))
    if (len(files_calibrated_scriptforfluxcal)>0):
        for fn in files_calibrated_scriptforfluxcal:
            special_cp(fn, join(delivery_root, subdirs, 'script', base(fn)),
                       mode=mode)

    # special case: copy over the scriptForPI if provided
    if PIscript!='':
        print "Including scriptForPI specified by user: ", PIscript
        special_cp(PIscript,
                   join(delivery_root, subdirs, 'script', base(PIscript)),
                   mode=mode)

    if readme == '':
        readme = join(staged_root, subdirs, 'README')

    postProcess(delivery_root, mode)   

    make_readme(delivery_root,
                readme,
                join(delivery_root, subdirs, 'README'))

    # Generate ticket tar file (if there is more than one member id, it will append):
    make_tar_ticket(packpath=delivery_root, group=1, member=1)

    print "Done."
   
def QA_Pipeline_Stager(pipeline_root, staged_root, mode='copy',
                       PIscript=None,
                       fake_imaging=False,
                       fake_flux_calibration=False,
                       PIPE_README='',cycle=4):
    """
    pre-packaging staging for pipeline-processed data *before* imaging.
    
    This will create a directory structure of the form:
    |-- staged_root/
    |   |-- sg_ouss_id/
    |   |   |-- group_ouss_id/
    |   |   |   |-- member_ouss_id/
    |   |   |   |   |-- log/
    |   |   |   |   |-- qa/
    |   |   |   |   |-- product/
    |   |   |   |   |-- script/
    |   |   |   |   |-- calibration/

    Parameters
    ----------
    pipeline_root : str
        The root directory of the processed data
    staged_root : str
        The root directory of the output renamed, restructured, processed data
    mode : ['fake','hard','copy','move']
        The copy mode.  'fake' will create empty files, 'hard' will create
        hardlinks, 'move' will move the file and create symbolic links back to
        the original location, 'copy' will make a regular copy.
    fake_imaging : bool
        If set, this will created a `calibrated/` directory and populate it
        with an empty `.ms` to mimic the results of running `scriptForPI.py`
    fake_flux_calibration : bool or 'calibrated'
        If set, this will create a `scriptForFluxCalibration.py` file in the
        script directory with contents: "# content of scriptForFluxCalibration.py"
        If set to 'calibrated', will create the file in the 'calibrated/'
        directory instead (From cycle3 on, this is called "scriptForImagingPrep.py")
    PIPE_README: Specify the location and name of the pipeline-related README file.

    Examples
    --------
    Do a "dry run"
    >>> QA_Pipeline_Stager('2013.1.00760.S_2014_07_22T11_41_12.638',
    ...                      '2013.1.00760.S', mode='fake')

    Do a real run, copying the files
    >>> QA_Pipeline_Stager('2013.1.00760.S_2014_07_22T11_41_12.638',
    ...                      '2013.1.00760.S', mode='copy')
    """

    print "Staging pipeline results from ", pipeline_root, " to directory ", staged_root, " in mode \""+mode+"\""

    if len(PIPE_README)>0 and not os.path.isfile(PIPE_README):
      raise IOError('PIPE_README file %s does NOT exist!'%PIPE_README)
  
    if mode == 'ticket':
        mode = 'copy'

    global logfile

    packLog = 'PipelineStager.log.txt'
    logfile = open(packLog,'w')

    from time import gmtime, strftime
    lprint('Pipeline Packaging:Staging script log file. Date: %s\n\n' % 
           strftime("%Y-%m-%d %H:%M:%S", gmtime()),False,True)

    subdirs = 'sg_ouss_id/group_ouss_id/member_ouss_id'
    dirs = ['log', 'qa', 'product', 'script', 'calibration']


    for dd in dirs:
        mkdir_p(join(staged_root, subdirs, dd))

    # Wildcard for finding the appropriate (deep) subdirectory
    sousgousmous = join(pipeline_root,
                                'S*OUS*/G*OUS*/M*OUS*/')


    Total_Power = len(glob.glob(os.path.join(sousgousmous,'working/jyperk.csv')))>0

    # Copy the pipeline scripts and PPR files to the output scripts/ directory
    if Total_Power:
      lprint('\nThis is a Total Power dataset\n',True,True)
      script_dict = {'products/*casa_pipescript.py': 'casa_pipescript.py',
                   'products/PPR*': '',}
    else:
      script_dict = {'products/*casa_piperestorescript.py': 'casa_piperestorescript.py',
                   'products/*casa_pipescript.py': 'casa_pipescript.py',
                   'products/PPR*': '',}


    for script_wildcard, script_target in script_dict.iteritems():

        # Use glob + wildcards to find the file
        scripts = glob.glob(join(sousgousmous, script_wildcard))
        # Then check and make sure exactly one of each file is found
        if len(scripts) > 1:
            raise IOError("Found more than one {wildcard}".format(wildcard=script_wildcard))
        elif len(scripts) == 0:
            raise IOError("Did not find {wildcard} in {directory}".
                          format(directory=sousgousmous,
                                 wildcard=script_wildcard))
        else:
            filename = scripts[0]
            if script_target == '':
                script_target = base(filename)

        script_destination = join(staged_root, subdirs, 'script/')
        target = join(script_destination, script_target)

        special_cp(filename, target, mode=mode)


    if os.path.exists(PIscript):
        special_cp(PIscript, join(staged_root, subdirs,
                                  'script/scriptForPI.py'), mode=mode)
    elif PIscript is not None:
        raise IOError("PIscript {0} not found".format(PIscript))


    if not Total_Power:

      if fake_flux_calibration == 'calibrated':
        mkdir_p(join(staged_root, subdirs, 'calibrated'))
        if cycle in [0,1,2]:
          with open(join(staged_root, subdirs,
                       'calibrated/scriptForFluxCalibration.py'), 'w') as f:
            f.write("# Content of scriptForFluxCalibration.py in calibrated/")
        else:
   #     elif cycle in [3]:
          with open(join(staged_root, subdirs,
                       'calibrated/scriptForImagingPrep.py'), 'w') as f:
            f.write("# Content of scriptForImagingPrep.py in calibrated/")
      elif fake_flux_calibration:
        if cycle in [0,1,2]:
          with open(join(staged_root, subdirs,
                       'script/scriptForFluxCalibration.py'), 'w') as f:
            f.write("# Content of scriptForFluxCalibration.py in script/")
        else:
     #   elif cycle in [3]:
          with open(join(staged_root, subdirs,
                       'script/scriptForImagingPrep.py'), 'w') as f:
            f.write("# Content of scriptForImagingPrep.py in script/")


    # Copy the calibration tables and products of various sorts
    # see proposedpipeline_transfertoarc_and_delivery_mapping
    if Total_Power:

      wildcard_target_dict = {'products/*calapply.txt':'calibration',
                            'products/*flagversions.t*gz':'calibration',
                            'products/*caltables.t*gz':'calibration',
                            'working/jyperk.csv':'calibration',
                            'working/*flagtemplate.txt':'calibration',
                            'products/*.fits':'product',
                            'products/*.flux.fits.gz':'product',
                            'products/*weblog.t*gz':'qa',
                            'products/*casa_commands.log':'log',
                            'working/casa*.log':'log',
                           }

      optional_wildcards = ['working/casa*.log',
                        #  'products/*.mask',
                          'products/*.flux.fits.gz',
                          'working/*flagtemplate.txt']

    else:

      wildcard_target_dict = {'products/*calapply.txt':'calibration',
                            'products/*flagversions.t*gz':'calibration',
                            'products/*caltables.t*gz':'calibration',
                            'working/flux.csv':'calibration',
                        #    'working/antennapos.csv':'calibration',
                            'working/*flagtemplate.txt':'calibration',
                        #    'working/*.mask':'product',
                            'products/*.fits':'product',
                            'products/*.flux.fits.gz':'product',
                            'products/*.mask':'product',
                            'products/*.mask.tgz':'product',
                            'products/*weblog.t*gz':'qa',
                            'products/*casa_commands.log':'log',
                            'working/casa*.log':'log',
                           }


      if cycle in [3,4]:
         wildcard_target_dict['working/antennapos.csv'] = 'calibration'
         wildcard_target_dict['products/*flagtargetstemplate.txt'] = 'calibration'
         wildcard_target_dict['products/*cont.dat'] = 'calibration'



      optional_wildcards = ['working/casa*.log',
                          'products/*.mask',
                          'products/*.mask.tgz',
                          'products/*.flux.fits.gz',
                          'working/antennapos.csv',
                          'working/*flagtemplate.txt']

      if cycle in [3,4]:
          optional_wildcards.append('products/*flagtargetstemplate.txt')
          optional_wildcards.append('products/*cont.dat')



    for wildcard, target in wildcard_target_dict.iteritems():

        target_dir = join(staged_root, subdirs, target)

        searchpath = join(sousgousmous, wildcard)

        files = glob.glob(searchpath)

        if len(files) == 0:
            if not (wildcard in optional_wildcards):
                lprint('ERROR: No files found matching'
                       ' {wildcard}'.format(wildcard=searchpath))
                break
        else:
            for fn in files:
                target = join(target_dir, os.path.basename(fn))
                #print "special_cp({0}, {1}, mode={2})".format(fn, target, mode)
                special_cp(fn, target, mode=mode)

    if fake_imaging: # and not Total_Power:
        _make_fake_imaging_files(join(staged_root, subdirs))



# Copy the PIPELINE README file:
    if len(PIPE_README)>0:
      special_cp(PIPE_README, join(staged_root, subdirs), mode=mode)


    logfile.close()
    print "Done."

def check_if_ephemerides_unique(filelist):

    # Check that any non-unique elements (i.e., files with the same name) are
    # identical; if not, complain.
    basenames = [base(fn) for fn in filelist]
    for bn in basenames:
        if basenames.count(bn) > 1:
            # get all the files with the same basename
            matched_list = [filelist[ii]
                            for ii,bb in enumerate(basenames)
                            if bb==bn]
            for f1,f2 in itertools.combinations(matched_list, 2):
                if not are_ephems_equal(f1,f2):
                    raise ValueError("Found multiple ephemeris tables with "
                                     "the same name, at least two of which "
                                     "do not have the same contents: {0}, {1}\n"
                                     "You will need to rename them such that "
                                     "they all have unique names and correct "
                                     "the references to them in the "
                                     "calibration scripts.".format(f1,f2))


def get_unique_files(filelist):
    """
    Examples
    --------
    >>> x = ['/a/b/c', '/b/a/c', '/c/c/c', '/a/b/c']
    >>> get_unique_files(x)
    ['/a/b/c']
    """
    basenames = [base(fn) for fn in filelist]
    unique_filelist = [filelist[ii]
                       for ii,bb in enumerate(basenames)
                       if ii == basenames.index(bb)]
    return unique_filelist


def are_ephems_equal(dir1, dir2):
    """
    Compare two directories recursively. Files in each directory are
    assumed to be equal if their names and contents are equal.

    From:
    http://stackoverflow.com/questions/4187564/recursive-dircmp-compare-two-directories-to-ensure-they-have-the-same-files-and

    @param dir1: First directory path
    @param dir2: Second directory path

    @return: True if the directory trees are the same and 
        there were no errors while accessing the directories or files, 
        False otherwise.
    """

    if ((os.path.isfile(dir1) and not os.path.isfile(dir2))
        or 
        (os.path.isfile(dir2) and not os.path.isfile(dir1))
       ):
        return False
    elif os.path.isfile(dir1) and os.path.isfile(dir2):
        return filecmp.cmp(dir1, dir2)

    dirs_cmp = filecmp.dircmp(dir1, dir2)
    if (len(dirs_cmp.left_only)>0 or
        len(dirs_cmp.right_only)>0 or
        len(dirs_cmp.funny_files)>0):
       return False
    (junk, mismatch, errors) =  filecmp.cmpfiles(dir1, dir2,
                                                 dirs_cmp.common_files,
                                                 shallow=False)
    if len(mismatch)>0 or len(errors)>0:
       return False
    for common_dir in dirs_cmp.common_dirs:
       new_dir1 = os.path.join(dir1, common_dir)
       new_dir2 = os.path.join(dir2, common_dir)
       if not are_ephems_equal(new_dir1, new_dir2):
           return False
    return True

def touch(fname, times=None):
    """ from http://stackoverflow.com/questions/1158076/implement-touch-using-python """
    with open(fname, 'a'):
        os.utime(fname, times)

def _make_fake_imaging_files(outdir, uid='uid___A001_X111_X111'):
    """
    Create fake calibrated and raw data structures

    Parameters
    ----------
    outdir : str
        A directory name in which the imaging files should be created
    uid : str
        The UID for the data.  A file {uid}.ms will be created.
    """

    print "UNIT TEST: Making fake imaging directory."

    touch(join(outdir, 'README'))

    calibrated_outdir = join(outdir, 'calibrated')

    # Need to match all of the ms / table types listed in datafind
    # ANTENNA and CAL_DESC should be in different files from DATA_DESCRIPTION
    # for the non-merged data (apparently)
    table_name_pairs = [('DATA_DESCRIPTION',''),
                        ('DATA_DESCRIPTION','.split.cal'),
                        ('ANTENNA','.tsys'),
                        ('CAL_DESC','.tsys')]
    for table,suffix in table_name_pairs:
        fake_ms = '{uid}.ms{suffix}/{table}/'.format(table=table,
                                                     suffix=suffix,
                                                     uid=uid)

        mkdir_p(join(calibrated_outdir,fake_ms))

    mkdir_p(join(calibrated_outdir,'calibrated_imaged_junk.mask'))
    fileo = open(join(calibrated_outdir,'calibrated_imaged_junk.mask/FORTEST'),'w')
    print >> fileo, 'I am a test'
    fileo.close()
    fileo = open(join(calibrated_outdir,'calibrated_imaged_junk.mask.gz'),'w')
    print >> fileo, 'I am a test'
    fileo.close()


    mkdir_p(join(calibrated_outdir,'calibrated_imaged_junk.image')) # "noise" file for testing purposes
    touch(join(calibrated_outdir,'calibrated_imaged_junk.fits'))
    touch(join(calibrated_outdir,'calibrated_imaged_junk.flux.fits'))
    touch(join(calibrated_outdir,'calibrated_imaged_junk.flux.fits.gz'))
    touch(join(calibrated_outdir,'calibrated_imaged_junk.pbcor.fits'))
    touch(join(calibrated_outdir,'scriptForImaging.py'))

    rawdata_outdir = join(outdir, 'rawdata')
    mkdir_p(join(rawdata_outdir,'uid___A001_X111_X111/ASDMBinary'))

def contains_fixplanets(script_filename):
    with open(script_filename, 'r') as ff:
        if 'fixplanets' in ff.read():
            return True
    return False

def make_ephems(ephem_style='none', target_directory=None,
                second_target_directory=None,cycle=1):
    """
    Make fake ephemeris file and add 'fixplanets' to a calibration script for
    unit testing

    Parameters
    ----------
    ephem_style : str
        If 'ok', create a single, plausibly-valid Ephemerides directory
        (it will not be a real ephemeris table, but will resemble one)
        If 'none', do not create ephemerides
        If 'two_ok', two Ephemeris directories will be created that *do* match
        If 'two_bad', two Ephemeris directories will be created that *do not* match
        If 'file', will create an ephemeride file instead of a directory
        If 'two_ok_file', two Ephemeris *files* will be created that *do* match
        If 'two_bad_file', two Ephemeris *files* will be created that *do not* match
    target_directory : str
        The directory in which the calibration file and the ephemeris file will
        live
    second_target_directory : str
        The directory in which the second ephemeris file will live
    """
    valid_ephems = ('ok', 'two_ok', 'two_bad', 'none', 'file', 'two_ok_file',
                    'two_bad_file')
    if ephem_style in ('ok','two_ok','two_bad'):
        mkdir_p(join(target_directory,'ephemeris.eph'))
        touch(join(target_directory,'ephemeris.eph','file1.txt'))
    elif 'file' in ephem_style:
        touch(join(target_directory,'ephemeris.file.eph'))
    elif ephem_style not in valid_ephems:
        raise ValueError("ephem must be one of {0}".format(valid_ephems))

    if ephem_style == 'two_ok':
        # different name, different content
        mkdir_p(join(second_target_directory,'ephemeris2.eph'))
        touch(join(second_target_directory,'ephemeris2.eph','file2.txt'))
    elif ephem_style == 'two_bad':
        # same name, different content
        mkdir_p(join(second_target_directory,'ephemeris.eph'))
        touch(join(second_target_directory,'ephemeris.eph','file2.txt'))
    elif ephem_style == 'two_bad_file':
        touch(join(second_target_directory,'ephemeris.file.eph'))
        with open(join(second_target_directory,'ephemeris.file.eph'), 'w') as f:
            f.write("Not the same as the other one.")
    elif ephem_style == 'two_ok_file':
        touch(join(second_target_directory,'ephemeris.file.eph'))

    if ephem_style in ('ok', 'two_ok', 'file', 'two_bad', 'two_ok_file',
                       'two_bad_file'):
        # Add the text 'fixplanets' to the calibration script
        if cycle in [0,1,2]:
          cal_scripts = [join(target_directory, 'scriptForFluxCalibration.py')]
        else:
      #  elif cycle in [3]:
          cal_scripts = [join(target_directory, 'scriptForImagingPrep.py')]
 
        if 'two' in ephem_style:
          if cycle in [0,1,2]:
            cal_scripts.append(join(second_target_directory,
                                    'scriptForFluxCalibration.py'))
          else:
       #   elif cycle in [3]:
            cal_scripts.append(join(second_target_directory,
                                    'scriptForImagingPrep.py'))
        for cal_script in cal_scripts:
            if os.path.exists(cal_script):
                with open(cal_script, 'r') as f:
                    txt = f.read()+"\n"
            else:
                txt = ""
            with open(cal_script, 'w') as f:
                f.write(txt+"fixplanets")

def cleanup_ephems(ephem_style='none', target_directory=None,
                   second_target_directory=None):
    """
    Remove the files created by make_ephems
    """
    if ephem_style == 'none':
        return

    valid_ephems = ('ok', 'two_ok', 'two_bad', 'none', 'file', 'two_ok_file',
                    'two_bad_file')
    if ephem_style in ('ok','two_ok','two_bad'):
        rmtarget = join(target_directory,'ephemeris.eph')
        shutil.rmtree(rmtarget)
    elif 'file' in ephem_style:
        rmtarget= join(target_directory,'ephemeris.file.eph')
        os.remove(rmtarget)

    if 'two' in ephem_style:
        if ephem_style == 'two_ok':
            rmtarget = join(second_target_directory,'ephemeris2.eph')
            shutil.rmtree(rmtarget)
        elif 'file' in ephem_style:
            rmtarget = join(second_target_directory,'ephemeris.file.eph')
            os.remove(rmtarget)
        else:
            rmtarget = join(second_target_directory,'ephemeris.eph')
            shutil.rmtree(rmtarget)

    #if ephem_style in ('ok', 'two_ok', 'file', 'two_bad'):
    #    # Add the text 'fixplanets' to the calibration script
    #    cal_script = join(target_directory, 'scriptForFluxCalibration.py')
    #    if os.path.exists(cal_script):
    #        with open(cal_script, 'r') as f:
    #            txt = f.read().replace("fixplanets","")
    #        with open(cal_script, 'w') as f:
    #            f.write(txt)
    #    else:
    #        # this should not be called
    #        print "WARNING: removing scriptForFluxCalibration.py"
    #        os.remove(cal_script)

