
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

# flag.py

# Description:
# ------------
# This module runs the flagging statistics subsystem of the QA2 system.

# Functions:
# ----------
# flag       - This function runs the flagging statistics subsystem of the QA2
#              system.
# flag_calc  - This function calculates the flagging statistics.
# flag_write - This function writes the flagging statistics to a CASA table.

# Non-user functions:
# -------------------
# flag_desc - This function returns the description for the flag statistics
#             table.

# Modification history:
# ---------------------
# 2011 May 20 - Nick Elias, NRAO
#               Initial version created with functions flag(), flag_calc(), and
#               flag_write().
# 2011 Jun 12 - Nick Elias, NRAO
#               Function flag_desc() added.
# 2011 Sep 01 - Nick Elias, NRAO
#               Logging functionality incorporated.
# 2011 Sep 10 - Nick Elias, NRAO
#               Local logging classes employed.

# ------------------------------------------------------------------------------

# Imports
# -------

import os

import casa

import utility.logs as logs

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

# flag

# Description:
# ------------
# This function runs the flagging statistics subsystem of the QA2 system.

# Inputs:
# -------
# in_ms   - This python string contains the name of the input MS.
# out_dir - This python string contains the output directory.
# logobj  - This python variable contains the logging object information
#           ('PYTHON' creates a local python Logger class, 'CASA' creates a
#           local python logger, <python> and <casa> are actual instances of the
#           logger classes; default = 'PYTHON').

# Outputs:
# --------
# If the function finishes successfully, True is returned via the function
# value.  If the function does not finish successfully, it raises an exception.

# Modification history:
# ---------------------
# 2011 May 20 - Nick Elias, NRAO
#               Initial version.
# 2011 Jun 13 - Nick Elias, NRAO
#               Improved error handling.
# 2011 Jun 20 - Nick Elias, NRAO
#               Raise an exception if the output table already exists.
# 2011 Sep 10 - Added logging capability.

# ------------------------------------------------------------------------------

def flag( in_ms, out_dir, logobj='PYTHON' ):

	# Initialize the logger

	root = 'flag.flag'
	level = logs.INFO

	log_local, logger = logs.init( out_dir, root, level, logobj )


	# Print the first log message

	msg = 'Flagging statistics started for ' + in_ms + ' ...'
	origin = root
	logger.info( msg, origin=origin )


	# Check to see if the output file already exists

	out_table_root = os.path.basename( os.path.splitext( in_ms )[0] )
	out_table = out_dir + '/' + out_table_root + '.flag.stats'

	if os.path.exists( out_table ):
		msg = 'Output table ' + out_table + ' already exists ...\n'
		origin = root
		logger.error( msg, origin=origin )
		raise IOError( msg )


	# Calculate the flagging statistics

	flag_stats_dict = flag_calc( in_ms )

	msg = 'Flagging statistics of ' + in_ms + ' calculated ...'
	origin = root
	logger.info( msg, origin=origin )


	# Write the flagging statistics to the table

	status = flag_write( flag_stats_dict, out_table )

	msg = 'Flagging statistics written in ' + out_table + ' ...'
	origin = root
	logger.info( msg, origin=origin )


	# Print the last log message

	msg = 'Flagging statistics finished for ' + in_ms + ' ...\n'
	origin = root
	logger.info( msg, origin=origin )


	# Delete the logger, if it was created locally

	if log_local: del logger


	# Return True

	return True

# ------------------------------------------------------------------------------

# flag_calc

# Description:
# ------------
# This function calculates the flagging statistics.

# Inputs:
# -------
# in_ms - This python string contains the name of the input MS.

# Outputs:
# --------
# The dictionary containing the flagging statistics for the MS, returned via
# the function value.

# Modification history:
# ---------------------
# 2011 May 20 - Nick Elias, NRAO
#               Initial version.

# ------------------------------------------------------------------------------

def flag_calc( in_ms ):

	# Create the local instance of the flag tool and open it

	fgLoc = casa.__flaggerhome__.create()

	fgLoc.open( in_ms )


	# Get the flagging statistics

	fgLoc.setdata()
	fgLoc.setflagsummary()

	flag_stats_dict = fgLoc.run()


	# Close and delete the local flag tool

	fgLoc.done()

	del fgLoc


	# Return the dictionary containing the flagging statistics

	return flag_stats_dict

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

# flag_write

# Description:
# ------------
# This function writes the flagging statistics to a CASA table.

# TBD:
# ---
# A simple-minded approach to write information using the table tool was used.
# There may be a more elegant way to do it.

# Inputs:
# -------
# flag_stats_dict - The python dictionary that contains the flagging statistics.
# out_table       - The python string that contains the output table name.

# Outputs:
# --------
# The flagging statistics table is created in the output directory.  If the
# function finishes successfully, True is returned via the function value.

# Modification history:
# ---------------------
# 2011 May 20 - Nick Elias, NRAO
#               Initial version (stub).
# 2011 Jun 12 - Nick Elias, NRAO
#               First working version.
# 2011 Jun 13 - Nick Elias, NRAO
#               Improved error handling.

# ------------------------------------------------------------------------------

def flag_write( flag_stats_dict, out_table ):

	# Create the local instance of the table tool and create the output
	# table according to the description

	tbLoc = casa.__tablehome__.create()

	tbLoc.create( out_table, flag_desc() )


	# Put the global flagging statistics into the output table

	total = flag_stats_dict['total']
	flagged = flag_stats_dict['flagged']
	fraction = float( flagged ) / float( total )

	row = 0

	tbLoc.addrows()
	tbLoc.putcol( 'type', 'global', row, 1, 1 )
	tbLoc.putcol( 'id', '', row, 1, 1 )
	tbLoc.putcol( 'total', total, row, 1, 1 )
	tbLoc.putcol( 'flagged', flagged, row, 1, 1 )
	tbLoc.putcol( 'fraction', fraction, row, 1, 1 )


	# Put the rest of the flagging statistics into the output table

	keys1 = sorted( flag_stats_dict )
	keys1.remove( 'total' )   # Global flagging statistics already written
	keys1.remove( 'flagged' ) # Global flagging statistics already written

	for k1 in keys1:

		keys2 = sorted( flag_stats_dict[k1] )

		for k2 in keys2:

			total = flag_stats_dict[k1][k2]['total']
			flagged = flag_stats_dict[k1][k2]['flagged']
			fraction = float( flagged ) / float( total )

			row += 1

			tbLoc.addrows()
			tbLoc.putcol( 'type', k1, row, 1, 1 )
			tbLoc.putcol( 'id', k2, row, 1, 1 )
			tbLoc.putcol( 'total', total, row, 1, 1 )
			tbLoc.putcol( 'flagged', flagged, row, 1, 1 )
			tbLoc.putcol( 'fraction', fraction, row, 1, 1 )


	# Close and delete the local table tool

	tbLoc.done()

	del tbLoc


	# Return True

	return True

# ------------------------------------------------------------------------------

# flag_desc

# Description:
# ------------
# This function creates the description of the flag statistics table.  The
# description is a python dictionary whose elements correspond to table
# columns.

# Inputs:
# -------
# None.

# Outputs:
# --------
# The python dictionary containing the description of the flag statistics table,
# returned via the function value.

# Modification history:
# ---------------------
# 2011 Jun 12 - Nick Elias, NRAO
#               Initial version.

# ------------------------------------------------------------------------------

def flag_desc():

	# Create the description of the flag statistics table

	desc = dict()

	desc['type'] = {
		'comment': 'Type of flag statistics',
		'dataManagerGroup': 'StandardStMan',
		'dataManagerType': 'StandardStMan',
		'maxlen': 0,
		'option': 0,
		'valueType': 'string'
	}

	desc['id'] = {
		'comment': 'ID for flag statistics',
		'dataManagerGroup': 'StandardStMan',
		'dataManagerType': 'StandardStMan',
		'maxlen': 0,
		'option': 0,
		'valueType': 'string'
	}

	desc['total'] = {
		'comment': 'Total number of data',
		'dataManagerGroup': 'StandardStMan',
		'dataManagerType': 'StandardStMan',
		'maxlen': 0,
		'option': 0,
		'valueType': 'integer'
	}

	desc['flagged'] = {
		'comment': 'Number of flagged data',
		'dataManagerGroup': 'StandardStMan',
		'dataManagerType': 'StandardStMan',
		'maxlen': 0,
		'option': 0,
		'valueType': 'integer'
	}

	desc['fraction'] = {
		'comment': 'Fraction of flagged data',
		'dataManagerGroup': 'StandardStMan',
		'dataManagerType': 'StandardStMan',
		'maxlen': 0,
		'option': 0,
		'valueType': 'float'
	}


	# Return the description of the flag statistics table

	return desc
