#!/bin/bash
# testpackage.sh
# test ALMA data packaging and calibration script correctness
# D.Petry (ESO), Mar, July, Sept 2014, Mar, June, July, Aug, Oct, Nov 2015, Mar, Apr, Aug, Oct, Nov 2016,
#                Feb 2017
# $Id: testpackage.sh,v 1.21 2017/03/31 10:17:12 dpetry Exp $

# Exit values: 
#  0 - all OK
#  1 - errors occurred (package not OK)
#  2 - wrong usage
#  3 - there were no errors but there were important warnings (package potentially not OK)

# Set env var. AIVDIR to the path to local copy of the AIV directory
#    to enable scriptForPI version testing

# Default values for env. variables:
if [ "a$AIVDIR" == "a" ]; then
    export AIVDIR='/lustre/opsw/alma/jao-mirror/AIV'
fi

# default casa version (may be overridden later by the casa version found in the README)
export USECASAVERSION=4.7.2


if [ "a"$1 == "a" ]; then
    echo "Usage 1: testpackage.sh <single tarballname or single package directory"
    echo "  Unpacks the given delivery tarball and scrutinizes the structure the contained scripts."
    echo
    echo "  Example: testpackage.sh 2012.1.00293.S_uid___A002_X6802f4_X6_001_of_001.tar"
    echo "            Should, when successful, result in a final message \"Done.\" and no ERROR messages."
    echo
    echo "Usage 2: testpackage.sh '<expression selecting a group of tarballs>'"
    echo "  Works as Usage 1 except that it processes a group of tarballs belonging to the same MOUS."
    echo '  NOTE: the single quotes around the expression are required!' 
    echo 
    echo "  Example: testpackage.sh '2012.1.00293.S_*.tar'"
    echo
    echo "Usage 3: testpackage.sh <single package directory>"
    echo "  Works as Usage 1 except that it skips the untarring step and immediately processes"
    echo '  the given directory as if it was the result of unpacking a given tarball.' 
    echo 
    echo "  Example: testpackage.sh 2012.1.00293.S"
    exit 2
fi

export TBALL=$1
export HAVEDIR=False

if [ ! -d $TBALL ]; then
    NUMTARBALLS=`ls $TBALL | wc -l`
    echo "Found $NUMTARBALLS tarballs belonging to this package."
    export TBALLS=$TBALL
    if [ $NUMTARBALLS -gt 1 ]; then
	export TBALL=`ls $TBALL | head -n 1`
    fi

    if [ ! -e $TBALL ]; then
	echo "ERROR: "${TBALL}" doesn not exist."
	exit 2
    fi

    export TARNAMEROOT=`echo $TBALL | sed 's/_001_of_/+/g' | cut -d "+" -f 1`
else
    export HAVEDIR=True
fi

STATSFILE=${PWD}/${TBALL}_STATS.txt
echo $TBALLS > $STATSFILE
date >> $STATSFILE
STATSSTART=`date +%s`

LOGFILE=${PWD}/test${TBALL}.log

echo "$TBALLS" > $LOGFILE

if [ $HAVEDIR == "False" ]; then
    ls -d  20??.?.*.? >> $LOGFILE 2>&1
    if [ $? == 0 ]; then
	echo "ERROR: a directory matching pattern 20??.?.*.? already exists" | tee -a $LOGFILE
	exit 2
    fi

    echo "Checking README ..." | tee -a $LOGFILE 
    checkREADME.py $TARNAMEROOT -tidy >> $LOGFILE
    if [ ! $? == 0 ]; then
	echo "ERROR: README contains incorrect file tree (see log file $LOGFILE for details)." | tee -a $LOGFILE
	exit 1
    fi

    echo "Unpacking $TBALLS ..."
    for NAME in $TBALLS; do
	echo "Processing "$NAME" ..."
	tar xvf $NAME >> $LOGFILE
    done
else
    which checkREADME2.py > /dev/null 2>&1
    if [ $? == 0 ]; then
	echo "Checking README ..." | tee -a $LOGFILE 
	checkREADME2.py $TBALL -tidy >> $LOGFILE
	if [ ! $? == 0 ]; then
    	    echo "ERROR: README contains incorrect file tree (see log file $LOGFILE for details)." | tee -a $LOGFILE
    	    exit 1
	fi
    else
	echo "Checking README on directories (instead of tarballs) not yet implemented ..." | tee -a $LOGFILE 
    fi
fi


export MYCYCLE="unknown"
ls -d  2012.?.*.? > /dev/null 2>&1
if [ $? == 0 ]; then
    export MYCYCLE=1
fi
ls -d  2013.?.*.? > /dev/null 2>&1
if [ $? == 0 ]; then
    export MYCYCLE=2
fi
ls -d  2015.?.*.? > /dev/null 2>&1
if [ $? == 0 ]; then
    export MYCYCLE=3
fi
ls -d  2016.?.*.? > /dev/null 2>&1
if [ $? == 0 ]; then
    export MYCYCLE=4
fi
ls -d  2017.?.*.? > /dev/null 2>&1
if [ $? == 0 ]; then
    export MYCYCLE=5
fi
echo "This is a project from Cycle "$MYCYCLE"." | tee -a $LOGFILE

cd 20??.?.*.?
if [ ! $? == 0 ]; then
    echo "ERROR: "${TBALLS}" does not contain a directory 20??.?.*.? or there is more than one" | tee -a $LOGFILE
    exit 2
fi

cd *
if [ ! $? == 0 ]; then
    echo "ERROR: on sg_ouss_id level: multiple directories or none" | tee -a $LOGFILE
    exit 1
fi
cd *
if [ ! $? == 0 ]; then
    echo "ERROR: on group_ouss_id level: multiple directories or none" | tee -a $LOGFILE
    exit 1
fi
cd *
if [ ! $? == 0 ]; then
    echo "ERROR: on member_ouss_id level: multiple directories or none" | tee -a $LOGFILE
    exit 1
fi

if [ ! -e README ]; then
    echo "ERROR: the file README does not exist." | tee -a $LOGFILE
    exit 1
fi

echo "Checking CASA version used according to README ..." | tee -a $LOGFILE
grep "CASA version used for reduction" README | grep $USECASAVERSION
if [ ! $? == 0 ]; then
    export USECASAVERSION=0
    for CVERSION in 4.5.0  4.5.1  4.5.2  4.5.3  4.5  4.6  4.7.0-1 4.7.0 4.7.1 4.7.2 4.7  4.2.2  4.3.1; do
      grep "CASA version used for reduction" README | grep $CVERSION
      if [ $? == 0 ]; then
	  export USECASAVERSION=$CVERSION
	  break
      fi
    done
    if [ $USECASAVERSION == 0 ]; then
	grep "CASA version used for reduction" README | grep 4.3
	if [ ! $? == 0 ]; then
	    export USECASAVERSION=`grep "CASA version used for reduction" README | grep -Eo "[3456789].[0123456789].*[0123456789]*"`
	    echo "*** NOTE: Required CASA version is different from 4.2.2, 4.3.1, 4.4, 4.5.0/1/2/3, 4.6, 4.7, 4.7.0-1, 4.7.1, 4.7.2" | tee -a $LOGFILE
	else
	    echo "ERROR: CASA version used for reduction is 4.3. This version should not be used" | tee -a $LOGFILE
	    echo "       but CASA 4.3.1 should be used instead. Please correct the calibration script CASA version checks." | tee -a $LOGFILE
	    exit 1
	fi
    fi
fi

    
echo "According to README, needs CASA version "$USECASAVERSION >> $LOGFILE

echo "Checking directories script, product, qa, log, and calibration ..." | tee -a $LOGFILE
if [ ! -d script ] || [ ! -d product ] || [ ! -d qa ] || [ ! -d log ] | [ ! -d calibration ]; then
    ls -l
    echo "ERROR: at least one of the directories script, product, qa, log, and calibration does not exist." | tee -a $LOGFILE
    exit 1
fi

cd product

echo "********************************************************" | tee -a $LOGFILE
echo "Testing the images in the product directory ..." | tee -a $LOGFILE


if [ $MYCYCLE == 1 ]; then

    
    ls -l *.fits
    if [ ! $? == 0 ]; then
	echo "ERROR: the images (.fits) are missing." | tee -a $LOGFILE
	exit 1
    fi

    echo "Testing phase centers ..." | tee -a $LOGFILE

    for NAME in *.fits
      do
      OBJECTNAME=`grep -Eo 'OBJECT  = [[:alnum:][:punct:]]* ' $NAME` 
      echo "$NAME  uses phase center $OBJECTNAME" | tee -a $LOGFILE
    done

else
    export ISPOLA=False

    ls -l *.pbcor.fits
    if [ ! $? == 0 ]; then
	ls -l ../script/scriptForPolCalibration.py > /dev/null
	if [ ! $? == 0 ]; then
	    echo "ERROR: the primary beam corrected images (.pbcor.fits) are missing." | tee -a $LOGFILE
	    exit 1
	else
	    export ISPOLA=True
	    echo "The primary beam corrected images (.pbcor.fits) are not present but this is a full polarisation dataset. Continuing ..." | tee -a $LOGFILE
	fi
    fi

    ls -l *.flux.fits 
    if [ ! $? == 0 ]; then
	ls -l *.flux.fits.gz
	if [ ! $? == 0 ]; then
	    ls -l *.pb.fits
	    if [ ! $? == 0 ]; then
		ls -l *.pb.fits.gz
		if [ ! $? == 0 ]; then
		    if [ $ISPOLA == "True" ]; then
			echo "The primary beam (.flux.fits[.gz] or .pb.fits[.gz]) images are not present but this is a full polarisation dataset. Continuing ..." | tee -a $LOGFILE		
		    else
			echo "ERROR: the primary beam (.flux.fits[.gz] or .pb.fits[.gz]) images are missing." | tee -a $LOGFILE
			exit 1
		    fi
		fi
	    fi
	fi
    fi

    echo "Testing phase centers ..." | tee -a $LOGFILE
 
    export IMAGELIST=`ls *.pbcor.fits`
    if [ $ISPOLA == "True" ]; then
	export IMAGELIST=`ls *.fits`
    fi

    for NAME in $IMAGELIST
      do
      OBJECTNAME=`grep -Eo 'OBJECT  = [[:alnum:][:punct:]]* ' $NAME` 
      echo "$NAME  uses phase center $OBJECTNAME" | tee -a $LOGFILE
    done

fi

for NAME in `ls oussid*.fits 2> /dev/null`; do
    echo "ERROR: file $NAME in directory products in not well named." | tee -a $LOGFILE
    OUSSIDFITS=True
done
if [ "a$OUSSIDFITS" == "aTrue" ]; then
    exit 1
fi

echo "********************************************************" | tee -a $LOGFILE

cd ../script

rm -f casapy*.log ipython*.log

echo "Checking scriptForPI.py ..."
if [ ! -e scriptForPI.py ]; then
    echo "ERROR: "${TBALLS}" does not contain scriptForPI.py"  | tee -a $LOGFILE
    exit 1
fi
if [ -d $AIVDIR ]; then
    diff ${AIVDIR}/science/qa2/scriptForPI.py scriptForPI.py >> $LOGFILE 2>&1
    if [ ! $? == 0 ]; then
	echo "WARNING: scriptForPI.py differs from repository!" | tee -a $LOGFILE
	echo "Will try to continue ..." | tee -a $LOGFILE
	export SFPIDIFFERS=1
    fi
else
    echo "WARNING: cannot test if scriptForPI.py differs from repository."  | tee -a $LOGFILE
    echo "         Environment variable AIVDIR is set to non-existing dir $AIVDIR" | tee -a $LOGFILE
fi 


echo "Checking scriptForImaging.py ..."  | tee -a $LOGFILE
if [ ! -e scriptForImaging.py ]; then
    echo "ERROR: scriptForImaging.py is missing."  | tee -a $LOGFILE
    exit 1
fi

grep outframe scriptForImaging.py >> $LOGFILE 2>&1
if [ ! $? == 0 ]; then
    echo "WARNING: scriptForImaging.py does not set outframe parameter!"  | tee -a $LOGFILE
    export NOOUTFRAME=1
fi

if [ $USECASAVERSION == "4.7" ] || [ $USECASAVERSION == "4.7.0" ]; then
    echo "   Checking presence of CAS-9131 workaround for CASA version "$USECASAVERSION  | tee -a $LOGFILE
    export CAS9131A=0
    export CAS9131B=0
    grep unsetenv scriptForImaging.py >> $LOGFILE 2>&1
    if [ ! $? == 0 ]; then 
	export CAS9131A=1
    fi
    grep CASAVERSION scriptForImaging.py >> $LOGFILE 2>&1
    if [ ! $? == 0 ]; then 
	export CAS9131B=1
    fi
    if [ $CAS9131A == 1 ] || [ $CAS9131B == 1 ]; then
	grep "no manual imaging was performed" scriptForImaging.py >> $LOGFILE 2>&1
	if [ $? == 0 ] || [ "a`wc -l scriptForImaging.py`" == "a2 scriptForImaging.py" ]; then
            export CAS9131WARN=1
	    echo "WARNING: This package was produced with CASA 4.7.0 but the scriptForImaging does not contain the line"  | tee -a $LOGFILE
            echo "   os.unsetenv('CASAVERSION') # necessary to avoid a CASA 4.7.0 bug which leads to a stray CR character in exported FITS"  | tee -a $LOGFILE
            echo "  This is probably OK since it seems no manual imaging was performed."
	else 
	    echo "ERROR: This package was produced with CASA 4.7.0 but the scriptForImaging does not contain the line"  | tee -a $LOGFILE
	    echo "   os.unsetenv('CASAVERSION') # necessary to avoid a CASA 4.7.0 bug which leads to a stray CR character in exported FITS"  | tee -a $LOGFILE
	    echo "  Please add it at the beginning of the script before the first exportfits call, and try again."  | tee -a $LOGFILE
	    exit 1
	fi
    fi
fi


if [ $USECASAVERSION == "4.1" ]; then
    echo "Checking weights recipe usage for CASA version "$USECASAVERSION  | tee -a $LOGFILE
    export WEIGHTSINIM=0
    export WEIGHTSINCAL=0
    export WEIGHTSINFCAL=0
    grep weights.py scriptForImaging.py >> $LOGFILE 2>&1
    if [ $? == 0 ]; then 
	export WEIGHTSINIM=1
    fi
    grep weights.py uid*scriptForCalibration.py  >> $LOGFILE 2>&1
    if [ $? == 0 ]; then  
	export WEIGHTSINCAL=1
    fi
    grep weights.py scriptForFluxCalibration.py >> $LOGFILE 2>&1
    if [ $? == 0 ]; then
	export WEIGHTSINFCAL=1
    fi
    if [ $WEIGHTSINIM == 1 ] || [ $WEIGHTSINCAL == 1 ] || [ $WEIGHTSINFCAL == 1 ]; then
	if [ ! -e weights.py ]; then
	    echo "ERROR: weights.py is missing."  | tee -a $LOGFILE
	    exit 1
	fi
    fi
fi

cd ..
if [ -d calibrated ]; then
    echo "ERROR: directory calibrated already exists."  | tee -a $LOGFILE
    exit 1
fi

cd script

echo "Checking for thesteps variable modifications and for es, aU, and np module calls ..." | tee -a $LOGFILE
for NAME in `ls uid*.ms.scriptForCalibration.py`; do
    export ANAME=`echo $NAME | cut -d'.' -f 1`

    grep "thesteps = \[\]" ${ANAME}.ms.scriptForCalibration.py >> $LOGFILE 2>&1
    if [ ! $? == 0 ]; then
	echo "ERROR: The variable \"thesteps\" was apparently modfied in "${ANAME}".ms.scriptForCalibration.py . Script will not run correctly!"  | tee -a $LOGFILE
	exit 1
    fi

    grep " es\." ${ANAME}.ms.scriptForCalibration.py >> $LOGFILE 2>&1
    grep  -B 1 " es\." ${ANAME}.ms.scriptForCalibration.py | tr "\n" " " | tr "[-*2]" "\n" | grep -v "#[[:space:]]*es\." | grep -v "if applyonly != True" | grep "es"
    if [ $? == 0 ]; then
	echo "ERROR: "${ANAME}".ms.scriptForCalibration.py seems to use es module methods even if applyonly is true!"  | tee -a $LOGFILE
	exit 1
    fi

    grep " aU\." ${ANAME}.ms.scriptForCalibration.py >> $LOGFILE 2>&1
    grep -B 1 " aU\." ${ANAME}.ms.scriptForCalibration.py | tr "\n" " " | tr "[-*2]" "\n" | grep -v "#[[:space:]]*aU\." | grep -v "if applyonly != True" | grep "aU"
    if [ $? == 0 ]; then
    	echo "ERROR: "${ANAME}".ms.scriptForCalibration.py seems to use aU module methods even if applyonly is true!" | tee -a $LOGFILE
    	exit 1
    fi

    grep " np\." ${ANAME}.ms.scriptForCalibration.py >> $LOGFILE 2>&1
    if [ $? == 0 ]; then
	grep "import numpy" ${ANAME}.ms.scriptForCalibration.py >> $LOGFILE 2>&1
	if [ ! $? == 0 ]; then
	    echo "ERROR: "${ANAME}".ms.scriptForCalibration.py seems to use np module methods but numpy is not imported!" | tee -a $LOGFILE
	    exit 1
	fi
    fi
done

cd ..

if [ -d raw ]; then
    echo "ERROR: directory raw already exists." | tee -a $LOGFILE
    exit 1
fi


export CASAOPTION=""

cd script

export ANAME=""

echo "Checking for files matching uid*.ms.scriptForCalibration.py ..." | tee -a $LOGFILE
for NAME in `ls uid*.ms.scriptForCalibration.py`; do
    export ANAME=`echo $NAME | cut -d'.' -f 1`
    echo $ANAME | tee -a $LOGFILE
done

echo "Checking for scriptForFluxCalibration.py ..." | tee -a $LOGFILE
ls scriptForFluxCalibration.py
if [ $? == 0 ]; then
    if [ $MYCYCLE -gt 2 ]; then
	echo "ERROR: in Cycle "$MYCYCLE" the name scriptForFluxCalibration.py should no longer be used." | tee -a $LOGFILE
	echo "       Please rename to scriptForImagingPrep.py ." | tee -a $LOGFILE
	exit 1
    fi
fi

echo "Checking for files matching PPR* ..." | tee -a $LOGFILE
for NAME in `extractasdms.py`; do
    export ANAME=$NAME
    echo $ANAME | tee -a $LOGFILE
    export CASAOPTION=" --pipeline"
done

cd ..

if [ "a$CASAOPTION" == "a --pipeline" ]; then
    echo "Checking pipeline version ..."  | tee -a $LOGFILE
    cd qa
    WEBLOGEXCERPT=`tar xzOf *weblog.t*gz "pipeline*/html/t1-1.html" | grep "<td>"` 
    echo $WEBLOGEXCERPT | grep "4\.2\.2" >> $LOGFILE
    if [ $? == 0 ]; then
	echo "pipeline version used was 4.2.2"  | tee -a $LOGFILE
	if [ ! $USECASAVERSION == "4.2.2" ]; then
	    echo "ERROR: expected CASA version is $USECASAVERSION" | tee -a $LOGFILE
	    exit 1
	fi
    else
	echo $WEBLOGEXCERPT | grep "4\.3\.1" >> $LOGFILE
	if [ $? == 0 ]; then
	    echo "pipeline version used was 4.3.1"  | tee -a $LOGFILE
	    if [ ! $USECASAVERSION == "4.3.1" ]; then
		echo "ERROR: expected CASA version is $USECASAVERSION" | tee -a $LOGFILE
		exit 1
	    fi
	else
	    echo $WEBLOGEXCERPT | grep "4\.5\.0" >> $LOGFILE
	    if [ $? == 0 ]; then
		echo "pipeline version used was 4.5"  | tee -a $LOGFILE
		if [ ! $USECASAVERSION == "4.5" ]; then
		    echo "ERROR: expected CASA version is $USECASAVERSION" | tee -a $LOGFILE
		    exit 1
		fi
	    else
		echo $WEBLOGEXCERPT | grep "4\.5\.1" >> $LOGFILE
		if [ $? == 0 ]; then
		    echo "pipeline version used was 4.5.1"  | tee -a $LOGFILE
		    if [ ! $USECASAVERSION == "4.5.1" ]; then
			echo "ERROR: expected CASA version is $USECASAVERSION" | tee -a $LOGFILE
			exit 1
		    fi
		else
		    echo $WEBLOGEXCERPT | grep "4\.5\.2" >> $LOGFILE
		    if [ $? == 0 ]; then
			echo "pipeline version used was 4.5.2"  | tee -a $LOGFILE
			if [ ! $USECASAVERSION == "4.5.2" ] && [ ! $USECASAVERSION == "4.5.3" ]; then
			    echo "ERROR: expected CASA version is $USECASAVERSION" | tee -a $LOGFILE
			    exit 1
			fi
		    else
			echo $WEBLOGEXCERPT | grep "4\.5\.3" >> $LOGFILE
			if [ $? == 0 ]; then
			    echo "pipeline version used was 4.5.3"  | tee -a $LOGFILE
			    if [ ! $USECASAVERSION == "4.5.3" ]; then
				echo "ERROR: expected CASA version is $USECASAVERSION" | tee -a $LOGFILE
				exit 1
			    fi
			else
			    echo $WEBLOGEXCERPT | grep "4\.7\.0" >> $LOGFILE
			    if [ $? == 0 ]; then
				echo "pipeline version used was 4.7.0"  | tee -a $LOGFILE
				if [ ! $USECASAVERSION == "4.7.0" ] && [ ! $USECASAVERSION == "4.7" ] && [ ! $USECASAVERSION == "4.7.0-1" ]; then
				    echo "ERROR: expected CASA version is $USECASAVERSION" | tee -a $LOGFILE
				    exit 1
				fi
			    else
				echo $WEBLOGEXCERPT | grep "4\.7\.2" >> $LOGFILE  # note: 4.7.1 was never used in production
				if [ $? == 0 ]; then
				    echo "pipeline version used was 4.7.2"  | tee -a $LOGFILE
				    if [ ! $USECASAVERSION == "4.7.2" ]; then
					echo "ERROR: expected CASA version is $USECASAVERSION" | tee -a $LOGFILE
					exit 1
				    fi
				else
				    echo "ERROR: pipeline version is neither 4.2.2 nor 4.3.1 nor 4.5.0/1/2/3 nor 4.7.0 nor 4.7.2" | tee -a $LOGFILE
				    exit 1
				fi
			    fi
			fi
		    fi
		fi
	    fi
	fi
    fi

    echo "Checking for forbidden SSO flux calibrators ..."  | tee -a $LOGFILE
    # (still in directory qa)
    WEBLOGEXCERPT=`tar xzOf *weblog.t*gz "pipeline*/html/casa_commands.log" | grep "reference='"` 
    for FORBIDDENSSO in Ceres Pallas; do
	echo $WEBLOGEXCERPT | grep "reference='$FORBIDDENSSO'" >> $LOGFILE
	if [ $? == 0 ]; then
	    if [ $USECASAVERSION == "4.5.3" ]; then
		grep setjy ../script/scriptForImagingPrep.py
		if [ ! $? == 0 ]; then
		    echo "ERROR: pipeline calibration used forbidden SSO $FORBIDDENSSO as flux calibrator and there is no appropriate scriptForImagingPrep." 
		    echo "       Need to use Quasar instead of the SSO (scaling in scriptForImagingPrep.py) or use CASA version 4.7 or later."
		    exit 1
		else
		    echo "NOTE: pipeline calibration used forbidden SSO $FORBIDDENSSO as flux calibrator but there seems to be a scriptForImagingPrep.py to fix this."
		fi
	    else
		echo "NOTE: pipeline calibration used SSO $FORBIDDENSSO as flux calibrator. Please only proceed with delivery if you know that CASA $USECASAVERSION can handle this."
	    fi
	fi
    done
    cd ..
fi

if [ "a$CASAOPTION" == "a" ]; then 
    echo "scriptForPI.py should be executed using CASA version ${USECASAVERSION}" | tee -a $LOGFILE
else
    echo "scriptForPI.py should be executed using CASA version ${USECASAVERSION} with option $CASAOPTION" | tee -a $LOGFILE
fi

if [ "a$ANAME" == "a" ]; then
    echo "ERROR: No uid*.ms.scriptForCalibration.py or PPR*.xml found." | tee -a $LOGFILE
    exit 1
fi

RVAL=0
   
if [ ! "a"$SFPIDIFFERS == "a" ]; then
    echo "WARNING (repeated): found that scriptForPI.py differs from repository!"  | tee -a $LOGFILE
    RVAL=3
fi
if [ ! "a"$NOOUTFRAME == "a" ]; then
    echo "WARNING (repeated): scriptForImaging.py does not set outframe!" | tee -a $LOGFILE
    RVAL=3
fi
if [ ! "a"$CAS9131WARN == "a" ]; then
    echo "WARNING (repeated): This package was produced with CASA 4.7.0 but the scriptForImaging does not contain the line"  | tee -a $LOGFILE
    echo "   os.unsetenv('CASAVERSION') # necessary to avoid a CASA 4.7.0 bug which leads to a stray CR character in exported FITS"  | tee -a $LOGFILE
    echo "  This is probably OK since it seems no manual imaging was performed."
fi

echo "Done." | tee -a $LOGFILE
exit $RVAL


