# This script generates ALMA QA2 products for a
# pipeline-reduced dataset for the purposes of comparing the pipeline
# calibration to the manual calibration.  Requires ImageMagick's
# convert command to convert pngs to pdfs, and either pdftk or
# ghostscript to concatenate pdfs.
#
# - Todd Hunter, November 2012

ANALYZEMSCAL_REVISION_STRING = "$Id: analyzemscal.py,v 1.519 2017/04/29 16:51:15 thunter Exp $" 

global HEURISTICS_VERSION 
HEURISTICS_VERSION = 2013.09  # will get reset below 

import os, re, sys, time, datetime, pwd
import getpass
import textwrap
import fnmatch
import inspect
import difflib
import shutil
import pprint # for printing dictionaries
import math, functools
import numpy as np
import analysisUtils as au
import casadef
from taskinit import *  # needed for tb and casalog
from clearstat_cli import clearstat_cli as clearstat
try:
    from predictcomp_cli   import predictcomp_cli   as predictcomp
except:
    print "ERROR: predictcomp is not available as a task in this old version of casa"
    
from plotcal_cli   import plotcal_cli   as plotcal
from plotants_cli  import plotants_cli  as plotants
from plotms_cli    import plotms_cli    as plotms
from listobs_cli   import listobs_cli   as listobs
from flagcmd_cli   import flagcmd_cli   as flagcmd
try:
    from wvrgcal_cli   import wvrgcal_cli   as wvrgcal
except:
    print "ERROR: wvrgcal is not available as a task in this old version of casa"
try:
    from plotbandpass_cli import plotbandpass_cli  as plotbandpass
except:
    print "ERROR: plotbandpass is not available as a task in this version of casa"
import glob,commands
import pdb
import pylab as pl
use_plotants = False  # if True, try to run plotants; otherwise (or if it fails) use es.listobs2
rsyncPermissions = '--no-p'

# The following list is used to decide whether to query for the angular diameter
# of the flux calibrator, and to limit the uvrange when making spectrum plots.
SSO_LIST = ['mercury','venus','mars','ceres','pallas','vesta',
            'juno','jupiter','ganymede','callisto','io','europa',
            'saturn','titan','uranus','neptune','pluto']

if (casadef.casa_version < '4.0.0'):
    # prevent crash on import in casa < 4.0
    aftool = None

# Example execution:
def example(run=0, test=1):
    """
    Prints an example execution of the main function of this module.
    To simply run the example: copy and paste the following two lines:
import analyzemscal
analyzemscal.example(1)
    """
    if (run==2):
        cleanup = True
    else:
        cleanup = False
    if (run>0):
        # Band 9
        analyzemscal(asdm='uid___A002_X4b58a4_X31',
                     pipecaldir='/lustre/naasc/ahale/pipeline/root/2011.0.00747.S_2013_01_09T15_36_13.072/MOUS_uid___A001_X74_X130/working',
                     manualcaldir='/lustre/naasc/thunter/Cycle0/2011.0.00747.S/X31_cleanlog/',
                     webpagedir='', cleanup=cleanup, uvrange='0~50m')
    else:
        print """
import analyzemscal
analyzemscal.analyzemscal(asdm='uid___A002_X49990a_X1f',
  pipecaldir='/lustre/naasc/pipeproc/pipeline/root/2011.0.00958.S_2012_11_07T15_17_41.333/MOUS_uid___A002_X314ea1_X233/working',
  manualcaldir='',
  webpagedir='')

import analyzemscal
analyzemscal.analyzemscal(asdm='uid___A002_X383b50_Xc45',
  pipecaldir='/lustre/naasc/pipeproc/pipeline/root/2011.0.00010.S_2012_11_07T14_24_09.283/MOUS_uid___A002_X327408_X217/working',
  manualcaldir='',
  webpagedir='')

              """
def myCopyfile(a, b):
    copyIndent = ' '*10         
    println(lineno(),"%srsync from %s to %s" % (copyIndent, a, b))
    #shutil.copyfile(a,b)
    os.system('rsync -vau %s %s %s' % (rsyncPermissions,a,b))

def lineno():
    """Returns the current line number in our program."""
    return inspect.currentframe().f_back.f_lineno
        
def println(linenumber, mystring):
    global terminal
    mystring = "v%s_%04d: " % (version().split()[2], linenumber) + mystring
    print mystring
    if (len(mystring) > 0):
        mystring += '\n'
    try:  # check if global is defined
        terminal.write(mystring)
        terminal.flush()
    except:
        return
    
def help(match='', debug=False):
    """
    Print an alphabetized list of all the defined functions at the top level in
    analyzemscal.py.
    match: limit the list to those functions containing this string (case insensitive)
    -- Todd Hunter
    """
    myfile = __file__
    if (myfile[-1] == 'c'):
        if (debug): print "amc loaded from .pyc file, looking at .py file instead"
        myfile = myfile[:-1]
    aufile = open(myfile,'r')
    lines = aufile.readlines()
    if (debug): print "Read %d lines from %s" % (len(lines), __file__)
    aufile.close()
    commands = []
    for line in lines:
        if (line.find('def ') == 0):
            commandline = line.split('def ')[1]
            tokens = commandline.split('(')
            if (len(tokens) > 1):
                command = tokens[0]
            else:
                command = commandline.strip('\n\r').strip('\r\n')
            if (match == '' or command.lower().find(match.lower()) >= 0):
                commands.append(command)
    commands.sort()
    for command in commands:
        print command


def version():
    return(ANALYZEMSCAL_REVISION_STRING)
 
def heuristicsversion():
    global HEURISTICS_VERSION
    return(HEURISTICS_VERSION)

def setheuristicsversion(pipecaldir,pipelineYYYYMMDD):
    weblogp1=pipecaldir+pipelineYYYYMMDD+'/html/t1-1.html'
    if os.path.exists(weblogp1):
        f = open(weblogp1,'r')
        lines = f.readlines()
        f.close()
        plversion=0
        foundVer=False
        for line in lines:
            if not foundVer:
                if line.find('Pipeline Version') >= 0:
                    foundVer=True
            else:
                vstr=line.strip().replace("/","").replace("<td>","")
                vstr=vstr.split()[0] # they append the branch now
                if vstr.isdigit():
                    plversion=int(vstr)
                elif vstr[:-1].isdigit(): # in case of "M" at end
                    plversion=int(vstr[:-1])
                else:
                    print "ERROR: can't determine pipeline version from string '"+vstr+"' in file "+weblogp1
                foundVer=False        
        global HEURISTICS_VERSION
        if plversion >= 28322:
            HEURISTICS_VERSION=2014.02
        elif plversion >= 26428:
            HEURISTICS_VERSION=2013.09
        else: HEURISTICS_VERSION=2013.04 # or earlier - foundVer will be F
        print "Set pipeline heuristics version to "+str(HEURISTICS_VERSION)+" based on finding a version number of '%d'" % plversion
    else:
        print "ERROR: Can't determine pipeline version from "+weblogp1
 
def addplot(webpagedir, myplot, width=800, br=True, debug=False):
    """
    Copies a plot file into the web area, and adds the plot to the
    current html page
    """
    myplot0=myplot.split('/')[-1]
    if (br):
        brk = '<br>'
    else:
        brk = ''
    mystring = """<a href="%s"><img width="%d" src="%s"></img></a>%s\n\n""" % (myplot0, width, myplot0, brk)
    destination = webpagedir+myplot0
    myCopyfile(myplot, destination)
    return(mystring)

def addplotrow(webpagedir, title, plotForColumn1, plotForColumn2, width=400, 
               pdf1='', pdf2='', textForColumn1='', textForColumn2='', debug=False):
    """
    Copies up to 2 plot files into the web area, and adds a row to the current html table
    If pdf1 is not blank, then the link is made to the PDF file rather than the full-sized
    version of the png.
    """
    copyIndent = ' '*10         
    plot1="c1"+plotForColumn1.split('/')[-1]
    plot2="c2"+plotForColumn2.split('/')[-1]
    if (textForColumn1 != ''):
        textForColumn1 += '<br>'
    if (1==0):    # debugging use
        if (textForColumn1 == ''):
            textForColumn1 = plotForColumn1
        if (textForColumn2 == ''):
            textForColumn2 = plotForColumn2

    if (pdf1 == ''):
        if (plot1 == 'c1'):
            mystring = """<tr><td><b>%s</td><td>%s</td>\n\n""" % (title,textForColumn1)
        else:
            mystring = """<tr> <td><b>%s</td> <td>%s<br><a href="%s"><img width="%d" src="%s"></img></a></td>\n""" % (title, textForColumn1, plot1, width, plot1)
    else:
        mystring = """<tr> <td><b>%s</td> <td>%s<br><a href="%s"><img width="%d" src="%s"></img></a></td>\n""" % (title, textForColumn1, os.path.basename(pdf1), width, plot1)
    if (pdf2 == ''):
        if (plot2 == 'c2'):
            mystring += """<td>%s</td> </tr>\n\n""" % (textForColumn2)
        else:
            mystring += """<td>%s<br><a href="%s"><img width="%d" src="%s"></img></a></td> </tr>\n\n""" % (textForColumn2, plot2, width, plot2)
    else:
        mystring += """<td>%s<br><a href="%s"><img width="%d" src="%s"></img></a></td> </tr>\n\n""" % (textForColumn2, os.path.basename(pdf2), width, plot2)
    # Make sure that the plots and PDFs get copied to the webpage, automatically.
    if os.path.exists(plotForColumn1):
        destination = webpagedir+plot1
        myCopyfile(plotForColumn1, destination)
    elif (plotForColumn1 != ''):
        println(lineno(),"ERROR: plotForColumn1 = %s does not exist!" % (plotForColumn1))
    if os.path.exists(plotForColumn2):
        destination = webpagedir+plot2
        myCopyfile(plotForColumn2, destination)
    elif (plotForColumn2 != ''):
        println(lineno(),"ERROR: plotForColumn2 = %s does not exist!" % (plotForColumn2))
        
    if os.path.exists(pdf1):
        destination = webpagedir+os.path.basename(pdf1)
        myCopyfile(pdf1, destination)
    elif (pdf1 != ''):
        println(lineno(),"ERROR: pdf1 = %s does not exist!" % (pdf1))
    if os.path.exists(pdf2):
        destination = webpagedir+os.path.basename(pdf2)
        myCopyfile(pdf2, destination)
    elif (pdf2 != ''):
        println(lineno(),"ERROR: pdf2 = %s does not exist!" % (pdf2))
    return(mystring)

def findManualQA2ScienceTarget(mydir):
    textfile = mydir + '/textfile.txt'
    if (os.path.exists(textfile) != False):
        f = open(mydir + '/textfile.txt','r')
        lines = f.readlines()
        for line in lines:
            if (line.find('target') >= 0):
                tokens = line.split()
                f.close()
                return(tokens[2])
        f.close()
    return('?')
    
def findGainTable(pipecaldir, ms, searchString, searchString2=None, verbose=True, excludeString=None):
    """
    Search for the name of a cal table created by the pipeline.
    Returns just the file name, not the preceding path.
    Returns blank string if it does not find one.
    """
    gaintable = ''
    filelist = os.listdir(pipecaldir)
    if (verbose):
        println(lineno(),"Searching %s for a gaintable with %s and %s in the name" % (pipecaldir, ms, searchString))
    possibleTables = []
    for filename in filelist:
        if (filename.find(searchString) > 0 and
            filename.find(ms) >= 0 and 
            filename.find('.plot')<0 and filename.find('.pdf')!=len(filename)-4  
            and filename.find('.png')!=len(filename)-4):
            found=True
            if searchString2:
                found = filename.find(searchString2) > 0
            if found:
                if (excludeString == None):
                    possibleTables.append(filename)
                elif (filename.find(excludeString) < 0):
                    possibleTables.append(filename)
    if (len(possibleTables) > 0):
        gaintable = possibleTables[0]   # Just take the first one, for now
    if (verbose):
        println(lineno(),"  Found gaintable = %s" % (gaintable))

    return(gaintable)
    
def findTsysTable(pipecaldir, ms):
    """
    Search for the name of the Tsys cal table created by the pipeline.
    It is either .tsys. (pre May 2013) or .tsyscal. (May2013- forward)
    """
    gaintable = findGainTable(pipecaldir, ms, '.tsys')
    if (gaintable == ''):
        # for Sep 2013 but since there is only one tsyscal table we don't
        # need to be this specific, but could keep this working for May 
        # heuristics as well
        # gaintable = findGainTable(pipecaldir, ms, '.hif_tsyscal')
        gaintable = findGainTable(pipecaldir, ms, '.tsyscal')
        if (gaintable == ''):
            println(lineno(),"  ERROR: Could not find tsys table")
        else:
            println(lineno(),"  Found Tsys table = %s" % (gaintable))
    else:
        println(lineno(),"  Found Tsys table = %s" % (gaintable))
    return(gaintable)

def findPreBandpassGainTable(pipecaldir, ms, debug=False):
    """
    Search for the name of the pre-bandpass phase temporal gain cal
    table created by the pipeline.
    """
    if heuristicsversion()<2013.08:
        stage = 8
        # prevent it from finding the test table with wvr corrections applied
        gaintable = findGainTable(pipecaldir, ms, '.gpcal.',searchString2='s%d'%stage, excludeString='wvr.')
        while (gaintable == '' and stage<12):
            println(lineno(),".........WARNING: Did not find s%d*.gpcal, searching for s%d*gpcal....."%(stage,stage+1))
            stage += 1
            gaintable = findGainTable(pipecaldir, ms, '.gpcal.',searchString2='s%d'%stage, excludeString='wvr.')

    else:
        if heuristicsversion()>=2014.02:
            bptaskname="hif_bandpass"   # why they didn't change this one is a mystery
        else:
            bptaskname="hif_bandpass"
        # only works post-sep 2013:
        gaintable = findGainTable(pipecaldir, ms, '.gpcal.',searchString2=bptaskname, excludeString='wvr.')

    if (gaintable == ''):
        println(lineno(),"ERROR: I can't find the pre-bandpass phase temporal table!")
        if debug: pdb.set_trace()
        println # TODO make script exit here
    else:
        println(lineno(),"  Found pre-bandpass phase temporal table = %s" % (gaintable))
    return(gaintable)
    
def findBandpassGainTable(pipecaldir, ms):
    """
    Search for the name of the bandpass gain cal table created by the pipeline.
    Here we want the later one (typically s9), not the earlier one (typically s3).
    """
    if (pipecaldir == '.'):
        pipecaldir = os.getcwd() + '/'

    if heuristicsversion()<2013.08:
        searchString = pipecaldir+'%s*.bcal.s*.tbl' % (ms)
        print "search string = %s" % (searchString)
        dirlist = sorted(glob.glob(searchString), key=os.path.getmtime, reverse=True)
        maxstage = 0
        gaintable = ''
        for f in dirlist:
            stage = int(f.split('.bcal.s')[1].split('_')[0])
            if (stage > maxstage):
                maxstage = stage
                gaintable = f
    else:
        if heuristicsversion()>=2014.02:
            bptaskname="hif_bandpass" # dunno why its not hifa_bandpass
        else:
            bptaskname="hif_bandpass"
        gaintable=findGainTable(pipecaldir, ms, '.bcal.',searchString2=bptaskname, excludeString='plotCal')
            
    if (gaintable == ''):
        println(lineno(),"ERROR:  $$$$$$$$$$ Did not find a bandpass table (.bcal.)")
    else:
        println(lineno(),"  Found bandpass table = %s" % (gaintable))
    return(pipecaldir+"/"+gaintable)
    
def findManualBandpassTable(caldir, ms):
    """
    Search for the name of the manual bandpass gain cal table to be used.
    First search for _smooth if present, otherwise use the default.
    """
    gaintable = findGainTable(caldir, ms, '.bandpass_smooth')
    if (gaintable == ''):
        gaintable = findGainTable(caldir, ms, '.bandpass')
    if (gaintable == ''):
        println(lineno(),"ERROR: $$$$$$$$$$ Did not find a manual bandpass table (.bandpass)")
    else:
        println(lineno(),"  Found bandpass table = %s" % (gaintable))
    return(gaintable)

def findAntGainCheckCalTable(pipecaldir, ms, minstage=10, maxstage=11):
    if heuristicsversion()<2013.08:
        for stage in range(minstage,maxstage+1):
            gaintable = findGainTable(pipecaldir, ms, 'solintinf.gacal.s%d' % (stage))
            if (gaintable != ''):
                println(lineno(),"  Found ant_gain_check gain table = %s" % (gaintable))
                break
    else:
        gaintable=findGainTable(pipecaldir, ms, '.solintinf.gacal.',searchString2='gfluxscale',excludeString='plots')            
    if (gaintable == ''):
        println(lineno(),"ERROR:  Did not find ant_gain_check gain table (solintinf.gacal.s%d..%d)" % (minstage,maxstage))
    return(gaintable)

def findFirstGACal(pipecaldir, ms, debug=False):
    """
    Finds the pipeline stage that produced the first .gacal. file, which
    is normally hif_newgaincal().
    """
    if (pipecaldir[-1] != '/'):
        pipecaldir += '/'
    filelist = glob.glob(pipecaldir+ms+'*.gacal.*')
    if (debug): print "filelist = ", filelist
    minstage = 10000
    maxstage = 0
    stages = []
    for f in filelist:
        if heuristicsversion()>2013.08:
            ff=f.split('/')[-1]
            stage = int( (ff.split('.')[3]).split('_')[0][1:] )
        else:
            stage = int(f[f.find('.gacal.')+8:].split('_')[0])        
        if (debug): print f
        stages.append(stage)
        if (stage < minstage):
            minstage = stage
        if (stage > maxstage):
            maxstage = stage
    if heuristicsversion()>2013.08:
        maxstages=3
    else:
        maxstages=2
    if (len(np.unique(stages)) > maxstages):
        println(lineno(),"WARNING: More than %d stages (%d)show a .gacal. table.  Did you re-run the pipeline with a different number of commands/stages?" % (maxstages,len(np.unique(stages))))
    if heuristicsversion()>2013.08:
        filelist = glob.glob(pipecaldir+ms+'*s%d_*.gacal.tbl' % (minstage))
    else:
        filelist = glob.glob(pipecaldir+ms+'*.gacal.s%d_*' % (minstage))
    if (len(filelist) < 1):
        print "ERROR: Did not find a .gacal. file.  The pipeline run must have been unsuccessful."
        return(None, None)
    firstGACal = filelist[0]
    println(lineno(),"Found first .gacal. file = %s" % (firstGACal))
    return(minstage, maxstage)
    
def findPhaseTemporalGainTableAllCals(pipecaldir, ms, maxstage=12):
    """
    Search for the name of the phase temporal gain cal table created
    by the pipeline.
    """
    if heuristicsversion()<2013.08:
        # gpcal = 'p', gpcal.s9 is 'int' but only has the bandpass calibrator
        # gacal = 'a', gacal.s10 is 'inf' and all cals, but only subset of antennas
        # gpcal = 'p', gpcal.s10 is 'int' and all cals, but only subset of antennas
        # gcal = 'ap', gcal.s11 is 'inf' and all cals, and all antennas
        gaintable = ''
        for stage in range(maxstage,maxstage-4,-1): # [12,11,10,9]:
            gaintable = findGainTable(pipecaldir, ms, 'solintint.gpcal.s%d' % stage)  
            if (gaintable != ''):
                println(lineno(),"  Found phase temporal table (AllCals) = %s" % (gaintable))
                break
        if (gaintable == ''): println(lineno(),"ERROR: Could not find solintint.gpcal.s%d..%d" % (maxstage,maxstage-3))
    else:
        if heuristicsversion()>=2014.02:
            gctaskname="hifa_timegaincal"
        else:
            gctaskname="hif_timegaincal"
        gaintable = findGainTable(pipecaldir, ms, 'solintint.gpcal',searchString2=gctaskname)  
        if (gaintable == ''): println(lineno(),"ERROR: Could not find "+gctaskname+"*solintint.gpcal")
    return(gaintable)
    
def findPhaseTemporalGainTablePhaseCal(pipecaldir, ms, maxstage=12, debug=False):
    """
    Search for the name of the phase temporal gain cal table created
    by the pipeline.
    """
    # gpcal = 'p', gpcal.s9 is 'int' but only has the bandpass calibrator
    # gacal = 'a', gacal.s10 is 'inf' and all cals, but only subset of antennas
    # gpcal = 'p', gpcal.s10 is 'int' and all cals, but only subset of antennas
    # gcal = 'ap', gcal.s11 is 'inf' and all cals, and all antennas
    gaintable = ''
    if heuristicsversion()<2013.08:
        for stage in range(maxstage,maxstage-4,-1): # [12,11,10,9]:
            if (debug):
                print "Checking stage %d" % (stage)
            gaintable = findGainTable(pipecaldir, ms, '.gpcal.s%d'%stage)  
            if (gaintable != ''):
                println(lineno(),"  Found phase temporal table (PhaseCal) = %s" % (gaintable))
                break
        if (gaintable == ''): println(lineno(),"ERROR: Could not find gpcal.s%d..%d" % (maxstage,maxstage-3))
    else:
        if heuristicsversion()>=2014.02:
            gctaskname="hifa_timegaincal"
        else:
            gctaskname="hif_timegaincal"

        gaintable = findGainTable(pipecaldir, ms, '.gpcal',searchString2=gctaskname)
        if (gaintable == ''): println(lineno(),"ERROR: Could not find "+gctaskname+"*gpcal")
    return(gaintable)
    
def findPipelineAmpTemporalGainTablePhaseCal(pipecaldir, ms, maxstage=12):
    """
    Search for the name of the amp temporal gain cal table that
    includes all antennas (created by the pipeline), without any preceeding path.
    """    
    gaintable = ''
    if heuristicsversion()<2013.08:
        for stage in range(maxstage,maxstage-4,-1): # [12,11,10,9]:
            gaintable = findGainTable(pipecaldir, ms, '.solintinf.gacal.s%d' % stage) 
            if (gaintable != ''):
                println(lineno(),"  Found amp temporal table = %s" % (gaintable))
                break
        if  (gaintable == ''): println(lineno(),"ERROR: Could not find .solintinf.gacal.s%d..%d" % (maxstage,maxstage-3))
    else:
        if heuristicsversion()>=2014.02:
            gctaskname="hifa_timegaincal"
        else:
            gctaskname="hif_timegaincal"

        gaintable = findGainTable(pipecaldir, ms, '.solintinf.gacal', searchString2=gctaskname)
        if  (gaintable == ''): println(lineno(),"ERROR: Could not find "+gctaskname+"*.solintinf.gacal")
    return(gaintable)
    
def findPipelineAmpTemporalGainTableAllCals(pipecaldir, ms, maxstage):
    """
    Search for the name of the amp temporal gain cal table that
    includes all antennas (created by the pipeline), without any preceeding path.
    """
    return(findPipelineAmpTemporalGainTablePhaseCal(pipecaldir, ms, maxstage))
    
def findWvrGainTable(pipecaldir, ms):
    """
    Search for the name of the wvr gain cal table created by the pipeline.
    """
    if heuristicsversion()<2013.08:
        gaintable = findGainTable(pipecaldir, ms, '.wvr.')
    else:
        gaintable = findGainTable(pipecaldir, ms, '.wvrcal.')
    if (gaintable == ''):
        println(lineno(),"  Found wvr table = %s" % (gaintable))
    return(gaintable)

def findOverlayTimeAntennaPlot(mydir, caltableNoPath):
    """
    Locates the plotbandpass plot of overlay='time,antenna' because it
    will be of the format: CALTABLE.spwX.tY.png
    by looking in directory 'mydir' for files that include the name
    'caltableNoPath'
    """
    myplot = ''
    pnglist = os.listdir(mydir)
    println(lineno(),"findOverlayTimeAntennaPlot() pnglist = %s" % (pnglist))
    for p in range(len(pnglist)):
        if (pnglist[p].find(caltableNoPath+'.spw') >= 0):
            myplot = mydir + pnglist[p]
            break
    return(myplot)

def findAntennasUsedForFluxCalibrationFromManualScript(script):
    """
    Searches the manual calibration script for the gain calibration on the flux
    calibrator and extracts the antenna argument.
    """
    vis = open(script,'r')
    subset = None
    foundGaincal = False
    foundGaincalStep = False
    for line in vis.readlines():
        if (foundGaincal):
            if (line.find('antenna =') >= 0):
                subset = line.split('=')[1].split('#')[0].strip()
                break
            elif (line.find('checkCalTable') >= 0):
                foundGaincal = False
                foundGaincalStep = False
        elif (foundGaincalStep):
            if (line.find('gaincal(') >= 0):
                foundGaincal = True
        elif (line.find('# Gain calibration') >= 0):
            foundGaincalStep = True
    vis.close()
    if (subset == None):
        println(lineno(),"ERROR: Did not find antenna subset for gaincal on flux calibrator in manual script.")
        subset = ''
    return(subset)
    
def findAntennasUsedForFluxCalibrationFromLogfile(logfile, asdmNoPath, debug=False):
    """
    Determine which subset of antennas was used for gaincal on the flux
    calibrator by parsing the casapy log file.  Yes, this is a kludge until
    we figure out where else we can find this information.
    """
#    debug = True
    if (logfile == None):
        return('')
    casalogfile = open(logfile,"r")
    if (debug): println(lineno(),"Searching file %s" % (logfile))
    lines = casalogfile.readlines()
    antennas = ''
    if heuristicsversion()>=2014.02:
        gftaskname="hifa_gfluxscale"
    else:
        gftaskname="hif_gfluxscale"

    for line in lines:
        if (((line.find('intent="*PHASE*,*AMPLI*"') > 0) or  # Jan 2013 heuristics
             (line.find("intent='*AMPLI*'") > 0) or          # Mar 2013 heuristics
             (line.find('intent="*AMPLI*"') > 0) or          # Mar 2013 heuristics
             (line.find("intent='*FLUX*,*AMPLI*'") > 0) or   # Mar 2013 heuristics when both intents present
             (line.find('intent="*FLUX*,*AMPLI*"') > 0) or   # Mar 2013 heuristics (just in case)
             (line.find("intent='*PHASE*,*AMPLI*'") > 0))    # Jan 2013 heuristics
            and line.find(gftaskname) >= 0 
            and ((line.find(asdmNoPath) > 0) or heuristicsversion()>2013.08)):
            tokens = line.split()
            if (debug): println(lineno(),"tokens = %s" % tokens)
            if (tokens[4].find('antenna') > 0):
                # Nov 2012 products version
                antenna = tokens[4].split(',antenna="')[-1]
                antennas = antenna.split('"')[0]
                if (antennas[-1] == '&'):
                    antennas = antennas[:-1]
            else:
                # Jan 2013 products version (where token[4] = 'Executing')
                for t in tokens:
                    if (debug): println(lineno(),"Searching for antenna in %s" % t)
                    if (t.find('antenna') >= 0):
                        antenna = t.split("antenna='")[-1]
                        if (debug): println(lineno(),"antenna = %s" % antenna)
                        # antenna now looks like:  DV19,DV20,DV08,DV21,DV04,DV12,DA41,DA42&',",  or 0~30'
                        antennas = antenna.split('&')[0]
                        # antennas now looks like:  DV19,DV20,DV08,DV21,DV04,DV12,DA41,DA42   or 0~30'
                        break
            if (debug): println(lineno(),"antennas = %s" % antennas)
            antennas = antennas.split("'")[0]
            if (debug): println(lineno(),"antennas = %s" % antennas)
            # antennas now looks like:  DV19,DV20,DV08,DV21,DV04,DV12,DA41,DA42   or 0~30
            println(lineno(),"Found antennas for flux cal in %s" % (logfile))
            break
    casalogfile.close()
    return(antennas)
    
def findRefAnt_grep(pipecaldir):
    """
    Determine the reference antenna used by the pipeline and return it, by
    looking in gaincal.last.  Note: this will fail if you calibrate more
    than one ms in the same directory!
    """
    lastfile = 'gaincal.last'
    infile = open(pipecaldir + lastfile, 'r')
    lines = infile.readlines()
    antlist = '0'  # use first antenna in case refant is not found
    for line in lines:
        line = re.sub(r'\s','',line)
        line = re.sub(r'"','',line)
        if (line.find("refant=") >= 0):
            antlist = line.split('=')[1]
            break
    infile.close()
    refAnt = antlist.split(',')[0]
    println(lineno(),"  Found first refAnt=%s in %s" % (refAnt,lastfile))
    return(refAnt)

def findCasaVersionFromLogfile(logfile):
    """
    Reads the CASA version from the specified log file
    """
    if (logfile == None):
        return "ERROR: Did not find a log file with an applycal"
    f = open(logfile,'r')
    lines = f.readlines()
    casaVersion = None 
    for line in lines:
        location = line.find('CASA Version')
        if (location > 0):
            casaVersion = line[location:-1]
            f.close()
            return(casaVersion)

def tailPipelineCasaLogfile(mydir, arg='-f', asdmNoPath=None, ignoredir='html', grep=''):
    """
    Finds the newest casapy logfile in the specified directory tree, then runs
    "tail -f" on it. It is usually of the format: casa-YYYYMMDD-HHMMSS.log. 
    asdmNoPath: if specified, then find the one that contains "applycal(" and 
        the specified asdm on the same line. This avoids the chafe of the lots 
        of little log files created when restarting casa frequently that may
        be newer than the actual pipeline run.
    ignoredir: a string to ignore in the directory name when recursively
              searching directories
    grep: pipe the result to grep on this string (e.g. 'Begin ')
    """
    mydir = mydir.replace('file://','')
    print "Searching for file..."
    logfile = findPipelineCasaLogfile(mydir,asdmNoPath,ignoredir)
    if (len(logfile) < 1):
        return
    else:
        if (len(grep) > 0):
            cmd = "tail %s %s | grep %s" % (arg,logfile,grep)
        else:
            cmd = "tail %s %s" % (arg,logfile)
        print cmd
        if (arg.find('-f') >= 0):
            print "Use control-C to return to the CASA prompt.\n"
        os.system(cmd)

def findPipelineCasaLogfile(mydir, asdmNoPath=None, ignoredir='html',
                            verbose=False, searchString='Analyzing pipeline',
                            sortByModificationTime=False):
    """
    Finds the newest casapy logfile in the specified directory tree that contains
    "Analyzing pipeline".  It is usually of the format:  casa-YYYYMMDD-HHMMSS.log. 
    asdmNoPath: if specified, then find the one that contains "applycal(" and 
        the specified asdm on the same line. This avoids the chafe of the lots 
        of little log files created when restarting casa frequently that may
        be newer than the actual pipeline run.
    searchString:  set to "Beginning pipeline" to find the initial importdata log
    sortByModificationTime: if False, then sort by name (which is creation time)
    """
    debug = False
    if (mydir==''):
        mydir = './'
    if (mydir[-1] != '/'):
        mydir += '/'
    # Make a list with the most recent file first
    if verbose:
        println(lineno(),"Search dir = %s" % mydir)
    if sortByModificationTime:
        # old style
        searchString = mydir+'*casa*.log'
        # The following would work to search the tree in python 3.5+.
        #    searchString = mydir+'**/*casa*.log'
        if verbose:
            println(lineno(),"Search string = %s" % searchString)
        dirlist = sorted(glob.glob(searchString), key=os.path.getmtime, reverse=True)
    else:
        dirlist = []
        for root, dirnames, filenames in os.walk(mydir):
            for filename in fnmatch.filter(filenames, 'casa*.log'):
                if (ignoredir == ''):
                    dirlist.append(os.path.join(root, filename))
                elif (root.find(ignoredir) < 0):
                    dirlist.append(os.path.join(root, filename))
        # The following sorts by filename, so latest one created will get returned
        dirlist = sorted(dirlist, reverse=True)
    if verbose:
        println(lineno(),"dirlist = %s"% dirlist)
    if (dirlist == []):
        println(lineno(),'WARNING: unknown (no log file found in %s)'%(mydir))
        return None
    else:
        newdirlist = []
        for mydir in dirlist:
            if (au.grep(mydir,searchString)[0] != ''):
                newdirlist.append(mydir)
        dirlist = newdirlist
        if (asdmNoPath == None):
            if (len(dirlist) < 1):
                print "No files found containing: ", searchString
                return ''
            else:
                return(dirlist[0])
        for logfile in dirlist:
            if (debug): print "Checking logfile = %s" % (logfile)
            f = open(logfile,'r')
            lines = f.readlines()
            f.close()
            for line in lines:
                if (line.find('applycal(')>0):
                    if (debug): print "    Checking line = %s" % (line)
                    if (line.find(asdmNoPath) > 0):
                        print "Found most recent logfile containing applycal(%s): %s" % (asdmNoPath, logfile)
                        return(logfile)
    println(lineno(),"EROR: I didn't find a pipeline casa log with an applycal(%s)!" % (asdmNoPath))
    return None

def tailCalibPipeLogfile(mydir, arg='-f'):
    """
    Finds the newest casaPipeIF.*log file in the specified directory tree, then runs
    "tail -f" on it. 
    """
    mydir = mydir.replace('file://','')
    print "Searching for file..."
    logfile = findCalibPipeLogfile(mydir)
    if (len(logfile) < 1):
        return
    else:
        print "tail %s %s" % (arg,logfile)
        if (arg.find('-f') >= 0):
            print "Use control-C to return to the CASA prompt.\n"
        os.system('tail %s %s' % (arg,logfile))

def computeDurationCalibPipeLogfile(logfile, units='minutes'):
    """
    logfile: a calibPipeIF logfile, found in the MOUS directory
    units: 'minutes' (default) or 'hours'
    """
    with open(logfile, "rb") as f:
        first = f.readline()     # Read the first line.
        for last in f: pass      # Loop through the whole file reading it all.
    startDatestamp = first.split(': ')[0]
    endDatestamp = last.split(': ')[0]
    durationMinutes = au.dateStringDifference(startDatestamp, endDatestamp)
    if units == 'hours':
        duration = durationMinutes/60.
    else:
        duration = durationMinutes
    return duration

def pipelineDuration(mydir, units='minutes', verbose=False):
    """
    Finds the calibPipeIF log file and computes the duration from
    the timestamps.
    """
    if (mydir[-4:] == '.log'):
        logfile = mydir
    else:
        logfile = findCalibPipeLogfile(mydir)
    if verbose: print "Found logfile = ", logfile
    if logfile == None: return
    duration = computeDurationCalibPipeLogfile(logfile, units)
    return duration

def findCalibPipeLogfile(mydir, verbose=False):
    """
    Finds the newest calibPipeIF.*log file in the specified directory tree.
    """
    debug = False
    if (mydir==''):
        mydir = './'
    if (mydir[-1] != '/'):
        mydir += '/'
    # Make a list with the most recent file first
    if verbose:
        println(lineno(),"Search dir = %s" % mydir)
    if False:
        # old style
        searchString = mydir+'calibPipeIF.*log'
        # The following would work to search the tree in python 3.5+.
        #    searchString = mydir+'**/*casa*.log'
        if verbose:
            println(lineno(),"Search string = %s" % searchString)
        dirlist = sorted(glob.glob(searchString), key=os.path.getmtime, reverse=True)
    else:
        dirlist = []
        for root, dirnames, filenames in os.walk(mydir):
            for filename in fnmatch.filter(filenames, 'calibPipeIF.*log'):
                dirlist.append(os.path.join(root, filename))
    if verbose:
        println(lineno(),"dirlist = %s"% dirlist)
    if (dirlist == []):
        println(lineno(),'WARNING: unknown (no log file found in %s)'%(mydir))
        return None
    else:
        return(dirlist[0])

def findCasaVersion(mydir):
    """
    Looks for a casapy*.log files in the specified directory, finds the first
    appearance of an applycal command, then reports the version number string
    found immediately prior to it, and reports it.  The format is:
    'CASA Version 3.4.0 (release r19988)'
    """
    if (mydir==''):
        mydir = './'
    if (mydir[-1] != '/'):
        mydir += '/'
    dirlist = sorted(glob.glob(mydir+'*casapy*.log'), key=os.path.getmtime, reverse=True)
    if (dirlist == []):
        return('unknown (no log file found in %s)'%(mydir))
    else:
        for logfile in dirlist:
            f = open(logfile,'r')
            lines = f.readlines()
            casaVersion = None 
            for line in lines:
                location = line.find('CASA Version')
                if (location > 0):
#                    println(lineno(),"Saw %s in %s" % (line[location:-1], logfile))
                    casaVersion = line[location:-1]
                elif (line.find('Begin Task: applycal')>0 and casaVersion!=None):
                    f.close()
                    return(casaVersion)
            f.close()
        return('unknown (found but failed to parse the log file)')

def findRefAntFromCCL(ccl):
    """
    Finds the final occurrence of the refant string in the casa_commands.log
    and returns it.  The final value is used because it might change during
    execution of the pipeline.
    """
    print "findRefAntFromCCL: ccl=%s" % (ccl)
    f = open(ccl,'r')
    lines = f.readlines()
    f.close()
    antString = ''
    inaGaincal=False
    for line in lines:
        if not inaGaincal:
            if line.find('gaincal(') >= 0:
                inaGaincal=True
        if inaGaincal:
            if line.find(')') >= 0:
                inaGaincal=False
        if inaGaincal:
            if line.find('refant=') >= 0:
                startAntString = line.find('refant')
                antString = line[startAntString:].split("=")[1]
                if (len(antString.split("'")) > 1):
                    antString = antString.split("'")[1]
                elif (len(antString.split('"')) > 1):
                    antString = antString.split('"')[1]
                else:
                    continue
#                print "Found refant=%s" % (antString)
    return(antString)
    
def findRefAnt(pipecaldir):
    """
    Determine the reference antenna used by the pipeline and return it.
    """
    # when h_resume returns the context we can get refant from 
    # context.observing_run.get_measurement_sets()[0].reference_antenna
    if pipeContext(pipecaldir):
        from hif_refant import hif_refant
        refant=hif_refant()
        return(refant._refant)
    else:
        return(findRefAnt_grep(pipecaldir))

def build_amc_exec_list(prefix='/home/safe.nrao.edu/content/alma/PipelineTestResults/Sep2013_heuristics/', outname='amc_exec.py',searchString='', webarea=True, webpagedir='Sep2013_heuristics', otherParameters=''):
    """
    Builds a list of analyzemscal commands corresponding to a list of ms's in
    a pipeline test results area by searching the html results table. The
    searchString parameter can be used to limit the list to those html files
    that contain the specified string (such as a username).
      webarea, webpagedir: parameters passed to analyemscal
      otherParameters: a string containing a list of other parameters, with
         internal comma delimiters, e.g. "refant='DV03', manualcaldir='mydir'"
    """
    build_ms_list(prefix, outname, searchString)
    f = open(outname,'r')
    lines = f.readlines()
    f.close()
    f = open(outname,'w')
    for line in lines:
        vis = line.split('\n')[0]
        asdm = os.path.basename(vis).split('.ms')[0]
        pipecaldir = os.path.dirname(vis)
        cmd = "amc.analyzemscal(asdm='%s', pipecaldir='%s', webarea=%s, webpagedir='%s'" % (asdm, pipecaldir, webarea, webpagedir)
        if (len(otherParameters) > 0):
            cmd += ', %s' % (otherParameters)
        cmd += ')\n'
        f.write(cmd)
    print "list of %d amc commands left in = %s" % (len(lines),outname)
    f.close()

def build_ms_list(prefix='/home/safe.nrao.edu/content/alma/PipelineTestResults/Sep2013_heuristics/', outname='mslist.txt',searchString=''):
    """
    Builds a list of ms's in a pipeline test results area by searching the
    html results table. The searchString parameter can be used to limit the
    list to those html files that contain the specified string (such as a
    username).
    """
    dirs = glob.glob(prefix+'uid*')
    mslist = []
    output = open(outname,'w')
    for dir in dirs:
        pipecaldir = ''
        foundSearchString = False
        if (os.path.exists(dir+'/index.html')):
            f = open(dir+'/index.html')
            lines = f.readlines()
            for line in lines:
                if (line.find('Pipeline reduction directory')>=0):
                    pipecaldir = line.split()[4].split('<br>')[0]
                if (line.find(searchString) >= 0):
                    foundSearchString = True
                    if (pipecaldir != ''): break
            f.close()
            if (foundSearchString):
                mslist.append(pipecaldir+'/'+os.path.basename(dir)+'.ms')
                output.write(mslist[-1]+'\n')
    output.close()
    print "list of %d ms left in file = %s" % (len(mslist),outname)

def checkFieldTables(file, verbose=True):
    """
    Takes a file with a list of measurement sets (full paths, one per line)
    and runs checkFieldTable on them and prints the results.
    verbose: False means only print the offenders (multiple SOURCE_ID per FIELD name)
    """
    f = open(file,'r')
    vislist = f.readlines()
    f.close()
    results = {}
    for vis in vislist:
        myvis = vis[:-1] # strip off the newline
        results[vis] = checkFieldTable(myvis)
        if (results[vis] == None):
            if (verbose):
                print "done %s (no FIELD table found)" % (os.path.basename(myvis))
        elif (len(results[vis]) > 0):
            print "done %s (offender: %s)" % (os.path.basename(myvis), str(results[vis]))
        elif (verbose):
            print "done %s" % (os.path.basename(myvis))
    return(results)
    
def checkFieldTable(vis):
    """
    Checks the FIELD table to see if any field name points to multiple SOURCE IDs,
    and returns a dictionary of fields that do so: key=fieldName, value=sourceID.
    """
    offenders = {}
    if (os.path.exists(vis+'/FIELD')==False):
        print "Dataset/table not found."
        return None
    mytb = createCasaTool(tbtool)
    mytb.open(vis+'/FIELD')
    names = mytb.getcol('NAME')
    sources = mytb.getcol('SOURCE_ID')
    for name in names:
        subtable = mytb.query('NAME == "%s"' % name)
        src = np.unique(subtable.getcol('SOURCE_ID'))
        subtable.close()
        if (len(src) > 1):
            offenders[name] = src
    mytb.close()
    return offenders

def indexContainingSubstring(the_list, substring):
    """
    Used to find which png file from a list contains two different
    strings, such as the reference antenna.
    """
    for i, s in enumerate(the_list):
        if (substring in s):
            return i
    return -1

def indexContainingSubstrings(the_list, substring1, substring2):
    """
    Used to find which png file from a list contains two different
    strings, such as the reference antenna, and the word 'phase'.
    """
    for i, s in enumerate(the_list):
        if ((substring1 in s) and (substring2 in s)):
            return i
    return -1

def indexContaining3Substrings(the_list, substring1, substring2, substring3):
    """
    Used to find which png file from a list contains three different
    strings, such as the reference antenna, and the word 'phase'.
    """
    for i, s in enumerate(the_list):
        if ((substring1 in s) and (substring2 in s) and (substring3 in s)):
            return i
    return -1

def pipelineHostname(mousdir):
    """
    Starting in pipeline 5.1p1 (Oct 2017), the hostname will appear in the casa log.
    mousdir: either the MOUS dir or any subdir thereof down to 'working'
    """
    casalog = findPipelineCasaLogfile(mousdir)
    if casalog == None:
        return
    result = au.grep(casalog,' running on ')[0]
    if result != '':
        hostname = result.split(' running on ')[1]
    else:
        hostname = 'unknown'
    return hostname
    
def pipeContext(pipecaldir):
    """
    Load pipeline context.  Return True if successful or already loaded.
    """
    # we should use a global variable with the previously loaded context file, 
    # to avoid reloading this over and over
    from h_resume import h_resume
    if not sys.modules.has_key('task_h_resume'):
       println(lineno(),"WARNING: Pipeline tasks not loaded - will attempt to continue but YMMV")
       return False

    contextfile = ''
    for f in os.listdir(pipecaldir): 
       if (f.find('context') >= 0):
          contextfile = pipecaldir + f
          break
    try:
       h_resume(contextfile) # nov 2012 version is broken and eventually the task should return the actual context object but right now it doesn't
       return True
    except:
       println(lineno(),"WARNING: Pipeline context cannot be loaded - using fragile 'grep' methods")
       return False


def getTsysChanRangeToPlot(ms):
    """
    Determines the channel range to pass to plotbandpass so that the resulting
    plots are not dominated by huge values at the edges of the band.
    """
    tsysNumChans = []
    es = au.stuffForScienceDataReduction()
    tsysSpwInfo = es.getSpwInfo(ms, intent='CALIBRATE_ATMOSPHERE')
    for i in tsysSpwInfo:
        tsysNumChans.append(tsysSpwInfo[i]['numChans'])
    tsysNumChans = sorted(dict.fromkeys(tsysNumChans).keys())
    if (len(tsysNumChans) != 1): 
        println(lineno(),'ERROR: Configuration not supported, assuming 128 channels.')
        tsysNumChans = [128]
    tsysNumChans = tsysNumChans[0]
    chanEdge = 0.0390625
    startChan = int(tsysNumChans * chanEdge)
    endChan = int(tsysNumChans * (1-chanEdge))
    chanrange = str(startChan)+'~'+str(endChan)
    return(chanrange)

def buildPngList(mydir):
    """
    Builds a list of .png files in the specified directory.
    """
    if (mydir[-1] != '/'):
        mydir += '/'
    pnglist = glob.glob(mydir+'*.png')  # glob.glob returns the full path for each file
    for p in range(len(pnglist)):
        pnglist[p] = pnglist[p]
    return(pnglist)

def buildFileListWildcard(mystr):
    """
    Builds a list of filenames whose names contain the specified string.
    """
    pnglist = glob.glob(mystr)  # glob.glob returns the full path for each file
    for p in range(len(pnglist)):
        pnglist[p] = pnglist[p]
    return(sorted(pnglist))

def createCasaTool(mytool):
    """
    A wrapper to handle the changing ways in which casa tools are invoked.
    Todd Hunter
    """
    if (type(casac.Quantity) != type):  # casa 4.x
        myt = mytool()
    else:  # casa 3.x
        myt = mytool.create()
    return(myt)

def determineSolutionType(caltable):
    """
    Open a caltable to determine the solution type (T or G).
    """
    println(lineno(),"Opening %s" % (caltable))
    mytb = createCasaTool(tbtool)
    mytb.open(caltable)
    names = mytb.colnames()
    if ('SPECTRAL_WINDOW_ID' in names):  # this is a table from >= 3.4
        soltype = len(mytb.getcell('CPARAM',0))
    else:
        soltype = len(mytb.getcell('GAIN',0))
    if (soltype == 1):
        ampGainType = 'T solution'
    elif (soltype == 2):
        ampGainType = 'G solution'
    else:
        ampGainType = 'Unknown solution type (len=%d)' % (soltype)
    mytb.close()
    return ampGainType

def computeMaxGainIgnoring1(cparam):
    cparam = np.abs(cparam.flatten())
    indices = np.where(np.round(cparam*1000000.)/1000000. != 1.0)[0]
    cparam = cparam[indices]
    if (len(cparam) == 0):
        return(1.0)  # in case all values are 1.0
    else:
        return(np.max(cparam))

def callCheckCalTable(calTableName, msName='', interactive=False, subplot=411, fontsize=5.0,temppath=""):
    """
    A wrapper for es.checkCalTable() to prevent file permission problems.
    """
    if not os.path.exists(temppath): temppath = 'tmp'
    if (os.access(calTableName, os.W_OK) == False):
        if (os.path.exists(temppath) == False):
            mkdirGroupWrite(temppath)
        println(lineno(),"Copying %s to %s/ to avoid permission problem..........." % (os.path.basename(calTableName),temppath))
        os.system('rsync -vau %s %s %s' % (rsyncPermissions, calTableName,temppath))
        if (os.path.exists('%s/%s' % (temppath,os.path.basename(msName))) == False):
            softlink = 'ln -s %s %s/%s' % (msName, temppath, os.path.basename(msName))
            println(lineno(),'Creating softlink: %s' % (softlink))
            os.system(softlink)
        calTableName = temppath + '/' + os.path.basename(calTableName)
        mytb = createCasaTool(tbtool)
        mytb.open(calTableName, nomodify=False)
        msName = temppath+'/'+os.path.basename(msName)
        mytb.putkeyword('MSName',msName)
        mytb.close()
        removeTemporary = True
    else:
        removeTemporary = False
#        println(lineno(),"I appear to have Write access to %s............." % (calTableName))
        plotdir = calTableName + '.plots'
    es = au.stuffForScienceDataReduction()
    println(lineno(),"Calling es.checkCalTable('%s', '%s', %s, %s, fontsize=%.1f)" % (calTableName,msName,interactive,str(subplot),fontsize))
    es.checkCalTable(calTableName, msName, interactive, subplot, fontsize=fontsize)
    if (removeTemporary and False):
        os.system('rm -rf %s' % (calTableName))
    plotdir = calTableName + '.plots/'
    return(plotdir)
    
def checkCalTableVariant(calTableName, msName='', interactive=False, subplot=411,
                         gaintype='a', plotfile='', fontsize=6.0, temppath=""):
    """
    Meant for G Jones or T Jones only.  Overlays all antennas on one N-panel plot where
    N is the number of spws. -- Todd Hunter
    """
    if not os.path.exists(temppath): temppath = 'tmp'
    pl.clf()
    pl.gcf().clear()
    plotcalshowgui = False
    subplotRows = subplot/100
    if (subplotRows < 1 or subplotRows > 9 or subplot-subplotRows*100 != 11):
        println(lineno(),"ERROR: subplot must be x11 where x=1..9 (and it only applies to G and T solutions)")
        return

    if (calTableName[-1] == '/'):
        println(lineno(),"checkCalTable(): stripping off the trailing slash from the name")
        calTableName = calTableName[:-1]  # strip off trailing slash if present
    removeTemporary = False
    if (os.access(calTableName, os.W_OK) == False):
        if (os.path.exists(temppath) == False):
            mkdirGroupWrite(temppath)
        println(lineno(),"Copying %s to %s/ to avoid permission problem..........."%(os.path.basename(calTableName),temppath))
        os.system('rsync -vau %s %s %s' % (rsyncPermissions, calTableName,temppath))
        if (os.path.exists('%s/%s' % (temppath,os.path.basename(calTableName))) == False):
            os.system('ln -s %s %s/' % (calTableName,temppath))
        calTableName = temppath + '/' + os.path.basename(calTableName)
        mytb = createCasaTool(tbtool)
        mytb.open(calTableName, nomodify=False)
        k=mytb.keywordnames()
        if (k.__contains__("MSName")==False):
            mytb.putkeyword('MSName',msName)
        if (os.access(mytb.getkeyword('MSName'),os.W_OK)==False):
            # overwrite name in caltable with writeable one if can't find or write to the one that's already in the caltable
            mytb.putkeyword('MSName',msName)            
        mytb.close()
        removeTemporary = True        
#    else:
#        println(lineno(),"I appear to have Write access to %s............." % calTableName)
    supportedCalTypes = ['G Jones', 'T Jones']
    mytb = createCasaTool(tbtool)
    if msName == '':
        mytb.open(calTableName+'/CAL_DESC')
        msName = dict.fromkeys(tb.getcol('MS_NAME')).keys()
        if len(msName) != 1: sys.exit('ERROR: Too many names.')
        msName = msName[0]
        mytb.close()

    mytb.open(calTableName)
    calType = (mytb.info())['subType']
    if calType not in supportedCalTypes: sys.exit('ERROR: Cal type not supported.')
    if calType in ['G Jones', 'T Jones']:
        # Check the number of spws in the caltable. If greater than 4, need 
        # to run plotcal multiple times for multiple pages.
        calStats = {}
        if re.search('^3.3', casadef.casa_version) != None:
            calStats['GAIN_amp'] = (mytb.statistics(column='GAIN', complex_value='amp'))['GAIN']
            calStats['GAIN_phase'] = (mytb.statistics(column='GAIN', complex_value='phase'))['GAIN']
            calStats['GAIN_amp']['max'] = computeMaxGainIgnoring1(mytb.getcol('GAIN'))
            spwsInSolution = np.unique(mytb.getcol('CAL_DESC_ID'))
        else:
            names = mytb.colnames()
            if ('SPECTRAL_WINDOW_ID' in names):  # this is a table from >= 3.4
                calStats['GAIN_amp'] = (mytb.statistics(column='CPARAM', complex_value='amp'))['CPARAM']
                calStats['GAIN_phase'] = (mytb.statistics(column='CPARAM', complex_value='phase'))['CPARAM']
                calStats['GAIN_amp']['max'] = computeMaxGainIgnoring1(mytb.getcol('CPARAM'))
                spwsInSolution = np.unique(mytb.getcol('SPECTRAL_WINDOW_ID'))
            else:
                # This is needed in case the user is running casa 3.4 and wants to examine a 3.3 table
                calStats['GAIN_amp'] = (mytb.statistics(column='GAIN', complex_value='amp'))['GAIN']
                calStats['GAIN_phase'] = (mytb.statistics(column='GAIN', complex_value='phase'))['GAIN']
                calStats['GAIN_amp']['max'] = computeMaxGainIgnoring1(mytb.getcol('GAIN'))
                spwsInSolution = np.unique(mytb.getcol('CAL_DESC_ID'))
        mytb.close()
        plotdir = calTableName+'.plots'
        if (os.path.exists(plotdir) == False):
            mkdirGroupWrite(plotdir)
        println(lineno(),"GAIN_amp_min, max = %s, %s" % (calStats['GAIN_amp']['min'], calStats['GAIN_amp']['max']))
        if (gaintype == 'a'):
          if abs(calStats['GAIN_amp']['min']-1) > 0.001 or abs(calStats['GAIN_amp']['max']-1) > 0.001:
            minAmp = calStats['GAIN_amp']['min']
            maxAmp = calStats['GAIN_amp']['max']
            for startSpw in range(0,len(spwsInSolution)):
                if (len(spwsInSolution) <= subplotRows):
                    plotspw = ''
                    spwnames = ''
                else:
                    plotspw = range(startSpw,np.min([startSpw+subplotRows,len(spwsInSolution)]))
                    plotspw = list(np.array(spwsInSolution)[plotspw])
                    plotspw = str(plotspw).replace(' ','').strip('[]')
                    spwnames = '.spw' + plotspw[0] + '-' + plotspw[-1]
                if (startSpw < len(spwsInSolution)-1):
                    figfile = ''
                else:
                    if (plotfile == ''):
                        figfile = calTableName+'.amp.png'
                    else:
                        figfile = plotfile
                if (plotcalshowgui):
                    println(lineno(), "checkCalTableVariant(): Running plotcal(caltable='%s', xaxis='time', yaxis='amp', iteration='', subplot=%d, plotrange=[0, 0, %f, %f], figfile='%s', spw='%s', showgui=%s, fontsize=%.1f)" % (calTableName,subplot+startSpw,minAmp, maxAmp, figfile, plotspw, interactive, fontsize) ) 
                    plotcal(caltable=calTableName, xaxis='time', yaxis='amp', iteration='', subplot=subplot+startSpw, plotrange=[0, 0, minAmp, maxAmp], figfile=figfile, spw=plotspw, showgui=interactive, fontsize=fontsize)
                else:
                    println(lineno(), "checkCalTableVariant(): Running plotcal(caltable='%s', xaxis='time', yaxis='amp', iteration='', subplot=%d, plotrange=[0, 0, %f, %f], figfile='%s', spw='%s', fontsize=%.1f)" % (calTableName,subplot+startSpw,minAmp, maxAmp, figfile, plotspw, fontsize) ) 
                    plotcal(caltable=calTableName, xaxis='time', yaxis='amp', iteration='', subplot=subplot+startSpw, plotrange=[0, 0, minAmp, maxAmp], figfile=figfile, spw=plotspw, fontsize=fontsize,showgui=False)
#                println(lineno(),"Finished")
                if (interactive == True and startSpw < subplotRows-1):
                    userRawInput = raw_input("Press Enter to continue, q to quit.")
                    if userRawInput.lower() == 'q': break

        elif (gaintype == 'p'):
          if abs(calStats['GAIN_phase']['min']) > 0.001 or abs(calStats['GAIN_phase']['max']) > 0.001:
            minPhase = math.degrees(calStats['GAIN_phase']['min'])
            maxPhase = math.degrees(calStats['GAIN_phase']['max'])
            for startSpw in range(0,len(spwsInSolution)):
                if (len(spwsInSolution) <= subplotRows):
                    plotspw = ''
                    spwnames = ''
                else:
                    plotspw = range(startSpw,np.min([startSpw+subplotRows,len(spwsInSolution)]))
                    plotspw = list(np.array(spwsInSolution)[plotspw])
                    plotspw = str(plotspw).replace(' ','').strip('[]')
                    spwnames = '.spw' + plotspw[0] + '-' + plotspw[-1]
                if (startSpw < len(spwsInSolution)-1):
                    figfile = ''
                else:
                    if (plotfile == ''):
                        figfile = calTableName+'.phase.png'
                    else:
                        figfile = plotfile
                if (plotcalshowgui):
                    println(lineno(), "checkCalTableVariant():Running plotcal(caltable='%s', xaxis='time', yaxis='phase', iteration='', subplot=%d, plotrange=[0, 0,%f,%f], figfile='%s', spw='%s', showgui=%s, fontsize=%.1f)\n" % (calTableName, subplot+startSpw, minPhase, maxPhase, figfile, plotspw, interactive, fontsize) )
                    plotcal(caltable=calTableName, xaxis='time', yaxis='phase', iteration='', subplot=subplot+startSpw, plotrange=[0, 0, minPhase, maxPhase], figfile=figfile, spw=plotspw, showgui=interactive, fontsize=fontsize)
                else:
                    println(lineno(), "checkCalTableVariant():Running plotcal(caltable='%s', xaxis='time', yaxis='phase', iteration='', subplot=%d, plotrange=[0, 0,%f,%f], figfile='%s', spw='%s', fontsize=%.1f)\n" % (calTableName, subplot+startSpw, minPhase, maxPhase, figfile, plotspw, fontsize) )
                    plotcal(caltable=calTableName, xaxis='time', yaxis='phase', iteration='', subplot=subplot+startSpw, plotrange=[0, 0, minPhase, maxPhase], figfile=figfile, spw=plotspw, fontsize=fontsize)
#                println(lineno(), "Finished")
                if (interactive == True and startSpw < subplotRows-1):
                    userRawInput = raw_input("Press Enter to continue, q to quit.")
                    if userRawInput.lower() == 'q': break
        else:
            println(lineno(), "ERROR: Unrecognized gain type = %s." % gaintype)
    if (removeTemporary and False):
        os.system('rm -rf %s' % (calTableName))

def setdirGroupWrite(mydir):
    """
    Recursively sets a directory and its contents to group read/write
    """
    #myname = os.getlogin()
    # liz/EU
    myname = os.getenv('USER')
    for r,d,f in os.walk(mydir):
        if (pwd.getpwuid(os.stat(r).st_uid).pw_name == myname):
            try:
                os.chmod(r, 02775)
            except:
                a=0
#        else:
#            println(lineno(),"Operation not permitted: os.chmod(%s,02775)" % (r))
        for myfile in f:
            if (pwd.getpwuid(os.stat(r+'/'+myfile).st_uid).pw_name == myname):
                try:
                    os.chmod(r+'/'+myfile, 0664)
                except:
                    a=0
#            else:
#                println(lineno(),"Operation not permitted: os.chmod(%s/%s,0664)" % (r,myfile))

def mkdirGroupWrite(mydir):
    os.makedirs(mydir)
    setdirGroupWrite(mydir)

def runPlotms(cmd):
    """
    Plotms should run sychronously with showgui=False, removing the need 
    to parse the logfile for "end", but we'll continue checking just in case
    """
    println(lineno(),"Running: "+cmd)
    exec "status="+cmd
    if status==False:
        print "WARNING: plotms command returned False status"
    else:
        # check just in case:
        mylogfile = casalog.logfile()
        if (os.path.exists(mylogfile) == False):
            # someone may have deleted it
            println(lineno(),"WARNING: runPlotms() is returning immediately because it cannot find the casa log file.")
            return
        foundend = False
        searchText = 'End Task: plotms'
        # sometimes plotms never prints 'End Task: plotms', 
        # but instead prints 'END       Time:'
        # so also check for that string to signify the end of the process
        searchText2 = 'END       Time:'
        f = os.popen('tail --lines=10 '+mylogfile)
        fstrs = f.readlines()
        f.close()
        for fstr in fstrs:
            if (fstr.count(searchText)>0  or fstr.count(searchText2)>0):
                foundend=True
        if foundend==False:
            print 
            print "ERROR: PLOTMS should be synchronous"
            print 
            print "ERROR: you should not be here - press 'q' to exit the debugger, the restrt casapy"
            pdb.set_trace()            
    return


def getManualFluxscaleResults(manualresult):
    """
    Parses the manual reduction's .fluxscale file for fluxscale results.
    """
    webpage = ''
    script=open(manualresult)
    for line in script:
        if (line.find('Found')>0):
            location = line.find('Found')
            webpage += line[location:]+"<br>\n"
        if (line.find('Flux density')>0 and line.find('Fitted spectrum')<0):
            location = line.find('Flux density')
            webpage += beautifyFluxLine(line[location+17:]) + "<br>\n"
    return(webpage)

def beautifyFluxLine(mystring):
    return(mystring.replace(" +/- ","+/-").replace(" = ","=").replace(" is:",":").replace(" in "," "))

def getPipelineFluxscaleResults(casalogfile, asdmNoPath, msWithPath):
    """
    Parses the casapy log file looking for fluxscale results.
    """
    if (casalogfile == '' or casalogfile == None):
        return('casa log file not found')
    f = open(casalogfile,'r')
    lines = f.readlines()
    string = ''
    foundFluxscale = False
    finished = False
    allFields = au.getFields(msWithPath)
    for l in range(len(lines)):
        line = lines[l]
        location = line.find(' fluxscale(')
        if (location > 0):
            println(lineno(),"Found a fluxscale command at line %d." % (l))
            if (foundFluxscale):
                # we've hit the next one, so we are finished
                println(lineno(),"Found next fluxscale command at line %d. Stopping." % (l))
                finished = True
                break
            location = line.find(asdmNoPath)
            if (location > 0):
                println(lineno(),"Found fluxscale with %s at line %d. Stopping" % (asdmNoPath, l))
                foundFluxscale = True
        if (finished):
            break
        if (foundFluxscale):
            location = line.find('Found reference')
            allFields = au.getFields(msWithPath)
            if (location > 0):
                println(lineno(),"Found reference")
                fieldNames = line.split('field(s):')[1].strip().split()
                fieldIDs = convertListOfFieldNamesToIds(fieldNames, msWithPath)
                fieldIDs = '(ID=' + str(fieldIDs).strip('[').strip(']') + ')'
                string += line[location:] + fieldIDs + "<br>\n"
            location = line.find('Found transfer')
            if (location > 0):
                println(lineno(),"Found transfer")
                fieldNames = line.split('field(s):')[1].strip().split()
                k=0
                while k<len(fieldNames):
                    if fieldNames[k][-1]==';':
                        tmp=fieldNames[0:k]
                        # assumes semicolon name had a space after ;
                        tmp.append(fieldNames[k]+" "+fieldNames[k+1])
                        for kk in k+2+pl.array(range(len(fieldNames)-k-2)):
                            tmp.append(fieldNames[kk])
                        fieldNames=tmp
                    k=k+1
                #deal with field names that include a space (e.g. "J1256-057 3c279")
                for field in allFields:
                    if ' ' in field:
                        split_field = field.split()
                        k=0
                        while k < len(fieldNames):
                            if split_field[0] == fieldNames[k] and \
                               split_field[1] == fieldNames[k+1]:
                                fieldNames.insert(k, fieldNames[k] + ' ' + fieldNames[k+1])
                                fieldNames.remove(fieldNames[k+1])
                                fieldNames.remove(fieldNames[k+1])
                            k+=1
                fieldIDs = convertListOfFieldNamesToIds(fieldNames, msWithPath)
                fieldIDs = '(ID=' + str(fieldIDs).strip('[').strip(']') + ')'
                string += line[location:] + fieldIDs + "<br>\n"
            location = line.find('Flux density for ')
            if (location>0 and line.find('INSUFFICIENT DATA') < 0):
                println(lineno(),"Found a flux density")
                string += beautifyFluxLine(line[location+17:]) + "<br>\n"
    f.close()
    return(string)

def convertListOfFieldNamesToIds(fieldNames, msWithPath):
    """
    Convert a list of field name strings into a list of integer IDs, given the full
    path to an existing MS.
    """
    fieldIDs = []
    allFields = list(au.getFields(msWithPath))
    for fieldName in fieldNames:
        fieldIDs.append(allFields.index(fieldName))
    return(fieldIDs)

def printFlagdict(vis):
    mydict = getFlagdict(vis)
    vis = os.path.basename(vis)
    fname = vis + '.flagdict'
    try:
        f = open(fname,'w')
    except:
        fname = '/tmp/' + vis + '.flagdict'
        f = open(fname,'w')
    mystr = pprint.pformat(mydict)
    f.write(mystr)
    f.close()
    print "Result left in %s" % (fname)
    return(mydict)

def flagMatrix(vis, tool=aftool, report='report0', field='', spw='', array='',
               feed='', scan='', correlation='', intent='', observation='', 
               thresholdPercent=95, showAutocorrelations=False):
    """
    Runs the agent flagger on every baseline and generates an ASCII plot showing which
    baselines are mostly flagged.
    -Todd Hunter
    """
    if (not os.path.exists(vis)):
        print "Could not find measurement set."
        return
    mymsmd = createCasaTool(msmdtool)
    mymsmd.open(vis)
    antennaNames = mymsmd.antennanames()
    nantennas = len(antennaNames)
    mymsmd.close()
    myaf = createCasaTool(aftool)
    myaf.open(vis)
    line = ''
    doubleSpace = 36
    for i in range(4):
        line += ' '*6
        for ant1 in range(nantennas):
            line += antennaNames[ant1][i]
            if (nantennas < doubleSpace): 
                line += ' '
        if (i<3): line += '\n'
    print line
    baselinesFlagged = 0
    nBaselines = 0
    for ant1 in range(nantennas):
        line = ''
        line += antennaNames[ant1] + ' '*(1+ant1)
        if (nantennas < doubleSpace): 
            line += ' '*(1+ant1)
        if not showAutocorrelations:
            antstart = ant1+1
            line += '.'
            if (nantennas < doubleSpace): 
                line += ' '
        else:
            antstart = ant1
        for ant2 in range(antstart, nantennas):
            antenna = '%s&%s' % (antennaNames[ant1],antennaNames[ant2])
            nBaselines += 1
            try:
                myaf.selectdata(field=field,spw=spw,array=array,feed=feed,scan=scan,antenna=antenna,
                                correlation=correlation,intent=intent,observation=observation)
                myaf.parsesummaryparameters()
                myaf.init()
                flag_dict = myaf.run()
                flag_dict = stuffPercentIntoAFdict(flag_dict)
                flag_dict = flag_dict[report]
                if (flag_dict['antenna'][antennaNames[ant1]]['percent'] > thresholdPercent):
                    line += 'X'
                    baselinesFlagged += 1
                else:
                    line += ' '
            except:
                line += '.'
            if (nantennas < doubleSpace): 
                line += ' '
        print line
    myaf.done()
    print "%d/%d baselines (%f%%) are flagged" % (baselinesFlagged,nBaselines, 100*baselinesFlagged/nBaselines)
    
def getFlagdict(vis, tool=aftool, report='report0', field='', spw='', array='',
                feed='', scan='', antenna='', uvrange='', 
                correlation='', intent='', observation='', percentOnly=False):
    """
    Calls either the fg or the af tool to return a dictionary of the percentage of flagged
    data as a function of field, spw, array, scan, baseline, correlation, observation.
    percentonly: only return keys with 'percent'
    intent: restrict the calculation to a single intent (wildcard character works)
    spw: restrict calculation to the specified spw list (a comma-delimited string or single int)
    similar for: field, array, feed, scan, antenna, uvrange, correlation, observation
    """
    if (type(spw) == int):
        spw = str(spw)
    if (tool==aftool):
        myaf = createCasaTool(aftool)
        myaf.open(vis)
        myaf.selectdata(field=field,spw=spw,array=array,feed=feed,scan=scan,
                        antenna=antenna,uvrange=uvrange,
                        correlation=correlation,intent=intent,observation=observation)
        myaf.parsesummaryparameters()
        myaf.init()
        flag_dict = myaf.run()
        myaf.done()
        flag_dict = stuffPercentIntoAFdict(flag_dict, percentOnly=percentOnly)
        flag_dict = flag_dict[report]
    else:
        if (antenna != ''):
            print "ERROR: The antenna field is not supported by the fg tool.  Use the af tool."
            return
        myfg = createCasaTool(fgtool)
        myfg.open(vis)
        myfg.setdata(field=field, spw=spw, array=array, feed=feed, scan=scan,
                     uvrange=uvrange, correlation=correlation,
                     intent=intent, observation=observation)
        myfg.setflagsummary()
        flag_dict = myfg.run()
        myfg.done()
        flag_dict = stuffPercentIntoFlagdict(flag_dict)    
    return(flag_dict)
    
def stuffPercentIntoAFdict(flag_dict, report='report0', percentOnly=False):
    """
    Computes and inserts 'percent' keys into the specified flag dictionary
    returned by the af tool
    """
    for key in flag_dict[report].keys():
        if (type(flag_dict[report][key]) == dict):
            for subkey in flag_dict[report][key].keys():
                flag_dict[report][key][subkey]['percent'] = flag_dict[report][key][subkey]['flagged']*100.0 / flag_dict[report][key][subkey]['total']
                if (percentOnly):
                    flag_dict[report][key][subkey].pop('flagged',None)
                    flag_dict[report][key][subkey].pop('total', None)
    flag_dict[report]['percent'] = flag_dict[report]['flagged']*100.0 / flag_dict[report]['total']
    return(flag_dict)

def stuffPercentIntoFlagdict(flag_dict):
    """
    Computes and inserts 'percent' keys into the specified flag dictionary
    returned by the fg tool.
    """
    for key in flag_dict.keys():
        if (key != 'total' and key != 'flagged'):
            for subkey in flag_dict[key].keys():
                flag_dict[key][subkey]['percent'] = flag_dict[key][subkey]['flagged']*100.0 / flag_dict[key][subkey]['total']
    flag_dict['percent'] = flag_dict['flagged']*100.0 / flag_dict['total']
    return(flag_dict)

def computePercentageFlagged(flags, edge):
    """
    Given a 3-dimensional array of flags (as from a Tsys cal table),
    compute the percentage flagged, and the percentage of non-edge channels
    flagged.
    """
    flatflags = flags.flatten()
    nflags = len(np.where(flatflags == 1)[0])
    percentageFlagged = 100.0*nflags/len(flatflags)
    if (edge > 0):
        flatflags = flags[:,edge:-edge].flatten()
        percentageFlaggedIgnoringEdge = 100.0*nflags/len(flatflags)
    else:
        # Avoid divide-by-zero error if edge==0
        percentageFlaggedIgnoringEdge = percentageFlagged
    return(percentageFlagged, percentageFlaggedIgnoringEdge)
    
def printTsysFlags(caltable='', pipecaldir='./', ms='', edge=10,
                   outputfile=None, debug=False, warnFraction=0.5):
    """
    Opens the Tsys Cal table and prints lists of channels flagged, detecting
    whether it is the whole spw, or not. Returns a string suitable for
    adding into the html table.  If caltable is specified, also print it
    to the screen.
    edge: this is the number of edge channels to ignore
    outputfile: write out an html file, for debugging purposes
    warnFraction: threshold to warn about flagged fraction (remaining antennas)
    Example usage: amc.printTsysFlags(caltable='mytable.tsys')
    """
    if (caltable == ''):
        if (pipecaldir[-1] != '/'):
            pipecaldir += '/'
        caltable = pipecaldir + findTsysTable(pipecaldir, ms)
    mytb = createCasaTool(tbtool)
    mytb.open(caltable)
    flags = mytb.getcol("FLAG") # This assumes that all Tsys solutions have same number of channels.
    percentageFlagged, percentageFlaggedIgnoringEdge = computePercentageFlagged(flags,edge)
    print "percentage flagged = %.3f" % (percentageFlagged) 
    print "percentage flagged ignoring edge = %.3f" % (percentageFlaggedIgnoringEdge) 
    npol = len(flags)
    antIndices = mytb.getcol("ANTENNA1")
    colnames = mytb.colnames()
    if ('SPECTRAL_WINDOW_ID' not in colnames):
        println(lineno(),"ERROR: printTsysFlags does not operate on old-style calibration tables.")
        mytb.close()
        return('')
    spws = mytb.getcol("SPECTRAL_WINDOW_ID") 
    times = mytb.getcol("TIME")
    fields = mytb.getcol("FIELD_ID")
    scans = mytb.getcol("SCAN_NUMBER")
    uniqueTimes = np.unique(times)
    mytb.close()
    mytb.open(caltable+"/ANTENNA")
    antNames = mytb.getcol("NAME")
    mytb.close()
    if (ms != ''):
        myms = pipecaldir+ms
        if (os.path.exists(myms) == False):
            if (os.path.exists(ms)):
#                print "Setting ms = ", ms
                myms = ms
        if (casadef.casa_version < '4.1.0'):
            vm = au.ValueMapping(myms)
            fieldnames = vm.uniqueFields
        else:
            mymsmd = createCasaTool(msmdtool)
            mymsmd.open(myms)
            fieldnames = mymsmd.namesforfields()
    else:
#        fieldnames = [''] * len(np.unique(fields))
        fieldnames = [''] * (np.max(np.unique(fields)) + 1)
    webpage = ''
    nEdgeChan = edge
    lowerEdgeChannels = range(nEdgeChan)
    allFlagged = dict.fromkeys(antNames)
    edgedict = {}
    commonChannelsFlagged = {}
    commonChannelList = {}
    for iant in range(len(antNames)):
        commonChannelsFlagged[iant] = {}
        commonChannelList[iant] = {}
        edgedict[antNames[iant]] = {}
        antwebpage = ''
        allFlagged[antNames[iant]] = dict.fromkeys(spws)
        for spw in pl.unique(spws):
            commonChannelsFlagged[iant][spw] = True # assume True until proven otherwise in the loop below
            commonChannelList[iant][spw] = None
            edgedict[antNames[iant]][spw] = {}
            # zsel will be an array of (a few) row numbers that match the ant and spw
            zsel = pl.where((antIndices==iant)*(spws==spw))[0]
            allFlagged[antNames[iant]][spw] = dict.fromkeys(range(npol))
            # print antNames[iant],spw,len(zsel),len(zflag)
            for pol in range(npol):
                edgedict[antNames[iant]][spw][pol] = {}
                allFlagged[antNames[iant]][spw][pol] = dict.fromkeys(range(len(zsel)))
                for t in range(len(zsel)):
                    edgedict[antNames[iant]][spw][pol][t] = {0: 0, 1:0}
                    allFlagged[antNames[iant]][spw][pol][t] = False
                    upperEdgeChannels = range(len(flags[pol,:,zsel[t]])-nEdgeChan, len(flags[pol,:,zsel[t]]))
                    zflag = pl.where(flags[pol,:,zsel[t]])
                    if (len(zflag[0]) > 0):
                        show_me = False
                        if (len(zflag[0]) == len(flags[0,:,zsel[t]])):
                            chans = 'all chan'
                            allFlagged[antNames[iant]][spw][pol][t] = True
                            commonChannelsFlagged[iant][spw] = False
                            if (debug): print "Setting commonChannelsFlagged[%d][%d] time%d to False, all chans" % (iant,spw,t)
                            show_me = True
                        else:
                            channelList = []
                            for ch in zflag[0]:
                                if (ch in lowerEdgeChannels):
                                    edgedict[antNames[iant]][spw][pol][t][0] += 1
                                if (ch in upperEdgeChannels):
                                    edgedict[antNames[iant]][spw][pol][t][1] += 1
                                if ((ch not in lowerEdgeChannels) and (ch not in upperEdgeChannels)):
                                    show_me = True
                                    channelList.append(ch)
                            if (commonChannelList[iant][spw] == None):
                                commonChannelList[iant][spw] = channelList
                            elif (commonChannelList[iant][spw] != channelList or channelList==[]):
                                if (debug): print "Setting commonChannelsFlagged[%d][%d] time%d to False, channelList=%s" % (iant,spw,t,str(channelList))
                                commonChannelsFlagged[iant][spw] = False
                            elif (debug):
                                print "channelList == commonChannelList[%d][%d]" % (iant,spw)
                            chans = '%d channels: '%(len(zflag[0])) + str(zflag[0])
                        if (show_me):
                            timeIndex = list(uniqueTimes).index(times[zsel[t]])
#                            print "iant=%d, len(antNames)=%d, t=%d, zsel[t]=%s, len(fields)=%d, fields[%d]=%d, len(fieldnames)=%d, fieldnames=%s" % (iant,len(antNames),t,str(zsel[t]),len(fields),zsel[t],fields[zsel[t]],len(fieldnames),fieldnames)
                            myline = antNames[iant] + " (#%02d), field %d (%s), time%d=%s, scan %d, pol %d, spw %2d, "%(iant,fields[zsel[t]],fieldnames[fields[zsel[t]]],timeIndex,au.mjdsecToUTHMS(times[zsel[t]]),scans[zsel[t]],pol,spw) + chans + "<br>\n"
#                            println(lineno(),myline[0:-1].replace('<br>',''))
                            antwebpage += myline
                    else:
                        if (debug):
                            print "Setting commonChannelsFlagged[%d][%d] time%d to False, because no flags found" % (iant,spw,t)
                        commonChannelsFlagged[iant][spw] = False
        if (len(antwebpage) > 0):
            # add a row gap between antennas
            webpage += antwebpage + "<br>\n"
    preamble = '%s<br>Percentage of solutions flagged (including edge channels) = %.3f<br>' % (os.path.basename(caltable),percentageFlagged)
    preamble += "The following report ignores edge channels (up to %d)<br>" % (edge)
    uniqueAntNames = np.unique(antNames)
    # zsel will be an array of (a few) row numbers that match the final antenna
    # and final spw from the loop above, and thus represents individual scans
    for t in range(len(zsel)):
#        print "Testing time %d for being fully flagged" % (t)
        timerangeFullyFlagged = True
        antNamesNotFullyFlagged = []
        antNamesFullyFlagged = []
        for iant in range(len(antNames)):  # effectively, the number of rows
            for spw in pl.unique(spws):
                for pol in range(npol):
                    if (t in allFlagged[antNames[iant]][spw][pol].keys()):
                        if (allFlagged[antNames[iant]][spw][pol][t] == False):
                            antNamesNotFullyFlagged.append(antNames[iant])
                            timerangeFullyFlagged = False
                        else:
                            antNamesFullyFlagged.append(antNames[iant])
        if (ms != ''):
            if (casadef.casa_version < '4.1.0'):
                intents = getMajorIntentsForFieldID(field=fields[zsel[t]], vm=vm)
            else:
                intents = getMajorIntentsForFieldID(field=fields[zsel[t]], mymsmd=mymsmd)
        else:
            intents = ''
        antNamesFullyFlagged = np.unique(antNamesFullyFlagged)
        antNamesNotFullyFlagged = np.unique(antNamesNotFullyFlagged)
        if (timerangeFullyFlagged):
            preamble += "Time%d=%s (field %s = %s, scan %d) is completely flagged on all antennas! (%s)<br>" % (t,au.mjdsecToUTHMS(times[zsel[t]]),fields[zsel[t]],fieldnames[fields[zsel[t]]], scans[zsel[t]], str(intents))
            # remove all mention of this timerange in the preceding text
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                checkForString = 'time%d=%s' % (t,au.mjdsecToUTHMS(times[zsel[t]]))
#                if (line.find(checkForString) < 0):
#                    print "Did not find %s in %s" % (checkForString,line)
#                    newpage += line + '<br>'
            webpage = newpage
        elif (len(antNamesNotFullyFlagged) < len(uniqueAntNames)*warnFraction): # if more than half the antennas are flagged
            percent = 100.0*(len(uniqueAntNames) - len(antNamesNotFullyFlagged)) / len(uniqueAntNames)
            preamble += "Time%d=%s (field %s = %s, scan %d) is completely flagged on %d/%d antennas (%.0f%%)! (%s)<br>" % (t,au.mjdsecToUTHMS(times[zsel[t]]),fields[zsel[t]],fieldnames[fields[zsel[t]]],scans[zsel[t]],len(uniqueAntNames)-len(antNamesNotFullyFlagged), len(uniqueAntNames), percent, str(intents))
            # remove all mention of this timerange on these antennas in the preceding text (someday)
            preamble += "   %s<br>" % (str(antNamesFullyFlagged))
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                checkForString = 'time%d=%s' % (t,au.mjdsecToUTHMS(times[zsel[t]]))
                if (line.find(checkForString) < 0 or line[:5] in antNamesNotFullyFlagged):
                    newpage += line + '<br>'
            webpage = newpage
    antennasFlagged = []
    for iant in range(len(uniqueAntNames)): 
        antSpwCombination = []
        for spw in pl.unique(spws):
            antSpwCombinationFlaggedForAllTimes = True
            for pol in range(npol):
                for t in range(len(zsel)):
                    if (t in allFlagged[uniqueAntNames[iant]][spw][pol].keys()):
                        if (allFlagged[uniqueAntNames[iant]][spw][pol][t] == False):
                            antSpwCombinationFlaggedForAllTimes = False
            if (antSpwCombinationFlaggedForAllTimes):
                antSpwCombination.append(spw)
        if (len(antSpwCombination) == len(pl.unique(spws))):
            preamble += "Antenna %s (#%2d) is flagged in all spws at all times<br>" % (uniqueAntNames[iant], iant)
            antennasFlagged.append(uniqueAntNames[iant])
            # remove all mention of this antenna in the preceding text
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                if (line.find(uniqueAntNames[iant]) < 0):
                    newpage += line + '<br>'
            webpage = newpage
        elif (len(antSpwCombination) > 1):
            preamble += "Antenna %s is flagged in spws %s at all times<br>" % (uniqueAntNames[iant],str(antSpwCombination))
            # remove all mention of this antenna/spws combination in the preceding text
            for asc in range(len(antSpwCombination)):
                newpage = ''
                lines = webpage.split('<br>')
                for line in lines:
                    if (line.find(uniqueAntNames[iant]) < 0 or line.find('spw %d'%(antSpwCombination[asc])) < 0):
                        newpage += line + '<br>'
                webpage = newpage
        elif (len(antSpwCombination) == 1):
            preamble += "Antenna %s is flagged in spw %d at all times<br>" % (uniqueAntNames[iant],antSpwCombination[0])
            # remove all mention of this antenna/spw combination in the preceding text
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                if (line.find(uniqueAntNames[iant]) < 0 or line.find('spw %d'%(antSpwCombination[0])) < 0):
                    newpage += line + '<br>'
            webpage = newpage
        else:
            for spw in pl.unique(spws):
                if (commonChannelsFlagged[iant][spw]):
                    preamble += "Antenna %s is flagged in spw %d at all times in %d channels: %s<br>" % (uniqueAntNames[iant],spw,len(commonChannelList[iant][spw]),str(commonChannelList[iant][spw]))
                    # remove all mention of this antenna/spw combination in the preceding text
                    if (debug): print "Removing ant%d/spw%d from text" % (iant,spw)
                    newpage = ''
                    lines = webpage.split('<br>\n')
                    for line in lines:
                        if (line.find(uniqueAntNames[iant]) < 0 or line.find('spw %d'%(spw)) < 0):
                            newpage += line + '<br>'
                    webpage = newpage
                else:
                    if (debug): print "Not consolidating any text for ant%d/spw%d" % (iant,spw)

            
    if (ms != '' and casadef.casa_version >= '4.1.0'):
        mymsmd.close()

#    webpage += '<a href="edgeChannelReport.html">edge channel flagging report</a>'
    
    ecwebpage = '<FONT FACE="courier">'
    ecwebpage += 'Edge channel report (values are: pol0,pol1)<br>'
#    ecwebpage += 'Antenna    '
    ecwebpage += 3*'&nbsp;'
    for spw in pl.unique(spws):
        ecwebpage += 4*'&nbsp;' + 'spw%2d' % (spw) + 3*'&nbsp;'
    ecwebpage += '<br>'
    ecwebpage += 5*'&nbsp;'
    for spw in pl.unique(spws):
        ecwebpage += 2*'&nbsp;' + 'low--high' + 1*'&nbsp;'
    ecwebpage += '<br>'
#    for iant in range(len(antNames)):
# report is the same for all antennas
    for iant in range(1):
        if (antNames[iant] in antennasFlagged): continue
        for t in range(len(zsel)):
#            if (t==0):
#                ecwebpage += '%4s' % (antNames[iant])
#            else:
#                ecwebpage += 4*'&nbsp;'
            ecwebpage += '&nbsp;t=%d&nbsp;' % (t)
            for spw in pl.unique(spws):
                ecwebpage += '&nbsp;'*2
                for pol in range(npol):
                    if (t in edgedict[antNames[iant]][spw][pol]):
                        if (pol>0):
                            ecwebpage += ','
#                        print "iant=%d, spw=%d, pol=%d, t=%d" % (iant,spw,pol,t)
                        ecwebpage += '%d' % (edgedict[antNames[iant]][spw][pol][t][0])
                ecwebpage += '--'
                for pol in range(npol):
                    if (t in edgedict[antNames[iant]][spw][pol]):
                        if (pol>0):
                            ecwebpage += ','
                        ecwebpage += '%d' % (edgedict[antNames[iant]][spw][pol][t][1])
                ecwebpage += 2*'&nbsp;'
                
            ecwebpage += '<br>'
    ecwebpage += '</font>'
    webpage += ecwebpage
    if (False):
        # Might put this on a separate page someday
        f.open('edgeChannelReport.html','w')
        f.write(ecwebpage.replace('<br>','<br>\n'))
        f.close()
        myCopyfile()
    webpage = preamble + webpage
    plaintext = webpage.replace('<br>','\n').replace('&nbsp;', ' ')
    if (outputfile != None):
        f = open(outputfile,'w')
        f.write(webpage)
        f.close()
        print "html file left in: %s" % (outputfile)
    return(webpage,plaintext)
# end of printTsysFlags()

def analyzeTsysFlags(caltable='', pipecaldir='./', ms='', edge=3):
    """
    Opens the Tsys Cal table and builds a dictionary of channels flagged, 
    detecting whether it is the whole spw, or not.  This is an unfinished
    work, and needs to be reconciled with recent changes to printTsysFlags
    before proceeding.

    edge: this is the number of edge channels to ignore
    Example usage: amc.analyzeTsysFlags(caltable='mytable.tsys')
    """
    if (caltable == ''):
        caltable = pipecaldir + findTsysTable(pipecaldir, ms)
    mytb = createCasaTool(tbtool)
    mytb.open(caltable)
    flags = mytb.getcol("FLAG") # This assumes that all Tsys solutions have same number of channels.
    percentageFlagged, percentageFlaggedIgnoringEdge = computePercentageFlagged(flags,edge)
    tsysinfo = {}
    tsysinfo['percentFlagged'] = percentageFlagged
    tsysinfo['percentFlaggedIgnoringEdge'] = percentageFlaggedIgnoringEdge
    print "percentage flagged = %.3f" % (percentageFlagged) 
    print "percentage flagged ignoring edge = %.3f" % (percentageFlaggedIgnoringEdge) 
    npol = len(flags)
    antIndices = mytb.getcol("ANTENNA1")
    colnames = mytb.colnames()
    if ('SPECTRAL_WINDOW_ID' not in colnames):
        println("ERROR: printTsysFlags does not operate on old-style calibration tables.")
        mytb.close()
        return('')
    spws = mytb.getcol("SPECTRAL_WINDOW_ID") 
    times = mytb.getcol("TIME")
    fields = mytb.getcol("FIELD_ID")
    scans = mytb.getcol("SCAN_NUMBER")
    uniqueTimes = np.unique(times)
    mytb.close()
    mytb.open(caltable+"/ANTENNA")
    antNames = mytb.getcol("NAME")
    mytb.close()
    if (ms != ''):
        myms = pipecaldir+ms
        if (os.path.exists(myms) == False):
            if (os.path.exists(ms)):
                myms = ms
        if (casadef.casa_version < '4.1.0'):
            vm = au.ValueMapping(myms)
            fieldnames = vm.uniqueFields
        else:
            mymsmd = createCasaTool(msmdtool)
            mymsmd.open(myms)
            fieldnames = mymsmd.namesforfields()
    else:
        fieldnames = [''] * len(np.unique(fields))
    nEdgeChan = edge
    lowerEdgeChannels = range(nEdgeChan)
    allFlagged = dict.fromkeys(antNames)
    for iant in range(len(antNames)):
        allFlagged[antNames[iant]] = dict.fromkeys(spws)
        for spw in pl.unique(spws):
            # zsel will be an array of (a few) row numbers that match the ant and spw
            zsel = pl.where((antIndices==iant)*(spws==spw))[0]
            allFlagged[antNames[iant]][spw] = dict.fromkeys(range(npol))
            # print antNames[iant],spw,len(zsel),len(zflag)
            for pol in range(npol):
                allFlagged[antNames[iant]][spw][pol] = dict.fromkeys(range(len(zsel)))
                for t in range(len(zsel)):
                    allFlagged[antNames[iant]][spw][pol][t] = False
                    upperEdgeChannels = range(len(flags[pol,:,zsel[t]])-nEdgeChan, len(flags[pol,:,zsel[t]]))
                    zflag=pl.where(flags[pol,:,zsel[t]])
                    if (len(zflag[0]) > 0):
                        show_me = False
                        if (len(zflag[0]) == len(flags[0,:,zsel[t]])):
                            chans = 'all chan'
                            allFlagged[antNames[iant]][spw][pol][t] = True
                            show_me = True
                        else:
                            for ch in zflag[0]:
                                if ((ch not in lowerEdgeChannels) and (ch not in upperEdgeChannels)):
                                    show_me = True
                                    break
                            chans = '%d channels: '%(len(zflag[0])) + str(zflag[0])
                        if (show_me):
                            timeIndex = list(uniqueTimes).index(times[zsel[t]])
                            if (antNames[iant] not in tsysInfo.keys()):
                                tsysInfo['antNames[iant]'] = {}
                            fieldname = fields[zsel[t]]
                            if (fieldnames[fieldname] not in tsysInfo['antNames[iant]'].keys()):
                                tsysInfo['antNames[iant]'][fieldname] = {}
                            if (scans[zsel[t]] not in tsysInfo['antNames[iant]'][fieldname].keys()):
                                tsysInfo['antNames[iant]'][fieldname][scans[zsel[t]]] = {}
                            tsysInfo['antNames[iant]'][fieldname][scans[zsel[t]]][pol] = chans

    uniqueAntNames = np.unique(antNames)
    # zsel will be an array of (a few) row numbers that match the final antenna
    # and final spw from the loop above, and thus represents individual scans
    for t in range(len(zsel)):
#        print "Testing time %d for being fully flagged" % (t)
        timerangeFullyFlagged = True
        antNamesNotFullyFlagged = []
        for iant in range(len(antNames)):  # effectively, the number of rows
            for spw in pl.unique(spws):
                for pol in range(npol):
                    if (t in allFlagged[antNames[iant]][spw][pol].keys()):
                        if (allFlagged[antNames[iant]][spw][pol][t] == False):
                            antNamesNotFullyFlagged.append(antNames[iant])
                            timerangeFullyFlagged = False
        if (ms != ''):
            if (casadef.casa_version < '4.1.0'):
                intents = getMajorIntentsForFieldID(field=fields[zsel[t]], vm=vm)
            else:
                intents = getMajorIntentsForFieldID(field=fields[zsel[t]], mymsmd=mymsmd)
        else:
            intents = ''
        antNamesNotFullyFlagged = np.unique(antNamesNotFullyFlagged)
        if (timerangeFullyFlagged):
            if ('allAntennas' not in tsysInfo.keys()):
                tsysInfo['allAntennas'] = {}
            if ('scan' not in tsysInfo['allAntennas'].keys()):
                tsysInfo['allAntennas']['scan'] = {}
            tsysInfo['allAntennas']['scan'].append(scans[zsel[t]])
#            preamble += "Time%d=%s (field %s = %s, scan %d) is completely flagged on all antennas! (%s)<br>" % (t,au.mjdsecToUTHM(times[zsel[t]]),fields[zsel[t]],fieldnames[fields[zsel[t]]], scans[zsel[t]], str(intents))
        elif (len(antNamesNotFullyFlagged) < len(uniqueAntNames)/2): # if more than half the antennas are flagged
            percent = 100.0*(len(uniqueAntNames) - len(antNamesNotFullyFlagged)) / len(uniqueAntNames)
            if ('mostAntennas' not in tsysInfo.keys()):
                tsysInfo['mostAntennas'] = {}
            if ('scan' not in tsysInfo['mostAntennas'].keys()):
                tsysInfo['mostAntennas']['scan'] = {}
            tsysInfo['mostAntennas']['scan'].append(scans[zsel[t]])
#            preamble += "Time%d=%s (field %s = %s, scan %d) is completely flagged on %d/%d antennas (%.0f%%)! (%s)<br>" % (t,au.mjdsecToUTHM(times[zsel[t]]),fields[zsel[t]],fieldnames[fields[zsel[t]]],scans[zsel[t]],len(uniqueAntNames)-len(antNamesNotFullyFlagged), len(uniqueAntNames), percent, str(intents))
#
# here is where I am 
#
    for iant in range(len(uniqueAntNames)): 
        antSpwCombination = []
        for spw in pl.unique(spws):
            antSpwCombinationFlaggedForAllTimes = True
            for pol in range(npol):
                for t in range(len(zsel)):
                    if (t in allFlagged[uniqueAntNames[iant]][spw][pol].keys()):
                        if (allFlagged[uniqueAntNames[iant]][spw][pol][t] == False):
                            antSpwCombinationFlaggedForAllTimes = False
            if (antSpwCombinationFlaggedForAllTimes):
                antSpwCombination.append(spw)
        if (len(antSpwCombination) == len(pl.unique(spws))):
            preamble += "Antenna %s is flagged in all spws at all times<br>" % (uniqueAntNames[iant])
            tsysInfo[uniqueAntNames[iant]]['allspws'] = {}
            # remove all mention of this antenna in the preceding text
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                if (line.find(uniqueAntNames[iant]) < 0):
                    newpage += line + '<br>'
            webpage = newpage
        elif (len(antSpwCombination) > 1):
            preamble += "Antenna %s is flagged in spws %s at all times<br>" % (uniqueAntNames[iant],str(antSpwCombination))
        elif (len(antSpwCombination) == 1):
            preamble += "Antenna %s is flagged in spw %d at all times<br>" % (uniqueAntNames[iant],antSpwCombination[0])
            # remove all mention of this antenna/spw combination in the preceding text
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                if (line.find(uniqueAntNames[iant]) < 0 or line.find('spw %d'%(antSpwCombination[0])) < 0):
                    newpage += line + '<br>'
            webpage = newpage

    if (ms != '' and casadef.casa_version >= '4.1.0'):
        mymsmd.close()
    return(tsysInfo)
# end of analyzeTsysFlags()

def printCaltableFlags(caltable='', caldir='', vis='', edge=3):
    """
    This is currently identical to printTsysFlags() except that the
    scan number is not printed, since it is not generally filled for
    non-Tsys tables.
    Opens a Cal table and prints lists of channels flagged, detecting
    whether it is the whole spw, or not. Returns a string suitable for
    adding into the html table.  If caltable is specified, also print it
    to the screen.
    edge: this is the number of edge channels to ignore
    Example usage: amc.printCaltableFlags(caltable='mytable.cal')
    """
    if (caltable == ''):
        print "ERROR: No caltable specified"
        return
    mytb = createCasaTool(tbtool)
#    print "printCaltableFlags(): caltable=%s" % (caltable)
    mytb.open(caltable)
    antIndices = mytb.getcol("ANTENNA1")
    colnames = mytb.colnames()
#    print colnames
    if ('SPECTRAL_WINDOW_ID' not in colnames):
        spws = mytb.getcol("CAL_DESC_ID") 
    else:
        spws = mytb.getcol("SPECTRAL_WINDOW_ID") 
    times = mytb.getcol("TIME")
    fields = mytb.getcol("FIELD_ID")
    scans = mytb.getcol("SCAN_NUMBER")  # these are all -1 for non-Tsys tables
    uniqueTimes = np.unique(times)
    mytb.close()
    if ('SPECTRAL_WINDOW_ID' not in colnames):
        mytb.open(vis+"/ANTENNA")
        antNames = mytb.getcol("NAME")
    else:
        mytb.open(caltable+"/ANTENNA")
        antNames = mytb.getcol("NAME")
    mytb.close()
    if (vis != ''):
        myms = caldir+vis
        if (os.path.exists(myms) == False):
            if (os.path.exists(vis)):
                myms = vis
        if (casadef.casa_version < '4.1.0'):
            vm = au.ValueMapping(myms)
            fieldnames = vm.uniqueFields
        else:
            mymsmd = createCasaTool(msmdtool)
            mymsmd.open(myms)
            fieldnames = mymsmd.namesforfields()
    else:
        fieldnames = [''] * len(np.unique(fields))
    webpage = ''
    nEdgeChan = edge
    lowerEdgeChannels = range(nEdgeChan)
    allFlagged = dict.fromkeys(antNames)
    mytb.open(caltable)
    totalflags = 0
    totalsolutions = 0
    for iant in range(len(antNames)):
        antwebpage = ''
        allFlagged[antNames[iant]] = dict.fromkeys(spws)
        for spw in pl.unique(spws):
            # zsel will be an array of (a few) row numbers that match the ant and spw
            zsel = pl.where((antIndices==iant)*(spws==spw))[0]
            flags = mytb.getcell("FLAG",zsel[0])
            flatflags = flags.flatten()
            totalflags += len(np.where(flatflags == 1)[0])
            totalsolutions += len(flatflags)
            npol = len(flags)
            allFlagged[antNames[iant]][spw] = dict.fromkeys(range(npol))
            # print antNames[iant],spw,len(zsel),len(zflag)
            for pol in range(npol):
                allFlagged[antNames[iant]][spw][pol] = dict.fromkeys(range(len(zsel)))
                for t in range(len(zsel)):
                    flags = mytb.getcell("FLAG",zsel[t])
                    allFlagged[antNames[iant]][spw][pol][t] = False
                    upperEdgeChannels = range(len(flags[pol,:])-nEdgeChan, len(flags[pol,:]))
                    zflag=pl.where(flags[pol,:])
                    if (len(zflag[0]) > 0):
                        show_me = False
                        if (len(zflag[0]) == len(flags[0,:])):
                            chans = 'all chan'
                            allFlagged[antNames[iant]][spw][pol][t] = True
                            show_me = True
                        else:
                            for ch in zflag[0]:
                                if ((ch not in lowerEdgeChannels) and (ch not in upperEdgeChannels)):
                                    show_me = True
                                    break
                            chans = '%d channels: '%(len(zflag[0])) + str(zflag[0])
                        if (show_me):
                            timeIndex = list(uniqueTimes).index(times[zsel[t]])
                            myline = antNames[iant] + " (#%02d), field %d (%s), time%d=%s, pol %d, spw %2d"%(iant,fields[zsel[t]],fieldnames[fields[zsel[t]]],timeIndex,au.mjdsecToUTHMS(times[zsel[t]]),pol,spw) + "<br>\n"
#                            println(lineno(),myline[0:-1].replace('<br>',''))
                            antwebpage += myline
        if (len(antwebpage) > 0):
            # add a row gap between antennas
            webpage += antwebpage + "<br>\n"
    mytb.close()
    preamble = 'Percentage of solutions flagged = %.3f<br>' % (totalflags*100.0/totalsolutions)
    uniqueAntNames = np.unique(antNames)
    # zsel will be an array of (a few) row numbers that match the final antenna
    # and final spw from the loop above, and thus represents individual scans
#    print "zsel = ", zsel
    for t in range(len(zsel)):
#        print "Testing time %d for being fully flagged" % (t)
        timerangeFullyFlagged = True
        antNamesNotFullyFlagged = []
        for iant in range(len(antNames)):  # effectively, the number of rows
            for spw in pl.unique(spws):
                for pol in range(npol):
#                    print "iant, spw, pol, t = ", iant, spw, pol, t
#                    print "antNames[iant] = ", antNames[iant]
#                    print "allFlagged[antNames[iant]][spw][pol].keys() =  ", allFlagged[antNames[iant]][spw][pol].keys()
                    if (t in allFlagged[antNames[iant]][spw][pol].keys()):
                        if (allFlagged[antNames[iant]][spw][pol][t] == False):
                            antNamesNotFullyFlagged.append(antNames[iant])
                            timerangeFullyFlagged = False
        if (vis != ''):
            if (casadef.casa_version < '4.1.0'):
                intents = getMajorIntentsForFieldID(field=fields[zsel[t]], vm=vm)
            else:
                intents = getMajorIntentsForFieldID(field=fields[zsel[t]], mymsmd=mymsmd)
        else:
            intents = ''
        antNamesNotFullyFlagged = np.unique(antNamesNotFullyFlagged)
        if (timerangeFullyFlagged):
            preamble += "Time%d=%s (field %s = %s) is completely flagged on all antennas! (%s)<br>" % (t,au.mjdsecToUTHMS(times[zsel[t]]),fields[zsel[t]],fieldnames[fields[zsel[t]]], str(intents))
        elif (len(antNamesNotFullyFlagged) < len(uniqueAntNames)/2): # if more than half the antennas are flagged
            percent = 100.0*(len(uniqueAntNames) - len(antNamesNotFullyFlagged)) / len(uniqueAntNames)
            preamble += "Time%d=%s (field %s = %s) is completely flagged on %d/%d antennas (%.0f%%)! (%s)<br>" % (t,au.mjdsecToUTHMS(times[zsel[t]]),fields[zsel[t]],fieldnames[fields[zsel[t]]],len(uniqueAntNames)-len(antNamesNotFullyFlagged), len(uniqueAntNames), percent, str(intents))
    for iant in range(len(uniqueAntNames)): 
        antSpwCombination = []
        for spw in pl.unique(spws):
            antSpwCombinationFlaggedForAllTimes = True
            for pol in range(npol):
                for t in range(len(zsel)):
                    if (t in allFlagged[uniqueAntNames[iant]][spw][pol].keys()):
                        if (allFlagged[uniqueAntNames[iant]][spw][pol][t] == False):
                            antSpwCombinationFlaggedForAllTimes = False
            if (antSpwCombinationFlaggedForAllTimes):
                antSpwCombination.append(spw)
        if (len(antSpwCombination) == len(pl.unique(spws))):
            preamble += "Antenna %s is flagged in all spws at all times<br>" % (uniqueAntNames[iant])
            # remove all mention of this antenna in the preceding text
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                if (line.find(uniqueAntNames[iant]) < 0):
                    newpage += line + '<br>'
            webpage = newpage
        elif (len(antSpwCombination) > 1):
            preamble += "Antenna %s is flagged in spws %s at all times<br>" % (uniqueAntNames[iant],str(antSpwCombination))
            # remove all mention of this antenna/spw combination in the preceding text
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                spwOkay =  True
                for spwcheck in antSpwCombination:
                    if (line.find('spw %d'%(spwcheck)) >= 0 or line.find('spw  %d'%(spwcheck)) >= 0):
                        spwOkay = False
                if (line.find(uniqueAntNames[iant]) < 0 or spwOkay):
                    newpage += line + '<br>'
            webpage = newpage
        elif (len(antSpwCombination) == 1):
            preamble += "Antenna %s is flagged in spw %d at all times<br>" % (uniqueAntNames[iant],antSpwCombination[0])
            # remove all mention of this antenna/spw combination in the preceding text
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                if (line.find(uniqueAntNames[iant]) < 0 or (line.find('spw %d'%(antSpwCombination[0])) < 0
                                                        and line.find('spw  %d'%(antSpwCombination[0])) < 0)):
                    newpage += line + '<br>'
            webpage = newpage

    if (vis != '' and casadef.casa_version >= '4.1.0'):
        mymsmd.close()
    webpage = preamble + webpage  
    return(preamble, webpage)
# end of printCaltableFlags

def printBandpassCaltableFlags(caltable='', caldir='', vis='', edge=3, debug=False):
    """
    Similar to printCaltableFlags() but examines each data descriptor
    id, one at a time, to avoid array conformance errors.
    Opens a Cal table and prints lists of channels flagged, detecting
    whether it is the whole spw, or not. Returns a string suitable for
    adding into the html table.  If caltable is specified, also print it
    to the screen.
    edge: this is the number of edge channels to ignore
    Example usage: amc.printBandpassCaltableFlags(caltable='mytable.cal')
    """
    if (caltable == ''):
        print "ERROR: No caltable specified"
        return
    mytb = createCasaTool(tbtool)
    mytb.open(caltable)
    colnames = mytb.colnames()
    if ('SPECTRAL_WINDOW_ID' in colnames):
        tableFormat = 34
        spws = mytb.getcol('SPECTRAL_WINDOW_ID')
    else:
        tableFormat = 33
        spws = mytb.getcol('CAL_DESC_ID')
    times = mytb.getcol("TIME")
    antennas = mytb.getcol("ANTENNA1")
    fields = mytb.getcol("FIELD_ID")
    scans = mytb.getcol("SCAN_NUMBER")  # these are all -1 for non-Tsys tables
    uniqueTimes = np.unique(times)
    mytb.close()
    if (tableFormat == 34):
        mytb.open(caltable+"/ANTENNA")
        antNames = mytb.getcol("NAME")
    else:
        mytb.open(vis+'/ANTENNA')
        antNames = mytb.getcol("NAME")
    mytb.close()
    if (vis != ''):
        myvis = caldir+vis
        if (os.path.exists(myvis) == False):
            if (os.path.exists(vis)):
                myvis = vis
        if (casadef.casa_version < '4.1.0'):
            vm = au.ValueMapping(myvis)
            fieldnames = vm.uniqueFields
        else:
            mymsmd = createCasaTool(msmdtool)
            mymsmd.open(myvis)
            fieldnames = mymsmd.namesforfields()
    else:
        fieldnames = [''] * len(np.unique(fields))
    nEdgeChan = edge
    lowerEdgeChannels = range(nEdgeChan)
    allFlagged = dict.fromkeys(antNames)
    mytb.open(caltable)
    preamble = ''
    webpage = ''
    for iant in range(len(antNames)):
        antwebpage = ''
        allFlagged[antNames[iant]] = dict.fromkeys(spws)
        for spw in pl.unique(spws):
            # zsel will be an array of (one or a few) row numbers that match the ant and spw
            zsel = pl.where((antennas==iant)*(spws==spw))[0]
            npol = len(mytb.getcell('FLAG',zsel[0]))
            allFlagged[antNames[iant]][spw] = dict.fromkeys(range(npol))
            for pol in range(npol):
                allFlagged[antNames[iant]][spw][pol] = dict.fromkeys(range(len(zsel)))
                for t in range(len(zsel)):
                    allFlagged[antNames[iant]][spw][pol][t] = False
                    flags = mytb.getcell("FLAG",zsel[t])
                    upperEdgeChannels = range(len(flags[pol,:])-nEdgeChan, len(flags[pol,:]))
                    zflag=pl.where(flags[pol,:])
                    if (len(zflag[0]) > 0):
                        show_me = False
                        if (len(zflag[0]) == len(flags[0,:])):
                            chans = 'all chan'
                            allFlagged[antNames[iant]][spw][pol][t] = True
                            show_me = True
                        else:
                            for ch in zflag[0]:
                                if ((ch not in lowerEdgeChannels) and (ch not in upperEdgeChannels)):
                                    show_me = True
                                    break
                            chans = '%d channels: '%(len(zflag[0])) + str(zflag[0])
                        if (show_me):
                            timeIndex = list(uniqueTimes).index(times[zsel[t]])
                            myline = antNames[iant] + " (#%02d), field %d (%s), time%d=%s, pol %d, spw %2d"%(iant,fields[zsel[t]],fieldnames[fields[zsel[t]]],timeIndex,au.mjdsecToUTHMS(times[zsel[t]]),pol,spw) + "<br>\n"
#                            println(lineno(),myline[0:-1].replace('<br>',''))
                            antwebpage += myline
        if (len(antwebpage) > 0):
            # add a row gap between antennas
            webpage += antwebpage + "<br>\n"
    mytb.close()
    uniqueAntNames = np.unique(antNames)
    for t in range(len(zsel)):
        if (debug):
            print "Testing time %d for being fully flagged" % (t)
        timerangeFullyFlagged = True
        antNamesNotFullyFlagged = []
        for iant in range(len(antNames)):  # effectively, the number of rows
            for spw in pl.unique(spws):
                for pol in range(npol):
                    if (t in allFlagged[antNames[iant]][spw][pol].keys()):
                        if (allFlagged[antNames[iant]][spw][pol][t] == False):
                            antNamesNotFullyFlagged.append(antNames[iant])
                            timerangeFullyFlagged = False
        if (ms != ''):
            if (casadef.casa_version < '4.1.0'):
                intents = getMajorIntentsForFieldID(field=fields[zsel[t]], vm=vm)
            else:
                intents = getMajorIntentsForFieldID(field=fields[zsel[t]], mymsmd=mymsmd)
        else:
            intents = ''
        antNamesNotFullyFlagged = np.unique(antNamesNotFullyFlagged)
        if (timerangeFullyFlagged):
            preamble += "Time%d=%s (field %s = %s) is completely flagged on all antennas! (%s)<br>" % (t,au.mjdsecToUTHMS(times[zsel[t]]),fields[zsel[t]],fieldnames[fields[zsel[t]]], str(intents))
        elif (len(antNamesNotFullyFlagged) < len(uniqueAntNames)/2): # if more than half the antennas are flagged
            percent = 100.0*(len(uniqueAntNames) - len(antNamesNotFullyFlagged)) / len(uniqueAntNames)
            preamble += "Time%d=%s (field %s = %s) is completely flagged on %d/%d antennas (%.0f%%)! (%s)<br>" % (t,au.mjdsecToUTHMS(times[zsel[t]]),fields[zsel[t]],fieldnames[fields[zsel[t]]],len(uniqueAntNames)-len(antNamesNotFullyFlagged), len(uniqueAntNames), percent, str(intents))
    for iant in range(len(uniqueAntNames)): 
        antSpwCombination = []
        for spw in pl.unique(spws):
            antSpwCombinationFlaggedForAllTimes = True
            for pol in range(npol):
                for t in range(len(zsel)):
                    if (t in allFlagged[uniqueAntNames[iant]][spw][pol].keys()):
                        if (allFlagged[uniqueAntNames[iant]][spw][pol][t] == False):
                            antSpwCombinationFlaggedForAllTimes = False
            if (antSpwCombinationFlaggedForAllTimes):
                antSpwCombination.append(spw)
        if (len(antSpwCombination) == len(pl.unique(spws))):
            preamble += "Antenna %s is flagged in all spws at all times<br>" % (uniqueAntNames[iant])
            # remove all mention of this antenna in the preceding text
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                if (line.find(uniqueAntNames[iant]) < 0):
                    newpage += line + '<br>'
            webpage = newpage
        elif (len(antSpwCombination) > 1):
            preamble += "Antenna %s is flagged in spws %s at all times<br>" % (uniqueAntNames[iant],str(antSpwCombination))
            # remove all mention of this antenna/spw combination in the preceding text
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                spwOkay =  True
                for spwcheck in antSpwCombination:
                    if (line.find('spw %d'%(spwcheck)) >= 0 or line.find('spw  %d'%(spwcheck)) >= 0):
                        spwOkay = False
                if (line.find(uniqueAntNames[iant]) < 0 or spwOkay):
                    newpage += line + '<br>'
            webpage = newpage
        elif (len(antSpwCombination) == 1):
            preamble += "Antenna %s is flagged in spw %d at all times<br>" % (uniqueAntNames[iant],antSpwCombination[0])
            # remove all mention of this antenna/spw combination in the preceding text
            newpage = ''
            lines = webpage.split('<br>')
            for line in lines:
                if (line.find(uniqueAntNames[iant]) < 0 or (line.find('spw %d'%(antSpwCombination[0])) < 0
                                                        and line.find('spw  %d'%(antSpwCombination[0])) < 0)):
                    newpage += line + '<br>'
            webpage = newpage

    if (vis != '' and casadef.casa_version >= '4.1.0'):
        mymsmd.close()
    webpage = preamble + webpage  
    return(preamble, webpage)
# end of printBandpassCaltableFlags  zsel[t]

def getMajorIntentsForFieldID(field, vm='', mymsmd=''):
    """
    Gets all the major intents for which a field was observed, either via
    ValueMapping or via the msmd tool, both passed in.  Returns a comma delimited string.
    By major, we mean everything except: CALIBRATE_WVR, CALIBRATE_ATMOSPHERE,
    CALIBRATE_SIDEBAND_RATIO, and CALIBRATE_POINTING.
    """
    if (vm == '' and mymsmd == ''):
        print "ERROR: You must specify either the vm or mymsmd parameters."
        return([])
    intents = getIntentsForFieldID(field,vm,mymsmd)
    myintents = []
    for i in intents:
        caltype = i.split('#')[0]
        if (caltype not in ['CALIBRATE_ATMOSPHERE','CALIBRATE_WVR','CALIBRATE_SIDEBAND_RATIO','CALIBRATE_POINTING']):
            myintents.append(caltype)
    intents = ''
    for i in myintents:
        if (intents != ''):
            intents += ','
        intents += i
    return(intents)

def getIntentsForFieldID(field, vm='', mymsmd=''):
    """
    Gets all the intents for which a field was observed, either via ValueMapping
    or via the msmd tool, both passed in.  Returns a list of strings.
    """
    intents = []
    if (vm == '' and mymsmd == ''):
        print "ERROR: You must specify either the vm or mymsmd parameters."
        return([])
    if (vm == ''):
        scans = mymsmd.scansforfield(field)
        for scan in scans:
            intents += mymsmd.intentsforscan(scan)
    else:
        scans = vm.getScansForFieldID(field)
        for scan in scans:
            intents += vm.getIntentsForScan(scan)
    intents = np.unique(intents)
    return(intents)
        
def getCvsVersion(versionString):
    """
    Extracts the meaningful fields from a CVS version string.
    Example: 'analyzemscal.py,v 1.24 2013/01/18 19:26:56'
    """
    tokens = versionString.split()        
    myVersion = tokens[1] + ' ' + tokens[2] + ' ' + tokens[3] + ' ' + tokens[4]
    return(myVersion)

def findPipelineTieParameter(ccl, asdmNoPath):
    """
    Accepts the pipeline casa_commands.log file as the argument, and extracts
    the wvrgcal tie and flag parameters that were used.
    """
    #grab the tie and flag parameters we want
    cclF = open(ccl, 'r')
    tieParameter = None
    flagParameter = None
    inaWVRgcal = False
    wvrCommand = ''
    nlines = 0
    for line in cclF:
        nlines += 1
        if not inaWVRgcal:
            if line.find('wvrgcal(') >= 0:
                inaWVRgcal = True
        if inaWVRgcal:
            wvrCommand += line.strip()
            if line.find(')') >= 0:
                #check we have the correct MS if more than one in PL execution
                if wvrCommand.find(asdmNoPath) > 0:
                    #check we have the call *WITH* wvrflagging
                    if wvrCommand.find('wvrflag=') > 0:
                        if wvrCommand.split('wvrflag=')[1].split(']')[0] != '[':
                            break
                    #make sure we retrieve the tie parameter even if there is
                    #only one wvrgcal call which has no wvrflagging
                    if wvrCommand.find('tie=') > 0:
                        tieParameter = 'tie=' + \
                            wvrCommand.split('tie=')[1].split(']')[0] + \
                            ']'
                wvrCommand = ''
                inaWVRgcal = False
                continue
    cclF.close()
    if wvrCommand.find('tie=') > 0:
        tieParameter = 'tie=' + wvrCommand.split('tie=')[1].split(']')[0] + ']'
    if wvrCommand.find('wvrflag=') > 0:
        flagParameter = 'wvrflag=' + wvrCommand.split('wvrflag=')[1].split(']')[0] + ']'

    #check if we weren't successful
    if (tieParameter == None):
        println(lineno(),"ERROR: Did not find wvrgcal for %s in all %d lines in %s" % (asdmNoPath,nlines,ccl))
        tieParameter = 'not found'
        flagParameter = 'not found'
    tieParameter = replaceEveryNthOccurrenceWith(tieParameter, 4, ',', ',<br>')
    if (flagParameter == None):
        flagParameter = 'not found'
    return(tieParameter, flagParameter)

def findPipelineWVRgcalAutoFlags(cpl, asdmNoPath):
    """
    Accepts the pipeline casapy.log file as the argument and extracts the
    antennas that were automatically wvrflagged by wvrgcal itself. This
    does assume that the automatic flagging is independent of the user
    specified wvrflag parameter.
    """
    clF = open(cpl, 'r')
    autoFlag = False
    lines = clF.readlines()
    clF.close()
    for i in range(len(lines)):
        if 'End Task: wvrgcal' in lines[i] and \
           'wvrgcal has flagged antennas:' in lines[i+2] and \
           asdmNoPath in lines[i-1]:
            autoFlag = True
            antList = lines[i+2].split('wvrgcal has flagged antennas:')[1]
            antList = antList.strip()
            break
    if not autoFlag:
        return 'not found'
    else:
       return antList

def findManualTieParameter(sfc):
    """
    Accepts the manual reduction script as the argument, and extracts the
    wvrgcal tie parameter that was used.
    """
    c = open(sfc,'r')
    foundwvrgcalStep = False
    tieParameter = ''
    for line in c.readlines():
        if (line.find('wvrgcal(') >= 0):
            foundwvrgcalStep = True
        if foundwvrgcalStep:
            if (line.find('tie')>=0):
                tieParameter = 'tie' + line.split('tie')[1].split(']')[0] + ']'
                break
    c.close()
    return(tieParameter)

def findManualWVRFlagParameter(sfc):
    """
    Accepts the manual reduction script as the argument, and extracts the
    antennas that were flagged in wvrgcal (if any).
    """
    c = open(sfc,'r')
    foundwvrgcal = False
    flagParameter = 'none'
    for line in c.readlines():
        if (line.find('wvrgcal') >= 0 or foundwvrgcal):
            if (line.find('wvrflag')>=0):
                flags = line.split('wvrflag')
                if flagParameter=='none': flagParameter="wvrflag"
                flagParameter += flags[1].split('--')[0]
                break
            elif (line.find('smoothcal')>=0 or line.find('mystep')>=0):
                break
            else:
                foundwvrgcal = True
    c.close()
    return(flagParameter)

def sevenMeterAntennasOnly(msWithPath):
    """
    Checks for 'CM' in antenna names, and returns False if at least one
    antenna is not a CM antenna.
    """
    antnames = au.getAntennaNames(msWithPath)
    for name in antnames:
        if ('CM' not in name):
            return(False)
    return(True)

def badCharacters(mystr):
    if (mystr.find('[') >= 0 or mystr.find(']') >= 0):
        return(True)
    else:
        return(False)

def fluxscaleRatioFromFiles(pipelineFluxscaleResults, manualFluxscaleResults, debug=False):
    """
    Takes two files one containing the pipeline fluxscale results and the
    second one containing the manual fluxscale results, and computes the
    mean ratio across all spws and fields.  Returns the
    mean and a list of all ratios.
    """
    with open(pipelineFluxscaleResults,'r') as f:
        pipe = f.read()
    with open(manualFluxscaleResults,'r') as f:
        manual = f.read()
    return(fluxscaleRatio(pipe,manual,debug))

def expectedFluxscaleRatio(body, freq):
    """
    Uses predictcomp to compute the expected flux density ratio of a planet
    between the 2012/2010 casa models.  Returns 1.0 if the body is not
    supported.  Returns NaN if predictcomp fails.
    """
    ratio = 1.0
    date = '2013-01-01'
    supportedBodies = ['venus','mars','jupiter','callisto','io','ganymede','europa',
                       'titan','uranus','neptune','ceres','pallas','vesta','juno']
    config = 'alma_cycle1_1.cfg'
    minfreq = str(freq)
    maxfreq = str(freq+1)  # Hz
    if (body.lower() in supportedBodies):
        try:
            for hour in range(0,24,4):
                epoch = date + '-%02d:00:00' % hour
                println(lineno(), "Running predictcomp('%s',minfreq='%s',maxfreq='%s',nfreqs=1,showplot=False,epoch='%s',antennalist='%s',standard='Butler-JPL-Horizons 2010')" % (body,minfreq,maxfreq,epoch,config))
                a = predictcomp(body,minfreq=minfreq,maxfreq=maxfreq,nfreqs=2,showplot=False,
                                epoch=epoch,antennalist=config,standard='Butler-JPL-Horizons 2010')
                if ('ival' in a['spectrum']):
                    break
            println(lineno(),"a['spectrum'] = %s" % (str(a['spectrum'])))
            for hour in range(0,24,4):
                epoch = date + '-%02d:00:00' % (hour)
                println(lineno(), "Running predictcomp('%s',minfreq='%s',maxfreq='%s',nfreqs=1,showplot=False,epoch='%s',antennalist='%s',standard='Butler-JPL-Horizons 2012')" % (body,minfreq,maxfreq,epoch,config))
                b = predictcomp(body,minfreq=minfreq,maxfreq=maxfreq,nfreqs=1,showplot=False,
                                epoch=epoch,antennalist=config,standard='Butler-JPL-Horizons 2012')
                if ('ival' in b['spectrum']):
                    break
            if ('ival' in b['spectrum'] and 'ival' in a['spectrum']):
                flux2010 = a['spectrum']['ival'][0] # ['bl0flux']['value']
                println(lineno(), "flux2010 = %f" % (flux2010))
                flux2012 = b['spectrum']['ival'][0] # ['bl0flux']['value']
                println(lineno(), "flux2012 = %f" % (flux2012))
                ratio = flux2012/flux2010
                println(lineno(), "ratio = %f" % (ratio))
            else:
                println(lineno(), "ERROR: predictcomp did not return an 'ival' list containing flux densities")
        except:
            println(lineno(), 'ERROR: predictcomp returned an exception')
            ratio = float('NaN')
    return(ratio)

def fluxscaleRatio(pipelineFluxscaleResults, manualFluxscaleResults, debug=False):
    """
    Takes two strings containing the pipeline and manual fluxscale results
    and computes the mean ratio across all spws and fields.  Returns the
    mean and a list of all ratios.
    """
    pipeline = pipelineFluxscaleResults.split('<br>')
    manual = manualFluxscaleResults.split('<br>')
    ratios = []
    mctr = 0
    for myline in pipeline:
        line = myline.replace('\n','')
        if (line.find('SpW=') > 0):
            try:
                pflux = float(line.split(':')[1].split('+')[0])
            except:
                continue
            for m in range(mctr, len(manual)):
                manual[m] = manual[m].replace('\n','')
                if (manual[m].find('SpW=') > 0):
#                    if (manual[m].split('SpW=')[0] != line.split('SpW=')[0]):
                    if (manual[m].split('SpW=')[0].replace(' ','').replace(',','').replace(';','') !=
                        line.split('SpW=')[0].replace(' ','').replace(',','').replace(';','')):
                        if (debug):
                            print "Source mismatch:\n%s\n%s" % (line,manual[m])
                        continue
                    if (debug):
                        print "Match:\n%s\n%s" % (line,manual[m])
                    try:
                        mflux = float(manual[m].split(':')[1].split('+')[0])
                    except:
                        break
                    mctr = m
                    ratios.append(pflux/mflux)
                    break
            if (m != mctr):
                # We did not find a match
                break
            mctr += 1
    ratio = np.mean(ratios)
    print "Found %d matches.  Mean ratio = %.3f" % (len(ratios),ratio)        
    return(ratio,ratios)

def writeStringToTextFile(filename, mystring):
    try:
        f = open(filename,'w')
    except:
        print "ERROR: Could not open file = %s" % (filename)
        return
    lines = mystring.split('<br>')
    linesWritten = 0
    for line in lines:
        if (len(line) > 0):
            if (line[-1] != '\n'):
                line += '\n'
            if (line[0] == '\n'):
                line = line[1:]
            f.write(line)
            if (len(line) > 1):
                linesWritten += 1
    f.close()
    return(linesWritten)

if False:
    # test non-default command-line generator
    amc.analyzemscal('uida', pipecaldir='a', manualcaldir='a', manualqa2dir='a',
                     webpagedir='w', target_source=1, cleanup=True,
                     uvrange='1~30m', pipelinedir='a', webarea=True,
                     refant='DV01', ampGainTableManual='a',
                     wvrstat_intent='ABANDPASS', pipeqa2dir='a',
                     force_au_plotbandpass=True, debug_plotms=False,
                     phase_cal=1, plotlevel=2, wvrstat_spw='3', wvrstat_pol='Y')

def appendIfNecessary(caltableFlags, preamble, caltableFlagsFile, nLines=0):
    """
    Takes a list of calibration table flags, and decides whether a link needs to
    be made to a longer list in a separate text file.
    """
    caltableFlagsStrip = caltableFlags.replace('\n','').replace('<br>','').replace('.','').strip()
    preambleStrip           = preamble.replace('\n','').replace('<br>','').replace('.','').strip()
    if (caltableFlagsStrip != preambleStrip):
        return('<br><a href="%s">complete listing</a> (%d lines)' % (caltableFlagsFile,nLines))
    else:
        return('')

def findMinMaxInCalibratedData(vis, spw=None, field=None):
    """
    This is an experimental function, not ready for use.
    """
    mytb = createCasaTool(tbtool)
    mytb.open(vis+'/DATA_DESCRIPTION')
    spws = mytb.getcol('SPECTRAL_WINDOW_ID')
    mytb.close()
    mytb.open(vis)
    # get the DATA_DESC_ID for the requested spw
    if (spw == None and field == None):
        mydd = mytb
    elif (spw != None and field == None):
        spw = int(spw)
        dd = spws.index(spw)
        print "Making query for spw %d via data descriptor %d" % (spw,dd)
        mydd = mytb.query('DATA_DESC_ID == %d' % (dd))
    elif (spw == None and field != None):
        mydd = mytb.query('FIELD_ID == %d' % (field))
    else:
        spw = int(spw)
        dd = spws.index(spw)
        mydd = mytb.query('FIELD_ID == %d && DATA_DESC_ID==%d' % (field,dd))
    data = mydd.getcol('CORRECTED_DATA')
    print "Got %d rows of data" % (len(data))
    mytb.close()
    # convert to amplitude, then compute min/max
    data = np.abs(data)
    mymin = np.min(data)
    mymax = np.max(data)
    return(mymin,mymax)

def getFlagStatistics(msWithPath, scienceSpws='', qa2dir='', cleanup=False, debug=False, 
                      htmlfile='', digits=5, totalOnly=False, includePointing=False):
    """
    Calls the af tool 7 times using the following 7 categories:
        'Total','Science spws','Bandpass','Flux','Phase','Target','WVR spw','Pointing'
    and collects the results into an html string.
    scienceSpws: restrict calculation to the specified spw list (a comma-delimited string)
                 default = automatically determine the science spws by intent
    cleanup: if True, then overwrite the existing htmlfile, if present
    htmlfile: the name of the one-line html file to produce, default = <ms>.flagstats.html
    totalOnly: if True, then stop after computing over the whole dataset, and return the
               full dictionary;  if False, return just the one line text information
    includePointing: if True, then also check pointing intent (against spectral spws)
    """
    if (not os.path.exists(msWithPath)):
        print "Could not find measurement set."
        return
    if (htmlfile == ''):
        htmlfile = msWithPath + '.flagstats.html'
    statsfile = qa2dir + htmlfile
    if (os.path.exists(statsfile) and cleanup==False):
        f = open(statsfile,'r')
        println(lineno(),'-------------- Reading flag statistics from existing file (%s) rather than calculating them.' % (statsfile))
        println(lineno(),'-------------- Either rename it, or set cleanup=True to overwrite it.')
        webpage = f.read()
        f.close()
        return(webpage)
    flagdict = {}
    println(lineno(),"   calculating for all data...")
    if (os.path.exists(msWithPath+'/SORTED_TABLE')):
        if (os.access(msWithPath+'/SORTED_TABLE', os.W_OK) == False):
            println(lineno(),"ERROR *** no write permission to %s/SORTED_TABLE" % (os.path.basename(msWithPath)))
            webpage = "<td>no write permission to %s/SORTED_TABLE</td>" % (os.path.basename(msWithPath))
            return(webpage)
    flagdict['Total'] = getFlagdict(msWithPath)
    if (totalOnly):
        return(flagdict)
    mymsmd = createCasaTool(msmdtool)
    mymsmd.open(msWithPath)
    intents = mymsmd.intents()
    spectralSpws = mymsmd.almaspws(tdm=True,fdm=True)
    if (scienceSpws == ''):
        if ('OBSERVE_TARGET#ON_SOURCE' in intents):
            scienceSpws = mymsmd.spwsforintent('OBSERVE_TARGET#ON_SOURCE')
            scienceSpws = ','.join([str(i) for i in np.intersect1d(scienceSpws,spectralSpws)])
    if (scienceSpws != ''):
        println(lineno(),"   for science spws: %s." % (scienceSpws))
        flagdict['Science spws'] = getFlagdict(msWithPath,spw=scienceSpws)
    else:
        flagdict['Science spws'] = None
    wvrSpw = mymsmd.almaspws(wvr=True)
    if (len(wvrSpw) < 1):
        wvrSpw == None
    else:
        wvrSpw = wvrSpw[0]
    mymsmd.close()
    try:
        println(lineno(),"   for bandpass cal")
        if ('CALIBRATE_BANDPASS#ON_SOURCE' not in intents):
            print "None observed"
            raise
        flagdict['Bandpass'] = getFlagdict(msWithPath,intent='*BANDPASS*',spw=scienceSpws)
    except:
        flagdict['Bandpass'] = None
    keys = ['Total','Science spws','Bandpass','Flux','Phase','Target','WVR spw']
    if (includePointing):
        keys.append('Pointing')
        try:
            println(lineno(),"   for pointing cal")
            if ('CALIBRATE_POINTING#ON_SOURCE' not in intents):
                print "None observed"
                raise
            spectralSpwString = ','.join([str(i) for i in spectralSpws])
            flagdict['Pointing'] = getFlagdict(msWithPath,intent='*POINTING*',spw=spectralSpwString)
        except:
            flagdict['Pointing'] = None
    try:
        println(lineno(),"   for phase cal")
        if ('CALIBRATE_PHASE#ON_SOURCE' not in intents):
            print "None observed"
            raise
        flagdict['Phase'] = getFlagdict(msWithPath,intent='*PHASE*',spw=scienceSpws)
    except:
        flagdict['Phase'] = None
    try:
        println(lineno(),"   for target")
        if ('OBSERVE_TARGET#ON_SOURCE' not in intents):
            print "None observed"
            raise
        flagdict['Target'] = getFlagdict(msWithPath,intent='*OBSERVE_TARGET*',spw=scienceSpws)
    except:
        flagdict['Target'] = None
    try:
        println(lineno(),"   for flux/amplitude")
        if ('CALIBRATE_FLUX#ON_SOURCE' not in intents):
            raise
        flagdict['Flux'] = getFlagdict(msWithPath,intent='*FLUX*',spw=scienceSpws)
    except:
        try:
            if ('CALIBRATE_AMPLI#ON_SOURCE' not in intents):
                print "None observed"
                raise
            flagdict['Flux'] = getFlagdict(msWithPath,intent='*AMPLI*',spw=scienceSpws)
        except:
            flagdict['Flux'] = None
    try:
        if (wvrSpw == None):
            print "No WVR spw."
            raise
        println(lineno(),"   for WVR spw (%d)" % wvrSpw)
        if debug: print "Calling getFlagdict('%s',spw=%s)" % (msWithPath,str(wvrSpw))
        flagdict['WVR spw'] = getFlagdict(msWithPath,spw=str(wvrSpw))
    except:
        flagdict['WVR spw'] = None
    webpage = "<td>"
    for key in keys:
        if (flagdict[key] == None):
            webpage += "%s: n/a,  " % (key)
        else:
            webpage += "%s: %.*f%%,  " % (key, digits, flagdict[key]['percent'])
        if (key == 'Bandpass'):
            webpage += "<br>"
    webpage += "</td>"
    f = open(statsfile, 'w')
    f.write(webpage)
    f.close()
    return(webpage)

def wvr_stat_all(vis, refAnt='', qa2_output_dir='', autoscale_yaxis=False, wvr_gaintable='',
                 scanintent='BANDPASS', solint='int', spw=[], pol=['X','Y'],
                 nsigma_coherence=5, unwrap_phase=False):
    """
    Runs es.wvr_stat several times, once for each spw/pol combination, and assembles all plots
    into a PDF.
    """
    es = au.stuffForScienceDataReduction()
    results = []
    println(lineno(),"Entered wvr_stat_all with spw=%s, pol=%s, gaintable=%s" % (str(spw), str(pol), wvr_gaintable))
    if (len(spw) < 1):
        println(lineno(),"ERROR: You must specify a list of spws, e.g.  spw=[9,11]")
        return
    for myspw in spw:
        for mypol in pol:
            if (scanintent == 'BANDPASS'):
                plotfile = os.path.basename(wvr_gaintable) + '.spw%s.pol%s.png' % (str(myspw),mypol)
            else:
                plotfile = os.path.basename(wvr_gaintable) + '.spw%s.pol%s.%s.png' % (str(myspw),mypol,scanintent)
            plotfileWithPath = qa2_output_dir + plotfile
            results.append(plotfileWithPath)
            if (os.path.exists(plotfileWithPath) == False):
                cmd="es.wvr_stat(ms1='%s', refAnt='%s', qa2_output_dir='%s', autoscale_yaxis=%s, wvr_gaintable='%s', scanintent='%s', solint='%s', spw='%s', pol='%s', plotfile='%s', nsigma_coherence=%f, unwrap_phase=%s)" % (vis, refAnt, qa2_output_dir, str(autoscale_yaxis), wvr_gaintable, scanintent, solint, str(myspw), mypol, plotfile, nsigma_coherence, unwrap_phase)
                println(lineno(),"     Running "+cmd)
                try:
                    exec cmd
                except:
                    println(lineno(),"ERROR: in wvr_stat_all, wvr_stat gaincal failed for spw=%s, intent=%s, pol=%s" % (str(myspw),scanintent,mypol) )
                    clearstat()
            else:
                println(lineno(),"    Skipping es.wvr_stat(spw='%s', pol='%s') because plot already exists" % (str(myspw),mypol))
    println(lineno(),"    len(results) = %d" % (len(results)))
    if (len(results) > 0):
        if (scanintent == 'BANDPASS'):
            pdfname = qa2_output_dir + os.path.basename(wvr_gaintable) + '.pdf'
        else:
            pdfname = qa2_output_dir + os.path.basename(wvr_gaintable) + '.%s.pdf' % (scanintent)
        au.buildPdfFromPngs(results, pdfname=pdfname)
        println(lineno(),"   Summary PDF left in %s" % (pdfname))
        print


def findPipelineRefspwmap(ccl, asdmNoPath, scienceSpwsIntegerList=[]):
    """
    Accepts the pipeline casa_commands.log file as the argument, and extracts
    the refspwmap parameter that was used.
    """
    c = open(ccl,'r')
    refspwmapString = None
    lines = c.readlines()
    c.close()
    # in the Sep vsn they put in linebreaks to the log as we asked, 
    # so you can't just search the line in the log
    # instead we need a flag for starting to parse that 
    # multiline command string, and collect both the ASDM and refspwmap until
    # all lines of the command are expended.
    inFluxscaleCommand=False
    for line in lines:
        if (line.find('fluxscale(') == 0): 
            inFluxscaleCommand=True
            correctASDM=False
        if inFluxscaleCommand:            
            if (line.find(asdmNoPath) >=0): correctASDM=True
            if (line.find("refspwmap") >=0):
                refspwmapString = line.split('refspwmap=[')[1].split(']')[0]
    # TODO what if refspwmap extends over multiple lines?
            if (line.find(")") >=0): 
                inFluxscaleCommand=False
                if correctASDM:
                    break

    if (refspwmapString == None):
        println(lineno(),"ERROR: Did not find refspwmap for %s in all %d lines of %s" % (asdmNoPath,len(lines),ccl))
        refspwmapParameter = 'refspwmap not found'
    else:
        refspwmapParameter = 'refspwmap=[%s]' % (refspwmapString)
        if (len(scienceSpwsIntegerList) > 0):
            tokens = refspwmapString.split(',')
            if (len(tokens) > np.max(scienceSpwsIntegerList)):
                refspwmap = [int(x) for x in tokens]
                mymap = str(list(np.array(refspwmap)[scienceSpwsIntegerList]))
                refspwmapParameter = 'refspwmap%s = %s' % (str(scienceSpwsIntegerList),mymap)
    return(refspwmapParameter)


def extractWvrgcalOutputFromLog(casalog, asdm, debug=False, nsigma_discrepancy=5):
    """
    Searches for the first appearance of a wvrgcal command with the
    specified ASDM on the same line, then
    searches for the beginning of the wvrgcal output table.  Copies
    it to a list.  Then searches for a second version and, if present,
    replaces the list.  Stops if it sees:
    1) an importasdm -- to treat sequential single=ASDM pipeline runs
    2) or a wvrgcal command that does not contain the asdm name -- to
         properly treat multi-ASDM pipeline runs.
    Returns the final list of lines.
    """
    clog = open(casalog,"r")
    lines = clog.readlines()
    clog.close()
    if (debug): println(lineno(),"extractWvrgcalOutputFromLog(): Read %d lines" % (len(lines)))
    if (debug): println(lineno(),"Searching for wvrgcal command on asdm = %s" % (asdm))
    copy = False
    foundAsdm = False
    copylines = []
    discrepancy = []
    antNames = []
    startLine = 'Retrieved parameters' # 'Antenna/WVR information'
    if heuristicsversion()>=2014.02:
        importdata_taskname='hifa_importdata'
    else:
        importdata_taskname='hif_importdata'
    for l in range(len(lines)):
        line = lines[l]
        if (foundAsdm == False):
            if (line.find(asdm)>=0 and line.find('wvrgcal(vis') >= 0):
                foundAsdm = True
                if (debug): println(lineno(),"Found wvrgcal ASDM on line %d" % (l))
        else:
            if (line.find(startLine) >= 0):
                if (debug): println(lineno(),"Found '%s' on line %d" % (startLine,l))
                copy = True
                copylines = []  # start fresh
                discrepancy = []
                antNames = []
            if (copy):
                copylines.append(line)
                if (line.find('Yes')>0 or line.find('No')>0):
                    if (debug): print "Splitting line = ", line
                    # extract discrepancy, after the casa\t -------v      
                    # 2013-06-26 14:17:36   INFO    wvrgcal::::casa 0       DA41    Yes     No      7.91e+03                8.61e+04
                    a,b,c,d,e,f = line.split('casa\t')[1].split()
                    discrepancy.append(float(f))
                    antNames.append(b)
            if (line.find('Rough estimate path error due') >= 0):
                copy = False
            if (len(copylines)>0 and line.find(importdata_taskname) >= 0):
                if (debug): println(lineno(),"Found "+importdata_taskname+" line %d. Stopping." % (l))
                break
            if (len(copylines)>0 and (line.find('wvrgcal(vis')>0 and line.find(asdm)<0)):
                if (debug): println(lineno(),"Found wvrgcal on a different ASDM on line %d. Stopping." % (l))
                break
    if (len(discrepancy) > 0):
        discrepancyMAD = mad(discrepancy)
        discrepancyMedian = np.median(discrepancy)
        outliers = []
        for i in range(len(discrepancy)):
            if (abs(discrepancy[i] - discrepancyMedian) > nsigma_discrepancy*discrepancyMAD):
                outliers.append(antNames[i])
        if (len(outliers) == 0):
            outliers = 'none'
        else:
            outliers = ','.join(outliers)
        outlierReport =  '-----------------------------------<br>'
        outlierReport += 'Outliers in discrepancy [>%g*MAD (>%g) from the median (%g)]:<br>  %s' % (nsigma_discrepancy,
                   nsigma_discrepancy*discrepancyMAD, discrepancyMedian, outliers)
    return(copylines, outlierReport)

def picktile(pnglist):
    tile = '1x1'
    if (len(pnglist) == 2):
        tile = '2x1'
    if (len(pnglist) > 2):
        tile = '2x2'
    return(tile)

def findFluxcalflagLinesFile(myppr):
    """
    Search the PPR for the keyword linesfile, and return its value if present.
    """
    linesfile = None
    try:
        f = open(myppr,'r')
    except:
        return(linesfile)
    lines = f.readlines()
    for l in range(len(lines)):
        if (lines[l].find('linesfile') > 0):
            linesfile = lines[l+1].split('<Value>')[1].split('</Value>')[0]
            break
    f.close()
    return(linesfile)

def getManualWvrstatRefAnt(manualqa2dir, refAnt0):
    """
    Tries to open the manual reduction's wvr_stat.txt file and extract the
    reference antenna.  If it doesn't find it, it returns refAnt0, which was
    passed into the function as the default value.
    """
    manualWvrstatRefAnt = refAnt0
    firstLine = []
    if (os.path.exists(manualqa2dir+'/wvr_stat.txt')):
        wvrstatfile = open(manualqa2dir+'/wvr_stat.txt','r')
        firstLine = wvrstatfile.readlines()[0].split('refAnt=')
        wvrstatfile.close()
    if (len(firstLine)>0):
        manualWvrstatRefAnt = firstLine[1].split()[0]
    return(manualWvrstatRefAnt)

def makeAntennaConfigZoomPlotIfNecessary(maxBaseline, antennaConfigZoomPlot,
                                         msWithPath, qa2dir):
    es = au.stuffForScienceDataReduction()
    if (maxBaseline > 400):
        xlim = [-140,240]
        ylim = [-150,150]
        # if the configuration is large, make another plot zoomed
        # to the array center
        if (os.path.exists(antennaConfigZoomPlot)):
            println(lineno(),"Found zoomed antenna configuration plot: %s" % (antennaConfigZoomPlot))
        else:
            println(lineno(),"Running es.listobs2('%s', qa2_output_dir='%s', plotAntennasActualSize=True, xlim=%s, ylim=%s)" % (msWithPath,qa2dir,str(xlim),str(ylim)))
            es.listobs2(ms2=msWithPath, qa2_output_dir=qa2dir,
                        plotAntennasActualSize=True, xlim=xlim, ylim=ylim)

def printWvrStatErrorMessage(msWithPath, lineNumber):
    if (os.access(msWithPath,os.W_OK) == False):
        message = "This is likely because you do not have write privilege on the ms = %s" % (msWithPath)
    else:
        message = "It could be a problem with tsysspwmap."
    println(lineNumber,message)
    return(message + '<br>\n')

def gatherPipelineStats(asdm, pipeCalDir, writeableMSWithPathPL, pipeQA2Dir, \
                        ccl, webPageDir, sevenMeterOnly, pipelineCasaVersion, \
                        manCalDir=''):
    """
    Currently (3/4/14) gathers WVR related statistics from pipeline
    execution and manual reduction and saves them to an XML file in
    the web page directory. This will eventually be expanded to also
    gather low gain flagging information.
    """
    #first check if this is an ACA data set
    if sevenMeterOnly:
        println(lineno(), 'Data set contains only 7m antennas so no WVR ' + \
                'data is available. No WVR statistics file will be ' + \
                'generated from this ASDM.')
        return
    #all deepest level list elements are stored as strings
    minedStats = dict()
    if pipeCalDir[-1] != '/': pipeCalDir += '/'
    if pipeQA2Dir[-1] != '/': pipeQA2Dir += '/'
    if len(manCalDir) > 0:
        if manCalDir[-1] != '/': manCalDir += '/'
    if webPageDir[-1] != '/': webPageDir += '/'
    #store the project code
    if pipeCalDir.find('/SOUS') > 0:
        tmp = pipeCalDir.split('/SOUS')
        tmp = tmp[0].split('/')
        tmp = tmp[-1].split('_')
    else:
        tmp = pipeCalDir.split('/MOUS')
        tmp = tmp[0].split('/')
        tmp = tmp[-1].split('_')
    minedStats['project'] = tmp[0]
    #store the ASDM disk name
    minedStats['ASDM'] = asdm
    #initialize the frequency dictionary
    minedStats['frequencies'] = dict()
    #initialize the antenna dictionary
    minedStats['antennas'] = dict()

    #fill in the per antenna phase RMS before and after WVR and coherences
    #for wvrgcal with and without wvrflagging
    wvrStatList = glob.glob(pipeQA2Dir+'wvr_stat*.txt')
    if len(wvrStatList) == 0:
        println(lineno(), 'ERROR: Cannot find pipeline wvr_stat files. ' + \
                          'These files must be present to gather all WVR ' + \
                          'statistics.')
        return
    else:
        for wvrStatF in wvrStatList:
            justFile = wvrStatF.split('/')[-1]
            if justFile != 'wvr_stat.txt' and \
               justFile != 'wvr_stat_wvr_plot_autoscale.txt'and \
               'CALIBRATE_PHASE' not in justFile:
                splitF = justFile.split('.')
                spw = splitF[-3].split('spw')[1]
                pol = splitF[-2].split('pol')[1]
                #fill out the SPWs we want the frequencies for
                if spw not in minedStats['frequencies']:
                    minedStats['frequencies'][spw] = ''
                fileObject = open(wvrStatF)
                for line in fileObject:
                    if (line.find('wvrgcal output of average wvr')>=0):
                        break
                    if (line.find("*******************")>=0):
                        continue
                    if (line.find('DA')==0 or \
                        line.find('DV')==0 or \
                        line.find('PM')==0 or \
                        line.find('CM')==0):
                        tokens = line.split()
                        if tokens[0] not in minedStats['antennas']:
                            minedStats['antennas'][tokens[0]] = \
                                {'NoFlagPrePhaseRMS': dict(), \
                                 'NoFlagPostPhaseRMS': dict(), \
                                 'FlagPrePhaseRMS': dict(), \
                                 'FlagPostPhaseRMS': dict(), \
                                 'NoFlagPreCoherence': dict(), \
                                 'NoFlagPostCoherence': dict(), \
                                 'FlagPreCoherence': dict(), \
                                 'FlagPostCoherence': dict(), \
                                 'AvgPWV': '', \
                                 'AvgCalEl': '', \
                                 'AvgBaselineLength': '', \
                                 'ManWVRFlagged': '', \
                                 'PLWVRFlagged': ''}
                        if 'noflag' in justFile:
                            minedStats['antennas'][tokens[0]]['NoFlagPrePhaseRMS'][pol+spw] = tokens[2]
                            minedStats['antennas'][tokens[0]]['NoFlagPostPhaseRMS'][pol+spw] = tokens[3]
                            minedStats['antennas'][tokens[0]]['NoFlagPreCoherence'][pol+spw] = tokens[5]
                            minedStats['antennas'][tokens[0]]['NoFlagPostCoherence'][pol+spw] = tokens[6]
                        else:
                            minedStats['antennas'][tokens[0]]['FlagPrePhaseRMS'][pol+spw] = tokens[2]
                            minedStats['antennas'][tokens[0]]['FlagPostPhaseRMS'][pol+spw] = tokens[3]
                            minedStats['antennas'][tokens[0]]['FlagPreCoherence'][pol+spw] = tokens[5]
                            minedStats['antennas'][tokens[0]]['FlagPostCoherence'][pol+spw] = tokens[6]
                fileObject.close()

    #fill in the frequencies
    mymsmd = au.createCasaTool(msmdtool)
    mymsmd.open(writeableMSWithPathPL)
    for spw in minedStats['frequencies']:
        minedStats['frequencies'][spw] = str(mymsmd.meanfreq(int(spw)))
    mymsmd.close()

    #fill in the average PWV for the EB (per antenna)
    pwvRecords = au.readpwv(pipeCalDir + '../rawdata/' + asdm)
    tmpDict = dict()
    for i in range(len(pwvRecords[2])):
        strippedAnt = pwvRecords[2][i].strip()
        if strippedAnt not in tmpDict:
            tmpDict[strippedAnt] = np.array(pwvRecords[1][i])
        else:
            tmpDict[strippedAnt] = np.append(tmpDict[strippedAnt], \
                                                pwvRecords[1][i])
    #put the average PWVs into minedStats
    for ant in tmpDict:
        if ant in minedStats['antennas']:
            minedStats['antennas'][ant]['AvgPWV'] = str(tmpDict[ant].mean())

    #fill in the average elevation, per antenna, **for bandpass calibrator**
    vm = au.ValueMapping(writeableMSWithPathPL)
    bpScans = au.getScansForIntentFast(vm, vm.uniqueScans, \
                                       'CALIBRATE_BANDPASS#ON_SOURCE')
    for ant in minedStats['antennas']:
        minedStats['antennas'][ant]['AvgCalEl'] = str(au.listazel(vis=writeableMSWithPathPL, \
                                                      scan=bpScans, \
                                                      vm=vm, antenna=ant)[1])

    #fill in the average baseline length to each antenna, projected to
    #phase calibrator field
    tmpDict = dict()
    phaseCalScans = au.getScansForIntentFast(vm, vm.uniqueScans, \
                                             'CALIBRATE_PHASE#ON_SOURCE')
    phaseCalField = vm.getFieldsForScan(phaseCalScans[0], returnName=False)[0]
    allBaselines = au.getBaselineLengths(writeableMSWithPathPL, \
                                         field=phaseCalField)
    for i in range(len(allBaselines)):
        ant1 = allBaselines[i][0].split('-')[0]
        ant2 = allBaselines[i][0].split('-')[1]
        if ant1 not in tmpDict:
            tmpDict[ant1] = np.array(allBaselines[i][1])
        else:
            tmpDict[ant1] = np.append(tmpDict[ant1], allBaselines[i][1])
        if ant2 not in tmpDict:
            tmpDict[ant2] = np.array(allBaselines[i][1])
        else:
            tmpDict[ant2] = np.append(tmpDict[ant2], allBaselines[i][1])
    #put the average baselines into minedStats
    for ant in tmpDict:
        if ant in minedStats['antennas']:
            minedStats['antennas'][ant]['AvgBaselineLength'] = \
                str(tmpDict[ant].mean())

    #fill in whether the manual reducer wvrflagged antennas
    if len(manCalDir) > 0:
        sfc = find_manual_script(manCalDir)
        manWVRFlags = findManualWVRFlagParameter(manCalDir + sfc)
        #check the returned flags and fill out minedStats accordingly
        if manWVRFlags == 'none':
            for ant in minedStats['antennas']:
                minedStats['antennas'][ant]['ManWVRFlagged'] = 'False'
        else:
            for ant in minedStats['antennas']:
                if ant in manWVRFlags:
                    minedStats['antennas'][ant]['ManWVRFlagged'] = 'True'
                else:
                    minedStats['antennas'][ant]['ManWVRFlagged'] = 'False'
    else:
        for ant in minedStats['antennas']:
            minedStats['antennas'][ant]['ManWVRFlagged'] = 'None'

    #fill in whether the pipeline wvrflagged antennas
    plTie, plWVRFlags = findPipelineTieParameter(ccl, asdm)
    print plWVRFlags
    if plWVRFlags == 'not found':
        for ant in minedStats['antennas']:
            minedStats['antennas'][ant]['PLWVRFlagged'] = 'False'
    else:
        for ant in minedStats['antennas']:
            if ant in plWVRFlags:
                minedStats['antennas'][ant]['PLWVRFlagged'] = 'True'
            else:
                minedStats['antennas'][ant]['PLWVRFlagged'] = 'False'

    #fill in whether wvrgcal wvrflagged antennas
    cpl = findPipelineCasaLogfile(pipeCalDir, asdm)
    wvrgFlags = findPipelineWVRgcalAutoFlags(cpl, asdm)
    if wvrgFlags == 'not found':
        for ant in minedStats['antennas']:
            minedStats['antennas'][ant]['WVRgcalWVRFlagged'] = 'False'
    else:
        for ant in minedStats['antennas']:
            if ant in wvrgFlags:
                minedStats['antennas'][ant]['WVRgcalWVRFlagged'] = 'True'
            else:
                minedStats['antennas'][ant]['WVRgcalWVRFlagged'] = 'False'

    #write minedStats out to XML file
    #setup the header and EB wide information
    statsLines = ['<?xml version="1.0" encoding="UTF-8"?>\n']
    statsLines.append('<!--\n')
#    statsLines.append('Header information can be automatically input\n')
    statsLines.append('Pipeline reduction used: ' + pipelineCasaVersion + '\n')
    statsLines.append('Using analyzemscal.py version: ' + \
                      version().split()[2] + '\n')
    statsLines.append('-->\n')
    statsLines.append('<SciPipeStats>\n')
    statsLines.append('    <Project>' + minedStats['project'] + '</Project>\n')
    statsLines.append('    <ASDM>' + minedStats['ASDM'] + '</ASDM>\n')
    for spw in sorted(minedStats['frequencies'].keys(), key=int):
        statsLines.append('    <SPW>\n')
        statsLines.append(' '*8 + '<ID>' + spw + '</ID>\n')
        statsLines.append(' '*8 + '<Freq>' + minedStats['frequencies'][spw] + \
                          '</Freq>\n')
        statsLines.append('    </SPW>\n')
    statsLines.append('    <WVR>\n')
    #loop through antennas making their individual entries
    for ant in sorted(minedStats['antennas']):
        statsLines.append(' '*8 + '<Antenna>\n')
        statsLines.append(' '*12 + '<Name>' + ant + '</Name>\n')
        for val in minedStats['antennas'][ant]:
            if val == 'NoFlagPrePhaseRMS' or val == 'NoFlagPostPhaseRMS' or \
               val == 'FlagPrePhaseRMS' or val == 'FlagPostPhaseRMS' or \
               val == 'NoFlagPreCoherence' or val == 'NoFlagPostCoherence' or \
               val == 'FlagPreCoherence' or val == 'FlagPostCoherence':
                for iVal in minedStats['antennas'][ant][val]:
                    statsLines.append(' '*12 + '<' + val + '>\n')
                    statsLines.append(' '*16 + '<PolSPW>' + iVal + '</PolSPW>\n')
                    if val == 'NoFlagPreCoherence' or \
                       val == 'NoFlagPostCoherence' or \
                       val == 'FlagPreCoherence' or \
                       val == 'FlagPostCoherence':
                        statsLines.append(' '*16 + '<Coherence>' + \
                                          minedStats['antennas'][ant][val][iVal] + \
                                      '</Coherence>\n')
                    else:
                        statsLines.append(' '*16 + '<RMS>' + \
                                          minedStats['antennas'][ant][val][iVal] + \
                                          '</RMS>\n')
                    statsLines.append(' '*12 + '</' + val + '>\n')
            else:
                statsLines.append(' '*12 + '<' + val + '>' + \
                                  minedStats['antennas'][ant][val] + \
                                  '</' + val + '>\n')
        statsLines.append(' '*8 + '</Antenna>\n')
    #close out the top level tags
    statsLines.append('    </WVR>\n')
    statsLines.append('</SciPipeStats>\n')
    #save the file with name [project code]_[ASDM]_stats.xml to webpage dir
    outFileSave = pipeQA2Dir + minedStats['project'] + '_' + \
                  minedStats['ASDM'] + '_stats.xml'
    outFileCopy = webPageDir + minedStats['project'] + '_' + \
                  minedStats['ASDM'] + '_stats.xml'
    with open(outFileSave, 'w') as outFile:
        outFile.writelines(statsLines)
    myCopyfile(outFileSave, outFileCopy)
#end gatherPipelineStats

















































#===================================================================
# MAIN
#===================================================================
#===================================================================
#===================================================================
uvrange_default = '0~40m'
def analyzemscal(asdm, 
                 pipecaldir='', # the pipeline run
                 pipelinedir='', # the subdir asdm+date
                 pipeqa2dir='', # pipeline qa2 output
                 manualcaldir='', # the manual reduction
                 manualqa2dir='', # manual qa2
                 manualqa2dir_extra='', # qa2 output on manual reduction
                 amc_outdir="", # place to put writeable qa2 dirs
                 webarea=False,
                 webpagedir='web', 
                 webpagedir_suffix='', 
                 cleanup=False,
                 target_source=0, 
                 uvrange=uvrange_default, 
                 refant='', ampGainTableManual='',
                 wvrstat_intent='BANDPASS', 
                 force_au_plotbandpass=False, debug_plotms=False, debug=False,
                 phase_cal='', plotlevel=1, wvrstat_spw='', wvrstat_pol='',
                 correlation='', bandpass_cal='', 
                 showFlagStats=True, edgeChannels=10, avgchannelTarget=5,
                 solutionTimeThresholdSeconds=30.0, nsigma_coherence=5,
                 unwrap_phase=False, autoscale_yaxis=False, wvr_refant='',
                 nsigma_discrepancy=5,
                 match_man_refant=True, ignoreZeroCoords=False):
    """
    This script generates ALMA Cycle 0+1+2 QA2 products for a
    pipeline-reduced dataset for the purposes of comparing the
    pipeline calibration to the manual calibration.  You do not need
    to be in any specific directory when you run it.  - Todd Hunter

    Parameters - (*) required, (+) common:

  * asdm:  the name of the ASDM (or ms) to plot (any path will be stripped)
  * pipecaldir:  the absolute directory (or '.') to find the pipeline result
                 for a single execution
    pipelinedir: the pipeline-YYYYMMDDTHHMMSS directory to search for the
            casa_commands.log file (default = most recent)
    pipeqa2dir: optional (default = pipecaldir+'qa2/'+asdm)

  + manualcaldir:  the absolute directory to find the manual result (optional)
    manualqa2dir:  the absolute directory to find the existing manual QA2
                   results (default = manualcaldir+'/qa2')
    manualqa2dir_extra:  the absolute directory to write extra manual QA2 plots
        (default: pipecaldir=='': pwd/manualqa2dir_extra/uid...
                            else: pipecaldir+'/qa2')
    match_man_refant: use the manual refant for all diagnostics

  + amc_outdir: the directory to put analyzemscal results in - especially
        important if you don't have write permission in either the pipeline 
        or manual directories.  default = "./amc/"

    webarea: True: precede webpagedir with
                   /home/safe.nrao.edu/content/alma/PipelineTestResults/
  + webpagedir: the relative directory to write a subdirectory (called the
                asdm name) which will contain the html & pngs (default='web/')
    webpagedir_suffix: optional suffix to add to the ASDM name to distinguish
                       multiple runs in the same webpagedir

    cleanup:  True will cause all old plots and directories created by this
              script to be removed prior to creating new plots.

    wvrstat_intent: the scan intent to use for the thumbnail image and the
                    statistics from es.wvr_stat()
    wvrstat_spw: the one spw to use in es.wvr_stat() as the thumbnail image
    wvrstat_pol: the one polarization to use in es.wvr_stat() as the thumbnail
                 image ('X' or 'Y')
    wvr_refant: override the pipeline's choice of refant when calling
                es.wvr_stat and re-generate all plots
    unwrap_phase: passed to es.wvr_stat()
    autoscale_yaxis: passed to es.wvr_stat()
    target_source:  integer, the Nth science target, which is used to pick a
                    phase calibrator for display on the phase vs. time plot
                    (default=0)
    uvrange:  Used to generate average target_spectrum only (default='0~40m')
    ampGainTableManual:  in case the manual reduction's split.flux_inf table
              is called something else, specify it here
    force_au_plotbandpass: default=True (as it has some newer features in it)
              False: if the task plotbandpass is present in casa, then use it.
              True: use au.plotbandpass()
    solutionTimeThresholdSeconds: passed to plotbandpass (default=30)
    debug_plotms: if True (default), it will detect missing plotms plotfiles
          and create a python script to run manually to try to make them again.
    phase_cal: the phase_cal ID to use instead of calling es.getPhaseCal()
    bandpass_cal: the bandpass_cal ID to use instead of the result from
                  es.getIntentsAndSourceNames
    plotlevel: 1=basic plots only;  2=per antenna plots as multipage PDFs
    correlation: the correlations to pass to plotms calls
    showFlagStats: default=True, set to False to speed up the results
    edgeChannels: default=10, Num of channels to ignore in counting flagged
                  Tsys channels
    avgchannelTarget: default=5, Num of channels to average in target spectrum
    nsigma_coherence: value to define outlier antennas in es.wvr_stat 
    nsigma_discrepancy: value to define outliers in extractWvrgcalOutputFromLog
    ignoreZeroCoords: if True, then don't stop if planets have 0,0 coordinates
    """
    global terminal

    # plotms options to be set in every plotms call
    if casadef.casa_version >= '4.2.0':
        plotmsparms="showgui=False,overwrite=True"
    else:
        plotmsparms="showgui=False"
    plotcalparms="showgui=False"
    plotcalshowgui=False
    # keep buildpdffrompsngs quiet
    auQuiet=True

    phase_cal = str(phase_cal)
    if (phase_cal != ''):
        if (phase_cal.isdigit() == False):
            print "ERROR: The phase_cal (if specified) must be an integer."
            return

    startTime = time.time()
    pbp_version = '4.1.0'  # first casa version with plotbandpass
    backend = pl.get_backend()

    if (backend == 'Agg' or backend == 'agg'):
        print "ERROR: The %s backend does not allow plotcal to produce multi-panel plots, even with showgui=False." % backend
        print "ERROR: Please run this script in a normal casapy, with the TkAgg backend."
        return

    print "analyzemscal's casalogfile = ", casalog.logfile()

    if (len(pipelinedir) > 0):
        # protect against the case where someone gives full path to this.
        pipelinedir = os.path.basename(pipelinedir)

    if (False):
        # old style: show all args  (not up to date)
        commandLine = "amc.analyzemscal(asdm='%s', pipecaldir='%s', manualcaldir='%s', manualqa2dir='%s', webpagedir='%s', target_source=%d, cleanup=%s, uvrange='%s', pipelinedir='%s', webarea=%s)" % (asdm, pipecaldir, manualcaldir, manualqa2dir, webpagedir, target_source, cleanup, uvrange, pipelinedir, webarea)
    else:
        # show only the non-default args
        commandLine = "amc.analyzemscal(asdm='%s'" % asdm
        if (pipecaldir != ''):
            commandLine += ", pipecaldir='%s'" % pipecaldir
        if (manualcaldir != ''):
            commandLine += ", manualcaldir='%s'" % manualcaldir
        if (manualqa2dir != ''):
            commandLine += ", manualqa2dir='%s'" % manualqa2dir
        if (manualqa2dir_extra != ''):
            commandLine += ", manualqa2dir_extra='%s'" % manualqa2dir_extra
        if (webpagedir != 'web'):
            commandLine += ", webpagedir='%s'" % webpagedir
        if (target_source != 0):
            commandLine += ", target_source=%d" % target_source
        if (cleanup != False):
            commandLine += ", cleanup=%s" % cleanup
        if (uvrange != uvrange_default):
            commandLine += ", uvrange='%s'" % uvrange
        if (pipelinedir != ''):
            commandLine += ", pipelinedir='%s'" % pipelinedir
        if (webarea != False):
            commandLine += ", webarea=%s" % webarea
        if (refant != ''):
            commandLine += ", refant='%s'" % refant
        if (ampGainTableManual != ''):
            commandLine += ", ampGainTableManual='%s'" % ampGainTableManual
        if (wvrstat_intent != 'BANDPASS'):
            commandLine += ", wvrstat_intent='%s'" % wvrstat_intent
        if (pipeqa2dir != ''):
            commandLine += ", pipeqa2dir='%s'" % pipeqa2dir
        if (force_au_plotbandpass != True):
            commandLine += ", force_au_plotbandpass=%s" % force_au_plotbandpass
#        if (debug_plotms != True):
#            commandLine += ", debug_plotms=%s" % debug_plotms
        if (phase_cal != ''):
            commandLine += ", phase_cal='%s'" % phase_cal
        if (plotlevel != 1):
            commandLine += ", plotlevel=%d" % plotlevel
        if (wvrstat_spw != ''):
            commandLine += ", wvrstat_spw='%s'" % wvrstat_spw
        if (wvrstat_pol != ''):
            commandLine += ", wvrstat_pol='%s'" % wvrstat_pol
        if (correlation != ''):
            commandLine += ", correlation='%s'" % correlation
        if (showFlagStats != True):
            commandLine += ", showFlagStats=%s" % showFlagStats
        if (edgeChannels != 10):
            commandLine += ", edgeChannels=%d" % edgeChannels
        if (avgchannelTarget != 5):
            commandLine += ", avgChannelTarget=%d" % avgChannelTarget
        if (solutionTimeThresholdSeconds != 30.0):
            commandLine += ", solutionTimeThresholdSeconds=%g" % solutionTimeThresholdSeconds
        if (nsigma_coherence != 5):
            commandLine += ", nsigma_coherence=%g" % nsigma_coherence
        if (unwrap_phase != False):
            commandLine += ", unwrap_phase=%s" % str(unwrap_phase)
        if (autoscale_yaxis != False):
            commandLine += ", autoscale_yaxis=%s" % str(autoscale_yaxis)
        if (wvr_refant != ''):
            commandLine += ", wvr_refant='%s'" % str(wvr_refant)
        if (webpagedir_suffix != ''):
            commandLine += ", webpagedir_suffix='%s'" % webpagedir_suffix
        commandLine += ")"
        print "commandLine = ", commandLine
        
    showSkippingMessages = True
#    if (plotlevel == 2 and casadef.casa_version >= '4.1.0' and os.getenv('USER').find('thunter')<0 and os.getenv('USER').find('rindebet')<0):
#        print "For plotlevel=2, please use casa 4.0.1 until further notice, due to CAS-5102, 5214, 5301."
#        return
    skipIndent = ' '*10
    copyIndent = skipIndent
    pipecaldir = pipecaldir.strip()
    if (pipecaldir == './' or pipecaldir == '.'):
        # This is needed to find the pipeline-YYYYMMDD file later
        pipecaldir = os.getcwd()
    manualcaldir = manualcaldir.strip()
    if (manualqa2dir == ''):
        # can leave double // before qa2 but that's ok
        manualqa2dir = manualcaldir + '/qa2/'
    eid = asdm
    if (eid[-3:] == '.ms'):
        ms = eid
        asdmNoPath = eid[:-3]
    else:
        asdmNoPath = eid
        ms = eid + '.ms'
    if (asdmNoPath.find('uid') < 0):
        print "ERROR: The asdm name is invalid, it does not start with 'uid'."
        return
    ms = os.path.basename(ms)

    if (pipecaldir != ''):
        if (pipecaldir[-1] != '/'): 
            pipecaldir += '/'  # be sure that pipecaldir ends in a slash
        asdm = pipecaldir+'../rawdata/'+asdmNoPath
        showingPipelineQA2 = True
    elif (manualcaldir == ''):
        print "ERROR: You must specify either pipecaldir or manualcaldir (or both)."
        return
    else:
        if (manualcaldir[-1] != '/'): 
            manualcaldir += '/'  # be sure that manualcaldir ends in a slash
        asdm = manualcaldir + asdmNoPath
        showingPipelineQA2 = False

    # insure that all directories have trailing slash
    if (manualcaldir != ''):
        showingManualQA2 = True
        if (os.path.exists(manualcaldir) == False):
            print "ERROR: The manualcaldir does not exist: ", manualcaldir
            return
        if (manualcaldir[-1] != '/'): 
            manualcaldir += '/'  # be sure that manualcaldir ends in a slash
        msSplitWithPath = manualcaldir+ms+'.split'
        if (len(os.listdir(manualqa2dir)) < 1):
            print "ERROR: The manualqa2dir directory is empty.  You need to go there and run es.generateQA2Report(),"
            # TODO run generateQA2Report for the user
            print "or specify a different location (on the command line) if it exists elsewhere."
            return
    else:
        showingManualQA2 = False

    # msWithPath gets defined as the ms in the pipeline area, unless we
    # are not showing the pipeline results, in which case it gets defined as
    # the ms in the manual reduction area.  It is only used by functions where
    # it doesn't matter which of these two ms we get.
    if (showingPipelineQA2):
        msWithPath = pipecaldir+ms
        msWithPathPipeline = msWithPath
        if (os.path.exists(pipecaldir) == False):
            print "ERROR: The pipecaldir does not exist: ", pipecaldir
            return
        if (pipelinedir != ''):
            pipelineYYYYMMDD = os.path.basename(pipelinedir)
        else:
            pipelineYYYYMMDD = findPipelineDir(pipecaldir, asdmNoPath)
            if (pipelineYYYYMMDD == []):
                print "ERROR: Could not find pipeline execution directory.  Check the ASDM name?"
                return
        if (pipelineYYYYMMDD == []):
            print "ERROR: Could not find pipeline execution directory.  Check the ASDM name?"
            return
        ccl = find_casa_commands_log(pipecaldir, pipelineYYYYMMDD)
        if (ccl == None): return
        ccl = pipecaldir + ccl
        setheuristicsversion(pipecaldir,pipelineYYYYMMDD)
    else:
        msWithPath = manualcaldir+ms


    if (showingManualQA2):
        msWithPathManual = manualcaldir+ms

    if (os.path.exists(msWithPath) == False):
        print "ERROR: ms not found: ", msWithPath
        return
        



    #================================================
    # set up output directories
    #===============================================
    if (webarea):
        webpagedir = '/home/safe.nrao.edu/content/alma/PipelineTestResults/' + webpagedir
    if (webpagedir[-1] != '/'):
        webpagedir += '/'        # be sure that webpagedir ends in a slash
    webpagedir += asdmNoPath + webpagedir_suffix + '/'

    if (os.path.exists(webpagedir) == False):
        println(lineno(),"----- Making directories:  %s -----" % (webpagedir))
        os.makedirs(webpagedir)


    # -------------
    if (cleanup):
        # TODO make robust to not having write permission 
        println(lineno(),"WARNING: Removing webpagedir %s and plot files in %s" % (webpagedir, pipecaldir))
        os.system('rm -rf %s' % (webpagedir))
        if (showingPipelineQA2):
            if not os.access(pipecaldir,os.W_OK):
                print "ERROR: can't cleanup "+pipecaldir
                print "because no write permission there"
                return
            os.system('rm -rf %s/%s*.pdf' % (pipecaldir,asdmNoPath))
            os.system('rm -rf %s/%s*.plots' % (pipecaldir,asdmNoPath))


    # -------------
    if len(amc_outdir)>0:
        if amc_outdir[-1]!="/": amc_outdir=amc_outdir+"/"        
        if os.access(amc_outdir, os.W_OK)==False:
            if os.access("./amc/", os.W_OK)==False:
                amc_outdir="./amc/"            
    # at this point amc_outdir is set to something writeable, or not set at all

    # -------------
    if showingPipelineQA2:
        if not os.path.exists(pipeqa2dir):
            if os.path.exists(amc_outdir): # if this is set, put qa2 there
                pipeqa2dir=amc_outdir+"%s/%s/" % (asdmNoPath,pipelineYYYYMMDD)
            else: # otherwise if its unset, try to put it in the pipecaldir
                pipeqa2dir=pipecaldir+'qa2/%s/' % (asdmNoPath)
        # either way, try to create pipeqa2dir if its not there
        if not os.path.exists(pipeqa2dir):
            try:
                os.makedirs(pipeqa2dir)
            except:
                println(lineno(),"ERROR: can't create %s" % pipeqa2dir)
                # fall back to putting it in ./amc
                pipeqa2dir="./amc/%s/%s/" % (asdmNoPath,pipelineYYYYMMDD)
                if not os.path.exists(pipeqa2dir): os.makedirs(pipeqa2dir)

        # now make sure we can write to it
        if os.access(pipeqa2dir, os.W_OK)==False:
            println(lineno(),"WARNING: can't write to %s" % pipeqa2dir)
            # and fall back again to ./amc
            pipeqa2dir='./amc/%s/%s/' % (asdmNoPath,pipelineYYYYMMDD)
            if not os.path.exists(pipeqa2dir): os.makedirs(pipeqa2dir)
            println(lineno(),' Pipeline plots will go to '+pipeqa2dir)
    
        if cleanup:
            println(lineno(),"WARNING: Removing files in %s" % pipeqa2dir)
            os.system('rm -rf %s/*' % (pipeqa2dir))


        # if we can't write to the MS...
        if os.access(msWithPathPipeline, os.W_OK)==True:
            writeableMSWithPathPipeline=msWithPathPipeline
        else:
            writeableMSWithPathPipeline=pipeqa2dir+ms
            
            bar=writeableMSWithPathPipeline
            if bar[-1]=="/": bar=bar[:-1]
            myCopyfile(msWithPathPipeline+"/",bar) # needs trailing slash on msWithPathPipeline, NOT on destination, because of how rsync works

            # do we only need to copy SORTED_TABLE and any locks?
            #os.system("mkdir "+writeableMSWithPathPipeline)
            #os.system("cp -r "+msWithPathPipeline+"/*lock* "+writeableMSWithPathPipeline+"/")
            #os.system("cp -r "+msWithPathPipeline+"/SORTED_TABLE "+writeableMSWithPathPipeline+"/")
            #os.system("ln -s "+msWithPathPipeline+"/* "+writeableMSWithPathPipeline+"/")



    # ------------
    if showingManualQA2:
       if not os.path.exists(manualqa2dir): 
           println(lineno(),"WARNING: Manual QA2dir %s doesn't exist - may cause issues" % manualqa2dir)
           
       if not os.path.exists(manualqa2dir_extra):
           if os.path.exists(amc_outdir): # if this is set put qa2 there
               manualqa2dir_extra=amc_outdir+"%s/manqa2/" % (asdmNoPath)
           else: # otherwise try to put it in ./amc
               manualqa2dir_extra='./amc/%s/manqa2/' % (asdmNoPath)
       if not os.path.exists(manualqa2dir_extra):       
           try:
               os.makedirs(manualqa2dir_extra)
           except:
               println(lineno(),"WARNING: can't create %s" % manualqa2dir_extra)
               manualqa2dir_extra='./amc/%s/manqa2/' % (asdmNoPath)
               if not os.path.exists(manualqa2dir_extra): os.makedirs(manualqa2dir_extra)
   
       if os.access(manualqa2dir_extra, os.W_OK)==False:
           println(lineno(),"WARNING: can't write to %s" % manualqa2dir_extra)
           manualqa2dir_extra='./amc/%s/manqa2/' % (asdmNoPath)
           os.makedirs(manualqa2dir_extra)
   
       if cleanup:
           println(lineno(),"WARNING: Removing files in %s" % manualqa2dir_extra)
           os.system('rm -rf %s/*' % (manualqa2dir_extra))

       writeableMSWithPathManual=msWithPathManual
       copyms=False
       if not os.access(msWithPathManual,os.W_OK): copyms=True
       if not os.access(msWithPathManual+"/SORTED_TABLE",os.W_OK): copyms=True
       if copyms:
           writeableMSWithPathManual=manualqa2dir_extra+"/"+ms

           bar=writeableMSWithPathManual
           if bar[-1]=="/": bar=bar[:-1]
           myCopyfile(msWithPathManual+"/",bar) # needs trailing slash on msWithPathPipeline, NOT on destination, because of how rsync works

           # we only need to copy SORTED_TABLE and any locks           
           #os.system("mkdir "+msWithPathManualWriteable)
           #os.system("cp -r "+msWithPathManual+"/*lock* "+msWithPathManualWriteable+"/")
           #os.system("cp -r "+msWithPathManual+"/SORTED_TABLE "+msWithPathManualWriteable+"/")
           #os.system("ln -s "+msWithPathManual+"/* "+msWithPathManualWriteable+"/")           

       writeableMSSplitWithPath=msSplitWithPath
       copyms=False
       if not os.access(msSplitWithPath,os.W_OK): copyms=True
       if not os.access(msSplitWithPath+"/SORTED_TABLE",os.W_OK): copyms=True
       if copyms:
           writeableMSSplitWithPath=manualqa2dir_extra+"/"+ms+".split"
           bar=writeableMSSplitWithPath
           if bar[-1]=="/": bar=bar[:-1]
           myCopyfile(msSplitWithPath+"/",bar) # needs trailing slash on msWithPathPipeline, NOT on destination, because of how rsync works



    # -------------
    if len(amc_outdir)<=0:
        amc_outdir="./amc/"
    if not os.path.exists(amc_outdir):
        os.makedirs(amc_outdir)


    #===========================================
    # Check for 0,0 position before proceeding
    es = au.stuffForScienceDataReduction()
    fields = es.getFieldsForFixPlanets(msWithPath)
    if (len(fields) > 0):
        if (casadef.casa_version >= '4.1.0'):
            mymsmd = createCasaTool(msmdtool)
            mymsmd.open(msWithPath)
            fieldnames = mymsmd.namesforfields(fields)
            if (len(fields) > 1):
                print "WARNING: %d fields (%s) have coordinates of (0,0)." % (len(fields),fieldnames)
            else:
                print "WARNING:%d field (%s) has coordinates of (0,0)." % (len(fields),fieldnames)
            print "ERROR: After running hif_importdata, you need to run fixplanets before running the full pipeline!"
            mymsmd.close()
        else:
            print "ERROR: %d fields have coordinates of (0,0).  You need to run fixplanets before running the pipeline!" % (len(fields))
        if (not ignoreZeroCoords):
            return
        

    #---------------------------------------------
    # open logfile
    terminalfile = amc_outdir+'terminal.' + asdmNoPath
    if (cleanup and os.path.exists(terminalfile)):
        os.system('rm -rf %s %s.cleanupFalse' % (terminalfile, terminalfile))
        
    terminalfileToOpen = terminalfile
    if (cleanup==False and os.path.exists(terminalfile)):
        terminalfileToOpen += '.cleanupFalse' # don't overwrite the most recent non-cleanup=False terminal file

    try:
        terminal = open(terminalfileToOpen,'w')
    except:        
        print "ERROR: Could not write %s" % terminalfileToOpen
        return
    println(lineno(),"version " + version() + ' ' + str(datetime.datetime.now()) + '\n')
    if (pl.isinteractive()):
        wasInteractive = True
#        pl.ioff()
    else:
        wasInteractive = False
    println(lineno(),"start time = %s" % (time.asctime()))
    println(lineno(),"PWD = %s" % (os.getcwd()))
    println(lineno(),"command line = %s" % (commandLine))
    println(lineno(),"pylab.isinteractive() = %s" % (pl.isinteractive()))
    print
    print 40*"-"






    #---------------------------------------------

    if (showingPipelineQA2):
        print "finding logfile:"
        print 
        if heuristicsversion()<2013.08:
            firstGACalStage, lastGACalStage = findFirstGACal(pipecaldir, ms)
            if (firstGACalStage == None and lastGACalStage == None):
                return
        pipelineCasaLogfile = findPipelineCasaLogfile(pipecaldir, asdmNoPath)
        println(lineno(),"pipelineCasaLogfile = %s"% pipelineCasaLogfile)
        print 
        print 40*"-"
        pipelineCasaVersion = findCasaVersionFromLogfile(pipelineCasaLogfile)
#        pipelineCasaVersion = findCasaVersion(pipecaldir)
    else:
        pipelineCasaVersion = ''

    # manual reduction information:
    manualQA2ScienceTarget = ''
    if (showingManualQA2):
        manualCasaVersion = findCasaVersion(manualcaldir)
        manualQA2ScienceTarget = findManualQA2ScienceTarget(manualqa2dir)  # parses the qa2/textfile.txt
        calScript=glob.glob(manualcaldir+"*scriptForCalibration.py")
        manualRefAnt="manual script not found"
        if len(calScript)==0:
            print "ERROR: Cannot find manual cal script - some output may be missing"
        else:
            inScript=commands.getoutput("grep refant "+calScript[0])
            if len(inScript)>0:
                manualRefAnt=inScript.split("\n")
                for i in range(len(manualRefAnt)):
                    #check if whole line is commented
                    if manualRefAnt[i].strip()[0] != '#':
                        x = manualRefAnt[i].split('=')[1]
                        #check if string specified with double-quotes
                        x = x.split('"')
                        if len(x) > 1:
                            mrf2 = x[1][0:2]
                            if mrf2 == 'DV' or \
                               mrf2 == "DA" or \
                               mrf2 == "PM" or \
                               mrf2 =="CM":
                                manualRefAnt = x[1]
                                break
                            else:
                                continue
                        else:
                            x = manualRefAnt[i].split('=')[1]
                            #check if string is specified with single-quotes
                            x = x.split("'")
                            if len(x) > 1:
                                mrf2 = x[1][0:2]
                                if mrf2 == 'DV' or \
                                   mrf2 == "DA" or \
                                   mrf2 == "PM" or \
                                   mrf2 =="CM":
                                    manualRefAnt = x[1]
                                    break
                                else:
                                    continue
                            else:
                                #try the next occurence of "refant"
                                continue
                #never set to str so we know we failed
                if type(manualRefAnt) == list:
                    println(lineno(), 'ERROR: could not find refant ' + \
                            'in this list of strings:\n', manualRefAnt)
                    return
            else:
                manualRefAnt=''
    else:
        manualCasaVersion = ''
        manualQA2ScienceTarget = ''
        match_man_refant=False








    ############################################################
    # Run Cycle 0 QA2 functions on pipeline calibration tables
    ############################################################


    ##########
    # Weather
    ##########
    if (showingPipelineQA2):
        preliminaryPlotsDestination = pipeqa2dir
    else:
        preliminaryPlotsDestination = manualqa2dir_extra
    weatherplot = preliminaryPlotsDestination + asdmNoPath + '.weather.png'
    if (os.path.exists(weatherplot) == False):
        if (os.path.exists(asdm)):
#            println(lineno(),"Running au.plotWeatherFromASDM('%s',plotfile='%s',titlefontsize=8)" % (asdm,weatherplot))
#            au.plotWeatherFromASDM(asdm,plotfile=weatherplot,titlefontsize=8)
            println(lineno(),"Running au.plotWeather('%s',figfile='%s',asdm='%s')" % (msWithPath,weatherplot,asdm))
            au.plotWeather(msWithPath, figfile=weatherplot, asdm=asdm)
            print 
        else:
            println(lineno(),"ERROR: Did not find asdm = %s"% asdm)
            terminal.close()
            return        
    elif (showSkippingMessages):
        println(lineno(),"%sSkipping au.plotWeather() because plot already exists" % (skipIndent))
    print 40*"-"


    ##########
    # PWV
    ##########
    pwvplot = preliminaryPlotsDestination + asdmNoPath + '.pwv.png'
    if (os.path.exists(pwvplot) == False):
        if (os.path.exists(asdm)):
            println(lineno(),"Running au.plotPWVFromASDM('%s',figfile='%s',showMedianOnly=True)" % (asdm,pwvplot))
            au.plotPWVFromASDM(asdm,figfile=pwvplot,showMedianOnly=True)
            print
        else:
            println(lineno(),"ERROR: plotPWVFromASDM() did not find asdm = %s"% asdm)
            terminal.close()
            return
    elif (showSkippingMessages):
        println(lineno(),"%sSkipping au.plotPWVFromASDM() because plot already exists" % (skipIndent))
    print 40*"-"


    #####################
    # Sideband gain ratio
    #####################    
    sbrplot = preliminaryPlotsDestination + asdmNoPath + '.sbr.png'
    sbr_suspect = ''
    if True:
        print "Sideband ratio:"
        print
        # can't skip because sbr_suspect accessed below
        # not(os.path.exists(sbrplot)):
        sbr_suspect = ''
        SBR_THRESHOLD_2SB=0.80
        SBR_THRESHOLD_DSB=0.25
        if (os.path.exists(sbrplot) == False):
            println(lineno(),"Running au.plotSBGain('%s',figfile='%s',SBR_THRESHOLD_2SB='%f',SBR_THRESHOLD_DSB='%f')" % (writeableMSWithPathPipeline,sbrplot,SBR_THRESHOLD_2SB,SBR_THRESHOLD_DSB))
            # return(figfile, suspect, suspectDict, valuesAreDefault)
            result = au.plotSBGain(writeableMSWithPathPipeline,figfile=sbrplot,SBR_THRESHOLD_2SB=SBR_THRESHOLD_2SB,
                                                 SBR_THRESHOLD_DSB=SBR_THRESHOLD_DSB)
            if (result != None):
                sbrplot, sbr_suspect, foo, bar = result
            else:
                sbrplot = ''
        else:
            println(lineno(),"Running au.plotSBGain('%s',figfile='%s', doplot=False,SBR_THRESHOLD_2SB='%f',SBR_THRESHOLD_DSB='%f')" % (writeableMSWithPathPipeline,sbrplot,SBR_THRESHOLD_2SB,SBR_THRESHOLD_DSB))
            result = au.plotSBGain(writeableMSWithPathPipeline, figfile=sbrplot, doplot=False,
                                               SBR_THRESHOLD_2SB=SBR_THRESHOLD_2SB,
                                               SBR_THRESHOLD_DSB=SBR_THRESHOLD_DSB)
            if (result != None):
                dummy, sbr_suspect, foo, bar = result
    
        println(lineno(),"sbrplot = %s" % (sbrplot))
        print 
        
    elif (showSkippingMessages):
        println(lineno(),"%sSkipping SBR plot because plot already exists" % (skipIndent))
    print 40*"-"


    #############################
    println(lineno(),"Running es.getIntentsAndSourceNames(%s)" % (writeableMSWithPathPipeline))
    intentsAndSourcesInfo = es.getIntentsAndSourceNames(writeableMSWithPathPipeline)
    target_source_name = intentsAndSourcesInfo['OBSERVE_TARGET']['name'][target_source]
    target_source_id = intentsAndSourcesInfo['OBSERVE_TARGET']['sourceid'][target_source]
    myfields = au.getFields(msWithPath)
    if (bandpass_cal == ''):
        bandpass_source_id = intentsAndSourcesInfo['CALIBRATE_BANDPASS']['sourceid'][0]
        bandpass_source_name = intentsAndSourcesInfo['CALIBRATE_BANDPASS']['name'][0]
        original_bandpass_source_id = bandpass_source_id
    else:
        bandpass_source_id = bandpass_cal
        bandpass_source_name = myfields[bandpass_source_id]
        original_bandpass_source_id = intentsAndSourcesInfo['CALIBRATE_BANDPASS']['sourceid'][0]
        original_bandpass_source_name = intentsAndSourcesInfo['CALIBRATE_BANDPASS']['name'][0]
        print "Using bandpass_cal = %d = %s" % (bandpass_source_id, bandpass_source_name)
    if ('CALIBRATE_AMPLI' in intentsAndSourcesInfo):
#       fluxcal_source_id = intentsAndSourcesInfo['CALIBRATE_AMPLI']['sourceid'][0]
        fluxcal_source_id = intentsAndSourcesInfo['CALIBRATE_AMPLI']['id'][0]
        fluxcal_source_name = intentsAndSourcesInfo['CALIBRATE_AMPLI']['name'][0]
    elif ('CALIBRATE_FLUX' in intentsAndSourcesInfo):
#        fluxcal_source_id = intentsAndSourcesInfo['CALIBRATE_FLUX']['sourceid'][0]
        fluxcal_source_id = intentsAndSourcesInfo['CALIBRATE_FLUX']['id'][0]
        fluxcal_source_name = intentsAndSourcesInfo['CALIBRATE_FLUX']['name'][0]
    else:
        fluxcal_source_id = -1
        fluxcal_source_name = ''
        println(lineno(),"WARNING: No flux calibrator in this dataset")
    if (fluxcal_source_id == ''):
        println(lineno(),"+++++++++++++++++  WARNING: No flux calibrator in this dataset!!")
        fluxcal_uvrange = ''
    else:
        if (fluxcal_source_name.lower() in SSO_LIST):
            fluxcal_uvrange = '0~50m'
        else:
            fluxcal_uvrange = ''
        println(lineno(),"+++++++++++++++++  flux cal source ID = %d = %s" % (fluxcal_source_id, fluxcal_source_name))
    print
         
   
    #######################
    # Get phase calibrator
    #######################
    if (phase_cal == ''):
        println(lineno(),"Calling es.getPhaseCal() which is going to have to run ValueMapping")
        phaseCalInfo = es.getPhaseCal(writeableMSWithPathPipeline)
        if (phaseCalInfo == None):
            print "ERROR: No field returned by es.getPhaseCal()."
            print "ERROR: You need to specify the field ID of the phase calibrator using the phase_cal option."
            return
        phase_cal = str(phaseCalInfo[target_source_name]['phaseCalId'])
    phase_cal_source_id = int(phase_cal)
    if (phase_cal_source_id > len(myfields)):
        print "ERROR: No such phase calibrator in ms."
        terminal.close()
        return
    phase_cal_source_name = myfields[phase_cal_source_id]

    if (bandpass_source_name.find(',') > 0):
        bandpass_source_name = bandpass_source_name.split(',')[0] + '*'
        println(lineno(),"Changing comma to * in bandpass calibrator name: %s" % (bandpass_source_name))
    if (showingManualQA2):
        bandpass_source_name_manual = bandpass_source_name
        if (bandpass_source_name_manual.find(';') > 0):
            # The manual reduction might still have a comma, even if pipeline ASDM was changed to ; manually
            # So, only pass the first name before the semicolon to the plotbandpass for manual reduction.
            bandpass_source_name_manual = bandpass_source_name_manual.split(';')[0] + '*'
            println(lineno(),"Changing semicolon to * in bandpass calibrator name for use in plotbandpass of the manual reduction: %s" % (bandpass_source_name_manual))
    print
    print 40*"-"
            
    ##################
    # Pointing results
    ##################
    pointingPlot = preliminaryPlotsDestination + asdmNoPath + '.pointing.scan*.png'
    pointingPlots = glob.glob(pointingPlot)
    thresholdArcsec = 10.0
    fractionOfScansBad = 0.60
    antwebpage = ''
    if (os.path.exists(asdm+'/CalPointing.xml')):
      if (pointingPlots == []):
        println(lineno(),"Running au.plotPointingResultsFromASDM('%s', figfile=True, interactive=False, figfiledir='%s',thresholdArcsec='%.1f',fractionOfScansBad='%.4f')" % (asdm,preliminaryPlotsDestination,thresholdArcsec,fractionOfScansBad))
        pointingPlots, badAntennas = au.plotPointingResultsFromASDM(asdm, figfile=True, interactive=False,
                                                                    figfiledir=preliminaryPlotsDestination, thresholdArcsec=thresholdArcsec,fractionOfScansBad=fractionOfScansBad)
      else:
        println(lineno(),"Running au.plotPointingResultsFromASDM('%s', figfile=False, interactive=False, thresholdArcsec='%.1f', fractionOfScansBad='%.4f', doplot=False)" % (asdm,thresholdArcsec,fractionOfScansBad))
        junk, badAntennas = au.plotPointingResultsFromASDM(asdm, figfile=False, interactive=False,thresholdArcsec=thresholdArcsec,fractionOfScansBad=fractionOfScansBad, doplot=False)
      criteria = '(criteria = >%.1f arcsec for >%.0f%% of scans)' % (thresholdArcsec,fractionOfScansBad*100)
      if (badAntennas == {}):
          antwebpage += 'No suspect antennas  %s<br>' % (criteria)
      elif (len(badAntennas) == 1):
          antwebpage += '1 suspect antenna  %s<br>' % (criteria)
      elif (len(badAntennas) > 1):
          antwebpage += '%d suspect antennas  %s<br>' % (len(badAntennas), criteria)
      if (len(badAntennas) > 0):
        for ba in range(len(badAntennas)):
            ant = badAntennas.keys()[ba]
            scans = badAntennas[ant].keys()
            antwebpage += ant 
            for sc in np.sort(scans):
                antwebpage += ' scan %d: %.2f arcsec' % (sc, badAntennas[ant][sc])
                if (sc != scans[-1]): antwebpage += ',  '
            antwebpage += '<br>'
    else:
        antwebpage += 'No pointing scans in this dataset!<br>'
    println(lineno(),"pointing summary = %s" % (antwebpage))
    print
    print 40*"-"
    
    ##################
    # Solar position 
    ##################
    sunplot = preliminaryPlotsDestination + asdmNoPath + '.ms.sun.png'
    if (os.path.exists(sunplot) == False):
        println(lineno(),"Running au.plotSunDuringTrack(ms='%s', plotfile=True, figfiledir='%s')" % (writeableMSWithPathPipeline,preliminaryPlotsDestination))
        sunplot = au.plotSunDuringTrack(ms=writeableMSWithPathPipeline, plotfile=True, figfiledir=preliminaryPlotsDestination)
    elif (showSkippingMessages):
        println(lineno(),"%sSkipping au.plotSunDuringTrack() because plot already exists" % (skipIndent))
    pointingPlots.append(sunplot)
    maxChannels = '9000'
    elevationplot = preliminaryPlotsDestination + asdmNoPath + '.elevation.png'
    runMessage = "plotms("+plotmsparms+",vis='%s', plotfile='%s', xaxis='time', yaxis='elevation', coloraxis='field',avgchannel='%s',avgtime='10s', avgbaseline=True)" % (writeableMSWithPathPipeline, elevationplot,maxChannels)
    if (os.path.exists(elevationplot) == False): runPlotms(runMessage)
    elif (showSkippingMessages):
        println(lineno(),"%sSkipping plotms of elevation vs. time because plot already exists at %s" % (skipIndent,elevationplot))
        println(lineno(),"%sI would have run: %s" % (skipIndent,runMessage))
    pointingPlots.append(elevationplot)
    print
    print 40*"-"
    


    ##########
    # Mosaic
    ##########
    mosaicplot = preliminaryPlotsDestination + 'mosaic_plot.png'
    if (os.path.exists(mosaicplot) == False):
        println(lineno(),"Running au.plotmosaic('%s', sourceid='%s', figfile='%s', sciencespws=True)" % (writeableMSWithPathPipeline,target_source_id,mosaicplot))
        au.plotmosaic(writeableMSWithPathPipeline, sourceid=target_source_id, figfile=mosaicplot, sciencespws=True)
    elif (showSkippingMessages):
        println(lineno(),"%sSkipping au.plotmosaic() because plot already exists" % (skipIndent))

    print
    print 40*"-"

    ##########
    # WVR
    ##########
    print "WVR analysis"
    println(lineno(),"Running es.getSpwInfo() to find science spws.")
    spwdict = es.getSpwInfo(writeableMSWithPathPipeline)

    # an alternative here would be: set(msmd.targetspws()).difference(set(msmd.chanavgspws())).difference(msmd.wvrspws())
    scienceSpwsIntegerList = spwdict.keys()
    narrowestSpwBW = 3e11
    nchanforspw = {}
    if (casadef.casa_version >= '4.1.0'):
        mymsmd = createCasaTool(msmdtool)
        mymsmd.open(writeableMSWithPathPipeline)
        for myspw in scienceSpwsIntegerList:
            mybw = (np.max(mymsmd.chanfreqs(myspw)) - np.min(mymsmd.chanfreqs(myspw)))*(1+mymsmd.nchan(myspw))/mymsmd.nchan(myspw)
            nchanforspw[myspw] = mymsmd.nchan(myspw)
            if (mybw < narrowestSpwBW):
                narrowestSpwBW = mybw
        mymsmd.close()

    spwdictkeys=spwdict.keys()
    spwdictkeys.sort()
    scienceSpws = str(spwdictkeys)[1:-1].replace(' ','')
    scienceSpwsManualIntegerList = range(len(spwdict.keys()))
    scienceSpwsManual = str(scienceSpwsManualIntegerList)[1:-1].replace(' ','')
    println(lineno(),"  Found science spws = %s = %s" % (scienceSpws, scienceSpwsManual))
    allpols = ['X','Y']
    tsysSpwInfo = es.getSpwInfo(writeableMSWithPathPipeline, intent='CALIBRATE_ATMOSPHERE')
    tsysSpwsIntegerList = tsysSpwInfo.keys()
    
    if (showingPipelineQA2):
        sevenMeterOnly = sevenMeterAntennasOnly(msWithPathPipeline)
    else:
        sevenMeterOnly = sevenMeterAntennasOnly(msSplitWithPath)
    pipelinePhaseTemporalPlotPhaseCalWithPath = ''


    # Pipeline WVR
    pipelineWvrPlot = ''
    pipelineWvrPdf = ''
    pipelineWvrText = ''
    manualWvrText = ''
    wvr_gaintable = ''
    amcRefAnt = ''
    if (showingPipelineQA2):
        pipeRefAnt = findRefAntFromCCL(ccl)

        # set refant first from manual if requested
        if match_man_refant:            
            amcRefAnt=manualRefAnt
            # set wvr_refant, which will trigger remaking of all plots
            wvr_refant=manualRefAnt

        # explicit refant parameter takes precedence
        if (refant != ''):
            amcRefAnt = refant

        # if neither, set from PL CCL:
        if amcRefAnt=="":
            amcRefAnt = pipeRefAnt

        # get only the first antenna:
        if (len(amcRefAnt)>4):
            amcRefAnt0 = amcRefAnt.split(',')[0]
        else:
            amcRefAnt0 = amcRefAnt

        # get only the first antenna:
        if (len(pipeRefAnt)>4):
            pipeRefAnt0 = pipeRefAnt.split(',')[0]
        else:
            pipeRefAnt0 = pipeRefAnt

        # explicit wvr_refant parameter overrides refant parameter
        if (wvr_refant != ''):
            amcRefAnt0 = wvr_refant

        if (sevenMeterOnly == False):
          pipelineWvrPlot = pipeqa2dir + 'wvr_plot.png'
          wvr_gaintable = findWvrGainTable(pipecaldir, ms)
          pipelineWvrPdf = pipeqa2dir + wvr_gaintable + '.pdf'
          wvrstat_intent2 = 'CALIBRATE_PHASE'
          pipelineWvrPdf2 = pipeqa2dir + wvr_gaintable + '.%s.pdf' % (wvrstat_intent2)
          newplot = False
          if (wvr_gaintable != ''):
              wvr_gaintable = pipecaldir + wvr_gaintable              
              if (os.path.exists(pipelineWvrPlot) == False):
#                  or wvrstat_spw != '' or wvrstat_pol != '' or wvr_refant != ''):
                  cmd="es.wvr_stat('%s', refAnt='%s', qa2_output_dir='%s', wvr_gaintable='%s', scanintent='%s', spw='%s', pol='%s', nsigma_coherence=%f, unwrap_phase=%s, autoscale_yaxis=%s)" % (writeableMSWithPathPipeline,amcRefAnt0,pipeqa2dir,wvr_gaintable,wvrstat_intent, wvrstat_spw, wvrstat_pol, nsigma_coherence, unwrap_phase, str(autoscale_yaxis))
                  println(lineno(),"\npipeline: Calling "+cmd)
                  try:
                      exec cmd
                  except:
                      msg="The gaincal failed in es.wvr_stat("+wvrstat_intent+"spw='%s')." % wvrstat_spw
                      println(lineno(), "\n ERROR: "+msg)
                      pipelineWvrText =msg+"<br>\n"+printWvrStatErrorMessage(writeableMSWithPathPipeline,lineno())
              elif (showSkippingMessages):
                  println(lineno(),"%sSkipping es.wvr_stat() since the result already exists in the qa2 subdirectory." % (skipIndent))

              if (os.path.exists(pipelineWvrPdf) == False):
                  cmd="wvr_stat_all('%s', refAnt='%s', qa2_output_dir='%s', wvr_gaintable='%s', scanintent='%s', spw=%s, pol=%s, nsigma_coherence=%f)" % (writeableMSWithPathPipeline,amcRefAnt0,pipeqa2dir,wvr_gaintable,wvrstat_intent,scienceSpwsIntegerList,allpols,nsigma_coherence)
                  println(lineno(),"pipeline: Calling "+cmd)                  
                  try:
                      wvr_stat_all(writeableMSWithPathPipeline, refAnt=amcRefAnt0, qa2_output_dir=pipeqa2dir, wvr_gaintable=wvr_gaintable, scanintent=wvrstat_intent, spw=scienceSpwsIntegerList, pol=allpols, nsigma_coherence=nsigma_coherence)
                  except:
                      pipelineWvrPdf = ''
                      msg="At least one spw's gaincal failed in wvr_stat_all(%s,spw='%s')." % (wvrstat_intent,scienceSpwsIntegerList)
                      println(lineno(), "\n ERROR: "+msg)
                      pipelineWvrText = msg+"<br>\n"+printWvrStatErrorMessage(writeableMSWithPathPipeline,lineno())
              elif (showSkippingMessages):
                  println(lineno(),"%sSkipping es.wvr_stat() since the result already exists in the qa2 subdirectory." % (skipIndent))

              if (os.path.exists(pipelineWvrPdf2) == False):
                  cmd="wvr_stat_all('%s', refAnt='%s', qa2_output_dir='%s', wvr_gaintable='%s', scanintent='%s', spw=%s, pol=%s, nsigma_coherence=%f)" % (writeableMSWithPathPipeline,amcRefAnt0,pipeqa2dir,wvr_gaintable,wvrstat_intent2,scienceSpwsIntegerList,allpols,nsigma_coherence)
                  println(lineno(),"pipeline: Calling "+cmd)
                  try:
                      wvr_stat_all(writeableMSWithPathPipeline, refAnt=amcRefAnt0, qa2_output_dir=pipeqa2dir, wvr_gaintable=wvr_gaintable, scanintent=wvrstat_intent2, spw=scienceSpwsIntegerList, pol=allpols, nsigma_coherence=nsigma_coherence)
                  except:
                      pipelineWvrPdf2 = ''                      
                      msg="At least one spw's gaincal failed in wvr_stat_all(%s,spw='%s')." % (wvrstat_intent2,scienceSpwsIntegerList)
                      println(lineno(), "\n ERROR: "+msg)
                      pipelineWvrText = msg+"<br>\n"+printWvrStatErrorMessage(writeableMSWithPathPipeline,lineno())
              elif (showSkippingMessages):
                  println(lineno(),"%sSkipping es.wvr_stat(%s) since the result already exists in the qa2 subdirectory." % (skipIndent,wvrstat_intent2))
    else:
        amcRefAnt = manualRefAnt
        amcRefAnt0 = manualRefAnt
        

    print
    print 40*"-"




    #######################
    # Pre-bandpass phase-up
    #######################
    nonRefAntIndex = -1
    # Avoid the reference antenna
    antennaNames = au.getAntennaNames(msWithPath)
    for a in antennaNames:
        if (a != amcRefAnt0)and(a != pipeRefAnt0):
            nonRefAnt = a
            println(lineno(),"WARNING: Picked non-reference antenna for bandpass plots: %s" % (nonRefAnt))
            break
    if (showingPipelineQA2):
        caltable = pipecaldir + findPreBandpassGainTable(pipecaldir, ms, debug=debug)
        caltableNoPath = caltable.split('/')[-1]
        pipelinePrebandpassCaltable = caltableNoPath
        plotdir = caltable +'.plots/'
        temppath=pipeqa2dir
        pipelinePreBandpassPdfName = plotdir + caltableNoPath + '.pdf'
        buildplots = True
        # look for plots, in pipeline run dir, and then in pipeqa2dir
        if (os.path.exists(plotdir)):
            if (len(os.listdir(plotdir)) > 0):
                buildplots = False
        else:
            println(lineno(),"-------WARNING: Did not find directory %s - will keep looking for plots" % (plotdir))
            if (os.path.exists(temppath +caltableNoPath+'.plots')):
                plotdir = temppath +caltableNoPath+'.plots/'
                if (len(os.listdir(plotdir)) > 0):
                    buildplots = False
            else:
                println(lineno(),"-------WARNING: Did not find directory %s" % (temppath +caltableNoPath+'.plots - will attempt buildplots'))

        if (buildplots):            
            println(lineno(),"1:Calling es.checkCalTable(calTableName='%s', msName='%s', interactive=False, subplot=411)" % (caltable,msWithPathPipeline))
            plotdir = callCheckCalTable(calTableName=caltable, msName=writeableMSWithPathPipeline, interactive=False, subplot=411,temppath=temppath)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping es.checkCalTable(pre_bandpass) since the plot directory already exists in %s." % (skipIndent,plotdir))
            
        # if the plots go to temp dir then pdf also
        # RI CHECK - this was done for BP but not here - bug?
        pipelinePreBandpassPdfName = plotdir + caltableNoPath + '.pdf'
 
        if (os.path.exists(plotdir)):
            pnglist = sorted(os.listdir(plotdir))
        else:
            pnglist = []

        if (os.path.exists(pipelinePreBandpassPdfName) == False or buildplots):
            if (os.path.exists(pipelinePreBandpassPdfName) == False):
                println(lineno(),"------WARNING: Did not find %s - will attempt to build" % (pipelinePreBandpassPdfName))
            if (buildplots):
                println(lineno(),"------buildplots = True")
            for p in range(len(pnglist)):
                pnglist[p] = plotdir + pnglist[p]

            bPFP = au.buildPdfFromPngs(pnglist=pnglist, pdfname=pipelinePreBandpassPdfName, quiet=auQuiet)
            #println(lineno(),"bPFP="+bPFP)
        if (os.path.exists(plotdir)):
            pnglist = sorted(os.listdir(plotdir))
        for scienceSpw in scienceSpwsIntegerList:
            nonRefAntIndex = indexContaining3Substrings(pnglist, nonRefAnt, 'phase','spw%d'%(scienceSpw))
            if (nonRefAntIndex >= 0):
                break
#        println(lineno(),"png file index of non-ref antenna (%s) = %d" % (nonRefAnt, nonRefAntIndex))
        if (nonRefAntIndex >= 0):
            pipelinePreBandpassPhasePlot = plotdir + pnglist[nonRefAntIndex]
        else: # a plot for the non-refAnt was not found, so show a blank plot to alert us
            println(lineno(),"ERROR ****** No pre-bandpass phase plot found for antenna %s **** " % (nonRefAnt))
            if debug: pdb.set_trace()
            pipelinePreBandpassPhasePlot = ''
        print
        print 40*"-"
    else:
        pipelinePreBandpassPdfName = ''
        pipelinePreBandpassPhasePlot = ''




    #######################
    # Bandpass
    #######################
    pipelineBandpassPdfName = ''
    pipelineBandpassPhasePlot = ''
    if (showingPipelineQA2):
      print "Bandpass:"
      print
      caltable = findBandpassGainTable(pipecaldir, ms)
      pipelineBandpassTableWithPath = caltable
      if (caltable != ''):
        caltableNoPath = caltable.split('/')[-1]
        buildplots = True
        plotdir = caltable+'.plots/'
        temppath = pipeqa2dir.rstrip("/")
        if (os.path.exists(plotdir) == True):
            if (len(os.listdir(plotdir)) > 0):
                buildplots = False
        elif (os.path.exists(temppath +caltableNoPath+'.plots')):
            plotdir = temppath +caltableNoPath+'.plots/'
            if (len(os.listdir(plotdir)) > 0):
                buildplots = False
        if (buildplots):
            println(lineno(),"2:Calling amc.callCheckCalTable(calTableName='%s', msName='%s', interactive=False, temppath='%s')" % (caltable,msWithPathPipeline,temppath))
            plotdir = callCheckCalTable(calTableName=caltable, msName=msWithPathPipeline, interactive=False,temppath=temppath)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping es.checkCalTable(bandpass) since the plot directory already exists in pipeline working directory." % (skipIndent))

        pipelineBandpassPdfName = plotdir + caltableNoPath+'.plots.pdf'
        if (os.path.exists(pipelineBandpassPdfName) == False or buildplots):
            
            if (os.access(os.path.dirname(pipelineBandpassPdfName), os.W_OK) == False):
                pipelineBandpassPdfName=pipeqa2dir+os.path.basename(pipelineBandpassPdfName)
            println(lineno(),"bandpass plotdir = "+plotdir)
            pnglist = sorted(os.listdir(plotdir))
            for p in range(len(pnglist)):
                pnglist[p] = plotdir + pnglist[p]
            if debug: auQuiet=False
            bPFP = au.buildPdfFromPngs(pnglist=pnglist, pdfname=pipelineBandpassPdfName,quiet=auQuiet)
            println(lineno(),"bPFP = "+bPFP)
            
        pnglist = sorted(os.listdir(plotdir))
        for scienceSpw in scienceSpwsIntegerList:
            nonRefAntIndex = indexContainingSubstrings(pnglist, nonRefAnt, '.spw%d.' % (scienceSpw))
            if (nonRefAntIndex >= 0):
                break
        # Assume the same reference antenna was used as for pre-bandpass, so pick the same non-refAnt. 
        if (nonRefAntIndex >= 0):
            pipelineBandpassPhasePlot = plotdir + pnglist[nonRefAntIndex]
        else: # a plot for the non-refAnt was not found, so pick a different one
            for a in antennaNames:
                if (a != amcRefAnt0 and a != nonRefAnt and a != pipeRefAnt0):
                    nonRefAnt = a
                    println(lineno(),"WARNING: Picked a different non-reference antenna for bandpass plots: %s" % (nonRefAnt))
                    break
            nonRefAntIndex = -1
            for scienceSpw in scienceSpwsIntegerList:
                nonRefAntIndex = indexContainingSubstrings(pnglist, nonRefAnt, '.spw%d.' % (scienceSpw))
                if (nonRefAntIndex >= 0):
                    break
            if (nonRefAntIndex >= 0):
                pipelineBandpassPhasePlot = plotdir + pnglist[nonRefAntIndex]
            else:
                pipelineBandpassPhasePlot = ''
      print
      print 40*"-"  



    #######################
    # Phase temporal
    #######################
    ampGainTypePipeline = ''
    pipelineAmpTemporalGainTableAllCalsNoPath = ''
    pipelineAmpTemporalGainTableAllCalsWithPath = ''
    pipelineAmpTemporalGainTablePhaseCalNoPath = ''
    pipelineAmpTemporalGainTablePhaseCalWithPath = ''
    pipelineAmpTemporalPlotAllCalsNoPath = ''
    pipelineAmpTemporalPlotAllCalsWithPath = ''
    pipelineAmpTemporalPlotPhaseCalsNoPath = ''
    pipelineAmpTemporalPlotPhaseCalsWithPath = ''
    pipelineAmpTemporalPdfAllCalsNoPath = ''
    pipelineAmpTemporalPdfAllCalsWithPath = ''

    pipelinePhaseTemporalGainTableAllCalsNoPath = ''
    pipelinePhaseTemporalGainTableAllCalsWithPath = ''
    pipelinePhaseTemporalGainTablePhaseCalNoPath = ''
    pipelinePhaseTemporalGainTablePhaseCalWithPath = ''
    pipelinePhaseTemporalPlotAllCalsNoPath = ''
    pipelinePhaseTemporalPlotAllCalsWithPath = ''
    pipelinePhaseTemporalPlotPhaseCalsWithPath = ''
    pipelinePhaseTemporalPlotPhaseCalsNoPath = ''
    pipelinePhaseTemporalPdfAllCalsNoPath = ''
    pipelinePhaseTemporalPdfAllCalsWithPath = ''
    if (showingPipelineQA2):
      print "Phase Temporal (Pipeline):"
      print
      if heuristicsversion()>2013.08: lastGACalStage=1000
      pipelinePhaseTemporalGainTableAllCalsNoPath = findPhaseTemporalGainTableAllCals(pipecaldir, ms, lastGACalStage)
      pipelinePhaseTemporalGainTableAllCalsWithPath = pipecaldir + pipelinePhaseTemporalGainTableAllCalsNoPath
      pipelinePhaseTemporalGainTablePhaseCalNoPath = findPhaseTemporalGainTablePhaseCal(pipecaldir, ms, lastGACalStage)

      if (pipelinePhaseTemporalGainTableAllCalsNoPath != '' and
          pipelinePhaseTemporalGainTablePhaseCalNoPath != ''):
        pipelinePhaseTemporalGainTablePhaseCalWithPath = pipecaldir + pipelinePhaseTemporalGainTablePhaseCalNoPath
        pipelinePhaseTemporalPlotPhaseCalNoPath = 'ant_phase_temporal.png'
        pipelinePhaseTemporalPlotPhaseCalWithPath = pipeqa2dir + pipelinePhaseTemporalPlotPhaseCalNoPath
        if (os.path.exists(pipelinePhaseTemporalPlotPhaseCalWithPath) == False):
            println(lineno(),"Calling es.ant_phase_temporal(ms2='%s', caltable='%s', phase_cal='%s', qa2_output_dir='%s')" % (writeableMSWithPathPipeline,pipelinePhaseTemporalGainTablePhaseCalWithPath,phase_cal,pipeqa2dir))
            es.ant_phase_temporal(ms2=writeableMSWithPathPipeline, caltable=pipelinePhaseTemporalGainTablePhaseCalWithPath, phase_cal=phase_cal, qa2_output_dir=pipeqa2dir)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping es.ant_phase_temporal() since the result already exists in the qa2 subdirectory." % (skipIndent))

        temppath=pipeqa2dir
        # Build checkCalTable(phase) pngs for pipeline
        plotdir = pipelinePhaseTemporalGainTableAllCalsWithPath + '.plots/'
        if (os.path.exists(plotdir) == False):
            println(lineno(),"-------WARNING: Did not find directory %s - will attempt to build plots elsewhere" % (plotdir))
            if (os.path.exists(temppath + os.path.basename(plotdir.rstrip('/')))):
                plotdir = temppath + os.path.basename(plotdir.rstrip('/')) + '/'
        if (cleanup):
            os.system('rm -rf %s/*' % (plotdir))
        if (os.path.exists(plotdir) == False or glob.glob(plotdir+'*.png') == []):
            println(lineno(),"-------WARNING: Did not find directory %s or any PNGs there - will attempt to build" % (plotdir))
            println(lineno(),"Running phase es.checkCalTable('%s', msName='%s', interactive=False)" % (pipelinePhaseTemporalGainTableAllCalsWithPath, pipecaldir+ms))
            plotdir = callCheckCalTable(pipelinePhaseTemporalGainTableAllCalsWithPath, msName=writeableMSWithPathPipeline, interactive=False,temppath=temppath)
            # This should produce *phase.XXNN.png files
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping es.checkCalTable() on the pipeline phase vs. time solutions since the PNGs already exist in %s" % (skipIndent, plotdir))

        # Assemble checkCalTable(phase) pngs for pipeline
        # RI TODO write permission for this pdf?
        pipelinePhaseTemporalPdfAllCalsNoPath = '%s.phase.pdf' % (pipelinePhaseTemporalGainTableAllCalsNoPath) 
        pipelinePhaseTemporalPdfAllCalsWithPath = '%s.phase.pdf' % (pipelinePhaseTemporalGainTableAllCalsWithPath) 
        if (os.path.exists(pipelinePhaseTemporalPdfAllCalsWithPath) == False):
            pnglist = buildFileListWildcard(plotdir+'*phase*png')
            println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (pnglist, pipelinePhaseTemporalPdfAllCalsWithPath))
            bPFP = au.buildPdfFromPngs(pnglist=pnglist, pdfname=pipelinePhaseTemporalPdfAllCalsWithPath,quiet=auQuiet)
            println(lineno(),"bPFP = "+bPFP)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping the building of a PDF from the pipeline phase vs. time solution PNGs." % (skipIndent))

        pipelinePhaseTemporalPlotAllCalsNoPath = pipelinePhaseTemporalGainTableAllCalsNoPath+'.phase.png'  # all antennas overlaid
        pipelinePhaseTemporalPlotAllCalsWithPath = pipelinePhaseTemporalGainTableAllCalsWithPath+'.phase.png'  # all antennas overlaid
        if (cleanup):
            os.system('rm -rf %s' % (pipelinePhaseTemporalPlotAllCalsWithPath))
        if (os.path.exists(pipelinePhaseTemporalPlotAllCalsWithPath) == False):
            if (os.access(pipelinePhaseTemporalPlotAllCalsWithPath,os.W_OK) == False):
                pipelinePhaseTemporalPlotAllCalsWithPath=pipeqa2dir + pipelinePhaseTemporalGainTableAllCalsNoPath+'.phase.png'
            cmd="checkCalTableVariant(calTableName='%s', msName='%s', gaintype='p', plotfile='%s', temppath=pipeqa2dir)" % (pipelinePhaseTemporalGainTableAllCalsWithPath,writeableMSWithPathPipeline,pipelinePhaseTemporalPlotAllCalsWithPath)
            println(lineno(),"Running "+cmd)
            exec cmd
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping checkCalTableVariant(phase) on the pipeline reduction because the PNG already exists: %s" % (skipIndent,pipelinePhaseTemporalPlotAllCalsWithPath))
      else:
          println(lineno(),"     pipelinePhaseTemporalGainTableAllCalsNoPath = %s" % (pipelinePhaseTemporalGainTableAllCalsNoPath))
          println(lineno(),"     pipelinePhaseTemporalGainTablePhaseCalNoPath = %s" % (pipelinePhaseTemporalGainTablePhaseCalNoPath))
          
      # Build checkCalTable(amp) pngs for pipeline
      pipelineAmpTemporalGainTableAllCalsNoPath = findPipelineAmpTemporalGainTableAllCals(pipecaldir, ms, lastGACalStage)  # no directory is preprended here
      pipelineAmpTemporalGainTableAllCalsWithPath = pipecaldir + pipelineAmpTemporalGainTableAllCalsNoPath
      pipelineAmpTemporalGainTablePhaseCalNoPath = findPipelineAmpTemporalGainTablePhaseCal(pipecaldir, ms, lastGACalStage)  # no directory is preprended here
      pipelineAmpTemporalPlotPhaseCalWithPath = ''
      if (pipelineAmpTemporalGainTableAllCalsNoPath != '' and
          pipelineAmpTemporalGainTablePhaseCalNoPath != ''):
        pipelineAmpTemporalGainTablePhaseCalWithPath = pipecaldir + pipelineAmpTemporalGainTablePhaseCalNoPath
        pipelineAmpTemporalPlotPhaseCalNoPath = 'ant_amp_temporal.png'
        pipelineAmpTemporalPlotPhaseCalWithPath = pipeqa2dir + pipelineAmpTemporalPlotPhaseCalNoPath
        ampGainTypePipeline = determineSolutionType(caltable)
        if (os.path.exists(pipelineAmpTemporalPlotPhaseCalWithPath) == False):
            println(lineno(),"Calling es.ant_amp_temporal(ms2='%s', caltable='%s', phase_cal='%s', qa2_output_dir='%s')" % (msWithPath,pipelineAmpTemporalGainTablePhaseCalWithPath,phase_cal,pipeqa2dir))
            es.ant_amp_temporal(ms2=msWithPath, caltable=pipelineAmpTemporalGainTablePhaseCalWithPath, phase_cal=phase_cal, qa2_output_dir=pipeqa2dir)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping es.ant_amp_temporal() since the result already exists in the qa2 subdirectory." % (skipIndent))

        pipelineAmpTemporalPlotAllCalsNoPath = pipelineAmpTemporalGainTableAllCalsNoPath + '.amp.png' # all antennas overlaid
        pipelineAmpTemporalPlotAllCalsWithPath = pipelineAmpTemporalGainTableAllCalsWithPath + '.amp.png' # all antennas overlaid
        if (cleanup):
            os.system('rm -rf %s' % (pipelineAmpTemporalPlotAllCalsWithPath))
        if (os.path.exists(pipelineAmpTemporalPlotAllCalsWithPath) == False):
            if (os.access(pipelineAmpTemporalPlotAllCalsWithPath,os.W_OK) == False):
                pipelineAmpTemporalPlotAllCalsWithPath=pipeqa2dir+pipelineAmpTemporalPlotAllCalsNoPath
            println(lineno(),"Running checkCalTableVariant(calTableName='%s', msName='%s', gaintype='a', plotfile='%s')" % (pipelineAmpTemporalGainTableAllCalsWithPath,writeableMSWithPathPipeline,pipelineAmpTemporalPlotAllCalsWithPath))
            checkCalTableVariant(calTableName=pipelineAmpTemporalGainTableAllCalsWithPath,
                                 msName=writeableMSWithPathPipeline, gaintype='a', plotfile=pipelineAmpTemporalPlotAllCalsWithPath,temppath=temppath)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping checkCalTableVariant(amp) because the PNG already exists: %s" % (skipIndent,pipelineAmpTemporalPlotAllCalsWithPath))

        plotdir = pipelineAmpTemporalGainTableAllCalsWithPath + '.plots/'
        if (os.path.exists(plotdir) == False):
            println(lineno(),"-------WARNING Did not find directory %s" % (plotdir))
#            if (os.path.exists('tmp/'+os.path.basename(plotdir.rstrip('/')))):
#                plotdir = 'tmp/' + os.path.basename(plotdir.rstrip('/'))  + '/'
            if (os.path.exists(pipeqa2dir + os.path.basename(plotdir.rstrip('/')))):
                plotdir = pipeqa2dir + os.path.basename(plotdir.rstrip('/'))  + '/'

        if (os.path.exists(plotdir) == False or glob.glob(plotdir + '*.png') == []):
            println(lineno(),"-------WARNING Did not find directory %s or any PNGs there" % (plotdir))
            println(lineno(),"Running phase es.checkCalTable('%s', msName='%s', interactive=False)" % (pipelineAmpTemporalGainTableAllCalsWithPath, pipecaldir+ms))
            plotdir = callCheckCalTable(pipelineAmpTemporalGainTableAllCalsWithPath, msName=msWithPath, interactive=False,temppath=pipeqa2dir)
            # This should produce *amp.XXNN.png files, and may also produce *phase.XXNN.png files but we will ignore those
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping es.checkCalTable() on the pipeline amp vs. time solutions since the PNGs already exist in %s" % (skipIndent, plotdir))

        # Assemble checkCalTable(amp) pngs for pipeline
        # RI TODO check write permuission for these pdfs
        pipelineAmpTemporalPdfAllCalsNoPath = '%s.amp.pdf' % (pipelineAmpTemporalGainTableAllCalsNoPath)
        pipelineAmpTemporalPdfAllCalsWithPath = '%s.amp.pdf' % (pipelineAmpTemporalGainTableAllCalsWithPath)
        if (os.path.exists(pipelineAmpTemporalPdfAllCalsWithPath) == False):
            pnglist = buildFileListWildcard(plotdir + '*amp*png')
            if (len(pnglist) == 0):
                println(lineno(),"---------WARNING: Did not find any amplitude .png's in %s - can't buuld amp v time pdf" % (plotdir))
            else:
                println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (pnglist, pipelineAmpTemporalPdfAllCalsWithPath))
                bPFP = au.buildPdfFromPngs(pnglist=pnglist, pdfname=pipelineAmpTemporalPdfAllCalsWithPath,quiet=auQuiet) 
                println(lineno(),bPFP)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping the building of a PDF from the pipeline amplitude vs. time solution PNGs." % (skipIndent))


    print
    print 40*"-"


    # assemble phase_int manual pngs into PDF
    phase_int_caltable = ms + '.split.phase_int'
    phase_inf_caltable = ms + '.split.phase_inf'
    if (os.path.exists(manualcaldir+phase_inf_caltable) == False):
        phase_inf_caltable = ms + '.split.cal_phase'  # low S/N case
    manualPhaseTemporalGainTablePdf = ''
    manualPhaseTemporalPlotAllCals = ''
    temppath=manualqa2dir_extra
    if (showingManualQA2):
        print "Phase Temporal (Manual processing)"
        print
        if (os.path.exists(manualcaldir+phase_int_caltable) == False):
            print "WARNING: Did not find %s, looking for .split.cal_phase instead" % (manualcaldir+phase_int_caltable)  # phase_short_int
            phase_int_caltable = ms + '.split.cal_phase' 
        manualPhaseTemporalGainTablePdf = '%s/%s.pdf'% (manualqa2dir_extra, phase_int_caltable)
        if (os.path.exists(manualPhaseTemporalGainTablePdf) == False):
            pnglist = buildPngList(manualcaldir+phase_int_caltable+'.plots')
            println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s/%s.pdf')\n\n" % (pnglist, manualqa2dir_extra, phase_int_caltable))
            bPFP = au.buildPdfFromPngs(pnglist=pnglist, pdfname=manualPhaseTemporalGainTablePdf,quiet=auQuiet)
            println(lineno(),bPFP)
        else:
            println(lineno(),"%sSkipping the building of a PDF from the manual phase vs. time solution PNGs." % (skipIndent))

        manualPhaseTemporalPlotAllCals = manualqa2dir_extra + phase_int_caltable + '.phase.png' # all antennas overlaid
        if (os.path.exists(manualPhaseTemporalPlotAllCals) == False):
            println(lineno(),"Running checkCalTableVariant(calTableName='%s', msName='%s', gaintype='p', plotfile='%s')" % (manualcaldir+phase_int_caltable,writeableMSSplitWithPath,manualPhaseTemporalPlotAllCals))
            checkCalTableVariant(calTableName=manualcaldir+phase_int_caltable, msName=writeableMSSplitWithPath,
                                 gaintype='p', plotfile=manualPhaseTemporalPlotAllCals,temppath=temppath)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping checkCalTableVariant(phase) on the manual reduction because the PNG already exists: %s" % (skipIndent,manualPhaseTemporalPlotAllCals))

    print
    print 40*"-"
        

    #######################################
    # Amplitude temporal manual processing
    #######################################
    # We don't need to run es.checkCalTable() for manual reduction because it was run in the Phase temporal section above
    if (showingManualQA2):
        print "Amplitude Gain temporal, Manual processing"
        print
        if (ampGainTableManual == ''):
            ampGainTableManual = msSplitWithPath + '.flux_inf'  # it may not always be this
            if (os.path.exists(ampGainTableManual) == False):
                ampGainTableManual = msSplitWithPath + '.cal_amp'  # low S/N case
                if (os.path.exists(ampGainTableManual) == False):
                    print "ERROR: Could not find amplitude gain table for manual reduction"
                    terminal.close()
                    return
        ampGainTypeManual = determineSolutionType(ampGainTableManual)
    else:
        ampGainTypeManual = ''
        ampGainTableManual = ''

    if (showingPipelineQA2):
        intentsAndSourcesInfo = es.getIntentsAndSourceNames(writeableMSWithPathPipeline)
    else:
        intentsAndSourcesInfo = es.getIntentsAndSourceNames(writeableMSWithPathManual)
    
    # Assemble the .flux_inf manually-generated pngs into PDF
    # If they are not present then build them (which is necessary for tracks
    # with unresolved flux calibrators because the manual script does not build them.)
    manualAmpTemporalPlot = ''
    manualAmpTemporalTable = ''
    manualAmpTemporalPdf = ''
    temppath=manualqa2dir_extra
    if (showingManualQA2):
        flux_inf_caltable = ms + '.split.flux_inf'
        if (os.path.exists(manualcaldir+flux_inf_caltable) == False):
            flux_inf_caltable = ms + '.split.cal_amp'  # low S/N case
        manualAmpTemporalPdf = '%s/%s.pdf' % (manualqa2dir_extra, flux_inf_caltable)
        if (os.path.exists(manualAmpTemporalPdf) == False):
            plotdir = manualcaldir+flux_inf_caltable+'.plots'
            if (os.path.exists(plotdir) == False):
                println(lineno(),"-------WARNING: Did not find directory %s - will look elsewhere for manual amp v time plots" % (plotdir))
                if (os.path.exists(temppath+os.path.basename(plotdir.rstrip('/')))):
                    plotdir = temppath+os.path.basename(plotdir.rstrip('/')) + '/'
            if (cleanup):
                os.system('rm -rf %s' % (plotdir))
            pnglist = buildPngList(plotdir)
            if (pnglist == []):
                # We need to run checkCalTable to build them!
                # note, callCheckCalTable will do this in a temp dir if 
                # we can't write to manualcaldir, and return pnglist accoringly
                println(lineno(),"Running flux_inf es.checkCalTable('%s', msName='%s', interactive=False)" % (manualcaldir+flux_inf_caltable, manualcaldir+ms))
                plotdir = callCheckCalTable(manualcaldir+flux_inf_caltable, msName=writeableMSWithPathManual, interactive=False, temppath=manualqa2dir_extra)
                pnglist = buildPngList(plotdir)
                
            println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s/%s.pdf')\n\n" % (pnglist, manualqa2dir_extra, flux_inf_caltable))
            bPFP = au.buildPdfFromPngs(pnglist=pnglist, pdfname=manualAmpTemporalPdf,quiet=auQuiet)
            println(lineno(),bPFP)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping the building of a PDF from the manual amplitude vs. time solution PNGs." % (skipIndent))
        manualAmpTemporalTable = flux_inf_caltable
        manualAmpTemporalPlot = manualqa2dir_extra + manualAmpTemporalTable + '.amp.png'  # all antennas overlaid
        if (os.path.exists(manualAmpTemporalPlot) == False):
            println(lineno(),"Running checkCalTableVariant(calTableName='%s', msName='%s', gaintype='a', plotfile='%s')" % (manualcaldir+manualAmpTemporalTable, writeableMSSplitWithPath,manualAmpTemporalPlot))
            checkCalTableVariant(calTableName=manualcaldir+manualAmpTemporalTable, msName=writeableMSSplitWithPath, gaintype='a',
                                 plotfile = manualAmpTemporalPlot,temppath=temppath)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping checkCalTableVariant(amp) on the manual reduction because the PNG already exists: %s" % (skipIndent,manualAmpTemporalPlot))
        print
        print 40*"-"

    ##############################################
    # Tsys overlay time (not showing flagged data)
    ##############################################
    if (showingPipelineQA2):
      print "Tsys overlay time (pipeline):"
      print
      caltable = pipecaldir + findTsysTable(pipecaldir, ms)
      caltableNoPath = caltable.split('/')[-1]
      buildplots = True
      suffix = '.plots.overlayTime/'
      if (cleanup):
          os.system('rm -rf %s' % (caltable+suffix))
      pipelineTsysPdf = caltable + suffix + caltableNoPath + '.pdf'

      if (os.path.exists(caltable+suffix) == True):
          if (len(os.listdir(caltable+suffix)) > 0):
              if (os.path.exists(pipelineTsysPdf)):
                  # 20131119 todd changed the suffix from t0 to t00 - hardcode
                  # that here as well.
                  if (os.path.exists(caltable+suffix+caltableNoPath+'.spw%02d.t00.png'%(tsysSpwsIntegerList[0]))):
                      buildplots = False
                  else:
                      println(lineno(),"WARNING: Did not find: %s - will attempt to build tsys v time plots" % (caltable+suffix+caltableNoPath+'.spw%02d.t0.png'%(tsysSpwsIntegerList[0])))
              else:
                  println(lineno(),"WARNING: Did not find: %s - will attempt to build" % (pipelineTsysPdf))

      if (buildplots):
          println(lineno(),"Building Tsys overlay time plots (showflagged=False)")
          tsysNumChans = []
          for i in tsysSpwInfo:
              tsysNumChans.append(tsysSpwInfo[i]['numChans'])
          tsysNumChans = sorted(dict.fromkeys(tsysNumChans).keys())[0]
          chanEdge = 0.0390625
          startChan = int(tsysNumChans * chanEdge)
          endChan = int(tsysNumChans * (1-chanEdge))
          chanrange = str(startChan)+'~'+str(endChan)
          pipelineTsysPlot = caltable + suffix + caltableNoPath
          # fails if the plotdir doesn't exist yet in the pipecaldir
#          if (os.path.exists(caltable+suffix) == True):
#              if (os.access(caltable+suffix, os.W_OK)==False):
#                  pipelineTsysPlot = pipeqa2dir + caltableNoPath + suffix + caltableNoPath
#                  pipelineTsysPdf = pipeqa2dir + caltableNoPath + suffix + caltableNoPath + '.pdf'
          if (os.access(caltable, os.W_OK)==False):
              pipelineTsysPlot = pipeqa2dir + caltableNoPath + suffix + caltableNoPath
              pipelineTsysPdf = pipeqa2dir + caltableNoPath + suffix + caltableNoPath + '.pdf'
                  
          if ('task_plotbandpass' in sys.modules and force_au_plotbandpass==False and casadef.casa_version >= pbp_version):
              println(lineno(),"Calling plotbandpass(caltable='%s', vis='%s', overlay='time', xaxis='freq', yaxis='amp', buildpdf=True, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=True, figfile='%s')" % (caltable,writeableMSWithPathPipeline,chanrange,pipelineTsysPlot))
              plotbandpass(caltable=caltable, vis=writeableMSWithPathPipeline, overlay='time', xaxis='freq',
                           yaxis='amp',buildpdf=True,interactive=False,showatm=True,pwv='auto',chanrange=chanrange,
                           showfdm=True, figfile=pipelineTsysPlot, solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)
              println(lineno(),"Calling plotbandpass(caltable='%s', vis='%s', overlay='time,antenna', xaxis='freq', yaxis='amp', buildpdf=False, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=True, figfile='%s')" % (caltable,writeableMSWithPathPipeline,chanrange,pipelineTsysPlot))
              plotbandpass(caltable=caltable, vis=writeableMSWithPathPipeline, overlay='time,antenna', xaxis='freq',
                        yaxis='amp',buildpdf=False,interactive=False,showatm=True,pwv='auto',chanrange=chanrange,
                        showfdm=True, figfile=pipelineTsysPlot, solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)
          else:
              println(lineno(),"Calling au.plotbandpass(caltable='%s', ms='%s', overlay='time', xaxis='freq', yaxis='amp', buildpdf=True, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=True, figfile='%s')" % (caltable,writeableMSWithPathPipeline,chanrange,pipelineTsysPlot))
              vm = au.plotbandpass(caltable=caltable, ms=writeableMSWithPathPipeline, overlay='time', xaxis='freq',
                               yaxis='amp',buildpdf=True,interactive=False,showatm=True,pwv='auto',chanrange=chanrange,
                               showfdm=True, figfile=pipelineTsysPlot, solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)
              println(lineno(),"Calling au.plotbandpass(caltable='%s', ms='%s', overlay='time,antenna', xaxis='freq', yaxis='amp', buildpdf=False, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=True, figfile='%s')" % (caltable,writeableMSWithPathPipeline,chanrange,pipelineTsysPlot))
              au.plotbandpass(caltable=caltable, ms=writeableMSWithPathPipeline, overlay='time,antenna', xaxis='freq',
                        yaxis='amp',buildpdf=False,interactive=False,showatm=True,pwv='auto',chanrange=chanrange,
                        showfdm=True, figfile=pipelineTsysPlot, vm='', solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)

      elif (showSkippingMessages):
          println(lineno(),"%sSkipping plotbandpass(Tsys, showflagged=False) since the results already exist in the pipeline working directory." % (skipIndent))

      if os.path.exists(caltable+suffix):
          pipelineTsysPlot = findOverlayTimeAntennaPlot(caltable+suffix, caltableNoPath)
      else:
          pipelineTsysPlot = findOverlayTimeAntennaPlot(pipeqa2dir + caltableNoPath + suffix, caltableNoPath)
      print


    ##############################################
    # Tsys overlay time (showing flagged data)
    ##############################################
    if (showingPipelineQA2):
      caltable = pipecaldir + findTsysTable(pipecaldir, ms)
      caltableNoPath = caltable.split('/')[-1]
      buildplots = True
      suffix = '.plots.overlayTimeIncludingFlaggedSolutions/'

      if (cleanup):
          os.system('rm -rf %s' % (caltable+suffix))

      pipelineTsysPlotIncludingFlaggedSolutions = caltable + suffix + caltableNoPath + '.incflagged'
      pipelineTsysPdfIncludingFlaggedSolutions = caltable + suffix + caltableNoPath + '.incflagged.pdf'

      if (os.path.exists(caltable+suffix) == True):
          if (len(os.listdir(caltable+suffix)) > 0):
              if (os.path.exists(pipelineTsysPdfIncludingFlaggedSolutions)):
                  if (os.path.exists(pipelineTsysPlotIncludingFlaggedSolutions + '.spw%02d.t00.png'%(tsysSpwsIntegerList[0]))):
                      buildplots = False
                  else:
                      println(lineno(),"WARNING: Did not find: %s - will look elsewhere" % (pipelineTsysPlotIncludingFlaggedSolutions + '.spw%02d.t00.png'%(tsysSpwsIntegerList[0])))
              else:
                  println(lineno(),"WARNING: Did not find: %s - will attempt to build" % (pipelineTsysPdfIncludingFlaggedSolutions))

      if (buildplots):
          println(lineno(),"Building Tsys overlay time plots (showflagged=False)")
          tsysNumChans = []
          for i in tsysSpwInfo:
              tsysNumChans.append(tsysSpwInfo[i]['numChans'])
          tsysNumChans = sorted(dict.fromkeys(tsysNumChans).keys())[0]
          chanEdge = 0.0390625
          startChan = int(tsysNumChans * chanEdge)
          endChan = int(tsysNumChans * (1-chanEdge))
          chanrange = str(startChan)+'~'+str(endChan)
#          if (os.path.exists(caltable+suffix)):
          if (os.access(caltable, os.W_OK)==False):
              pipelineTsysPlotIncludingFlaggedSolutions = pipeqa2dir + caltableNoPath + suffix + caltableNoPath + '.incflagged'
              pipelineTsysPdfIncludingFlaggedSolutions = pipeqa2dir + caltableNoPath + suffix + caltableNoPath + '.incflagged.pdf'
              
          if ('task_plotbandpass' in sys.modules and force_au_plotbandpass==False and casadef.casa_version >= pbp_version):
              println(lineno(),"Calling plotbandpass(caltable='%s', vis='%s', overlay='time', xaxis='freq', yaxis='amp', buildpdf=True, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=True, figfile='%s', showflagged=True)" % (caltable,writeableMSWithPathPipeline,chanrange,pipelineTsysPlotIncludingFlaggedSolutions))
              plotbandpass(caltable=caltable, vis=writeableMSWithPathPipeline, overlay='time', xaxis='freq',
                           yaxis='amp',buildpdf=True,interactive=False,showatm=True,pwv='auto',chanrange=chanrange,
                           showfdm=True, figfile=pipelineTsysPlotIncludingFlaggedSolutions, showflagged=True,
                           solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)

              println(lineno(),"Calling plotbandpass(caltable='%s', vis='%s', overlay='time,antenna', xaxis='freq', yaxis='amp', buildpdf=False, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=True, figfile='%s', showflagged=True)" % (caltable,writeableMSWithPathPipeline,chanrange,pipelineTsysPlotIncludingFlaggedSolutions))
              plotbandpass(caltable=caltable, vis=writeableMSWithPathPipeline, overlay='time,antenna', xaxis='freq',
                        yaxis='amp',buildpdf=False,interactive=False,showatm=True,pwv='auto',chanrange=chanrange,
                        showfdm=True, figfile=pipelineTsysPlotIncludingFlaggedSolutions, showflagged=True,
                           solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)
          else:
              println(lineno(),"Calling au.plotbandpass(caltable='%s', ms='%s', overlay='time', xaxis='freq', yaxis='amp', buildpdf=True, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=True, figfile='%s', showflagged=True)" % (caltable,writeableMSWithPathPipeline,chanrange,pipelineTsysPlotIncludingFlaggedSolutions))
              vm = au.plotbandpass(caltable=caltable, ms=writeableMSWithPathPipeline, overlay='time', xaxis='freq',
                                   yaxis='amp',buildpdf=True,interactive=False,showatm=True,pwv='auto',chanrange=chanrange,
                                   showfdm=True, figfile=pipelineTsysPlotIncludingFlaggedSolutions, showflagged=True,
                                   solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)

              println(lineno(),"Calling au.plotbandpass(caltable='%s', ms='%s', overlay='time,antenna', xaxis='freq', yaxis='amp', buildpdf=False, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=True, figfile='%s', showflagged=True)" % (caltable,writeableMSWithPathPipeline,chanrange,pipelineTsysPlotIncludingFlaggedSolutions))
              au.plotbandpass(caltable=caltable, ms=writeableMSWithPathPipeline, overlay='time,antenna', xaxis='freq',
                              yaxis='amp',buildpdf=False,interactive=False,showatm=True,pwv='auto',chanrange=chanrange,
                              showfdm=True, figfile=pipelineTsysPlotIncludingFlaggedSolutions, vm='', showflagged=True,
                              solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)
      elif (showSkippingMessages):
          println(lineno(),"%sSkipping plotbandpass(Tsys,showflagged=True) since the result already exists in the pipeline working directory." % (skipIndent))

      if os.path.exists(caltable+suffix):
          pipelineTsysPlotIncludingFlaggedSolutions = findOverlayTimeAntennaPlot(caltable+suffix, caltableNoPath+'.incflagged')
      else:
          pipelineTsysPlotIncludingFlaggedSolutions = findOverlayTimeAntennaPlot(pipeqa2dir + caltableNoPath + suffix, caltableNoPath+'.incflagged')

      print
      print 40*"-"

    ###############################################
    # Tsys derivative, overlay time, field by field
    ###############################################

    # implement once plotbandpass can do it

    ##############################################
    # manual reduction Tsys
    ##############################################
    buildplots = True
    manualcaltable = manualcaldir + ms + '.tsys'
    manualcaltableNoPath = manualcaltable.split('/')[-1]
    if (os.path.exists(manualcaltable + '.plots.overlayTime/') == True):
        if (len(os.listdir(manualcaltable+'.plots.overlayTime')) > 0):
            buildplots = False

    # RI TODO if not write permission to add .plots.overlayTime 
    if (buildplots and showingManualQA2):
        print "Tsys overlay time (manual):"
        print
        manualTsysPlot = manualcaltable + '.plots.overlayTime/' + manualcaltableNoPath
        if ('task_plotbandpass' in sys.modules and force_au_plotbandpass==False and casadef.casa_version >= pbp_version):
            println(lineno(),"Calling plotbandpass(caltable='%s', vis='%s', overlay='time,antenna', xaxis='freq', yaxis='amp', buildpdf=False, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=False, figfile='%s')" % (manualcaltable,manualcaldir+ms,chanrange,manualTsysPlot))
            plotbandpass(caltable=manualcaltable, vis=manualcaldir+ms, overlay='time,antenna', xaxis='freq',
                         yaxis='amp',buildpdf=False,interactive=False,showatm=True,pwv='auto',chanrange=chanrange,
                         showfdm=False, figfile=manualTsysPlot,
                         solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)
        else:
            println(lineno(),"Calling au.plotbandpass(caltable='%s', ms='%s', overlay='time,antenna', xaxis='freq', yaxis='amp', buildpdf=False, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=False, figfile='%s')" % (manualcaltable,manualcaldir+ms,chanrange,manualTsysPlot))
            au.plotbandpass(caltable=manualcaltable, ms=manualcaldir+ms, overlay='time,antenna', xaxis='freq',
                            yaxis='amp',buildpdf=False,interactive=False,showatm=True,pwv='auto',chanrange=chanrange,
                            showfdm=False, figfile=manualTsysPlot,
                            solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)
        print
        print 40*"-"


    ########################################
    # Flux calibrator amplitude vs. uvdist
    ########################################
    ptypes = ['amp','phase']
    pipelineTargetSpectrumPlot = ''
    pipelineCalibratedDataPlotsAmp = ''
    pipelineCalibratedDataPlotsPhase = ''
    pipelineCalibratedDataPlotsAmpPdf = ''
    pipelineCalibratedDataPlotsPhasePdf = ''
    pipelineBandpassSpectraAmpPlot = ''
    pipelineBandpassSpectraPhasePlot = ''
    plotrangeAmp = [0,0,0,0]

    if (sevenMeterOnly):
        plotrangePhase = [0,0,-90,90]
        plotrangePhase_phasecal = [0,0,-180,180]
    else:
        plotrangePhase = [0,0,-45,45]
        plotrangePhase_phasecal = [0,0,-90,90]

    pipelineBandpassSpectraAmpPlotsPdf = ''
    pipelineBandpassSpectraPhasePlotsPdf = ''
    if (showingPipelineQA2):
      print "Fluxcal amp vs uvdist:"
      print
      println(lineno(),"Calling findAntennasUsedForFluxCalibrationFromLogfile('%s', '%s')" % (pipelineCasaLogfile,asdmNoPath))
      antennaFluxCal = findAntennasUsedForFluxCalibrationFromLogfile(pipelineCasaLogfile, asdmNoPath)
      if (len(antennaFluxCal) == 0):
          filelist = ['ampcal_uvdist.png']
      else:
          filelist = ['ampcal_uvdist.png','ampcal_uvdist_antenna_subset.png']
      runAmpcalUvdist = True
      for pngfile in filelist:
          if (os.path.exists(pipeqa2dir+pngfile) == False):        
              println(lineno(),"----------------WARNING: Did not find: %s (will create)" % (pipeqa2dir+pngfile))
              runAmpcalUvdist = True
      if (runAmpcalUvdist):
          println(lineno(),"Running es.ampcal_uvdist(ms2='%s', qa2_output_dir='%s', spw='%s', correlation='%s')" % (writeableMSWithPathPipeline, pipeqa2dir, scienceSpws, correlation))
          clearstat()  # This is needed in some cases, but why does the table lock happen?
          es.ampcal_uvdist(ms2=writeableMSWithPathPipeline, qa2_output_dir=pipeqa2dir, spw=scienceSpws, antenna='', correlation=correlation)
          println(lineno(),"Antennas used by the pipeline for gaincal on the flux calibrator = %s" % antennaFluxCal)
          if (len(antennaFluxCal) > 0):
              antennaFluxCalIntrabaselines = antennaFluxCal + '&'
          else:
              antennaFluxCalIntrabaselines = antennaFluxCal
          println(lineno(),"Running es.ampcal_uvdist(ms2='%s', qa2_output_dir='%s', spw='%s', antenna='%s', correlation='%s')" % (writeableMSWithPathPipeline, pipeqa2dir, scienceSpws,antennaFluxCalIntrabaselines, correlation))
          es.ampcal_uvdist(ms2=writeableMSWithPathPipeline, qa2_output_dir=pipeqa2dir, spw=scienceSpws, antenna=antennaFluxCalIntrabaselines, correlation=correlation)
      elif (showSkippingMessages):
          println(lineno(),"%sSkipping es.ampcal_uvdist() calls since the results already exists in the pipeline QA2 directory: %s " % (skipIndent, pipeqa2dir))          
      print
      print 40*"-"


      ######################################################################
      ## create target_spectrum from pipeline data
      ######################################################################
      pipelineTargetSpectrumPlot = pipeqa2dir + 'target_spectrum.png'
      if (os.path.exists(pipelineTargetSpectrumPlot+"foo") == False):
          if (badCharacters(target_source_name)):
              cmd="es.target_spectrum(ms2='%s', target='%s', dospw='%s', qa2_output_dir='%s', uvrange='%s', correlation='%s', avgchannel=%d)" % (writeableMSWithPathPipeline, str(target_source_id), scienceSpws, pipeqa2dir, uvrange, correlation, avgchannelTarget)
          else:
              cmd="es.target_spectrum(ms2='%s', target='%s', dospw='%s', qa2_output_dir='%s', uvrange='%s', correlation='%s', avgchannel=%d)" % (writeableMSWithPathPipeline, target_source_name, scienceSpws, pipeqa2dir, uvrange, correlation,avgchannelTarget)

          println(lineno(),"Running for pipeline data: "+cmd)
          exec cmd
          
      print

      #####################################################################
      ## create bandpass_spectrum of bp calibrator from pipeline data w/plotms
      #####################################################################
      clearstat()
      pnglist = {}
      pnglist['amp'] = ''
      pnglist['phase'] = ''
      pipelineBandpassSpectraAmpPlot = pipeqa2dir + 'bandpass_spectrum_amp.png'
      pipelineBandpassSpectraPhasePlot = pipeqa2dir + 'bandpass_spectrum_phase.png'
      pipelineBandpassSpectraAmpPlots = []
      pipelineBandpassSpectraPhasePlots = []
      if (plotlevel > 1):
          pipelineBandpassSpectraAmpPlotsPdf = pipeqa2dir + 'bandpass_spectrum_amp_pipeline.pdf'
          pipelineBandpassSpectraPhasePlotsPdf = pipeqa2dir + 'bandpass_spectrum_phase_pipeline.pdf'
      overwrite = True
      interactive = False
      avgtime = '100000'
      for spwinteger in scienceSpwsIntegerList:
          spw = str(spwinteger)
          for ptype in ptypes:
              figfile = pipeqa2dir + 'bandpass_spectrum_%s_spw%s.png'%(ptype,spw)
              pnglist[ptype] += ' %s' % figfile
              if (ptype == 'amp'):
                  plotrange = plotrangeAmp 
              else:
                  plotrange = plotrangePhase
              if (os.path.exists(figfile) == False):
                  coloraxis = 'antenna1'
                  runPlotms("plotms("+plotmsparms+",vis='%s', xaxis='frequency', yaxis='%s', coloraxis='%s', field='%s', spw='%s', ydatacolumn='corrected', title='Calibrated spw %s (%s)', avgtime='%s', avgscan=True, avgantenna=True, plotfile='%s', plotrange=%s, correlation='%s')" % (writeableMSWithPathPipeline,ptype,coloraxis,bandpass_source_name,spw,spw,bandpass_source_name,avgtime,figfile,plotrange,correlation))
                  # If any of the component plots were missing, then remove the old montage result.
                  if (ptype == 'amp'):
                      if (os.path.exists(pipelineBandpassSpectraAmpPlot)):
                          os.system('rm -f '+pipelineBandpassSpectraAmpPlot)
                  else:
                      if (os.path.exists(pipelineBandpassSpectraPhasePlot)):
                          os.system('rm -f '+pipelineBandpassSpectraPhasePlot)
              # now loop over individual antennas
              if (plotlevel > 1 and ((ptype == 'amp' and os.path.exists(pipelineBandpassSpectraAmpPlotsPdf) == False)
                  or (ptype == 'phase' and os.path.exists(pipelineBandpassSpectraPhasePlotsPdf) == False))):
                  for antname in antennaNames:
                      figfile = pipeqa2dir + 'bandpass_spectrum_%s_spw%s_%s.png'%(ptype,spw,antname)
                      if (os.path.exists(figfile) == False):
                          runPlotms("plotms("+plotmsparms+",vis='%s', antenna='%s', xaxis='frequency', yaxis='%s', coloraxis='corr', field='%s', spw='%s', ydatacolumn='corrected', avgbaseline=True, title='%s spw %s (%s) (pipeline)', avgtime='%s', avgscan=True, plotfile='%s', plotrange=%s, correlation='%s')" % (writeableMSWithPathPipeline,antname,ptype,bandpass_source_name,spw,antname,spw,bandpass_source_name,avgtime,figfile,plotrange, correlation))
                      else:
                          println(lineno(),"Skipping plotms of %s %s %s for pipeline bandpass spectra" % (antname, spw, ptype))
                      # If any of the component plots were missing, then remove the old montage result.
                      if (ptype == 'amp'):
                          pipelineBandpassSpectraAmpPlots.append(figfile)
                          if (os.path.exists(pipelineBandpassSpectraAmpPlotsPdf)):
                              os.system('rm -f '+pipelineBandpassSpectraAmpPlotsPdf)
                      else:
                          pipelineBandpassSpectraPhasePlots.append(figfile)
                          if (os.path.exists(pipelineBandpassSpectraPhasePlotsPdf)):
                              os.system('rm -f '+pipelineBandpassSpectraPhasePlotsPdf)
              elif (showSkippingMessages):
                  println(lineno(),"%sSkipping plotms of pipeline calibrated BPcal %s spectrum for spw %s" % (skipIndent,ptype,spw))
      if (os.path.exists(pipelineBandpassSpectraAmpPlot) == False or
          os.path.exists(pipelineBandpassSpectraPhasePlot) == False):
          tile = picktile(pnglist['amp'])
          println(lineno(),"tile = %s"% tile)
          println(lineno(),"pnglist['amp'] = %s"% pnglist['amp'])
          println(lineno(),"pnglist['phase'] = %s"% pnglist['phase'])
          println(lineno(),"pipelineBandpassSpectraAmpPlot = %s"% pipelineBandpassSpectraAmpPlot)
          os.system("montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist['amp'], pipelineBandpassSpectraAmpPlot))
          os.system("montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist['phase'], pipelineBandpassSpectraPhasePlot))
      if (os.path.exists(pipelineBandpassSpectraAmpPlotsPdf) == False and plotlevel > 1):
          # Now montage them into pipelineBandpassSpectraAmpPlotsPdf
          println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (pipelineBandpassSpectraAmpPlots,pipelineBandpassSpectraAmpPlotsPdf))
          au.buildPdfFromPngs(pipelineBandpassSpectraAmpPlots, pdfname=pipelineBandpassSpectraAmpPlotsPdf)
      if (os.path.exists(pipelineBandpassSpectraPhasePlotsPdf) == False and plotlevel > 1):
          println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (pipelineBandpassSpectraPhasePlots,pipelineBandpassSpectraPhasePlotsPdf))
          au.buildPdfFromPngs(pipelineBandpassSpectraPhasePlots, pdfname=pipelineBandpassSpectraPhasePlotsPdf,quiet=auQuiet)
    
    manualBandpassSpectraAmpPlot = ''
    manualBandpassSpectraPhasePlot = ''
    manualBandpassSpectraAmpPlots = []
    manualBandpassSpectraPhasePlots = []
    manualBandpassSpectraAmpPlotsPdf = ''
    manualBandpassSpectraPhasePlotsPdf = ''
    directoriesAndVersions =  ''
    if (showingManualQA2):
      if (plotlevel > 1):
          manualBandpassSpectraAmpPlotsPdf = manualqa2dir_extra + 'bandpass_spectrum_amp_manual.pdf'
          manualBandpassSpectraPhasePlotsPdf = manualqa2dir_extra + 'bandpass_spectrum_phase_manual.pdf'
      sfc = find_manual_script(manualcaldir)  # *scriptForCalibration.py
      sfcDotTxt =  sfc + '.txt' # add .txt so that firefox will show it
      if (sfc != None):
          myCopyfile(manualcaldir+sfc, webpagedir+sfcDotTxt)
      clearstat()
      antennaFluxCalManual = findAntennasUsedForFluxCalibrationFromManualScript(manualcaldir+sfc)
      println(lineno(),"antenna subset for flux cal manual reduction = %s" % (antennaFluxCalManual))
      ## create bandpass_spectrum from manual data using plotms
      pnglist = {}
      pnglist['amp'] = ''
      pnglist['phase'] = ''
      manualBandpassSpectraAmpPlot = manualqa2dir_extra + 'bandpass_spectrum_amp.png'
      manualBandpassSpectraPhasePlot = manualqa2dir_extra + 'bandpass_spectrum_phase.png'
      overwrite = True
      interactive = False
      avgtime = '100000'
      for spwinteger in range(len(scienceSpwsIntegerList)):
          spw = str(spwinteger)
          for ptype in ptypes:
              figfile = manualqa2dir_extra + 'bandpass_spectrum_%s_spw%s.png'%(ptype,spw)
              pnglist[ptype] += ' %s' % figfile
              if (ptype == 'amp'):
                  plotrange = plotrangeAmp 
              else:
                  plotrange = plotrangePhase
              if (os.path.exists(figfile) == False):
                  coloraxis='antenna1'
                  runPlotms("plotms("+plotmsparms+",vis='%s', xaxis='frequency', yaxis='%s', coloraxis='%s', field='%s', spw='%s', ydatacolumn='corrected', title='Calibrated spw %s (%s)', avgtime='%s', avgscan=True, avgantenna=True, plotfile='%s', plotrange=%s, correlation='%s')" % (msSplitWithPath,ptype,coloraxis,bandpass_source_name_manual,spw,spw,bandpass_source_name_manual,avgtime,figfile,plotrange,correlation))
                  # If any of the component plots were missing, then remove the old montage result.
                  if (ptype == 'amp'):
                      manualBandpassSpectraAmpPlots.append(figfile)
                      if (os.path.exists(manualBandpassSpectraAmpPlot)):
                          os.system('rm -f '+manualBandpassSpectraAmpPlot)
                  else:
                      manualBandpassSpectraPhasePlots.append(figfile)
                      if (os.path.exists(manualBandpassSpectraPhasePlot)):
                          os.system('rm -f '+manualBandpassSpectraPhasePlot)
              # now loop over individual antennas
              if (plotlevel > 1 and ((ptype == 'amp' and os.path.exists(manualBandpassSpectraAmpPlotsPdf) == False)
                  or (ptype == 'phase' and os.path.exists(manualBandpassSpectraPhasePlotsPdf) == False))):
                  for antname in antennaNames:
                      figfile = manualqa2dir_extra + 'bandpass_spectrum_%s_spw%s_%s.png'%(ptype,spw,antname)
                      if (os.path.exists(figfile) == False):
                          runPlotms("plotms("+plotmsparms+",vis='%s', antenna='%s', xaxis='frequency', yaxis='%s', coloraxis='corr', field='%s', spw='%s', ydatacolumn='corrected', avgbaseline=True, title='%s spw %s (%s) (manual)', avgtime='%s', avgscan=True, plotfile='%s', plotrange=%s, correlation='%s')" % (msSplitWithPath,antname,ptype,bandpass_source_name,spw,antname,spw,bandpass_source_name,avgtime,figfile,plotrange,correlation))
                      else:
                          println(lineno(),"Skipping plotms of %s %s %s for manual bandpass spectra" % (antname, spw, ptype))
                      # If any of the component plots were missing, then remove the old montage result.
                      if (ptype == 'amp'):
                          manualBandpassSpectraAmpPlots.append(figfile)
                          if (os.path.exists(manualBandpassSpectraAmpPlotsPdf)):
                              os.system('rm -f '+manualBandpassSpectraAmpPlotsPdf)
                      else:
                          manualBandpassSpectraPhasePlots.append(figfile)
                          if (os.path.exists(manualBandpassSpectraPhasePlotsPdf)):
                              os.system('rm -f '+manualBandpassSpectraPhasePlotsPdf)
              elif (showSkippingMessages):
                  println(lineno(),"%sSkipping plotms of manual calibrated BPcal %s spectrum for spw %s" % (skipIndent,ptype,spw))
      if (os.path.exists(manualBandpassSpectraAmpPlot) == False or
          os.path.exists(manualBandpassSpectraPhasePlot) == False):
          tile = picktile(pnglist['amp'])
          println(lineno(),"tile = %s"% tile)
          println(lineno(),"pnglist['amp'] = %s"% pnglist['amp'])
          println(lineno(),"pnglist['phase'] = %s"% pnglist['phase'])
          println(lineno(),"manualBandpassSpectraAmpPlot = %s"% manualBandpassSpectraAmpPlot)
          os.system("montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist['amp'], manualBandpassSpectraAmpPlot))
          os.system("montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist['phase'], manualBandpassSpectraPhasePlot))
      if (os.path.exists(manualBandpassSpectraAmpPlotsPdf) == False and plotlevel > 1):
          # Now montage them into manualBandpassSpectraAmpPlotsPdf
          println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (manualBandpassSpectraAmpPlots,manualBandpassSpectraAmpPlotsPdf))
          au.buildPdfFromPngs(manualBandpassSpectraAmpPlots, pdfname=manualBandpassSpectraAmpPlotsPdf,quiet=auQuiet)
      if (os.path.exists(manualBandpassSpectraPhasePlotsPdf) == False and plotlevel > 1):
          println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (manualBandpassSpectraPhasePlots,manualBandpassSpectraPhasePlotsPdf))
          au.buildPdfFromPngs(manualBandpassSpectraPhasePlots, pdfname=manualBandpassSpectraPhasePlotsPdf,quiet=auQuiet)

      print
      print 40*"-"


    showPhasecalBandpass = True
    if (showPhasecalBandpass):    
      ###################################################################
      ## create bandpass_spectrum of ph cal from pipeline data w/ plotms
      ###################################################################
      pipelinePhasecalSpectraAmpPlot = ''
      pipelinePhasecalSpectraPhasePlot = ''
      pipelinePhasecalSpectraAmpPlotsPdf = ''
      pipelinePhasecalSpectraPhasePlotsPdf = ''
      avgtime = '100000'
      overwrite = True
      interactive = False
      phasecal_avgchan_default = 8  # but only if more than 128 channels
      if (showingPipelineQA2):
          print "Bandpass spectrum of phase cal with plotms"
          print
          clearstat()
          pnglist = {}
          pnglist['amp'] = ''
          pnglist['phase'] = ''
          pipelinePhasecalSpectraAmpPlot = pipeqa2dir + 'phasecal_spectrum_amp.png'
          pipelinePhasecalSpectraPhasePlot = pipeqa2dir + 'phasecal_spectrum_phase.png'
          pipelinePhasecalSpectraAmpPlots = []
          pipelinePhasecalSpectraPhasePlots = []
          if (plotlevel > 1):
              pipelinePhasecalSpectraAmpPlotsPdf = pipeqa2dir + 'phasecal_spectrum_amp_pipeline.pdf'
              pipelinePhasecalSpectraPhasePlotsPdf = pipeqa2dir + 'phasecal_spectrum_phase_pipeline.pdf'
          for spwinteger in scienceSpwsIntegerList:
              spw = str(spwinteger)
              phasecal_avgchan = phasecal_avgchan_default
              if (casadef.casa_version >= '4.1.0'):
                  if (nchanforspw[spwinteger] <= 128):
                      phasecal_avgchan = 0
                  
              for ptype in ptypes:
                  figfile = pipeqa2dir + 'phasecal_spectrum_%s_spw%s.png'%(ptype,spw)
                  pnglist[ptype] += ' %s' % figfile
                  if (ptype == 'amp'):
                      plotrange = plotrangeAmp 
                  else:
                      plotrange = plotrangePhase_phasecal
                  if (os.path.exists(figfile) == False):
                      coloraxis = 'antenna1'
                      runPlotms("plotms("+plotmsparms+",vis='%s', xaxis='frequency', yaxis='%s', coloraxis='%s', field='%s', spw='%s', ydatacolumn='corrected', title='Calibrated spw %s (%s)', avgtime='%s', avgscan=True, avgantenna=True, plotfile='%s', plotrange=%s, correlation='%s', avgchannel='%d')" % (writeableMSWithPathPipeline,ptype,coloraxis,phase_cal_source_name,spw,spw,phase_cal_source_name,avgtime,figfile,plotrange, correlation,phasecal_avgchan))
                      # If any of the component plots were missing, then remove the old montage result.
                      if (ptype == 'amp'):
                          if (os.path.exists(pipelinePhasecalSpectraAmpPlot)):
                              os.system('rm -f '+pipelinePhasecalSpectraAmpPlot)
                      else:
                          if (os.path.exists(pipelinePhasecalSpectraPhasePlot)):
                              os.system('rm -f '+pipelinePhasecalSpectraPhasePlot)
                  # now loop over individual antennas
                  if (plotlevel > 1 and ((ptype == 'amp' and os.path.exists(pipelinePhasecalSpectraAmpPlotsPdf) == False)
                      or (ptype == 'phase' and os.path.exists(pipelinePhasecalSpectraPhasePlotsPdf) == False))):
                      for antname in antennaNames:
                          figfile = pipeqa2dir + 'phasecal_spectrum_%s_spw%s_%s.png'%(ptype,spw,antname)
                          if (os.path.exists(figfile) == False):
                              runPlotms("plotms("+plotmsparms+",vis='%s', antenna='%s', xaxis='frequency', yaxis='%s', coloraxis='corr', field='%s', spw='%s', ydatacolumn='corrected', avgbaseline=True, title='%s spw %s (%s) (pipeline)', avgscan=True, avgtime='%s', plotfile='%s', plotrange=%s, correlation='%s', avgchannel='%d')" % (writeableMSWithPathPipeline,antname,ptype,phase_cal_source_name,spw,antname,spw,phase_cal_source_name,avgtime,figfile,plotrange,correlation,phasecal_avgchan))
                          else:
                              println(lineno(),"Skipping plotms of %s %s %s for pipeline phasecal spectra" % (antname, spw, ptype))
                          # If any of the component plots were missing, then remove the old montage result.
                          if (ptype == 'amp'):
                              pipelinePhasecalSpectraAmpPlots.append(figfile)
                              if (os.path.exists(pipelinePhasecalSpectraAmpPlotsPdf)):
                                  os.system('rm -f '+pipelinePhasecalSpectraAmpPlotsPdf)
                          else:
                              pipelinePhasecalSpectraPhasePlots.append(figfile)
                              if (os.path.exists(pipelinePhasecalSpectraPhasePlotsPdf)):
                                  os.system('rm -f '+pipelinePhasecalSpectraPhasePlotsPdf)
                  elif (showSkippingMessages):
                      println(lineno(),"%sSkipping plotms of pipeline calibrated phasecal %s spectrum for spw %s" % (skipIndent,ptype,spw))
                      
          if (os.path.exists(pipelinePhasecalSpectraAmpPlot) == False or
              os.path.exists(pipelinePhasecalSpectraPhasePlot) == False):
              tile = picktile(pnglist['amp'])
              println(lineno(),"tile = %s"% tile)
              println(lineno(),"pnglist['amp'] = %s"% pnglist['amp'])
              println(lineno(),"pnglist['phase'] = %s"% pnglist['phase'])
              println(lineno(),"pipelinePhasecalSpectraAmpPlot = %s"% pipelinePhasecalSpectraAmpPlot)
              os.system("montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist['amp'], pipelinePhasecalSpectraAmpPlot))
              os.system("montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist['phase'], pipelinePhasecalSpectraPhasePlot))
          if (os.path.exists(pipelinePhasecalSpectraAmpPlotsPdf) == False and plotlevel > 1):
              # Now montage them into pipelinePhasecalSpectraAmpPlotsPdf
              println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (pipelinePhasecalSpectraAmpPlots,pipelinePhasecalSpectraAmpPlotsPdf))
              au.buildPdfFromPngs(pipelinePhasecalSpectraAmpPlots, pdfname=pipelinePhasecalSpectraAmpPlotsPdf,quiet=auQuiet)
          if (os.path.exists(pipelinePhasecalSpectraPhasePlotsPdf) == False and plotlevel > 1):
              println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (pipelinePhasecalSpectraPhasePlots,pipelinePhasecalSpectraPhasePlotsPdf))
              au.buildPdfFromPngs(pipelinePhasecalSpectraPhasePlots, pdfname=pipelinePhasecalSpectraPhasePlotsPdf,quiet=auQuiet)
                      
        
          print
          print 40*"-"




      manualPhasecalSpectraAmpPlot = ''
      manualPhasecalSpectraPhasePlot = ''
      manualPhasecalSpectraAmpPlotsPdf = ''
      manualPhasecalSpectraPhasePlotsPdf = ''
      if (showingManualQA2):
          ## create phasecal_spectrum from manual data using plotms
          pnglist = {}
          pnglist['amp'] = ''
          pnglist['phase'] = ''
          manualPhasecalSpectraAmpPlot = manualqa2dir_extra + 'phasecal_spectrum_amp.png'
          manualPhasecalSpectraPhasePlot = manualqa2dir_extra + 'phasecal_spectrum_phase.png'
          manualPhasecalSpectraAmpPlots = []
          manualPhasecalSpectraPhasePlots = []
          if (plotlevel > 1):
              manualPhasecalSpectraAmpPlotsPdf = manualqa2dir_extra + 'phasecal_spectrum_amp_manual.pdf'
              manualPhasecalSpectraPhasePlotsPdf = manualqa2dir_extra + 'phasecal_spectrum_phase_manual.pdf'
          for spwinteger in range(len(scienceSpwsIntegerList)):
              spw = str(spwinteger)
              for ptype in ptypes:
                  figfile = manualqa2dir_extra + 'phasecal_spectrum_%s_spw%s.png'%(ptype,spw)
                  pnglist[ptype] += ' %s' % figfile
                  if (ptype == 'amp'):
                      plotrange = plotrangeAmp 
                  else:
                      plotrange = plotrangePhase_phasecal
                  if (os.path.exists(figfile) == False):
                      overwrite = True
                      interactive = False
                      coloraxis='antenna1'
                      runPlotms("plotms("+plotmsparms+",vis='%s', xaxis='frequency', yaxis='%s', coloraxis='%s', field='%s', spw='%s', ydatacolumn='corrected', title='Calibrated spw %s (%s)', avgtime='%s', avgscan=True, avgantenna=True, plotfile='%s', plotrange=%s, correlation='%s', avgchannel='%d')" % (msSplitWithPath,ptype,coloraxis,phase_cal_source_name,spw,spw,phase_cal_source_name,avgtime,figfile,plotrange,correlation,phasecal_avgchan))
                      # If any of the component plots were missing, then remove the old montage result.
                      if (ptype == 'amp'):
                          if (os.path.exists(manualPhasecalSpectraAmpPlot)):
                              os.system('rm -f '+manualPhasecalSpectraAmpPlot)
                      else:
                          if (os.path.exists(manualPhasecalSpectraPhasePlot)):
                              os.system('rm -f '+manualPhasecalSpectraPhasePlot)
                  # now loop over individual antennas
                  if (plotlevel > 1 and ((ptype == 'amp' and os.path.exists(manualPhasecalSpectraAmpPlotsPdf) == False)
                      or (ptype == 'phase' and os.path.exists(manualPhasecalSpectraPhasePlotsPdf) == False))):
                      for antname in antennaNames:
                          figfile = manualqa2dir_extra + 'phasecal_spectrum_%s_spw%s_%s.png'%(ptype,spw,antname)
                          if (os.path.exists(figfile) == False):
                              runPlotms("plotms("+plotmsparms+",vis='%s', antenna='%s', xaxis='frequency', yaxis='%s', coloraxis='corr', field='%s', spw='%s', ydatacolumn='corrected', avgbaseline=True, title='%s spw %s (%s) (manual)', avgtime='%s', avgscan=True, plotfile='%s', plotrange=%s, correlation='%s', avgchannel='%d')" % (msSplitWithPath,antname,ptype,phase_cal_source_name,spw,antname,spw,phase_cal_source_name,avgtime,figfile,plotrange,correlation,phasecal_avgchan))
                          else:
                              println(lineno(),"Skipping plotms of %s %s %s for manual phasecal spectra" % (antname, spw, ptype))
                          # If any of the component plots were missing, then remove the old montage result.
                          if (ptype == 'amp'):
                              manualPhasecalSpectraAmpPlots.append(figfile)
                              if (os.path.exists(manualPhasecalSpectraAmpPlotsPdf)):
                                  os.system('rm -f '+manualPhasecalSpectraAmpPlotsPdf)
                          else:
                              manualPhasecalSpectraPhasePlots.append(figfile)
                              if (os.path.exists(manualPhasecalSpectraPhasePlotsPdf)):
                                  os.system('rm -f '+manualPhasecalSpectraPhasePlotsPdf)
                  elif (showSkippingMessages):
                      println(lineno(),"%sSkipping plotms of manual calibrated phasecal %s spectrum for spw %s" % (skipIndent,ptype,spw))
          if (os.path.exists(manualPhasecalSpectraAmpPlot) == False or
              os.path.exists(manualPhasecalSpectraPhasePlot) == False):
              tile = picktile(pnglist['amp'])
              println(lineno(),"tile = %s"% tile)
              println(lineno(),"pnglist['amp'] = %s"% pnglist['amp'])
              println(lineno(),"pnglist['phase'] = %s"% pnglist['phase'])
              println(lineno(),"manualPhasecalSpectraAmpPlot = %s"% manualPhasecalSpectraAmpPlot)
              os.system("montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist['amp'], manualPhasecalSpectraAmpPlot))
              os.system("montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist['phase'], manualPhasecalSpectraPhasePlot))
          if (os.path.exists(manualPhasecalSpectraAmpPlotsPdf) == False and plotlevel > 1):
              # Now montage them into manualPhasecalSpectraAmpPlotsPdf
              println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (manualPhasecalSpectraAmpPlots,manualPhasecalSpectraAmpPlotsPdf))
              au.buildPdfFromPngs(manualPhasecalSpectraAmpPlots, pdfname=manualPhasecalSpectraAmpPlotsPdf,quiet=auQuiet)
          if (os.path.exists(manualPhasecalSpectraPhasePlotsPdf) == False and plotlevel > 1):
              println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (manualPhasecalSpectraPhasePlots,manualPhasecalSpectraPhasePlotsPdf))
              au.buildPdfFromPngs(manualPhasecalSpectraPhasePlots, pdfname=manualPhasecalSpectraPhasePlotsPdf,quiet=auQuiet)

      print
      print 40*"-"

#  end of pipelineBandpass  phasecal_spectrum manualBandpass





    ##################################
    # Make plots of the corrected data
    ##################################
    mytb = createCasaTool(tbtool)
    pipelineCalibratedFluxCalPlotsAmp = ''
    if (showingPipelineQA2):
        print "Plots of corrected data:"
        print
        mytb.open(writeableMSWithPathPipeline)
        if ('CORRECTED_DATA' not in mytb.colnames()):
            pipelineCorrectedDataPresent = False
            println(lineno(),"WARNING: The CORRECTED_DATA column is not present in the pipeline reduction.")
        else:
            pipelineCorrectedDataPresent = True
        mytb.close()
        if (pipelineCorrectedDataPresent):
            # Calibrated data from pipeline
            pipelineCalibratedDataPlots = {}
            pipelineCalibratedDataPlots['amp'] = []
            pipelineCalibratedDataPlots['phase'] = []
            pipelineCalibratedDataPlotsPerAntenna = {}
            pipelineCalibratedDataPlotsPerAntenna['amp'] = []
            pipelineCalibratedDataPlotsPerAntenna['phase'] = []
            if (plotlevel > 1):
                pipelineCalibratedDataPlotsPhasePdf = pipeqa2dir + 'calibrated_data_phase_pipeline.pdf'
                pipelineCalibratedDataPlotsAmpPdf = pipeqa2dir + 'calibrated_data_amp_pipeline.pdf'
            pnglist = {}
            pnglist['amp'] = ''
            pnglist['phase'] = ''
            clearstat()
            for spwinteger in scienceSpwsIntegerList:
                spw = str(spwinteger)
                for ptype in ptypes:
                    figfile = pipeqa2dir+'calibrated_data_%s_spw%s.png'%(ptype,spw)
                    pipelineCalibratedDataPlots[ptype].append(figfile)
                    pnglist[ptype] += ' %s' % figfile
                    if (os.path.exists(figfile) == False):
                        overwrite = True
                        interactive = False
                        avgtime = '6'
                        avgchannel = maxChannels
                        runPlotms("plotms("+plotmsparms+",vis='%s', xaxis='time', yaxis='%s', coloraxis='field', title='Calibrated %s vs. time (spw %s)', avgchannel='%s', spw='%s', ydatacolumn='corrected', plotfile='%s', correlation='%s', avgtime='%s')" % (writeableMSWithPathPipeline, ptype, ptype, spw, avgchannel, spw, figfile, correlation,avgtime))
                        # If any of the component plots were missing, then remove the old montage result.
                        if (ptype == 'amp'):
                            if (os.path.exists(pipelineCalibratedDataPlotsAmp)):
                                os.system('rm -f '+pipelineCalibratedDataPlotsAmp)
                        else:
                            if (os.path.exists(pipelineCalibratedDataPlotsPhase)):
                                os.system('rm -f '+pipelineCalibratedDataPlotsPhase)
#                        if (os.path.exists(figfile) == False and debug_plotms):
#                            # Write to an execfile for later processing
#                            fplotms.write(cmd)
#                            writeWaitForPlotms(fplotms)
#                            fplotmsHasCommands += 1
                    elif (showSkippingMessages):
                        println(lineno(),"%sSkipping plotms of pipeline calibrated %s vs time (antennas overlaid) for spw %s" % (skipIndent,ptype,spw))
                    # now loop over individual antennas
                    if (plotlevel > 1 and ((ptype == 'amp' and os.path.exists(pipelineCalibratedDataPlotsAmpPdf) == False)
                        or (ptype == 'phase' and os.path.exists(pipelineCalibratedDataPlotsPhasePdf) == False))):
                        for antname in antennaNames:
                            figfile = pipeqa2dir + 'calibrated_data_%s_spw%s_%s.png'%(ptype,spw,antname)
                            if (os.path.exists(figfile) == False):
                                runPlotms("plotms("+plotmsparms+",vis='%s', antenna='%s', xaxis='time', yaxis='%s', coloraxis='field', spw='%s', ydatacolumn='corrected', avgbaseline=True, title='%s spw %s (pipeline)', plotfile='%s', correlation='%s')" % (writeableMSWithPathPipeline,antname,ptype,spw,antname,spw,figfile,correlation))
                            else:
                                println(lineno(),"Skipping plotms of %s %s %s for calibrated data" % (antname, spw, ptype))
                            pipelineCalibratedDataPlotsPerAntenna[ptype].append(figfile)
                            # If any of the component plots were missing, then remove the old montage result.
                            if (ptype == 'phase'):
                                if (os.path.exists(pipelineCalibratedDataPlotsPhasePdf)):
                                    os.system('rm -f '+pipelineCalibratedDataPlotsPhasePdf)
                            else:
                                if (os.path.exists(pipelineCalibratedDataPlotsAmpPdf)):
                                    os.system('rm -f '+pipelineCalibratedDataPlotsAmpPdf)
            pipelineCalibratedDataPlotsAmp = pipeqa2dir + 'calibrated_data_amp.png'
            pipelineCalibratedDataPlotsPhase = pipeqa2dir + 'calibrated_data_phase.png'
            if (os.path.exists(pipelineCalibratedDataPlotsAmp) == False or
                os.path.exists(pipelineCalibratedDataPlotsPhase) == False):
                tile = picktile(pnglist['amp'])
                println(lineno(),"tile = %s"% tile)
                println(lineno(),"pnglist['amp'] = %s"% pnglist['amp'])
                println(lineno(),"pnglist['phase'] = %s"% pnglist['phase'])
                println(lineno(),"pipelineCalibratedDataPlotsAmp = %s"% pipelineCalibratedDataPlotsAmp)
                os.system("montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist['amp'], pipelineCalibratedDataPlotsAmp))
                os.system("montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist['phase'], pipelineCalibratedDataPlotsPhase))
            if (os.path.exists(pipelineCalibratedDataPlotsAmpPdf) == False and plotlevel > 1):
                # Now montage them into pipelinePhasecalSpectraAmpPlotsPdf
                println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (pipelineCalibratedDataPlotsPerAntenna['amp'],pipelineCalibratedDataPlotsAmpPdf))
                au.buildPdfFromPngs(pipelineCalibratedDataPlotsPerAntenna['amp'], pdfname=pipelineCalibratedDataPlotsAmpPdf,quiet=auQuiet)
            if (os.path.exists(pipelineCalibratedDataPlotsPhasePdf) == False and plotlevel > 1):
                println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (pipelineCalibratedDataPlotsPerAntenna['phase'],pipelineCalibratedDataPlotsPhasePdf))
                au.buildPdfFromPngs(pipelineCalibratedDataPlotsPerAntenna['phase'], pdfname=pipelineCalibratedDataPlotsPhasePdf,quiet=auQuiet)
                            
            if (fluxcal_source_id >= 0 and fluxcal_source_id != ''):
                pipelineCalibratedFluxCalPlotsAmp = pipeqa2dir + 'calibrated_data_fluxcal_amp.png'
                if (os.path.exists(pipelineCalibratedFluxCalPlotsAmp) == False):
                    pnglist = ''
                    for spwinteger in scienceSpwsIntegerList:
                        spw = str(spwinteger)
                        figfile = pipeqa2dir + 'calibrated_data_fluxcal_amp_spw%s.png'%(spw)
                        pnglist += ' %s' % figfile
                        if (os.path.exists(figfile) == False):
                            overwrite = True
                            interactive = False
                            avgbaseline = True
                            avgchannel = ''
                            avgtime = '1e9'
                            correlation = ''
                            correctedfigfile = pipeqa2dir + 'calibrated_data_fluxcal_amp_spw%s_corrected.png'%(spw)
                            runPlotms("plotms("+plotmsparms+",vis='%s', xaxis='freq', yaxis='amp', coloraxis='corr',title='Calibrated amp vs. freq (spw %s)', avgchannel='%s', spw='%s', ydatacolumn='corrected', plotfile='%s', correlation='%s', field='%d', avgbaseline=%s,uvrange='%s', avgtime='%s')" % (writeableMSWithPathPipeline, spw, avgchannel, spw, correctedfigfile, correlation,fluxcal_source_id,avgbaseline,fluxcal_uvrange,avgtime))
                            modelfigfile = pipeqa2dir + 'calibrated_data_fluxcal_amp_spw%s_model.png'%(spw)
                            runPlotms("plotms("+plotmsparms+",vis='%s', xaxis='freq', yaxis='amp', coloraxis='corr',title='Calibrated amp vs. freq (spw %s)', avgchannel='%s', spw='%s', ydatacolumn='model', plotfile='%s', correlation='%s', field='%d', avgbaseline=%s, uvrange='%s', avgtime='%s')" % (writeableMSWithPathPipeline, spw, avgchannel, spw, modelfigfile, correlation,fluxcal_source_id,avgbaseline,fluxcal_uvrange,avgtime))
                            command = "montage -geometry 1000x1000+1+1 %s %s %s" % (correctedfigfile, modelfigfile, figfile)
                            println(lineno(),"%sRunning: %s" % (skipIndent,command))
                            os.system(command)
                            # If any of the component plots were missing, then remove the old montage result.
                            if (os.path.exists(pipelineCalibratedFluxCalPlotsAmp)):
                                os.system('rm -f '+pipelineCalibratedFluxCalPlotsAmp)
                        elif (showSkippingMessages):
                            println(lineno(),"%sSkipping plotms of pipeline calibrated flux cal amp vs freq (spw%s)" % (skipIndent,spw))
                    tile = picktile(pnglist)
                    command = "montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist, pipelineCalibratedFluxCalPlotsAmp)
                    println(lineno(),"%sRunning: %s" % (skipIndent,command))
                    os.system(command)
                elif (showSkippingMessages):
                    println(lineno(),"%sSkipping plotms of pipeline calibrated flux cal amp vs time" % (skipIndent))
    

    print
    print 40*"-"

    ###################################################
    manualCalibratedDataPlotsAmp = ''
    manualCalibratedDataPlotsPhase = ''
    manualCalibratedDataPlotsAmpPdf = ''
    manualCalibratedDataPlotsPhasePdf = ''
    clearstat()
    manualCalibratedFluxCalPlotsAmp = ''
    if (showingManualQA2):
        print "Calibrated data spectrum (manual reduction):"
        print
        mytb.open(msSplitWithPath)
        if ('CORRECTED_DATA' not in mytb.colnames()):
            manualCorrectedDataPresent = False
            println(lineno(),"WARNING: The CORRECTED_DATA column is not present in the manual reduction.")
        else:
           manualCorrectedDataPresent = True
        mytb.close()
        if (manualCorrectedDataPresent):
            # Calibrated data from manual reduction
            manualCalibratedDataPlots = {}
            manualCalibratedDataPlots['amp'] = []
            manualCalibratedDataPlots['phase'] = []
            manualCalibratedDataPlotsPerAntenna = {}
            manualCalibratedDataPlotsPerAntenna['amp'] = []
            manualCalibratedDataPlotsPerAntenna['phase'] = []
            if (plotlevel > 1):
                manualCalibratedDataPlotsPhasePdf = manualqa2dir_extra + 'calibrated_data_phase.pdf'
                manualCalibratedDataPlotsAmpPdf = manualqa2dir_extra + 'calibrated_data_amp.pdf'
            pnglist = {}
            pnglist['amp'] = ''
            pnglist['phase'] = ''

            for spwinteger in range(len(scienceSpwsIntegerList)):
              spw = str(spwinteger)
              unsplitspw = str(scienceSpwsIntegerList[spwinteger])

              for ptype in ptypes:
                figfile = manualqa2dir_extra+'calibrated_data_%s_spw%s.png'%(ptype,spw)
                manualCalibratedDataPlots[ptype].append(figfile)
                pnglist[ptype] += ' %s' % figfile
                if (os.path.exists(figfile) == False):
                    overwrite = True
                    interactive = False
                    avgchannel = maxChannels
                    avgtime = '6s'
                    runPlotms("plotms("+plotmsparms+",vis='%s', xaxis='time', yaxis='%s', coloraxis='field',title='Calibrated %s vs. time (spw %s)', avgchannel='%s', spw='%s', ydatacolumn='corrected', plotfile='%s', correlation='%s', avgtime='%s')" % (msSplitWithPath, ptype, ptype, spw, avgchannel, spw, figfile, correlation,avgtime))
                    # If any of the component plots were missing, then remove the old montage result.
                    if (ptype == 'amp'):
                        if (os.path.exists(manualCalibratedDataPlotsAmp)):
                            os.system('rm -f '+manualCalibratedDataPlotsAmp)
                    else:
                        if (os.path.exists(manualCalibratedDataPlotsPhase)):
                            os.system('rm -f '+manualCalibratedDataPlotsPhase)
#                    if (os.path.exists(figfile) == False and debug_plotms):
#                        # Write to an execfile for later processing
#                        fplotms.write(cmd)
#                        writeWaitForPlotms(fplotms)
#                        fplotmsHasCommands += 1
                elif (showSkippingMessages):
                    println(lineno(),"%sSkipping plotms of manually calibrated %s vs time" % (skipIndent,ptype))
                # now loop over individual antennas
                if (plotlevel > 1 and ((ptype == 'amp' and os.path.exists(manualCalibratedDataPlotsAmpPdf) == False)
                                    or (ptype == 'phase' and os.path.exists(manualCalibratedDataPlotsPhasePdf)==False))):
                    for antname in antennaNames:
                        figfile = manualqa2dir_extra + 'calibrated_data_%s_spw%s_%s.png'%(ptype,spw,antname)
                        if (os.path.exists(figfile) == False):
                            runPlotms("plotms("+plotmsparms+",vis='%s', antenna='%s', xaxis='time', yaxis='%s', coloraxis='field', spw='%s', ydatacolumn='corrected', avgbaseline=True, title='%s spw %s (manual)', plotfile='%s', correlation='%s')" % (msSplitWithPath,antname,ptype,spw,antname,spw,figfile,correlation))
# is this supposed to be msWithPath as Todd originally coded it (which then 
# needs unsplitspw)  or msSplitWithPath?
                        else:
                            println(lineno(),"Skipping plotms of %s %s %s for calibrated data" % (antname, spw, ptype))
                        manualCalibratedDataPlotsPerAntenna[ptype].append(figfile)
                        # If any of the component plots were missing, then remove the old montage result.
                        if (ptype == 'phase'):
                            if (os.path.exists(manualCalibratedDataPlotsPhasePdf)):
                                os.system('rm -f '+manualCalibratedDataPlotsPhasePdf)
                        else:
                            if (os.path.exists(manualCalibratedDataPlotsAmpPdf)):
                                os.system('rm -f '+manualCalibratedDataPlotsAmpPdf)
            # end 'for' loop over spws                    
            manualCalibratedDataPlotsAmp = manualqa2dir_extra + 'calibrated_data_amp.png'
            manualCalibratedDataPlotsPhase = manualqa2dir_extra + 'calibrated_data_phase.png'
            if (os.path.exists(manualCalibratedDataPlotsAmp) == False):
                tile = picktile(pnglist['amp'])
                os.system("montage -tile %s -geometry 1200x1200+0+0 %s %s" % (tile, pnglist['amp'], manualCalibratedDataPlotsAmp))
            if (os.path.exists(manualCalibratedDataPlotsPhase) == False):
                tile = picktile(pnglist['phase'])
                os.system("montage -tile %s -geometry 1200x1200+0+0 %s %s" % (tile, pnglist['phase'], manualCalibratedDataPlotsPhase))
            if (os.path.exists(manualCalibratedDataPlotsAmpPdf) == False and plotlevel > 1):
                # Now montage them into manualPhasecalSpectraAmpPlotsPdf
                println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (manualCalibratedDataPlotsPerAntenna['amp'],manualCalibratedDataPlotsAmpPdf))
                au.buildPdfFromPngs(manualCalibratedDataPlotsPerAntenna['amp'], pdfname=manualCalibratedDataPlotsAmpPdf,quiet=auQuiet)
            if (os.path.exists(manualCalibratedDataPlotsPhasePdf) == False and plotlevel > 1):
                println(lineno(),"Running au.buildPdfFromPngs(pnglist='%s', pdfname='%s')\n\n" % (manualCalibratedDataPlotsPerAntenna['phase'],manualCalibratedDataPlotsPhasePdf))
                au.buildPdfFromPngs(manualCalibratedDataPlotsPerAntenna['phase'], pdfname=manualCalibratedDataPlotsPhasePdf,quiet=auQuiet)
      
            if (fluxcal_source_id >= 0 and fluxcal_source_id != ''):
                manualCalibratedFluxCalPlotsAmp = manualqa2dir_extra + 'calibrated_data_fluxcal_amp.png'
                if (os.path.exists(manualCalibratedFluxCalPlotsAmp) == False):
                    pnglist = ''
                    for spwinteger in range(len(scienceSpwsIntegerList)):
                        spw = str(spwinteger)
                        figfile = manualqa2dir_extra + 'calibrated_data_fluxcal_amp_spw%s.png'%(spw)
                        pnglist += ' %s' % figfile
                        if (os.path.exists(figfile) == False):
                            overwrite = True
                            interactive = False
                            avgchannel = ''
                            avgtime = '1e9'
                            avgbaseline = True
                            correlation = ''
                            correctedfigfile = manualqa2dir_extra + 'calibrated_data_fluxcal_amp_spw%s_corrected.png'%(spw)
                            runPlotms("plotms("+plotmsparms+",vis='%s', xaxis='freq', yaxis='amp', coloraxis='corr',title='Calibrated amp vs. freq (spw %s)', avgchannel='%s', spw='%s', ydatacolumn='corrected', plotfile='%s', correlation='%s', field='%d', avgbaseline=%s, uvrange='%s', avgtime='%s')" % (msSplitWithPath, spw, avgchannel, spw, correctedfigfile, correlation,fluxcal_source_id,avgbaseline,fluxcal_uvrange,avgtime))
                            modelfigfile = manualqa2dir_extra + 'calibrated_data_fluxcal_amp_spw%s_model.png'%(spw)
                            runPlotms("plotms("+plotmsparms+",vis='%s', xaxis='freq', yaxis='amp', coloraxis='corr',title='Calibrated amp vs. freq (spw %s)', avgchannel='%s', spw='%s', ydatacolumn='model', plotfile='%s',  correlation='%s', field='%d', avgbaseline=%s, uvrange='%s', avgtime='%s')" % (msSplitWithPath, spw, avgchannel, spw, modelfigfile, correlation,fluxcal_source_id,avgbaseline,fluxcal_uvrange,avgtime))
                            command = "montage -geometry 1000x1000+1+1 %s %s %s" % (correctedfigfile, modelfigfile, figfile)
                            println(lineno(),"%sRunning: %s" % (skipIndent,command))
                            os.system(command)
                            # If any of the component plots were missing, then remove the old montage result.
                            if (os.path.exists(manualCalibratedFluxCalPlotsAmp)):
                                os.system('rm -f '+manualCalibratedFluxCalPlotsAmp)
                        elif (showSkippingMessages):
                            println(lineno(),"%sSkipping plotms of manually calibrated flux cal amp vs time (spw%s)" % (skipIndent,spw))
                    tile = picktile(pnglist)
                    command = "montage -tile %s -geometry 1000x1000+0+0 %s %s" % (tile, pnglist, manualCalibratedFluxCalPlotsAmp)
                    println(lineno(),"%sRunning: %s" % (skipIndent,command))
                    os.system(command)



























    print
    print
    print 40*"="
    print "building webpage"
    print

    htmlfile = ms.split("/")[-1]+".html"
    webfile = webpagedir + htmlfile
#    if (webfile.find('safe.nrao.edu') >= 0):
#        webfile2 = webfile[webfile.find('content/alma')+12:]
#        location = 'http://safe.nrao.edu/alma/%s' % (webfile2)
#    else:
#        if (webfile[0] == '/'):
#            location = 'file://%s' % (webfile)
#        else:
#            location = 'file://%s/%s' % (os.getcwd(), webfile)
    if (webpagedir.find('safe.nrao.edu') >= 0):
        webpagedir2 = webpagedir[webpagedir.find('content/alma')+12:]
        weblocation = 'http://safe.nrao.edu/alma/%s' % (webpagedir2)
    else:
        if (webpagedir[0] == '/'):
            weblocation = 'file://%s' % (webpagedir)
        else:
            weblocation = 'file://%s/%s' % (os.getcwd(), webpagedir)
    location=weblocation+htmlfile


    ############################################################
    # Create header for the webpage of the results
    ############################################################
    if (showingPipelineQA2):
        ppr = findPPR(pipecaldir, pipelineCasaLogfile)
        directoriesAndVersions +=  'OUS = %s<br>\n' % (au.getOUS(asdm))
        directoriesAndVersions +=  'Pipeline reduction directory = %s<br>\n' % (pipecaldir)
        directoriesAndVersions += 'Pipeline reduction used %s' % (pipelineCasaVersion)
        if (ppr == None):
            directoriesAndVersions += ' (PPR not found!)<br>\n'
            ppr = "(PPR not found!)"
            if debug: pdb.set_trace()
        else:
            directoriesAndVersions += ' and this PPR:<br>&nbsp;&nbsp;&nbsp;<a href="%s">%s</a><br>\n' % (ppr,ppr)
            myCopyfile(pipecaldir+ppr, webpagedir+ppr)
        if (ccl != None):
            cclNoPathNoTxt = os.path.basename(ccl)
            cclNoPath = cclNoPathNoTxt + '.txt' # add .txt so that firefox will show it
            println(lineno(),"Calling unwrapCasaCommandLog('%s',newlog='%s.txt',strippath=True)"%(ccl,ccl))
            unwrapCasaCommandLog(ccl, newlog=ccl+'.txt', strippath=True,temppath=pipeqa2dir)
            myCopyfile(ccl, webpagedir+cclNoPathNoTxt)
            myCopyfile(ccl+'.txt', webpagedir+cclNoPath)
        if (pipelineCasaLogfile != None):
            pipelineCasaLogfileNoPath = pipelineCasaLogfile.split('/')[-1]
            myCopyfile(pipelineCasaLogfile, webpagedir+pipelineCasaLogfileNoPath+'.txt')
        else:
            pipelineCasaLogfileNoPath = None
    if (showingManualQA2):
        directoriesAndVersions += 'Manual reduction directory = %s<br>\n' % (manualcaldir)
        directoriesAndVersions += 'Manual reduction used %s<br>\n' % (manualCasaVersion)
        
    auVersion = getCvsVersion(au.version())
    amcVersion = getCvsVersion(version())
    py0,py1,py2,py3,py4 = sys.version_info
#    if (terminalfileToOpen == terminalfile):
#        terminalfileString = 'Terminal output = <a href="%s.txt">%s.txt</a>' % (terminalfile, terminalfile)
#    else:
#        terminalfileString = 'Terminal output = <a href="%s.txt">%s.txt</a>, &nbsp; <a href="%s.txt">%s.txt</a><br>' % (terminalfile, terminalfile, terminalfileToOpen, terminalfileToOpen)
    terminalfileString = 'Terminal output = <a href="%s.txt">%s.txt</a><br>' % (weblocation+os.path.basename(terminalfileToOpen), terminalfileToOpen)

    webpage = """
    <html>
      <head>
         <title>ASDM = %s</title>
         <style type="text/css">
            html, td {font-size: 90%%;}
            .headline {font-size:200%%; font-weight: bold;}              
            table {border-collapse:collapse;}
            table, th, td {border: 1px solid black;}           
            td {padding:0.25em}
            td {vertical-align:top;}
            td[colspan] {background-color: #F3F3F3;}
            .innertable td, .innertable tr, .innertable {border: none; padding: 0; padding-right: 1em;}            
            b.passquality {color: #227722;}
            b.nopassquality {color: #772222;}
         </style>
      </head>
      <body>
      <div class="headline">ASDM = %s</div><br>
      <H3>Location of data and version information</H3>
      %s
      This report was generated by %s on %s at <b>%s</b> from CASA %s (python %d.%d.%d %s %d) and %s,<br>using %s.<br>
      PWD = %s<br>
      Command line = %s<br>
      %s<br>
      <table border=1>
    """ % (asdmNoPath, asdmNoPath, directoriesAndVersions, os.getenv('USER'), os.getenv('HOST'), str(datetime.datetime.now()), casadef.casa_version+' r'+casadef.subversion_revision, py0,py1,py2,py3,py4,auVersion,amcVersion, os.getcwd(), commandLine, terminalfileString)

    ###########################
    # add images to the webpage
    ###########################

    webpage += '<H3>Plots of the antenna configuration and observing schedule</H3>'
    if (showingPipelineQA2):
        outputfile = pipeqa2dir+ms+'.cfg'
        println(lineno(),"Running au.buildConfigurationFile('%s', output='%s', includeAntennaNames=True)" % (writeableMSWithPathPipeline,outputfile))
        nAnts, totalAnts = au.buildConfigurationFile(writeableMSWithPathPipeline, output=outputfile, includeAntennaNames=True)
    else:
        outputfile = manualqa2dir_extra+ms+'.cfg'
        println(lineno(),"Running au.buildConfigurationFile('%s', output='%s', includeAntennaNames=True)" % (msSplitWithPath,outputfile))
        nAnts, totalAnts = au.buildConfigurationFile(msSplitWithPath, output=outputfile, includeAntennaNames=True)
    myCopyfile(outputfile, webpagedir+ms+'.cfg.txt')
    webpage += '<a href="%s.cfg.txt">configuration file</a>' % (ms)
    if (nAnts < totalAnts):
        webpage += ' (%d of %d antennas written)' % (nAnts,totalAnts)
    webpage += '<BR>'
    
    maxBaseline = au.getBaselineStats(writeableMSWithPathPipeline)[2]
    println(lineno(),"Max baseline length = %.2f m" % (maxBaseline))
    if (showingManualQA2):
        if (False):
            # use the antenna config plot from the manual reduction
            antennaConfigPlot = manualqa2dir+'antenna_config.png'
            if (os.path.exists(antennaConfigPlot)):
                webpage += addplot(webpagedir, antennaConfigPlot, debug=debug)
            elif (showingPipelineQA2):
                antennaConfigPlot = pipeqa2dir+'antenna_config.png'
                if (os.path.exists(antennaConfigPlot) == False and use_plotants):
                    plotants(writeableMSWithPathManual, figfile=antennaConfigPlot)
                if (os.path.exists(antennaConfigPlot) == False):
                    es.listobs2(ms2=writeableMSWithPathManual, qa2_output_dir=manualqa2dir_extra, plotAntennasActualSize=True)
                if (os.path.exists(antennaConfigPlot)):
                    webpage += addplot(webpagedir, antennaConfigPlot, debug=debug)
                else:
                    webpage += "missing plot: %s<BR>" % (antennaConfigPlot)
            else:
                webpage += "missing plot: %s<BR>" % (antennaConfigPlot)
        else:
            antennaConfigPlot = manualqa2dir_extra+'antenna_config.png'
            antennaConfigZoomPlot = manualqa2dir_extra+'antenna_config_zoom.png'
            if (os.path.exists(antennaConfigPlot) == False):
                println(lineno(),"WARNING: Did not find %s - will attempt to create" % (antennaConfigPlot))
                if (use_plotants):
                    println(lineno(),"Running plotants('%s', figfile='%s')" % (writeableMSWithPathManual,antennaConfigPlot))
                    plotants(writeableMSWithPathManual, figfile=antennaConfigPlot)
            if (os.path.exists(antennaConfigPlot)):
                println(lineno(),"Found antenna configuration plot: %s" % (antennaConfigPlot))
                webpage += addplot(webpagedir, antennaConfigPlot,debug=debug)
            else:
                println(lineno(),"Running es.listobs2('%s', qa2_output_dir='%s', plotAntennasActualSize=True)" % (writeableMSWithPathManual,manualqa2dir_extra))
                es.listobs2(ms2=writeableMSWithPathManual, qa2_output_dir=manualqa2dir_extra, plotAntennasActualSize=True)
                if (os.path.exists(antennaConfigPlot)):
                    webpage += addplot(webpagedir, antennaConfigPlot,debug=debug)
                else:
                    webpage += "missing plot: %s<BR>" % (antennaConfigPlot)
            makeAntennaConfigZoomPlotIfNecessary(maxBaseline,
                      antennaConfigZoomPlot,writeableMSWithPathManual,manualqa2dir_extra)
            if (os.path.exists(antennaConfigZoomPlot)):
                webpage += addplot(webpagedir, antennaConfigZoomPlot,debug=debug)

        # use the observing schedule plot from the manual reduction
        obsDisplayPlot = manualqa2dir+'obs_display.png'
        if (os.path.exists(obsDisplayPlot)):
            webpage += addplot(webpagedir, obsDisplayPlot,debug=debug)
        elif (showingPipelineQA2):
            obsDisplayPlot = pipeqa2dir+'obs_display.png'
            if (os.path.exists(obsDisplayPlot) == False):
                es.listobs3(msName=writeableMSWithPathManual, figfile=obsDisplayPlot)
            if (os.path.exists(obsDisplayPlot)):
                webpage += addplot(webpagedir, obsDisplayPlot,debug=debug)
            else:
                webpage += "missing plot: %s<BR>" % (obsDisplayPlot)
        else:
            webpage += "missing plot: %s<BR>" % (obsDisplayPlot)
    else:
        # If we are only plotting the pipeline result, then we need to
        # make both these plots.
        antennaConfigPlot = pipeqa2dir+'antenna_config.png'
        antennaConfigZoomPlot = pipeqa2dir+'antenna_config_zoom.png'
        if (os.path.exists(antennaConfigPlot) == False and use_plotants):
            println(lineno(),"Running plotants('%s', figfile='%s')" % (writeableMSWithPathPipeline,antennaConfigPlot))
            plotants(writeableMSWithPathPipeline, figfile=antennaConfigPlot)
        if (os.path.exists(antennaConfigPlot) == False):
            println(lineno(),"Running es.listobs2('%s', qa2_output_dir='%s', plotAntennasActualSize=True)" % (writeableMSWithPathPipeline,pipeqa2dir))
            es.listobs2(ms2=writeableMSWithPathPipeline, qa2_output_dir=pipeqa2dir, plotAntennasActualSize=True)
        if (os.path.exists(antennaConfigPlot)):
            webpage += addplot(webpagedir, antennaConfigPlot,debug=debug)
        else:
            webpage += "missing plot: %s<BR>" % (antennaConfigPlot)
        makeAntennaConfigZoomPlotIfNecessary(maxBaseline,
                  antennaConfigZoomPlot, writeableMSWithPathPipeline, pipeqa2dir)
        if (os.path.exists(antennaConfigZoomPlot)):
            webpage += addplot(webpagedir, antennaConfigZoomPlot,debug=debug)

        obsDisplayPlot = pipeqa2dir+'obs_display.png'
        if (os.path.exists(obsDisplayPlot) == False):
            println(lineno(),"Running es.listobs3(msName='%s', figfile='%s')" % (writeableMSWithPathPipeline,obsDisplayPlot))
            es.listobs3(msName=writeableMSWithPathPipeline, figfile=obsDisplayPlot)
        if (os.path.exists(obsDisplayPlot)):
            webpage += addplot(webpagedir, obsDisplayPlot,debug=debug)
        else:
            webpage += "missing plot: %s<BR>" % (obsDisplayPlot)

    webpage += '<H3>Plots of the weather conditions</H3>'
        
    if (os.path.exists(weatherplot)):
        webpage += addplot(webpagedir, weatherplot,debug=debug)
    else:
        webpage += "missing weather plot: %s" % (weatherplot)

    if (os.path.exists(pwvplot)):
        webpage += addplot(webpagedir, pwvplot, width=500, br='',debug=debug)
    elif (sevenMeterOnly == False):
        webpage += "missing PWV plot: %s" % (pwvplot)

    if (os.path.exists(mosaicplot)):
        webpage += addplot(webpagedir, mosaicplot, width=500,debug=debug)
    else:
        webpage += "missing mosaic plot: %s" % (mosaicplot)

    webpage += '<H3>Plots of pointing solutions</H3>'
    webpage += antwebpage
    for pointingPlot in pointingPlots:
        if (os.path.exists(pointingPlot)):
            if (pointingPlot == pointingPlots[-1]):
                br = True
            else:
                br = False
            webpage += addplot(webpagedir, pointingPlot, width=400, br=br,debug=debug)
        else:
            webpage += "missing pointing plot: %s" % (pointingPlot)

    webpage += '<H3>Plots of the sideband ratio measurement</H3>'

    if (os.path.exists(sbrplot)):
        webpage += addplot(webpagedir, sbrplot, width=500,debug=debug)
        if (sbr_suspect == ''):
            webpage += '<br>No suspect sideband ratio measurements. Criteria: < %.2f for 2SB, < %.2f for DSB<br>' % (SBR_THRESHOLD_2SB, SBR_THRESHOLD_DSB)
        else:
            sbr_suspect_file = asdmNoPath + '_sbr_suspect.txt'
            writeStringToTextFile(pipeqa2dir+sbr_suspect_file, sbr_suspect)
            myCopyfile(pipeqa2dir+sbr_suspect_file, webpagedir+sbr_suspect_file)
            webpage += '<br><a href="%s">%d suspect sideband ratio measurements</a>. Criteria: < %.2f for 2SB, < %.2f for DSB<br>' % (sbr_suspect_file,sbr_suspect.count('<br>'), SBR_THRESHOLD_2SB, SBR_THRESHOLD_DSB)
    else:
        webpage += "missing sideband ratio plot: %s" % (sbrplot)
    
    # Write the column titles of the web table
    webpage += """<tr> <th>Item</th> <th>Pipeline results</th> <th>Manual results</th> </tr>\n"""

    # Information:
    if (showingPipelineQA2):
        showNAntennas = 6
        if (len(pipeRefAnt)>(showNAntennas*5-1)):
            pipeRefAntStr=pipeRefAnt[0:(showNAntennas*5-1)]+",etc."
        else:
            pipeRefAntStr=pipeRefAnt
    else:
        pipeRefAntStr = 'n/a'

    webpage += "<tr> <td><b>listobs</td>"
    if (showingPipelineQA2):
        # need to generate a listobs
        listfile = "%s.listobs.pipeline" % (ms)
        if (os.path.exists(pipeqa2dir+listfile) == False):
            listobs(writeableMSWithPathPipeline, listfile=pipeqa2dir+listfile, verbose=True)
        if (casadef.casa_version >= '4.1.0'):
            webpage += '<td><a href="%s">%s</a> &nbsp; narrowest spw = %.0f MHz</td>' % (listfile+'.txt', listfile, narrowestSpwBW*1e-6)
        else:
            webpage += '<td><a href="%s">%s</a></td>' % (listfile+'.txt', listfile)
        myCopyfile(pipeqa2dir+listfile, webpagedir+listfile+'.txt')
    else:
        webpage += "<td></td>"
    if (showingManualQA2):
        listfile = '%s.listobs' % (ms)
        listfile2 = '%s.split.listobs' % (ms)
        webpage += '<td><a href="%s">listobs</a> &nbsp; <a href="%s">split.listobs</a></td>' % (listfile+'.txt', listfile2+'.txt')
        myCopyfile(manualcaldir+listfile, webpagedir+listfile+'.txt')
        myCopyfile(manualcaldir+listfile2, webpagedir+listfile2+'.txt')
    else:
        webpage += "<td></td>"
    
    webpage += "<tr> <td><b>listobs (only FULL_RES spws)</td>"
    if (showingPipelineQA2):
        # need to generate a listobs
        listfile = "%s.listobs.full_res" % (ms)
        if (os.path.exists(pipeqa2dir+listfile) == False):
            listobs(writeableMSWithPathPipeline, listfile=pipeqa2dir+listfile, verbose=True, spw='*FULL_RES*')
        webpage += '<td><a href="%s">%s</a></td>' % (listfile+'.txt', listfile)
        myCopyfile(pipeqa2dir+listfile, webpagedir+listfile+'.txt')
    else:
        webpage += "<td></td>"
    webpage += "<td></td>"
    
    webpage += "<tr> <td><b>Command log or script</td>"
    if (showingPipelineQA2):
        if (ccl != None):
            webpage += '<td><a href="%s">casa_commands.log</a> from %s</td>' % (cclNoPath,pipelineYYYYMMDD)
        else:
            if debug:
                print "ERROR: can't find casa_commands.log"
                pdb.set_trace()
            webpage += "<td>not found</td>"
    else:
            webpage += "<td></td>"
    if (showingManualQA2):
        if (sfc != None):
            webpage += '<td><a href="%s">%s</a></td>' % (sfcDotTxt, sfc)
        else:
            if debug:
                print "ERROR: can't find casa_commands.log"
                pdb.set_trace()
            webpage += "<td>not found</td>"
    else:
        webpage += "<td></td>"
    webpage += "</tr><br>\n"
        
    webpage += "<tr> <td><b>CASA log</td>"
    if (showingPipelineQA2):
        if (pipelineCasaLogfileNoPath == None):
            webpage += '<td>not found</a></td>'
            if debug:
                print "ERROR: can't find casa log"
                pdb.set_trace()
        else:
            webpage += '<td><a href="%s.txt">%s</a></td>' % (pipelineCasaLogfileNoPath, pipelineCasaLogfileNoPath)
    else:
        webpage += "<td></td>"
    if (showingManualQA2):
        webpage += "<td>copy not attempted</td>"
    else:
        webpage += "<td></td>"
    webpage += "</tr><br>\n"

    webpage += "<tr> <td><b>Deterministic<br>flag command file</td>"
    if (showingPipelineQA2):
        flagcmdfileNoPath = asdmNoPath + '_flagcmds.inp'
        flagcmdfile = pipecaldir + flagcmdfileNoPath
        if (os.path.exists(flagcmdfile) == False):
            # name changed in May 2013 release
            flagcmdfileNoPath = asdmNoPath + '_flagcmds.txt'
            flagcmdfile = pipecaldir + flagcmdfileNoPath
        if (os.path.exists(flagcmdfile)):
            myCopyfile(flagcmdfile, webpagedir+flagcmdfileNoPath)
            # Put first N lines there that start with 'mode'
            flagfile = open(flagcmdfile,'r')
            lines = flagfile.readlines()
            webpage += '<td>'
            for line in lines:
                if (line.find('mode=') < 0):
                    break
                webpage += line + '<br>'
            flagfile.close()
            webpage += 'complete file: <a href="%s">%s</a>' % (flagcmdfileNoPath, flagcmdfileNoPath)
            if (flagcmdfile.find('.txt') > 0):
                onlineflagcmdfileNoPath = asdmNoPath + '_flagonline.txt'
                onlineflagcmdfile = pipecaldir + onlineflagcmdfileNoPath
                if (os.path.exists(onlineflagcmdfile)):
                    ofc = open(onlineflagcmdfile,'r')
                    numberOnlineFlags = len(ofc.readlines())
                    ofc.close()
                    myCopyfile(onlineflagcmdfile, webpagedir+onlineflagcmdfileNoPath)
                    webpage += '<br>%d online flags: <a href="%s">%s</a></td>' % (numberOnlineFlags,onlineflagcmdfileNoPath, onlineflagcmdfileNoPath)
                else:
                    println(lineno(),"ERROR: Did not find %s" % onlineflagcmdfile)
                    webpage += "</td>"
            else:
                println(lineno(),"This is pre-May 2013 heuristics")
                webpage += "</td>"
        else:
            webpage += "<td>No _flagcmds.inp nor .txt file found<br>(AMC will only find the default named file)</td>"
    else:
        webpage += "<td></td>"
    if (showingManualQA2):
        webpage += "<td>n/a</td>"
    else:
        webpage += "<td></td>"
    webpage += "</tr><br>\n"

    if (showingPipelineQA2):
        flagplot = pipeqa2dir + asdmNoPath + '.flags.png'
        plotForColumn1 = flagplot
        plotForColumn2 = ''
    else:
        flagplot = manualqa2dir_extra + asdmNoPath + '.flags.png'
        plotForColumn1 = ''
        plotForColumn2 = flagplot
    textForColumn1 = ''
    if (os.path.exists(flagplot) == False):
        println(lineno(), "Running flagcmd('%s', inpfile='%s', action='plot', inpmode='xml', plotfile='%s')" % (writeableMSWithPathPipeline, asdm, flagplot))
        flagcmd(writeableMSWithPathPipeline, inpfile=asdm, action='plot', inpmode='xml', plotfile=flagplot)
        if (os.path.exists(flagplot) == False):
            plotForColumn1 = ''
            if (os.access(writeableMSWithPathPipeline,os.W_OK) == False):
                textForColumn1 = 'no write permission to ms'
                println(lineno(),"ERROR: flagcmd failed due to lack of write permission on ms")
            else:
                println(lineno(),"ERROR: flagcmd failed, but not due to lack of write permission on ms")
    else:
        println(lineno(), "Skipping flagcmd plot because it already exists")
    webpage += addplotrow(webpagedir, title='Plot of online flags', 
                          plotForColumn1=plotForColumn1,
                          plotForColumn2=plotForColumn2,
                          textForColumn1=textForColumn1,debug=debug)
    
    if (showFlagStats):
      webpage += "<tr> <td><b>Percentage of data<br>flagged (from the af tool)</td>"
      if (showingPipelineQA2):
        print "Computing flag statistics in pipeline reduction dataset"
        webpage += getFlagStatistics(writeableMSWithPathPipeline, scienceSpws, pipeqa2dir, cleanup, debug=debug)
      else:
        webpage += "<td></td>"
        
      if (showingManualQA2):
        print "Computing flag statistics in manual reduction dataset"
        webpage += getFlagStatistics(writeableMSSplitWithPath, qa2dir=manualqa2dir_extra,cleanup=cleanup,debug=debug) # it only contains science spws
      else:
        webpage += "<td></td>"
        
    webpage += "<tr> <td><b>Reference Antenna</td>"
    if (showingPipelineQA2):
        webpage += "<td>pipeline: "+pipeRefAntStr+"<br>AMC: "+str(amcRefAnt)+"</td>"
    else:
        webpage += "<td></td><br>\n"
    if (showingManualQA2):
        webpage += "<td> aU.getRefAntenna: "+es.getRefAntenna(msSplitWithPath)+"<br>"
        webpage += "manual script: "+manualRefAnt+"<br>AMC: "+str(amcRefAnt)+"</td> </tr>\n"
    else:
        webpage += "<td></td><br>\n"

    if (showingPipelineQA2):
        webpage += "<tr><td><b>Web log</td>"
        webpage += '<td><a href="html/t1-1.html">html</a></td>\n'
        if (showingManualQA2):
            webpage += "<td>n/a</td><br>\n"
        else:
            webpage += "<td></td><br>\n"
        println(lineno(),"rsync weblog to %s" % (webpagedir))
        os.system('rsync -vau %s %s %s' % (rsyncPermissions, pipecaldir+pipelineYYYYMMDD+'/html', webpagedir))

    # Tsys
    manualTsysPlot = ''
    manualTsysPdf = ''
    if (showingManualQA2):
        figfile = manualqa2dir_extra + manualcaltableNoPath
        manualTsysPlot = findOverlayTimeAntennaPlot(manualqa2dir_extra, manualcaltableNoPath)
        println(lineno(),"       picked manualTsysPlot = '%s'" % manualTsysPlot)
        if (manualTsysPlot == ''):
            # Then we need to build this plot, as it was not generated in the earlier versions of the manual script.
            chanrange = getTsysChanRangeToPlot(manualcaldir+ms)
            if ('task_plotbandpass' in sys.modules and force_au_plotbandpass==False and casadef.casa_version >= pbp_version):
                println(lineno(),"Running plotbandpass('%s', overlay='antenna,time', xaxis='freq', yaxis='amp', subplot=22, buildpdf=False, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=True, vis='%s', field='', figfile='%s')" % (manualcaltable, chanrange, manualcaldir+ms, figfile))
                plotbandpass(manualcaltable, overlay='antenna,time', xaxis='freq', yaxis='amp', 
                             subplot=22, buildpdf=False, interactive=False, showatm=True, pwv='auto', 
                             chanrange=chanrange, showfdm=True, vis=manualcaldir+ms, field='', 
                             figfile=figfile, solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)
            else:
                println(lineno(),"Running au.plotbandpass('%s', overlay='antenna,time', xaxis='freq', yaxis='amp', subplot=22, buildpdf=False, interactive=False, showatm=True, pwv='auto', chanrange='%s', showfdm=True, ms='%s', field='', figfile='%s')" % (manualcaltable, chanrange, manualcaldir+ms, figfile))
                au.plotbandpass(manualcaltable, overlay='antenna,time', xaxis='freq', yaxis='amp', 
                                subplot=22, buildpdf=False, interactive=False, showatm=True, pwv='auto', 
                                chanrange=chanrange, showfdm=True, ms=manualcaldir+ms, field='', 
                                figfile=figfile, solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)
                
            manualTsysPlot = findOverlayTimeAntennaPlot(manualqa2dir_extra, manualcaltableNoPath)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping au.plotbandpass overlay='antenna,time' for manual reduction" % (skipIndent))

        pnglist = buildPngList(manualcaldir+ms+'.tsys.plots.overlayTime/')
        manualTsysPdf = manualqa2dir_extra + manualcaltableNoPath + '.pdf'
        println(lineno(),"      using manualTsysPdf = %s" % manualTsysPdf)
        if (os.path.exists(manualTsysPdf) == False):
            bPFP = au.buildPdfFromPngs(pnglist=pnglist, pdfname=manualTsysPdf,quiet=auQuiet)
            println(lineno(),bPFP)

    if (showingPipelineQA2 == False):
        pipelineTsysPlot = ''
        pipelineTsysPdf = ''
        pipelineTsysPlotIncludingFlaggedSolutions = ''
        pipelineTsysPdfIncludingFlaggedSolutions = ''

    textForColumn2 = ''
    if (showingManualQA2):
        if os.path.exists(calScript[0]):
            f = open(calScript[0])
            script = f.readlines()
            f.close()
            textForColumn2 = 'tsysChanTol=default'
            for line in script:
                tsysChanTolLoc = line.find('tsysChanTol=')
                if (tsysChanTolLoc >= 0):
                    textForColumn2 = line[tsysChanTolLoc:].strip('\r\n').strip(')')
                    break
        else:
            textForColumn2 = "cannot find manual script"
        
    println(lineno(),"Adding row for Tsys plots prior to flagging")
    webpage += addplotrow(webpagedir, 
                          title="Tsys before flagging<br> with overlay=<br>'antenna,time'<br><br></b>(click image for<br> multi-page PDF of<br> overlay='time')",
                          plotForColumn1=pipelineTsysPlotIncludingFlaggedSolutions,
                          plotForColumn2=manualTsysPlot,
                          pdf1=pipelineTsysPdfIncludingFlaggedSolutions,
                          pdf2=manualTsysPdf,
                          textForColumn2=textForColumn2,debug=debug)

    println(lineno(),"Adding row for Tsys plots after flagging")
    webpage += addplotrow(webpagedir, 
                          title="Tsys after flagging<br> with overlay=<br>'antenna,time'<br><br></b>(click image for<br>multi-page PDF of<br> overlay='time')",
                          plotForColumn1=pipelineTsysPlot,
                          plotForColumn2='',
                          pdf1=pipelineTsysPdf,
                          pdf2='',debug=debug)


    # Tsys flagging:
    println(lineno(),"Adding row for Tsys flagging")
    webpage += "<tr><td><b>Tsys flagging: </td> <td>"
    
    # grep the log for whether it flagged
    pipelog = glob.glob(pipecaldir+"casa.pipelinerun.log")
    if (len(pipelog) < 1):
        pipecaldirpaginated = pipecaldir
        charlimit = 50
        if (len(pipecaldirpaginated) > charlimit):
            pipecaldirpaginated = pipecaldirpaginated[0:charlimit] + '<br>' + pipecaldirpaginated[charlimit:]
            
#        webpage += "could not find casa.pipelinerun.log in<br>%s<br>" % (pipecaldir)
    elif os.path.exists(pipelog[0]):
        inLog = commands.getoutput("grep viewflaggers "+pipelog[0])
        webpage += inLog+"<br>"
#    else:
#        webpage += "could not find casa.pipelinerun.log in %s<br>" % (pipecaldir)
    if (showingPipelineQA2):
        # open the caltable and attempt to figure out what was flagged
        tsysFlags, tsysFlagsText = printTsysFlags(pipecaldir=pipecaldir, ms=writeableMSWithPathPipeline, edge=edgeChannels, caltable=pipecaldir + findTsysTable(pipecaldir, ms) )
        if (tsysFlags == ''):
            tsysFlags = "No flags detected!"
        webpage += tsysFlags

    webpage += "</td> <td>"
    if (showingManualQA2):
        webpage += "Manual Flagging in script:<br>"
        if os.path.exists(calScript[0]):
            script=open(calScript[0])
            flagline=False
            ct=5
            for line in script:
                if line[0:14]=="# Initial flag": 
                    flagline=True
                if line[0:6]=="mystep" and ct<=0: 
                    flagline=False
                if flagline:
                    webpage += line+"<br>\n"
                    ct=ct-1                
        else:
            webpage += "cannot find manual script"
    webpage += "</td></tr>"


    # WVR
    if (sevenMeterOnly):
        webpage += "<tr><td><b>no WVRs on 7m antennas</td>"
    else:
        webpage += "<tr><td><b>WVR tie</td>"
        if (showingPipelineQA2):
            pipelineTieParameter, pipelineWVRFlagParameter = findPipelineTieParameter(ccl, asdmNoPath)
            webpage += "<td>%s</td>" % (pipelineTieParameter)
        else:
            webpage += "<td></td>"
        if (showingManualQA2):
            manualTieParameter = findManualTieParameter(manualcaldir+sfc)
            webpage += "<td>%s</td>" % (manualTieParameter)
        else:
            webpage += "<td></td>"
    
        webpage += "<tr><td><b>WVR flag</td>"
        if (showingPipelineQA2):
            webpage += "<td>%s</td>" % (pipelineWVRFlagParameter)
        else:
            webpage += "<td></td>"
        if (showingManualQA2):
            manualWVRFlagParameter =  findManualWVRFlagParameter(manualcaldir+sfc)
            webpage += "<td>%s</td>" % (manualWVRFlagParameter)
        else:
            webpage += "<td></td>"
        

        manualplot = ''
        manualWvrPdf = ''
        wvr_gaintable_manual = ''
        showPipelineThumbnailAutoscale = False
        if (showingManualQA2):
            wvr_gaintable_manual = msWithPathManual + '.wvr'       
            manualplot = manualqa2dir + 'wvr_plot_fixedscale.png'
            if (os.path.exists(manualplot) == False): # or match_man_refant==True):
                manualplot = manualqa2dir_extra + 'wvr_plot.png'
                if (os.path.exists(manualplot) == False): # or match_man_refant==True):
                    ### TEMP FIX pending Moellenbrock - need write permission
                    # to wvr_gaintable for some reason even though its only 
                    # used as gaintable in gaincal
                    if not os.access(wvr_gaintable_manual,os.W_OK):
                        tblcopy=manualqa2dir_extra+"/"+os.path.basename(wvr_gaintable_manual)                        
                        if tblcopy[-1]=="/": tblcopy=tblcopy[:-1]
                        myCopyfile(wvr_gaintable_manual+"/",tblcopy) # needs trailing slash on source, NOT on destination
                        wvr_gaintable_manual=tblcopy
                    cmd="es.wvr_stat('%s', refAnt='%s', qa2_output_dir='%s', wvr_gaintable='%s', scanintent='%s', spw='%s', pol='%s', nsigma_coherence=%f, unwrap_phase=%s, autoscale_yaxis=%s)" % (writeableMSWithPathManual,amcRefAnt0,manualqa2dir_extra,wvr_gaintable_manual,wvrstat_intent,wvrstat_spw,wvrstat_pol, nsigma_coherence, unwrap_phase, str(autoscale_yaxis))
                    print
                    println(lineno(),"manual: Calling "+cmd)
                    try:
                        exec cmd
                    except:
                        msg="The gaincal failed in es.wvr_stat(spw='%s')." % wvrstat_spw
                        println(lineno(), "\n ERROR: "+msg)
                        manualWvrText =msg+"<br>\n"+printWvrStatErrorMessage(writeableMSWithPathManual,lineno())
            else:
                showPipelineThumbnailAutoscale = True

            manualWvrPdf = manualqa2dir_extra + os.path.basename(wvr_gaintable_manual) + '.pdf'
            if (os.path.exists(manualWvrPdf) == False): # or match_man_refant==True):
                cmd="wvr_stat_all('%s', refAnt='%s', qa2_output_dir='%s', wvr_gaintable='%s', scanintent='%s', spw='%s', pol='%s', nsigma_coherence=%f)" % (writeableMSWithPathManual,amcRefAnt0,manualqa2dir_extra,wvr_gaintable_manual,wvrstat_intent,scienceSpwsIntegerList,allpols,nsigma_coherence)
                println(lineno(),"manual: Calling "+cmd)
                try:
                    wvr_stat_all(writeableMSWithPathManual, refAnt=amcRefAnt0, qa2_output_dir=manualqa2dir_extra, wvr_gaintable=wvr_gaintable_manual, scanintent=wvrstat_intent,spw=scienceSpwsIntegerList, pol=allpols, nsigma_coherence=nsigma_coherence)                    
                except:
                    msg="At least one spw's gaincal failed in wvr_stat_all(spw='%s')." % scienceSpwsIntegerList
                    println(lineno(), "\n ERROR: "+msg)
                    manualWvrText =msg+"<br>\n"+printWvrStatErrorMessage(writeableMSWithPathManual,lineno())
                    manualWvrPdf = ''
                    showPipelineThumbnailAutoscale = True

            if (showingPipelineQA2 and showPipelineThumbnailAutoscale):
                # Show the autoscale plot for the pipeline column
                pipelineWvrPlot = pipeqa2dir + 'wvr_plot_autoscale.png'
                if (wvr_refant == ''):
                    manualWvrstatRefAnt = getManualWvrstatRefAnt(manualqa2dir, refAnt0)
                else:
                    manualWvrstatRefAnt = wvr_refant # which in turn is equal to refAnt0
                if (os.path.exists(pipelineWvrPlot) == False or wvrstat_spw != '' or wvrstat_pol != ''
                    or wvr_refant != '' or manualWvrstatRefAnt != amcRefAnt0):
                    cmd="es.wvr_stat('%s', refAnt='%s', qa2_output_dir='%s', wvr_gaintable='%s', plotfile='%s')" % (writeableMSWithPathPipeline,manualWvrstatRefAnt,pipeqa2dir,wvr_gaintable,'wvr_plot_autoscale.png')
                    println(lineno(),"pipeline: Calling "+cmd)
                    try:
                        exec cmd
                    except:
                        msg="The gaincal failed in es.wvr_stat(pipeline_autoscale)." 
                        println(lineno(), "\n ERROR: "+msg)
                        pipelineWvrText += "\n"+msg+"<br>\n"+printWvrStatErrorMessage(writeableMSWithPathPipeline,lineno())

                elif (showSkippingMessages):
                    println(lineno(),"%sSkipping es.wvr_stat(pipeline autoscale) since the result already exists in the qa2 subdirectory." % (skipIndent))
                  
        println(lineno(), "Calling addplotrow with pipelineWvrPdf=%s" % (pipelineWvrPdf))
        pipelineWvrText += os.path.basename(wvr_gaintable)
        manualWvrText += os.path.basename(wvr_gaintable_manual)
        webpage += addplotrow(webpagedir, title='Phase vs. time<br> on the bandpass calibrator<br> before and after WVR corrections<br></b><br><br>(Click for PDF with fixed y-scale)<br><br>Note: antennas<br>showing only zeros<br> are either:<br>1) the refant<br>2) flagged for Tsys<br>3) flagged by applycal', 
                              plotForColumn1=pipelineWvrPlot,
                              plotForColumn2=manualplot,
                              textForColumn1=pipelineWvrText,
                              textForColumn2=manualWvrText,
                              pdf1=pipelineWvrPdf,
                              pdf2=manualWvrPdf,debug=debug)
        if (pipelineWvrPdf2 != ''):
            println(lineno(), "Calling addplotrow with pipelineWvrPdf2=%s" % (pipelineWvrPdf2))
            webpage += '<tr><td><b>Phase vs. time<br> on the phase calibrator<br> before and after WVR corrections</td><td><a href="%s">PDF for phase calibrator</a></td><td></td></tr>' % (os.path.basename(pipelineWvrPdf2))
            myCopyfile(pipelineWvrPdf2, webpagedir+os.path.basename(pipelineWvrPdf2))
    
        # wvrgcal information table
        webpage += "<tr><td><b>wvrgcal information</td><td>"
        webpage += '<FONT FACE="courier">'
        if (showingPipelineQA2):
            if not pipelineCasaLogfile==None:
                wvrfile = open(pipeqa2dir+"/"+ms+".wvr_output.txt", 'w')
                println(lineno(),"Running extractWvrgcalOutputFromLog('%s','%s',debug=debug,nsigma_discrepancy=%f)" % (pipelineCasaLogfile, asdmNoPath, nsigma_discrepancy))
                wvrgcalOutput,outlierReport = extractWvrgcalOutputFromLog(pipelineCasaLogfile, asdmNoPath,
                                        debug=debug, nsigma_discrepancy=nsigma_discrepancy)
                for line in wvrgcalOutput:
 #                    webpage += line.split('casa\t')[1] + '<br>'
                    wvrfile.write(line.split('casa\t')[1])
                wvrfile.close()
                webpage += '<a href='+ms+'.wvr_output.txt'+'>wvrgcal output</a><br>'
                myCopyfile(pipeqa2dir+"/"+ms+".wvr_output.txt", webpagedir+ms+".wvr_output.txt")
                webpage += outlierReport
            else:
                webpage += "no pipeline casalog"
        webpage += "</td>"
        if (showingManualQA2):
            if (False):
                wvrgcalOutput,outlierReport = extractWvrgcalOutputFromLog(manualCasaLogfile, asdmNoPath,
                                  debug=debug, nsigma_discrepancy=nsigma_discrepancy)
                webpage += "<td>"
                for line in wvrgcalOutput:
#                    println(lineno(), line)
                    webpage += line.split('casa\t')[1] + '<br>'
                webpage += outlierReport
                webpage += "</td>"
            else:
                webpage += "<td>copy not attempted</td></tr>"
        else:
            webpage += "<td></td></tr>"
            
        # WVR statistics
        #rerun wvrgcal to regenerate PL overwritten caltable w/out wvrflagging
        #grab the original command from the casa command log
        cclF = open(ccl, 'r')
        inaWVRgcal = False
        wvrCommand = ''
        for line in cclF:
            if not inaWVRgcal:
                if line.find('wvrgcal(') >= 0:
                    inaWVRgcal = True
            if inaWVRgcal:
                wvrCommand += line.strip()
                if line.find(')') >= 0:
                    #check we have the correct MS if more than one in PL execution
                    if wvrCommand.find(asdmNoPath) < 0:
                        wvrCommand = ''
                        inaWVRgcal = False
                        continue
                    #check we have the call with NO wvrflagging
                    if wvrCommand.split('wvrflag=')[1].split(']')[0] != '[':
                        wvrCommand = ''
                        inaWVRgcal = False
                        continue
                    break
        cclF.close()

        #fully qualify some paths and prevent overwriting
        wvrCommand = wvrCommand.replace("vis='" + asdmNoPath + '.ms', \
                                        "vis='" + \
                                        writeableMSWithPathPipeline)
        cTbl = wvrCommand.split('caltable=')[1].split("'")[1]
        newCTbl = cTbl.split('.tbl')[0]
        newCTbl += '_noflag.tbl'
        newCTbl = pipeqa2dir + newCTbl
        wvrCommand = wvrCommand.replace("caltable='" + cTbl + "'", "caltable='" + \
                                      newCTbl + "'")

        #execute wvrgcal command and then run wvr_stat_all on the resulting table
        if os.path.exists(newCTbl):
            println(lineno(), newCTbl+" exists - skipping wvrgcal command "+wvrCommand)
        else:
            println(lineno(), 'Running %s' % (wvrCommand))
            clearstat()
            exec wvrCommand
        wvrStatCmd = "wvr_stat_all('%s', refAnt='%s', qa2_output_dir='%s', wvr_gaintable='%s', scanintent='%s', spw=%s, pol=%s, nsigma_coherence=%f)" % (writeableMSWithPathPipeline, amcRefAnt0, pipeqa2dir, newCTbl, wvrstat_intent, scienceSpwsIntegerList, allpols, nsigma_coherence)
        println(lineno(), 'Running %s' % (wvrStatCmd))
        exec wvrStatCmd

        #make the webpage table from the non-flagged wvr_stat output
#        charlimit = 51
#        webpage += "<tr><td><b>WVR statistics</td><td>"
        charlimit = 101
        webpage += "<tr><td><b>WVR statistics</td><td colspan=2>"
        useCourierForWvrstat = True
        if (useCourierForWvrstat): webpage += '<FONT FACE="courier">'
        # first the pipeline
        if (showingPipelineQA2):
            script = glob.glob(pipeqa2dir + '*noflag*txt')[0]
            if script != []:
                script = open(script, 'r')
                for line in script:
                    if (line.find('wvrgcal output of average wvr') >= 0):
                        break
                    elif (line.find("*******************") >= 0):
                        continue
                    elif 'A' in line and 'B' in line and \
                         'C' in line and 'D' in line and \
                         'E' in line and 'F' in line and \
                         'G' in line and 'H' in line:
                        formattedLine = re.sub(' ', '&nbsp;', line)
                        webpage += formattedLine + "<br>\n"
                    elif (line.find('DA') == 0 or \
                        line.find('DV') == 0 or \
                        line.find('PM') == 0 or \
                        line.find('CM') == 0):
                        tokens = line.split()
                        formattedLine = tokens[0] + ' %7.1f'%float(tokens[1]) + \
                                        '  %6.1f'%float(tokens[2]) + \
                                        '  %6.1f'%float(tokens[3]) +  \
                                        '  %4d'%int(tokens[4]) + \
                                        '  %3d'%int(tokens[5]) + \
                                        '  %3d'%int(tokens[6])
                        formattedLine = re.sub(' ', '&nbsp;', formattedLine)
                        webpage += formattedLine
                        for i in range(7, len(tokens)):
                            webpage += ' &nbsp;&nbsp;' + tokens[i]
                        webpage += "<br>\n"
                    else:
                        if (len(line) < charlimit):
                            webpage += line + "<br>\n"
                        else:
                            webpage += line[0:charlimit] + "<br>\n" + line[charlimit:] + '<br>\n'
                script.close()
            else:
                webpage += 'cannot find non-flagged pipeline wvr_stat.txt files<br>\n'
                if debug: pdb.set_trace()
        
    
        if (useCourierForWvrstat): webpage += '</font>'
#        webpage += '</td><td>'
        if (useCourierForWvrstat): webpage += '<FONT FACE="courier">'

        #mine the pipeline WVR statistics and save them to the web page dir
        if len(pipecaldir) > 0:
            gatherPipelineStats(asdmNoPath, pipecaldir, \
                                writeableMSWithPathPipeline, pipeqa2dir, ccl, \
                                webpagedir, sevenMeterOnly, \
                                pipelineCasaVersion, manCalDir=manualcaldir)
    
        # now the manual reduction
# RI removed in favor of the pre-flagging one that Nathan implemented above
# could consider also putting in the post-wvrflag stats but meh...
#        if (showingManualQA2):
#          if os.path.exists(manualqa2dir+'wvr_stat.txt'):
#            script=open(manualqa2dir+'wvr_stat.txt')
#            for line in script:
#                if (line.find("wvrgcal output of average wvr")>=0):
#                    break
#                if (line.find("*******************")>=0):
#                    continue
#                if (line.find("DA")==0 or line.find("DV")==0 or line.find("PM")==0 or line.find("CM")==0):
#                    tokens = line.split()
#                    formattedLine = tokens[0] + ' %6.1f'%float(tokens[1]) + ' %5.1f'%float(tokens[2]) + ' %5.1f'%float(tokens[3]) +  ' %3d'%int(tokens[4])  + ' %3d'%int(tokens[5])
#                    formattedLine = re.sub(' ','&nbsp;',formattedLine)
#                    webpage += formattedLine
#                    for i in range(6,len(tokens)):
#                        webpage += ' ' + tokens[i]
#                    webpage += "<br>\n"
#                else:
#                    if (len(line) < charlimit):
#                        webpage += line+"<br>\n"
#                    else:
#                        webpage += line[0:charlimit]+"<br>\n"+line[charlimit:]+"<br>\n"
#          else:
#              webpage += "cannot find manual reduction's wvr_stat.txt<br>\n"
#        if (useCourierForWvrstat): webpage += "</FONT>\n"
        webpage += "</td></tr>\n"
    
      
    if (showingPipelineQA2):
        webpage += "<tr><td><b>Flux densities<br>loaded from flux.csv</b></td>"
        webpage += '<td>Link = <a href="flux.csv.txt">flux.csv</a></br>'
        # now show the contents
        try:
            fluxcsv = open(pipecaldir+'flux.csv','r')
            lines = fluxcsv.readlines()
            allFields = au.getFields(writeableMSWithPathPipeline)
            webpage += "Contents with field names appended:\n"
            webpage += "<pre>\n"
            for line in lines:
                if (line.find('.ms')<0):
                    webpage += line
                else:
                    fieldID = int(line.split(',')[1])
                    try:
                        webpage += line.strip('\r\n') + ' # ' + allFields[fieldID] + '\r\n'
                    except:
                        webpage += line
            webpage += "</pre>\n"
        except:
            print "WARNING: Could not open flux.csv"
        webpage += "</td>"
        myCopyfile(pipecaldir+'flux.csv', webpagedir+'flux.csv.txt')
    else:
        # we must be showing the manual reduction
        webpage += "<tr><td><b>First setjy in script</b></td>"
        
    webpage += "<td>"
    if (showingManualQA2):
        if os.path.exists(calScript[0]):
            script = open(calScript[0])
            getlines = False
            sawfield = False
            mystep = -1
            for line in script:
                if (line.find('mystep') == 0):
                    try:
                        mystep = int(line.split('=')[1])
                    except:
                        pass
                findsetjy = line.find('setjy')
                if (findsetjy >= 0 and line.find('(')>findsetjy):
                    if (showingPipelineQA2):
                        if (mystep < 0):
                            # then we did not find the step number
                            webpage += "First setjy in script:<br>"
                        else:
                            webpage += "First setjy in script (in step %d):<br>" % (mystep)
                    webpage += "<pre>"
                    webpage += line
                    getlines=True
                elif (getlines):
                    if (line.find('field')>=0):
                        sawfield = True
                    webpage += line
                    if (line.find(')')>=0 and sawfield):
                        # this must have been the last line of the command
                        break
            if (getlines):
                webpage += "</pre>"
        else:
            webpage += "cannot find manual script"
    webpage += "</td></tr>"

    # ===============================================
    # Flux calibrator amplitude vs. uvdist
    # TODO put in same plot for manual as for pipeline
    textForColumn1 = ''
    textForColumn2 = ''
    if (showingManualQA2):
        manualplot = manualqa2dir+'ampcal_uvdist.png'
        manualplot2 = manualqa2dir_extra+'ampcal_uvdist_antenna_subset.png'
    else:
        manualplot = ''
        manualplot2 = ''
    if (showingPipelineQA2):
        pipelineplot = pipeqa2dir+'ampcal_uvdist.png'
        pipelineplot2 = pipeqa2dir+'ampcal_uvdist_antenna_subset.png'
    else:
        pipelineplot = ''
    if (fluxcal_source_id == ''):
        fluxtext = 'No flux calibrator in this dataset!'
        pipelineplot = ''
        manualplot = ''
        if (showingPipelineQA2):
            textForColumn1 = 'No flux calibrator in this dataset!'
        if (showingManualQA2):
            textForColumn2 = 'No flux calibrator in this dataset!'
    elif (casadef.casa_version >= '4.0.0'):
        fluxtext = 'Flux calibrator<br> data and model<br><br></b>'
    else:
        fluxtext = 'Flux calibrator<br> data (green) and model (blue)<br><br></b>'

    useJPL=True
    if (fluxcal_source_name.lower() in SSO_LIST):
        println(lineno(),"Running au.planet(vis='%s',bodyForScan='%s')['angularDiameter']" % (writeableMSWithPathPipeline,fluxcal_source_name))
        try:
            angularDiameter = au.planet(vis=writeableMSWithPathPipeline,bodyForScan=fluxcal_source_name)['angularDiameter']
            fluxtext += 'Mean diameter = %.2f arcsec<br><br>' % (angularDiameter)
        except:
            useJPL = False
            try:
                println(lineno(),"Running au.planet(vis='%s',bodyForScan='%s',useJPL=%s)['angularDiameter']" % (writeableMSWithPathPipeline,fluxcal_source_name,str(useJPL)))
                angularDiameter = au.planet(vis=writeableMSWithPathPipeline,bodyForScan=fluxcal_source_name,useJPL=useJPL)['angularDiameter']
                fluxtext += 'Mean diameter = %.2f arcsec<br><br>' % (angularDiameter)
            except:
                println(lineno(),"ERROR: au.planet Failed!")
                fluxtext += 'Mean diameter = [JPL Horizons failed!]<br><br>'
    else:
        println(lineno(),"WARNING: %s is not in the list of Solar System Objects"%(fluxcal_source_name))
    
    webpage += addplotrow(webpagedir, title=fluxtext+'(all antennas)', 
                          plotForColumn1=pipelineplot,
                          plotForColumn2=manualplot,
                          textForColumn1 = textForColumn1,
                          textForColumn2 = textForColumn2,debug=debug)

    if (fluxcal_source_id != ''):
        println(lineno(),"Adding row for fluxscale")
        if (showingPipelineQA2):
            # Here we assume that the pipeline will use one of these two formats:
            #    '0~8'  or 'DV01,DV02'  and not  '0,2~8')
            if (antennaFluxCal.find('~')>0):
                nAntennas = 1 + int(antennaFluxCal.split('~')[1]) - int(antennaFluxCal.split('~')[0])
            else:
                nAntennas = len(antennaFluxCal.split(','))
            # should sort the names
            antennaFluxCal = ",".join(np.sort(antennaFluxCal.split(',')))
            # list only 3 per line to avoid making first column too wide
#           antennaFluxCalList = replaceEveryNthOccurrenceWith(antennaFluxCal, 3, ',', '<br>')
            if (nAntennas > 1):
                nAnt = str(nAntennas)
            else:
                nAnt = ''
            if (showingManualQA2):
                if (antennaFluxCalManual.find(',') > 0):
                    if (antennaFluxCalManual[-1] == '\n'):
                        antennaFluxCalManual = antennaFluxCalManual[:-1]
                    if (antennaFluxCalManual[-1] == ','):
                        antennaFluxCalManual = antennaFluxCalManual[:-1]
                    nAnts = len(antennaFluxCalManual.split(','))
                    if (nAnts==1 and antennaFluxCalManual=="''"):
                        textForColumn2 = "all antennas: %s" % (antennaFluxCalManual)
                    else:
                        textForColumn2 = "%d antennas: %s" % (nAnts, antennaFluxCalManual)
                else:
                    textForColumn2 = antennaFluxCalManual
                if len(antennaFluxCalManual)<=1: 
                    textForColumn2="all antennas (same as above)"
                else:
                    # create subset figure
                    # antennafluxcalmanual has single quotes and ends with &
                    if antennaFluxCalManual[0]=="'":
                        antennaFluxCalManual=antennaFluxCalManual[1:]
                    if antennaFluxCalManual[-1]=="'":
                        antennaFluxCalManual=antennaFluxCalManual[:-1]
                    correlation=''
                    println(lineno(),"Running es.ampcal_uvdist(ms2='%s', qa2_output_dir='%s', antenna=%s, correlation='%s')" % (msSplitWithPath, manualqa2dir_extra,antennaFluxCalManual, correlation))
                    es.ampcal_uvdist(ms2=msSplitWithPath, qa2_output_dir=manualqa2dir_extra, antenna=antennaFluxCalManual, correlation=correlation)
            else:
                textForColumn2 = ''
            mytb.open(writeableMSWithPathPipeline+'/ANTENNA')
            nAntennasInDataset = len(mytb.getcol('NAME'))
            mytb.close()
            if (nAntennasInDataset > nAntennas):
                subsetString = 'subset of'
            else:
                subsetString = 'all'
            textForColumn1 = "%s antennas used in gaincal: %s"%(nAnt,antennaFluxCal)
            if (os.path.exists(pipelineplot2) == False or subsetString=='all'):
                pipelineplot2 = ''
                textForColumn1 = 'same as above (all antennas used in fluxcal)'
            if (os.path.exists(manualplot2) == False):
                manualplot2=""
            webpage += addplotrow(webpagedir, 
                                  title=fluxtext,
#                                  title='%s<br> for %s %s antennas<br> used in gaincal:<br></b>%s'%(fluxtext,subsetString,nAnt,antennaFluxCalList), 
                                  plotForColumn1=pipelineplot2, 
                                  plotForColumn2=manualplot2,
                                  textForColumn1=textForColumn1,
                                  textForColumn2=textForColumn2,debug=debug)
            
        println(lineno(),"Adding row for flux calibrator spectrum")
        if (fluxcal_uvrange == ''):
            uvr = ''
        else:
            uvr = ", uvrange='%s'" % (fluxcal_uvrange)
        textForColumn1 = ''
        if (showingPipelineQA2):
            fluxcalflag_linesfile = findFluxcalflagLinesFile(pipecaldir+ppr)
            textForColumn1 = findPipelineRefspwmap(ccl, asdmNoPath, scienceSpwsIntegerList)
            if (fluxcalflag_linesfile != None):
                textForColumn1 += ', extra lines to flag = <a href="%s">%s</a>' % (fluxcalflag_linesfile+'.txt',fluxcalflag_linesfile)
                myCopyfile(pipecaldir+fluxcalflag_linesfile, webpagedir+fluxcalflag_linesfile+'.txt')

        if (showingManualQA2):
            from recipes.almahelpers import tsysspwmap
            tsysmap = tsysspwmap(vis = writeableMSWithPathManual, tsystable = msWithPathManual+".tsys", tsysChanTol = 1)
            # TODO get actual tsysChanTol from manual script
            textForColumn2="tsysmap from recipes.almahelpers = "+str(tsysmap)
        else:
            textforColumn2=""
        webpage += addplotrow(webpagedir,title='Flux calibrator<br>corrected spectrum<br>and model spectrum<br><br></b>(weighted average of all baselines%s)'%(uvr),
                                plotForColumn1=pipelineCalibratedFluxCalPlotsAmp,
                                plotForColumn2=manualCalibratedFluxCalPlotsAmp,
                                pdf1='',
                                pdf2='',
                                textForColumn1=textForColumn1,
                                textForColumn2=textForColumn2,debug=debug)
          
        webpage += "<tr><td><b>Flux scale results</b>"
        if (showingPipelineQA2):
            println(lineno(),"~~~~~~~~ fluxscaleResults = amc.getPipelineFluxscaleResults('%s','%s','%s')" % (pipelineCasaLogfile, asdmNoPath, msWithPath))
            fluxscaleResults = getPipelineFluxscaleResults(pipelineCasaLogfile, asdmNoPath, msWithPath)
        else:
            fluxscaleResults = ''
        if (showingManualQA2):
            manualresult = msSplitWithPath+'.fluxscale'
            if os.path.exists(manualresult):
                println(lineno(),"~~~~~~~~ manualFluxscaleResults = amc.getManualFluxscaleResults('%s')" % (manualresult))
                manualFluxscaleResults = getManualFluxscaleResults(manualresult)
                if (showingPipelineQA2):
                    println(lineno(),"~~~~~~~~ Calling amc.fluxscaleRatio(fluxscaleResults, manualFluxscaleResults)")
                    fluxRatio, fluxRatios = fluxscaleRatio(fluxscaleResults, manualFluxscaleResults)
                    webpage += "<br><br>Found %d matches<br>ratio = %.3f" % (len(fluxRatios),fluxRatio)
                    if (manualCasaVersion.split()[2] < '4.0.0'):
                        if (casadef.casa_version >= '4.1.0'):
                            mymsmd = createCasaTool(msmdtool)
                            mymsmd.open(writeableMSWithPathPipeline)
                            freqs = []
                            for spw in scienceSpwsIntegerList:
                                freqs.append(mymsmd.meanfreq(spw))
                            frequency = np.mean(freqs)
                            mymsmd.close()
                        else:
                            vm = au.ValueMapping(writeableMSWithPathPipeline)
                            freqs = []
                            for spw in scienceSpwsIntegerList:
                                freqs.append(vm.spwInfo[spw]['meanFreq'])
                            frequency = np.mean(freqs)
                        println(lineno(),"Mean frequency of science spws = %.6f GHz" % (frequency*1e-9))
                        expectedRatio = expectedFluxscaleRatio(fluxcal_source_name, frequency)
                    else:
                        expectedRatio = 1.0
                    if (expectedRatio != 1.0):
                        println(lineno(),"Expected flux ratio of 2012/2010 models = %f" % (expectedRatio))
                        webpage += "<br><br>expected ratio of <br>2012/2010 model = %.3f<br><br>new ratio = %.3f" % (expectedRatio,fluxRatio/expectedRatio)
        if (fluxcal_source_name.lower() in ['titan','callisto','io','ganymede','europa']):
            if (fluxcal_source_name.lower() in ['callisto','io','ganymede','europa']):
                println(lineno(),"Running au.angularSeparationOfPlanets(vis='%s', useJPL=%s)" % (writeableMSWithPathPipeline,str(useJPL)))
                (separation, beams, good) = au.angularSeparationOfPlanets(vis=writeableMSWithPathPipeline, useJPL=useJPL)
                webpage += '<br><br>Separation from Jupiter<br> = %.1f arcsec = %.1f beams<br>' % (separation*206264.8, beams)
            if (fluxcal_source_name.lower() == 'titan'):
                println(lineno(),"Running au.angularSeparationOfPlanets(vis='%s', useJPL=%s)" % (writeableMSWithPathPipeline,str(useJPL)))
                (separation, beams, good) = au.angularSeparationOfPlanets(vis=writeableMSWithPathPipeline, useJPL=useJPL)
                webpage += '<br><br>Separation from Saturn<br> = %.1f arcsec = %.1f beams<br>' % (separation*206264.8, beams)
            println(lineno(),"%s is %.1f arcsec from parent planet (%.1f beams)" % (fluxcal_source_name, separation*206264.8,beams))
            if (good == False):
                webpage += '<b>Too close to planet!'
        webpage += '</td><td>'
        useCourierForFluxscale = True
        if (useCourierForFluxscale): webpage += '<FONT FACE="courier">'
        webpage += fluxscaleResults
        if (useCourierForFluxscale): webpage += '</font>'
        webpage += '</td>'
        webpage += '<td>'
        if (useCourierForFluxscale): webpage += '<FONT FACE="courier">'
        # Flux scale information
        if (showingManualQA2):
          if os.path.exists(manualresult):
              webpage += manualFluxscaleResults
          else:
              webpage += "cannot find manual reduction's .fluxscale file<br>\n"
        if (useCourierForFluxscale): webpage += "</FONT>\n"
        webpage += "</td></tr>\n"
        
    # Pre-bandpass phase
    manualPreBandpassPhasePlot = ''
    manualPreBandpassPdfName = ''
    if (showingManualQA2):
        manpreBPplotdir=msSplitWithPath+'.ap_pre_bandpass.plots'
        if not os.path.exists(manpreBPplotdir):
            println(lineno(),"ERROR: Did not find the manual dataset's *.ap_pre_bandpass.plots.  Did you run es.checkCalTable during manual reduction?")
            println(lineno(),"attempting to run that for you...")
            if os.access(msSplitWithPath,os.W_OK):
                es.checkCalTable(msSplitWithPath+'.ap_pre_bandpass', msName=msSplitWithPath, interactive=False)
            else:
                manpreBPplotdir=writeableMSWithPathManual+".split.ap_pre_bandpass.plots"
                bar=writeableMSWithPathManual
                if bar[-1]=="/": bar=bar[:-1]
                es.checkCalTable(bar+".split.ap_pre_bandpass", msName=writeableMSSplitWithPath, interactive=False)
        if not os.path.exists(manpreBPplotdir):
            println(lineno(),"ERROR: Still can't find the manual dataset's *.ap_pre_bandpass.plots")
            if debug: pdb.set_trace()
            return

        pnglist = sorted(os.listdir(manpreBPplotdir))
        for p in range(len(pnglist)):
            pnglist[p] = manpreBPplotdir+'/' + pnglist[p]
        
        # RI TODO: use the same antenna as the pipeline used:        
        if (showingPipelineQA2):
            pipeBPfile = pipelinePreBandpassPhasePlot.split("/")[-1]
            print "Found pipeBPfile = %s" % (pipeBPfile)
            if (pipeBPfile == ''):
                manualPreBandpassPhasePlot = ''
            else:
                pipeBPant = pipeBPfile.split(".")[-2]
                myindex = indexContainingSubstring(pnglist, 'phase.'+pipeBPant)
                if (myindex == -1):
                    println(lineno(),"ERROR: Did not find the manual dataset's *.ap_pre_bandpass.plot %s.  Did you run es.checkCalTable during manual reduction?" % 'phase.'+pipeBPant)                    
                    #return
                    manualPreBandpassPhasePlot = ''
                else:
                    manualPreBandpassPhasePlot = pnglist[myindex]
        else:
            # pick first antenna plot file in the list, avoid the reference antenna
            manualRefAntPlotIndex = indexContainingSubstring(pnglist, 'phase.'+manualRefAnt)
            if (manualRefAntPlotIndex == 0):
                manualPlotIndex = 1
            else:
                manualPlotIndex = 0
            manualPreBandpassPhasePlot = pnglist[manualPlotIndex]
        manualPreBandpassPdfName = manpreBPplotdir+'.pdf'
        if (os.path.exists(manualPreBandpassPdfName) == False):
            bPFP = au.buildPdfFromPngs(pnglist=pnglist, pdfname=manualPreBandpassPdfName,quiet=auQuiet)
            println(lineno(),bPFP)

    println(lineno(),"Adding row for prebandpass phase")
    textForColumn1 = ''
    textForColumn2 = ''
    if (showingPipelineQA2):
        if (pipelinePreBandpassPhasePlot == ''):
            textForColumn1 = 'no pre-bandpass phase plot found for antenna %s' % (nonRefAnt)
        else:
            textForColumn1 = os.path.basename(pipelinePrebandpassCaltable)
    if (showingManualQA2):
        if (manualPreBandpassPhasePlot == ''):
            textForColumn2 = 'no pre-bandpass phase plot found for antenna %s' % (nonRefAnt)
    webpage += addplotrow(webpagedir, 
                          title='Pre-bandpass phase vs. time<br> on bandpass calibrator<br><br></b>(click image for<br> multi-page PDF)',
                          plotForColumn1=pipelinePreBandpassPhasePlot, 
                          plotForColumn2=manualPreBandpassPhasePlot,
                          pdf1=pipelinePreBandpassPdfName, 
                          pdf2=manualPreBandpassPdfName,
                          textForColumn1=textForColumn1,
                          textForColumn2=textForColumn2,debug=debug
                          )


    ########################################
    # amplitude table
    ########################################
    println(lineno(),"Adding row for Amp vs time plot")
    pipelineAmpVsTimePlot = ''
    pipelineAmpVsTimePdf = ''
    textForColumn1 = ''
    textForColumn2 = ''
    temppath=pipeqa2dir
    if heuristicsversion() > 2013.08: firstGACalStage=0
    if (showingPipelineQA2):
        caltable = pipecaldir + findAntGainCheckCalTable(pipecaldir, ms, minstage=firstGACalStage, maxstage=lastGACalStage)
        pipelineAmpVsTimePlot = pipeqa2dir + os.path.basename(caltable) + '.png'
        if (os.path.exists(pipelineAmpVsTimePlot) == False):
            println(lineno(),"Running checkCalTableVariant(calTableName='%s', msName='%s', gaintype='a', plotfile='%s')" % (caltable,writeableMSWithPathPipeline, pipelineAmpVsTimePlot))
            checkCalTableVariant(calTableName=caltable, msName=writeableMSWithPathPipeline, gaintype='a',interactive=False,
                                 plotfile = pipelineAmpVsTimePlot,temppath=temppath)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping checkCalTableVariant(amp vs time) on the pipeline reduction because the PNG already exists: %s" % (skipIndent,pipelineAmpVsTimePlot))
        pipelineAmpVsTimePdf = pipeqa2dir + os.path.basename(caltable) + '.pdf'
        textForColumn1 = os.path.basename(caltable)
        if (os.path.exists(pipelineAmpVsTimePdf) == False):
            plotdir = caltable + '.plots'
            if (cleanup):
                os.system('rm -rf %s' % (plotdir))
            if (os.path.exists(plotdir) == False):
                println(lineno(),"-------WARNING: Did not find directory %s - will look elsewhere" % (plotdir))
                if (os.path.exists(temppath+os.path.basename(plotdir.rstrip('/')))):
                    plotdir = temppath+os.path.basename(plotdir.rstrip('/')) + '/'
            println(lineno(),"Running buildPngList('%s')" % (plotdir))
            pnglist = buildPngList(plotdir)
            if (pnglist == []):  # why does this happen?
                println(lineno(),"Running checkCalTable('%s', msName='%s', interactive=False)" % (caltable,pipecaldir+ms))
                plotdir = callCheckCalTable(caltable, msName=pipecaldir+ms, interactive=False,temppath=temppath)
                println(lineno(),"Running buildPngList('%s')" % (plotdir))
#                println(lineno(),"------------------------------------------------")
#                println(lineno(),"Sleeping for 30 seconds, check: ls %s"%(plotdir))
#                time.sleep(30)
                pnglist = buildPngList(plotdir)
            println(lineno(),"Running buildPdfFromPngs(pnglist=%s, pdfname='%s')" % (str(pnglist),pipelineAmpVsTimePdf))
            bPFP = au.buildPdfFromPngs(pnglist=pnglist, pdfname=pipelineAmpVsTimePdf,quiet=auQuiet)
            println(lineno(),bPFP)
        elif (showSkippingMessages):
            println(lineno(),"%sSkipping the building of a PDF from the pipeline amplitude vs. time solution PNGs." % (skipIndent))
    if (showingManualQA2):
        textForColumn2 = 'This space intentionally left blank'
    webpage += addplotrow(webpagedir, title='Amplitude vs. time<br></b>all antennas overlaid<br><br>(click image<br>for PDF of<br>individual<br>antennas)',
                          plotForColumn1=pipelineAmpVsTimePlot, plotForColumn2='',
                          textForColumn1=textForColumn1,
                          textForColumn2=textForColumn2,
                          pdf1=pipelineAmpVsTimePdf,debug=debug)
        
    ########################################
    # gain statistics 
    ########################################
    println(lineno(),"Adding row for Gain statistics")
    webpage += "<tr><td><b>Amplitude<br>gain statistics</td>\n"
    webpage += '<td><pre>'
    if (showingPipelineQA2):
        caltable = pipecaldir + findAntGainCheckCalTable(pipecaldir, ms, minstage=firstGACalStage, maxstage=lastGACalStage)
        println(lineno(),"Running es.ant_gain_check(ms2='%s',qa2_output_dir='%s',caltable='%s',verbose=False, field='%s')" % (writeableMSWithPathPipeline,pipeqa2dir,caltable,bandpass_source_id))
        es.ant_gain_check(ms2=writeableMSWithPathPipeline,qa2_output_dir=pipeqa2dir,caltable=caltable,verbose=False,field=bandpass_source_id)
        gainCheckPipe = pipeqa2dir+'ant_gain_check.txt'   
        script = open(gainCheckPipe)
        webpage += "%s<br>" % (os.path.basename(caltable))
        for line in script:
            webpage += line    
        script.close()
    webpage += "</pre></td>\n"

    webpage += "<td><pre>\n"
    if (showingManualQA2):
      gainCheckManual = manualqa2dir_extra + 'ant_gain_check.txt'
# 20140210 the manual qa2 one comes out with all gains=1.0
# because ap_pre_bandpass is p not ap
#      caltable = msSplitWithPath + '.ap_pre_bandpass'
      caltable = msSplitWithPath + '.ampli_inf'
#      if (os.path.exists(gainCheckManual)):
      if False:
          script = open(gainCheckManual)
          webpage += "%s<br>" % (os.path.basename(caltable))
          for line in script:
              webpage += line
          script.close()
      else:
          if not (os.path.exists(caltable)):
              caltable = msSplitWithPath + '.flux_inf'
          if (os.path.exists(caltable)):
              println(lineno(),"Running es.ant_gain_check(ms2='%s',qa2_output_dir='%s',caltable='%s',verbose=False, field='%s')" % (msSplitWithPath,manualqa2dir_extra,caltable,bandpass_source_id))
              es.ant_gain_check(ms2=msSplitWithPath,qa2_output_dir=manualqa2dir_extra,caltable=caltable,verbose=False,field=bandpass_source_id)
              gainCheckManual = manualqa2dir_extra+'ant_gain_check.txt'   
              script = open(gainCheckManual)
              webpage += "%s<br>" % (os.path.basename(caltable))
              for line in script:
                  webpage += line    
              script.close()
          else:
              webpage += 'ant_gain_check.txt and '+caltable+' missing!'
    webpage += "</pre></td>\n"
    webpage += "</td></tr>\n"


    ########################################
    # Bandpass solution
    ########################################
    manualBandpassPhasePlot = ''
    manualBandpassPdfName = ''
    if (showingManualQA2):
      manualBandpassTable = findManualBandpassTable(manualcaldir, ms)
      if (manualBandpassTable != ''):
        if (os.path.exists(manualcaldir+manualBandpassTable+'.plots/')):
            println(lineno(),"Found %s" % (manualcaldir+manualBandpassTable+'.plots/'))
            figfile = manualcaldir+manualBandpassTable+'.plots/'
        else: # we need to build these plots
            figfile = manualqa2dir_extra + manualBandpassTable + '.plots/'
            if (os.path.exists(figfile) == False):
                println(lineno(),"Running plotbandpass('%s',xaxis='freq', yaxis='both', subplot=22, buildpdf=False, interactive=False, showatm=True, pwv='auto', showfdm=True, ms='%s', field='', figfile='%s/%s')" % (manualcaldir+manualBandpassTable, msSplitWithPath, figfile, manualBandpassTable))
                plotbandpass(manualcaldir+manualBandpassTable, xaxis='freq', yaxis='both', 
                                subplot=22, buildpdf=False, interactive=False, showatm=True, pwv='auto', 
                                showfdm=True, vis=msSplitWithPath, field='', 
                                figfile=figfile+'/%s'%(manualBandpassTable),
                                solutionTimeThresholdSeconds=solutionTimeThresholdSeconds)
            else:
                println(lineno(),"Skipping plotbandpass(manualBandpassTable) because .plots already exists")
        pnglist = sorted(os.listdir(figfile))
        for p in range(len(pnglist)):
            pnglist[p] = figfile + pnglist[p]
        if (manualBandpassTable.find('smooth20') > 0):
            searchString = 'bandpass_smooth20ch.'+nonRefAnt
        else:
            searchString = 'bandpass.'+nonRefAnt
        manualPlotIndex = indexContainingSubstrings(pnglist, searchString, 'spw0')
        manualBandpassPhasePlot = pnglist[manualPlotIndex]
        if (os.access(manualcaldir,os.W_OK)):
            manualBandpassPdfName = manualcaldir+manualBandpassTable+'.plots.pdf'
        else:
            manualBandpassPdfName = manualqa2dir_extra+manualBandpassTable+'.plots.pdf'
        println(lineno(),"manualBandpassPdfName = %s" % (manualBandpassPdfName))
        if (os.path.exists(manualBandpassPdfName) == False):
            bPFP = au.buildPdfFromPngs(pnglist=pnglist, pdfname=manualBandpassPdfName,quiet=auQuiet)
            println(lineno(),bPFP)

    println(lineno(),"Adding row for bandpass calibrator amp and phase")
    textForColumn1 = ''
    textForColumn2 = ''
    if (pipelineBandpassPhasePlot == '' and showingPipelineQA2):
        textForColumn1 = 'No bandpass amp/phase plot found for antenna %s' % (nonRefAnt)
    if (manualBandpassPhasePlot == '' and showingManualQA2):
        textForColumn2 = 'No bandpass amp/phase plot found for antenna %s' % (nonRefAnt)
    webpage += addplotrow(webpagedir, title='Bandpass<br><br></b>(click image for<br> multi-page PDF of<br>individual antennas)',
                          plotForColumn1=pipelineBandpassPhasePlot, 
                          plotForColumn2=manualBandpassPhasePlot,
                          pdf1=pipelineBandpassPdfName,
                          pdf2=manualBandpassPdfName,
                          textForColumn1=textForColumn1,
                          textForColumn2=textForColumn2, debug=debug
                          )

    #########################
    # Bandpass caltable flags
    #########################
    try:
        if (showingPipelineQA2):
            preamble, caltableFlags = printBandpassCaltableFlags(pipelineBandpassTableWithPath, '', msWithPathPipeline)
        else:
            preamble, caltableFlags = printBandpassCaltableFlags(manualcaldir+manualBandpassTable, '', msSplitWithPath)
        if (preamble != None):
            bandpassCaltableFlags = True
        else:
            bandpassCaltableFlags = False
    except:
        bandpassCaltableFlags = False
    if (bandpassCaltableFlags):
      println(lineno(),"Adding row for bandpass Caltable flagging")
      webpage += "<tr><td><b>Caltable flagging<br></b>(failed solutions<br>for bandpass):</td> <td>"
      if (showingPipelineQA2):
        # open the caltable and attempt to figure out what was flagged
        preamble, caltableFlags = printCaltableFlags(pipelineBandpassTableWithPath, '', msWithPathPipeline)
        if (caltableFlags == ''):
            webpage += "No flags detected!"
        else:
            caltableFlagsFile = 'pipeline.bandpass.caltable.flags.txt'
            nLines = writeStringToTextFile(pipeqa2dir+caltableFlagsFile, caltableFlags)
            myCopyfile(pipeqa2dir+caltableFlagsFile, webpagedir+caltableFlagsFile)
            webpage += preamble
            webpage += appendIfNecessary(caltableFlags, preamble, caltableFlagsFile, nLines)
      webpage += "</td> <td>"
      if (showingManualQA2):
        # open the caltable and attempt to figure out what was flagged
        print "Calling printCaltableFlags('%s', '', '%s')" % (manualBandpassTable,msSplitWithPath)
        preamble, caltableFlags = printCaltableFlags(manualcaldir+manualBandpassTable, '', msSplitWithPath)
        if (caltableFlags == ''):
            webpage += "No flags detected!"
        else:
            caltableFlagsFile = 'manual.bandpass.caltable.flags.txt'
            nLines = writeStringToTextFile(manualqa2dir_extra+caltableFlagsFile, caltableFlags)
            myCopyfile(manualqa2dir_extra+caltableFlagsFile, webpagedir+caltableFlagsFile)
            webpage += preamble
            webpage += appendIfNecessary(caltableFlags, preamble, caltableFlagsFile, nLines)
      webpage += "</td></tr>"
    

    ########################################
    # Phase vs. time on phase calibrator from es.ant_phase_temporal()
    ########################################
    manualplot = ''
    textForColumn2 = ''
    if (showingManualQA2):
        manualplot = manualqa2dir+'ant_phase_temporal.png'
        textForColumn2 = 'usually this is %s.split.phase_inf' % (ms)
    println(lineno(),"Adding row for Phase vs. time on phase calibrator")
    webpage += addplotrow(webpagedir, title='Phase vs. time<br> on phase calibrator<br> (field ID %s),<br> all spws overlaid'%(phase_cal),
                          plotForColumn1=pipelinePhaseTemporalPlotPhaseCalWithPath,
                          textForColumn1=pipelinePhaseTemporalGainTablePhaseCalNoPath,
                          textForColumn2=textForColumn2,
                          plotForColumn2=manualplot,debug=debug)

    # Phase vs. time on all calibrators using checkCalTable results
    println(lineno(),"Adding row for Phase vs. time on all calibrators")
    if (os.path.exists(manualcaldir+phase_inf_caltable) == False):
        textForColumn2 = ''
    else:
        textForColumn2 = phase_inf_caltable
    webpage += addplotrow(webpagedir,title='Phase vs. time<br> on all calibrators,<br> spws not overlaid<br><br></b>(click image for<br> multi-page PDF of<br>individual antennas)',
                          plotForColumn1 = pipelinePhaseTemporalPlotAllCalsWithPath,
                          plotForColumn2 = manualPhaseTemporalPlotAllCals, 
                          pdf1=pipelinePhaseTemporalPdfAllCalsWithPath,
                          pdf2=manualPhaseTemporalGainTablePdf,
                          textForColumn1=pipelinePhaseTemporalGainTableAllCalsNoPath,
                          textForColumn2=textForColumn2, debug=debug)

    ########################################
    # Amplitude vs. time on phase calibrator
    ########################################
    manualplot = ''
    pipelineplot = ''
    if (showingManualQA2):
        manualplot = manualqa2dir+'ant_amp_temporal.png'
        textForColumn2 = ampGainTypeManual + ': ' + ampGainTableManual.split('/')[-1]
    else:
        textForColumn2 = ''
    if (showingPipelineQA2):
        pipelineplot = pipelineAmpTemporalPlotPhaseCalWithPath
        textForColumn1 = ampGainTypePipeline + ': ' + pipelineAmpTemporalGainTablePhaseCalNoPath 
    else:
        textForColumn1 = ''
    println(lineno(),"Adding row for Amp vs. time on phase calibrator")
    webpage += addplotrow(webpagedir, 
                          title='amplitude vs. time<br> on phase calibrator<br> (field ID %s),<br> all spws overlaid<br></b><br>Note: only unflagged<br>solutions are shown.'%(phase_cal),
                          plotForColumn1=pipelineplot, 
                          plotForColumn2=manualplot, 
                          textForColumn1=textForColumn1,
                          textForColumn2=textForColumn2, debug=debug)

    # Amplitude vs. time on all calibrators using checkCalTable results
    println(lineno(),"Adding row for Amp vs. time on all calibrators")
    webpage += addplotrow(webpagedir,
                          title='Amplitude vs. time<br> on all calibrators,<br> spws not overlaid<br><br></b>(click image for<br> multi-page PDF of<br>individual antennas)',
                          plotForColumn1=pipelineAmpTemporalPlotAllCalsWithPath,
                          plotForColumn2=manualAmpTemporalPlot, 
                          pdf1=pipelineAmpTemporalPdfAllCalsWithPath,
                          pdf2=manualAmpTemporalPdf,
                          textForColumn1=pipelineAmpTemporalGainTableAllCalsNoPath,
                          textForColumn2=manualAmpTemporalTable, debug=debug)

    #######################
    # Caltable flags
    #######################
    println(lineno(),"Adding row for amp Caltable flagging")
    webpage += "<tr><td><b>Caltable flagging<br></b>(failed solutions<br>in amplitude<br>solint='inf'):</td> <td>"
    if (showingPipelineQA2):
        # open the caltable and attempt to figure out what was flagged
        preamble, caltableFlags = printCaltableFlags(pipelineAmpTemporalGainTableAllCalsWithPath, pipecaldir, ms)
        if (caltableFlags == ''):
            webpage += "No flags detected!"
        else:
            caltableFlagsFile = 'pipeline.amp.caltable.flags.txt'
            nLines = writeStringToTextFile(pipeqa2dir+caltableFlagsFile, caltableFlags)
            myCopyfile(pipeqa2dir+caltableFlagsFile, webpagedir+caltableFlagsFile)
            webpage += preamble
            webpage += appendIfNecessary(caltableFlags, preamble, caltableFlagsFile, nLines)
    webpage += "</td> <td>"
    if (showingManualQA2):
        # open the caltable and attempt to figure out what was flagged
        print "Calling printCaltableFlags('%s', '', '%s')" % (manualAmpTemporalTable,msSplitWithPath)
        preamble, caltableFlags = printCaltableFlags(manualcaldir+manualAmpTemporalTable, '', msSplitWithPath)
        if (caltableFlags == ''):
            webpage += "No flags detected!"
        else:
            caltableFlagsFile = 'manual.amp.caltable.flags.txt'
            nLines = writeStringToTextFile(manualqa2dir_extra+caltableFlagsFile, caltableFlags)
            myCopyfile(manualqa2dir_extra+caltableFlagsFile, webpagedir+caltableFlagsFile)
            webpage += preamble
            webpage += appendIfNecessary(caltableFlags, preamble, caltableFlagsFile, nLines)
    webpage += "</td></tr>"
    

    #######################
    # Caltable flags (phase)
    #######################
    println(lineno(),"Adding row for phase Caltable flagging")
    webpage += "<tr><td><b>Caltable flagging<br></b>(failed solutions<br>in phase<br>solint='inf'):</td> <td>"
    if (showingPipelineQA2):
        # open the caltable and attempt to figure out what was flagged
        preamble, caltableFlags = printCaltableFlags(pipelinePhaseTemporalGainTablePhaseCalWithPath, pipecaldir, ms)
        if (caltableFlags == ''):
            webpage += "No flags detected!"
        else:
            caltableFlagsFile = 'pipeline.phase.caltable.flags.txt'
            nLines = writeStringToTextFile(pipeqa2dir+caltableFlagsFile, caltableFlags)
            myCopyfile(pipeqa2dir+caltableFlagsFile, webpagedir+caltableFlagsFile)
            webpage += preamble
            webpage += appendIfNecessary(caltableFlags, preamble, caltableFlagsFile, nLines)
    webpage += "</td> <td>"
    if (showingManualQA2):
        if (os.path.exists(manualcaldir+phase_inf_caltable)):
            # open the caltable and attempt to figure out what was flagged
            print "Calling printCaltableFlags('%s', '', '%s')" % (manualcaldir+phase_inf_caltable, msSplitWithPath)
            preamble, caltableFlags = printCaltableFlags(manualcaldir+phase_inf_caltable, '', msSplitWithPath)
            if (caltableFlags == ''):
                webpage += "No flags detected!"
            else:
                caltableFlagsFile = 'manual.phase.caltable.flags.txt'
                nLines = writeStringToTextFile(manualqa2dir_extra+caltableFlagsFile, caltableFlags)
                myCopyfile(manualqa2dir_extra+caltableFlagsFile, webpagedir+caltableFlagsFile)
                webpage += preamble
                webpage += appendIfNecessary(caltableFlags, preamble, caltableFlagsFile, nLines)
    webpage += "</td></tr>"

    #######################
    # Calibrated data (amp)
    #######################
    if (plotlevel > 1):
        clickspectra = '<br><br></b>(click image for<br> multi-page PDF)'
    else:
        clickspectra = '<br><br></b>(re-run with plotlevel=2 to<br>generate a multi-page PDF)'
    println(lineno(),"Adding row for Calibrated amplitude vs. time")
    webpage += addplotrow(webpagedir,title='Calibrated amplitude<br> for science spws<br>(all sources)' + clickspectra,
                              plotForColumn1=pipelineCalibratedDataPlotsAmp,
                              plotForColumn2=manualCalibratedDataPlotsAmp,
                              pdf1=pipelineCalibratedDataPlotsAmpPdf,
                              pdf2=manualCalibratedDataPlotsAmpPdf,
                              textForColumn1='',
                              textForColumn2='', debug=debug)

    #########################
    # Calibrated data (phase)
    #########################
    println(lineno(),"Adding row for Calibrated phase vs. time")
    webpage += addplotrow(webpagedir,title='Calibrated phase<br> for science spws<br>(all sources)'+clickspectra,
                          plotForColumn1=pipelineCalibratedDataPlotsPhase,
                          plotForColumn2=manualCalibratedDataPlotsPhase,
                          pdf1=pipelineCalibratedDataPlotsPhasePdf,
                          pdf2=manualCalibratedDataPlotsPhasePdf,
                          textForColumn1='',
                          textForColumn2='', debug=debug)

    ################################
    # Bandpass calibrator spectrum
    #################################
    manualBandpassSpectrumPlot = ''
    if (showingManualQA2):
        manualBandpassSpectrumPlot = manualqa2dir + 'bandpass_spectrum.png'

    println(lineno(),"Adding row for Bandpass calibrator amplitude spectrum")
    webpage += addplotrow(webpagedir,title='Bandpass calibrator<br>amplitude spectrum<br> for each spw<br>colored by antenna<br><br>field: %s = %s%s' % (bandpass_source_id, bandpass_source_name, clickspectra),
                          plotForColumn1=pipelineBandpassSpectraAmpPlot,
                          plotForColumn2=manualBandpassSpectraAmpPlot,
                          pdf1=pipelineBandpassSpectraAmpPlotsPdf,
                          pdf2=manualBandpassSpectraAmpPlotsPdf,
                          textForColumn1='',
                          textForColumn2='', debug=debug)

    println(lineno(),"Adding row for Bandpass calibrator phase spectrum")
    webpage += addplotrow(webpagedir,title='Bandpass calibrator<br>phase spectrum<br> for each spw<br>colored by antenna<br><br>field: %s = %s%s' % (bandpass_source_id, bandpass_source_name, clickspectra),
                          plotForColumn1=pipelineBandpassSpectraPhasePlot,
                          plotForColumn2=manualBandpassSpectraPhasePlot,
                          pdf1=pipelineBandpassSpectraPhasePlotsPdf,
                          pdf2=manualBandpassSpectraPhasePlotsPdf,
                          textForColumn1='',
                          textForColumn2='', debug=debug)

    ################################
    # Phase calibrator spectrum
    #################################
    if (showPhasecalBandpass):    
        manualPhasecalSpectrumPlot = ''
        if (showingManualQA2):
            manualPhasecalSpectrumPlot = manualqa2dir + 'bandpass_spectrum.png'
    
        println(lineno(),"Adding row for Phase calibrator amplitude spectrum")
        webpage += addplotrow(webpagedir,title='Phase calibrator<br>amplitude spectrum<br> for each spw<br>colored by antenna<br><br>field: %s = %s%s<br><br>avgchan=%d' % (phase_cal_source_id, phase_cal_source_name, clickspectra,phasecal_avgchan),
                              plotForColumn1=pipelinePhasecalSpectraAmpPlot,
                              plotForColumn2=manualPhasecalSpectraAmpPlot,
                              pdf1=pipelinePhasecalSpectraAmpPlotsPdf,
                              pdf2=manualPhasecalSpectraAmpPlotsPdf,
                              textForColumn1='',
                              textForColumn2='', debug=debug)
    
        println(lineno(),"Adding row for Phase calibrator phase spectrum")
        webpage += addplotrow(webpagedir,title='Phase calibrator<br>phase spectrum<br> for each spw<br>colored by antenna<br><br>field: %s = %s%s<br><br>avgchan=%d' % (phase_cal_source_id, phase_cal_source_name, clickspectra,phasecal_avgchan),
                              plotForColumn1=pipelinePhasecalSpectraPhasePlot,
                              plotForColumn2=manualPhasecalSpectraPhasePlot,
                              pdf1=pipelinePhasecalSpectraPhasePlotsPdf,
                              pdf2=manualPhasecalSpectraPhasePlotsPdf,
                              textForColumn1='',
                              textForColumn2='', debug=debug)

    #################
    # Target spectrum
    #################
    manualTargetSpectrumPlot = ''
    if (showingManualQA2):
#        manualTargetSpectrumPlot = manualqa2dir + 'target_spectrum.png'
      manualTargetSpectrumPlot = manualqa2dir_extra + 'target_spectrum.png'
      if (os.path.exists(manualTargetSpectrumPlot) == False):
          if (badCharacters(target_source_name)):
              println(lineno(),"Running for manual data: es.target_spectrum(ms2='%s', target='%s', dospw='%s', qa2_output_dir='%s', uvrange='%s', correlation='%s', avgchannel=%d)" % (msSplitWithPath, str(target_source_id), scienceSpwsManual, manualqa2dir_extra, uvrange, correlation, avgchannelTarget))
              es.target_spectrum(ms2=msSplitWithPath, target=str(target_source_id),
                                 dospw=scienceSpwsManual, qa2_output_dir=manualqa2dir_extra,
                                 uvrange=uvrange, correlation=correlation, avgchannel=avgchannelTarget)
          else:
              println(lineno(),"Running for manual data: es.target_spectrum(ms2='%s', target='%s', dospw='%s', qa2_output_dir='%s', uvrange='%s', correlation='%s', avgchannel=%d)" % (msSplitWithPath, target_source_name, scienceSpwsManual, manualqa2dir_extra, uvrange, correlation, avgchannelTarget))
              es.target_spectrum(ms2=msSplitWithPath, target=target_source_name,
                                 dospw=scienceSpwsManual, qa2_output_dir=manualqa2dir_extra,
                                 uvrange=uvrange, correlation=correlation, avgchannel=avgchannelTarget)
      textForColumn2 = 'field = %s' % (target_source_name)
    else:
        textForColumn2 = ''
    if (showingPipelineQA2):
        textForColumn1 = 'field = %s' % (target_source_name)
    else:
        textForColumn1 = ''
    println(lineno(),"Adding row for Target spectrum")
    webpage += addplotrow(webpagedir,title='Target spectrum<br> for each spw<br><br></b>(uv range<br> restricted)',
                          plotForColumn1=pipelineTargetSpectrumPlot,
                          plotForColumn2=manualTargetSpectrumPlot,
                          pdf1='', pdf2='',
                          textForColumn1=textForColumn1,
                          textForColumn2=textForColumn2, debug=debug)
#                          textForColumn2='field = %s' % (manualQA2ScienceTarget))

    ###########
    # Finish up    
    ###########
    f = open(webfile,"w")
    print >>f, webpage + """
         </table>
         </body>
         </html>
      """
    f.close()
    if (os.path.exists(webpagedir+'/index.html')):
        cmd = webpagedir+'/index.html'
#        println(lineno(),"Removing %s" % (cmd))
        os.remove(cmd)
#    os.system('ln -s %s/%s %s/index.html' % (os.getcwd(), webfile, webpagedir)) # absolute path
    cmd = 'ln -s %s %s/index.html' % (htmlfile, webpagedir) # relative path, requested by analysts
#    println(lineno(),"Running %s" % (cmd))
    os.system(cmd)  
    if (showingPipelineQA2):
        setdirGroupWrite(pipeqa2dir)
        temppath=pipeqa2dir
    if (showingManualQA2):
        setdirGroupWrite(manualqa2dir_extra)
        if not showingPipelineQA2: temppath=manualqa2dir_extra
    setdirGroupWrite(webpagedir)
#    if (debug_plotms):
#        fplotms.close()
    println(lineno(),"Done!")
    print "analyzemscal terminal output left in: %s" % (terminalfileToOpen)

    if (os.path.exists(terminalfile)):
        println(lineno(),"Calling unwrapCasaCommandLog('%s',width=80,newlog='%s.txt')" % (terminalfile,terminalfile))
        unwrapCasaCommandLog(terminalfile, width=80, newlog=terminalfile+'.txt',temppath=temppath)
        myCopyfile(terminalfile+'.txt', webpagedir+os.path.basename(terminalfile)+'.txt')
    if (os.path.exists(terminalfileToOpen) and terminalfileToOpen != terminalfile):
        println(lineno(),"Calling unwrapCasaCommandLog('%s',width=80,newlog='%s.txt')" % (terminalfileToOpen,terminalfileToOpen))
        unwrapCasaCommandLog(terminalfileToOpen, width=80, newlog=terminalfileToOpen+'.txt',temppath=temppath)
        myCopyfile(terminalfileToOpen+'.txt', webpagedir+os.path.basename(terminalfileToOpen)+'.txt') # might be .cleanupFalse

    if (wasInteractive):
        pl.ion()
#    if (debug_plotms):
#        if (fplotmsHasCommands > 0):
#            println(lineno(),"For some reason, plotms did not create all of its plots.  To make the %d missing plots," % (fplotmsHasCommands))
#            println(lineno(),"restart casapy, then run: execfile('%s'), then re-run analyzemscal with cleanup=False" % (runplotms))
    println(lineno(),"   Wrote the web page: %s"  % (webfile))
    println(lineno(),"   which will be accessible in your web browser at\n  %s" % (location) )
    duration = (time.time()-startTime)/60.0
    println(lineno(),"Execution time = %.2f minutes" % (duration))
    terminal.close()










































def replaceNthWith(n, replacement):
    def replace(match, c=[0]):
        c[0] += 1
        return replacement if c[0] in n else match.group(0)
    return replace

def replaceEveryNthOccurrenceWith(mystring, n, token, replacement):
    """
    In <mystring>, replace every <n>th <token> string with <replacement> string.
    """
    return(re.sub(token,replaceNthWith(range(n,len(mystring),n),replacement),mystring))
    
def unwrapCasaCommandLog(mylog, width=80, newlog='',strippath=False, maxlines=-1,temppath="./tmp/"):
    """
    Takes an ASCII file and unwraps the long lines (without breaking words)
    and writes a new file.  Requires:  "import textwrap"
    mylog: the full path to the casa log file to be processed
    width: the character width parameter passed to textwrap.wrap()
    newlog: (optional) name of output file
    strippath: if True, then cull the leading directories and put a cd command at top
    maxlines: how many lines to convert (for debugging purposes only)
    """
    if (newlog == ''):
        newlog = mylog + '.txt'
    if (os.path.exists(mylog) == False):
        print "ERROR: log file does not exist: %s" % (mylog)
        return
    f = open(mylog,'r')
    if (os.path.exists(newlog)):
        if (os.access(newlog, os.W_OK)):
            println(lineno(),"WARNING: Overwriting existing file")
        else:
            println(lineno(),"WARNING: Cannot overwrite existing file, writing to "+temppath)
            newlog='%s/%s'%(temppath,os.path.basename(newlog))
    else:
        mydir = os.path.dirname(newlog)
        if (mydir == ''):
            mydir = os.getcwd()
        if (os.access(mydir, os.W_OK)):
            println(lineno(),"Writing new file")
        else:
            println(lineno(),"WARNING: Cannot write to directory = %s, writing to /tmp instead" % (mydir))
            newlog='/%s/%s'%(temppath,os.path.basename(newlog))
    newf = open(newlog,'w')
    if (strippath):
        if (mylog[0] != '/'):
            mylog = os.getcwd() + '/' + mylog
        basedir = os.path.dirname(mylog.split('pipeline-20')[0])
        basedir_upone = os.path.dirname(basedir)
        basedir += '/'
        newf.write('cd %s\n' % basedir)
    lines = f.readlines()
    if (maxlines > 0):
        lines = lines[:maxlines]
    for line in lines:
        # task(vis='/lustre/mypath/myfile',output='/lustre/mypath/myfile2')
        # task(vis='  /lustre/mypath/  myfile',output='    /lustre/mypath/     myfile2')
        #    0                      1                                       2
        if (strippath):
            while (line.find(basedir) > 0):
                tokens = line.split(basedir)
                newline = tokens[0]    # task(vis='
                if (len(tokens) > 1):
                    # there is still a long path name on the line
                    newline += os.path.basename(tokens[1].split("'")[0]) + tokens[1][tokens[1].find("'"):]   # task(vis='myfile',output='
                    for t in range(2,len(tokens)):
                        newline += basedir + tokens[t]  # task(vis='myfile',output='/lustre/mypath/myfile2')
                line = newline
            while (line.find(basedir_upone) > 0):
                tokens = line.split(basedir_upone)
                newline = tokens[0] 
                if (len(tokens) > 1):
                    # there is still a long path name on the line
                    rawdata = tokens[1].lstrip('/').split('/')[0]+'/'  # typically '/rawdata/uid' --> 'rawdata/'
                    newline += '../' + rawdata + os.path.basename(tokens[1].split("'")[0]) + tokens[1][tokens[1].find("'"):]   
                    for t in range(2,len(tokens)):
                        newline += basedir_upone + tokens[t]  # task(vis='myfile',output='/lustre/mypath/myfile2')
                line = newline
        sublines = textwrap.wrap(line,subsequent_indent=10*' ',width=width,break_long_words=False)
        for subline in sublines:
            newf.write(subline+'\n')
    f.close()
    newf.close()
    println(lineno(),"unwrapCasaCommandLog(): New unwrapped file left in: %s" % (newlog))
    return newlog

def findPPR(mydir, casalog=''):
    """
    Finds and returns the name of the PPR file, without any preceding path because
    it expects to find it in the specified directory
    """
    println(lineno(),"findPPR called with mydir=%s, casalog=%s" % (mydir,casalog))
    if (casalog != '' and casalog != None):
        inLog = commands.getoutput('grep "For processing request" ' + casalog)
        # Get the first PPR mentioned in the file that also exists in mydir
        pprlist = inLog.split('For processing request: ')
        for pprtext in pprlist[1:]:
            ppr = pprtext.split('\n')[0].strip()
            if (os.path.exists(mydir+ppr)):
                println(lineno(),"Found PPR in casa log file: %s (%s)" % (ppr, casalog))
                return(ppr)
            else:
                println(lineno(),"WARNING: Rejecting PPR=%s because it is not found in the %s"%(ppr,mydir))
        println(lineno(),"ERROR: No valid PPRs found in %s" % (mydir))
        return(None)
#        ppr = inLog.split('For processing request: ')[-1].split('\n')[0].strip()
    else:
        filelist = sorted(glob.glob(mydir+'/PPR*xml'), key=os.path.getmtime, reverse=True)
        if (filelist == []):
            return(None)
        else:
#            mostRecent = pickMostRecentFile(filelist)
            mostRecent = os.path.basename(filelist[0])
            return(mostRecent)
    
def find_manual_script(mydir):
    """
    Finds and returns the name of the manual calibrations script, without any preceding path because
    it expects to find it in the specified directory
    """
    searchToken = mydir+"*scriptForCalibration*.py"
#    print "search for ", searchToken
    calScript = sorted(glob.glob(searchToken), key=os.path.getmtime, reverse=True)
#    print "calScript = %s." % (calScript)
    if (calScript == []):
        return(None)
    else:
        return(os.path.basename(calScript[0]))

def pickMostRecentFile(filelist):
    # Unused, once I learned about sorted(glob.glob(), key=os.path.getmtime)
    max_mtime = 0
    max_file = ''
    for fname in filelist:
        mtime = os.stat(fname).st_mtime
        if mtime > max_mtime:
            max_mtime = mtime
            max_file = fname
    return(max_file, max_mtime)

def findPipelineDir(pipecaldir, asdmNoPath='', debug=False):
    """
    Finds and returns a string like:  pipeline-20130325T134833
    """
    if (pipecaldir[-1] != '/'): pipecaldir += '/'
    pipelineFiles = sorted(glob.glob(pipecaldir + 'pipeline-20*'), key=os.path.getmtime, reverse=True)
    pipelineYYYYMMDD = []
    for pf in pipelineFiles:
        # screen out the .context files
        if (pf.find('.context') < 0):
            if (debug): print "Found directory: %s" % (pf)
            pipelineYYYYMMDD.append(pf)
    if (len(pipelineYYYYMMDD) > 1):
        # Search for asdm in the html/casa_commands.log file to try to limit the possibilities
        pipelineFiles = []
        for pf in pipelineYYYYMMDD:
            if (os.path.exists(pf+'/html/casa_commands.log')):
                if (asdmNoPath == ''):
                    result = commands.getoutput("grep importasdm %s/html/casa_commands.log" % (pf))
                else:
                    result = commands.getoutput("grep %s %s/html/casa_commands.log | grep importasdm" % (asdmNoPath,pf))
                if (result != ''):
                    pipelineFiles.append(pf)
                    if (debug): print "Found importasdm(%s) in %s/html/casa_commands.log" % (asdmNoPath, pf)
        if (len(pipelineFiles) > 1):
            println(lineno(),"WARNING: There is some uncertainty as to which pipeline-YYYYMMDD directory is the right one.")
            println(lineno(),"WARNING: I picked the most recent one where the casa_commands.log contained importasdm(%s)." % (asdmNoPath))
        elif (len(pipelineFiles) == 0):
            println(lineno(),"ERROR: None of the %d pipeline-YYYYMMDD directories contain the asdm in the casa_commands.log file." % (len(pipelineYYYYMMDD)))
            println(lineno(),"ERROR: Do you have the ASDM name correct?")
            return([])
        if (len(pipelineFiles) > 0):
            pipelineYYYYMMDD = os.path.basename(pipelineFiles[0])
    elif (len(pipelineYYYYMMDD) == 1):
        pipelineYYYYMMDD = os.path.basename(pipelineYYYYMMDD[0])
    else:
        print "ERROR: You don't seem to have a pipeline-YYYYMMDDTHHMMSS directory in pipecaldir.  Did you run the pipeline yet?"
        print "ERROR: If so, then please check the <RelativePath> keyword in the PPR.  It probably does not "
        print "ERROR: match what you entered for pipecaldir."
    println(lineno(),"pipelineYYYYMMDD = %s"% pipelineYYYYMMDD)
    print 
    return(pipelineYYYYMMDD)

def find_casa_commands_log(mydir, pipelinedir):
    """
    Finds and returns the name of the casa_commands.log file, preceded
    by the portion of the path below the specified directory. In principle,
    this can work for either the pipeline or the manual reduction, but is
    currently only called for the pipeline.
    """
    if (pipelinedir == ''):
        # This will likely never be used again since findPipelineDir was added.
        searchpath = mydir+'pipeline*/html/casa_commands.log'
    else:
        print "mydir=",mydir
        print "pipelinedir=",pipelinedir
        searchpath = mydir+pipelinedir+'/html/casa_commands.log'
    filelist = sorted(glob.glob(searchpath), key=os.path.getmtime, reverse=True)
#    filelist = glob.glob(searchpath)
    if (filelist == []):
        println(lineno(),"mydir = %s" % (mydir))
        println(lineno(),"--------- WARNING: Could not find casa_commands.log in %s" % (searchpath))
        print 
        return(None)
    elif (len(filelist) == 1):
        println(lineno(),"---------  Found casa_commands.log: %s" % (filelist[0]))
        print
        return(filelist[0][len(mydir):])
    else:
        # pick the most recent one
#        max_file, max_mtime = pickMostRecentFile(filelist)
#        println(lineno(),"Picking most recently modified casa_commands.log file (from %d possible): %s" % (len(filelist),max_file[len(mydir):]))
        println(lineno(),"------WARNING: Picking most recently modified casa_commands.log file (from %d possible): %s" % (len(filelist), filelist[0][len(mydir):]))
        print
        return(filelist[0][len(mydir):])


def writeToPPR(desc, line, extra=' '):
    global linectr, globaldebug
    desc.write(line)
    linectr = linectr + 1
    if (globaldebug):
        print "Wrote line %3d %s: %s" % (linectr,extra,line[:-1])

def insertFlagData(newppr,flagdata):
    writeToPPR(newppr,'                        <Parameter>\n')
    writeToPPR(newppr,'                            <Keyword>template</Keyword>\n')
    writeToPPR(newppr,'                            <Value>True</Value>\n')
    writeToPPR(newppr,'                        </Parameter>\n')
    if (flagdata != True):
        writeToPPR(newppr,'                        <Parameter>\n')
        writeToPPR(newppr,'                            <Keyword>filetemplate</Keyword>\n')
        writeToPPR(newppr,'                            <Value>%s</Value>\n'%flagdata)
        writeToPPR(newppr,'                        </Parameter>\n')

def insertFlagBackup(newppr):
    writeToPPR(newppr,'                        <Parameter>\n')
    writeToPPR(newppr,'                            <Keyword>flagbackup</Keyword>\n')
    writeToPPR(newppr,'                            <Value>True</Value>\n')
    writeToPPR(newppr,'                        </Parameter>\n')

default_sharps_limit = 0.1
default_sharps2_limit = 0.1

#===============================================================
#===============================================================
#===============================================================
def checkPPR(xmlfile):
    """
    Check PPR to make sure all tasks exist in the CASA task dictionary. - Joe Masters
    Returns: True if things look OK.
    """
    import xml.dom.minidom as minidom
    from pipeline.infrastructure import casataskdict

    if not os.path.isfile(xmlfile) or not xmlfile.endswith('xml'):
        LOG.info('Please provide a single valid xml PPR filename as input and try again.')
        return False
    casatasks = set(casataskdict.CasaTaskDict.keys())
    doc = minidom.parse(xmlfile)
    tasks_in_ppr = set([str(x.firstChild.data) for x in doc.getElementsByTagName('Command')])
    diffs = (tasks_in_ppr - casatasks)
    for suspect_task in diffs:
        LOG.info("{} is not a recognized CASA task.  Check your PPR and try again.\n".format(suspect_task))
        return False
    if not diffs:
        LOG.info("Passed.  All tasks exist in the CASA task dictionary.")
        return True 

def addDoubleQuotesToFieldList(imageSources, delimiter=','):
    """
    Converts  ['ab','cd']  or 'ab,cd' into '"ab","cd"'
    imageSources: a comma-delimited string, or a python list or array of source name strings
    delimiter: only relevant if imageSources is a string
    """
    if (type(imageSources) == str):
        return ','.join(['"'+imageSource+'"' for imageSource in imageSources.split(delimiter)])
    else:
        return ','.join(['"'+imageSource+'"' for imageSource in imageSources])

def modifyPPR(ppr, refant='', newpprname='', wvrflag='', 
              stopAfterCalibration=False,
              phaseupbw='', tie=True, wvrqa2intent='', wvrgcalflag='',
              asdm='', resolvedcals='', bpsolint='', toffset=None,
              fluxdensity=None, vis=[], tsysflag_shape_fhi_limit='',
              stopaftertsysflag=False, debug=False, overwrite=False,
              usems=False, stopafterimport=False, skipwvr=False,
              skipflagdata=False, stopafterwvr=False, flagdata=False,
              wvrsmooth='',spw=None,wvrgcalflag_intent='',wvr_fhi_limit='',
              tiesources='',field=None, flag_edges=False, flag_sharps='',
              flag_sharps2='',edge_limit='',sharps_limit=default_sharps_limit,
              sharps2_limit=default_sharps2_limit, metric='antenna_median',
              startbeforetsysflagchans=False, tsysflag_intentgroups='',
              fluxcal_linesfile='',metricb='median',flag_sharps2b='',
              sharps2b_limit='', atmflag_nmedian=False,
              atmflag_intent='', stopaftertsyscal=False,
              tsysflagchans_intentgroups='',bpflagchans=False, fitorder=None,
              fitfunc=None,edge=None,tsysChanTol=None, asdmDiskName='',asis='',
              maxdistm=None, minnumants=None, bdfflags=None, imageSources=[],
              dbservice=None, tsysNormalize=None, nbins='', clearcals=True,
              moveBpBefore='', skipcontsub=False, imageSpws=[], cubeSpws=[],
              flagbackup=False, fixpath=False):
    """
    Modifies a PPR.xml file. You must specify all the WVR-related modifications
    that you need, even if they already exist in the input file from an earlier
    execution of modifyPPR(), because the contents of any selected sections are
    overwritten.

    General parameters:
       ppr : the filename of the existing PPR file
       newpprname : the name of the new PPR file (default: appends selected
                    options to the current filename)
             Set to False in order to modify the existing file (but it will first
             copy the original PPR to <ppr>.original).
       asdm: if a PPR contains multiple ASDMs, use this parameter to only
             include the one specified ASDM (the last four letters is enough)
             or a list of ASDMs: e.g.  'Xcde,Xeff' 
       asdmDiskName: if not blank, use this string as the name of the ASDM
       usems: If True, then change the existing AsdmDiskName entries to *.ms
       fixpath: if True, and if the SchedBlockSet RelativePath is not found in the current working path,
                then change the RelativePath to match your current working path
       vis: use this parameter to pass the name of an ms to hif_importdata,
            instead of the ASDM
       stopafterimport : Only copy commands until the first hif_flagdata
       stopaftertsyscal : Only copy commands until the first hif_tsysflag
       stopaftertsysflag : Only copy commands until the first hif_setjy or hif_antpos
       stopafterwvr : Only copy commands until the first hif_tsyscal
       startbeforetsysflagchans : Only copy the importdata, tsyscal & tsysflagchans stages
       stopAfterCalibration : True will strip out the imaging portion if present
       overwrite: set the overwrite parameter in hif_importdata
       clearcals: if False, then set clearcals=False in importdata
       moveBpBefore: if specified as a hif(a) task, then move the breakpoint to before the
                first execution of this task,  e.g. 'hifa_flagtargets'.

    Calibration parameters:
       asis: set the asis parameter in hif_importdata
       atmflag_nmedian: if set to a value, then insert atmflag(flag_nmedian=True,fnm_limit=value)
       atmflag_intent: which intents to flag (default: '' = calibrators only)
       bdfflags: sets the bdfflags option of hifa_importdata (True or False)
       bpflagchans: if True, then insert this stage before hif_gfluxscale
       bpsolint: set to 'inf,20ch' to emulate manual script for FDM data
                 set to 'inf,7.8125MHz' to get 16 channel smoothing on widest FDM
       dbservice: sets the dbservice option of hifa_importdata (True or False)
       edge_limit: floating point, set this parameter in a new hif_tsysflagchans stage
       field: the field (name or ID) to use for the fluxdensity values
       flagbackup: if True, then set flagbackup=True in hifa_flagdata
       flag_edges: Boolean, set this parameter in a new hif_tsysflagchans stage
       flag_sharps: Boolean, set this parameter in a new hif_tsysflagchans stage
       flag_sharps2: Boolean, set this parameter in a new hif_tsysflagchans stage
       flag_sharps2b: Boolean, set this parameter in a 2nd new hif_tsysflagchans stage
       fluxcal_linesfile: a list of additional lines to flag in the flux calibrator
       flagdata: True: insert the template=True option into hif_flagdata
             string filename:  same as True, but also insert the filetemplate parameter=filename
       fluxdensity: the value (or list) to give to hif_setjy for the I Stokes parameter,
                   e.g. [7.8,7.8,7.4,7.4]
       maxdistm: for wvrgcal (default = 500.0m)
       metric: metric to use in a new hif_tsysflagchans stage (default='antenna_median')
       metricb: metric to use in a 2nd new hif_tsysflagchans stage (default='median')
       minnumants: for wvrgcal (default = 2)
       phaseupbw : the central spw bandwidth to use in the pre-bandpass gaincal
       refant : the antenna name to use as reference (e.g. 'DV04')
       resolvedcals : set to 'manual' to use all antennas
       sharps_limit: floating point, set this parameter in a new hif_tsysflagchans stage
       sharps2_limit: floating point, set this parameter in a new hif_tsysflagchans stage
       sharps2b_limit: Boolean, set this parameter in a 2nd new hif_tsysflagchans stage
       skipflagdata: do not run hif_flagdata
       skipwvr: do not run hif_wvrgcal
       spw: a list of spws, used with the fluxdensity parameter to specify the order
            of spw flux densities (e.g. [17,19,21,23]) 
       tie: default=True: automatic,  False: sets hm_tie to 'manual' and tie=tiesources
       tiesources:  if tie==False, sets the value of tie (e.g. '' or '3c273,3c279')
       toffset: The value in seconds to pass to wvrgcal or wvrgcalflag
       tsysChanTol: integer value for tsysspwmap, pipeline defaults to 1
       tsysflag_shape_fhi_limit: Set the fhi_limit parameter of tsysflag shape
       tsysflag_intentgroups: Set intentgroups in tsysflag stages
       tsysflagchan_intentgroups: Set intentgroups tsysflagchans stages
       tsysNormalize: sets the normalize_tsys option of hifa_tsysflag 
                      (not compatible with specifying other tsysflag options)
       wvr_fhi_limit:  a floating point value (default: '' = 5.0)
       wvrflag : a comma-delimited string list of antennas to flag
       wvrgcalflag : if True, replaces hif_wvrgcal with hif_wvrgcalflag
       wvrgcalflag_intent : default='*BANDPASS*', but can be '*PHASE*,*BANDPASS*'
       wvrqa2intent: the qa2_intent to pass to hif_wvrgcal (e.g. '*BANDPASS*')
       wvrsmooth: '' (default), 'automatic', or an integer # of samples

    Single-dish calibration parameters:
       fitorder: set the fit order in hsd_baseline
       fitfunc: set the fitting function in hsd_baseline (poly or spline)
       edge: set the edge parameter in hsd_baseline (must be a list of length 2)

    Imaging parameters:
       imageSources: a list or comma-delimited string of source names to image,
                     if this is an imaging PPR.  Also inserts into the calls
                     to hif_uvcontfit and hif_uvcontsub.
       nbins: inserts the specified nbins parameter into the second and fourth 
              calls to makeimlist:  e.g. ':2' or '17:2,19:2' etc.
       skipcontsub: if True, removes findcont, uvcontfit, uvcontsub, and cube
              imaging
       imageSpws: a list or comma-delimited string of spw IDs to image for the
                  science target continuum images, if this is an imaging PPR.
       cubeSpws: a list or comma-delimited string of spw IDs to image for the
                 science target line cubes, if this is an imaging PPR.
       Note: if you specify both imageSpws and cubeSpws, every spw in cubeSpw must
             also appear in imageSpws so that continuum subtraction will happen

    -- Todd Hunter
    """
    global linectr, globaldebug
    globaldebug = debug
    
    commandLine = "amc.modifyPPR(ppr='%s'" % ppr
    if (refant != ''):
        commandLine += ", refant='%s'" % str(refant)
    if (nbins != ''):
        nbins = str(nbins)
        commandLine += ", nbins='%s'" % nbins
    if (asdm != ''):
        commandLine += ", asdm='%s'" % asdm
    if (newpprname != ''):
        if (type(newpprname) == str):
            commandLine += ", newpprname='%s'" % newpprname
        else:
            commandLine += ", newpprname=%s" % str(newpprname)
    if (len(imageSources) > 0):
        if (type(imageSources) == str):
            commandLine += ", imageSources='%s'" % imageSources
            imageSources = imageSources.split(',')
        else:
            commandLine += ", imageSources=%s" % str(imageSources)
    if (type(imageSpws) == int):
        imageSpws = [str(imageSpws)]
    if (len(imageSpws) > 0):
        if (type(imageSpws) == str):
            commandLine += ", imageSpws='%s'" % imageSpws
            imageSpws = imageSpws.split(',')
        else:
            commandLine += ", imageSpws=%s" % str(imageSpws)
            if (type(imageSpws[0]) != str):
                imageSpws = [str(i) for i in imageSpws]
    if (type(cubeSpws) == int):
        cubeSpws = [str(cubeSpws)]
    if (len(cubeSpws) > 0):
        if (type(cubeSpws) == str):
            commandLine += ", cubeSpws='%s'" % cubeSpws
            cubeSpws = cubeSpws.split(',')
        else:
            commandLine += ", cubeSpws=%s" % str(cubeSpws)
            if (type(cubeSpws[0]) != str):
                cubeSpws = [str(i) for i in cubeSpws]
    commandLine += ')'
    if len(cubeSpws) > 0 and len(imageSpws) > 0:
        for cubeSpw in cubeSpws:
            if cubeSpw not in imageSpws:
                print "Aborting: spw %s must also be included in imageSpws in order for continuum subtraction to happen." % (cubeSpw)
                return
    print "imageSpws=%s, cubeSpws=%s" % (imageSpws, cubeSpws)

    if (refant=='' and newpprname=='' and wvrflag=='' and stopAfterCalibration==False
        and phaseupbw=='' and tie==True and wvrqa2intent=='' 
        and wvrgcalflag=='' and (asdm=='' or asdm==['']) and resolvedcals==''
        and bpsolint=='' and toffset==None and vis==[]
        and tsysflag_shape_fhi_limit=='' and overwrite==False
        and stopaftertsysflag==False and usems==False
        and stopaftertsyscal==False
        and stopafterimport==False and skipwvr==False and flagdata==False
        and skipflagdata==False and stopafterwvr==False
        and wvrsmooth=='' and fluxdensity==None and wvrgcalflag_intent==''
        and wvr_fhi_limit != '' and flag_sharps=='' and flag_sharps2==''
        and flag_edges==False and sharps_limit==default_sharps_limit
        and sharps2_limit==default_sharps2_limit 
        and edge_limit=='' and metric=='antenna_median'
        and startbeforetsysflagchans==False and tsysflag_intentgroups==''
        and fluxcal_linesfile=='' and flag_sharps2b=='' and metricb=='median'
        and sharps2b_limit==0.05 and atmflag_nmedian==False
        and atmflag_intent=='' and tsysflagchans_intentgroups==''
        and bpflagchans==False and fitfunc==None and fitorder==None
        and edge==None and tsysChanTol==None and asis==''
        and maxdistm==None and minnumants==None and bdfflags==None
        and dbservice==None and tsysNormalize==None and len(imageSources)==0
        and len(imageSpws)==0 and len(cubeSpws)==0 and not flagbackup
        and nbins=='' and clearcals and moveBpBefore=='' 
        and not skipcontsub):
        print "WARNING: No changes were requested.  Stopping."
        return
    if (tiesources != '' and tie==True):
        print "ERROR: You have set a value for tiesources, but tie==True (meaning automatic)."
        print "ERROR: You need to set tie=False to change from automatic to manual specification of the tie parameter."
        return
    if (fluxdensity != None and spw == None):
        print "ERROR: You must specify the spw parameter if you specify the fluxdensity parameter"
        return
    if (fluxdensity == None and spw != None):
        print "ERROR: You must specify the fluxdensity parameter if you specify the spw parameter"
        return
    if (fluxdensity != None and field == None):
        print "ERROR: You must specify the field parameter if you specify the fluxdensity parameter"
        return
    if (stopafterimport and newpprname==''):
        # Don't bother putting this in the filename
        stopaftertsysflag = False
    if (os.path.exists(ppr)==False):
        print "ERROR: Cannot find ppr file = ", ppr
        return
    if (edge != None):
        if (type(edge) != list):
            print "ERROR: The edge parameter must be a list of length 2"
            return
        if (len(edge) != 2):
            print "ERROR: The edge parameter must be a list of length 2"
            return


    oldppr = open(ppr, 'r')
    if (type(asdm) == list):
        asdm = ",".join(asdm)
    asdms = asdm.replace(' ','').split(',')
    if (asdms != ['']):
        print "asdms to search for: ", asdms
    if (newpprname == ''):
        suffix = ''
        if (overwrite):
            addToSuffix = '__overwrite'
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (usems):
            addToSuffix = '__usems_%s' % (usems)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (asdmDiskName != ''):
            addToSuffix = '__asdmDiskName_%s' % (asdmDiskName)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (refant != ''):
            addToSuffix = '__refant_%s' % (refant)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (tsysflag_shape_fhi_limit != ''):
            addToSuffix = '__tsysflag_shape_fhi_limit_%g' % (tsysflag_shape_fhi_limit)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (startbeforetsysflagchans):
            addToSuffix = '__startbeforetsysflagchans'
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (stopaftertsysflag):
            addToSuffix = '__stopaftertsysflag'
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (tsysChanTol != None):
            tsysChanTol = int(tsysChanTol)
            addToSuffix = '__tsysChanTol_%d' % (tsysChanTol)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (stopaftertsyscal):
            addToSuffix = '__stopaftertsyscal'
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (stopafterwvr):
            addToSuffix = '__stopafterwvr'
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (fluxcal_linesfile != ''):
            addToSuffix = '__fluxcal_linesfile__%s' % (fluxcal_linesfile)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (stopafterimport):
            addToSuffix = '__stopafterimport'
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (asis != ''):
            addToSuffix = '__asis_%s' % (asis.replace('*','star'))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (wvrflag != ''):
            addToSuffix = '__wvrflag_%s' % (wvrflag)
            addToSuffix.replace(',','_')
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (phaseupbw != ''):
            addToSuffix = '__phaseupbw_%s' % (phaseupbw)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (atmflag_nmedian != False):
            if (atmflag_intent == ''):
                addToSuffix = '__atmflag_nmedian_%g' % (atmflag_nmedian)
            else:
                addToSuffix = '__atmflag_nmedian_%g_intent_%s' % (atmflag_nmedian,atmflag_intent)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (tie == False):
            if (tiesources == ''):
                addToSuffix = '__no_tie' % ()
            else:
                addToSuffix = '__tie_manual' % ()
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (skipflagdata):
            addToSuffix = '__skipflagdata' % ()
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (flagbackup):
            addToSuffix = '__flagbackup' % ()
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (skipwvr):
            addToSuffix = '__skipwvr' % ()
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (wvrsmooth != ''):
            addToSuffix = '__wvrsmooth_%s' % (str(wvrsmooth))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (flagdata):
            addToSuffix = '__flagdata_%s' % (str(flagdata))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (wvrqa2intent != ''):
            addToSuffix = '__wvrqa2_%s' % (wvrqa2intent.replace('*','-'))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (wvrgcalflag == True):
            addToSuffix = '__wvrgcalflag' % ()
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (maxdistm != None):
            addToSuffix = '__maxdistm_%g' % (maxdistm)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (minnumants != None):
            addToSuffix = '__minnumants_%d' % (minnumants)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (wvr_fhi_limit != ''):
            addToSuffix = '__wvr_fhi_limit_%.2f' % (wvr_fhi_limit)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (wvrgcalflag_intent != ''):
            addToSuffix = '__wvrgcalflag_intent_%s' % (wvrgcalflag_intent.replace('*','-'))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (toffset != None):
            addToSuffix = '__toffset_%g' % (toffset)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (resolvedcals != ''):
            addToSuffix = '__resolvedcals_%s' % (resolvedcals)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (bpsolint != ''):
            addToSuffix = '__bpsolint_%s' % (bpsolint)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (bpflagchans):
            addToSuffix = '__bpflagchans'
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (edge_limit != ''):
            addToSuffix = '__edge_limit_%g' % (edge_limit)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (sharps_limit != default_sharps_limit):
            addToSuffix = '__sharps_limit_%g' % (sharps_limit)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (sharps2_limit != default_sharps2_limit):
            addToSuffix = '__sharps2_limit_%g' % (sharps2_limit)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (metric != 'antenna_median'):
            addToSuffix = '__metric_%s' % (str(metric))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (metricb != 'median'):
            addToSuffix = '__metricb_%s' % (str(metricb))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (flag_sharps2b != ''):
            addToSuffix = '__flag_sharps2b_%s' % (str(flag_sharps2b))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (sharps2b_limit != ''):
            addToSuffix = '__sharps2b_limit_%g' % (sharps2b_limit)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (flag_edges != False):
            addToSuffix = '__flag_edges_%s' % (str(flag_edges))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (bdfflags != None):
            addToSuffix = '__bdfflags_%s' % (str(bdfflags))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (not clearcals):
            addToSuffix = '__clearcals_False'
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (tsysNormalize != None):
            addToSuffix = '__tsysNormalize_%s' % (str(tsysNormalize))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (dbservice != None):
            addToSuffix = '__dbservice_%s' % (str(dbservice))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (flag_sharps != ''):
            addToSuffix = '__flag_sharps_%s' % (str(flag_sharps))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (flag_sharps2 != ''):
            addToSuffix = '__flag_sharps2_%s' % (str(flag_sharps2))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (tsysflag_intentgroups != ''):
            addToSuffix = '__tsysflag_intentgroups_%s' % (tsysflag_intentgroups.replace('*',''))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (tsysflagchans_intentgroups != ''):
            addToSuffix = '__tsysflagchans_intentgroups_%s' % (tsysflagchans_intentgroups.replace('*',''))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (vis != []):
            if (type(vis) == str):
                vis = [vis]
            addToSuffix = '__vis'
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (fluxdensity != None):
            if (type(fluxdensity) != list):
                fluxdensity = [[fluxdensity,0,0,0]]
            else: # it is a list, assumed to be I for each spw
                fds = []
                for fd in fluxdensity:
                    fds.append([fd,0,0,0])
                fluxdensity = fds
            addToSuffix = '__fluxdensity_%g' % (fluxdensity[0][0])
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (stopAfterCalibration == True):
            addToSuffix = '__no_imaging'
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (fitfunc != None):
            addToSuffix = '__fitfunc_%s' % (fitfunc)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (fitorder != None):
            addToSuffix = '__fitorder_%d' % (fitorder)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (len(imageSources) > 0):
            addToSuffix = '__image%dsources' % (len(imageSources))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (len(imageSpws) > 0):
            addToSuffix = '__image%dspws' % (len(imageSpws))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (len(cubeSpws) > 0):
            addToSuffix = '__cube%dspws' % (len(cubeSpws))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (len(nbins) > 0):
            addToSuffix = '__nbins_%s' % (nbins.replace(',','_').replace(' ','').replace(':','_'))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (edge != None):
            addToSuffix = '__edge_%d_%d' % (edge[0],edge[1])
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (skipcontsub):
            addToSuffix = '__skipcontsub'
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (moveBpBefore != ''):
            addToSuffix = '__moveBpBefore_%s' % (moveBpBefore)
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        if (asdm != '' and asdm != ['']):
            addToSuffix = '__%s' % (asdm.replace(' ','').replace(',','__').replace(':','_').replace('/','_'))
            if (ppr.find(addToSuffix) < 0):
                suffix += addToSuffix
        newpprname = ppr[:-4] + suffix + '.xml'
    elif newpprname == False:
        newpprname = ppr
    lines = oldppr.readlines()
    oldppr.close()
    oldpprlines = len(lines)
    if (len(lines) < 2):
        print "WARNING: This is a very short or blank PPR.  Stopping"
        return
    if (os.path.exists(newpprname)):
        print "WARNING: I am going to overwrite the existing file: %s" % (newpprname)
        if (newpprname == ppr):
            print "Copying the original PPR to %s.original" % (ppr)
            os.system('cp %s %s.original' % (ppr,ppr))
    newppr = open(newpprname, 'w')
    username = getpass.getuser()
    modified = au.unixTimeToDateString(au.getCurrentUnixTime())
    linectr = 0
    skiptoebi = False   # ebi = next ExecBlockId
    skiptoeop = False  # eop = end of Parameter
    skipToNextParameter = False
    skipToNextParameterEnd = False
    skiptoeops = False  # eops = end of ParameterSet
    skiptoeopc = False  # eopc = end of ProcessingCommand
    skiptoeopp = False  # eopp = end of ProcessingProcedure
    sawtsysflag = False
    sawwvrgcal = False
    processingCommandLine = None
    processingASDMLine = None
    processingTsysflag = False
    parameterSetOpen = False
    tsysflagIntentGroupsSet = False
    processingTsysflagShape = None
    wroteASDMLine = 0
    wroteAsdmDiskName = False
#    if plversion >= 28322:
#        HEURISTICS_VERSION=2014.02
#    elif plversion >= 26428:
#        HEURISTICS_VERSION=2013.09
#    else:
#        HEURISTICS_VERSION=2013.04 # or earlier - foundVer will be F
    findcont_taskname = 'hif_findcont'
    makeimages_taskname = 'hif_makeimages'
    if casadef.subversion_revision >= 28322:
        print "Using heurististics >= 2014.02"
        importdata_taskname = 'hifa_importdata'
        tsysflag_taskname = 'hifa_tsysflag'
        tsysflagchans_taskname = 'hifa_tsysflagchans'
        tsyscal_taskname = 'hifa_tsyscal'
        wvrgcal_taskname = 'hifa_wvrgcal'
        wvrgcalflag_taskname = 'hifa_wvrgcalflag'
        flagdata_taskname = 'hifa_flagdata'
        fluxcalflag_taskname = 'hifa_fluxcalflag'
        gfluxscale_taskname = 'hifa_gfluxscale'
        # tasks shared with VLA
        bandpass_taskname = 'hif_bandpass'
        bpflagchans_taskname = 'hif_bpflagchans'
        refant_taskname = 'hif_refant'
        setjy_taskname = 'hif_setjy'
        makeimlist_taskname = 'hif_makeimlist'
    else:
        print "Using old heurististics < 2014.02"
        importdata_taskname='hif_importdata'
        tsysflag_taskname = 'hif_tsysflag'
        tsysflagchans_taskname = 'hif_tsysflagchans'
        tsyscal_taskname = 'hif_tsyscal'
        wvrgcal_taskname = 'hif_wvrgcal'
        wvrgcalflag_taskname = 'hif_wvrgcalflag'
        flagdata_taskname = 'hif_flagdata'
        bandpass_taskname = 'hif_bandpass'
        bpflagchans_taskname = 'hif_bpflagchans'
        fluxcalflag_taskname = 'hif_fluxcalflag'
        gfluxscale_taskname = 'hif_gfluxscale'
        refant_taskname = 'hif_refant'
        setjy_taskname = 'hif_setjy'
        makeimlist_taskname = 'hif_makeimlist'
    makeimlist = 0  # number of times the makeimlist task has been encountered as we read through the PPR
    uvcontfit_taskname = 'hif_uvcontfit'
    uvcontsub_taskname = 'hif_uvcontsub'
    firstImaging_taskname = 'hif_mstransform'
    removeImlist = -1
    for line in lines:
        if (debug):
            print "Read line: %s" % (line[:-1])
        if (skiptoeop):
            if (line.find('</Parameter')>=0):
                if (not skipToNextParameter):
                    writeToPPR(newppr,line)
                skiptoeop = False
            continue
        if skipToNextParameter and line.find('<Parameter>') >= 0:
            skipToNextParameter = False
            skipToNextParameterEnd = True
            continue
        if skipToNextParameterEnd and line.find('</Parameter>') >= 0:
            writeToPPR(newppr, line)
            skipToNextParameterEnd = False
            continue
        if (skiptoeopp):
            if (line.find('</ProcessingProcedure')>=0):
                writeToPPR(newppr, line, 'z')
                skiptoeopp = False
            continue
        if (skiptoeopc):
            if (line.find('</ProcessingCommand')>=0):
                skiptoeopc = False
            continue
        if (skiptoeops):
            # We wish to ignore the pre-existing lines of code after a <ParameterSet>
            # that we are modifying, until we reach the end of it.
            if (line.find('</ParameterSet>')>=0 or line.find('<ParameterSet/>')>=0):
                if (debug): print "setting processingCommandLine = None"
                processingCommandLine = None
                skiptoeops = False
            continue
        if (skiptoebi):
            if (line.find('ExecBlockId') > 0 or line.find('SchedBlockIdentifier') > 0):
                if (line.find('SchedBlockIdentifier') > 0):
                    processingASDMLine = None
                    skiptoebi = False
                else:
                    skiptoebi = False
                    # It might be the one we want, so keep processing it
            else:
                continue
        if (processingCommandLine != None):
            # If we have seen a <ProcessingCommand> line, then
            # check to see if we are supposed to stop after this command.  If so,
            # then don't print any lines until we reach
            # the </ProcessingCommand> line, which we will also not print.
            if (line.find(tsysflag_taskname)>0):
                sawtsysflag = True
            if (line.find(wvrgcal_taskname)>0):
                sawwvrgcal = True
            if ((line.find(wvrgcal_taskname)>0 and skipwvr) or
                (line.find(flagdata_taskname)>0 and skipflagdata) or
                ((line.find(findcont_taskname)>0 or
                  line.find(uvcontfit_taskname)>0 or
                  line.find(uvcontsub_taskname)>0) 
                  and skipcontsub) or
                ((line.find(tsysflag_taskname+'<')>0 or
                  line.find(fluxcalflag_taskname)>0 or
                  line.find(flagdata_taskname)>0 or
                  line.find(refant_taskname)>0)
                and startbeforetsysflagchans) 
                ):
                if line.find(uvcontsub_taskname) > 0:
                    removeImlist = 4
                skiptoeopc = True
                processingCommandLine = None
                if (debug): print "Setting processingCommandLine = None"
                continue
            elif (((line.find(setjy_taskname)>0 or line.find('hifa_antpos')>0) and stopaftertsysflag) or
                  (line.find(setjy_taskname)>0 and stopafterwvr)  or
                  (line.find(tsysflag_taskname)>0 and stopaftertsyscal)  or
                  (line.find(wvrgcal_taskname)>0 and stopaftertsysflag and sawtsysflag) or
                  (line.find(flagdata_taskname)>0 and stopafterimport) or
                  (line.find(firstImaging_taskname)>0 and stopAfterCalibration) or
                  (line.find(tsyscal_taskname)>0 and stopafterwvr and sawwvrgcal)
                  ): 
                skiptoeopp = True
                needToWritePCLine = False
                if (debug): print "Setting processingCommandLine = None"
                continue
            elif (moveBpBefore != '' and (line.find(moveBpBefore)>0 or line.find('breakpoint')>0)):
                if line.find('breakpoint') > 0:
                    skiptoeopc = True
                    needToWritePCLine = False
                    continue
                elif (line.find(moveBpBefore) > 0):
                    writeToPPR(newppr, '                <ProcessingCommand>\n')
                    writeToPPR(newppr, '                    <Command>breakpoint</Command>\n')
                    writeToPPR(newppr, '                    <ParameterSet/>\n')
                    writeToPPR(newppr, '                </ProcessingCommand>\n')
                    writeToPPR(newppr, '                <ProcessingCommand>\n')
            elif (needToWritePCLine):
                # If it is not, then write out the prior <ProcessingCommand> line normally,
                # then go on to write out the new parameter set.
                if (makeimlist != removeImlist):
                    if debug: print "writing because needToWritePCLine"
                    writeToPPR(newppr, processingCommandLine, 'a')
                needToWritePCLine = False
            else:
                # normally, you only get here when line=='<ParameterSet>'
                if (line.find("<ProcessingCommand>") < 0):
                    if (processingTsysflag):
                        if (line.find("</ProcessingCommand>")>=0 or
                            line.find("<ParameterSet/>")>=0):
                            if (debug): print "Setting processingTsysflag = False"
                            processingTsysflag = False
                            tsysflagIntentGroupsSet = False
                            if (line.find("<ParameterSet/>") >= 0 and tsysflag_intentgroups != ''):
                                # There were no prior parameters in this hif_tsysflag
                                writeToPPR(newppr, line.replace('<ParameterSet/>', '<ParameterSet>'))
                                insertIntentGroups(newppr, tsysflag_intentgroups)
                                writeToPPR(newppr, line.replace('<ParameterSet/>', '</ParameterSet>'))
                                continue
                        elif (line.find("<ParameterSet>") < 0):
                            writeToPPR(newppr, line, 'B')
                if (line.find("</ProcessingCommand>") >=0 and tsysflag_intentgroups == ''):
                    continue
        # endif processingCommandLine
        if (processingASDMLine != None):
            if (line.find('ExecBlockId') < 0):
                processingASDMLine2 = line
                continue
        if (line.find(refant_taskname) > 0 and refant != ''):
            writeToPPR(newppr, line)
            writeToPPR(newppr,'                    <ParameterSet>\n')
            writeToPPR(newppr,'                        <Parameter>\n')
            writeToPPR(newppr,'                            <Keyword>hm_refant</Keyword>\n')
            writeToPPR(newppr,'                            <Value>manual</Value>\n')
            writeToPPR(newppr,'                        </Parameter>\n')
            writeToPPR(newppr,'                        <Parameter>\n')
            writeToPPR(newppr,'                            <Keyword>refant</Keyword>\n')
            writeToPPR(newppr,'                            <Value>%s</Value>\n'%refant)
            writeToPPR(newppr,'                        </Parameter>\n')
            writeToPPR(newppr,'                    </ParameterSet>\n')
            skiptoeops = True
        elif (line.find(flagdata_taskname) > 0 and (flagdata!=False or flagbackup)):
            writeToPPR(newppr, line, 'f')
            writeToPPR(newppr,'                    <ParameterSet>\n')
            if flagdata != False:
                insertFlagData(newppr, flagdata)
            if flagbackup:
                insertFlagBackup(newppr)
            writeToPPR(newppr,'                    </ParameterSet>\n')
            skiptoeops = True
        elif (line.find(tsyscal_taskname) > 0 and tsysChanTol != None):
            writeToPPR(newppr, line, 'tc')
            insertTsysChanTol(newppr, tsysChanTol)
            skiptoeops = True
        elif (line.find(fluxcalflag_taskname) > 0 and fluxcal_linesfile != ''):
            writeToPPR(newppr, line,'f')
            writeToPPR(newppr,'                    <ParameterSet>\n')
            writeToPPR(newppr,'                        <Parameter>\n')
            writeToPPR(newppr,'                            <Keyword>appendlines</Keyword>\n')
            writeToPPR(newppr,'                            <Value>True</Value>\n')
            writeToPPR(newppr,'                        </Parameter>\n')
            writeToPPR(newppr,'                        <Parameter>\n')
            writeToPPR(newppr,'                            <Keyword>linesfile</Keyword>\n')
            writeToPPR(newppr,'                            <Value>%s</Value>\n'%fluxcal_linesfile)
            writeToPPR(newppr,'                        </Parameter>\n')
            writeToPPR(newppr,'                    </ParameterSet>\n')
            skiptoeops = True
        elif (line.find(wvrgcal_taskname) > 0):
          if (wvrgcalflag == True):
              line = line.replace(wvrgcal_taskname, wvrgcalflag_taskname)
          elif (wvrgcalflag == False):
              line = line.replace(wvrgcalflag_taskname, wvrgcal_taskname)
          writeToPPR(newppr,line,'w')
          if (wvrflag != '' or tie==False or toffset!=None or
              wvrqa2intent != '' or wvrgcalflag!='' or wvrsmooth != '' or
              wvrgcalflag_intent != '' or wvr_fhi_limit != '' or
              minnumants != None or maxdistm != None):
            writeToPPR(newppr,'                    <ParameterSet>\n','w')
            if (wvrsmooth != ''):
                if (wvrsmooth == 'automatic'):
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>hm_smooth</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>automatic</Value>\n')
                    writeToPPR(newppr,'                        </Parameter>\n')
                else:
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>hm_smooth</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>manual</Value>\n')
                    writeToPPR(newppr,'                        </Parameter>\n')
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>smooth</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>%s</Value>\n',str(wvrsmooth))
                    writeToPPR(newppr,'                        </Parameter>\n')
            if (wvr_fhi_limit != ''):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>fhi_limit</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%f</Value>\n' % (wvr_fhi_limit))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (wvrqa2intent != ''):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>qa2_intent</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n' % (wvrqa2intent))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (wvrgcalflag_intent != ''):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>flag_intent</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n' % (wvrgcalflag_intent))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (wvrflag != ''):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>wvrflag</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n'%(wvrflag))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (toffset != None):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>toffset</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%g</Value>\n'%(toffset))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (tie == False):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>hm_tie</Keyword>\n')
                writeToPPR(newppr,'                            <Value>manual</Value>\n')
                writeToPPR(newppr,'                        </Parameter>\n')
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>tie</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n' % (tiesources))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (maxdistm != None):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>maxdistm</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%g</Value>\n' % (maxdistm))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (minnumants != None):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>minnumants</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%d</Value>\n' % (minnumants))
                writeToPPR(newppr,'                        </Parameter>\n')
            writeToPPR(newppr,'                    </ParameterSet>\n')
            skiptoeops = True
        elif (line.find(importdata_taskname) > 0 and (vis != [] or overwrite or asis != '' or bdfflags != None or dbservice != None or not clearcals)):
            writeToPPR(newppr,line)
            writeToPPR(newppr,'                    <ParameterSet>\n')
            if (overwrite):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>overwrite</Keyword>\n')
                writeToPPR(newppr,'                            <Value>True</Value>\n')
                writeToPPR(newppr,'                        </Parameter>\n')
            if (not clearcals):  # default is True
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>clearcals</Keyword>\n')
                writeToPPR(newppr,'                            <Value>False</Value>\n')
                writeToPPR(newppr,'                        </Parameter>\n')
            if (bdfflags != None): # default is True
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>bdfflags</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n' % (str(bdfflags)))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (dbservice != None):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>dbservice</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n' % (str(dbservice)))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (vis != []):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>vis</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n'%(vis))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (asis != ''):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>asis</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n'%(asis))
                writeToPPR(newppr,'                        </Parameter>\n')
            writeToPPR(newppr,'                    </ParameterSet>\n')
            if (atmflag_nmedian != False):
                insertAtmFlag(newppr,atmflag_nmedian,atmflag_intent)
            skiptoeops = True
        elif (line.find(importdata_taskname) > 0 and atmflag_nmedian != False):
            writeToPPR(newppr,line)
            writeToPPR(newppr,'                    <ParameterSet/>\n')
            writeToPPR(newppr,'                </ProcessingCommand>\n')
            insertAtmFlag(newppr,atmflag_nmedian,atmflag_intent)
            skiptoeops = True
        elif (line.find(tsysflagchans_taskname) > 0 and (flag_sharps or flag_sharps2 or
                                                      edge_limit!='' or tsysflagchans_intentgroups != '')):
            writeToPPR(newppr,line)
            if (edge_limit == '' and tsysflagchans_intentgroups == ''):
                writeToPPR(newppr,'                    <ParameterSet/>\n')
            else:
                # Change the existing tsysflagchans stage, which only flags edge channels
                writeToPPR(newppr,'                    <ParameterSet>\n')
                if (edge_limit != ''):
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>edge_limit</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>%g</Value>\n'%(edge_limit))
                    writeToPPR(newppr,'                        </Parameter>\n')
                if (tsysflagchans_intentgroups != ''):
                    insertIntentGroups(newppr, tsysflagchans_intentgroups)
                writeToPPR(newppr,'                    </ParameterSet>\n')
            if (flag_sharps or flag_sharps2):
                # Insert a new tsysflagchans stage
                writeToPPR(newppr,'                </ProcessingCommand>\n')
                writeToPPR(newppr,'                <ProcessingCommand>\n')
                writeToPPR(newppr,'                    <Command>%s</Command>\n' % tsysflagchans_taskname)
                writeToPPR(newppr,'                    <ParameterSet>\n')
                if (flag_edges != ''):
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>flag_edges</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>%s</Value>\n'%(flag_edges))
                    writeToPPR(newppr,'                        </Parameter>\n')
                if (flag_sharps != ''):
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>flag_sharps</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>%s</Value>\n'%(flag_sharps))
                    writeToPPR(newppr,'                        </Parameter>\n')
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>sharps_limit</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>%g</Value>\n'%(sharps_limit))
                    writeToPPR(newppr,'                        </Parameter>\n')
                if (flag_sharps2 != ''):
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>flag_sharps2</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>%s</Value>\n'%(flag_sharps2))
                    writeToPPR(newppr,'                        </Parameter>\n')
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>sharps2_limit</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>%g</Value>\n'%(sharps2_limit))
                    writeToPPR(newppr,'                        </Parameter>\n')
                if (metric != ''):
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>metric</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>%s</Value>\n'%(metric))
                    writeToPPR(newppr,'                        </Parameter>\n')
                if (tsysflagchans_intentgroups != ''):
                    insertIntentGroups(newppr, tsysflagchans_intentgroups)
                writeToPPR(newppr,'                    </ParameterSet>\n')
                if (flag_sharps2b != ''):
                    # Insert a second new tsysflagchans stage
                    writeToPPR(newppr,'                </ProcessingCommand>\n')
                    writeToPPR(newppr,'                <ProcessingCommand>\n')
                    writeToPPR(newppr,'                    <Command>%s</Command>\n' % tsysflagchans_taskname)
                    writeToPPR(newppr,'                    <ParameterSet>\n')
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>flag_edges</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>False</Value>\n')
                    writeToPPR(newppr,'                        </Parameter>\n')
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>metric</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>%s</Value>\n' % (metricb))
                    writeToPPR(newppr,'                        </Parameter>\n')
                    writeToPPR(newppr,'                        <Parameter>\n')
                    writeToPPR(newppr,'                            <Keyword>flag_sharps2</Keyword>\n')
                    writeToPPR(newppr,'                            <Value>%s</Value>\n' % str(flag_sharps2b))
                    writeToPPR(newppr,'                        </Parameter>\n')
                    if (tsysflagchans_intentgroups != ''):
                        insertIntentGroups(newppr, tsysflagchans_intentgroups)
                    if (sharps2b_limit != ''):
                        writeToPPR(newppr,'                        <Parameter>\n')
                        writeToPPR(newppr,'                            <Keyword>sharps2_limit</Keyword>\n')
                        writeToPPR(newppr,'                            <Value>%g</Value>\n'%(sharps2b_limit))
                        writeToPPR(newppr,'                        </Parameter>\n')
                    writeToPPR(newppr,'                    </ParameterSet>\n')
            skiptoeops = True
        elif (line.find('intentgroups') > 0 and processingTsysflag and tsysflag_intentgroups != ''):
            # Change any existing intentgroups to the newly-specified one
            writeToPPR(newppr,line,'T')
            newline = line.split('<Keyword>')[0] + "<Value>['%s']</Value>\n" % (tsysflag_intentgroups)
            writeToPPR(newppr,newline,'T')
            tsysflagIntentGroupsSet = True
            skiptoeop = True
        elif (line.find(bandpass_taskname) > 0 and (phaseupbw != '' or bpsolint != '')):
            writeToPPR(newppr,line)
            writeToPPR(newppr,'                    <ParameterSet>\n')
            if (phaseupbw != ''):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>phaseupbw</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n'%(phaseupbw))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (bpsolint != ''):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>solint</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n'%(bpsolint))
                writeToPPR(newppr,'                        </Parameter>\n')
            writeToPPR(newppr,'                    </ParameterSet>\n')
            skiptoeops = True
        elif (line.find(setjy_taskname) > 0 and fluxdensity != None and spw != None):
            for s in range(len(spw)):
                writeToPPR(newppr,line)
                writeToPPR(newppr,'                    <ParameterSet>\n')
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>fluxdensity</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n' % (str(fluxdensity[s])))
                writeToPPR(newppr,'                        </Parameter>\n')
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>spw</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n' % (str(spw[s])))
                writeToPPR(newppr,'                        </Parameter>\n')
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>field</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n' % (field))
                writeToPPR(newppr,'                        </Parameter>\n')
                writeToPPR(newppr,'                    </ParameterSet>\n')
                if (s < len(spw)-1):
                    writeToPPR(newppr,'                </ProcessingCommand>\n')
                    writeToPPR(newppr,'                <ProcessingCommand>\n')
            skiptoeops = True
        elif (line.find(makeimlist_taskname) > 0):
            makeimlist += 1
            print "Working on %s %d" % (makeimlist_taskname, makeimlist)
            if (removeImlist != makeimlist):
                writeToPPR(newppr,line)
                if ((makeimlist > 1 and len(imageSources) > 0) or
                    (makeimlist in [2,3] and len(imageSpws)>0) or 
                    (makeimlist in [2,4] and len(nbins) > 0) or
                    (makeimlist==4 and len(cubeSpws)>0)):
                    writeToPPR(newppr,'                    <ParameterSet>\n')
                    if (len(imageSources) > 0):
                        writeToPPR(newppr,'                        <Parameter>\n')
                        writeToPPR(newppr,'                            <Keyword>field</Keyword>\n')
                        fields = addDoubleQuotesToFieldList(imageSources)
                        writeToPPR(newppr,'                            <Value>%s</Value>\n'%fields)
                        writeToPPR(newppr,'                        </Parameter>\n')
                    if (len(imageSpws) > 0 and makeimlist<4):
                        writeToPPR(newppr,'                        <Parameter>\n')
                        writeToPPR(newppr,'                            <Keyword>spw</Keyword>\n')
                        ispws = ','.join(imageSpws)
                        writeToPPR(newppr,'                            <Value>%s</Value>\n'%ispws)
                        writeToPPR(newppr,'                        </Parameter>\n')
                    if (len(cubeSpws) > 0 and makeimlist==4):
                        writeToPPR(newppr,'                        <Parameter>\n')
                        writeToPPR(newppr,'                            <Keyword>spw</Keyword>\n')
                        ispws = ','.join(cubeSpws)
                        writeToPPR(newppr,'                            <Value>%s</Value>\n'%ispws)
                        writeToPPR(newppr,'                        </Parameter>\n')
                    if (len(nbins) > 0 and (makeimlist==2 or makeimlist==4)):
                        writeToPPR(newppr,'                        <Parameter>\n')
                        writeToPPR(newppr,'                            <Keyword>nbins</Keyword>\n')
                        writeToPPR(newppr,'                            <Value>%s</Value>\n' % nbins)
                        writeToPPR(newppr,'                        </Parameter>\n')
                    parameterSetOpen = True
            else:
                skiptoeopc = True
        elif (line.find(makeimages_taskname) > 0 and removeImlist == makeimlist):
            skiptoeopc = True
        elif ((line.find(uvcontfit_taskname) > 0 or line.find(uvcontsub_taskname) > 0 or line.find(findcont_taskname) > 0) and 
              (len(imageSources) > 0 or len(imageSpws) > 0)):
            writeToPPR(newppr,line,'uv')
            writeToPPR(newppr,'                    <ParameterSet>\n')
            if (len(imageSources) > 0):
                writeToPPR(newppr,'                        <Parameter>\n')
                fields = addDoubleQuotesToFieldList(imageSources)
                writeToPPR(newppr,'                            <Keyword>field</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n'%(fields))
                writeToPPR(newppr,'                        </Parameter>\n')
            contPlusCubeSpws = sorted(np.unique(imageSpws+cubeSpws), key=int)
            if (len(contPlusCubeSpws) > 0):
                writeToPPR(newppr,'                        <Parameter>\n')
                ispws = ','.join(contPlusCubeSpws)
                writeToPPR(newppr,'                            <Keyword>spw</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n'%(ispws))
                writeToPPR(newppr,'                        </Parameter>\n')
            writeToPPR(newppr,'                    </ParameterSet>\n')
            skiptoeops = True
        elif (line.find(gfluxscale_taskname) > 0 and (resolvedcals != '' or bpflagchans)):
            if (bpflagchans):
                loc = line.find("<Command") - 4
                space = max([loc,0])*' '
                writeToPPR(newppr,space+"    <Command>%s</Command>\n" % (bpflagchans_taskname))
                writeToPPR(newppr,space+"    <ParameterSet>\n")
                writeToPPR(newppr,space+"    </ParameterSet>\n")
                writeToPPR(newppr,space+"</ProcessingCommand>\n")
                writeToPPR(newppr,space+"<ProcessingCommand>\n")
            writeToPPR(newppr,line)
            if (resolvedcals != ''):
                writeToPPR(newppr,'                    <ParameterSet>\n')
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>hm_resolvedcals</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n'%(resolvedcals))
                writeToPPR(newppr,'                        </Parameter>\n')
                writeToPPR(newppr,'                    </ParameterSet>\n')
                skiptoeops = True
        elif (line.find(tsysflag_taskname) > 0 and (tsysflag_shape_fhi_limit != '' or tsysflag_intentgroups != '' or tsysNormalize!=None)):
            writeToPPR(newppr,line,'t')
            processingTsysflag = True
            if (tsysNormalize != None):  # CAS-8459
                writeToPPR(newppr,'                    <ParameterSet>\n')
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword xmlns="">normalize_tsys</Keyword>\n')
                writeToPPR(newppr,'                            <Value xmlns="">%s</Value>\n'%(tsysNormalize))
                writeToPPR(newppr,'                        </Parameter>\n')
                writeToPPR(newppr,'                    </ParameterSet>\n')
                skiptoeops = True
        elif (line.find('<RelativePath>') > 0):
            # This section fixes a PPR that you copied from some other project to point to the current working dir
            relativePath = line.split('<RelativePath>')[1].split('</RelativePath>')[0]
            print "Searching for %s\n in\n %s" % (relativePath, os.getcwd())
            if os.getcwd().find(relativePath) < 0:
                print "Did not find %s in %s" % (os.getcwd(), relativePath)
                if fixpath:
                    relativePath = '<RelativePath>'+os.getcwd().split('pipeline/root/')[1]+'</RelativePath>'
                    print "Fixing the path to ", relativePath
                    writeToPPR(newppr,'                        '+relativePath+'\n', 'r')
                else:
                    print "Warning: the relativePath in the PPR is not in the current path.  If this PPR came from a different project, then you can try to run with fixpath=True."
                    writeToPPR(newppr,line)
            else:
                print "Found it"
                writeToPPR(newppr,line)
        elif (line.find('AsdmDiskName') > 0):
            if (asdmDiskName != ''):  
                uid = asdmDiskName  # use the name specified on the command line
            else:
                uid = line.split("<AsdmDiskName>")[1].split("</AsdmDiskName>")[0]
            if (line.find('.ms') > 0 and usems==False):
                # Need to strip off the existing .ms
                uid = uid.split(".ms</AsdmDiskName>")[0]
                line = '                            <AsdmDiskName>%s</AsdmDiskName>\n' % (uid)
            elif (uid.find('.ms') < 0 and usems):
                # Need to insert the .ms
                uid = uid.split("</AsdmDiskName>")[0]
                line = '                            <AsdmDiskName>%s.ms</AsdmDiskName>\n' % (uid)
            else:
                line = '                            <AsdmDiskName>%s</AsdmDiskName>\n' % (uid)
            writeToPPR(newppr, line)
        elif (line.find('ExecBlockId') > 0 and (asdm != '') and (asdm != [''])):
            # Assume at first that none of them are the ASDM(s) we want.
            skiptoebi = True
            for oneasdm in asdms:
                if (line.find(oneasdm) > 0):
                    # This is one of the ASDM(s) that we want, so write it out.
                    # The rest will be written out in the loop in line 8875
                    writeToPPR(newppr,processingASDMLine)  # AsdmIdentifier
                    writeToPPR(newppr,processingASDMLine2)  # AsdmRef
                    writeToPPR(newppr,line)
                    processingASDMLine = None
                    wroteASDMLine += 1
                    skiptoebi = False
                    asdms.remove(oneasdm) # so we don't write it again in the later loop in line 8875
                    break
        elif (line.find('hsd_baseline') > 0 and (fitfunc != None or fitorder != None or edge != None)):
            # Here we assume that no other hsd_Baseline parameters are set in the original PPR.
            writeToPPR(newppr,line)
            writeToPPR(newppr,    '                    <ParameterSet>\n')
            if (edge != None):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>edge</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n' % (str(edge)))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (fitfunc != None):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>fitfunc</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%s</Value>\n' % (fitfunc))
                writeToPPR(newppr,'                        </Parameter>\n')
            if (fitorder != None):
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>fitorder</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%d</Value>\n' % (fitorder))
                writeToPPR(newppr,'                        </Parameter>\n')
            writeToPPR(newppr,    '                    </ParameterSet>\n')
            skiptoeops = True
        elif (asdm != '' and asdm != [''] and line.find('</SchedBlockIdentifier>') >= 0):
            for myasdm in asdms:
                writeToPPR(newppr,'                        <AsdmIdentifier>\n')
                writeToPPR(newppr,'                            <AsdmRef>\n')
                writeToPPR(newppr,'                                <ExecBlockId>%s</ExecBlockId>\n'% au.uidToSlashes(myasdm))
                writeToPPR(newppr,'                            </AsdmRef>\n')
                writeToPPR(newppr,'                            <AsdmDiskName>%s</AsdmDiskName>\n' % au.uidToUnderscores(myasdm))
                writeToPPR(newppr,'                        </AsdmIdentifier>\n')
            writeToPPR(newppr,'                    </SchedBlockIdentifier>\n')
        else:
            if (line.find('<ProcessingCommand>') >= 0):
                if (debug): print "Setting processingCommandLine = True"
                processingCommandLine = line
                needToWritePCLine = True
            elif (line.find('<AsdmIdentifier>') >= 0 and asdm != ''):
                processingASDMLine = line
            elif (line.find('<Value>shape</Value>') >= 0 and tsysflag_shape_fhi_limit != ''):
                processingTsysflagShape = True
                writeToPPR(newppr,line,'s')
            elif (processingTsysflagShape):
                writeToPPR(newppr,'                        </Parameter>\n','T')
                writeToPPR(newppr,'                        <Parameter>\n')
                writeToPPR(newppr,'                            <Keyword>fhi_limit</Keyword>\n')
                writeToPPR(newppr,'                            <Value>%g</Value>\n'%(tsysflag_shape_fhi_limit))
                writeToPPR(newppr,'                        </Parameter>\n')
                processingTsysflagShape = None
            else:
                # It is not a new "<ProcessingCommand>", so just write it out
                processingCommandLine = None
                if (debug): print "Setting processingCommandLine = None"
                if (line.find('</ParameterSet>') >= 0 and processingTsysflag and
                    tsysflag_intentgroups != '' and tsysflagIntentGroupsSet == False):
                    insertIntentGroups(newppr, tsysflag_intentgroups)
                    tsysflagIntentGroupsSet = True
                if (not parameterSetOpen or line.find('<ParameterSet>') < 0):
                    if (line.find('<Keyword>field') > 0 and parameterSetOpen):
                        skiptoeop = True
                        skipToNextParameter = True
                    else:
                        if (line.find('<ProjectSummary>') >= 0):
                            writeToPPR(newppr,'<!-- modified by %s at %s using amc v%s:\n' % (username,modified,version().split()[2]))
                            writeToPPR(newppr,'  %s\n-->\n' % (commandLine))
                        writeToPPR(newppr,line,'p')
                    if line.find('</ParameterSet>') >= 0:
                        parameterSetOpen = False
                if (line.find('</ProcessingCommand>') >= 0 and processingTsysflag):
                    if (debug): print "Setting processingTsysflag = False and bottom of the 'for' loop"
                    processingTsysflag = False
                    tsysflagIntentGroupsSet = False
    newppr.close()
    if (wroteASDMLine == 0):
        if (asdm != '' and asdm != ['']):
            print "WARNING: I did not find any ASDM from the list %s in the original PPR. If this is an E2E test of a CSV project, then this is expected." % (asdms)
    else:
        print "Kept %d ASDMs" % (wroteASDMLine)
    print "Original ppr had %d lines " % (oldpprlines)
    print "New ppr with %d lines left in: " % (linectr), newpprname
    oldppr = open(ppr, 'r')
    newppr = open(newpprname, 'r')
    mydiff = difflib.SequenceMatcher(None, oldppr.read(), newppr.read())
    oldppr.close()
    newppr.close()
    if (mydiff.ratio() == 1.0 and newpprname != ppr):
        print "WARNING: the new PPR is identical to the old!"

def insertTsysChanTol(newppr, tsysChanTol):
    writeToPPR(newppr,'                        <Parameter>\n')
    writeToPPR(newppr,'                            <Keyword>tsysChanTol</Keyword>\n')
    writeToPPR(newppr,"                            <Value>%d</Value>\n"%(tsysChanTol))
    writeToPPR(newppr,'                        </Parameter>\n')
    
def insertIntentGroups(newppr, tsysflag_intentgroups):
    writeToPPR(newppr,'                        <Parameter>\n')
    writeToPPR(newppr,'                            <Keyword>intentgroups</Keyword>\n')
    writeToPPR(newppr,"                            <Value>['%s']</Value>\n"%(tsysflag_intentgroups))
    writeToPPR(newppr,'                        </Parameter>\n')

def insertAtmFlag(newppr, atmflag_nmedian, intent):
    writeToPPR(newppr,'                <ProcessingCommand>\n')
    writeToPPR(newppr,'                    <Command>hif_atmflag</Command>\n')
    writeToPPR(newppr,'                    <ParameterSet>\n')
    writeToPPR(newppr,'                        <Parameter>\n')
    writeToPPR(newppr,'                           <Keyword>flag_minabs</Keyword>\n')
    writeToPPR(newppr,'                           <Value>False</Value>\n')
    writeToPPR(newppr,'                        </Parameter>\n')
    writeToPPR(newppr,'                        <Parameter>\n')
    writeToPPR(newppr,'                           <Keyword>flag_nmedian</Keyword>\n')
    writeToPPR(newppr,'                           <Value>True</Value>\n')
    writeToPPR(newppr,'                        </Parameter>\n')
    writeToPPR(newppr,'                        <Parameter>\n')
    writeToPPR(newppr,'                           <Keyword>fnm_limit</Keyword>\n')
    writeToPPR(newppr,'                           <Value>%g</Value>\n' % (atmflag_nmedian))
    writeToPPR(newppr,'                        </Parameter>\n')
    if (intent != ''):
        writeToPPR(newppr,'                        <Parameter>\n')
        writeToPPR(newppr,'                           <Keyword>intent</Keyword>\n')
        writeToPPR(newppr,'                           <Value>%s</Value>\n' % (intent))
        writeToPPR(newppr,'                        </Parameter>\n')
    writeToPPR(newppr,'                    </ParameterSet>\n')

def detectOutliers(values, threshold=10., edgetrim=5):
    """
    Detect outlying samples from an array of values.

    threshold: if a value is deviated from the mean by more than
               threshold*sigma, it is identified as an outlier,
               where mean and sigma are not "raw" ones but
               "robust" ones (see below)
    edgetrim: minimum and maximum 1/edgetrim samples are ignored
              to determine robust mean and sigma
    """
    n_samples = len(values)
    subsample = pl.sort(values)[n_samples/edgetrim:-n_samples/edgetrim]
    std = subsample.std()
    mean = subsample.mean()
    return (pl.fabs(values-mean) > threshold*std)

def mjd2iso8601(mjd_in_sec):
    """
    Convert MJD (in seconds) to ISO8601 string.
    """
    return time.strftime("%Y-%m-%dT%H:%M:%S", time.gmtime(mjd_in_sec-3506716800.))

def getAntNames(msname):
    """
    Read antenna names in msname.
    """
    tb.open('%s/ANTENNA' % msname)
    antnames = tb.getcol('NAME')
    tb.close()
    return antnames

def readCoeffs(bpcaltable, spw='1', nants=2):
    """
    Read polynomial coeffs and timestamps of bandpass table.
    """
    # obtain CAL_DESC_IDs corresponding to spws.
    # should be [0, 1, ...] ...
    spws = [int(item) for item in spw.split(',')]
    tb.open('%s/CAL_DESC' % bpcaltable)
    spwin = tb.getcol('SPECTRAL_WINDOW_ID').squeeze()
    tb.close()
    caldescs = [pl.arange(len(spwin))[spwin==s][0] for s in spws]

    coeffs = {}
    tb.open(bpcaltable)
    for i in range(nants):
        coeffs[i] = {}
        for j in range(len(spws)):
            q = tb.query('ANTENNA1==%d && CAL_DESC_ID==%d' % (i, caldescs[j]))
            coeffs[i][spws[j]] = {'timestamp': q.getcol('TIME').squeeze(),
                'acoeff': q.getcol('POLY_COEFF_AMP').squeeze(),
                'pcoeff': q.getcol('POLY_COEFF_PHASE').squeeze()}
    tb.close()
    return coeffs

def mad(a, c=0.6745, axis=0):
    """
    Median Absolute Deviation along given axis of an array:

    median(abs(a - median(a))) / c

    c = 0.6745 is the constant to convert from MAD to std; it is used by
    default

    """
    a = np.array(a)
    good = (a==a)
    a = np.asarray(a, np.float64)
    if a.ndim == 1:
        d = np.median(a[good])
        m = np.median(np.fabs(a[good] - d)) / c
#        print  "mad = %f" % (m)
    else:
        d = np.median(a[good], axis=axis)
        # I don't want the array to change so I have to copy it?
        if axis > 0:
            aswp = swapaxes(a[good],0,axis)
        else:
            aswp = a[good]
        m = np.median(np.fabs(aswp - d), axis=0) / c

    return m


#####################################################################################################
"""
delayJumpDetector.py: detect delay jumps by utilizing bandpass solution.
  -from T. Sawada.

usage:
import delayJumpDetector as djd
djd.detectDelayJumps('uid___XXX_XXX_XXX.ms')
"""
import os 
import sys
import time
import pylab as pl
from casa import *
from taskinit import *

def detectDelayJumps(msname='', spw='', scan='', refant=0,
                     intent='CALIBRATE_PHASE#ON_SOURCE'):
    """
    Detect delay jumps using bandpass(bandtype=BPOLY).

    Outliers in the amplitude and phase slopes of the bandpass solution
    are identified as delay jumps.
    spw: spws to be analyzed (can be a comma-separated list)
    scan: scans to be analyzed (can be a comma-separated list)
    refant: reference antenna ID (integer) for bandpass solution
    """

    if not os.path.exists(msname):
        sys.exit('ERROR: msname "%s" is not found.' % msname)

    antnames = getAntNames(msname)
    nants = len(antnames)
    if (spw == ''):
        if (casadef.casa_version >= '4.1.0'):
            mymsmd = createCasaTool(msmdtool)
            mymsmd.open(msname)
            phase_spw = mymsmd.spwsforintent(intent)
            wvrspws = mymsmd.wvrspws()
            spw = []
            for s in phase_spw:
                if (len(mymsmd.chanfreqs(s)) > 4 and s not in wvrspws):
                    spw.append(s)
            if (len(spw) < 1):
                print "ERROR: Found no non-wvr, multi-channel spws for intent=%s" % (intent)
                return
            print "Found %d non-wvr, multi-channel spws for intent=%s: %s" % (len(spw), str(spw), intent)
            spw = ','.join([str(i) for i in spw])
            mymsmd.close() 
        else:
            print "ERROR: You must specify 1 or more spws via the spw parameter."
            return
    
    bpcaltable = '%s.bpcal' % msname
    os.system('rm -rf %s' % bpcaltable)
    bandpass(vis = msname,
        caltable = bpcaltable,
        spw = spw,
        scan = scan,
        intent = intent,
        solint = 'int',
        refant = '%s' % refant,
        bandtype = 'BPOLY',
        degamp = 1,
        degphase = 1,
        visnorm = True,
        maskedge = 10)

    coeffs = readCoeffs(bpcaltable, spw, nants)
    spws = [int(item) for item in spw.split(',')]
    timestamp = coeffs[0][spws[0]]['timestamp']
    outliers = []
    # hereafter 2-pol spws are assumed
    for i in range(nants):
        for j in range(len(spws)):
            outX = (detectOutliers(coeffs[i][spws[j]]['acoeff'][1, :]) | \
                    detectOutliers(coeffs[i][spws[j]]['pcoeff'][1, :]))
            outY = (detectOutliers(coeffs[i][spws[j]]['acoeff'][3, :]) | \
                    detectOutliers(coeffs[i][spws[j]]['pcoeff'][3, :]))
            outliers.append(outX)
            outliers.append(outY)
    outliers = pl.array(outliers)
    olexist = outliers.sum(0, dtype=bool)
    antspwpol = pl.array([('%sspw%dpol%d' % (a, s, p)) \
        for a in antnames for s in spws for p in [0, 1]])
    if olexist.sum(dtype=bool):
        ts = timestamp[olexist]
        ol = outliers[:, olexist]
        for i in range(len(ts)):
            print mjd2iso8601(ts[i]), ', '.join(antspwpol[ol[:, i]])
    else:
        print 'no delay jumps were detected.'

