############################################################################
#
# plotbpass.py
#
#    plotbpass: Compares multiple bandpass ampiltude and phase solutions
#               observed with 1 or more basebands.  Designed to work for 2 or 6
#               consecutive timerange solutions. Generates plots of:
#
# * phase overlays between first & nth time: bpntime%d.0min_Xmin_spwY_phase.pdf
# * ampl. overlays between first & nth time: bpntime%d.0min_Xmin_spwY_amp.pdf
# * phase diff between first & nth time: bpntime%d.0min_Xmin_spwY_phasediff.pdf
# * ampl. ratio between first & nth time: bpntime%d.0min_Xmin_spwY_ampratio.pdf
#
# if ntime > 2, then it also plots how the solution decays with time:
# * phase st.dev vs. time between solutions: plotbpntime%d.phasestdev.spwX.pdf
# * ampl. st.dev vs. time between solutions: plotbpntime%d.ampstdev.spwX.pdf
#
#        The atmospheric transmission at elev=30 deg for ALMA is overlaid in 
#        blue, and ozone lines < 500 cm^-1 from splatalogue are labelled.
#
# Calls:
#    getAtmPlot(f0,f1): Computes typical ALMA atmospheric transmission between
#                the two specified frequencies (in GHz).  
#                Returns two arrays: frequency (GHz) & transmission at el=30deg
#
#    freqToBand(freq): Only needed if the ASDM_CALDELAY table does not
#                      contain the receiver band.
#
#    drawOzone(ozoneInfo, x0, x1, ymid): 
#                      Given a list of ozone lines (ozoneInfo), this draws 
#                      the ozone line labels between frequencies x0 and x1 
#                      (in GHz) and at the vertical location ymid which is
#                      on a scale of (0..1).
#
# Created T.R. Hunter 2010-12-02  v1.0
#
# Example usage:
#  sys.path.insert(0,'/users/thunter/AIV/science/analysis_scripts')
#  import plotbpass
#  plotbpass.plotbpass('uid___A002_X14427c_X31.ms',
#                      'bandpass2_20m.bcal',
#                      spwlist=[0,2,3],
#                      ignoreSpw=[0],
#                      ignoreAnt=[1],
#                      ignoreBeyondChannel=[96],
#                      ignoreBeforeChannel=[0])
#  plotbpass.plotbpass('uid___A002_X14427c_X31.ms',
#                      'bandpass2_60m.bcal',
#                      separate_spws=0,
#                      ntime=2,
#                      ntimeMinutes=60,
#                      spwlist=[0,2,3],
#                      ignoreSpw=[0],
#                      ignoreAnt=[1],
#                      ignoreBeyondChannel=[96],
#                      ignoreBeforeChannel=[0])
#  plotbpass.plotpbass('conjugatedvis_ngc7674_two.ms',
#                      'callisto_bandpass_60m.bcal',
#                      separate_spws=0,
#                      ntime=2,
#                      ntimeMinutes=60,
#                      spwlist=[0,1],
#                      ignoreBeyondChannel=[89],
#                      ignoreSpw=[0],
#                      ignoreAnt=[-1])
#
# Notes:
#   It is possible to have it ignore certain channel ranges when computing
#   the standard deviation statistic.  These are specified by the four
#   parallel sets of 'ignore' arrays.  Each entry corresponds to one range
#   on one spw on one antenna.  Specifying -1 for the antenna will apply
#   the channel range to all antennas for that spw.
#
from numpy import arctan2,std,mean,where,unique,intersect1d,pi,log10,exp,max
from taskinit import *   # needed to define tb.
import math
import casac
import pylab as pl
import sys
elevation = 30 # for transmission overlay
almaTemperature = 273 # K
almaAltitude = 5058 # meter
almaPressure = 592 # mbar

def plotbpass(vis, bpcalprefix, separate_spws=1, ntime=6, ntimeMinutes=20, 
              spwlist=[0,1,2,3], nedgePercent=8,
              ignoreSpw=[], ignoreAnt=[], 
              ignoreBeyondChannel=[], ignoreBeforeChannel=[]):
# ignoreAnt == -1 means all antennas
#
# FYI, this is the order of the data storage loops in the casa table:
# spw 0:
#    time0:
#        ant0:
#            [polX,polY]
#
 ozoneLabel = 0.75  # in transmission units (0..1)
 try:
     lines = ozoneLines
     showOzone = 2 # show ozone line labels and the atm trans. plot
     print "Read %d ozone lines" % (len(lines))
 except:
     showOzone = 1 # show only the atmospheric transmission plot
     print "Failed to load ozone line list file"
 if (showOzone == 2):
     ozoneFreqs = []
     ozoneEnergies = []
     for line in lines:
         if (line.find('Species') < 0):
             (a,b,fr1,er1,fr2,er2,trans,cdmsIntensity,lovasIntensity,energy,linelist) = line.split(':')
             if (len(fr1) > 0):
                 freq = fr1
             else:
                 freq = fr2
             ozoneFreqs.append(float(freq))
             ozoneEnergies.append(float(energy))
     ozoneInfo = [ozoneFreqs,ozoneEnergies]
     print "freqs[0] = %f, energies[0] = %f" % (ozoneFreqs[0],ozoneEnergies[0])
 
 tb.open(tablename='%s/ANTENNA'%(vis))
 antNames = tb.getcol('NAME')
 tb.close()
 try:
   tb.open("%s/ASDM_CALDELAY" % vis)
   antennaNames = tb.getcol("antennaName")
 except:
   antennaNames = []
 if (len(antennaNames) < 1):
   tb.open("%s/SPECTRAL_WINDOW" % vis)
   freq = tb.getcol("REF_FREQUENCY")
   print "Found freq in SPECTRAL_WINDOW table = %f" % (freq[0])
   rxband = freqToBand(freq[0])[0]
   print "converting to band = %d" % (rxband)
 else:
   rxbands = tb.getcol("receiverBand")    # format is 'ALMA_RB_%02d'
   rxband = int(rxbands[0][8:])
   print "Found rxband in ASDM_CALDELAY table = %d" % (rxband)
 tb.close()
 ampWiki = '%s.wikiAmp.txt'%(vis)
 phaseWiki = '%s.wikiPhase.txt'%(vis)
 stdevAmpWiki = open(ampWiki,'w')
 stdevAmpWiki = open(phaseWiki,'w')
 stdevPhaseWiki = open('%s.wikiPhase.txt'%(vis),'w')
 stdevAmpWiki.write('|  Antenna  |  spw  |  polarization  |  st.dev of amplitude ratio (dB)  |\n')
 stdevPhaseWiki.write('|  Antenna  |  spw  |  polarization  |  st.dev of phase difference (deg)  |\n')

 for one_spw in spwlist:
  tb.open("%s/SPECTRAL_WINDOW" % vis)
  try:
    bbc_no = tb.getcol("BBC_NO")
  except:
    cols = tb.getcol("REF_FREQUENCY")
    bbc_no = pl.array(range(1,len(cols)+1))
  print "Determined bbc_no from number of REF_FREQUENCY's = ", bbc_no
  num_chan = tb.getcol("NUM_CHAN")
  idx = where(num_chan > 1)  # ignore the channel averaged data
  idy = where(bbc_no == one_spw+1)
#  print 'num_chan = ', num_chan 
#  print 'one_spw+1 = ', one_spw+1
#  print 'idy = ', idy
#  print idx, idy
  try:
    newlist = intersect1d(idx[0],idy[0])
  except:
    print "Mismatch in requested number of spws with the existing number = ", idx[0], idy
    return
#  print newlist
  id2 = unique(newlist)

  chanfreqs = tb.getcell("CHAN_FREQ",id2)*1e-9
  if (freq[one_spw] > chanfreqs[0]):
      chanfreqs.reverse()
      print "spw %d = %f - %f GHz (reversed because LSB)" % (one_spw,chanfreqs[0], chanfreqs[-1])
  else:
      print "spw %d = %f - %f GHz" % (one_spw,chanfreqs[0], chanfreqs[-1])
  numChan = len(chanfreqs)
  nedge = (numChan * nedgePercent) / 100
  tb.close()
  stdevResults = open('bpntime%d.amp.spw%d.std'%(ntime,one_spw),'w')
  stdevResults.write('time spw pol ant ant stdev\n')
  pstdevResults = open('bpntime%d.phase.spw%d.std'%(ntime,one_spw),'w')
  pstdevResults.write('time spw pol ant ant stdev\n')
  if (separate_spws):
    bpcaltable = '%s_spw%d' % (bpcalprefix,one_spw)
  else:
    bpcaltable = '%s' % (bpcalprefix)
  tb.open(tablename=bpcaltable)
  gain = tb.getcol('GAIN')
  ant = tb.getcol('ANTENNA1')
  # tb.colnames()
  amp = gain.real**2+gain.imag**2
  phase = arctan2(gain.real,gain.imag)
  time = tb.getcol('TIME')
  spw = tb.getcol('CAL_DESC_ID')
  ulist = pl.unique(ant)
  nant = len(ulist)
  if (len(ignoreAnt) > 0):
    if (ignoreAnt[0] == -1):
      ignoreAnt[0] = 0
      for i in range(1,nant):
          ignoreAnt.append(i)
          ignoreSpw.append(ignoreSpw[0])
          if (len(ignoreBeyondChannel) > 0):
              ignoreBeyondChannel.append(ignoreBeyondChannel[0])
          if (len(ignoreBeforeChannel) > 0):
              ignoreBeforeChannel.append(ignoreBeforeChannel[0])
  for i in range(len(ignoreAnt)):
      print "Ant=%d, spw=%d, before=%d, beyond=%d" % (ignoreAnt[i],ignoreSpw[i],ignoreBeforeChannel[i],ignoreBeyondChannel[i])
  nfeed = len(tb.getcol('FEED1'))
  fields = tb.getcol('FIELD_ID')
  nfields = len(unique(fields))
  tb.close()
  nchan = len(amp[0,:,:])
  npol  = len(amp[:,:,0])
  nspw = nfeed/nant/npol/nfields/(ntime/2.)
  print 'File contains: nfeed=%d, nant=%d, nfield=%d, nspw=%d (spw %d), npol=%d' % (nfeed, nant, nfields, nspw, one_spw, npol)
  if (nfields > 1):
     print "Warning, solutions with more than 1 field are not supported by this script.  Use bandpass(combine='field,scan')."
     return

  list = ''
  ratiolist = ''
  plist = ''
  phasedifflist = ''
  polar = []
  polar.append('XX')
  polar.append('YY')
  pl.ioff()
  py0 = -15  # deg
  py1 = 15
  py0diff = -3
  py1diff = 3
  for timeDiff in range(1,ntime):
    if (separate_spws):
      spw = 0   # there is only one spw per bp solution file, so each is it's own "0"
    else:
      spw = one_spw
    for ant in range(nant):
        print "Working on timerange = %d of %d, spw = %d, antenna = %d = %s" % (timeDiff,ntime-1,one_spw,ant,antNames[ant])
        ratio = []
        phaseDiff = []
        # amplitude:
        for pol in range(npol):
            time = 0  # always compare to the first time
            j = spw*(ntime*nant) + time*nant + ant
            amp0 = amp[:,:,j][pol]
            for a in range(len(amp0)):
                if (amp0[a] != amp0[a]):
                    print "Replacing nan"
                    amp0[a] = 1
            time = timeDiff
            j = spw*(ntime*nant)+time*nant + ant
#            print "j=%d, pol=%d, spw=%d, ntime=%d, nant=%d, time=%d, ant=%d" % (j,pol,spw,ntime,nant,time,ant)
            amp1 = amp[:,:,j][pol]
            for a in range(len(amp0)):
                if (amp1[a] != amp1[a]):
                    print "Replacing nan"
                    amp1[a] = 1
            ratio.append(amp1/amp0)
            i = range(len(amp0))
            ax2 = []
            pl.clf()
            pl.plot(i[nedge:-nedge],amp0[nedge:-nedge],'k-',i[nedge:-nedge],amp1[nedge:-nedge],'r-')
            pl.xlim([nedge,len(amp0)-nedge])
            pl.xlabel('Channel')
            y0 = 0.5
            y1 = 1.5
            pl.ylim([y0,y1])
            pl.text(numChan/6,y1-(y1-y0)*0.92,'time range %d min - %d min'%(0,ntimeMinutes),color='k')
            pl.text(numChan/6,y1-(y1-y0)*0.85,'time range %d min - %d min'%(timeDiff*ntimeMinutes,ntimeMinutes*(timeDiff+1)),color='r')
            pl.text(numChan/4,y1-(y1-y0)*0.05,'Band %d   spw %d    %s   pol %s  timeDiff=%d minutes' % (rxband, spw,antNames[ant],polar[pol],timeDiff*ntimeMinutes))
            pl.ylabel('Amplitude')

            ax2 = pl.twiny()
            pl.xlabel('%s - Frequency (GHz)'%(vis))
            name = 'ant%dspw%d%samp.pdf'%(ant,spw,polar[pol])
            if (showOzone):
                [freqs,transmission] = getAtmPlot(chanfreqs[nedge],chanfreqs[-nedge])
                transmission = y0+(y1-y0)*transmission
                pl.plot(chanfreqs[nedge:-nedge],amp0[nedge:-nedge],'k-', freqs, transmission, 'b-')
                if (showOzone==2):
                  drawOzone(ozoneInfo,chanfreqs[nedge],chanfreqs[-nedge], y0+(y1-y0)*ozoneLabel)
                pl.ylim([y0,y1])
                ax2 = pl.twinx()
                pl.ylabel('Transmission at elev=%d (%%)'%(elevation),color='b')
            else:
                pl.plot(chanfreqs[nedge:-nedge],amp0[nedge:-nedge],'k-')
            pl.xlim([chanfreqs[nedge],chanfreqs[-nedge]])
            pl.ylim([y0,y1])
            ax2.yaxis.tick_right()
            ax2.xaxis.tick_top()
            pl.ylim([0,100]) 
            pl.savefig(name,format='pdf')
            list = list + name + ' '
        # phase:
        for pol in range(npol):
            time = 0  # always compare to the first time
            j = spw*(ntime*nant) + time*nant + ant
            phase0 = phase[:,:,j][pol]*180/pi
            time = timeDiff
            j = spw*(ntime*nant)+time*nant + ant
            phase1 = phase[:,:,j][pol]*180/pi
            phaseDiff.append(phase0-phase1)
            i = range(len(amp0))
            ax2 = []
            pl.clf()
            pl.plot(i[nedge:-nedge],phase0[nedge:-nedge],'k-',i[nedge:-nedge],phase1[nedge:-nedge],'r-')
            pl.xlabel('Channel')
            pl.xlim([nedge,len(phase0)-nedge])
            py2 = mean(phase0)+py0  # show same relative number of degrees, but move with the mean
            py3 = mean(phase0)+py1
            pl.ylim([py2,py3]) 
            pl.text(numChan/6,py3-(py3-py2)*0.92,'time range %d min - %d min'%(0,ntimeMinutes),color='k')
            pl.text(numChan/6,py3-(py3-py2)*0.85,'time range %d min - %d min'%(timeDiff*ntimeMinutes,ntimeMinutes*(timeDiff+1)),color='r')
            pl.text(numChan/4,py3-(py3-py2)*0.05,'Band %d   spw %d    %s   pol %s  timeDiff=%d minutes' % (rxband,spw,antNames[ant],polar[pol],timeDiff*ntimeMinutes))
            pl.ylabel('Phase (deg)')

            ax2 = pl.twiny()
            pl.xlabel('%s - Frequency (GHz)'%(vis))
            if (showOzone):
                [freqs,transmission] = getAtmPlot(chanfreqs[nedge],chanfreqs[-nedge])
                transmission = py2+(py3-py2)*transmission
                pl.plot(chanfreqs[nedge:-nedge],phase0[nedge:-nedge],'k-', freqs, transmission, 'b-')
                pl.hold('on')
                if (showOzone==2):
                    drawOzone(ozoneInfo,chanfreqs[nedge],chanfreqs[-nedge], py2+(py3-py2)*ozoneLabel)
                pl.ylim([py2,py3]) 
                ax2 = pl.twinx()
                pl.ylabel('Transmission at elev=%d deg (%%)'%(elevation),color='b')
            else:
                pl.plot(chanfreqs[nedge:-nedge],phase0[nedge:-nedge],'k-')

            pl.xlim([chanfreqs[nedge],chanfreqs[-nedge]])
            pl.ylim([py2,py3]) 
            ax2.yaxis.tick_right()
            ax2.xaxis.tick_top()
            pl.ylim([0,100])
            name = 'ant%dspw%d%sphase.pdf'%(ant,spw,polar[pol])
            pl.savefig(name,format='pdf')
            plist = plist + name + ' '
#
        # amplitude ratio
        pl.clf()
        color = ['k','r','b','g']
        y0 = 0.85
        y1 = 1.15
        for pol in range(npol):
            pl.plot(i[nedge:-nedge],ratio[pol][nedge:-nedge],color[pol])
            pl.text(nedge*2,y1-(y1-y0)*0.05*(1+pol),'%s'%(polar[pol]),color=color[pol])
            pl.hold('on')
        pl.text(numChan/4,y1-(y1-y0)*0.05,'Band %d    spw %d    %s      timeDiff=%d minutes' % (rxband,spw,antNames[ant],timeDiff*ntimeMinutes))
        pl.ylabel('Amplitude ratio')
        pl.xlabel('Channel')
        for pol in range(npol):
            stdev = std(ratio[pol][nedge:-nedge])
            for ix in range(len(ignoreSpw)):
                # ignore some bad channels
                if (spw == ignoreSpw[ix] and ant == ignoreAnt[ix]):
                    if (len(ignoreBeyondChannel) > 0 and len(ignoreBeforeChannel) > 0):
                        if (ignoreBeyondChannel[ix] > ignoreBeforeChannel[ix]):
                            stdev = std(ratio[pol][max(nedge,ignoreBeforeChannel[ix]):ignoreBeyondChannel[ix]])
                        else:
#                            print "joining 2 pieces"
                            twoPieces = ratio[pol][nedge:ignoreBeyondChannel[ix]].tolist() + ratio[pol][ignoreBeforeChannel[ix]:-nedge].tolist() 
                            stdev = std(twoPieces)
                    elif (len(ignoreBeforeChannel) > 0):
                        stdev = std(ratio[pol][max(nedge,ignoreBeforeChannel[ix]):-nedge])
                    elif (len(ignoreBeyondChannel) > 0):
                        stdev = std(ratio[pol][nedge:ignoreBeyondChannel[ix]])
            pl.text(64,y0+(y1-y0)*0.05*(2+pol),'st.dev = %.4f = %.2fdB' % (stdev,10*log10(stdev)), color=color[pol])  # amp
            stdevResults.write('%d %d %s %d %s %f\n' % (timeDiff*ntimeMinutes,spw,pol,ant,antNames[ant],10*log10(stdev)))
            name = 'ant%dspw%d_ampratio.pdf'%(ant,spw)
            if (ntime <= 2):
                stdevAmpWiki.write('|  %s  |  %d  |  %s  |  %.2f  |\n'%(antNames[ant],spw,pol,10*log10(stdev)))
            else:
                stdevAmpWiki.write('|  %s  |  %d  |  %s  |  %.2f  |  %d  |\n'%(antNames[ant],spw,pol,10*log10(stdev),ntimeMinutes*(timeDiff+1)))
        pl.xlim([nedge,len(phase0)-nedge])
        pl.ylim([y0,y1])  # for amplitude ratio

        ax2 = pl.twiny()
        pl.xlabel('%s - Frequency (GHz)'%(vis))
        if (showOzone):
            [freqs,transmission] = getAtmPlot(chanfreqs[nedge],chanfreqs[-nedge])
            transmission = y0+(y1-y0)*transmission
            for pol in range(npol):
                pl.plot(chanfreqs[nedge:-nedge],ratio[pol][nedge:-nedge],color[pol])
                pl.hold('on')
            pl.plot(freqs, transmission, 'b-')
            if (showOzone==2):
                drawOzone(ozoneInfo,chanfreqs[nedge],chanfreqs[-nedge], y0+(y1-y0)*ozoneLabel)
            pl.ylim([y0,y1])  # for amplitude ratio
            ax2 = pl.twinx()
            pl.ylabel('Transmission at elev=%d deg (%%)'%(elevation),color='b')
        else:
            for pol in range(npol):
                pl.plot(chanfreqs[nedge:-nedge],ratio[pol][nedge:-nedge],color[pol])
                pl.hold('on')

        pl.ylim([y0,y1])  # for amplitude ratio
        pl.xlim([chanfreqs[nedge],chanfreqs[len(phase0)-nedge]])

        ax2.yaxis.tick_right()
        ax2.xaxis.tick_top()
        pl.ylim([0,100])
        pl.savefig(name,format='pdf')
        ratiolist = ratiolist + name + ' '
        pl.hold('off')

        # phase difference
        pl.clf()
        color = ['k','r','b','g']
        for pol in range(npol):
            pl.plot(i[nedge:-nedge],phaseDiff[pol][nedge:-nedge],color[pol])
            pl.text(nedge*2,py1diff-(py1diff-py0diff)*0.05*(1+pol),'%s'%(polar[pol]),color=color[pol])
            pl.hold('on')
        pl.text(numChan/4,py1diff-(py1diff-py0diff)*0.05,'Band %d    spw %d    %s      timeDiff=%d minutes' % (rxband,spw,antNames[ant],timeDiff*ntimeMinutes))
        pl.xlabel('Channel')
        pl.ylabel('Phase difference (deg)')
        for pol in range(npol):
            pstdev = std(phaseDiff[pol][nedge:-nedge])
            for ix in range(len(ignoreSpw)):
                # ignore some bad channels
                if (spw == ignoreSpw[ix] and ant == ignoreAnt[ix]):
                    if (len(ignoreBeyondChannel) > 0 and len(ignoreBeforeChannel) > 0):
                        if (ignoreBeyondChannel[ix] > ignoreBeforeChannel[ix]):
                            pstdev = std(phaseDiff[pol][max(nedge,ignoreBeforeChannel[ix]):ignoreBeyondChannel[ix]])
                        else:
                            twoPieces = phaseDiff[pol][nedge:ignoreBeyondChannel[ix]].tolist() + phaseDiff[pol][ignoreBeforeChannel[ix]:-nedge].tolist() 
                            pstdev = std(twoPieces)
                    elif (len(ignoreBeforeChannel) > 0):
                        pstdev = std(phaseDiff[pol][max(nedge,ignoreBeforeChannel[ix]):-nedge])
                    elif (len(ignoreBeyondChannel) > 0):
                        pstdev = std(phaseDiff[pol][nedge:ignoreBeyondChannel[ix]])
            pl.text(64,py0diff+(py1diff-py0diff)*0.05*(1+pol),'st.dev = %.4f deg' % (pstdev),color=color[pol])  # phase
            pstdevResults.write('%d %d %s %d %s %f\n' % (timeDiff*ntimeMinutes,spw,pol,ant,antNames[ant],pstdev))
            name = 'ant%dspw%d_phasediff.pdf'%(ant,spw)
            if (ntime <= 2):
                stdevPhaseWiki.write('|  %s  |  %d  |  %s  |  %.2f  |\n'%(antNames[ant],spw,pol,pstdev))
            else:
                stdevPhaseWiki.write('|  %s  |  %d  |  %s  |  %.2f  |  %d  |\n'%(antNames[ant],spw,pol,pstdev,ntimeMinutes*(timeDiff+1)))
        pl.xlim([nedge,len(phase0)-nedge])
        pl.ylim([py0diff,py1diff])  # for phase diff

        ax2 = pl.twiny()
        pl.xlabel('%s - Frequency (GHz)'%(vis))
        if (showOzone):
            [freqs,transmission] = getAtmPlot(chanfreqs[nedge],chanfreqs[-nedge])
            transmission = py0diff+(py1diff-py0diff)*transmission
            for pol in range(npol):
                pl.plot(chanfreqs[nedge:-nedge],phaseDiff[pol][nedge:-nedge],color[pol])
                pl.hold('on')
            pl.plot(freqs, transmission, 'b-')
            if (showOzone==2):
                drawOzone(ozoneInfo,chanfreqs[nedge],chanfreqs[-nedge], py0diff+(py1diff-py0diff)*ozoneLabel)
            pl.ylim([py0diff,py1diff])
            ax2 = pl.twinx()
            pl.ylabel('Transmission at elev=%d deg (%%)'%(elevation),color='b')
        else:
            for pol in range(npol):
                pl.plot(chanfreqs[nedge:-nedge],phaseDiff[pol][nedge:-nedge],color[pol])
                pl.hold('on')
        pl.xlim([chanfreqs[nedge],chanfreqs[len(phase0)-nedge]])
        pl.ylim([py0diff,py1diff])  # for phase diff
        ax2.yaxis.tick_right()
        ax2.xaxis.tick_top()
        pl.ylim([0,100])
        pl.savefig(name,format='pdf')
        phasedifflist = phasedifflist + name + ' '

    os.system('gs -q -sPAPERSIZE=letter -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -sOutputFile=bp0min_%dmin_spw%d_phase.pdf %s'%(timeDiff*ntimeMinutes,one_spw,plist))
    os.system('gs -q -sPAPERSIZE=letter -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -sOutputFile=bp0min_%dmin_spw%d_phasediff.pdf %s'%(timeDiff*ntimeMinutes,one_spw,phasedifflist))
    os.system('gs -q -sPAPERSIZE=letter -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -sOutputFile=bp0min_%dmin_spw%d_amp.pdf %s'%(timeDiff*ntimeMinutes,one_spw,list))
    os.system('gs -q -sPAPERSIZE=letter -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -sOutputFile=bp0min_%dmin_spw%d_ampratio.pdf %s'%(timeDiff*ntimeMinutes,one_spw,ratiolist))
  # end of loop over time intervals
  stdevResults.close()
  pstdevResults.close()
 # end of loop over spws
 for timeDiff in range(1,ntime):
   for ftype in ('amp','ampratio','phase','phasediff'):
     outputpdf = 'bp0min_%dmin_%s.pdf' % (ntimeMinutes*timeDiff,ftype)
     cmd = 'gs -q -sPAPERSIZE=letter -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -sOutputFile=%s' % (outputpdf)
     print "Producing plot %s"% (outputpdf)
     filelist = ''
     for one_spw in spwlist:
       filelist = filelist + ' bp0min_%dmin_spw%d_%s.pdf' % (ntimeMinutes*timeDiff, one_spw, ftype)
     cmd = cmd + filelist
     print "Executing command = %s" % (cmd)
     os.system(cmd)
     os.system('rm -f %s'%(filelist))
 os.system('rm -f ant*.pdf')
 outputpdf = ''
 for ftype in ('amp','ampratio','phase','phasediff'):
   outputpdf = outputpdf+' bp0min_%dmin_%s.pdf' % (ntimeMinutes*timeDiff,ftype)
 singlepdf = 'bp0min_%dmin.pdf' % (ntimeMinutes*(ntime-1))
 cmd = 'gs -q -sPAPERSIZE=letter -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -sOutputFile=%s %s' % (singlepdf,outputpdf)
 os.system(cmd)
 print "Results left in %s, %s, and %s" % (singlepdf, ampWiki, phaseWiki)
 if (ntime > 2):
#
# Now here I need to make the stdev plots
#
  print 'Plotting standard deviations vs. time'
  pl.clf()
  pl.hold('off')
  if (nant <= 8):
   panels = 421
  elif (nant == 9):
   panels = 331
  else:
   print "You need to adjust the panel layout of the st.dev. plots for %d antennas." % (nant)

  # phases
  for one_spw in spwlist:
   for ant in range(nant):
    stdevResults = open('bpntime%d.phase.spw%d.std'%(ntime,one_spw),'r')
    lines = stdevResults.readlines()
    panel = panels+ant
    pl.subplot(panel)
    stdDev0 = []
    time0 = []
    stdDev1 = []
    time1 = []
    for line in lines:
        if (line.find('time') >= 0):
            continue
        (minutes,spw,pol,antenna,antName,stdev) = line.split()
        spw = int(spw)
        if (ant == int(antenna)):
#            print "Found antenna %d" % (int(antenna))
            if (int(pol) == 0):
                time0.append(int(minutes))
                stdDev0.append(float(stdev))
            else:
                time1.append(int(minutes))
                stdDev1.append(float(stdev))
    pl.plot(time0, stdDev0, 'ko-', time1, stdDev1, 'ro-')
    pl.xlim([0,nchan-nedge])
    pl.title('Band %d    Antenna %s   SPW %d' %(rxband,antNames[ant],one_spw),fontsize=10)
    if (panel == 425 or panel == 334):
        pl.ylabel('Standard Deviation of bandpass phase (deg)')
    if (panel == 426 or panel == 336):
        pl.text(45, 0.52, 'Reference antenna',fontsize=10)
    if (panel == 427 or panel == 338):
        pl.xlabel('Time between bandpass solutions (minutes)')
    pl.ylim([-0.05, 1.2])
    pl.text(24,0.64,'XX', color='k')
    pl.text(24,0.44,'YY', color='r')
    pl.grid('on')
    stdevResults.close()
   pl.savefig('plotbpntime%d.phasestdev.spw%d.png'%(ntime,one_spw),format='png')
   os.system('convert plotbpntime%d.phasestdev.spw%d.png plotbpntime%d.phasestdev.spw%d.pdf'%(ntime,one_spw,ntime,one_spw))
   os.remove('plotbpntime%d.phasestdev.spw%d.png'%(ntime,one_spw))

  # amplitudes
  for one_spw in spwlist:
   pl.clf()
   for ant in range(nant):
    stdevResults = open('bpntime%d.amp.spw%d.std'%(ntime,one_spw),'r')
    lines = stdevResults.readlines()
    panel = panels+ant
    pl.subplot(panel)
    stdDev0 = []
    time0 = []
    stdDev1 = []
    time1 = []
    for line in lines:
        if (line.find('time') >= 0):
            continue
        (minutes,spw,pol,antenna,antName,stdev) = line.split()
        spw = int(spw)
        if (ant == int(antenna)):
#            print "Found antenna %d" % (int(antenna))
            if (int(pol) == 0):
                time0.append(int(minutes))
                stdDev0.append(float(stdev))
            else:
                time1.append(int(minutes))
                stdDev1.append(float(stdev))
    pl.plot(time0, stdDev0, 'ko-', time1, stdDev1, 'ro-')
    pl.xlim([0,nchan-nedge])
    pl.title('Band %d    Antenna %s   SPW %d' %(rxband,antNames[ant],one_spw),fontsize=10)
    if (panel == 425 or panel == 334):
        pl.ylabel('Standard Deviation of bandpass amplitude (dB)')
    if (panel == 427 or panel == 338):
        pl.xlabel('Time between bandpass solutions (minutes)')
    pl.ylim([-30,-15])
    pl.text(10,-17.5,'XX', color='k')
    pl.text(10,-19.5,'YY', color='r')
    pl.grid('on')
    stdevResults.close()
   pl.savefig('plotbp%dntime.ampstdev.spw%d.png'%(ntime,one_spw),format='png')
   os.system('convert plotbp%dntime.ampstdev.spw%d.png plotbp%dntime.ampstdev.spw%d.pdf'%(ntime,one_spw,ntime,one_spw))
   os.remove('plotbp%dntime.ampstdev.spw%d.png'%(ntime,one_spw))

def freqToBand(freq):
    # This is a kludge that is only needed if the ASDM_CALDELAY table 
    # (which holds the rx band) is empty.
    band = []
    for f in [freq]:
        if (f > 750e9):
            band.append(10)
        elif (f > 550e9):
            band.append(9)
        elif (f > 379e9):
            band.append(8)
        elif (f > 275e9):
            band.append(7)
        elif (f > 211e9):
            band.append(6)
        elif (f > 163e9):
            band.append(5)
        elif (f > 120e9):
            band.append(4)
        elif (f > 83e9):
            band.append(3)
        elif (f > 60e9):
            band.append(2)
        elif (f > 20e9):
            band.append(1)
    return(band)
        
def drawOzone(ozoneInfo, x0, x1, ymid):
    freqs = ozoneInfo[0]
    energies = ozoneInfo[1]
    if (x0 < x1):
        lowfreq = x0
        highfreq = x1
    else:
        lowfreq = x1
        highfreq = x0
#    print "between %f and %f at %f" % (x0,x1,ymid)
    for i in range(len(freqs)):
        if (freqs[i] > lowfreq and freqs[i] < highfreq and energies[i] < 500):
            pl.text(freqs[i],ymid,'Ozone',rotation=90)
#            print "Drawing Ozone line at %f" % (freqs[i])

def getAtmPlot(f0, f1):
  tropical = 1
  midLatitudeSummer = 2
  midLatitudeWinter = 3
  at.initAtmProfile(humidity=20.,
                    temperature=casac.Quantity(almaTemperature,"K"),
                    altitude=casac.Quantity(almaAltitude,"m"),
                    pressure=casac.Quantity(almaPressure,'mbar'),
                    atmType=midLatitudeWinter)

#  print at.getStartupWaterContent().value, "mm H2O"
  reffreq=(f0+f1)*0.5
  numchan=200
  chansep=(f1-f0)/(numchan-1.0)
  nb = 1
  fC = casac.Quantity(reffreq,'GHz')
  fR = casac.Quantity(chansep,'GHz')
  fW = casac.Quantity(numchan*chansep,'GHz')
  at.initSpectralWindow(nb,fC,fW,fR)
  rf = at.getRefFreq()
  rc = at.getRefChan()
  cs = at.getChanSep()
  n = at.getNumChan()
  freq = fC.value+chansep*(pl.array(range(numchan))-0.5*numchan)
  dry=pl.array(at.getDryOpacitySpec(0)['dryOpacity'])
  wet=pl.array(at.getWetOpacitySpec(0)['wetOpacity'].value)
  airmass = 1/math.cos(elevation*pi/180.)
  transmission = exp(airmass*(-wet-dry))
  return(freq,transmission)

# main
vis = 'uid___A002_X14427c_X31.ms',
bpcalprefix = 'bandpass2_20m' # what you called the bandpass calibration table in casa
separate_spws = 1  # are the spws in separate calibration tables? 0=no, 1=yes
ntime = 6          # number of solution intervals
ntimeMinutes = 20  # minutes per solution interval
spwlist = [0,1,2,3]
nedgePercent = 8   # percent channels to skip on each edge (to avoid roll-off issues)

if (len(sys.argv) < 2):
  print('Usage: plotbpass(ms, bpcaltable, separate_spws=1, ntime=6, ntimeMinutes=20, spwlist=[0,1,2,3],')
  print('       nedgePercent=8, ignoreSpw=[], ignoreAnt=[], ignoreBeyondChannel=[],ignoreBeforeChannel=[])')
  print('    ms: the name of the parent ms file')
  print('    bpcaltable: the name of the bandpass cal table, or the prefix of a list with _spwX')
  print('    separate_spws: 0=no, 1=yes    yes means that each file ends in _spwX')
  print('    ntime:  the number of solution intervals, currently must be either 2 or 6 (default = 6)')
  print('    ntimeMinutes:  the length of each solution interval in minutes (default = 20)')
  print('    spwlist:  list of spws to analyze, e.g. [0,1,2,3] which is the default')
  print('    nedgePercent:  fraction of channel to clip from each edge (default=8)')
  print(' The rest of the arguments are to avoid certain channel ranges in calculating statistics.')
  print('    ignoreSpw: list of spws in which to ignore some channels')
  print('    ignoreAnt: list of antennas in which to ignore some channels in the corresponding spw')
  print('    ignoreBeyondChannel: list of channels beyond which to ignore in the corresponding spw, 0=none')
  print('    ignoreBeforeChannel: list of channels before which to ignore in the corresponding spw, 0=none')


ozoneLines = ['Species:Chemical Name:Freq-GHz:Freq Err:Meas Freq-GHz:Meas Freq Err:Resolved QNs:CDMS/JPL Intensity:Lovas/AST<br>Intensity:E<sub>L</sub> (cm<sup>-1</sup>):Linelist','O3:Ozone:::93.84435:0.05:8(2,6)-9(1,9):-5.33500::39.75060:JPL','O3:Ozone:::93.95505:0.05:13(3,11)-14(2,12):-5.17970::101.57910:JPL','O3:Ozone:::95.79640:0.05:31(3,29)-30(4,26):-5.56790::442.11250:JPL','O3:Ozone:::96.22834:0.05:2(1,1)-2(0,2):-4.92020::2.51950:JPL','O3:Ozone:::99.24715:0.03:37(4,34)-36(5,31):-5.83790::638.97690:JPL','O3:Ozone:::100.69141:0.06:35(6,30)-36(5,31):-5.83260::638.97690:JPL','O3:Ozone:::101.73687:0.05:4(1,3)-4(0,4):-4.64170::8.39100:JPL','O3:Ozone:::101.83542:0.05:28(5,23)-29(4,26):-5.45110::416.35260:JPL','O3:Ozone:::103.87839:0.05:14(3,11)-15(2,14):-5.10630::113.08700:JPL','O3:Ozone:::109.55933:0.05:20(2,18)-19(3,17):-5.00860::188.05470:JPL','O3:Ozone:::110.83604:0.05:6(1,5)-6(0,6):-4.44760::17.59740:JPL','O3v2=1:Ozone:112.84440:0.0092:::6(1,5)-6(0,6):-5.90440::718.45230:JPL','O3:Ozone:::114.97920:0.05:5(2,4)-6(1,5):-5.24810::21.29440:JPL','O3:Ozone:::118.36450:0.06:1(1,1)-0(0,0):-5.13030::0.00000:JPL','O3:Ozone:::119.27762:0.06:27(5,23)-28(4,24):-5.28010::392.17840:JPL','O3:Ozone:::123.34909:0.06:28(3,25)-27(4,24):-5.17380::368.29590:JPL','O3:Ozone:::124.08738:0.06:8(1,7)-8(0,8):-4.28860::30.11190:JPL','O3:Ozone:::125.38944:0.06:8(0,8)-7(1,7):-4.51960::25.92940:JPL','O3:Ozone:::125.41325:0.06:20(4,16)-21(3,19):-5.01300::222.57160:JPL','O3v2=1:Ozone:126.37682:0.0099:::8(1,7)-8(0,8):-5.74520::730.91130:JPL','O3:Ozone:::128.09482:0.1001:33(3,31)-32(4,28):-5.42850::495.51740:JPL','O3:Ozone:::128.31385:0.1001:6(2,4)-7(1,7):-5.14860::25.92940:JPL','O3:Ozone:::130.95482:0.06:36(4,32)-35(5,31):-5.52900::608.46860:JPL','O3:Ozone:::132.38501:0.06:34(6,28)-35(5,31):-5.54710::608.46860:JPL','O3:Ozone:136.33734:0.0016:::41(7,35)-42(6,36):-5.99630::872.22300:JPL','O3:Ozone:::136.86024:0.1001:19(4,16)-20(3,17):-4.92310::205.32850:JPL','O3:Ozone:::140.76727:0.06:26(2,24)-27(1,27):-5.97080::308.42460:JPL','O3:Ozone:::142.17504:0.06:10(1,9)-10(0,10):-4.15030::45.90080:JPL','O3v2=1:Ozone:144.85103:0.0113:::10(1,9)-10(0,10):-5.60650::746.62820:JPL','O3:Ozone:::144.91067:0.06:39(4,36)-38(5,33):-5.62980::702.31670:JPL','O3:Ozone:::144.91936:0.06:14(1,13)-13(2,12):-4.57720::88.83480:JPL','O3:Ozone:::148.74487:0.06:12(3,9)-13(2,12):-4.81880::88.83480:JPL','O3:Ozone:::153.72419:0.1001:35(3,33)-34(4,30):-5.40110::552.41750:JPL','O3:Ozone:::153.95319:0.06:26(5,21)-27(4,24):-5.03030::368.29590:JPL','O3:Ozone:::154.04644:0.06:11(3,9)-12(2,10):-4.80120::78.53860:JPL','O3:Ozone:::156.10681:0.06:33(6,28)-34(5,29):-5.35790::579.06130:JPL','O3:Ozone:164.76998:0.0019:::40(7,33)-41(6,36):-5.77110::836.80280:JPL','O3:Ozone:::164.95182:0.3:3(1,3)-2(0,2):-4.54600::2.51950:JPL','O3:Ozone:::165.78445:0.3:12(1,11)-12(0,12):-4.02840::64.92570:JPL','O3:Ozone:::167.57271:0.3:4(2,2)-5(1,5):-5.11360::15.35060:JPL','O3v2=1:Ozone:168.96740:0.0132:::12(1,11)-12(0,12):-5.48430::765.56410:JPL','O3:Ozone:::171.41253:0.1001:37(3,35)-36(4,32):-5.45620::612.83680:JPL','O3:Ozone:::173.48549:0.05:3(2,2)-4(1,3):-5.23270::11.78460:JPL','O3:Ozone:::175.18638:0.05:25(5,21)-26(4,22):-4.89110::345.69290:JPL','O3:Ozone:::175.44570:0.05:18(4,14)-19(3,17):-4.70520::188.05470:JPL','O3:Ozone:::178.57699:0.08:41(3,39)-40(4,36):-5.77600::744.30670:JPL','O3:Ozone:::180.00129:0.05:39(3,37)-38(4,34):-5.58230::676.79570:JPL','O3v2=1:Ozone:182.18003:0.0161:::10(0,10)-9(1,9):-5.56150::740.55130:JPL','O3:Ozone:::184.37782:0.05:10(0,10)-9(1,9):-4.07930::39.75060:JPL','O3:Ozone:::184.74869:0.05:22(2,20)-21(3,19):-4.56520::222.57160:JPL','O3:Ozone:::185.55685:0.05:32(6,26)-33(5,29):-5.16480::550.31340:JPL','O3:Ozone:::187.13248:0.05:41(4,38)-40(5,35):-5.54110::769.09590:JPL','O3:Ozone:::187.88484:0.05:28(2,26)-29(1,29):-5.89880::354.14660:JPL','O3:Ozone:::190.57259:0.05:39(7,33)-40(6,34):-5.58620::802.30940:JPL','O3:Ozone:::193.35112:0.05:17(4,14)-18(3,15):-4.61750::172.29740:JPL','O3:Ozone:::195.43036:0.05:14(1,13)-14(0,14):-3.92240::87.14990:JPL','O3:Ozone:::195.72118:0.05:10(3,7)-11(2,10):-4.63840::67.90280:JPL','O3v2=1:Ozone:199.25015:0.0154:::14(1,13)-14(0,14):-5.37800::787.68120:JPL','O3:Ozone:::199.38427:0.05:30(3,27)-29(4,26):-4.82180::416.35260:JPL','O3:Ozone:::203.45316:0.05:38(4,34)-37(5,33):-5.25280::670.00920:JPL','O3:Ozone:::206.13205:0.05:24(5,19)-25(4,22):-4.72740::323.62070:JPL','O3:Ozone:::207.48615:0.05:46(5,41)-45(6,40):-5.81680::983.34550:JPL','O3:Ozone:::208.64241:0.05:5(1,5)-4(0,4):-4.17100::8.39100:JPL','O3v2=1:Ozone:209.52894:0.0183:::5(1,5)-4(0,4):-5.63960::709.28610:JPL','O3v2=1:Ozone:210.20590:0.0193:::16(1,15)-15(2,14):-5.67690::813.80420:JPL','O3:Ozone:::210.42301:0.05:31(6,26)-32(5,27):-5.01490::522.55760:JPL','O3:Ozone:::210.76232:0.05:2(2,0)-3(1,3):-5.43550::8.02170:JPL','O3:Ozone:::210.80320:0.05:9(3,7)-10(2,8):-4.61360::58.98100:JPL','O3:Ozone:::214.95568:0.05:16(1,15)-15(2,14):-4.18610::113.08700:JPL','O3:Ozone:::218.11867:0.05:38(7,31)-39(6,34):-5.41280::768.60160:JPL','O3:Ozone:::223.89470:0.05:45(8,38)-46(7,39):-5.93400::1061.77290:JPL','O3:Ozone:::224.85329:0.05:43(4,40)-42(5,37):-5.52850::839.33290:JPL','O3:Ozone:::226.05404:0.05:16(4,12)-17(3,15):-4.48730::156.90330:JPL','O3:Ozone:::229.57495:0.05:23(5,19)-24(4,20):-4.61370::302.63690:JPL','O3:Ozone:::231.28151:0.05:16(1,15)-16(0,16):-3.83270::112.54240:JPL','O3v1=1,v3=1:Ozone:231.99094:0.0228:::15(2,13)-15(1,14),v3=1:-5.85730::1147.36060:JPL','O3:Ozone:::232.98680:0.05:49(5,45)-48(6,42):-5.95720::1102.35930:JPL','O3v1=1,v3=1:Ozone:234.35526:0.0254:::13(2,11)-13(1,12),v3=1:-5.88060::1122.76420:JPL','O3v1=1,v3=1:Ozone:234.55754:0.0211:::17(2,15)-17(1,16),v3=1:-5.84090::1175.26350:JPL','O3:Ozone:::235.70986:0.05:16(2,14)-16(1,15):-3.67410::120.25710:JPL','O3v2=1:Ozone:235.86765:0.0177:::16(1,15)-16(0,16):-5.28790::812.94820:JPL','O3v1=1,v3=1:Ozone:236.06776:0.0255:::16(2,14)-16(1,15),v1=1:-5.97560::1223.05940:JPL','O3:Ozone:::237.14612:0.05:14(2,12)-14(1,13):-3.69000::93.66870:JPL','O3:Ozone:::238.43190:0.05:30(6,24)-31(5,27):-4.86850::495.54370:JPL','O3v1=1,v3=1:Ozone:238.67541:0.0277:::14(2,12)-14(1,13),v1=1:-5.98920::1196.53280:JPL','O3:Ozone:::239.09328:0.05:18(2,16)-18(1,17):-3.66450::150.19000:JPL','O3v2=1:Ozone:239.29217:0.017:::16(4,12)-17(3,15):-5.91070::857.78780:JPL','O3v2=1:Ozone:239.86126:0.015:::16(2,14)-16(1,15):-5.13050::820.81590:JPL','O3v1=1,v3=1:Ozone:240.41080:0.0284:::11(2,9)-11(1,10),v3=1:-5.91630::1101.50670:JPL','O3:Ozone:::240.90425:0.05:30(2,28)-31(1,31):-5.86100::403.01940:JPL','O3v2=1:Ozone:241.20358:0.0193:::12(0,12)-11(1,11):-5.24060::757.51840:JPL','O3v2=1:Ozone:241.27323:0.0164:::14(2,12)-14(1,13):-5.14670::794.32750:JPL','O3:Ozone:::242.31869:0.05:12(2,10)-12(1,11):-3.71660::70.45570:JPL','O3v1=1,v3=1:Ozone:243.06350:0.0206:::19(2,17)-19(1,18),v3=1:-5.82780::1206.43400:JPL','O3v2=1:Ozone:243.38234:0.0143:::18(2,16)-18(1,17):-5.12050::850.63390:JPL','O3:Ozone:::243.45378:0.05:12(0,12)-11(1,11):-3.76090::56.80500:JPL','O3:Ozone:::244.14574:0.05:37(7,31)-38(6,32):-5.26120::735.78460:JPL','O3:Ozone:::244.15839:0.05:8(3,5)-9(2,8):-4.55030::50.30210:JPL','O3v1=1,v3=1:Ozone:244.21720:0.0236:::20(2,18)-20(1,19),v1=1:-5.96770::1286.12800:JPL','O3v2=1:Ozone:246.51265:0.0182:::12(2,10)-12(1,11):-5.17350::771.20020:JPL','O3:Ozone:::247.76175:0.05:15(4,12)-16(3,13):-4.41570::142.71280:JPL','O3:Ozone:::248.18339:0.05:20(2,18)-20(1,19):-3.65820::183.43070:JPL','O3v1=1,v3=1:Ozone:248.72080:0.0313:::9(2,7)-9(1,8),v3=1:-5.97080::1083.61420:JPL','O3:Ozone:::249.78855:0.05:7(1,7)-6(0,6):-3.89560::17.59740:JPL','O3:Ozone:::249.96196:0.05:10(2,8)-10(1,9):-3.75950::50.64320:JPL','O3v2=1:Ozone:250.27140:0.0216:::7(1,7)-6(0,6):-5.36610::718.45230:JPL','O3:Ozone:::250.72637:0.05:44(8,36)-45(7,39):-5.76730::1023.08130:JPL','O3v2=1:Ozone:252.74108:0.0144:::20(2,18)-20(1,19):-5.11370::883.74400:JPL','O3v2=1:Ozone:253.39209:0.021:::8(3,5)-9(2,8):-5.99090::751.29220:JPL','O3v2=1:Ozone:254.28713:0.0199:::10(2,8)-10(1,9):-5.21660::751.45990:JPL','O3v2=1:Ozone:254.89608:0.0213:::24(2,22)-23(3,21):-5.77440::960.89800:JPL','O3v1=1,v3=1:Ozone:256.46115:0.0231:::22(2,20)-22(1,21),v1=1:-5.96800::1322.60270:JPL','O3:Ozone:256.88561:0.0041:::45(4,42)-44(5,39):-5.57430::913.04880:JPL','O3:Ozone:::258.20205:0.12:22(5,17)-23(4,20):-4.49570::282.32880:JPL','O3v1=1,v3=1:Ozone:258.30919:0.0209:::21(2,19)-21(1,20),v3=1:-5.81610::1240.82780:JPL','O3:Ozone:::258.71618:0.12:8(2,6)-8(1,7):-3.82590::34.25100:JPL','O3v2=1:Ozone:260.79851:0.0185:::15(4,12)-16(3,13):-5.84410::843.66040:JPL','O3v1=1,v3=1:Ozone:262.47687:0.0282:::17(1,16)-17(0,17),v3=1:-5.94430::1166.50820:JPL','O3:Ozone:::262.85785:0.12:24(2,22)-23(3,21):-4.27690::260.44680:JPL','O3v2=1:Ozone:263.20667:0.0215:::8(2,6)-8(1,7):-5.28310::735.12680:JPL','O3:Ozone:::263.69260:0.12:22(2,20)-22(1,21):-3.65360::219.93830:JPL','O3v1=1,v3=1:Ozone:263.83191:0.0314:::8(1,8)-7(0,7),v3=1:-5.96740::1065.19840:JPL','O3:Ozone:263.88610:0.0032:::29(6,24)-30(5,25):-4.74520::469.45620:JPL','O3:Ozone:264.92589:0.0023:::7(3,5)-8(2,6):-4.55440::42.88090:JPL','O3:Ozone:::267.26657:0.12:6(2,4)-6(1,5):-3.92800::21.29440:JPL','O3v2=1:Ozone:::268.66368:0.05:22(2,20)-22(1,21):-5.10860::920.10420:JPL','O3v1=1,v3=1:Ozone:::270.92335:0.05:13(0,13)-12(1,12),v3=1:-5.79900::1107.54560:JPL','O3:Ozone:271.09189:0.003:::36(7,29)-37(6,32):-5.11890::703.77950:JPL','O3v2=1:Ozone:271.92638:0.0228:::6(2,4)-6(1,5):-5.38530::722.21640:JPL','O3:Ozone:::273.05090:0.12:18(1,17)-18(0,18):-3.75940::141.08200:JPL','O3v2=1:Ozone:273.94769:0.0228:::7(3,5)-8(2,6):-5.99820::743.90640:JPL','O3:Ozone:::274.47838:0.12:4(2,2)-4(1,3):-4.09530::11.78460:JPL','O3v1=1,v3=1:Ozone:::274.97493:0.05:24(2,22)-24(1,23),v1=1:-5.97050::1362.32010:JPL','O3v2=1:Ozone:275.17120:0.0193:::22(5,17)-23(4,20):-5.91340::983.16290:JPL','O3:Ozone:::276.92354:0.12:14(4,10)-15(3,13):-4.33480::129.12200:JPL','O3:Ozone:277.03826:0.0036:::43(8,36)-44(7,37):-5.61440::985.24580:JPL','O3v2=1:Ozone:::278.52128:0.09:18(1,17)-18(0,18):-5.21420::841.34340:JPL','O3v2=1:Ozone:279.28477:0.0238:::4(2,2)-4(1,3):-5.55280::712.74030:JPL','O3:Ozone:279.33534:0.0055:::51(5,47)-50(6,44):-5.96910::1185.90050:JPL','O3:Ozone:279.47297:0.0044:::48(5,43)-47(6,42):-5.70720::1061.68640:JPL','O3:Ozone:::279.48584:0.05:2(2,0)-2(1,1):-4.47030::5.72930:JPL','O3:Ozone:::279.89246:0.12:32(3,29)-31(4,28):-4.59880::467.78720:JPL','O3v1=1,v3=1:Ozone:280.89114:0.0215:::23(2,21)-23(1,22),v3=1:-5.80580::1278.39850:JPL','O3:Ozone:280.99471:0.0036:::40(4,36)-39(5,35):-5.08660::734.93370:JPL','O3v2=1:Ozone:281.53489:0.0208:::18(1,17)-17(2,16):-5.38820::841.24290:JPL','O3:Ozone:281.95853:0.0052:::47(4,44)-46(5,41):-5.67010::990.26640:JPL','O3:Ozone:::282.83720:0.12:21(5,17)-22(4,18):-4.40370::262.99510:JPL','O3v2=1:Ozone:284.39532:0.0245:::2(2,0)-2(1,1):-5.92780::706.70640:JPL','O3:Ozone:::286.08769:0.12:18(1,17)-17(2,16):-3.90330::140.64710:JPL','O3:Ozone:::286.15653:0.12:24(2,22)-24(1,23):-3.65090::259.66970:JPL','O3:Ozone:::286.29475:0.12:3(2,2)-3(1,3):-4.22440::8.02170:JPL','O3:Ozone:::288.95914:0.12:9(1,9)-8(0,8):-3.67970::30.11190:JPL','O3v2=1:Ozone:289.00057:0.0243:::9(1,9)-8(0,8):-5.15150::730.91130:JPL','O3v2=1:Ozone:289.95921:0.02:::14(4,10)-15(3,13):-5.76800::830.12270:JPL','O3:Ozone:::290.97512:0.12:28(6,22)-29(5,25):-4.62810::444.15880:JPL','O3v1=1,v3=1:Ozone:291.05309:0.0443:::9(1,9)-8(0,8),v1=1:-5.97430::1133.22860:JPL','O3v2=1:Ozone:291.34554:0.0246:::3(2,2)-3(1,3):-5.68190::708.98330:JPL','O3v2=1:Ozone:291.69508:0.0156:::24(2,22)-24(1,23):-5.10540::959.67050:JPL','O3:Ozone:::293.17127:0.12:5(2,4)-5(1,5):-3.98170::15.35060:JPL','O3:Ozone:::293.54823:0.12:6(3,3)-7(2,6):-4.57010::36.04190:JPL','O3:Ozone:297.17289:0.0034:::35(7,29)-36(6,30):-4.99020::672.64410:JPL','O3v2=1:Ozone:298.36498:0.0243:::5(2,4)-5(1,5):-5.43900::716.27530:JPL','O3:Ozone:298.60086:0.01:::32(2,30)-33(1,33):-5.85100::455.03920:JPL','O3:Ozone:298.79513:0.0064:::49(4,46)-48(5,43):-5.81260::1071.00860:JPL','O3v1=1,v3=1:Ozone:299.44581:0.0392:::14(0,14)-13(1,13),v1=1:-5.83300::1180.32470:JPL','O3v2=1:Ozone:299.48843:0.0217:::14(0,14)-13(1,13):-4.99850::777.69140:JPL','O3v2=1:Ozone:299.68012:0.0207:::21(5,17)-22(4,18):-5.82650::963.91090:JPL','O3v1=1,v3=1:Ozone:300.14006:0.0266:::26(2,24)-26(1,25),v1=1:-5.97580::1405.24110:JPL','O3:Ozone:::300.68515:0.12:13(4,10)-14(3,11):-4.28450::116.55200:JPL','O3v1=1,v3=1:Ozone:301.03651:0.0343:::10(1,10)-9(0,9),v3=1:-5.77280::1079.14420:JPL','O3:Ozone:::301.81300:0.12:14(0,14)-13(1,13):-3.52040::77.08250:JPL','O3:Ozone:::303.16492:0.12:7(2,6)-7(1,7):-3.83730::25.92940:JPL','O3:Ozone:::303.57018:0.05:42(8,34)-43(7,37):-5.47100::948.24220:JPL','O3v1=1,v3=1:Ozone:306.93221:0.0352:::8(2,7)-8(1,8),v3=1:-5.96090::1073.99890:JPL','O3v2=1:Ozone:308.56641:0.0239:::7(2,6)-7(1,7):-5.29450::726.80050:JPL','O3v1=1,v3=1:Ozone:309.85124:0.0346:::19(1,18)-19(0,19),v3=1:-5.87660::1196.09840:JPL','O3:Ozone:::310.06352:0.05:20(5,15)-21(4,18):-4.31520::244.42020:JPL','O3v1=1,v3=1:Ozone:311.13894:0.0226:::25(2,23)-25(1,24),v3=1:-5.79820::1319.10140:JPL','O3v2=1:Ozone:313.57498:0.0217:::13(4,10)-14(3,11):-5.72120::817.60720:JPL','O3:Ozone:::315.87512:0.12:26(2,24)-26(1,25):-3.65140::302.58360:JPL','O3:Ozone:::316.32725:0.12:9(2,8)-9(1,9):-3.73540::39.75060:JPL','O3:Ozone:316.68160:0.004:::27(6,22)-28(5,23):-4.52510::419.74940:JPL','O3:Ozone:::317.19506:0.12:5(3,3)-6(2,4):-4.64390::30.20950:JPL','O3:Ozone:::319.99654:0.12:20(1,19)-20(0,20):-3.70210::172.75680:JPL','O3v2=1:Ozone:322.00280:0.0236:::9(2,8)-9(1,9):-5.19240::740.55130:JPL','O3v2=1:Ozone:322.13713:0.0162:::26(2,24)-26(1,25):-5.10530::1002.40090:JPL','O3v1=1,v3=1:Ozone:322.68846:0.034:::10(2,9)-10(1,10),v3=1:-5.87080::1089.18570:JPL','O3:Ozone:323.70375:0.0037:::34(7,27)-35(6,30):-4.86950::642.33560:JPL','O3v1=1,v3=1:Ozone:326.05426:0.0256:::19(1,18)-18(2,17),v3=1:-5.92790::1195.55800:JPL','O3v2=1:Ozone:326.44687:0.0233:::20(1,19)-20(0,20):-5.15660::872.85490:JPL','O3v2=1:Ozone:326.48073:0.0263:::11(1,11)-10(0,10):-4.97720::746.62820:JPL','O3v2=1:Ozone:326.85820:0.0221:::20(5,15)-21(4,18):-5.74260::945.41190:JPL','O3:Ozone:326.90096:0.0031:::11(1,11)-10(0,10):-3.50430::45.90080:JPL','O3v1=1,v3=1:Ozone:327.69918:0.0268:::15(0,15)-14(1,14),v3=1:-5.58950::1129.06750:JPL','O3:Ozone:327.84456:0.0031:::12(4,8)-13(3,11):-4.23920::104.71310:JPL','O3v1=1,v3=1:Ozone:329.07955:0.0403:::11(2,10)-11(1,11),v1=1:-5.96230::1160.01600:JPL','O3v1=1,v3=1:Ozone:329.63532:0.0479:::11(1,11)-10(0,10),v1=1:-5.79930::1149.02060:JPL','O3:Ozone:329.78693:0.0041:::41(8,34)-42(7,35):-5.33730::912.08820:JPL','O3v1=1,v3=1:Ozone:332.10203:0.0315:::28(2,26)-28(1,27),v1=1:-5.98530::1451.33010:JPL','O3:Ozone:332.70452:0.003:::11(2,10)-11(1,11):-3.65900::56.80500:JPL','O3:Ozone:332.88170:0.0033:::2(2,0)-1(1,1):-4.07500::3.94820:JPL','O3:Ozone:335.27123:0.0039:::19(5,15)-20(4,16):-4.24250::226.75490:JPL','O3:Ozone:335.63757:0.0058:::48(9,39)-49(8,42):-5.91920::1228.81720:JPL','O3v2=1:Ozone:335.68485:0.0222:::26(2,24)-25(3,23):-5.55930::1001.94900:JPL','O3v2=1:Ozone:337.24071:0.0271:::27(6,22)-28(5,23):-5.94370::1120.61280:JPL','O3v1=1,v3=1:Ozone:337.55230:0.0361:::12(1,12)-11(0,11),v3=1:-5.61230::1096.28600:JPL','O3v2=1:Ozone:337.63983:0.0287:::2(2,0)-1(1,1):-5.53540::704.93040:JPL','O3v2=1:Ozone:338.72174:0.0232:::11(2,10)-11(1,11):-5.11580::757.51840:JPL','O3v2=1:Ozone:340.68001:0.0235:::12(4,8)-13(3,11):-5.67910::805.81620:JPL','O3v1=1,v3=1:Ozone:341.88534:0.0331:::12(2,11)-12(1,12),v3=1:-5.80170::1107.54560:JPL','O3:Ozone:343.18060:0.0044:::26(6,20)-27(5,23):-4.42890::396.15700:JPL','O3:Ozone:343.23742:0.0027:::26(2,24)-25(3,23):-4.06970::301.67090:JPL','O3:Ozone:343.50564:0.0031:::4(3,1)-5(2,4):-4.78630::25.12970:JPL','O3v1=1,v3=1:Ozone:347.19295:0.0405:::13(2,12)-13(1,13),v1=1:-5.90600::1180.32470:JPL','O3v1=1,v3=1:Ozone:348.57830:0.0316:::22(1,21)-22(0,22),v1=1:-5.98970::1310.97540:JPL','O3v1=1,v3=1:Ozone:349.00332:0.0249:::27(2,25)-27(1,26),v3=1:-5.79550::1362.89890:JPL','O3:Ozone:::349.74397:0.05:33(7,27)-34(6,28):-4.75840::612.88450:JPL','O3v2=1:Ozone:351.95790:0.0237:::19(5,15)-20(4,16):-5.67370::927.82090:JPL','O3:Ozone:::352.32408:0.05:13(2,12)-13(1,13):-3.60090::77.08250:JPL','O3:Ozone:::352.59288:0.05:11(4,8)-12(3,9):-4.21430::93.79630:JPL','O3v1=1,v3=1:Ozone:352.64425:0.037:::29(3,26)-29(2,27),v3=1:-5.79230::1422.90560:JPL','O3:Ozone:::352.81519:0.05:28(2,26)-28(1,27):-3.65680::348.64510:JPL','O3v2=1:Ozone:353.27197:0.0216:::20(1,19)-19(2,18):-5.16860::871.96010:JPL','O3v2=1:Ozone:353.27443:0.0339:::34(3,31)-33(4,30):-5.94320::1222.43700:JPL','O3v1=1,v3=1:Ozone:353.72502:0.0348:::27(3,24)-27(2,25),v3=1:-5.73120::1374.54040:JPL','O3:Ozone:::355.01814:0.05:3(2,2)-2(1,1):-3.97420::5.72930:JPL','O3:Ozone:::356.08712:0.05:40(8,32)-41(7,35):-5.21120::876.77070:JPL','O3v2=1:Ozone:::356.42325:11.9917:16(0,16)-15(1,15):-4.81350::801.05930:JPL','O3:Ozone:::356.42553:0.05:50(5,45)-49(6,44):-5.65290::1143.40520:JPL','O3v1=1,v3=1:Ozone:356.65046:0.0409:::16(0,16)-15(1,15),v1=1:-5.64750::1203.85320:JPL','O3:Ozone:::357.62982:0.05:20(1,19)-19(2,18):-3.68730::171.50150:JPL','O3:Ozone:::358.19983:0.05:30(3,27)-30(2,28):-3.65770::411.05510:JPL','O3v1=1,v3=1:Ozone:358.74276:0.0393:::31(3,28)-31(2,29),v3=1:-5.85370::1474.52670:JPL','O3v2=1:Ozone:358.75089:0.023:::13(2,12)-13(1,13):-5.05730::777.69140:JPL','O3:Ozone:::358.85335:0.05:16(0,16)-15(1,15):-3.33650::100.57230:JPL','O3v1=1,v3=1:Ozone:359.07326:0.0465:::30(3,27)-30(2,28),v1=1:-5.95720::1512.92270:JPL','O3v2=1:Ozone:359.60577:0.0303:::3(2,2)-2(1,1):-5.43570::706.70640:JPL','O3:Ozone:359.64957:0.0135:::34(2,32)-35(1,35):-5.86380::510.20250:JPL','O3:Ozone:::359.64966:0.05:28(3,25)-28(2,26):-3.59000::360.41370:JPL','O3v2=1:Ozone:359.94879:0.0177:::28(2,26)-28(1,27):-5.11020::1048.25970:JPL','O3v1=1,v3=1:Ozone:360.87476:0.0332:::25(3,22)-25(2,23),v3=1:-5.67220::1329.47990:JPL','O3v1=1,v3=1:Ozone:361.65918:0.046:::21(1,20)-21(0,21),v3=1:-5.82490::1228.76420:JPL','O3:Ozone:::361.66931:0.05:18(5,13)-19(4,16):-4.17640::209.89360:JPL','O3:Ozone:::361.82074:0.05:47(9,39)-48(8,40):-5.78000::1187.66570:JPL','O3:Ozone:::363.46927:0.05:32(3,29)-32(2,30):-3.72630::464.99940:JPL','O3v2=1:Ozone:363.56664:0.0276:::13(1,13)-12(0,12):-4.83280::765.56410:JPL','O3v2=1:Ozone:363.67279:0.0289:::26(6,20)-27(5,23):-5.85190::1097.11790:JPL','O3:Ozone:::363.71443:0.05:42(4,38)-41(5,37):-4.98340::803.23880:JPL','O3v1=1,v3=1:Ozone:363.76745:0.0416:::28(3,25)-28(2,26),v1=1:-5.88560::1462.40780:JPL','O3:Ozone:::364.40367:0.05:34(3,31)-33(4,30):-4.44660::522.59390:JPL','O3:Ozone:::364.44991:0.05:13(1,13)-12(0,12):-3.35910::64.92570:JPL','O3v2=1:Ozone:364.50174:0.0233:::30(3,27)-30(2,28):-5.11240::1110.71890:JPL','O3v1=1,v3=1:Ozone:364.52161:0.033:::14(2,13)-14(1,14),v3=1:-5.74850::1129.06750:JPL','O3v2=1:Ozone:365.30932:0.0254:::11(4,8)-12(3,9):-5.65690::794.94610:JPL','O3v2=1:Ozone:365.83285:0.0225:::28(3,25)-28(2,26):-5.04530::1060.26630:JPL','O3:Ozone:::366.77695:0.05:26(3,23)-26(2,24):-3.52500::313.12010:JPL','O3v1=1,v3=1:Ozone:367.68427:0.0499:::13(1,13)-12(0,12),v1=1:-5.65480::1168.06010:JPL','O3v1=1,v3=1:Ozone:368.31436:0.0408:::15(2,14)-15(1,15),v1=1:-5.86440::1203.85320:JPL','O3:Ozone:::368.33221:0.05:3(3,1)-4(2,2):-5.09690::20.94020:JPL','O3:Ozone:::368.94916:0.05:25(6,20)-26(5,21):-4.34270::373.43120:JPL','O3v2=1:Ozone:370.05430:0.0241:::32(3,29)-32(2,30):-5.18030::1164.45840:JPL','O3:Ozone:::371.02285:0.05:22(1,21)-22(0,22):-3.66030::207.56230:JPL','O3v1=1,v3=1:Ozone:372.68285:0.0327:::23(3,20)-23(2,21),v3=1:-5.61780::1287.76800:JPL','O3v1=1,v3=1:Ozone:372.82992:0.0418:::33(3,30)-33(2,31),v3=1:-5.91450::1529.35260:JPL','O3v2=1:Ozone:372.98783:0.022:::26(3,23)-26(2,24):-4.98070::1013.14620:JPL','O3v1=1,v3=1:Ozone:373.08174:0.0383:::26(3,23)-26(2,24),v1=1:-5.81810::1415.25270:JPL','O3v1=1,v3=1:Ozone:374.28723:0.0368:::14(1,14)-13(0,13),v3=1:-5.47870::1116.58260:JPL','O3:Ozone:::375.17918:0.05:15(2,14)-15(1,15):-3.55740::100.57230:JPL','O3:Ozone:::375.97798:0.05:32(7,25)-33(6,28):-4.65430::584.26840:JPL','O3:Ozone:::376.23827:0.05:34(3,31)-34(2,32):-3.79500::522.19910:JPL','O3:Ozone:::378.28123:0.05:24(3,21)-24(2,22):-3.46470::269.21480:JPL','O3v2=1:Ozone:378.28378:0.0254:::18(5,13)-19(4,16):-5.61100::911.02920:JPL','O3v2=1:Ozone:378.52047:0.0291:::22(1,21)-22(0,22):-5.11430::907.47810:JPL','O3:Ozone:::378.69487:0.05:10(4,6)-11(3,9):-4.20210::83.67700:JPL','O3v2=1:Ozone:382.08345:0.023:::15(2,14)-15(1,15):-5.01350::801.05930:JPL','O3:Ozone:::382.19013:0.05:39(8,32)-40(7,33):-5.09260::842.29900:JPL','O3v1=1,v3=1:Ozone:382.77305:0.0291:::17(0,17)-16(1,16),v3=1:-5.43000::1153.74030:JPL','O3v2=1:Ozone:383.28271:0.0249:::34(3,31)-34(2,32):-5.24820::1221.43600:JPL','O3v2=1:Ozone:::384.65471:11.9917:24(3,21)-24(2,22):-4.92080::969.40040:JPL','O3v1=1,v3=1:Ozone:385.73508:0.0373:::24(3,21)-24(2,22),v1=1:-5.75640::1371.49220:JPL','O3:Ozone:::387.09840:0.05:17(5,13)-18(4,14):-4.12190::193.90700:JPL','O3:Ozone:::387.28761:0.05:4(2,2)-3(1,3):-3.87400::8.02170:JPL','O3v1=1,v3=1:Ozone:387.49531:0.0336:::21(3,18)-21(2,19),v3=1:-5.57010::1249.44410:JPL','O3:Ozone:::387.98455:0.05:46(9,37)-47(8,40):-5.64750::1147.35150:JPL','O3v2=1:Ozone:389.35097:0.0308:::25(6,20)-26(5,21):-5.76960::1074.48690:JPL','O3v1=1,v3=1:Ozone:390.54576:0.0341:::16(2,15)-16(1,16),v3=1:-5.70890::1153.74030:JPL','O3v2=1:Ozone:391.32875:0.0274:::10(4,6)-11(3,9):-5.64700::784.86860:JPL','O3v2=1:Ozone:391.91597:0.0324:::4(2,2)-3(1,3):-5.33640::708.98330:JPL','O3v1=1,v3=1:Ozone:392.41457:0.0412:::17(2,16)-17(1,17),v1=1:-5.83540::1230.59110:JPL','O3:Ozone:::392.65403:0.05:22(3,19)-22(2,20):-3.41120::228.73420:JPL','O3v1=1,v3=1:Ozone:393.96306:0.0291:::29(2,27)-29(1,28),v3=1:-5.79950::1409.76440:JPL','O3:Ozone:::395.06298:0.05:24(6,18)-25(5,21):-4.26330::351.53640:JPL','O3v1=1,v3=1:Ozone:395.41516:0.0444:::35(3,32)-35(2,33),v3=1:-5.97480::1587.33150:JPL','O3:Ozone:396.52944:0.0039:::30(2,28)-30(1,29):-3.66900::397.82830:JPL','O3:Ozone:397.02761:0.0059:::36(3,33)-36(2,34):-3.86360::582.60590:JPL','O3v1=1,v3=1:Ozone:397.08178:0.0294:::21(1,20)-20(2,19),v3=1:-5.73790::1227.58260:JPL','O3v2=1:Ozone:399.25481:0.0225:::22(3,19)-22(2,20):-4.86770::929.06590:JPL','O3v1=1,v3=1:Ozone:399.45174:0.0368:::24(1,23)-24(0,24),v1=1:-5.96330::1348.99580:JPL','O3:Ozone:400.07697:0.0039:::5(2,4)-4(1,3):-3.78150::11.78460:JPL','O3v1=1,v3=1:Ozone:400.31607:0.0381:::22(3,19)-22(2,20),v1=1:-5.70260::1331.15730:JPL','O3v2=1:Ozone:401.06419:0.0282:::15(1,15)-14(0,14):-4.71280::787.68120:JPL','O3:Ozone:401.21773:0.0033:::17(2,16)-17(1,17):-3.52650::127.26390:JPL','O3:Ozone:401.93028:0.0051:::31(7,25)-32(6,26):-4.55760::556.50290:JPL','O3:Ozone:402.39533:0.0035:::15(1,15)-14(0,14):-3.23860::87.14990:JPL','O3v1=1,v3=1:Ozone:403.55125:0.0355:::19(3,16)-19(2,17),v3=1:-5.53160::1214.54170:JPL','O3v2=1:Ozone:403.61202:0.0271:::17(5,13)-18(4,14):-5.55930::895.10970:JPL','O3:Ozone:403.83734:0.0039:::9(4,6)-10(3,7):-4.21000::74.43140:JPL','O3v2=1:Ozone:404.33999:0.0334:::5(2,4)-4(1,3):-5.24490::712.74030:JPL','O3v2=1:Ozone:::404.66345:0.08:30(2,28)-30(1,29):-5.12180::1097.22080:JPL','O3v2=1:Ozone:::404.71309:0.06:36(3,33)-36(2,34):-5.31610::1281.60240:JPL','O3v1=1,v3=1:Ozone:405.91991:0.0504:::15(1,15)-14(0,14),v1=1:-5.53510::1190.31320:JPL','O3:Ozone:408.29399:0.0036:::20(3,17)-20(2,18):-3.36670::191.70920:JPL','O3:Ozone:408.30478:0.0051:::38(8,30)-39(7,33):-4.98040::808.66630:JPL','O3v2=1:Ozone:::408.66575:0.06:17(2,16)-17(1,17):-4.98230::827.61130:JPL','O3v1=1,v3=1:Ozone:410.49286:0.0493:::5(2,3)-4(1,4),v3=1:-5.96790::1053.18550:JPL','O3v2=1:Ozone:::411.67950:0.06:18(0,18)-17(1,17):-4.67260::827.61130:JPL','O3v1=1,v3=1:Ozone:411.97282:0.0368:::16(1,16)-15(0,15),v3=1:-5.36770::1139.99830:JPL','O3v1=1,v3=1:Ozone:412.40256:0.041:::18(0,18)-17(1,17),v1=1:-5.50550::1230.59110:JPL','O3:Ozone:413.01187:0.0049:::16(5,11)-17(4,14):-4.07550::178.74690:JPL','O3:Ozone:::414.04906:0.06:45(9,37)-46(8,38):-5.52140::1107.87730:JPL','O3:Ozone:414.25453:0.0032:::18(0,18)-17(1,17):-3.19640::127.26390:JPL','O3v1=1,v3=1:Ozone:414.38486:0.0502:::6(2,5)-5(1,4),v3=1:-5.88170::1057.99960:JPL','O3v2=1:Ozone:::415.18985:0.06:20(3,17)-20(2,18):-4.82340::892.17450:JPL','O3v2=1:Ozone:::415.38607:0.06:24(6,18)-25(5,21):-5.69350::1052.68320:JPL','O3v1=1,v3=1:Ozone:415.39750:0.0398:::20(3,17)-20(2,18),v1=1:-5.65850::1294.27420:JPL','O3v2=1:Ozone:::416.36299:0.06:9(4,6)-10(3,7):-5.65690::775.66210:JPL','O3v1=1,v3=1:Ozone:416.48694:0.064:::23(1,22)-23(0,23),v3=1:-5.78900::1264.50600:JPL','O3v2=1:Ozone:::418.03121:0.06:28(2,26)-27(3,25):-5.39890::1046.32230:JPL','O3v1=1,v3=1:Ozone:418.60929:0.0305:::22(1,21)-21(2,20),v1=1:-5.85500::1308.63940:JPL','O3v1=1,v3=1:Ozone:419.17734:0.038:::17(3,14)-17(2,15),v3=1:-5.50460::1183.08750:JPL','O3v1=1,v3=1:Ozone:419.42031:0.0416:::19(2,18)-19(1,19),v1=1:-5.81780::1260.52830:JPL','O3v1=1,v3=1:Ozone:419.84546:0.0375:::18(2,17)-18(1,18),v3=1:-5.68130::1181.55340:JPL','O3:Ozone:420.79492:0.0058:::23(6,18)-24(5,19):-4.19200::330.49650:JPL','O3:Ozone:422.69450:0.0181:::36(2,34)-37(1,37):-5.89640::568.50630:JPL','O3:Ozone:423.66948:0.0035:::18(3,15)-18(2,16):-3.33310::158.16530:JPL','O3v2=1:Ozone:424.60827:0.0219:::22(1,21)-21(2,20):-4.99820::905.94080:JPL','O3:Ozone:424.84593:0.0035:::24(1,23)-24(0,24):-3.63350::245.49830:JPL','O3:Ozone:425.15775:0.0035:::28(2,26)-27(3,25):-3.91440::346.23200:JPL','O3v2=1:Ozone:425.95659:0.0362:::31(7,25)-32(6,26):-5.98090::1257.52470:JPL','O3:Ozone:426.06936:0.0064:::38(3,35)-38(2,36):-3.93300::646.17380:JPL','O3:Ozone:427.94374:0.0057:::30(7,23)-31(6,26):-4.46720::529.57660:JPL','O3:Ozone:428.79222:0.0029:::22(1,21)-21(2,20):-3.51940::205.63540:JPL','O3:Ozone:429.42006:0.0041:::8(4,4)-9(3,7):-4.23920::66.01270:JPL','O3v2=1:Ozone:429.43937:0.029:::16(5,11)-17(4,14):-5.51540::880.01280:JPL','O3v1=1,v3=1:Ozone:430.06007:0.0714:::18(3,15)-18(2,16),v1=1:-5.69020::1260.85090:JPL','O3:Ozone:::430.33383:0.05:19(2,18)-19(1,19):-3.50720::157.14710:JPL','O3v2=1:Ozone:::430.90069:11.9917:18(3,15)-18(2,16):-4.79010::858.75230:JPL','O3v1=1,v3=1:Ozone:432.99661:0.0409:::15(3,12)-15(2,13),v3=1:-5.49200::1155.09900:JPL','O3v2=1:Ozone:433.42803:0.0391:::24(1,23)-24(0,24):-5.08710::945.21290:JPL','O3:Ozone:::434.28957:0.05:37(8,30)-38(7,31):-4.87440::775.87720:JPL','O3v2=1:Ozone:434.57444:0.0285:::38(3,35)-38(2,36):-5.38480::1344.91080:JPL','O3v1=1,v3=1:Ozone:436.04849:0.0308:::19(0,19)-18(1,18),v3=1:-5.30950::1181.55340:JPL','O3:Ozone:::437.49356:0.05:16(3,13)-16(2,14):-3.31300::128.11960:JPL','O3v2=1:Ozone:438.38927:0.0256:::19(2,18)-19(1,19):-4.96270::857.33700:JPL','O3:Ozone:::438.48073:0.05:15(5,11)-16(4,12):-4.03940::164.44370:JPL','O3:Ozone:::438.78799:0.05:52(5,47)-51(6,46):-5.63660::1228.49880:JPL','O3v2=1:Ozone:::439.58868:11.9917:17(1,17)-16(0,16):-4.61400::812.94820:JPL','O3:Ozone:::440.07704:0.05:44(9,35)-45(8,38):-5.40110::1069.24120:JPL','O3v2=1:Ozone:441.02730:0.0348:::23(6,18)-24(5,19):-5.62510::1031.73120:JPL','O3:Ozone:::441.34038:0.05:17(1,17)-16(0,16):-3.13960::112.54240:JPL','O3v2=1:Ozone:441.71895:0.0391:::36(3,33)-35(4,32):-5.82980::1280.36800:JPL','O3v2=1:Ozone:441.84939:0.0316:::8(4,4)-9(3,7):-5.68800::767.27870:JPL','O3:Ozone:::442.11757:0.05:7(2,6)-6(1,5):-3.62160::21.29440:JPL','O3v1=1,v3=1:Ozone:442.26961:0.0432:::16(3,13)-16(2,14),v1=1:-5.60660::1230.93380:JPL','O3v1=1,v3=1:Ozone:443.47066:0.0631:::7(2,6)-6(1,5),v1=1:-5.91730::1124.37270:JPL','O3v1=1,v3=1:Ozone:444.10519:0.0443:::13(3,10)-13(2,11),v3=1:-5.49690::1130.58150:JPL','O3v1=1,v3=1:Ozone:444.51983:0.0635:::6(2,4)-5(1,5),v1=1:-5.99920::1118.50560:JPL','O3v1=1,v3=1:Ozone:444.93158:0.0493:::17(1,17)-16(0,16),v1=1:-5.43680::1215.74980:JPL','O3v2=1:Ozone:444.99809:0.0264:::16(3,13)-16(2,14):-4.77020::828.81680:JPL','O3v1=1,v3=1:Ozone:445.01679:0.035:::31(2,29)-31(1,30),v3=1:-5.81180::1459.68260:JPL','O3:Ozone:::445.45871:0.05:6(2,4)-5(1,5):-3.70390::15.35060:JPL','O3v2=1:Ozone:445.99341:0.036:::7(2,6)-6(1,5):-5.08650::722.21640:JPL','O3:Ozone:::446.14377:0.05:32(2,30)-32(1,31):-3.68930::450.11770:JPL','O3:Ozone:::446.64910:0.05:22(6,16)-23(5,19):-4.12720::310.29470:JPL','O3:Ozone:448.14462:0.0051:::23(3,21)-24(0,24):-5.25870::245.49830:JPL','O3:Ozone:::448.87628:0.05:14(3,11)-14(2,12):-3.30910::101.57910:JPL','O3v1=1,v3=1:Ozone:449.21052:0.0426:::21(2,20)-21(1,21),v1=1:-5.81110::1293.65530:JPL','O3:Ozone:449.96678:0.0045:::21(3,19)-22(0,22):-5.23310::207.56230:JPL','O3v2=1:Ozone:450.03779:0.0359:::6(2,4)-5(1,5):-5.16770::716.27530:JPL','O3v1=1,v3=1:Ozone:451.04431:0.0364:::18(1,18)-17(0,17),v3=1:-5.27710::1166.50820:JPL','O3:Ozone:::451.42047:0.05:44(4,40)-43(5,39):-4.92230::874.91950:JPL','O3v2=1:Ozone:451.89055:0.0387:::30(7,23)-31(6,26):-5.89390::1230.71050:JPL','O3v1=1,v3=1:Ozone:452.16605:0.0484:::11(3,8)-11(2,9),v3=1:-5.52360::1109.52600:JPL','O3:Ozone:::452.23890:0.05:36(3,33)-35(4,32):-4.34040::580.76420:JPL','O3v1=1,v3=1:Ozone:452.24374:0.0446:::20(2,19)-20(1,20),v3=1:-5.66490::1212.49740:JPL','O3v1=1,v3=1:Ozone:452.33433:0.046:::14(3,11)-14(2,12),v1=1:-5.60370::1204.49420:JPL','O3v1=1,v3=1:Ozone:452.34094:0.0492:::26(1,25)-26(0,26),v1=1:-5.95120::1390.15260:JPL','O3:Ozone:452.82007:0.0061:::25(3,23)-26(0,26):-5.30050::286.56650:JPL','O3v1=1,v3=1:Ozone:453.59443:0.0525:::8(2,7)-7(1,6),v3=1:-5.73770::1069.10680:JPL','O3:Ozone:::453.78936:0.05:29(7,23)-30(6,24):-4.38320::503.49690:JPL','O3:Ozone:::454.66868:0.05:7(4,4)-8(3,5):-4.29860::58.44640:JPL','O3v2=1:Ozone:454.80908:0.0309:::15(5,11)-16(4,12):-5.48150::865.76970:JPL','O3v2=1:Ozone:455.37838:0.0302:::32(2,30)-32(1,31):-5.14160::1149.26860:JPL','O3v2=1:Ozone:::456.64387:11.9917:14(3,11)-14(2,12):-4.76650::802.37550:JPL','O3v1=1,v3=1:Ozone:457.37445:0.0527:::9(3,6)-9(2,7),v3=1:-5.57770::1091.91060:JPL','O3:Ozone:::457.41562:0.05:12(3,9)-12(2,10):-3.32480::78.53860:JPL','O3:Ozone:458.62071:0.0041:::19(3,17)-20(0,20):-5.23000::172.75680:JPL','O3v1=1,v3=1:Ozone:459.75323:0.0497:::12(3,9)-12(2,10),v1=1:-5.62050::1181.53410:JPL','O3:Ozone:::460.25099:0.05:36(8,28)-37(7,31):-4.77400::743.92840:JPL','O3v1=1,v3=1:Ozone:460.29672:0.057:::7(3,4)-7(2,5),v3=1:-5.67050::1077.70530:JPL','O3v1=1,v3=1:Ozone:461.65034:0.0607:::5(3,2)-5(2,3),v3=1:-5.83210::1066.87810:JPL','O3v1=1,v3=1:Ozone:462.27205:0.0623:::4(3,2)-4(2,3),v3=1:-5.96920::1062.71170:JPL','O3:Ozone:::462.36544:0.05:21(2,20)-21(1,21):-3.49890::190.21250:JPL','O3v1=1,v3=1:Ozone:462.58710:0.0589:::6(3,4)-6(2,5),v3=1:-5.73710::1071.82200:JPL','O3:Ozone:::463.18961:0.05:10(3,7)-10(2,8):-3.36460::58.98100:JPL','O3:Ozone:::463.24596:0.05:40(3,37)-40(2,38):-4.00470::712.86350:JPL','O3v1=1,v3=1:Ozone:463.39454:0.0548:::8(3,6)-8(2,7),v3=1:-5.61250::1084.23710:JPL','O3:Ozone:463.58009:0.0074:::27(3,25)-28(0,28):-5.35450::330.76860:JPL','O3:Ozone:::464.10869:0.05:14(5,9)-15(4,12):-4.01270::150.97720:JPL','O3v1=1,v3=1:Ozone:464.69764:0.0541:::10(3,7)-10(2,8),v1=1:-5.66120::1162.03640:JPL','O3v1=1,v3=1:Ozone:465.01472:0.0503:::10(3,8)-10(2,9),v3=1:-5.53370::1099.94940:JPL','O3v2=1:Ozone:::465.21794:11.9917:20(0,20)-19(1,19):-4.56710::857.33700:JPL','O3v2=1:Ozone:465.36924:0.0302:::12(3,9)-12(2,10):-4.78230::779.42300:JPL','O3:Ozone:::466.02900:0.05:43(9,35)-44(8,36):-5.28610::1031.44470:JPL','O3v1=1,v3=1:Ozone:466.57681:0.04:::20(0,20)-19(1,19),v1=1:-5.39850::1260.52830:JPL','O3:Ozone:::466.64125:0.05:8(3,5)-8(2,6):-3.43620::42.88090:JPL','O3v2=1:Ozone:466.79509:0.0368:::22(6,16)-23(5,19):-5.56300::1011.61370:JPL','O3v1=1,v3=1:Ozone:466.87528:0.0341:::23(1,22)-22(2,21),v3=1:-5.59090::1262.82520:JPL','O3v2=1:Ozone:466.99271:0.0337:::7(4,4)-8(3,5):-5.74900::759.74440:JPL','O3v1=1,v3=1:Ozone:467.61732:0.0586:::8(3,5)-8(2,6),v1=1:-5.73320::1145.97780:JPL','O3v1=1,v3=1:Ozone:467.81714:0.0457:::12(3,10)-12(2,11),v3=1:-5.48240::1118.94960:JPL','O3:Ozone:::467.96712:0.05:20(0,20)-19(1,19):-3.09140::157.14710:JPL','O3:Ozone:::468.39933:0.05:6(3,3)-6(2,4):-3.55680::30.20950:JPL','O3v1=1,v3=1:Ozone:469.08544:0.0627:::6(3,3)-6(2,4),v1=1:-5.85400::1133.33320:JPL','O3:Ozone:::469.10421:0.05:4(3,1)-4(2,2):-3.78710::20.94020:JPL','O3:Ozone:::469.32508:0.05:3(3,1)-3(2,2):-4.03620::17.57150:JPL','O3:Ozone:::469.48235:0.05:5(3,3)-5(2,4):-3.64890::25.12970:JPL','O3:Ozone:::469.95389:0.05:7(3,5)-7(2,6):-3.48550::36.04190:JPL','O3v1=1,v3=1:Ozone:469.99313:0.0645:::5(3,3)-5(2,4),v1=1:-5.94620::1128.26570:JPL','O3v1=1,v3=1:Ozone:470.42049:0.0609:::7(3,5)-7(2,6),v1=1:-5.78300::1139.16530:JPL','O3v1=1,v3=1:Ozone:470.95844:0.0527:::7(2,5)-6(1,6),v3=1:-5.82010::1061.99580:JPL','O3:Ozone:470.99087:0.0039:::9(3,7)-9(2,8):-3.38810::50.30210:JPL','O3v2=1:Ozone:471.08779:0.0298:::21(2,20)-21(1,21):-4.95390::890.22700:JPL','O3:Ozone:471.23288:0.0103:::50(10,40)-51(9,43):-5.91200::1365.86440:JPL','O3v2=1:Ozone:471.27834:0.0322:::10(3,7)-10(2,8):-4.82230::759.94200:JPL','O3v1=1,v3=1:Ozone:471.33768:0.0566:::9(3,7)-9(2,8),v1=1:-5.68610::1153.40990:JPL','O3v1=1,v3=1:Ozone:472.19790:0.0416:::14(3,12)-14(2,13),v3=1:-5.45020::1141.22660:JPL','O3:Ozone:472.30077:0.0067:::21(6,16)-22(5,17):-4.06960::290.94150:JPL','O3v2=1:Ozone:472.73550:0.0323:::40(3,37)-40(2,38):-5.45570::1411.32120:JPL','O3:Ozone:472.89885:0.0038:::11(3,9)-11(2,10):-3.32560::67.90280:JPL','O3v1=1,v3=1:Ozone:472.91110:0.0893:::25(1,24)-25(0,25),v3=1:-5.76860::1303.32680:JPL','O3v1=1,v3=1:Ozone:473.00787:0.0523:::11(3,9)-11(2,10),v1=1:-5.62400::1170.99290:JPL','O3:Ozone:473.07308:0.0093:::42(4,38)-42(3,39):-4.09490::799.59100:JPL','O3:Ozone:474.16522:0.0092:::44(4,40)-44(3,41):-4.22250::874.16090:JPL','O3:Ozone:474.31174:0.004:::17(3,15)-18(0,18):-5.25780::141.08200:JPL','O3v2=1:Ozone:::474.81129:11.9917:8(3,5)-8(2,6):-4.89390::743.90640:JPL','O3v1=1,v3=1:Ozone:475.72952:0.0481:::13(3,11)-13(2,12),v1=1:-5.58490::1191.90590:JPL','O3:Ozone:476.02061:0.0037:::13(3,11)-13(2,12):-3.28590::88.83480:JPL','O3v2=1:Ozone:476.61230:0.0358:::6(3,3)-6(2,4):-5.01470::731.28690:JPL','O3v2=2:Ozone:477.28627:0.0829:::19(1,19)-18(0,18):-5.97880::1539.00120:JPL','O3v2=1:Ozone:477.33437:0.0371:::4(3,1)-4(2,2):-5.24510::722.05620:JPL','O3v2=1:Ozone:477.56063:0.0376:::3(3,1)-3(2,2):-5.49410::718.70160:JPL','O3v2=1:Ozone:477.65113:0.0413:::29(7,23)-30(6,24):-5.81280::1204.73980:JPL','O3v2=1:Ozone:477.72133:0.0365:::5(3,3)-5(2,4):-5.10680::726.22760:JPL','O3v2=1:Ozone:478.20360:0.035:::7(3,5)-7(2,6):-4.94320::737.09310:JPL','O3v1=1,v3=1:Ozone:478.55581:0.0383:::16(3,14)-16(2,15),v3=1:-5.43290::1166.76750:JPL','O3v2=1:Ozone:479.26456:0.0333:::9(3,7)-9(2,8):-4.84580::751.29220:JPL','O3:Ozone:479.39567:0.0093:::40(4,36)-40(3,37):-3.96810::728.31570:JPL','O3v2=1:Ozone:479.47588:0.028:::19(1,19)-18(0,18):-4.53440::841.34340:JPL','O3:Ozone:479.63245:0.0069:::28(7,21)-29(6,24):-4.30490::478.25850:JPL','O3v1=1,v3=1:Ozone:479.82148:0.0445:::15(3,13)-15(2,14),v1=1:-5.56300::1216.13890:JPL','O3v1=1,v3=1:Ozone:479.86935:0.0819:::29(2,27)-28(3,26),v3=1:-5.96780::1406.89880:JPL','O3:Ozone:479.96295:0.0091:::29(3,27)-30(0,30):-5.41790::378.10630:JPL','O3:Ozone:480.01418:0.0046:::6(4,2)-7(3,5):-4.40130::51.71790:JPL','O3:Ozone:480.18236:0.0046:::26(1,25)-26(0,26):-3.62160::286.56650:JPL','O3v2=1:Ozone:480.34286:0.0329:::14(5,9)-15(4,12):-5.45680::852.35970:JPL','O3:Ozone:480.71725:0.0037:::15(3,13)-15(2,14):-3.26320::113.08700:JPL','O3v1=1,v3=1:Ozone:480.72137:0.1478:::37(4,33)-37(3,34),v3=1:-5.95780::1662.64780:JPL','O3v2=1:Ozone:481.17267:0.0399:::42(4,38)-42(3,39):-5.54750::1498.10330:JPL','O3:Ozone:481.19895:0.0045:::9(2,8)-8(1,7):-3.49150::34.25100:JPL','O3v2=1:Ozone:481.21686:0.0313:::11(3,9)-11(2,10):-4.78300::768.81690:JPL','O3v1=1,v3=1:Ozone:481.61770:0.0456:::23(2,22)-23(1,23),v1=1:-5.81480::1329.96320:JPL','O3:Ozone:481.62014:0.0039:::19(1,19)-18(0,18):-3.05980::141.08200:JPL','O3v2=1:Ozone:482.57236:0.0442:::44(4,40)-44(3,41):-5.67430::1572.39470:JPL','O3v1=1,v3=1:Ozone:483.36684:0.0667:::9(2,8)-8(1,7),v1=1:-5.78660::1137.28660:JPL','O3:Ozone:483.54445:0.009:::46(4,42)-46(3,43):-4.35010::951.97230:JPL','O3v2=1:Ozone:484.41124:0.0292:::13(3,11)-13(2,12):-4.74310::789.65800:JPL','O3v2=1:Ozone:484.62655:0.0378:::9(2,8)-8(1,7):-4.95760::735.12680:JPL','O3v1=1,v3=1:Ozone:485.09303:0.047:::19(1,19)-18(0,18),v1=1:-5.35770::1244.34740:JPL','O3v1=1,v3=1:Ozone:485.60723:0.0415:::17(3,15)-17(2,16),v1=1:-5.55480::1243.68070:JPL','O3:Ozone:486.12122:0.0066:::35(8,28)-36(7,29):-4.67910::712.82220:JPL','O3:Ozone:486.45101:0.0238:::38(2,36)-39(1,39):-5.94670::629.94750:JPL','O3v1=1,v3=1:Ozone:487.26934:0.036:::18(3,16)-18(2,17),v3=1:-5.42760::1195.55800:JPL','O3:Ozone:487.34853:0.0038:::17(3,15)-17(2,16):-3.25400::140.64710:JPL','O3v2=1:Ozone:487.38345:0.038:::40(4,36)-40(3,37):-5.42150::1427.08990:JPL','O3v1=1,v3=1:Ozone:487.50266:0.0565:::22(2,21)-22(1,22),v3=1:-5.65940::1246.56390:JPL','O3v1=1,v3=1:Ozone:487.66635:0.0324:::21(0,21)-20(1,20),v3=1:-5.22050::1212.49740:JPL','O3v1=1,v3=1:Ozone:488.41540:0.032:::24(1,23)-23(2,22),v1=1:-5.72390::1346.02830:JPL','O3v2=1:Ozone:489.21703:0.0273:::15(3,13)-15(2,14):-4.72020::813.80420:JPL','O3:Ozone:489.53710:0.0059:::13(5,9)-14(4,10):-3.99690::138.35920:JPL','O3v1=1,v3=1:Ozone:489.71651:0.0541:::10(2,9)-9(1,8),v3=1:-5.62030::1083.61420:JPL','O3v2=1:Ozone:489.85929:0.054:::26(1,25)-26(0,26):-5.07480::986.06090:JPL','O3v1=1,v3=1:Ozone:491.62833:0.0362:::20(1,20)-19(0,19),v3=1:-5.20500::1196.09840:JPL','O3:Ozone:491.93796:0.0078:::42(9,33)-43(8,36):-5.17620::994.48680:JPL','O3:Ozone:491.95117:0.0091:::38(4,34)-38(3,35):-3.84360::660.38590:JPL','O3v2=1:Ozone:492.23558:0.0358:::6(4,2)-7(3,5):-5.85320::753.04430:JPL','O3v2=1:Ozone:492.35403:0.0388:::21(6,16)-22(5,17):-5.50770::992.34160:JPL','O3v2=1:Ozone:492.46480:0.0518:::46(4,42)-46(3,43):-5.80090::1649.90970:JPL','O3v1=1,v3=1:Ozone:493.39920:0.0391:::19(3,17)-19(2,18),v1=1:-5.55820::1274.51870:JPL','O3v2=1:Ozone:::494.68154:11.9917:24(1,23)-23(2,22):-4.86690::943.16920:JPL','O3v2=1:Ozone:496.00203:0.0256:::17(3,15)-17(2,16):-4.71070::841.24290:JPL','O3:Ozone:496.25401:0.0039:::19(3,17)-19(2,18):-3.25630::171.50150:JPL','O3:Ozone:497.04308:0.004:::15(3,13)-16(0,16):-5.32750::112.54240:JPL','O3:Ozone:497.09787:0.0045:::23(2,22)-23(1,23):-3.50110::226.45190:JPL','O3:Ozone:497.11747:0.0107:::49(10,40)-50(9,41):-5.78820::1323.07780:JPL','O3v1=1,v3=1:Ozone:497.49971:0.1266:::35(4,31)-35(3,32),v3=1:-5.84340::1600.52110:JPL','O3:Ozone:497.97273:0.0072:::20(6,14)-21(5,17):-4.01870::272.42950:JPL','O3v1=1,v3=1:Ozone:498.67585:0.0347:::20(3,18)-20(2,19),v3=1:-5.43270::1227.58260:JPL','O3:Ozone:498.74743:0.0036:::24(1,23)-23(2,22):-3.38990::243.03330:JPL','O3v2=1:Ozone:500.00519:0.0375:::38(4,34)-38(3,35):-5.29760::1359.40660:JPL','O3:Ozone:500.43348:0.0056:::34(2,32)-34(1,33):-3.71880::505.50650:JPL','O3v1=1,v3=1:Ozone:500.78160:0.0434:::33(2,31)-33(1,32),v3=1:-5.83360::1512.64830:JPL','O3:Ozone:500.79324:0.0102:::37(4,34)-38(1,37):-5.55720::625.58270:JPL','O3v2=1:Ozone:501.11203:0.0254:::30(2,28)-29(3,27):-5.27780::1094.00360:JPL','O3:Ozone:501.47633:0.0114:::31(3,29)-32(0,32):-5.48940::428.58040:JPL','O3:Ozone:501.77138:0.0087:::48(4,44)-48(3,45):-4.47740::1032.97210:JPL','O3:Ozone:502.75876:0.0095:::35(4,32)-36(1,35):-5.45780::563.99400:JPL','O3v2=1:Ozone:503.40945:0.0439:::28(7,21)-29(6,24):-5.73720::1179.60690:JPL','O3v1=1,v3=1:Ozone:503.48414:0.0377:::21(3,19)-21(2,20),v1=1:-5.57180::1308.63940:JPL','O3v2=1:Ozone:505.11323:0.0243:::19(3,17)-19(2,18):-4.71270::871.96010:JPL','O3:Ozone:505.25722:0.0048:::5(4,2)-6(3,3):-4.57880::45.83360:JPL','O3:Ozone:505.36914:0.0075:::27(7,21)-28(6,22):-4.23230::453.86470:JPL','O3v2=1:Ozone:505.67129:0.035:::13(5,9)-14(4,10):-5.44280::839.79470:JPL','O3:Ozone:505.86899:0.0109:::39(4,36)-40(1,39):-5.66450::690.27640:JPL','O3v1=1,v3=1:Ozone:506.05662:0.0699:::8(2,6)-7(1,7),v1=1:-5.86880::1129.09750:JPL','O3v1=1,v3=1:Ozone:506.16196:0.0706:::28(1,27)-28(0,28),v1=1:-5.95330::1434.44640:JPL','O3v2=1:Ozone:::506.53833:11.9917:23(2,22)-23(1,23):-4.95570::926.27270:JPL','O3:Ozone:507.73614:0.0042:::21(3,19)-21(2,20):-3.26850::205.63540:JPL','O3:Ozone:507.81538:0.0045:::30(2,28)-29(3,27):-3.79700::394.11620:JPL','O3:Ozone:508.03103:0.0077:::42(3,39)-42(2,40):-4.08040::782.64490:JPL','O3:Ozone:508.19290:0.0047:::8(2,6)-7(1,7):-3.57540::25.92940:JPL','O3:Ozone:509.27249:0.0088:::36(4,32)-36(3,33):-3.72290::595.84930:JPL','O3v2=1:Ozone:::510.83387:0.07:34(2,32)-34(1,33):-5.17050::1204.39650:JPL','O3v2=1:Ozone:::511.41272:0.1:48(4,44)-48(3,45):-5.92730::1730.59430:JPL','O3:Ozone:::511.95322:0.05:34(8,26)-35(7,29):-4.58910::682.55670:JPL','O3:Ozone:512.26806:0.009:::33(4,30)-34(1,33):-5.36890::505.50650:JPL','O3v2=1:Ozone:512.82232:0.0392:::8(2,6)-7(1,7):-5.04020::726.80050:JPL','O3v1=1,v3=1:Ozone:513.05416:0.0346:::22(3,20)-22(2,21),v3=1:-5.44690::1262.82520:JPL','O3v2=2:Ozone:514.14853:0.0803:::22(0,22)-21(1,21):-5.93520::1587.63240:JPL','O3:Ozone:::514.99262:0.05:12(5,7)-13(4,10):-3.99270::126.58180:JPL','O3v1=1,v3=1:Ozone:516.10991:0.0379:::23(3,21)-23(2,22),v1=1:-5.59450::1346.02830:JPL','O3v1=1,v3=1:Ozone:516.43381:0.0526:::25(2,24)-25(1,25),v1=1:-5.82860::1369.44410:JPL','O3v2=1:Ozone:::516.85719:11.9917:21(3,19)-21(2,20):-4.72450::905.94080:JPL','O3v2=1:Ozone:517.17559:0.0253:::22(0,22)-21(1,21):-4.49030::890.22700:JPL','O3:Ozone:517.42652:0.0117:::41(4,38)-42(1,41):-5.77820::758.07850:JPL','O3:Ozone:::517.43040:0.05:11(2,10)-10(1,9):-3.38540::50.64320:JPL','O3v2=1:Ozone:::517.54469:0.1:36(4,32)-36(3,33):-5.17750::1295.10210:JPL','O3:Ozone:::517.78551:0.05:41(9,33)-42(8,34):-5.07100::958.36820:JPL','O3v2=1:Ozone:::517.93424:0.1:20(6,14)-21(5,17):-5.45890::973.90710:JPL','O3v2=2:Ozone:518.24060:0.086:::21(1,21)-20(0,20):-5.91680::1570.34570:JPL','O3v2=1:Ozone:::518.64442:0.15:42(3,39)-42(2,40):-5.53050::1480.80320:JPL','O3v1=1,v3=1:Ozone:518.91427:0.199:::33(4,29)-33(3,30),v3=1:-5.83640::1541.78890:JPL','O3v1=1,v3=1:Ozone:519.24151:0.0385:::22(0,22)-21(1,21),v1=1:-5.32050::1293.65530:JPL','O3:Ozone:::520.13480:0.05:22(0,22)-21(1,21):-3.01510::190.21250:JPL','O3v2=1:Ozone:520.35144:0.039:::11(2,10)-10(1,9):-4.85270::751.45990:JPL','O3v1=1,v3=1:Ozone:520.61064:0.0691:::11(2,10)-10(1,9),v1=1:-5.68010::1153.62720:JPL','O3v2=1:Ozone:::520.78747:11.9917:21(1,21)-20(0,20):-4.47250::872.85490:JPL','O3:Ozone:::522.04611:0.05:23(3,21)-23(2,22):-3.28960::243.03330:JPL','O3v1=1,v3=1:Ozone:522.92446:0.055:::12(2,11)-11(1,10),v3=1:-5.52430::1101.50670:JPL','O3:Ozone:522.95905:0.011:::48(10,38)-49(9,41):-5.66890::1281.12720:JPL','O3:Ozone:523.30883:0.0043:::21(1,21)-20(0,20):-2.99800::172.75680:JPL','O3:Ozone:523.53165:0.0077:::19(6,14)-20(5,15):-3.97470::254.76280:JPL','O3v1=1,v3=1:Ozone:525.33402:0.0747:::24(2,23)-24(1,24),v3=1:-5.66440::1283.74600:JPL','O3:Ozone:526.53166:0.0043:::13(3,11)-14(0,14):-5.45250::87.14990:JPL','O3v1=1,v3=1:Ozone:526.54524:0.044:::21(1,21)-20(0,20),v1=1:-5.29630::1276.09170:JPL','O3:Ozone:::526.73235:0.05:54(5,49)-53(6,48):-5.64850::1316.96290:JPL','O3:Ozone:527.60968:0.0144:::33(3,31)-34(0,34):-5.56820::482.19100:JPL','O3v2=1:Ozone:529.05792:0.0466:::27(7,21)-28(6,22):-5.66710::1155.31510:JPL','O3:Ozone:::529.08045:0.05:50(4,46)-50(3,47):-4.60500::1117.11000:JPL','O3:Ozone:::529.68386:0.05:34(4,30)-34(3,31):-3.60770::534.74910:JPL','O3v1=1,v3=1:Ozone:529.70591:0.1223:::27(1,26)-27(0,27),v3=1:-5.76350::1345.22980:JPL','O3:Ozone:529.72056:0.0084:::31(4,28)-32(1,31):-5.29410::450.11770:JPL','O3:Ozone:::530.50105:0.05:4(4,0)-5(3,3):-4.92420::40.79000:JPL','O3v1=1,v3=1:Ozone:530.61124:0.0362:::24(3,22)-24(2,23),v3=1:-5.46950::1301.26920:JPL','O3v2=1:Ozone:531.02776:0.0371:::12(5,7)-13(4,10):-5.44030::828.06690:JPL','O3:Ozone:::531.07622:0.05:26(7,19)-27(6,22):-4.16510::430.31280:JPL','O3v1=1,v3=1:Ozone:531.47442:0.0411:::25(3,23)-25(2,24),v1=1:-5.62550::1386.67050:JPL','O3v2=1:Ozone:::531.48406:11.9917:23(3,21)-23(2,22):-4.74520::943.16920:JPL','O3v2=1:Ozone:532.69813:0.0445:::38(3,35)-37(4,34):-5.75180::1341.63770:JPL','O3v1=1,v3=1:Ozone:533.62161:0.0366:::22(1,22)-21(0,21),v3=1:-5.15000::1228.76420:JPL','O3:Ozone:::534.27312:0.05:25(2,24)-25(1,25):-3.51360::265.85790:JPL','O3v1=1,v3=1:Ozone:534.59457:0.0398:::25(1,24)-24(2,23),v3=1:-5.47990::1301.26920:JPL','O3:Ozone:534.87792:0.013:::43(4,40)-44(1,43):-5.89750::828.99160:JPL','O3:Ozone:::535.92266:0.05:28(1,27)-28(0,28):-3.62450::330.76860:JPL','O3v1=1,v3=1:Ozone:536.74628:0.0699:::3(3,0)-2(2,1),v3=1:-5.63560::1056.91160:JPL','O3v1=1,v3=1:Ozone:536.98009:0.0555:::9(2,7)-8(1,8),v3=1:-5.71370::1073.99890:JPL','O3:Ozone:::537.71667:0.05:33(8,26)-34(7,27):-4.50410::653.13320:JPL','O3v1=1,v3=1:Ozone:537.89099:0.0343:::23(0,23)-22(1,22),v3=1:-5.15700::1246.56390:JPL','O3v2=1:Ozone:538.29165:0.0371:::34(4,30)-34(3,31):-5.06280::1234.22100:JPL','O3:Ozone:::539.37269:0.05:25(3,23)-25(2,24):-3.31910::283.67940:JPL','O3v1=1,v3=1:Ozone:539.86274:0.0877:::31(4,27)-31(3,28),v3=1:-5.63140::1486.49310:JPL','O3:Ozone:::540.35598:0.05:11(5,7)-12(4,8):-4.00240::115.64880:JPL','O3:Ozone:::542.58041:0.05:38(3,35)-37(4,34):-4.26750::642.28740:JPL','O3v2=1:Ozone:543.39793:0.0431:::19(6,14)-20(5,15):-5.41690::956.31470:JPL','O3:Ozone:::543.58908:0.05:40(9,31)-41(8,34):-4.97040::923.08870:JPL','O3:Ozone:::543.61341:0.05:46(4,42)-45(5,41):-4.89190::949.96860:JPL','O3v2=1:Ozone:544.48621:0.0483:::25(2,24)-25(1,25):-4.96760::965.46690:JPL','O3:Ozone:::544.85749:0.05:3(3,1)-2(2,0):-3.45270::15.05200:JPL','O3v1=1,v3=1:Ozone:545.28283:0.0773:::3(3,1)-2(2,0),v1=1:-5.74970::1118.20010:JPL','O3v1=1,v3=1:Ozone:545.67567:0.1339:::34(4,30)-34(3,31),v1=1:-5.89080::1636.13530:JPL','O3v2=1:Ozone:546.68499:0.0741:::28(1,27)-28(0,28):-5.07730::1030.02420:JPL','O3:Ozone:548.75368:0.0114:::47(10,38)-48(9,39):-5.55390::1240.01280:JPL','O3:Ozone:::549.07042:0.05:18(6,12)-19(5,15):-3.93780::237.93840:JPL','O3v2=1:Ozone:549.21769:0.0242:::25(3,23)-25(2,24):-4.77420::983.62900:JPL','O3v1=1,v3=1:Ozone:549.71756:0.0489:::27(3,25)-27(2,26),v1=1:-5.66440::1430.55090:JPL','O3:Ozone:::550.98947:0.05:13(2,12)-12(1,11):-3.29900::70.45570:JPL','O3:Ozone:::551.43574:0.05:32(4,28)-32(3,29):-3.49960::477.12350:JPL','O3v1=1,v3=1:Ozone:551.47295:0.0405:::26(3,24)-26(2,25),v3=1:-5.50000::1342.89820:JPL','O3v2=1:Ozone:552.77108:0.0434:::3(3,1)-2(2,0):-4.91330::716.19280:JPL','O3v2=1:Ozone:553.35013:0.0396:::13(2,12)-12(1,11):-4.76720::771.20020:JPL','O3v1=1,v3=1:Ozone:553.41939:0.0667:::27(2,26)-27(1,27),v1=1:-5.85230::1412.09080:JPL','O3v1=1,v3=1:Ozone:553.48838:0.055:::14(2,13)-13(1,12),v3=1:-5.44580::1122.76420:JPL','O3v2=1:Ozone:554.67604:0.0493:::26(7,19)-27(6,22):-5.60210::1131.86190:JPL','O3v1=1,v3=1:Ozone:555.34220:0.0701:::13(2,12)-12(1,11),v1=1:-5.59360::1173.38160:JPL','O3:Ozone:555.34496:0.0079:::29(4,26)-30(1,29):-5.23780::397.82830:JPL','O3:Ozone:556.20992:0.0155:::45(5,41)-46(2,44):-5.98050::931.41540:JPL','O3v2=1:Ozone:556.28943:0.0392:::11(5,7)-12(4,8):-5.45140::817.18010:JPL','O3:Ozone:::556.71038:0.05:25(7,19)-26(6,20):-4.10340::407.60430:JPL','O3v1=1,v3=1:Ozone:556.73368:0.0361:::26(1,25)-25(2,24),v1=1:-5.62480::1386.67050:JPL','O3:Ozone:557.84450:0.0181:::35(3,33)-36(0,36):-5.65410::538.93740:JPL','O3:Ozone:::557.97008:0.05:36(2,34)-36(1,35):-3.75840::563.99400:JPL','O3:Ozone:559.47741:0.0084:::44(3,41)-44(2,42):-4.16140::855.49870:JPL','O3v1=1,v3=1:Ozone:559.67398:0.058:::35(2,33)-35(1,34),v3=1:-5.86570::1568.66280:JPL','O3:Ozone:::559.83439:0.05:27(3,25)-27(2,26):-3.35630::327.55790:JPL','O3v1=1,v3=1:Ozone:560.03384:0.1028:::30(1,29)-30(0,30),v1=1:-5.96960::1481.87730:JPL','O3v2=1:Ozone:560.45589:0.0362:::32(4,28)-32(3,29):-4.95510::1176.80210:JPL','O3v2=2:Ozone:560.54038:0.0961:::23(1,23)-22(0,22):-5.87110::1604.78250:JPL','O3v1=1,v3=1:Ozone:561.52021:0.0708:::4(3,2)-3(2,1),v3=1:-5.58310::1059.40110:JPL','O3v1=1,v3=1:Ozone:561.88758:0.0664:::29(4,25)-29(3,26),v3=1:-5.53670::1434.66850:JPL','O3:Ozone:562.14960:0.0047:::11(3,9)-12(0,12):-5.64990::64.92570:JPL','O3v1=1,v3=1:Ozone:562.35468:0.1016:::31(2,29)-30(3,28),v3=1:-5.86590::1455.76860:JPL','O3v2=1:Ozone:562.76540:0.0223:::26(1,25)-25(2,24):-4.76890::983.62900:JPL','O3:Ozone:::563.43772:0.05:32(8,24)-33(7,27):-4.42380::624.55070:JPL','O3v2=1:Ozone:::563.44494:11.9917:23(1,23)-22(0,22):-4.42700::907.47810:JPL','O3v2=2:Ozone:564.55034:0.0936:::24(0,24)-23(1,23):-5.88170::1623.48010:JPL','O3v2=1:Ozone:564.93965:0.0496:::33(8,26)-34(7,27):-5.93600::1354.60720:JPL','O3:Ozone:::565.34147:0.05:52(4,48)-52(3,49):-4.73370::1204.34210:JPL','O3v1=1,v3=1:Ozone:565.41535:0.1001:::26(2,25)-26(1,26),v3=1:-5.67960::1324.03800:JPL','O3:Ozone:565.70348:0.0067:::10(5,5)-11(4,8):-4.02860::105.55760:JPL','O3:Ozone:566.29357:0.0047:::23(1,23)-22(0,22):-2.95260::207.56230:JPL','O3v1=1,v3=1:Ozone:566.63070:0.0978:::32(4,28)-32(3,29),v1=1:-5.78330::1578.76910:JPL','O3:Ozone:566.73498:0.0044:::26(1,25)-25(2,24):-3.29330::283.67940:JPL','O3v2=1:Ozone:::567.80392:11.9917:24(0,24)-23(1,23):-4.43730::926.27270:JPL','O3v2=1:Ozone:568.84100:0.0452:::18(6,12)-19(5,15):-5.38170::939.56090:JPL','O3v1=1,v3=1:Ozone:569.24218:0.0412:::23(1,23)-22(0,22),v1=1:-5.25140::1310.97540:JPL','O3:Ozone:::569.34126:0.05:39(9,31)-40(8,32):-4.87420::888.64850:JPL','O3v2=1:Ozone:569.56632:0.0603:::36(2,34)-36(1,35):-5.20950::1262.60370:JPL','O3:Ozone:::570.09714:0.05:4(3,1)-3(2,2):-3.40000::17.57150:JPL','O3v2=1:Ozone:570.14207:0.0256:::27(3,25)-27(2,26):-4.81100::1027.30440:JPL','O3v1=1,v3=1:Ozone:570.49040:0.0795:::4(3,1)-3(2,2),v1=1:-5.69690::1120.71660:JPL','O3v1=1,v3=1:Ozone:570.58136:0.0373:::24(0,24)-23(1,23),v1=1:-5.26660::1329.96320:JPL','O3v1=1,v3=1:Ozone:570.91656:0.0621:::29(3,27)-29(2,28),v1=1:-5.71090::1477.65450:JPL','O3:Ozone:::570.99937:0.05:24(0,24)-23(1,23):-2.96270::226.45190:JPL','O3v2=1:Ozone:571.31903:0.0445:::44(3,41)-44(2,42):-5.61080::1553.33750:JPL','O3v1=1,v3=1:Ozone:572.58143:0.0756:::10(2,8)-9(1,9),v1=1:-5.77630::1142.93710:JPL','O3:Ozone:::572.87722:0.05:30(4,26)-30(3,27):-3.40010::423.00330:JPL','O3:Ozone:573.60280:0.0064:::27(2,26)-27(1,27):-3.53610::308.42460:JPL','O3:Ozone:::574.50764:0.05:46(10,36)-47(9,39):-5.44310::1199.73480:JPL','O3:Ozone:::574.54045:0.05:17(6,12)-18(5,13):-3.90820::221.95760:JPL','O3v1=1,v3=1:Ozone:575.67907:0.0484:::28(3,26)-28(2,27),v3=1:-5.53800::1387.69620:JPL','O3:Ozone:::576.51487:0.05:10(2,8)-9(1,9):-3.48580::39.75060:JPL','O3v1=1,v3=1:Ozone:576.79973:0.0381:::24(1,24)-23(0,23),v3=1:-5.11100::1264.50600:JPL','O3v2=1:Ozone:::577.89493:11.9917:4(3,1)-3(2,2):-4.86130::718.70160:JPL','O3v2=1:Ozone:580.21890:0.052:::25(7,19)-26(6,20):-5.54240::1109.24870:JPL','O3v2=1:Ozone:::581.32156:11.9917:10(2,8)-9(1,9):-4.95140::740.55130:JPL','O3v2=1:Ozone:581.53487:0.0414:::10(5,5)-11(4,8):-5.47900::807.13150:JPL','O3v1=1,v3=1:Ozone:581.80309:0.0541:::16(2,15)-15(1,14),v3=1:-5.38190::1147.36060:JPL','O3:Ozone:::582.14412:0.05:15(2,14)-14(1,13):-3.22900::93.66870:JPL','O3:Ozone:582.22209:0.0153:::43(5,39)-44(2,42):-5.84610::855.49870:JPL','O3v1=1,v3=1:Ozone:582.24030:0.0513:::27(4,23)-27(3,24),v3=1:-5.45240::1386.33940:JPL','O3:Ozone:::582.30637:0.05:24(7,17)-25(6,20):-4.04710::385.73800:JPL','O3v2=1:Ozone:582.34335:0.0348:::30(4,26)-30(3,27):-4.85600::1122.87740:JPL','O3:Ozone:::582.43163:0.05:56(5,51)-56(4,52):-5.06840::1410.04750:JPL','O3:Ozone:::583.47500:0.05:29(3,27)-29(2,28):-3.40120::374.65350:JPL','O3v2=1:Ozone:::583.89378:11.9917:15(2,14)-14(1,13):-4.69800::794.32750:JPL','O3v2=1:Ozone:584.03296:0.0284:::32(2,30)-31(3,29):-5.18770::1144.97720:JPL','O3v2=1:Ozone:::584.62827:11.9917:27(2,26)-27(1,27):-4.98970::1007.80330:JPL','O3v1=1,v3=1:Ozone:585.98197:0.1644:::29(1,28)-29(0,29),v3=1:-5.77360::1390.21810:JPL','O3:Ozone:::586.13279:0.1:54(5,49)-54(4,50):-4.88780::1314.98150:JPL','O3v1=1,v3=1:Ozone:586.42319:0.0701:::30(4,26)-30(3,27),v1=1:-5.68510::1524.90010:JPL','O3v1=1,v3=1:Ozone:586.55492:0.0714:::5(3,2)-4(2,3),v3=1:-5.52720::1062.71170:JPL','O3v1=1,v3=1:Ozone:587.01748:0.0371:::25(0,25)-24(1,24),v3=1:-5.11500::1283.74600:JPL','O3:Ozone:::587.53605:0.05:58(5,53)-58(4,54):-5.24940::1508.34040:JPL','O3v1=1,v3=1:Ozone:587.77388:0.0698:::15(2,14)-14(1,13),v1=1:-5.52360::1196.53280:JPL','O3:Ozone:::589.10431:0.05:31(8,24)-32(7,25):-4.34820::596.80970:JPL','O3:Ozone:589.11577:0.0074:::27(4,24)-28(1,27):-5.20510::348.64510:JPL','O3:Ozone:::590.33754:0.05:32(2,30)-31(3,29):-3.70970::445.30790:JPL','O3v2=1:Ozone:590.57471:0.0524:::32(8,24)-33(7,27):-5.85800::1326.14450:JPL','O3:Ozone:::591.00434:0.05:9(5,5)-10(4,6):-4.07610::96.30890:JPL','O3:Ozone:::591.24892:0.05:30(1,29)-30(0,30):-3.64210::378.10630:JPL','O3:Ozone:591.66443:0.0229:::37(3,35)-38(0,38):-5.74720::598.81870:JPL','O3v1=1,v3=1:Ozone:592.31325:0.0912:::29(2,28)-29(1,29),v1=1:-5.88580::1457.89710:JPL','O3:Ozone:::592.63075:0.05:28(4,24)-28(3,25):-3.31060::372.41030:JPL','O3v2=1:Ozone:594.21330:0.0474:::17(6,12)-18(5,13):-5.35380::923.64740:JPL','O3v2=1:Ozone:::594.30857:11.9917:29(3,27)-29(2,28):-4.85540::1074.17950:JPL','O3:Ozone:::595.05120:0.1:38(9,29)-39(8,32):-4.78220::855.04750:JPL','O3:Ozone:::595.08091:0.05:5(3,3)-4(2,2):-3.34450::20.94020:JPL','O3v1=1,v3=1:Ozone:595.08452:0.0809:::31(3,29)-31(2,30),v1=1:-5.76480::1527.96690:JPL','O3v1=1,v3=1:Ozone:595.48245:0.0815:::5(3,3)-4(2,2),v1=1:-5.64120::1124.07990:JPL','O3:Ozone:::597.64857:0.05:52(5,47)-52(4,48):-4.70850::1223.19980:JPL','O3v1=1,v3=1:Ozone:599.60719:0.046:::27(1,26)-26(2,25),v3=1:-5.40030::1342.89820:JPL','O3v1=1,v3=1:Ozone:599.83856:0.0429:::25(4,21)-25(3,22),v3=1:-5.37980::1341.51740:JPL','O3:Ozone:::599.97876:0.05:16(6,10)-17(5,13):-3.88650::206.81920:JPL','O3:Ozone:::600.22004:0.05:45(10,36)-46(9,37):-5.33630::1160.29330:JPL','O3:Ozone:::602.09145:0.05:60(5,55)-60(4,56):-5.43030::1609.80300:JPL','O3v2=1:Ozone:602.53604:0.0329:::28(4,24)-28(3,25):-4.76690::1072.46910:JPL','O3v2=1:Ozone:::602.78370:11.9917:5(3,3)-4(2,2):-4.80630::722.05620:JPL','O3:Ozone:602.94024:0.0052:::9(3,7)-10(0,10):-5.94300::45.90080:JPL','O3v2=1:Ozone:603.07620:0.0999:::30(1,29)-30(0,30):-5.09440::1077.10440:JPL','O3v1=1,v3=1:Ozone:603.18296:0.0608:::30(3,28)-30(2,29),v3=1:-5.58360::1435.64860:JPL','O3v2=2:Ozone:603.96510:0.1152:::25(1,25)-24(0,24):-5.84070::1642.31150:JPL','O3v1=1,v3=1:Ozone:604.03376:0.0527:::28(4,24)-28(3,25),v1=1:-5.59770::1474.54180:JPL','O3v2=1:Ozone:605.72251:0.0547:::24(7,17)-25(6,20):-5.48790::1087.47420:JPL','O3v2=1:Ozone:606.73231:0.0435:::9(5,5)-10(4,6):-5.52770::797.92190:JPL','O3v2=1:Ozone:::607.19964:11.9917:25(1,25)-24(0,24):-4.39690::945.21290:JPL','O3v1=1,v3=1:Ozone:607.40854:0.1345:::28(2,27)-28(1,28),v3=1:-5.70500::1367.43530:JPL','O3:Ozone:::607.84851:0.05:23(7,17)-24(6,18):-3.99610::364.71430:JPL','O3v1=1,v3=1:Ozone:608.41290:0.0524:::18(2,17)-17(1,16),v3=1:-5.33010::1175.26350:JPL','O3:Ozone:::609.73059:0.05:26(4,22)-26(3,23):-3.23270::325.35440:JPL','O3v1=1,v3=1:Ozone:609.78754:0.0582:::11(2,9)-10(1,10),v3=1:-5.64550::1089.18570:JPL','O3:Ozone:::610.01665:0.05:54(4,50)-54(3,51):-4.86510::1294.63350:JPL','O3:Ozone:::610.26256:0.05:31(3,29)-31(2,30):-3.45370::424.95170:JPL','O3:Ozone:::610.36519:0.05:25(1,25)-24(0,24):-2.92270::245.49830:JPL','O3v1=1,v3=1:Ozone:610.80316:0.0719:::6(3,4)-5(2,3),v3=1:-5.47380::1066.87810:JPL','O3:Ozone:::611.27664:0.05:17(2,16)-16(1,15):-3.17260::120.25710:JPL','O3v2=1:Ozone:::612.36507:11.9917:17(2,16)-16(1,15):-4.64240::820.81590:JPL','O3v1=1,v3=1:Ozone:613.02471:0.0398:::25(1,25)-24(0,24),v1=1:-5.22180::1348.99580:JPL','O3v1=1,v3=1:Ozone:613.34761:0.1498:::32(1,31)-32(0,32),v1=1:-5.99980::1532.44530:JPL','O3v2=2:Ozone:613.88034:0.1158:::26(0,26)-25(1,25):-5.84850::1662.45760:JPL','O3v1=1,v3=1:Ozone:614.12519:0.0405:::23(4,19)-23(3,20),v3=1:-5.32030::1300.19940:JPL','O3:Ozone:::614.72953:0.05:30(8,22)-31(7,25):-4.27700::569.90980:JPL','O3:Ozone:::614.78419:0.05:29(2,28)-29(1,29):-3.56850::354.14660:JPL','O3:Ozone:::615.65404:0.08:50(5,45)-50(4,46):-4.53170::1134.75820:JPL','O3v2=1:Ozone:616.15346:0.0554:::31(8,24)-32(7,25):-5.78440::1298.51990:JPL','O3:Ozone:::616.27895:0.1:46(3,43)-46(2,44):-4.24910::931.41540:JPL','O3:Ozone:::616.28239:0.05:8(5,3)-9(4,6):-4.15260::87.90200:JPL','O3:Ozone:::617.30432:0.05:38(2,36)-38(1,37):-3.80880::625.58270:JPL','O3:Ozone:617.38965:0.0152:::41(5,37)-42(2,40):-5.72560::782.64490:JPL','O3v2=1:Ozone:617.39233:0.027:::26(0,26)-25(1,25):-4.40460::965.46690:JPL','O3v1=1,v3=1:Ozone:618.21000:0.0683:::17(2,16)-16(1,15),v1=1:-5.46780::1223.05940:JPL','O3v1=1,v3=1:Ozone:618.84737:0.0448:::26(4,22)-26(3,23),v1=1:-5.52200::1427.69740:JPL','O3v2=1:Ozone:619.55298:0.0496:::16(6,10)-17(5,13):-5.33350::908.57280:JPL','O3v2=1:Ozone:::620.05475:11.9917:26(4,22)-26(3,23):-4.68930::1025.58770:JPL','O3:Ozone:::620.10510:0.05:56(5,51)-55(6,50):-5.68280::1408.79080:JPL','O3v1=1,v3=1:Ozone:620.12625:0.0855:::37(2,35)-37(1,36),v3=1:-5.90920::1627.73050:JPL','O3:Ozone:::620.68670:0.05:6(3,3)-5(2,4):-3.29010::25.12970:JPL','O3:Ozone:::620.71704:0.05:37(9,29)-38(8,30):-4.69430::822.28580:JPL','O3:Ozone:::620.82569:0.05:26(0,26)-25(1,25):-2.93040::265.85790:JPL','O3v1=1,v3=1:Ozone:620.82655:0.0378:::26(0,26)-25(1,25),v1=1:-5.23330::1369.44410:JPL','O3v1=1,v3=1:Ozone:620.90575:0.0411:::26(1,26)-25(0,25),v3=1:-5.08690::1303.32680:JPL','O3v1=1,v3=1:Ozone:621.00255:0.0833:::6(3,3)-5(2,4),v1=1:-5.58670::1128.26570:JPL','O3v2=1:Ozone:::621.67962:11.9917:31(3,29)-31(2,30):-4.90730::1124.23960:JPL','O3v1=1,v3=1:Ozone:622.17132:0.1053:::33(3,31)-33(2,32),v1=1:-5.82610::1581.47410:JPL','O3v1=1,v3=1:Ozone:622.94659:0.0423:::28(1,27)-27(2,26),v1=1:-5.55400::1430.55090:JPL','O3:Ozone:::623.68773:0.05:24(4,20)-24(3,21):-3.16750::281.83290:JPL','O3v1=1,v3=1:Ozone:625.05127:0.0419:::21(4,17)-21(3,18),v3=1:-5.27480::1262.36960:JPL','O3v2=1:Ozone:625.28077:0.0499:::40(3,37)-39(4,36):-5.70100::1406.23280:JPL','O3v2=1:Ozone:625.32402:0.1044:::50(5,45)-50(4,46):-5.98340::1832.39760:JPL','O3:Ozone:::625.37111:0.12:15(6,10)-16(5,11):-3.87340::192.52350:JPL','O3:Ozone:::625.89487:0.05:44(10,34)-45(9,37):-5.23340::1121.68850:JPL','O3:Ozone:::626.39066:0.2:62(5,57)-62(4,58):-5.61120::1714.38070:JPL','O3v2=1:Ozone:626.65482:0.0851:::29(2,28)-29(1,29):-5.02160::1053.27650:JPL','O3v2=1:Ozone:::628.24507:11.9917:28(1,27)-27(2,26):-4.70030::1027.30440:JPL','O3v2=1:Ozone:628.28512:0.0473:::6(3,3)-5(2,4):-4.75250::726.22760:JPL','O3:Ozone:628.56547:0.029:::39(3,37)-40(0,40):-5.84750::661.83320:JPL','O3v2=1:Ozone:629.41202:0.0549:::46(3,43)-46(2,44):-5.69770::1628.91480:JPL','O3v2=1:Ozone:630.09556:0.082:::38(2,36)-38(1,37):-5.25930::1323.89300:JPL','O3:Ozone:::630.31748:0.05:51(11,41)-52(10,42):-5.88670::1467.93160:JPL','O3v1=1,v3=1:Ozone:630.65976:0.043:::24(4,20)-24(3,21),v1=1:-5.45900::1384.35900:JPL','O3:Ozone:630.67635:0.0069:::25(4,22)-26(1,25):-5.20070::302.58360:JPL','O3v2=1:Ozone:631.17057:0.0575:::23(7,17)-24(6,18):-5.43870::1066.53890:JPL','O3v2=1:Ozone:631.90629:0.0457:::8(5,3)-9(4,6):-5.60540::789.55050:JPL','O3:Ozone:::632.17694:0.05:28(1,27)-27(2,26):-3.22570::327.55790:JPL','O3v1=1,v3=1:Ozone:632.83384:0.1305:::31(2,30)-31(1,31),v1=1:-5.92930::1506.85780:JPL','O3v1=1,v3=1:Ozone:632.94983:0.0454:::19(4,15)-19(3,16),v3=1:-5.24390::1228.00270:JPL','O3:Ozone:::633.35269:0.05:22(7,15)-23(6,18):-3.95070::344.53270:JPL','O3v1=1,v3=1:Ozone:633.85589:0.0779:::32(3,30)-32(2,31),v3=1:-5.63670::1486.74170:JPL','O3v1=1,v3=1:Ozone:634.02083:0.0501:::20(2,19)-19(1,18),v3=1:-5.28900::1206.43400:JPL','O3v2=1:Ozone:634.33068:0.0291:::24(4,20)-24(3,21):-4.62430::982.23070:JPL','O3:Ozone:::634.46160:0.05:22(4,18)-22(3,19):-3.11580::241.83170:JPL','O3:Ozone:634.52017:0.0076:::40(3,37)-39(4,36):-4.22070::707.15040:JPL','O3v1=1,v3=1:Ozone:635.31664:0.0412:::27(0,27)-26(1,26),v3=1:-5.09120::1324.03800:JPL','O3v1=1,v3=1:Ozone:636.67242:0.072:::7(3,4)-6(2,5),v3=1:-5.42260::1071.82200:JPL','O3v1=1,v3=1:Ozone:638.36264:0.0504:::17(4,13)-17(3,14),v3=1:-5.22820::1197.06970:JPL','O3:Ozone:::638.53482:0.15:48(5,43)-48(4,44):-4.35880::1049.70940:JPL','O3:Ozone:::638.90310:0.05:19(2,18)-18(1,17):-3.12810::150.19000:JPL','O3v2=1:Ozone:::639.33140:11.9917:19(2,18)-18(1,17):-4.59850::850.63390:JPL','O3:Ozone:::639.54849:0.05:48(4,44)-47(5,43):-4.88560::1028.37630:JPL','O3v1=1,v3=1:Ozone:639.61073:0.044:::22(4,18)-22(3,19),v1=1:-5.40920::1344.51040:JPL','O3:Ozone:::640.09242:0.1:33(3,31)-33(2,32):-3.51360::478.43900:JPL','O3:Ozone:::640.30980:0.05:29(8,22)-30(7,23):-4.21040::543.85120:JPL','O3v1=1,v3=1:Ozone:641.22333:0.2178:::31(1,30)-31(0,31),v3=1:-5.79830::1438.29370:JPL','O3:Ozone:::641.53413:0.05:7(5,3)-8(4,4):-4.27290::80.33660:JPL','O3v2=1:Ozone:641.68950:0.0585:::30(8,22)-31(7,25):-5.71520::1271.73300:JPL','O3v1=1,v3=1:Ozone:641.88375:0.0568:::15(4,11)-15(3,12),v3=1:-5.22860::1169.54220:JPL','O3:Ozone:::642.34970:0.05:20(4,16)-20(3,17):-3.07810::205.32850:JPL','O3v1=1,v3=1:Ozone:643.58526:0.1209:::33(2,31)-32(3,30),v3=1:-5.79220::1507.88490:JPL','O3v1=1,v3=1:Ozone:644.05532:0.0644:::13(4,9)-13(3,10),v3=1:-5.24690::1145.39520:JPL','O3v1=1,v3=1:Ozone:644.47516:0.0483:::18(4,15)-18(3,16),v3=1:-5.22450::1211.81150:JPL','O3v1=1,v3=1:Ozone:644.65285:0.0446:::20(4,17)-20(3,18),v3=1:-5.24050::1244.21670:JPL','O3v1=1,v3=1:Ozone:644.72194:0.0537:::16(4,13)-16(3,14),v3=1:-5.22130::1182.73040:JPL','O3:Ozone:::644.80493:0.05:7(3,5)-6(2,4):-3.24050::30.20950:JPL','O3v2=1:Ozone:644.84545:0.0518:::15(6,10)-16(5,11):-5.32170::894.33730:JPL','O3v1=1,v3=1:Ozone:645.09532:0.0804:::12(2,10)-11(1,11),v1=1:-5.71870::1160.01600:JPL','O3v1=1,v3=1:Ozone:645.15839:0.0606:::14(4,11)-14(3,12),v3=1:-5.23300::1156.97740:JPL','O3v1=1,v3=1:Ozone:645.26290:0.0847:::7(3,5)-6(2,4),v1=1:-5.53680::1133.33320:JPL','O3v1=1,v3=1:Ozone:645.31862:0.0727:::11(4,7)-11(3,8),v3=1:-5.28690::1124.60860:JPL','O3v2=1:Ozone:645.37252:0.0281:::22(4,18)-22(3,19):-4.57290::942.38360:JPL','O3v1=1,v3=1:Ozone:645.55536:0.0422:::22(4,19)-22(3,20),v3=1:-5.26760::1279.93890:JPL','O3v1=1,v3=1:Ozone:645.62005:0.0685:::12(4,9)-12(3,10),v3=1:-5.26290::1134.55430:JPL','O3:Ozone:::645.67019:0.05:32(1,31)-32(0,32):-3.67380::428.58040:JPL','O3v1=1,v3=1:Ozone:646.00455:0.081:::9(4,5)-9(3,6),v3=1:-5.35640::1107.16700:JPL','O3v1=1,v3=1:Ozone:646.00954:0.0769:::10(4,7)-10(3,8),v3=1:-5.31690::1115.46060:JPL','O3v1=1,v3=1:Ozone:646.06937:0.0465:::20(4,16)-20(3,17),v1=1:-5.37310::1308.13040:JPL','O3v1=1,v3=1:Ozone:646.28633:0.085:::8(4,5)-8(3,6),v3=1:-5.40710::1099.69420:JPL','O3:Ozone:::646.34400:0.05:36(9,27)-37(8,30):-4.61040::790.36360:JPL','O3v1=1,v3=1:Ozone:646.34549:0.0887:::7(4,3)-7(3,4),v3=1:-5.47430::1093.05910:JPL','O3v1=1,v3=1:Ozone:646.45170:0.0921:::6(4,3)-6(3,4),v3=1:-5.56660::1087.25230:JPL','O3v1=1,v3=1:Ozone:646.49539:0.0952:::5(4,1)-5(3,2),v3=1:-5.70540::1082.27710:JPL','O3v1=1,v3=1:Ozone:646.53168:0.0978:::4(4,1)-4(3,2),v3=1:-5.96070::1078.13140:JPL','O3v1=1,v3=1:Ozone:647.06512:0.0655:::19(2,18)-18(1,17),v1=1:-5.42400::1252.93490:JPL','O3v1=1,v3=1:Ozone:647.53614:0.0412:::24(4,21)-24(3,22),v3=1:-5.30470::1318.96850:JPL','O3:Ozone:::647.83992:0.15:18(4,14)-18(3,15):-3.05500::172.29740:JPL','O3v2=2:Ozone:648.27073:0.1438:::27(1,27)-26(0,26):-5.82450::1682.93450:JPL','O3v2=1:Ozone:648.48970:0.0857:::48(5,43)-48(4,44):-5.81120::1747.65320:JPL','O3v1=1,v3=1:Ozone:648.89101:0.1238:::34(2,32)-33(3,31),v1=1:-5.99930::1602.22750:JPL','O3v1=1,v3=1:Ozone:650.51089:0.0507:::18(4,14)-18(3,15),v1=1:-5.35100::1275.19610:JPL','O3:Ozone:::650.73273:0.05:14(6,8)-15(5,11):-3.86980::179.06980:JPL','O3v1=1,v3=1:Ozone:650.97834:0.18:::30(2,29)-30(1,30),v3=1:-5.74020::1413.93430:JPL','O3v1=1,v3=1:Ozone:650.98303:0.0422:::26(4,23)-26(3,24),v3=1:-5.35090::1361.29330:JPL','O3:Ozone:::651.47584:0.05:16(4,12)-16(3,13):-3.04700::142.71280:JPL','O3:Ozone:::651.53256:0.05:43(10,34)-44(9,35):-5.13420::1083.92060:JPL','O3:Ozone:::651.55697:0.05:12(2,10)-11(1,11):-3.43200::56.80500:JPL','O3v2=1:Ozone:651.81989:0.0299:::27(1,27)-26(0,26):-4.38100::986.06090:JPL','O3v1=1,v3=1:Ozone:652.06684:0.136:::35(3,33)-35(2,34),v1=1:-5.89480::1638.16340:JPL','O3v2=1:Ozone:652.18823:0.04:::33(3,31)-33(2,32):-4.96660::1177.47090:JPL','O3v2=1:Ozone:652.27063:0.0482:::7(3,5)-6(2,4):-4.70350::731.28690:JPL','O3v1=1,v3=1:Ozone:653.41804:0.0567:::16(4,12)-16(3,13),v1=1:-5.34370::1245.68630:JPL','O3v2=1:Ozone:653.46027:0.0283:::20(4,16)-20(3,17):-4.53540::906.02380:JPL','O3:Ozone:::653.76245:0.05:14(4,10)-14(3,11):-3.05570::116.55200:JPL','O3:Ozone:::654.71349:0.05:19(4,16)-19(3,17):-3.05440::188.05470:JPL','O3:Ozone:::654.85304:0.05:17(4,14)-17(3,15):-3.04350::156.90330:JPL','O3:Ozone:::655.00526:0.05:21(4,18)-21(3,19):-3.07730::222.57160:JPL','O3:Ozone:::655.12033:0.05:12(4,8)-12(3,9):-3.08370::93.79630:JPL','O3:Ozone:::655.20358:0.05:15(4,12)-15(3,13):-3.04640::129.12200:JPL','O3v1=1,v3=1:Ozone:655.22016:0.0642:::14(4,10)-14(3,11),v1=1:-5.35280::1219.58240:JPL','O3:Ozone:::655.29020:0.05:27(1,27)-26(0,26):-2.90730::286.56650:JPL','O3:Ozone:::655.60631:0.05:13(4,10)-13(3,11):-3.06590::104.71310:JPL','O3:Ozone:::655.87259:0.05:10(4,6)-10(3,7):-3.13660::74.43140:JPL','O3v1=1,v3=1:Ozone:655.92346:0.048:::19(4,16)-19(3,17),v1=1:-5.35200::1290.97670:JPL','O3:Ozone:::655.93491:0.1:50(11,39)-51(10,42):-5.77320::1424.34030:JPL','O3:Ozone:::655.96210:0.05:11(4,8)-11(3,9):-3.10610::83.67700:JPL','O3:Ozone:::656.00573:0.05:23(4,20)-23(3,21):-3.11110::260.44680:JPL','O3v1=1,v3=1:Ozone:656.06907:0.0533:::17(4,14)-17(3,15),v1=1:-5.34110::1259.87880:JPL','O3v1=1,v3=1:Ozone:656.09229:0.0444:::21(4,18)-21(3,19),v1=1:-5.37520::1325.43380:JPL','O3:Ozone:::656.22376:0.05:9(4,6)-9(3,7):-3.17540::66.01270:JPL','O3:Ozone:::656.25256:0.05:8(4,4)-8(3,5):-3.22610::58.44640:JPL','O3v1=1,v3=1:Ozone:656.26656:0.0728:::12(4,8)-12(3,9),v1=1:-5.38100::1196.86980:JPL','O3v1=1,v3=1:Ozone:656.29569:0.0465:::28(4,25)-28(3,26),v3=1:-5.40540::1406.89880:JPL','O3v1=1,v3=1:Ozone:656.35473:0.0602:::15(4,12)-15(3,13),v1=1:-5.34390::1232.14400:JPL','O3:Ozone:::656.38400:0.05:7(4,4)-7(3,5):-3.29290::51.71790:JPL','O3:Ozone:::656.41975:0.05:6(4,2)-6(3,3):-3.38510::45.83360:JPL','O3:Ozone:::656.46154:0.05:5(4,2)-5(3,3):-3.52360::40.79000:JPL','O3:Ozone:::656.47770:0.05:4(4,0)-4(3,1):-3.77880::36.58780:JPL','O3v2=1:Ozone:656.57951:0.0602:::22(7,15)-23(6,18):-5.39480::1046.44230:JPL','O3v2=2:Ozone:656.65124:0.1069:::22(4,18)-22(3,19):-5.99930::1640.34650:JPL','O3v1=1,v3=1:Ozone:656.65915:0.0684:::13(4,10)-13(3,11),v1=1:-5.36330::1207.77450:JPL','O3v2=1:Ozone:656.68363:0.0449:::12(2,10)-11(1,11):-4.89810::757.51840:JPL','O3v1=1,v3=1:Ozone:656.80171:0.0425:::23(4,20)-23(3,21),v1=1:-5.40920::1363.24380:JPL','O3v1=1,v3=1:Ozone:656.82320:0.0817:::10(4,6)-10(3,7),v1=1:-5.43390::1177.53700:JPL','O3v1=1,v3=1:Ozone:656.91055:0.0772:::11(4,8)-11(3,9),v1=1:-5.40350::1186.77080:JPL','O3v2=1:Ozone:657.05315:0.0478:::7(5,3)-8(4,4):-5.72670::782.01720:JPL','O3v1=1,v3=1:Ozone:657.07889:0.0861:::9(4,6)-9(3,7),v1=1:-5.47280::1169.13210:JPL','O3v1=1,v3=1:Ozone:657.08198:0.0903:::8(4,4)-8(3,5),v1=1:-5.52340::1161.57580:JPL','O3v1=1,v3=1:Ozone:657.16533:0.0943:::7(4,4)-7(3,5),v1=1:-5.59020::1154.85680:JPL','O3v1=1,v3=1:Ozone:657.17479:0.0979:::6(4,2)-6(3,3),v1=1:-5.68240::1148.98020:JPL','O3v1=1,v3=1:Ozone:657.19038:0.1011:::5(4,2)-5(3,3),v1=1:-5.82090::1143.94300:JPL','O3:Ozone:::657.51512:0.05:31(2,30)-31(1,31):-3.61070::403.01940:JPL','O3v1=1,v3=1:Ozone:657.68897:0.0413:::27(1,27)-26(0,26),v1=1:-5.20670::1390.15260:JPL','O3:Ozone:::658.03864:0.05:25(4,22)-25(3,23):-3.15460::301.67090:JPL','O3v1=1,v3=1:Ozone:658.32315:0.0427:::25(4,22)-25(3,23),v1=1:-5.45310::1404.39850:JPL','O3v2=1:Ozone:658.53857:0.1323:::32(1,31)-32(0,32):-5.12560::1127.30220:JPL','O3:Ozone:::658.81498:0.05:21(7,15)-22(6,16):-3.91090::325.19330:JPL','O3v2=1:Ozone:::659.06374:11.9917:18(4,14)-18(3,15):-4.51250::873.12480:JPL','O3v1=1,v3=1:Ozone:659.21266:0.0721:::8(3,6)-7(2,5),v3=1:-5.37830::1077.70530:JPL','O3v1=1,v3=1:Ozone:659.46509:0.0478:::22(2,21)-21(1,20),v3=1:-5.25750::1240.82780:JPL','O3:Ozone:660.35245:0.0207:::64(5,59)-64(4,60):-5.79300::1822.02480:JPL','O3v1=1,v3=1:Ozone:660.96244:0.046:::27(4,24)-27(3,25),v1=1:-5.50610::1448.88750:JPL','O3:Ozone:661.45830:0.0066:::27(4,24)-27(3,25):-3.20710::346.23200:JPL','O3v1=1,v3=1:Ozone:661.58537:0.0524:::29(1,28)-28(2,27),v3=1:-5.34860::1387.69620:JPL','O3:Ozone:661.64692:0.015:::39(5,35)-40(2,38):-5.62220::712.86350:JPL','O3:Ozone:662.14814:0.0102:::56(4,52)-56(3,53):-5.00050::1387.96060:JPL','O3v2=2:Ozone:662.39163:0.148:::28(0,28)-27(1,27):-5.83270::1704.55840:JPL','O3v2=1:Ozone:::662.81115:11.9917:16(4,12)-16(3,13):-4.50480::843.66040:JPL','O3v1=1,v3=1:Ozone:663.86181:0.0559:::30(4,27)-30(3,28),v3=1:-5.46760::1455.76860:JPL','O3:Ozone:664.48665:0.0142:::46(5,41)-46(4,42):-4.19120::968.10160:JPL','O3v2=2:Ozone:664.94686:0.1065:::20(4,16)-20(3,17):-5.96210::1604.13280:JPL','O3v1=1,v3=1:Ozone:665.04547:0.0535:::29(4,26)-29(3,27),v1=1:-5.56770::1496.69820:JPL','O3v2=1:Ozone:::665.13753:11.9917:14(4,10)-14(3,11):-4.51370::817.60720:JPL','O3v2=1:Ozone:665.44299:0.0363:::21(2,20)-20(1,19):-4.56490::883.74400:JPL','O3:Ozone:665.67771:0.0058:::21(2,20)-20(1,19):-3.09390::183.43070:JPL','O3v1=1,v3=1:Ozone:665.70236:0.0457:::28(1,28)-27(0,27),v3=1:-5.07670::1345.22980:JPL','O3:Ozone:665.85207:0.0126:::28(8,20)-29(7,23):-4.14820::518.63370:JPL','O3v2=1:Ozone:665.85236:0.0322:::34(2,32)-33(3,31):-5.12420::1199.22560:JPL','O3v2=1:Ozone:666.14921:0.0286:::19(4,16)-19(3,17):-4.51170::888.80890:JPL','O3v2=1:Ozone:666.16672:0.0295:::28(0,28)-27(1,27):-4.38930::1007.80330:JPL','O3v2=1:Ozone:666.28914:0.0307:::17(4,14)-17(3,15):-4.50110::857.78780:JPL','O3v2=1:Ozone:666.45312:0.0274:::21(4,18)-21(3,19):-4.53430::923.18140:JPL','O3v2=1:Ozone:666.55949:0.038:::12(4,8)-12(3,9):-4.54190::794.94610:JPL','O3:Ozone:666.63090:0.0074:::29(4,26)-29(3,27):-3.26790::394.11620:JPL','O3v2=1:Ozone:666.64674:0.0334:::15(4,12)-15(3,13):-4.50430::830.12270:JPL','O3:Ozone:666.76773:0.0077:::6(5,1)-7(4,4):-4.46870::73.61250:JPL','O3v2=1:Ozone:::667.05021:11.9917:13(4,10)-13(3,11):-4.52390::805.81620:JPL','O3v2=1:Ozone:667.17905:0.0618:::29(8,22)-30(7,23):-5.65040::1245.78400:JPL','O3v2=1:Ozone:667.33158:0.041:::10(4,6)-10(3,7):-4.59490::775.66210:JPL','O3v2=1:Ozone:667.42359:0.0395:::11(4,8)-11(3,9):-4.56430::784.86860:JPL','O3v2=1:Ozone:667.48536:0.0272:::23(4,20)-23(3,21):-4.56770::960.89800:JPL','O3v1=1,v3=1:Ozone:667.49631:0.1003:::34(3,32)-34(2,33),v3=1:-5.69740::1540.96370:JPL','O3v2=1:Ozone:667.69218:0.0424:::9(4,6)-9(3,7):-4.63380::767.27870:JPL','O3v2=1:Ozone:667.72185:0.0437:::8(4,4)-8(3,5):-4.68460::759.74440:JPL','O3v2=1:Ozone:667.85700:0.045:::7(4,4)-7(3,5):-4.75140::753.04430:JPL','O3v2=1:Ozone:667.89391:0.0461:::6(4,2)-6(3,3):-4.84370::747.18500:JPL','O3v2=1:Ozone:667.93701:0.047:::5(4,2)-5(3,3):-4.98220::742.16270:JPL','O3v2=1:Ozone:667.95381:0.0478:::4(4,0)-4(3,1):-5.23740::737.97840:JPL','O3:Ozone:668.06608:0.0367:::41(3,39)-42(0,42):-5.95520::727.97910:JPL','O3v2=1:Ozone:669.57772:0.0279:::25(4,22)-25(3,23):-4.61090::1001.94900:JPL','O3:Ozone:669.85713:0.006:::28(0,28)-27(1,27):-2.91550::308.42460:JPL','O3v2=1:Ozone:670.10601:0.054:::14(6,8)-15(5,11):-5.31940::880.94050:JPL','O3v1=1,v3=1:Ozone:670.20402:0.0412:::28(0,28)-27(1,27),v1=1:-5.21770::1412.09080:JPL','O3v2=1:Ozone:670.26093:0.1124:::31(2,30)-31(1,31):-5.06320::1101.88210:JPL','O3v2=2:Ozone:670.72402:0.1096:::18(4,14)-18(3,15):-5.93950::1571.36840:JPL','O3v1=1,v3=1:Ozone:670.90207:0.0659:::31(4,28)-31(3,29),v1=1:-5.63710::1547.81680:JPL','O3:Ozone:671.66922:0.0058:::8(3,5)-7(2,6):-3.19310::36.04190:JPL','O3:Ozone:671.80443:0.0069:::34(2,32)-33(3,31):-3.64830::499.79020:JPL','O3v1=1,v3=1:Ozone:671.85034:0.086:::8(3,5)-7(2,6),v1=1:-5.48930::1139.16530:JPL','O3:Ozone:671.93232:0.0129:::35(9,27)-36(8,28):-4.53050::759.28080:JPL','O3:Ozone:672.79452:0.0092:::35(3,33)-35(2,34):-3.58110::535.10310:JPL','O3v2=1:Ozone:673.09312:0.0292:::27(4,24)-27(3,25):-4.66300::1046.32230:JPL','O3:Ozone:673.91436:0.0082:::31(4,28)-31(3,29):-3.33640::445.30790:JPL','O3v1=1,v3=1:Ozone:674.03442:0.0715:::32(4,29)-32(3,30),v3=1:-5.53710::1507.88490:JPL','O3v2=2:Ozone:674.55106:0.1169:::16(4,12)-16(3,13):-5.93200::1542.02640:JPL','O3v2=1:Ozone:674.86943:0.0708:::46(5,41)-46(4,42):-5.64430::1666.33660:JPL','O3v1=1,v3=1:Ozone:674.87378:0.0618:::21(2,20)-20(1,19),v1=1:-5.39100::1286.12800:JPL','O3:Ozone:676.06153:0.0104:::13(6,8)-14(5,9):-3.87720::166.45820:JPL','O3:Ozone:676.90246:0.0103:::48(3,45)-48(2,46):-4.34470::1010.39300:JPL','O3v2=2:Ozone:676.95857:0.1278:::14(4,10)-14(3,11):-5.94110::1516.08280:JPL','O3:Ozone:677.13592:0.0143:::42(10,32)-43(9,35):-5.03870::1046.98970:JPL','O3:Ozone:677.14321:0.0094:::40(2,38)-40(1,39):-3.87080::690.27640:JPL','O3v2=2:Ozone:677.97432:0.1063:::19(4,16)-19(3,17):-5.93840::1586.97760:JPL','O3v2=2:Ozone:678.11482:0.1123:::17(4,14)-17(3,15):-5.92810::1556.08930:JPL','O3v2=2:Ozone:678.29069:0.1044:::21(4,18)-21(3,19):-5.96080::1621.20290:JPL','O3v2=2:Ozone:678.38875:0.1409:::12(4,8)-12(3,9):-5.96950::1493.51800:JPL','O3v2=1:Ozone:::678.40335:11.9917:29(4,26)-29(3,27):-4.72330::1094.00360:JPL','O3v2=2:Ozone:678.47993:0.1219:::15(4,12)-15(3,13):-5.93160::1528.54280:JPL','O3v1=1,v3=1:Ozone:678.84933:0.0833:::33(4,30)-33(3,31),v1=1:-5.71410::1602.22750:JPL','O3v2=2:Ozone:678.90203:0.1343:::13(4,10)-13(3,11):-5.95140::1504.34070:JPL','O3v2=1:Ozone:::679.04491:11.9917:8(3,5)-7(2,6):-4.65650::737.09310:JPL','O3v2=2:Ozone:679.27617:0.1479:::11(4,8)-11(3,9):-5.99210::1483.48320:JPL','O3:Ozone:679.30443:0.0065:::23(4,20)-24(1,23):-5.22900::259.66970:JPL','O3v2=2:Ozone:679.35584:0.1059:::23(4,20)-23(3,21):-5.99380::1658.75780:JPL','O3v1=1,v3=1:Ozone:680.78639:0.1313:::39(2,37)-39(1,38),v3=1:-5.96450::1689.85680:JPL','O3:Ozone:681.52145:0.0175:::49(11,39)-50(10,40):-5.66330::1381.58300:JPL','O3v2=1:Ozone:681.94518:0.063:::21(7,15)-22(6,16):-5.35650::1027.18430:JPL','O3v2=1:Ozone:682.18138:0.0499:::6(5,1)-7(4,4):-5.92340::775.32160:JPL','O3v1=1,v3=1:Ozone:683.01194:0.0468:::29(0,29)-28(1,28),v3=1:-5.08350::1367.43530:JPL','O3:Ozone:683.63901:0.0091:::33(4,30)-33(3,31):-3.41230::499.79020:JPL','O3:Ozone:684.24310:0.0127:::20(7,13)-21(6,16):-3.87690::306.69580:JPL','O3v1=1,v3=1:Ozone:684.60703:0.1746:::37(3,35)-37(2,36),v1=1:-5.97100::1698.02270:JPL','O3v2=1:Ozone:685.60789:0.0507:::35(3,33)-35(2,34):-5.03350::1233.86120:JPL','O3v1=1,v3=1:Ozone:685.64680:0.0465:::24(2,23)-23(1,22),v3=1:-5.23530::1278.39850:JPL','O3v2=1:Ozone:685.88298:0.0324:::31(4,28)-31(3,29):-4.79140::1144.97720:JPL','O3v1=1,v3=1:Ozone:686.63137:0.0495:::30(1,29)-29(2,28),v1=1:-5.50880::1477.65450:JPL','O3v1=1,v3=1:Ozone:687.11221:0.0933:::34(4,31)-34(3,32),v3=1:-5.61350::1563.22900:JPL','O3v1=1,v3=1:Ozone:687.42232:0.0719:::9(3,6)-8(2,7),v3=1:-5.33450::1084.23710:JPL','O3v1=1,v3=1:Ozone:689.17594:0.106:::35(4,32)-35(3,33),v1=1:-5.79810::1659.91400:JPL','O3v2=2:Ozone:690.57130:0.1148:::23(2,22)-22(1,21):-5.98180::1617.66780:JPL','O3v1=1,v3=1:Ozone:690.59912:0.0607:::13(2,11)-12(1,12),v3=1:-5.61220::1107.54560:JPL','O3v2=1:Ozone:::690.76080:11.9917:30(1,29)-29(2,28):-4.65800::1074.17950:JPL','O3v2=1:Ozone:691.10645:0.1097:::40(2,38)-40(1,39):-5.32070::1388.26830:JPL','O3v2=1:Ozone:691.34934:0.0732:::48(3,45)-48(2,46):-5.79250::1707.53330:JPL','O3:Ozone:691.35650:0.0137:::27(8,20)-28(7,21):-4.09050::494.25730:JPL','O3v2=1:Ozone:691.46964:0.035:::23(2,22)-22(1,21):-4.54080::920.10420:JPL','O3:Ozone:691.66426:0.0141:::44(5,39)-44(4,40):-4.03040::889.97730:JPL','O3:Ozone:691.98594:0.0079:::5(5,1)-6(4,2):-4.83480::67.72940:JPL','O3:Ozone:692.36859:0.0062:::23(2,22)-22(1,21):-3.06950::219.93830:JPL','O3v2=1:Ozone:692.62939:0.0651:::28(8,20)-29(7,23):-5.58980::1220.67260:JPL','O3v2=2:Ozone:693.23317:0.1824:::29(1,29)-28(0,28):-5.82180::1726.65340:JPL','O3:Ozone:693.47371:0.0059:::9(3,7)-8(2,6):-3.15390::42.88090:JPL','O3v1=1,v3=1:Ozone:694.14897:0.0867:::9(3,7)-8(2,6),v1=1:-5.44940::1145.97780:JPL','O3:Ozone:694.76095:0.0064:::30(1,29)-29(2,28):-3.18420::374.65350:JPL','O3v1=1,v3=1:Ozone:695.23253:0.2862:::33(1,32)-33(0,33),v3=1:-5.83670::1489.45790:JPL','O3v2=1:Ozone:695.33292:0.0562:::13(6,8)-14(5,9):-5.32800::868.38220:JPL','O3v1=1,v3=1:Ozone:695.80848:0.2395:::32(2,31)-32(1,32),v3=1:-5.78520::1463.53210:JPL','O3v2=1:Ozone:695.86064:0.0337:::33(4,30)-33(3,31):-4.86670::1199.22560:JPL','O3:Ozone:696.09039:0.0101:::35(4,32)-35(3,33):-3.49510::557.54510:JPL','O3v2=1:Ozone:::697.08342:11.9917:29(1,29)-28(0,28):-4.37870::1030.02420:JPL','O3:Ozone:697.48560:0.014:::34(9,25)-35(8,28):-4.45450::729.03740:JPL','O3v1=1,v3=1:Ozone:698.91915:0.3144:::41(5,36)-41(4,37),v3=1:-5.98330::1812.39750:JPL','O3:Ozone:698.98063:0.0131:::34(1,33)-34(0,34):-3.71890::482.19100:JPL','O3:Ozone:699.36779:0.0493:::66(6,60)-66(5,61):-5.99860::1956.16140:JPL','O3v2=1:Ozone:700.68444:0.0496:::9(3,7)-8(2,6):-4.61780::743.90640:JPL','O3:Ozone:700.85303:0.0067:::29(1,29)-28(0,28):-2.90540::330.76860:JPL','O3:Ozone:701.36450:0.0108:::12(6,6)-13(5,9):-3.89770::154.68840:JPL','O3:Ozone:701.50823:0.0123:::33(2,32)-33(1,33):-3.66260::455.03920:JPL','O3v1=1,v3=1:Ozone:702.12831:0.1346:::37(4,34)-37(3,35),v1=1:-5.88900::1720.85870:JPL','O3v2=1:Ozone:702.23500:0.0642:::35(9,27)-36(8,28):-5.96830::1461.29130:JPL','O3v1=1,v3=1:Ozone:702.28158:0.0577:::23(2,22)-22(1,21),v1=1:-5.36790::1322.60270:JPL','O3v2=1:Ozone:702.57238:0.0599:::44(5,39)-44(4,40):-5.48410::1588.49150:JPL','O3:Ozone:702.70607:0.0151:::41(10,32)-42(9,33):-4.94680::1010.89600:JPL','O3v1=1,v3=1:Ozone:703.03396:0.0468:::29(1,29)-28(0,28),v1=1:-5.20510::1434.44640:JPL','O3v1=1,v3=1:Ozone:703.32399:0.1212:::36(4,33)-36(3,34),v3=1:-5.69640::1621.78190:JPL','O3:Ozone:703.49011:0.0295:::66(5,61)-66(4,62):-5.97650::1932.69550:JPL','O3v1=1,v3=1:Ozone:703.84341:0.1293:::36(3,34)-36(2,35),v3=1:-5.76580::1598.30420:JPL','O3v1=1,v3=1:Ozone:706.01042:0.0718:::10(3,8)-9(2,7),v3=1:-5.30180::1091.91060:JPL','O3:Ozone:707.07847:0.018:::48(11,37)-49(10,40):-5.55680::1339.65990:JPL','O3v2=1:Ozone:707.27568:0.0657:::20(7,13)-21(6,16):-5.32390::1008.76480:JPL','O3:Ozone:708.14331:0.0103:::37(3,35)-37(2,36):-3.65630::594.93340:JPL','O3v2=1:Ozone:708.63029:0.0348:::35(4,32)-35(3,33):-4.94890::1256.73060:JPL','O3:Ozone:709.63774:0.0134:::19(7,13)-20(6,14):-3.84920::289.04010:JPL','O3v2=2:Ozone:710.29417:0.1912:::30(0,30)-29(1,29):-5.83220::1749.77720:JPL','O3v1=1,v3=1:Ozone:710.99353:0.0518:::30(1,30)-29(0,29),v3=1:-5.07960::1390.21810:JPL','O3:Ozone:711.49517:0.011:::37(4,34)-37(3,35):-3.58460::618.55450:JPL','O3v2=1:Ozone:712.86823:0.173:::34(1,33)-34(0,34):-5.17010::1180.61770:JPL','O3v1=1,v3=1:Ozone:713.41000:0.0467:::26(2,25)-25(1,24),v3=1:-5.22240::1319.10140:JPL','O3v2=1:Ozone:::714.33947:11.9917:30(0,30)-29(1,29):-4.38940::1053.27650:JPL','O3:Ozone:714.56736:0.0147:::37(5,33)-38(2,36):-5.53910::646.17380:JPL','O3v2=1:Ozone:715.15238:0.1475:::33(2,32)-33(1,33):-5.11440::1153.61600:JPL','O3:Ozone:716.53356:0.0431:::64(6,58)-64(5,59):-5.76990::1844.05180:JPL','O3v2=2:Ozone:716.70736:0.1181:::25(2,24)-24(1,23):-5.96730::1657.06580:JPL','O3:Ozone:716.82732:0.0148:::26(8,18)-27(7,21):-4.03730::470.72200:JPL','O3v1=1,v3=1:Ozone:717.89938:0.1711:::39(4,36)-39(3,37),v1=1:-5.98640::1785.04410:JPL','O3v2=1:Ozone:718.04055:0.0686:::27(8,20)-28(7,21):-5.53370::1196.39880:JPL','O3v2=1:Ozone:718.25896:0.0343:::25(2,24)-24(1,23):-4.52600::959.67050:JPL','O3:Ozone:718.29621:0.007:::30(0,30)-29(1,29):-2.91610::354.14660:JPL','O3:Ozone:718.35863:0.0138:::42(5,37)-42(4,38):-3.87770::815.37100:JPL','O3:Ozone:718.42848:0.0238:::58(5,53)-57(6,52):-5.73540::1503.97430:JPL','O3v2=1:Ozone:718.47112:0.055:::42(3,39)-41(4,38):-5.67300::1474.13770:JPL','O3v1=1,v3=1:Ozone:718.91078:0.0478:::30(0,30)-29(1,29),v1=1:-5.21780::1457.89710:JPL','O3:Ozone:719.79235:0.0067:::25(2,24)-24(1,23):-3.05430::259.66970:JPL','O3:Ozone:720.40641:0.0124:::58(4,54)-58(3,55):-5.14110::1484.31020:JPL','O3v1=1,v3=1:Ozone:720.52084:0.059:::31(1,30)-30(2,29),v3=1:-5.32140::1435.64860:JPL','O3v2=1:Ozone:720.53318:0.0584:::12(6,6)-13(5,9):-5.34960::856.66210:JPL','O3v2=1:Ozone:721.72599:0.0661:::37(3,35)-37(2,36):-5.10800::1293.39940:JPL','O3v1=1,v3=1:Ozone:722.57431:0.1384:::35(2,33)-34(3,32),v3=1:-5.74380::1563.22900:JPL','O3v1=1,v3=1:Ozone:722.81798:0.1552:::38(4,35)-38(3,36),v3=1:-5.78570::1683.52470:JPL','O3:Ozone:723.00506:0.0152:::33(9,25)-34(8,26):-4.38240::699.63370:JPL','O3v1=1,v3=1:Ozone:723.31183:0.0877:::10(3,7)-9(2,8),v1=1:-5.40950::1153.40990:JPL','O3:Ozone:723.37727:0.0061:::10(3,7)-9(2,8):-3.11390::50.30210:JPL','O3v2=1:Ozone:724.42248:0.036:::37(4,34)-37(3,35):-5.03780::1317.47360:JPL','O3v1=1,v3=1:Ozone:724.58161:0.0839:::14(2,12)-13(1,13),v1=1:-5.69300::1180.32470:JPL','O3v1=1,v3=1:Ozone:724.78408:0.249:::39(5,34)-39(4,35),v3=1:-5.84470::1743.77750:JPL','O3:Ozone:726.64369:0.0112:::11(6,6)-12(5,7):-3.93440::143.76010:JPL','O3:Ozone:727.08665:0.0089:::42(3,39)-41(4,38):-4.19590::775.33800:JPL','O3v2=1:Ozone:727.69954:0.0667:::34(9,25)-35(8,28):-5.89400::1431.17530:JPL','O3:Ozone:728.24515:0.016:::40(10,30)-41(9,33):-4.85840::975.63970:JPL','O3v1=1,v3=1:Ozone:728.68213:0.1626:::36(2,34)-35(3,33),v1=1:-5.95910::1659.91400:JPL','O3v2=1:Ozone:729.84045:0.0527:::42(5,37)-42(4,38):-5.33180::1514.15350:JPL','O3v1=1,v3=1:Ozone:730.00678:0.0544:::25(2,24)-24(1,23),v1=1:-5.35420::1362.32010:JPL','O3:Ozone:730.01095:0.0118:::39(4,36)-39(3,37):-3.68060::682.79990:JPL','O3v1=1,v3=1:Ozone:730.27585:0.0539:::31(0,31)-30(1,30),v3=1:-5.09010::1413.93430:JPL','O3v2=1:Ozone:::730.58223:11.9917:10(3,7)-9(2,8):-4.57810::751.29220:JPL','O3v2=1:Ozone:732.57157:0.0685:::19(7,13)-20(6,14):-5.29750::991.18360:JPL','O3:Ozone:732.60732:0.0185:::47(11,37)-48(10,38):-5.45360::1298.57120:JPL','O3:Ozone:733.94919:0.0062:::21(4,18)-22(1,21):-5.29320::219.93830:JPL','O3:Ozone:734.38954:0.0061:::14(2,12)-13(1,13):-3.41080::77.08250:JPL','O3:Ozone:735.00339:0.0142:::18(7,11)-19(6,14):-3.82820::272.22600:JPL','O3:Ozone:736.48213:0.0112:::42(2,40)-42(1,41):-3.94460::758.07850:JPL','O3:Ozone:738.30975:0.0109:::50(4,46)-49(5,45):-4.89900::1110.13090:JPL','O3v2=2:Ozone:738.66645:0.2323:::31(1,31)-30(0,30):-5.83170::1773.47010:JPL','O3v1=1,v3=1:Ozone:739.26512:0.0711:::11(3,8)-10(2,9),v3=1:-5.26460::1099.94940:JPL','O3:Ozone:739.76168:0.0115:::50(3,47)-50(2,48):-4.44910::1092.43420:JPL','O3v2=1:Ozone:740.01181:0.0474:::14(2,12)-13(1,13):-4.87710::777.69140:JPL','O3:Ozone:740.13029:0.036:::62(6,56)-62(5,57):-5.54430::1735.27480:JPL','O3:Ozone:740.36730:0.0061:::11(3,9)-10(2,8):-3.08650::58.98100:JPL','O3v1=1,v3=1:Ozone:741.51832:0.0876:::11(3,9)-10(2,8),v1=1:-5.38090::1162.03640:JPL','O3v1=1,v3=1:Ozone:741.61361:0.317:::34(2,33)-34(1,34),v3=1:-5.83980::1516.22620:JPL','O3:Ozone:742.26604:0.016:::25(8,18)-26(7,19):-3.98880::448.02760:JPL','O3v1=1,v3=1:Ozone:742.59393:0.1669:::38(3,36)-38(2,37),v3=1:-5.84200::1658.75440:JPL','O3v2=1:Ozone:::742.81376:11.9917:31(1,31)-30(0,30):-4.38920::1077.10440:JPL','O3:Ozone:743.16441:0.0132:::40(5,35)-40(4,36):-3.73430::744.30670:JPL','O3v2=1:Ozone:743.39674:0.0382:::39(4,36)-39(3,37):-5.13320::1381.43570:JPL','O3v1=1,v3=1:Ozone:743.40499:0.0485:::28(2,27)-27(1,26),v3=1:-5.21910::1362.89890:JPL','O3v2=1:Ozone:743.41694:0.0721:::26(8,18)-27(7,21):-5.48200::1172.96260:JPL','O3v2=2:Ozone:744.42961:0.1265:::27(2,26)-26(1,25):-5.96200::1699.60890:JPL','O3v2=1:Ozone:745.63617:0.0365:::36(2,34)-35(3,33):-5.08460::1256.73060:JPL','O3v1=1,v3=1:Ozone:745.65619:0.1959:::40(4,37)-40(3,38),v3=1:-5.88140::1748.43910:JPL','O3v2=1:Ozone:745.70888:0.0606:::11(6,6)-12(5,7):-5.38730::845.78010:JPL','O3:Ozone:745.87214:0.0115:::39(3,37)-39(2,38):-3.73920::657.92030:JPL','O3v1=1,v3=1:Ozone:745.94027:0.1029:::4(4,1)-3(3,0),v3=1:-5.24740::1074.81550:JPL','O3:Ozone:746.50127:0.0154:::35(2,34)-35(1,35):-3.72380::510.20250:JPL','O3v2=1:Ozone:746.58775:0.0344:::27(2,26)-26(1,25):-4.52050::1002.40090:JPL','O3:Ozone:746.87412:0.0076:::31(1,31)-30(0,30):-2.91630::378.10630:JPL','O3v1=1,v3=1:Ozone:747.22878:0.2293:::17(7,10)-18(6,13),v3=1:-5.99100::1294.66780:JPL','O3v2=1:Ozone:747.28117:0.0503:::11(3,9)-10(2,8):-4.55110::759.94200:JPL','O3v1=1,v3=1:Ozone:747.60844:0.057:::32(1,31)-31(2,30),v1=1:-5.48680::1527.96690:JPL','O3v1=1,v3=1:Ozone:747.88674:0.1901:::37(5,32)-37(4,33),v3=1:-5.71620::1678.68300:JPL','O3v1=1,v3=1:Ozone:748.02879:0.3741:::35(1,34)-35(0,35),v3=1:-5.88790::1543.71120:JPL','O3:Ozone:748.49333:0.0166:::32(9,23)-33(8,26):-4.31410::671.06950:JPL','O3:Ozone:748.71066:0.0073:::27(2,26)-26(1,25):-3.04860::302.58360:JPL','O3v1=1,v3=1:Ozone:748.89570:0.0571:::31(1,31)-30(0,30),v1=1:-5.21640::1481.87730:JPL','O3v1=1,v3=1:Ozone:750.33080:0.071:::12(3,10)-11(2,9),v3=1:-5.24460::1109.52600:JPL','O3v2=1:Ozone:::750.35054:11.9917:32(1,31)-31(2,30):-4.63900::1124.23960:JPL','O3:Ozone:751.17612:0.0168:::36(1,35)-36(0,36):-3.77650::538.93740:JPL','O3:Ozone:751.30174:0.0083:::36(2,34)-35(3,33):-3.61040::557.54510:JPL','O3v2=1:Ozone:751.58242:0.1457:::42(2,40)-42(1,41):-5.39390::1455.73310:JPL','O3:Ozone:751.71993:0.0126:::41(4,38)-41(3,39):-3.78310::750.26330:JPL','O3:Ozone:751.90309:0.0116:::10(6,4)-11(5,7):-3.99220::133.67310:JPL','O3v2=1:Ozone:753.12899:0.0696:::33(9,25)-34(8,26):-5.82350::1401.89540:JPL','O3:Ozone:753.75458:0.017:::39(10,30)-40(9,31):-4.77350::941.22090:JPL','O3v2=2:Ozone:754.45012:0.1626:::11(3,9)-10(2,8):-5.98530::1458.31750:JPL','O3:Ozone:754.45636:0.0076:::32(1,31)-31(2,30):-3.16600::424.95170:JPL','O3v2=1:Ozone:755.22165:0.0483:::40(5,35)-40(4,36):-5.18880::1443.34730:JPL','O3v2=1:Ozone:755.50952:0.1052:::50(3,47)-50(2,48):-5.89620::1789.19600:JPL','O3v1=1,v3=1:Ozone:756.62790:0.0594:::32(1,32)-31(0,31),v3=1:-5.09490::1438.29370:JPL','O3:Ozone:757.24977:0.0065:::4(4,0)-3(3,1):-3.06510::33.22650:JPL','O3v2=2:Ozone:757.75163:0.2473:::32(0,32)-31(1,31):-5.84560::1798.10930:JPL','O3v1=1,v3=1:Ozone:757.83745:0.1127:::4(4,0)-3(3,1),v1=1:-5.36220::1136.38880:JPL','O3v2=1:Ozone:757.83752:0.0712:::18(7,11)-19(6,14):-5.27760::974.44050:JPL','O3:Ozone:758.10946:0.019:::46(11,35)-47(10,38):-5.35360::1258.31730:JPL','O3v1=1,v3=1:Ozone:758.76743:0.0535:::27(2,26)-26(1,25),v1=1:-5.34970::1405.24110:JPL','O3v2=1:Ozone:760.26916:0.0876:::39(3,37)-39(2,38):-5.19030::1356.07590:JPL','O3:Ozone:760.34211:0.0149:::17(7,11)-18(6,12):-3.81440::256.25340:JPL','O3v2=1:Ozone:761.06205:0.1926:::35(2,34)-35(1,35):-5.17510::1208.47490:JPL','O3v2=1:Ozone:::762.09941:11.9917:32(0,32)-31(1,31):-4.40330::1101.88210:JPL','O3:Ozone:765.10158:0.0125:::38(5,33)-38(4,34):-3.60130::676.79570:JPL','O3v2=1:Ozone:765.63534:0.0426:::41(4,38)-41(3,39):-5.23500::1448.59880:JPL','O3v2=1:Ozone:766.06536:0.2244:::36(1,35)-36(0,36):-5.22720::1237.05050:JPL','O3:Ozone:766.30133:0.008:::32(0,32)-31(1,31):-2.93060::403.01940:JPL','O3v1=1,v3=1:Ozone:767.09467:0.057:::32(0,32)-31(1,31),v1=1:-5.23220::1506.85780:JPL','O3v1=1,v3=1:Ozone:767.52745:0.1404:::35(5,30)-35(4,31),v3=1:-5.59910::1617.11590:JPL','O3:Ozone:767.67565:0.0171:::24(8,16)-25(7,19):-3.94500::426.17420:JPL','O3v2=1:Ozone:768.29797:0.0548:::4(4,0)-3(3,1):-4.52630::734.63130:JPL','O3:Ozone:768.41428:0.0293:::60(6,54)-60(5,55):-5.32290::1629.88660:JPL','O3v2=1:Ozone:768.76009:0.0756:::25(8,18)-26(7,19):-5.43480::1150.36390:JPL','O3v1=1,v3=1:Ozone:770.77827:0.1018:::5(4,1)-4(3,2),v3=1:-5.21560::1078.13140:JPL','O3v2=1:Ozone:770.86416:0.0627:::10(6,4)-11(5,7):-5.44600::835.73590:JPL','O3v1=1,v3=1:Ozone:771.81390:0.2449:::42(4,39)-42(3,40),v3=1:-5.98350::1816.50790:JPL','O3v1=1,v3=1:Ozone:772.33680:0.2428:::16(7,10)-17(6,11),v3=1:-5.98540::1279.74070:JPL','O3:Ozone:773.95204:0.018:::31(9,23)-32(8,24):-4.24970::643.34490:JPL','O3v2=2:Ozone:774.34723:0.1388:::29(2,28)-28(1,27):-5.96610::1745.26090:JPL','O3:Ozone:775.34529:0.0142:::35(5,31)-36(2,34):-5.47830::582.60590:JPL','O3v1=1,v3=1:Ozone:775.76900:0.0882:::12(3,9)-11(2,10),v1=1:-5.34750::1170.99290:JPL','O3v1=1,v3=1:Ozone:775.98990:0.0517:::30(2,29)-29(1,28),v3=1:-5.22590::1409.76440:JPL','O3:Ozone:776.26805:0.0063:::12(3,9)-11(2,10):-3.05270::67.90280:JPL','O3:Ozone:776.62696:0.0133:::43(4,40)-43(3,41):-3.89200::820.92770:JPL','O3v1=1,v3=1:Ozone:776.65955:0.0655:::33(1,32)-32(2,31),v3=1:-5.31580::1486.74170:JPL','O3v2=1:Ozone:777.05602:0.0356:::29(2,28)-28(1,27):-4.52450::1048.25970:JPL','O3:Ozone:777.14548:0.0119:::9(6,4)-10(5,5):-4.07910::124.42740:JPL','O3v1=1,v3=1:Ozone:777.23550:0.0624:::33(0,33)-32(1,32),v3=1:-5.10970::1463.53210:JPL','O3v2=1:Ozone:777.69715:0.0457:::38(5,33)-38(4,34):-5.05620::1376.08500:JPL','O3v2=1:Ozone:778.52605:0.0726:::32(9,23)-33(8,26):-5.75680::1373.45160:JPL','O3:Ozone:779.23615:0.0182:::38(10,28)-39(9,31):-4.69210::907.63970:JPL','O3:Ozone:779.71456:0.0081:::29(2,28)-28(1,27):-3.05250::348.64510:JPL','O3v2=2:Ozone:779.72986:0.2053:::4(4,0)-3(3,1):-5.95700::1433.46330:JPL','O3v1=1,v3=1:Ozone:780.40560:0.0633:::15(2,13)-14(1,14),v3=1:-5.60980::1129.06750:JPL','O3v1=1,v3=1:Ozone:781.60675:0.2672:::38(5,33)-38(4,34),v1=1:-5.88430::1778.13100:JPL','O3:Ozone:782.43825:0.0066:::5(4,2)-4(3,1):-3.03330::36.58780:JPL','O3v1=1,v3=1:Ozone:782.52231:0.2178:::33(5,28)-33(4,29),v3=1:-5.59100::1559.09800:JPL','O3v1=1,v3=1:Ozone:783.01097:0.1125:::5(4,2)-4(3,1),v1=1:-5.33030::1139.74610:JPL','O3v2=1:Ozone:783.07562:0.0739:::17(7,11)-18(6,12):-5.26490::958.53540:JPL','O3:Ozone:783.20356:0.016:::60(4,56)-60(3,57):-5.28820::1583.67810:JPL','O3v2=1:Ozone:783.33113:0.0507:::12(3,9)-11(2,10):-4.51760::768.81690:JPL','O3v1=1,v3=1:Ozone:783.42061:0.2162:::40(3,38)-40(2,39),v3=1:-5.92620::1722.30700:JPL','O3:Ozone:783.58613:0.0197:::45(11,35)-46(10,36):-5.25700::1218.89830:JPL','O3:Ozone:783.66059:0.0116:::36(5,31)-36(4,32):-3.47940::612.83680:JPL','O3v2=2:Ozone:784.42558:0.2958:::33(1,33)-32(0,32):-5.85360::1823.38520:JPL','O3:Ozone:784.69137:0.0064:::13(3,11)-12(2,10):-3.03760::78.53860:JPL','O3:Ozone:785.65707:0.0156:::16(7,9)-17(6,12):-3.80870::241.12220:JPL','O3:Ozone:785.68831:0.0128:::41(3,39)-41(2,40):-3.83010::724.05560:JPL','O3v1=1,v3=1:Ozone:786.66764:0.0874:::13(3,11)-12(2,10),v1=1:-5.33060::1181.53410:JPL','O3:Ozone:787.04691:0.0246:::52(12,40)-53(11,43):-5.93630::1576.92770:JPL','O3v1=1,v3=1:Ozone:788.14655:0.417:::36(2,35)-36(1,36),v3=1:-5.90380::1572.01450:JPL','O3v2=1:Ozone:::788.87088:11.9917:33(1,33)-32(0,32):-4.41170::1127.30220:JPL','O3v1=1,v3=1:Ozone:789.18525:0.0562:::29(2,28)-28(1,27),v1=1:-5.35470::1451.33010:JPL','O3v2=2:Ozone:790.65707:0.162:::12(3,9)-11(2,10):-5.95190::1467.14450:JPL','O3v2=1:Ozone:791.14122:0.0507:::43(4,40)-43(3,41):-5.34320::1518.94540:JPL','O3v2=1:Ozone:::791.24523:11.9917:13(3,11)-12(2,10):-4.50290::779.42300:JPL','O3v1=1,v3=1:Ozone:791.33102:0.07:::14(3,12)-13(2,11),v3=1:-5.20540::1130.58150:JPL','O3:Ozone:792.26388:0.0192:::37(2,36)-37(1,37):-3.79450::568.50630:JPL','O3v1=1,v3=1:Ozone:792.81897:0.0699:::13(3,10)-12(2,11),v3=1:-5.21280::1118.94960:JPL','O3:Ozone:793.05813:0.0183:::23(8,16)-24(7,17):-3.90610::405.16160:JPL','O3:Ozone:793.21306:0.0087:::33(1,33)-32(0,32):-2.93930::428.58040:JPL','O3:Ozone:793.33766:0.0061:::19(4,16)-20(1,19):-5.39630::183.43070:JPL','O3v2=1:Ozone:793.37919:0.0557:::5(4,2)-4(3,1):-4.49500::737.97840:JPL','O3v1=1,v3=1:Ozone:793.92803:0.1881:::33(1,33)-32(0,32),v1=1:-5.34550::1532.44530:JPL','O3v2=1:Ozone:794.07312:0.0792:::24(8,16)-25(7,19):-5.39220::1128.60270:JPL','O3:Ozone:794.66333:0.0133:::44(2,42)-44(1,43):-4.03040::828.99160:JPL','O3v1=1,v3=1:Ozone:795.60453:0.1003:::6(4,3)-5(3,2),v3=1:-5.18110::1082.27710:JPL','O3v2=1:Ozone:796.00184:0.0648:::9(6,4)-10(5,5):-5.53380::826.52940:JPL','O3v1=1,v3=1:Ozone:796.27546:0.0709:::31(5,26)-31(4,27),v3=1:-5.40110::1504.50100:JPL','O3v2=1:Ozone:::796.73143:11.9917:36(5,31)-36(4,32):-4.93480::1312.36560:JPL','O3v1=1,v3=1:Ozone:797.08139:0.1843:::36(5,31)-36(4,32),v1=1:-5.76540::1714.48390:JPL','O3v1=1,v3=1:Ozone:797.41541:0.2568:::15(7,8)-16(6,11),v3=1:-5.98900::1265.64430:JPL','O3v2=2:Ozone:798.05136:0.1605:::13(3,11)-12(2,10):-5.93780::1477.72060:JPL','O3v1=1,v3=1:Ozone:798.45973:0.1529:::37(2,35)-36(3,34),v3=1:-5.71900::1621.78190:JPL','O3:Ozone:798.76139:0.0107:::34(5,29)-34(4,30):-3.36920::552.41750:JPL','O3v1=1,v3=1:Ozone:799.02596:0.1336:::33(4,29),v3=1-32(0,32),v1=1:-5.90330::1532.44530:JPL','O3:Ozone:799.38331:0.0195:::30(9,21)-31(8,24):-4.18920::616.46010:JPL','O3:Ozone:799.48982:0.0238:::58(6,52)-58(5,53):-5.10720::1527.93850:JPL','O3v1=1,v3=1:Ozone:799.74687:0.4867:::37(1,36)-37(0,37),v3=1:-5.95090::1601.05380:JPL','O3v2=1:Ozone:800.93784:0.1175:::41(3,39)-41(2,40):-5.28050::1421.88240:JPL','O3:Ozone:802.36636:0.0212:::38(1,37)-38(0,38):-3.84580::598.81870:JPL','O3:Ozone:802.37380:0.0122:::8(6,2)-9(5,5):-4.21020::116.02270:JPL','O3v1=1,v3=1:Ozone:802.49355:0.0685:::34(1,34)-33(0,33),v3=1:-5.12200::1489.45790:JPL','O3:Ozone:803.39609:0.0131:::52(3,49)-52(2,50):-4.56330::1177.54370:JPL','O3v2=1:Ozone:803.89241:0.076:::31(9,23)-32(8,24):-5.69380::1345.84400:JPL','O3:Ozone:804.66156:0.014:::45(4,42)-45(3,43):-4.00750::894.77690:JPL','O3:Ozone:804.69131:0.0196:::37(10,28)-38(9,29):-4.61410::874.89620:JPL','O3v2=2:Ozone:804.70166:0.2054:::5(4,2)-4(3,1):-5.92620::1436.79590:JPL','O3v2=2:Ozone:804.88619:0.3192:::34(0,34)-33(1,33):-5.87170::1849.55080:JPL','O3v1=1,v3=1:Ozone:805.85738:0.0565:::29(5,24)-29(4,25),v3=1:-5.31890::1453.41110:JPL','O3v1=1,v3=1:Ozone:805.93146:0.0643:::34(1,33)-33(2,32),v1=1:-5.48570::1581.47410:JPL','O3v1=1,v3=1:Ozone:806.41152:0.205:::38(2,36)-37(3,35),v1=1:-5.94000::1720.85870:JPL','O3v2=2:Ozone:806.79973:0.1526:::31(2,30)-30(1,29):-5.98000::1793.99530:JPL','O3v2=1:Ozone:807.20670:0.0332:::34(1,33)-33(2,32):-4.64080::1177.47090:JPL','O3:Ozone:807.62410:0.0067:::6(4,2)-5(3,3):-2.99890::40.79000:JPL','O3v2=1:Ozone:807.75565:0.2503:::37(2,36)-37(1,37):-5.24500::1266.45560:JPL','O3v1=1,v3=1:Ozone:808.18421:0.1119:::6(4,2)-5(3,3),v1=1:-5.29570::1143.94300:JPL','O3v2=1:Ozone:808.28917:0.0765:::16(7,9)-17(6,12):-5.26030::943.46820:JPL','O3:Ozone:809.03869:0.0205:::44(11,33)-45(10,36):-5.16350::1180.31450:JPL','O3v1=1,v3=1:Ozone:809.43459:0.1262:::34(5,29)-34(4,30),v1=1:-5.65790::1654.33700:JPL','O3v2=1:Ozone:::809.48461:11.9917:34(0,34)-33(1,33):-4.43000::1153.61600:JPL','O3v2=1:Ozone:::810.02423:11.9917:31(2,30)-30(1,29):-4.53850::1097.22080:JPL','O3:Ozone:810.64613:0.0098:::32(5,27)-32(4,28):-3.27060::495.51740:JPL','O3v2=1:Ozone:810.86281:0.1937:::44(2,42)-44(1,43):-5.47910::1526.29000:JPL','O3:Ozone:810.95066:0.0163:::15(7,9)-16(6,10):-3.81220::226.83230:JPL','O3v1=1,v3=1:Ozone:810.98449:0.1214:::34(0,34),v1=1-33(4,29),v3=1:-5.93200::1559.09800:JPL','O3v1=1,v3=1:Ozone:811.21305:0.0557:::32(2,31)-31(1,30),v3=1:-5.24300::1459.68260:JPL','O3v2=1:Ozone:811.22710:0.0599:::44(3,41)-43(4,40):-5.66500::1545.33500:JPL','O3:Ozone:811.46344:0.0089:::34(1,33)-33(2,32):-3.16840::478.43900:JPL','O3v1=1,v3=1:Ozone:811.85438:0.0856:::16(2,14)-15(1,15),v1=1:-5.69590::1203.85320:JPL','O3v2=1:Ozone:::812.24569:11.9917:34(5,29)-34(4,30):-4.82490::1252.17640:JPL','O3:Ozone:812.47046:0.0245:::51(12,40)-52(11,41):-5.82830::1532.52610:JPL','O3v1=1,v3=1:Ozone:813.00590:0.0527:::27(5,22)-27(4,23),v3=1:-5.24870::1405.76090:JPL','O3:Ozone:813.14033:0.0089:::31(2,30)-30(1,29):-3.06660::397.82830:JPL','O3:Ozone:813.99105:0.0092:::34(0,34)-33(1,33):-2.95790::455.03920:JPL','O3v1=1,v3=1:Ozone:816.08242:0.2149:::34(0,34)-33(1,33),v1=1:-5.36250::1558.92790:JPL','O3v1=1,v3=1:Ozone:816.86733:0.0808:::34(5,30)-34(4,31),v3=1:-5.50670::1586.14860:JPL','O3v1=1,v3=1:Ozone:817.25572:0.0645:::32(5,28)-32(4,29),v3=1:-5.42010::1530.36830:JPL','O3v1=1,v3=1:Ozone:817.57617:0.1054:::36(5,32)-36(4,33),v3=1:-5.60120::1645.24230:JPL','O3v1=1,v3=1:Ozone:818.23657:0.0546:::25(5,20)-25(4,21),v3=1:-5.18990::1361.52580:JPL','O3v2=1:Ozone:818.24468:0.2898:::38(1,37)-38(0,38):-5.29590::1296.59930:JPL','O3v1=1,v3=1:Ozone:818.39052:0.0565:::30(5,26)-30(4,27),v3=1:-5.34180::1477.91270:JPL','O3:Ozone:818.41590:0.0195:::22(8,14)-23(7,17):-3.87240::384.99000:JPL','O3v2=1:Ozone:818.45769:0.0565:::6(4,2)-5(3,3):-4.46100::742.16270:JPL','O3v1=1,v3=1:Ozone:819.02298:0.0885:::32(5,27)-32(4,28),v1=1:-5.56160::1597.66990:JPL','O3:Ozone:819.26282:0.0103:::44(3,41)-43(4,40):-4.19040::846.83320:JPL','O3v2=1:Ozone:819.35802:0.0828:::23(8,16)-24(7,17):-5.35460::1107.67900:JPL','O3:Ozone:819.74575:0.0088:::30(5,25)-30(4,26):-3.18350::442.11250:JPL','O3v1=1,v3=1:Ozone:819.76162:0.1374:::38(5,34)-38(4,35),v3=1:-5.70320::1707.63530:JPL','O3v2=1:Ozone:819.84030:0.0643:::45(4,42)-45(3,43):-5.45780::1592.45910:JPL','O3v1=1,v3=1:Ozone:819.96512:0.0548:::28(5,24)-28(4,25),v3=1:-5.27230::1428.79050:JPL','O3v1=1,v3=1:Ozone:820.43080:0.0984:::7(4,3)-6(3,4),v3=1:-5.14690::1087.25230:JPL','O3:Ozone:820.94833:0.0324:::60(5,55)-59(6,54):-5.80380::1602.50270:JPL','O3v2=1:Ozone:821.12497:0.0668:::8(6,2)-9(5,5):-5.66570::818.16030:JPL','O3v1=1,v3=1:Ozone:821.69570:0.0628:::31(2,30)-30(1,29),v1=1:-5.36940::1500.55810:JPL','O3v1=1,v3=1:Ozone:821.72779:0.0565:::26(5,22)-26(4,23),v3=1:-5.21210::1383.00780:JPL','O3v1=1,v3=1:Ozone:822.01185:0.059:::23(5,18)-23(4,19),v3=1:-5.14260::1320.68440:JPL','O3v2=1:Ozone:822.54617:0.0408:::38(2,36)-37(3,35):-5.06750::1317.47360:JPL','O3v1=1,v3=1:Ozone:823.48634:0.0596:::24(5,20)-24(4,21),v3=1:-5.16170::1340.56800:JPL','O3v1=1,v3=1:Ozone:823.81234:0.1755:::40(5,36)-40(4,37),v3=1:-5.81230::1773.31150:JPL','O3v1=1,v3=1:Ozone:823.98147:0.0724:::35(0,35)-34(1,34),v3=1:-5.14140::1516.22620:JPL','O3v2=1:Ozone:::824.41427:11.9917:32(5,27)-32(4,28):-4.72670::1195.49690:JPL','O3v1=1,v3=1:Ozone:824.70851:0.0648:::21(5,16)-21(4,17),v3=1:-5.10710::1283.21900:JPL','O3:Ozone:824.78887:0.021:::29(9,21)-30(8,22):-4.13270::590.41500:JPL','O3v1=1,v3=1:Ozone:825.10667:0.0638:::22(5,18)-22(4,19),v3=1:-5.12200::1301.47230:JPL','O3v1=1,v3=1:Ozone:825.68795:0.1084:::18(2,16),v1=1-18(4,15),v3=1:-5.96000::1233.30890:JPL','O3:Ozone:825.71521:0.0066:::15(3,13)-14(2,12):-3.00590::101.57910:JPL','O3:Ozone:825.84466:0.0066:::16(2,14)-15(1,15):-3.41820::100.57230:JPL','O3v1=1,v3=1:Ozone:826.29264:0.0671:::30(5,25)-30(4,26),v1=1:-5.47640::1544.46100:JPL','O3v1=1,v3=1:Ozone:826.50271:0.0697:::20(5,16)-20(4,17),v3=1:-5.09390::1265.72000:JPL','O3:Ozone:826.56014:0.0079:::28(5,23)-28(4,24):-3.10770::392.17840:JPL','O3v1=1,v3=1:Ozone:826.61679:0.0721:::19(5,14)-19(4,15),v3=1:-5.08400::1249.11560:JPL','O3:Ozone:827.28887:0.0143:::43(3,41)-43(2,42):-3.92890::793.33230:JPL','O3:Ozone:827.59059:0.0125:::7(6,2)-8(5,3):-4.41780::108.45890:JPL','O3v1=1,v3=1:Ozone:827.95255:0.0814:::17(5,12)-17(4,13),v3=1:-5.07440::1218.36320:JPL','O3:Ozone:828.00629:0.0097:::38(2,36)-37(3,35):-3.59470::618.55450:JPL','O3v1=1,v3=1:Ozone:828.00874:0.0937:::18(5,14)-18(4,15),v3=1:-5.13800::1233.30890:JPL','O3v1=1,v3=1:Ozone:828.36796:0.0688:::16(3,14)-15(2,13),v3=1:-5.18240::1155.09900:JPL','O3v1=1,v3=1:Ozone:828.60720:0.0869:::16(5,12)-16(4,13),v3=1:-5.07510::1204.23600:JPL','O3v1=1,v3=1:Ozone:828.87349:0.0927:::15(5,10)-15(4,11),v3=1:-5.08010::1190.95320:JPL','O3v1=1,v3=1:Ozone:828.91995:0.0862:::15(3,13)-14(2,12),v1=1:-5.29750::1204.49420:JPL','O3v2=1:Ozone:829.23025:0.0795:::30(9,21)-31(8,24):-5.63460::1319.07260:JPL','O3v1=1,v3=1:Ozone:829.26699:0.0989:::14(5,10)-14(4,11),v3=1:-5.08970::1178.49760:JPL','O3v1=1,v3=1:Ozone:829.49443:0.1054:::13(5,8)-13(4,9),v3=1:-5.10450::1166.87860:JPL','O3v2=2:Ozone:829.67068:0.2047:::6(4,2)-5(3,3):-5.89260::1440.96200:JPL','O3v1=1,v3=1:Ozone:829.72299:0.0875:::14(3,11)-13(2,12),v1=1:-5.30280::1191.90590:JPL','O3v1=1,v3=1:Ozone:829.73480:0.1121:::12(5,8)-12(4,9),v3=1:-5.12530::1156.08990:JPL','O3v1=1,v3=1:Ozone:829.89977:0.1187:::11(5,6)-11(4,7),v3=1:-5.15340::1146.13410:JPL','O3v1=1,v3=1:Ozone:830.04547:0.1253:::10(5,6)-10(4,7),v3=1:-5.19050::1137.00920:JPL','O3v1=1,v3=1:Ozone:830.10553:0.2188:::42(5,38)-42(4,39),v3=1:-5.92810::1842.25290:JPL','O3:Ozone:830.12168:0.021:::36(10,26)-37(9,29):-4.53960::842.99070:JPL','O3v1=1,v3=1:Ozone:830.15257:0.1318:::9(5,4)-9(4,5),v3=1:-5.23940::1128.71540:JPL','O3v1=1,v3=1:Ozone:830.23778:0.1378:::8(5,4)-8(4,5),v3=1:-5.30500::1121.25200:JPL','O3v1=1,v3=1:Ozone:830.30066:0.1435:::7(5,2)-7(4,3),v3=1:-5.39680::1114.61890:JPL','O3v1=1,v3=1:Ozone:830.34725:0.1487:::6(5,2)-6(4,3),v3=1:-5.53630::1108.81560:JPL','O3v1=1,v3=1:Ozone:830.38042:0.1533:::5(5,0)-5(4,1),v3=1:-5.79530::1103.84180:JPL','O3v1=1,v3=1:Ozone:830.39664:0.0719:::35(1,34)-34(2,33),v3=1:-5.32890::1540.96370:JPL','O3v2=2:Ozone:830.40209:0.3766:::35(1,35)-34(0,34):-5.88710::1876.39890:JPL','O3:Ozone:830.55665:0.0114:::35(5,31)-35(4,32):-3.38690::580.76420:JPL','O3:Ozone:830.94178:0.0066:::14(3,11)-13(2,12):-3.00930::88.83480:JPL','O3:Ozone:831.00884:0.0104:::33(5,29)-33(4,30):-3.29430::522.59390:JPL','O3:Ozone:831.07847:0.0123:::37(5,33)-37(4,34):-3.48740::642.28740:JPL','O3:Ozone:831.46911:0.02:::56(6,50)-56(5,51):-4.89820::1429.47530:JPL','O3:Ozone:831.57389:0.0071:::26(5,21)-26(4,22):-3.04310::345.69290:JPL','O3v1=1,v3=1:Ozone:831.69684:0.0579:::28(5,23)-28(4,24),v1=1:-5.40200::1494.69020:JPL','O3v2=1:Ozone:831.84128:0.0499:::15(3,13)-14(2,12):-4.47190::802.37550:JPL','O3:Ozone:832.11791:0.0094:::31(5,27)-31(4,28):-3.21010::467.78720:JPL','O3v2=1:Ozone:832.15061:0.0496:::16(2,14)-15(1,15):-4.88460::801.05930:JPL','O3:Ozone:832.78960:0.0067:::7(4,4)-6(3,3):-2.96470::45.83360:JPL','O3:Ozone:832.92114:0.0132:::39(5,35)-39(4,36):-3.59550::707.15040:JPL','O3v1=1,v3=1:Ozone:833.12045:0.1284:::37(5,33)-37(4,34),v1=1:-5.78360::1744.27920:JPL','O3v1=1,v3=1:Ozone:833.24477:0.1005:::35(5,31)-35(4,32),v1=1:-5.68290::1682.90240:JPL','O3v1=1,v3=1:Ozone:833.34278:0.1109:::7(4,4)-6(3,3),v1=1:-5.26140::1148.98020:JPL','O3v2=1:Ozone:833.48059:0.0791:::15(7,9)-16(6,10):-5.26470::929.23880:JPL','O3:Ozone:833.60895:0.0084:::29(5,25)-29(4,26):-3.13470::416.35260:JPL','O3v2=1:Ozone:833.77591:0.0376:::30(5,25)-30(4,26):-4.64000::1142.30230:JPL','O3v1=1,v3=1:Ozone:833.98228:0.1633:::39(5,35)-39(4,36),v1=1:-5.89200::1808.99060:JPL','O3v1=1,v3=1:Ozone:834.05822:0.0795:::33(5,29)-33(4,30),v1=1:-5.59030::1624.87150:JPL','O3:Ozone:834.46833:0.0214:::43(11,33)-44(10,34):-5.07310::1142.56610:JPL','O3v2=1:Ozone:835.13710:0.0507:::35(1,35)-34(0,34):-4.44570::1180.61770:JPL','O3v1=1,v3=1:Ozone:835.20099:0.5449:::38(2,37)-38(1,38),v3=1:-5.97710::1630.89510:JPL','O3:Ozone:835.21020:0.0065:::24(5,19)-24(4,20):-2.98960::302.63690:JPL','O3:Ozone:835.25749:0.0076:::27(5,23)-27(4,24):-3.06840::368.29590:JPL','O3v1=1,v3=1:Ozone:835.29912:0.0653:::31(5,27)-31(4,28),v1=1:-5.50610::1570.19560:JPL','O3v1=1,v3=1:Ozone:835.64617:0.0563:::26(5,21)-26(4,22),v1=1:-5.33840::1448.33990:JPL','O3:Ozone:835.68395:0.0145:::47(4,44)-47(3,45):-4.12950::971.79620:JPL','O3:Ozone:836.22542:0.017:::14(7,7)-15(6,10):-3.82640::213.38360:JPL','O3:Ozone:836.44526:0.0139:::41(5,37)-41(4,38):-3.71090::775.33800:JPL','O3v1=1,v3=1:Ozone:836.74899:0.0578:::29(5,25)-29(4,26),v1=1:-5.43080::1518.88180:JPL','O3:Ozone:836.89272:0.0068:::25(5,21)-25(4,22):-3.01180::323.62070:JPL','O3:Ozone:837.81434:0.006:::22(5,17)-22(4,18):-2.94760::262.99510:JPL','O3:Ozone:837.87376:0.0245:::50(12,38)-51(11,41):-5.72340::1488.95670:JPL','O3v2=1:Ozone:837.89514:0.0506:::14(3,11)-13(2,12):-4.47460::789.65800:JPL','O3v2=2:Ozone:838.20002:0.1587:::15(3,13)-14(2,12):-5.90740::1500.58350:JPL','O3v1=1,v3=1:Ozone:838.23644:0.0555:::27(5,23)-27(4,24),v1=1:-5.36470::1470.93480:JPL','O3:Ozone:838.39610:0.0062:::23(5,19)-23(4,20):-2.96550::282.32880:JPL','O3v1=1,v3=1:Ozone:838.48624:0.0585:::24(5,19)-24(4,20),v1=1:-5.28570::1405.39550:JPL','O3:Ozone:838.60074:0.024:::39(2,38)-39(1,39):-3.87420::629.94750:JPL','O3:Ozone:838.87182:0.0131:::52(4,48)-51(5,47):-4.92940::1195.21810:JPL','O3v1=1,v3=1:Ozone:839.63714:0.0567:::25(5,21)-25(4,22),v1=1:-5.30840::1426.35780:JPL','O3:Ozone:839.65544:0.0057:::20(5,15)-20(4,16):-2.91750::226.75490:JPL','O3:Ozone:839.69619:0.0058:::21(5,17)-21(4,18):-2.93020::244.42020:JPL','O3:Ozone:839.76454:0.0099:::35(1,35)-34(0,34):-2.97390::482.19100:JPL','O3v1=1,v3=1:Ozone:840.49393:0.0633:::22(5,17)-22(4,18),v1=1:-5.24410::1365.84550:JPL','O3:Ozone:840.76067:0.0057:::19(5,15)-19(4,16):-2.90680::209.89360:JPL','O3v2=1:Ozone:840.77461:0.0354:::28(5,23)-28(4,24):-4.56450::1092.56760:JPL','O3v1=1,v3=1:Ozone:840.87038:0.0604:::23(5,19)-23(4,20),v1=1:-5.26230::1385.15240:JPL','O3:Ozone:840.93712:0.0058:::18(5,13)-18(4,14):-2.90020::193.90700:JPL','O3v1=1,v3=1:Ozone:841.38823:0.0771:::35(1,35)-34(0,34),v1=1:-5.27450::1586.14950:JPL','O3:Ozone:841.58731:0.0058:::17(5,13)-17(4,14):-2.89670::178.74690:JPL','O3:Ozone:841.81083:0.006:::16(5,11)-16(4,12):-2.89730::164.44370:JPL','O3v1=1,v3=1:Ozone:841.88385:0.0707:::20(5,15)-20(4,16),v1=1:-5.21440::1329.68090:JPL','O3v1=1,v3=1:Ozone:841.89312:0.0665:::21(5,17)-21(4,18),v1=1:-5.22720::1347.31870:JPL','O3:Ozone:842.00750:0.0145:::43(5,39)-43(4,40):-3.83310::846.83320:JPL','O3:Ozone:842.19479:0.0061:::15(5,11)-15(4,12):-2.90180::150.97720:JPL','O3:Ozone:842.38873:0.0063:::14(5,9)-14(4,10):-2.91110::138.35920:JPL','O3:Ozone:842.61442:0.0066:::13(5,9)-13(4,10):-2.92550::126.58180:JPL','O3v1=1,v3=1:Ozone:842.69261:0.0752:::19(5,15)-19(4,16),v1=1:-5.20390::1312.85600:JPL','O3:Ozone:842.75434:0.0068:::12(5,7)-12(4,8):-2.94610::115.64880:JPL','O3v1=1,v3=1:Ozone:842.81905:0.0806:::18(5,13)-18(4,14),v1=1:-5.19730::1296.89490:JPL','O3:Ozone:842.84342:0.0136:::33(5,29)-34(2,32):-5.44080::522.19910:JPL','O3:Ozone:842.88338:0.007:::11(5,7)-11(4,8):-2.97380::105.55760:JPL','O3:Ozone:842.97070:0.0072:::10(5,5)-10(4,6):-3.01060::96.30890:JPL','O3:Ozone:843.03956:0.0075:::9(5,5)-9(4,6):-3.05930::87.90200:JPL','O3:Ozone:843.08606:0.0077:::8(5,3)-8(4,4):-3.12470::80.33660:JPL','O3:Ozone:843.11798:0.0079:::7(5,3)-7(4,4):-3.21620::73.61250:JPL','O3:Ozone:843.13756:0.0081:::6(5,1)-6(4,2):-3.35550::67.72940:JPL','O3:Ozone:843.14849:0.0082:::5(5,1)-5(4,2):-3.61430::62.68720:JPL','O3v1=1,v3=1:Ozone:843.27848:0.0866:::17(5,13)-17(4,14),v1=1:-5.19390::1281.76290:JPL','O3v1=1,v3=1:Ozone:843.42235:0.0931:::16(5,11)-16(4,12),v1=1:-5.19450::1267.48200:JPL','O3v2=1:Ozone:843.42243:0.1586:::43(3,41)-43(2,42):-5.37850::1490.81190:JPL','O3v2=1:Ozone:843.51533:0.057:::7(4,4)-6(3,3):-4.42720::747.18500:JPL','O3v1=1,v3=1:Ozone:843.67536:0.1:::15(5,11)-15(4,12),v1=1:-5.19910::1254.03760:JPL','O3:Ozone:843.75095:0.0207:::21(8,14)-22(7,15):-3.84410::365.65910:JPL','O3v1=1,v3=1:Ozone:843.78641:0.1073:::14(5,9)-14(4,10),v1=1:-5.20840::1241.43820:JPL','O3v1=1,v3=1:Ozone:843.91632:0.1148:::13(5,9)-13(4,10),v1=1:-5.22290::1229.67830:JPL','O3v1=1,v3=1:Ozone:843.98181:0.1225:::12(5,7)-12(4,8),v1=1:-5.24340::1218.76050:JPL','O3v1=1,v3=1:Ozone:844.02442:0.1686:::5(5,1)-5(4,2),v1=1:-5.91170::1165.86450:JPL','O3v1=1,v3=1:Ozone:844.03776:0.1301:::11(5,7)-11(4,8),v1=1:-5.27120::1208.68290:JPL','O3v1=1,v3=1:Ozone:844.04458:0.1635:::6(5,1)-6(4,2),v1=1:-5.65290::1170.90120:JPL','O3v1=1,v3=1:Ozone:844.06183:0.1578:::7(5,3)-7(4,4),v1=1:-5.51350::1176.77750:JPL','O3v1=1,v3=1:Ozone:844.06309:0.1375:::10(5,5)-10(4,6),v1=1:-5.30800::1199.44630:JPL','O3v1=1,v3=1:Ozone:844.07289:0.1515:::8(5,3)-8(4,4),v1=1:-5.42200::1183.49370:JPL','O3v1=1,v3=1:Ozone:844.07544:0.1447:::9(5,5)-9(4,6),v1=1:-5.35660::1191.04990:JPL','O3v2=1:Ozone:844.61730:0.0863:::22(8,14)-23(7,17):-5.32200::1087.59250:JPL','O3v2=1:Ozone:844.95201:0.0399:::35(5,31)-35(4,32):-4.84200::1280.36800:JPL','O3v2=2:Ozone:845.11114:0.1587:::14(3,11)-13(2,12):-5.90950::1487.89290:JPL','O3v1=1,v3=1:Ozone:845.20227:0.0964:::8(4,5)-7(3,4),v3=1:-5.11440::1093.05910:JPL','O3v2=1:Ozone:845.39672:0.0388:::33(5,29)-33(4,30):-4.75000::1222.43700:JPL','O3v2=1:Ozone:845.48336:0.0406:::33(2,32)-32(1,31):-4.56260::1149.26860:JPL','O3v2=1:Ozone:845.51115:0.0408:::37(5,33)-37(4,34):-4.94200::1341.63770:JPL','O3v2=1:Ozone:::845.91539:11.9917:26(5,21)-26(4,22):-4.50030::1046.26990:JPL','O3v2=1:Ozone:846.23615:0.0687:::7(6,2)-8(5,3):-5.87410::810.62860:JPL','O3v2=1:Ozone:::846.51797:11.9917:31(5,27)-31(4,28):-4.66620::1167.85580:JPL','O3v2=1:Ozone:847.42979:0.0419:::39(5,35)-39(4,36):-5.04950::1406.23280:JPL','O3v2=1:Ozone:::848.04391:11.9917:29(5,25)-29(4,26):-4.59120::1116.63280:JPL','O3v1=1,v3=1:Ozone:848.51006:0.0792:::36(1,36)-35(0,35),v3=1:-5.16020::1543.71120:JPL','O3:Ozone:848.85468:0.0211:::62(4,58)-62(3,59):-5.44290::1686.06590:JPL','O3v1=1,v3=1:Ozone:848.87463:0.0602:::34(2,33)-33(1,32),v3=1:-5.27040::1512.64830:JPL','O3v1=1,v3=1:Ozone:848.88061:0.0687:::15(3,12)-14(2,13),v3=1:-5.17880::1141.22660:JPL','O3:Ozone:849.05116:0.0099:::33(2,32)-32(1,31):-3.09090::450.11770:JPL','O3:Ozone:849.44471:0.0075:::8(3,5)-8(0,8):-5.52270::30.11190:JPL','O3v2=1:Ozone:849.65796:0.0323:::24(5,19)-24(4,20):-4.44720::1003.38970:JPL','O3v2=1:Ozone:849.72223:0.0338:::27(5,23)-27(4,24):-4.52540::1068.77420:JPL','O3:Ozone:849.94254:0.0149:::45(5,41)-45(4,42):-3.96200::921.61750:JPL','O3:Ozone:850.17058:0.0227:::28(9,19)-29(8,22):-4.08020::565.20970:JPL','O3v2=1:Ozone:851.07732:0.0442:::41(5,37)-41(4,38):-5.16430::1474.13770:JPL','O3:Ozone:851.35708:0.0159:::46(2,44)-46(1,45):-4.12800::903.01720:JPL','O3v2=1:Ozone:851.39495:0.0325:::25(5,21)-25(4,22):-4.46920::1024.28370:JPL','O3v2=1:Ozone:851.58772:0.0854:::47(4,44)-47(3,45):-5.57910::1669.12480:JPL','O3v2=2:Ozone:851.78620:0.4111:::36(0,36)-35(1,35):-5.90960::1904.09820:JPL','O3v2=1:Ozone:::852.33094:11.9917:22(5,17)-22(4,18):-4.40550::963.91090:JPL','O3:Ozone:852.70743:0.0267:::40(1,39)-40(0,40):-3.92600::661.83320:JPL','O3:Ozone:852.79824:0.0127:::6(6,0)-7(5,3):-4.79720::101.73590:JPL','O3v2=1:Ozone:852.93446:0.032:::23(5,19)-23(4,20):-4.42320::983.16290:JPL','O3v1=1,v3=1:Ozone:853.41566:0.2803:::20(8,13)-21(7,14),v3=1:-5.99550::1384.25070:JPL','O3v2=1:Ozone:854.22253:0.0336:::20(5,15)-20(4,16):-4.37570::927.82090:JPL','O3v2=1:Ozone:854.26689:0.0328:::21(5,17)-21(4,18):-4.38820::945.41190:JPL','O3v2=1:Ozone:854.54135:0.0833:::29(9,21)-30(8,22):-5.57930::1293.13750:JPL','O3v2=2:Ozone:854.61827:0.2033:::7(4,4)-6(3,3):-5.85920::1445.96250:JPL','O3v2=1:Ozone:855.03475:0.3241:::39(2,38)-39(1,39):-5.32410::1327.55500:JPL','O3:Ozone:855.32663:0.0074:::10(3,7)-10(0,10):-5.09710::45.90080:JPL','O3v2=1:Ozone:855.35864:0.0347:::19(5,15)-19(4,16):-4.36510::911.02920:JPL','O3v2=2:Ozone:855.46184:0.1368:::28(5,23)-28(4,24):-5.99080::1790.37020:JPL','O3:Ozone:855.52868:0.0227:::35(10,26)-36(9,27):-4.46850::811.92330:JPL','O3v2=1:Ozone:855.53877:0.0361:::18(5,13)-18(4,14):-4.35870::895.10970:JPL','O3:Ozone:856.11388:0.0063:::17(4,14)-18(1,17):-5.54150::150.19000:JPL','O3v2=1:Ozone:::856.22525:11.9917:17(5,13)-17(4,14):-4.35530::880.01280:JPL','O3v2=1:Ozone:::856.44410:11.9917:16(5,11)-16(4,12):-4.35600::865.76970:JPL','O3v1=1,v3=1:Ozone:856.49889:0.0723:::33(2,32)-32(1,31),v1=1:-5.39420::1552.90440:JPL','O3v2=1:Ozone:856.67444:0.0536:::36(0,36)-35(1,35):-4.46860::1208.47490:JPL','O3v2=1:Ozone:856.81859:0.0493:::43(5,39)-43(4,40):-5.28580::1545.33500:JPL','O3v2=1:Ozone:856.83108:0.0414:::15(5,11)-15(4,12):-4.36070::852.35970:JPL','O3v2=1:Ozone:::857.02869:11.9917:14(5,9)-14(4,10):-4.37010::839.79470:JPL','O3v2=1:Ozone:857.26258:0.0453:::13(5,9)-13(4,10):-4.38460::828.06690:JPL','O3v2=1:Ozone:857.40663:0.0473:::12(5,7)-12(4,8):-4.40520::817.18010:JPL','O3v2=1:Ozone:857.53960:0.0493:::11(5,7)-11(4,8):-4.43310::807.13150:JPL','O3v2=1:Ozone:857.62971:0.0511:::10(5,5)-10(4,6):-4.46990::797.92190:JPL','O3v2=1:Ozone:857.70089:0.0529:::9(5,5)-9(4,6):-4.51870::789.55050:JPL','O3v2=1:Ozone:857.74908:0.0546:::8(5,3)-8(4,4):-4.58410::782.01720:JPL','O3v2=1:Ozone:857.78229:0.0561:::7(5,3)-7(4,4):-4.67570::775.32160:JPL','O3v2=1:Ozone:857.80279:0.0575:::6(5,1)-6(4,2):-4.81510::769.46350:JPL','O3v2=1:Ozone:857.81438:0.0587:::5(5,1)-5(4,2):-5.07390::764.44270:JPL','O3:Ozone:857.96788:0.0068:::8(4,4)-7(3,5):-2.93220::51.71790:JPL','O3v1=1,v3=1:Ozone:858.51183:0.1097:::8(4,4)-7(3,5),v1=1:-5.22880::1154.85680:JPL','O3v2=1:Ozone:858.65252:0.0817:::14(7,7)-15(6,10):-5.27990::915.84710:JPL','O3:Ozone:859.87632:0.0224:::42(11,31)-43(10,34):-4.98590::1105.65340:JPL','O3:Ozone:860.54628:0.0151:::47(5,43)-47(4,44):-4.09730::999.67160:JPL','O3v2=2:Ozone:860.75249:0.1392:::26(5,21)-26(4,22):-5.92690::1744.26380:JPL','O3v1=1,v3=1:Ozone:861.12470:0.0672:::18(3,16)-17(2,15),v3=1:-5.17330::1183.08750:JPL','O3:Ozone:861.45129:0.0105:::36(0,36)-35(1,35):-2.99710::510.20250:JPL','O3:Ozone:861.48361:0.0176:::13(7,7)-14(6,8):-3.85350::200.77590:JPL','O3v2=1:Ozone:861.67775:0.0374:::36(1,35)-35(2,34):-4.66080::1233.86120:JPL','O3v1=1,v3=1:Ozone:861.83016:0.0719:::36(1,35)-35(2,34),v1=1:-5.50310::1638.16340:JPL','O3v1=1,v3=1:Ozone:862.59828:0.0869:::36(0,36)-35(1,35),v1=1:-5.29820::1614.21520:JPL','O3:Ozone:862.64642:0.018:::54(6,48)-54(5,49):-4.69730::1334.53280:JPL','O3:Ozone:862.91533:0.0069:::17(3,15)-16(2,14):-2.98960::128.11960:JPL','O3:Ozone:863.25778:0.0247:::49(12,38)-50(11,39):-5.62140::1446.22000:JPL','O3v2=1:Ozone:863.29288:0.0875:::36(10,26)-37(9,29):-5.98340::1545.75290:JPL','O3v2=2:Ozone:864.58904:0.1438:::24(5,19)-24(4,20):-5.87420::1701.56260:JPL','O3v2=2:Ozone:864.67075:0.1388:::27(5,23)-27(4,24):-5.95170::1766.66750:JPL','O3v2=1:Ozone:864.99544:0.0584:::45(5,41)-45(4,42):-5.41400::1619.80600:JPL','O3:Ozone:865.51880:0.0073:::12(3,9)-12(0,12):-4.78000::64.92570:JPL','O3:Ozone:866.12615:0.0102:::36(1,35)-35(2,34):-3.18900::535.10310:JPL','O3v2=2:Ozone:866.38278:0.1418:::25(5,21)-25(4,22):-5.89590::1722.36510:JPL','O3:Ozone:866.61544:0.0157:::54(3,51)-54(2,52):-4.68780::1265.72630:JPL','O3v2=2:Ozone:867.33627:0.1499:::22(5,17)-22(4,18):-5.83280::1662.25010:JPL','O3v1=1,v3=1:Ozone:867.74946:0.0842:::17(3,15)-16(2,14),v1=1:-5.27990::1230.93380:JPL','O3v2=2:Ozone:867.96020:0.1466:::23(5,19)-23(4,20):-5.85030::1681.41870:JPL','O3v2=1:Ozone:868.52643:0.049:::17(3,15)-16(2,14):-4.45620::828.81680:JPL','O3v2=1:Ozone:868.58613:0.0573:::8(4,4)-7(3,5):-4.39510::753.04430:JPL','O3v2=1:Ozone:868.62138:0.258:::46(2,44)-46(1,45):-5.57590::1599.94070:JPL','O3:Ozone:869.06540:0.0219:::20(8,12)-21(7,15):-3.82170::347.16900:JPL','O3v2=2:Ozone:869.27861:0.1584:::20(5,15)-20(4,16):-5.80330::1626.31300:JPL','O3v2=2:Ozone:869.32662:0.1537:::21(5,17)-21(4,18):-5.81570::1643.82830:JPL','O3:Ozone:869.49468:0.015:::49(4,46)-49(3,47):-4.25820::1051.97210:JPL','O3v2=1:Ozone:869.56680:0.3728:::40(1,39)-40(0,40):-5.37530::1359.26270:JPL','O3v2=1:Ozone:869.85302:0.0899:::21(8,14)-22(7,15):-5.29480::1068.34340:JPL','O3v1=1,v3=1:Ozone:870.03233:0.0942:::9(4,5)-8(3,6),v3=1:-5.08440::1099.69420:JPL','O3:Ozone:870.37521:0.0161:::45(3,43)-45(2,44):-4.03570::865.74430:JPL','O3v2=2:Ozone:870.44708:0.1636:::19(5,15)-19(4,16):-5.79290::1609.59240:JPL','O3v1=1,v3=1:Ozone:870.57657:0.0842:::37(0,37)-36(1,36),v3=1:-5.18440::1572.01450:JPL','O3v1=1,v3=1:Ozone:870.61742:0.1642:::39(2,37)-38(3,36),v3=1:-5.71660::1683.52470:JPL','O3v2=2:Ozone:870.63117:0.1699:::18(5,13)-18(4,14):-5.78660::1593.74130:JPL','O3v2=2:Ozone:871.31851:0.1768:::17(5,13)-17(4,14):-5.78330::1578.70880:JPL','O3v2=2:Ozone:871.55389:0.1846:::16(5,11)-16(4,12):-5.78420::1564.52700:JPL','O3v2=2:Ozone:871.96000:0.193:::15(5,11)-15(4,12):-5.78900::1551.17450:JPL','O3v2=2:Ozone:872.16504:0.2019:::14(5,9)-14(4,10):-5.79850::1538.66370:JPL','O3v2=2:Ozone:872.40415:0.2111:::13(5,9)-13(4,10):-5.81310::1526.98640:JPL','O3v2=2:Ozone:872.55261:0.2205:::12(5,7)-12(4,8):-5.83390::1516.14660:JPL','O3v2=2:Ozone:872.68979:0.2299:::11(5,7)-11(4,8):-5.86180::1506.14140:JPL','O3v2=2:Ozone:872.78291:0.2392:::10(5,5)-10(4,6):-5.89880::1496.97180:JPL','O3v2=2:Ozone:872.85660:0.2481:::9(5,5)-9(4,6):-5.94760::1488.63660:JPL','O3:Ozone:874.06143:0.0152:::49(5,45)-49(4,46):-4.23880::1080.97530:JPL','O3v2=2:Ozone:874.35441:0.1582:::17(3,15)-16(2,14):-5.89220::1526.92400:JPL','O3:Ozone:875.53011:0.0244:::27(9,19)-28(8,20):-4.03180::540.84420:JPL','O3v2=1:Ozone:875.90947:0.0725:::47(5,43)-47(4,44):-5.54860::1697.53070:JPL','O3v2=2:Ozone:876.51703:0.4796:::37(1,37)-36(0,36):-5.93140::1932.51070:JPL','O3v1=1,v3=1:Ozone:878.56407:0.2952:::19(8,11)-20(7,14),v3=1:-5.97960::1366.84090:JPL','O3v2=2:Ozone:879.57943:0.2013:::8(4,4)-7(3,5):-5.82760::1451.79640:JPL','O3v1=1,v3=1:Ozone:879.80747:0.0661:::17(2,15)-16(1,16),v3=1:-5.63330::1153.74030:JPL','O3v2=1:Ozone:879.82764:0.0872:::28(9,19)-29(8,22):-5.52800::1268.03870:JPL','O3:Ozone:880.91378:0.0245:::34(10,24)-35(9,27):-4.40090::781.69400:JPL','O3v1=1,v3=1:Ozone:881.36720:0.2484:::40(2,38)-39(3,37),v1=1:-5.94130::1785.04410:JPL','O3:Ozone:881.45283:0.0072:::14(3,11)-14(0,14):-4.54610::87.14990:JPL','O3v2=1:Ozone:881.54174:0.0591:::37(1,37)-36(0,36):-4.49070::1237.05050:JPL','O3v1=1,v3=1:Ozone:882.17688:0.0781:::37(1,36)-36(2,35),v3=1:-5.35850::1598.30420:JPL','O3:Ozone:883.05621:0.0069:::9(4,6)-8(3,5):-2.90230::58.44640:JPL','O3v2=1:Ozone:883.33092:0.0437:::35(2,34)-34(1,33):-4.59710::1204.39650:JPL','O3v1=1,v3=1:Ozone:883.61055:0.1083:::9(4,6)-8(3,5),v1=1:-5.19870::1161.57580:JPL','O3v2=1:Ozone:883.80726:0.0842:::13(7,7)-14(6,8):-5.30790::903.29280:JPL','O3:Ozone:885.26380:0.0236:::41(11,31)-42(10,32):-4.90190::1069.57660:JPL','O3:Ozone:885.35139:0.0299:::41(2,40)-41(1,41):-3.96300::694.52350:JPL','O3v1=1,v3=1:Ozone:885.80963:0.0856:::16(3,13)-15(2,14),v1=1:-5.27520::1216.13890:JPL','O3v2=1:Ozone:886.17781:0.1167:::49(4,46)-49(3,47):-5.70700::1748.92910:JPL','O3:Ozone:886.45174:0.0112:::37(1,37)-36(0,36):-3.01950::538.93740:JPL','O3:Ozone:886.72744:0.0182:::12(7,5)-13(6,8):-3.89680::189.00920:JPL','O3:Ozone:887.28518:0.0109:::35(2,34)-34(1,33):-3.12560::505.50650:JPL','O3v2=1:Ozone:887.41818:0.2144:::45(3,43)-45(2,44):-5.48460::1562.85800:JPL','O3v1=1,v3=1:Ozone:887.98880:0.092:::37(1,37)-36(0,36),v1=1:-5.32010::1642.98840:JPL','O3:Ozone:888.15906:0.007:::16(3,13)-15(2,14):-2.98320::113.08700:JPL','O3v2=1:Ozone:888.60866:0.0893:::35(10,26)-36(9,27):-5.91360::1514.81680:JPL','O3:Ozone:888.62350:0.0251:::48(12,36)-49(11,39):-5.52240::1404.31610:JPL','O3v1=1,v3=1:Ozone:888.62783:0.0652:::36(2,35)-35(1,34),v3=1:-5.30830::1568.66280:JPL','O3v1=1,v3=1:Ozone:889.63318:0.0655:::20(3,18)-19(2,17),v3=1:-5.17560::1214.54170:JPL','O3v2=1:Ozone:889.80736:0.0922:::49(5,45)-49(4,46):-5.68930::1778.48880:JPL','O3:Ozone:890.66637:0.0151:::51(5,47)-51(4,48):-4.38640::1165.50860:JPL','O3:Ozone:891.65189:0.0172:::52(6,46)-52(5,47):-4.50550::1243.13520:JPL','O3v2=1:Ozone:893.56464:0.0575:::9(4,6)-8(3,5):-4.36560::759.74440:JPL','O3v1=1,v3=1:Ozone:893.56854:0.0836:::35(2,34)-34(1,33),v1=1:-5.42910::1608.35710:JPL','O3:Ozone:894.36113:0.0231:::19(8,12)-20(7,13):-3.80560::329.51970:JPL','O3v1=1,v3=1:Ozone:894.62073:0.0921:::38(1,38)-37(0,37),v3=1:-5.20920::1601.05380:JPL','O3v1=1,v3=1:Ozone:894.64551:0.0918:::10(4,7)-9(3,6),v3=1:-5.05740::1107.16700:JPL','O3v2=1:Ozone:::895.06636:11.9917:16(3,13)-15(2,14):-4.44890::813.80420:JPL','O3v2=1:Ozone:895.06734:0.0935:::20(8,12)-21(7,15):-5.27340::1049.93160:JPL','O3v2=1:Ozone:895.94201:0.045:::40(2,38)-39(3,37):-5.07180::1381.43570:JPL','O3:Ozone:896.06382:0.0073:::19(3,17)-18(2,16):-2.98650::158.16530:JPL','O3v2=2:Ozone:898.51392:0.5286:::38(0,38)-37(1,37):-5.95860::1961.74820:JPL','O3:Ozone:900.86916:0.0261:::26(9,17)-27(8,20):-3.98770::517.31850:JPL','O3v2=1:Ozone:901.07478:0.0481:::19(3,17)-18(2,16):-4.45350::858.75230:JPL','O3:Ozone:901.28516:0.0112:::40(2,38)-39(3,37):-3.60020::682.79990:JPL','O3v2=2:Ozone:902.23631:0.1554:::16(3,13)-15(2,14):-5.88420::1511.93100:JPL','O3:Ozone:902.35950:0.0333:::42(1,41)-42(0,42):-4.01640::727.97910:JPL','O3v2=1:Ozone:902.48717:0.0658:::46(3,43)-45(4,42):-5.67560::1619.80600:JPL','O3v2=1:Ozone:902.73663:0.4179:::41(2,40)-41(1,41):-5.41210::1391.77040:JPL','O3v1=1,v3=1:Ozone:903.14992:0.0992:::19(3,17)-18(2,16),v1=1:-5.33530::1260.85090:JPL','O3v2=1:Ozone:::903.67340:11.9917:38(0,38)-37(1,37):-4.51820::1266.45560:JPL','O3v1=1,v3=1:Ozone:903.68998:0.3115:::18(8,11)-19(7,12),v3=1:-5.97080::1350.26140:JPL','O3v2=2:Ozone:904.44585:0.1986:::9(4,6)-8(3,5):-5.79830::1458.46760:JPL','O3:Ozone:904.48488:0.0073:::16(3,13)-16(0,16):-4.38070::112.54240:JPL','O3v2=1:Ozone:905.09083:0.0914:::27(9,19)-28(8,20):-5.48080::1243.77620:JPL','O3v2=1:Ozone:905.64029:0.1863:::52(6,46)-52(5,47):-5.95770::1940.77490:JPL','O3:Ozone:905.84704:0.0158:::51(4,48)-51(3,49):-4.39390::1135.29280:JPL','O3:Ozone:906.27835:0.0264:::33(10,24)-34(9,25):-4.33680::752.30300:JPL','O3v2=2:Ozone:906.28376:0.1592:::19(3,17)-18(2,16):-5.89010::1556.74720:JPL','O3:Ozone:906.48952:0.0192:::48(2,46)-48(1,47):-4.23680::980.15580:JPL','O3v2=1:Ozone:906.86952:0.1179:::51(5,47)-51(4,48):-5.83610::1862.65970:JPL','O3v1=1,v3=1:Ozone:907.16517:0.1048:::18(2,16)-17(1,17),v1=1:-5.78730::1230.59110:JPL','O3:Ozone:908.25896:0.0069:::10(4,6)-9(3,7):-2.87520::66.01270:JPL','O3v1=1,v3=1:Ozone:908.43905:0.0677:::17(3,14)-16(2,15),v3=1:-5.16270::1166.76750:JPL','O3:Ozone:908.74277:0.0118:::38(0,38)-37(1,37):-3.04740::568.50630:JPL','O3v1=1,v3=1:Ozone:908.79735:0.1066:::10(4,6)-9(3,7),v1=1:-5.17140::1169.13210:JPL','O3v2=1:Ozone:908.94708:0.0866:::12(7,5)-13(6,8):-5.35190::891.57600:JPL','O3v1=1,v3=1:Ozone:909.91386:0.1002:::38(0,38)-37(1,37),v1=1:-5.34830::1672.60850:JPL','O3:Ozone:910.01158:0.0117:::46(3,43)-45(4,42):-4.20310::921.61750:JPL','O3:Ozone:910.46781:0.0151:::53(5,49)-53(4,50):-4.54010::1253.25180:JPL','O3:Ozone:910.63196:0.025:::40(11,29)-41(10,32):-4.82090::1034.33580:JPL','O3:Ozone:911.95890:0.0188:::11(7,5)-12(6,6):-3.96120::178.08340:JPL','O3v2=1:Ozone:913.90152:0.0915:::34(10,24)-35(9,27):-5.84720::1484.71530:JPL','O3:Ozone:913.97188:0.0256:::47(12,36)-48(11,37):-5.42630::1363.24550:JPL','O3v2=1:Ozone:914.17660:0.0414:::38(1,37)-37(2,36):-4.69690::1293.39940:JPL','O3v1=1,v3=1:Ozone:914.21005:0.0643:::22(3,20)-21(2,19),v3=1:-5.18720::1249.44410:JPL','O3:Ozone:914.66530:0.0186:::47(3,45)-47(2,46):-4.15070::941.28620:JPL','O3v1=1,v3=1:Ozone:915.63438:0.0804:::38(1,37)-37(2,36),v1=1:-5.53720::1698.02270:JPL','O3:Ozone:915.69468:0.0129:::31(5,27)-32(2,30):-5.42690::464.99940:JPL','O3:Ozone:915.75373:0.0282:::64(4,60)-64(3,61):-5.60600::1791.47850:JPL','O3v1=1,v3=1:Ozone:917.06331:0.0986:::39(0,39)-38(1,38),v3=1:-5.23810::1630.89510:JPL','O3:Ozone:917.55134:0.017:::50(6,44)-50(5,45):-4.32370::1155.29420:JPL','O3v2=1:Ozone:918.66076:0.0575:::10(4,6)-9(3,7):-4.33870::767.27870:JPL','O3:Ozone:918.84525:0.0116:::38(1,37)-37(2,36):-3.22570::594.93340:JPL','O3v1=1,v3=1:Ozone:919.56903:0.0894:::11(4,7)-10(3,8),v3=1:-5.03320::1115.46060:JPL','O3:Ozone:919.64002:0.0242:::18(8,10)-19(7,13):-3.79660::312.71100:JPL','O3v2=1:Ozone:920.19557:0.4779:::42(1,41)-42(0,42):-5.46500::1425.03870:JPL','O3v2=1:Ozone:920.26222:0.097:::19(8,12)-20(7,13):-5.25830::1032.35700:JPL','O3v2=1:Ozone:920.72701:0.0508:::16(3,13)-16(0,16):-5.83560::812.94820:JPL','O3:Ozone:920.96516:0.0067:::15(4,12)-16(1,15):-5.73340::120.25710:JPL','O3v2=2:Ozone:922.71403:0.6111:::39(1,39)-38(0,38):-5.98630::1991.71940:JPL','O3v2=1:Ozone:::923.21987:11.9917:37(2,36)-36(1,35):-4.64170::1262.60370:JPL','O3v2=1:Ozone:923.35697:0.1613:::51(4,48)-51(3,49):-5.84180::1831.85990:JPL','O3v2=1:Ozone:924.79093:0.3431:::48(2,46)-48(1,47):-5.68410::1676.68560:JPL','O3:Ozone:925.23044:0.0078:::21(3,19)-20(2,18):-2.99430::191.70920:JPL','O3:Ozone:926.18930:0.0278:::25(9,17)-26(8,18):-3.94800::494.63280:JPL','O3:Ozone:926.39869:0.0071:::18(2,16)-17(1,17):-3.44970::127.26390:JPL','O3:Ozone:926.70680:0.0434:::62(5,57)-61(6,56):-5.88600::1704.36320:JPL','O3v2=1:Ozone:927.20268:0.1504:::53(5,49)-53(4,50):-5.98900::1950.02340:JPL','O3:Ozone:927.53950:0.0121:::37(2,36)-36(1,35):-3.17060::563.99400:JPL','O3v2=1:Ozone:928.02783:0.0687:::39(1,39)-38(0,38):-4.54620::1296.59930:JPL','O3:Ozone:928.58033:0.0196:::56(3,53)-56(2,54):-4.82300::1356.98650:JPL','O3v1=1,v3=1:Ozone:928.79465:0.3286:::17(8,9)-18(7,12),v3=1:-5.96980::1334.51200:JPL','O3v2=2:Ozone:929.43318:0.1955:::10(4,6)-9(3,7):-5.77180::1465.96920:JPL','O3v2=1:Ozone:::929.58446:11.9917:21(3,19)-20(2,18):-4.46180::892.17450:JPL','O3v1=1,v3=1:Ozone:930.07485:0.0707:::38(2,37)-37(1,36),v3=1:-5.35650::1627.73050:JPL','O3v2=1:Ozone:930.33267:0.0956:::26(9,17)-27(8,20):-5.43780::1220.35010:JPL','O3:Ozone:931.62376:0.0284:::32(10,22)-33(9,25):-4.27620::723.75050:JPL','O3v2=1:Ozone:932.21274:0.1452:::50(6,44)-50(5,45):-5.77640::1853.25620:JPL','O3:Ozone:932.38779:0.0372:::43(2,42)-43(1,43):-4.06060::762.23120:JPL','O3v1=1,v3=1:Ozone:932.42495:0.0845:::39(1,38)-38(2,37),v3=1:-5.40260::1658.75440:JPL','O3v2=1:Ozone:932.63768:0.2889:::47(3,45)-47(2,46):-5.59870::1638.01540:JPL','O3v1=1,v3=1:Ozone:932.70713:0.0957:::37(2,36)-36(1,35),v1=1:-5.47430::1666.91090:JPL','O3:Ozone:933.13979:0.007:::11(4,8)-10(3,7):-2.85130::74.43140:JPL','O3:Ozone:933.21971:0.0126:::39(1,39)-38(0,38):-3.07570::598.81870:JPL','O3:Ozone:933.49742:0.0157:::55(5,51)-55(4,52):-4.69990::1344.18550:JPL','O3v2=1:Ozone:933.58301:0.0514:::18(2,16)-17(1,17):-4.91600::827.61130:JPL','O3v1=1,v3=1:Ozone:933.73123:0.1047:::11(4,8)-10(3,7),v1=1:-5.14730::1177.53700:JPL','O3v2=2:Ozone:934.06849:0.1617:::21(3,19)-20(2,18):-5.89880::1590.04580:JPL','O3v2=1:Ozone:934.07405:0.0889:::11(7,5)-12(6,6):-5.41710::880.69650:JPL','O3v1=1,v3=1:Ozone:934.14072:0.0787:::21(3,19)-20(2,18),v1=1:-5.28380::1294.27420:JPL','O3v1=1,v3=1:Ozone:934.66289:0.1075:::39(1,39)-38(0,38),v1=1:-5.37640::1702.96000:JPL','O3v1=1,v3=1:Ozone:935.36690:0.0651:::24(3,22)-23(2,21),v3=1:-5.20630::1287.76800:JPL','O3:Ozone:935.81364:0.0076:::18(3,15)-18(0,18):-4.27360::141.08200:JPL','O3:Ozone:935.98190:0.0266:::39(11,29)-40(10,30):-4.74310::999.93140:JPL','O3:Ozone:937.17990:0.0193:::10(7,3)-11(6,6):-4.05490::167.99830:JPL','O3v1=1,v3=1:Ozone:938.74151:0.1727:::41(2,39)-40(3,38),v3=1:-5.73530::1748.43910:JPL','O3v2=1:Ozone:939.17290:0.0941:::33(10,24)-34(9,25):-5.78430::1455.44870:JPL','O3:Ozone:939.30384:0.0262:::46(12,34)-47(11,37):-5.33310::1323.00840:JPL','O3:Ozone:939.86971:0.0169:::48(6,42)-48(5,43):-4.15240::1071.00860:JPL','O3:Ozone:940.14298:0.0158:::54(4,50)-53(5,49):-4.97510::1283.62170:JPL','O3v1=1,v3=1:Ozone:940.78597:0.1081:::40(1,40)-39(0,39),v3=1:-5.26840::1661.48510:JPL','O3v2=1:Ozone:::943.41389:11.9917:11(4,8)-10(3,7):-4.31510::775.66210:JPL','O3v1=1,v3=1:Ozone:943.78481:0.087:::12(4,9)-11(3,8),v3=1:-5.01260::1124.60860:JPL','O3:Ozone:944.46141:0.0172:::53(4,50)-53(3,51):-4.53650::1221.74790:JPL','O3v1=1,v3=1:Ozone:944.81066:0.0826:::18(3,15)-17(2,16),v1=1:-5.26440::1243.68070:JPL','O3:Ozone:944.90382:0.0254:::17(8,10)-18(7,11):-3.79550::296.74300:JPL','O3v2=1:Ozone:945.43957:0.1004:::18(8,10)-19(7,13):-5.25010::1015.61950:JPL','O3:Ozone:948.85044:0.0074:::18(3,15)-17(2,16):-2.97460::140.64710:JPL']
